Package: torchvision
Title: Models, Datasets and Transformations for Images
Version: 0.9.0
Authors@R: c(
    person("Tomasz", "Kalinowski", role = c("ctb", "cre"),
           email = "tomasz@posit.co"),
    person("Daniel", "Falbel", role = c("aut", "cph"),
           email = "dfalbel@gmail.com"),
    person("Christophe", "Regouby", role = c("ctb"),
           email = "christophe.regouby@free.fr"),
    person("Akanksha", "Koshti", role = c("ctb")),
    person("Derrick", "Richard", role = c("ctb"),
           email = "derrickrds@gmail.com"),
    person("ANAMASGARD", role = c("ctb")),
    person("Chandraveer", "Singh", role = c("ctb"),
           email = "chandraveersingh1717@gmail.com"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
    )
Description: Provides access to datasets, models and preprocessing
    facilities for deep learning with images. Integrates seamlessly
    with the 'torch' package and its API borrows heavily from
    the 'PyTorch' vision package.
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://torchvision.mlverse.org,
        https://github.com/mlverse/torchvision
RoxygenNote: 7.3.3
Imports: torch (>= 0.5.0), fs, rlang, rappdirs, utils, jpeg, tiff,
        magrittr, png, abind, jsonlite, withr, cli, glue, zeallot
Suggests: arrow, magick, prettyunits, testthat, coro, R.matlab, xml2,
        knitr, rmarkdown, torchvisionlib
BugReports: https://github.com/mlverse/torchvision/issues
Collate: 'collection-catalog.R' 'folder-dataset.R'
        'collection-rf100-doc.R' 'collection-rf100-biology.R'
        'collection-rf100-damage.R' 'collection-rf100-infrared.R'
        'collection-rf100-medical.R' 'collection-rf100-underwater.R'
        'conditions.R' 'dataset-caltech.R' 'dataset-cifar.R'
        'dataset-coco.R' 'dataset-eurosat.R' 'dataset-fer.R'
        'dataset-fgvc.R' 'dataset-flickr.R' 'dataset-flowers.R'
        'dataset-imagenet.R' 'dataset-lfw.R' 'dataset-mnist.R'
        'dataset-oxfordiiitpet.R' 'dataset-pascal.R'
        'dataset-places365.R' 'dataset-plankton.R'
        'dataset-rf100-peixos.R' 'dataset-vggface2.R' 'extension.R'
        'globals.R' 'models-alexnet.R' 'models-convnext.R'
        'models-convnext_detection.R' 'models-convnext_segmentation.R'
        'models-deeplabv3.R' 'models-efficientnet.R'
        'models-efficientnetv2.R' 'models-facenet.R'
        'models-faster_rcnn.R' 'models-fcn.R' 'models-inception.R'
        'models-mask_rcnn.R' 'models-maxvit.R' 'models-mobilenetv2.R'
        'models-mobilenetv3.R' 'models-mobilenetv3_large.R'
        'models-resnet.R' 'models-vgg.R' 'models-vit.R'
        'ops-box_convert.R' 'ops-boxes.R' 'transforms-array.R'
        'transforms-defaults.R' 'transforms-generics.R'
        'transforms-magick.R' 'transforms-segmentation.R'
        'transforms-tensor.R' 'utils.R' 'vision_utils.R'
Depends: R (>= 3.5)
LazyData: true
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-22 13:10:42 UTC; dfalbel
Author: Tomasz Kalinowski [ctb, cre],
  Daniel Falbel [aut, cph],
  Christophe Regouby [ctb],
  Akanksha Koshti [ctb],
  Derrick Richard [ctb],
  ANAMASGARD [ctb],
  Chandraveer Singh [ctb],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Tomasz Kalinowski <tomasz@posit.co>
Repository: CRAN
Date/Publication: 2026-04-22 14:20:02 UTC
Built: R 4.6.0; ; 2026-04-26 03:05:01 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(draw_bounding_boxes,default)
S3method(draw_bounding_boxes,image_with_bounding_box)
S3method(draw_bounding_boxes,torch_tensor)
S3method(draw_segmentation_masks,default)
S3method(draw_segmentation_masks,image_with_segmentation_mask)
S3method(draw_segmentation_masks,torch_tensor)
S3method(get_image_size,"magick-image")
S3method(transform_adjust_brightness,default)
S3method(transform_adjust_brightness,torch_tensor)
S3method(transform_adjust_contrast,default)
S3method(transform_adjust_contrast,torch_tensor)
S3method(transform_adjust_gamma,default)
S3method(transform_adjust_gamma,torch_tensor)
S3method(transform_adjust_hue,default)
S3method(transform_adjust_hue,torch_tensor)
S3method(transform_adjust_saturation,default)
S3method(transform_adjust_saturation,torch_tensor)
S3method(transform_affine,default)
S3method(transform_affine,torch_tensor)
S3method(transform_center_crop,default)
S3method(transform_color_jitter,default)
S3method(transform_convert_image_dtype,default)
S3method(transform_convert_image_dtype,torch_tensor)
S3method(transform_crop,"magick-image")
S3method(transform_crop,default)
S3method(transform_crop,torch_tensor)
S3method(transform_five_crop,default)
S3method(transform_five_crop,torch_tensor)
S3method(transform_grayscale,default)
S3method(transform_grayscale,torch_tensor)
S3method(transform_hflip,"magick-image")
S3method(transform_hflip,default)
S3method(transform_hflip,torch_tensor)
S3method(transform_linear_transformation,default)
S3method(transform_linear_transformation,torch_tensor)
S3method(transform_normalize,default)
S3method(transform_normalize,torch_tensor)
S3method(transform_pad,default)
S3method(transform_pad,torch_tensor)
S3method(transform_perspective,default)
S3method(transform_perspective,torch_tensor)
S3method(transform_random_affine,default)
S3method(transform_random_apply,default)
S3method(transform_random_choice,default)
S3method(transform_random_crop,default)
S3method(transform_random_erasing,default)
S3method(transform_random_grayscale,default)
S3method(transform_random_grayscale,torch_tensor)
S3method(transform_random_horizontal_flip,default)
S3method(transform_random_order,default)
S3method(transform_random_perspective,default)
S3method(transform_random_resized_crop,default)
S3method(transform_random_rotation,default)
S3method(transform_random_vertical_flip,default)
S3method(transform_resize,"magick-image")
S3method(transform_resize,default)
S3method(transform_resize,torch_tensor)
S3method(transform_resized_crop,default)
S3method(transform_rgb_to_grayscale,torch_tensor)
S3method(transform_rotate,default)
S3method(transform_rotate,torch_tensor)
S3method(transform_ten_crop,default)
S3method(transform_ten_crop,torch_tensor)
S3method(transform_to_tensor,"magick-image")
S3method(transform_to_tensor,array)
S3method(transform_to_tensor,default)
S3method(transform_to_tensor,list)
S3method(transform_to_tensor,matrix)
S3method(transform_vflip,default)
S3method(transform_vflip,torch_tensor)
export(base_loader)
export(batched_nms)
export(box_area)
export(box_convert)
export(box_iou)
export(caltech101_dataset)
export(caltech256_dataset)
export(caltech_classes)
export(cifar100_dataset)
export(cifar10_dataset)
export(clip_boxes_to_image)
export(coco_caption_dataset)
export(coco_classes)
export(coco_detection_dataset)
export(coco_segmentation_dataset)
export(draw_bounding_boxes)
export(draw_keypoints)
export(draw_segmentation_masks)
export(emnist_collection)
export(emnist_dataset)
export(eurosat100_dataset)
export(eurosat_all_bands_dataset)
export(eurosat_dataset)
export(fashion_mnist_dataset)
export(fer_dataset)
export(fgvc_aircraft_dataset)
export(flickr30k_caption_dataset)
export(flickr8k_caption_dataset)
export(flowers102_dataset)
export(generalized_box_iou)
export(get_collection_catalog)
export(image_folder_dataset)
export(imagenet_1k_classes)
export(imagenet_21k_classes)
export(imagenet_21k_df)
export(imagenet_classes)
export(kmnist_dataset)
export(lfw_pairs_dataset)
export(lfw_people_dataset)
export(list_collection_datasets)
export(magick_loader)
export(mnist_dataset)
export(model_alexnet)
export(model_convnext_base_1k)
export(model_convnext_base_22k)
export(model_convnext_base_detection)
export(model_convnext_base_fcn)
export(model_convnext_base_upernet)
export(model_convnext_large_1k)
export(model_convnext_large_22k)
export(model_convnext_small_22k)
export(model_convnext_small_22k1k)
export(model_convnext_small_detection)
export(model_convnext_small_fcn)
export(model_convnext_small_upernet)
export(model_convnext_tiny_1k)
export(model_convnext_tiny_22k)
export(model_convnext_tiny_detection)
export(model_convnext_tiny_fcn)
export(model_convnext_tiny_upernet)
export(model_deeplabv3_resnet101)
export(model_deeplabv3_resnet50)
export(model_efficientnet_b0)
export(model_efficientnet_b1)
export(model_efficientnet_b2)
export(model_efficientnet_b3)
export(model_efficientnet_b4)
export(model_efficientnet_b5)
export(model_efficientnet_b6)
export(model_efficientnet_b7)
export(model_efficientnet_v2_l)
export(model_efficientnet_v2_m)
export(model_efficientnet_v2_s)
export(model_facenet_inception_resnet_v1)
export(model_facenet_onet)
export(model_facenet_pnet)
export(model_facenet_rnet)
export(model_fasterrcnn_mobilenet_v3_large_320_fpn)
export(model_fasterrcnn_mobilenet_v3_large_fpn)
export(model_fasterrcnn_resnet50_fpn)
export(model_fasterrcnn_resnet50_fpn_v2)
export(model_fcn_resnet101)
export(model_fcn_resnet50)
export(model_inception_v3)
export(model_maskrcnn_resnet50_fpn)
export(model_maskrcnn_resnet50_fpn_v2)
export(model_maxvit)
export(model_mobilenet_v2)
export(model_mobilenet_v3_large)
export(model_mobilenet_v3_large_quantized)
export(model_mobilenet_v3_small)
export(model_mtcnn)
export(model_resnet101)
export(model_resnet152)
export(model_resnet18)
export(model_resnet34)
export(model_resnet50)
export(model_resnext101_32x8d)
export(model_resnext50_32x4d)
export(model_vgg11)
export(model_vgg11_bn)
export(model_vgg13)
export(model_vgg13_bn)
export(model_vgg16)
export(model_vgg16_bn)
export(model_vgg19)
export(model_vgg19_bn)
export(model_vit_b_16)
export(model_vit_b_32)
export(model_vit_h_14)
export(model_vit_l_16)
export(model_vit_l_32)
export(model_wide_resnet101_2)
export(model_wide_resnet50_2)
export(nms)
export(oxfordiiitpet_binary_dataset)
export(oxfordiiitpet_dataset)
export(oxfordiiitpet_segmentation_dataset)
export(pascal_detection_dataset)
export(pascal_segmentation_dataset)
export(pascal_voc_classes)
export(places365_dataset)
export(places365_dataset_large)
export(qmnist_dataset)
export(remove_small_boxes)
export(rf100_biology_collection)
export(rf100_damage_collection)
export(rf100_document_collection)
export(rf100_infrared_collection)
export(rf100_medical_collection)
export(rf100_peixos_segmentation_dataset)
export(rf100_underwater_collection)
export(search_collection)
export(target_transform_coco_masks)
export(target_transform_trimap_masks)
export(tensor_image_browse)
export(tensor_image_display)
export(tiny_imagenet_dataset)
export(transform_adjust_brightness)
export(transform_adjust_contrast)
export(transform_adjust_gamma)
export(transform_adjust_hue)
export(transform_adjust_saturation)
export(transform_affine)
export(transform_center_crop)
export(transform_color_jitter)
export(transform_convert_image_dtype)
export(transform_crop)
export(transform_five_crop)
export(transform_grayscale)
export(transform_hflip)
export(transform_linear_transformation)
export(transform_normalize)
export(transform_pad)
export(transform_perspective)
export(transform_random_affine)
export(transform_random_apply)
export(transform_random_choice)
export(transform_random_crop)
export(transform_random_erasing)
export(transform_random_grayscale)
export(transform_random_horizontal_flip)
export(transform_random_order)
export(transform_random_perspective)
export(transform_random_resized_crop)
export(transform_random_rotation)
export(transform_random_vertical_flip)
export(transform_resize)
export(transform_resized_crop)
export(transform_rgb_to_grayscale)
export(transform_rotate)
export(transform_ten_crop)
export(transform_to_tensor)
export(transform_vflip)
export(vggface2_dataset)
export(vision_make_grid)
export(whoi_plankton_dataset)
export(whoi_small_coralnet_dataset)
export(whoi_small_plankton_dataset)
importFrom(grDevices,dev.off)
importFrom(graphics,polygon)
importFrom(jsonlite,fromJSON)
importFrom(magrittr,"%>%")
importFrom(stats,setNames)
importFrom(torch,as_array)
importFrom(torch,dataset)
importFrom(torch,load_state_dict)
importFrom(torch,nn_adaptive_avg_pool2d)
importFrom(torch,nn_batch_norm1d)
importFrom(torch,nn_batch_norm2d)
importFrom(torch,nn_conv2d)
importFrom(torch,nn_conv_transpose2d)
importFrom(torch,nn_dropout)
importFrom(torch,nn_gelu)
importFrom(torch,nn_hardsigmoid)
importFrom(torch,nn_hardswish)
importFrom(torch,nn_identity)
importFrom(torch,nn_layer_norm)
importFrom(torch,nn_linear)
importFrom(torch,nn_max_pool2d)
importFrom(torch,nn_module)
importFrom(torch,nn_module_dict)
importFrom(torch,nn_module_list)
importFrom(torch,nn_parameter)
importFrom(torch,nn_prelu)
importFrom(torch,nn_relu)
importFrom(torch,nn_sequential)
importFrom(torch,nn_softmax)
importFrom(torch,nnf_gelu)
importFrom(torch,nnf_grid_sample)
importFrom(torch,nnf_interpolate)
importFrom(torch,nnf_layer_norm)
importFrom(torch,nnf_normalize)
importFrom(torch,nnf_relu)
importFrom(torch,nnf_softmax)
importFrom(torch,torch_arange)
importFrom(torch,torch_cat)
importFrom(torch,torch_chunk)
importFrom(torch,torch_clamp)
importFrom(torch,torch_empty)
importFrom(torch,torch_flatten)
importFrom(torch,torch_float)
importFrom(torch,torch_float32)
importFrom(torch,torch_int32)
importFrom(torch,torch_linspace)
importFrom(torch,torch_log2)
importFrom(torch,torch_long)
importFrom(torch,torch_matmul)
importFrom(torch,torch_max)
importFrom(torch,torch_meshgrid)
importFrom(torch,torch_nonzero)
importFrom(torch,torch_ones)
importFrom(torch,torch_randn)
importFrom(torch,torch_sigmoid)
importFrom(torch,torch_stack)
importFrom(torch,torch_tensor)
importFrom(torch,torch_zeros)
importFrom(torch,torch_zeros_like)
importFrom(utils,read.csv)
importFrom(utils,read.delim)
importFrom(utils,tail)
importFrom(zeallot,"%<-%")

# torchvision <img src='https://torch.mlverse.org/css/images/hex/torchvision.png' align="right" style="width: 15%"/>

<!-- badges: start -->
[![R build status](https://github.com/mlverse/torchvision/workflows/R-CMD-check/badge.svg)](https://github.com/mlverse/torchvision/actions)
[![CRAN status](https://www.r-pkg.org/badges/version/torchvision)](https://CRAN.R-project.org/package=torchvision)
[![](https://cranlogs.r-pkg.org/badges/torchvision)](https://cran.r-project.org/package=torchvision)
<!-- badges: end -->

torchvision is an extension for [torch](https://github.com/mlverse/torch) providing image loading, transformations, common architectures for computer vision, pre-trained weights and access to commonly used datasets. 

## Installation

The CRAN release can be installed with:

```r
install.packages("torchvision")
```

You can install the development version from GitHub with:

``` r
remotes::install_github("mlverse/torchvision@main")
```

# torchvision

[![total torchvision downloads](https://pepy.tech/badge/torchvision)](https://pepy.tech/project/torchvision)
[![documentation](https://img.shields.io/badge/dynamic/json.svg?label=docs&url=https%3A%2F%2Fpypi.org%2Fpypi%2Ftorchvision%2Fjson&query=%24.info.version&colorB=brightgreen&prefix=v)](https://pytorch.org/vision/stable/index.html)

The torchvision package consists of popular datasets, model architectures, and common image transformations for computer
vision.

## Installation

Please refer to the [official
instructions](https://pytorch.org/get-started/locally/) to install the stable
versions of `torch` and `torchvision` on your system.

To build source, refer to our [contributing
page](https://github.com/pytorch/vision/blob/main/CONTRIBUTING.md#development-installation).

The following is the corresponding `torchvision` versions and supported Python
versions.

| `torch`            | `torchvision`      | Python              |
| ------------------ | ------------------ | ------------------- |
| `main` / `nightly` | `main` / `nightly` | `>=3.9`, `<=3.12`   |
| `2.4`              | `0.19`             | `>=3.8`, `<=3.12`   |
| `2.3`              | `0.18`             | `>=3.8`, `<=3.12`   |
| `2.2`              | `0.17`             | `>=3.8`, `<=3.11`   |
| `2.1`              | `0.16`             | `>=3.8`, `<=3.11`   |
| `2.0`              | `0.15`             | `>=3.8`, `<=3.11`   |

<details>
    <summary>older versions</summary>

| `torch` | `torchvision`     | Python                    |
|---------|-------------------|---------------------------|
| `1.13`  | `0.14`            | `>=3.7.2`, `<=3.10`       |
| `1.12`  | `0.13`            | `>=3.7`, `<=3.10`         |
| `1.11`  | `0.12`            | `>=3.7`, `<=3.10`         |
| `1.10`  | `0.11`            | `>=3.6`, `<=3.9`          |
| `1.9`   | `0.10`            | `>=3.6`, `<=3.9`          |
| `1.8`   | `0.9`             | `>=3.6`, `<=3.9`          |
| `1.7`   | `0.8`             | `>=3.6`, `<=3.9`          |
| `1.6`   | `0.7`             | `>=3.6`, `<=3.8`          |
| `1.5`   | `0.6`             | `>=3.5`, `<=3.8`          |
| `1.4`   | `0.5`             | `==2.7`, `>=3.5`, `<=3.8` |
| `1.3`   | `0.4.2` / `0.4.3` | `==2.7`, `>=3.5`, `<=3.7` |
| `1.2`   | `0.4.1`           | `==2.7`, `>=3.5`, `<=3.7` |
| `1.1`   | `0.3`             | `==2.7`, `>=3.5`, `<=3.7` |
| `<=1.0` | `0.2`             | `==2.7`, `>=3.5`, `<=3.7` |

</details>

## Image Backends

Torchvision currently supports the following image backends:

- torch tensors
- PIL images:
    - [Pillow](https://python-pillow.org/)
    - [Pillow-SIMD](https://github.com/uploadcare/pillow-simd) - a **much faster** drop-in replacement for Pillow with SIMD.

Read more in in our [docs](https://pytorch.org/vision/stable/transforms.html).

## [UNSTABLE] Video Backend

Torchvision currently supports the following video backends:

- [pyav](https://github.com/PyAV-Org/PyAV) (default) - Pythonic binding for ffmpeg libraries.
- video_reader - This needs ffmpeg to be installed and torchvision to be built from source. There shouldn't be any
  conflicting version of ffmpeg installed. Currently, this is only supported on Linux.

```
conda install -c conda-forge 'ffmpeg<4.3'
python setup.py install
```

# Using the models on C++

Refer to [example/cpp](https://github.com/pytorch/vision/tree/main/examples/cpp).

**DISCLAIMER**: the `libtorchvision` library includes the torchvision
custom ops as well as most of the C++ torchvision APIs. Those APIs do not come
with any backward-compatibility guarantees and may change from one version to
the next. Only the Python APIs are stable and with backward-compatibility
guarantees. So, if you need stability within a C++ environment, your best bet is
to export the Python APIs via torchscript.

## Documentation

You can find the API documentation on the pytorch website: <https://pytorch.org/vision/stable/index.html>

## Contributing

See the [CONTRIBUTING](CONTRIBUTING.md) file for how to help out.

## Disclaimer on Datasets

This is a utility library that downloads and prepares public datasets. We do not host or distribute these datasets,
vouch for their quality or fairness, or claim that you have license to use the dataset. It is your responsibility to
determine whether you have permission to use the dataset under the dataset's license.

If you're a dataset owner and wish to update any part of it (description, citation, etc.), or do not want your dataset
to be included in this library, please get in touch through a GitHub issue. Thanks for your contribution to the ML
community!

## Pre-trained Model License

The pre-trained models provided in this library may have their own licenses or terms and conditions derived from the
dataset used for training. It is your responsibility to determine whether you have permission to use the models for your
use case.

More specifically, SWAG models are released under the CC-BY-NC 4.0 license. See
[SWAG LICENSE](https://github.com/facebookresearch/SWAG/blob/main/LICENSE) for additional details.

## Citing TorchVision

If you find TorchVision useful in your work, please consider citing the following BibTeX entry:

```bibtex
@software{torchvision2016,
    title        = {TorchVision: PyTorch's Computer Vision library},
    author       = {TorchVision maintainers and contributors},
    year         = 2016,
    journal      = {GitHub repository},
    publisher    = {GitHub},
    howpublished = {\url{https://github.com/pytorch/vision}}
}
```

# Video Classification

We present a simple training script that can be used for replicating the result of [resenet-based video models](https://research.fb.com/wp-content/uploads/2018/04/a-closer-look-at-spatiotemporal-convolutions-for-action-recognition.pdf). All models are trained on [Kinetics400 dataset](https://deepmind.com/research/open-source/kinetics), a benchmark dataset for human-action recognition. The accuracy is reported on the traditional validation split.

## Data preparation

If you already have downloaded [Kinetics400 dataset](https://deepmind.com/research/open-source/kinetics), 
please proceed directly to the next section.

To download videos, one can use https://github.com/Showmax/kinetics-downloader. Please note that the dataset can take up upwards of 400GB, depending on the quality setting during download.

## Training

We assume the training and validation AVI videos are stored at `/data/kinectics400/train` and 
`/data/kinectics400/val`. For training we suggest starting with the hyperparameters reported in the [paper](https://research.fb.com/wp-content/uploads/2018/04/a-closer-look-at-spatiotemporal-convolutions-for-action-recognition.pdf), in order to match the performance of said models. Clip sampling strategy is a particularly important parameter during training, and we suggest using random temporal jittering during training - in other words sampling multiple training clips from each video with random start times during at every epoch. This functionality is built into our training script, and optimal hyperparameters are set by default.  

### Multiple GPUs

Run the training on a single node with 8 GPUs:
```bash
torchrun --nproc_per_node=8 train.py --data-path=/data/kinectics400 --kinetics-version="400" --lr 0.08 --cache-dataset --sync-bn --amp 
```

**Note:** all our models were trained on 8 nodes with 8 V100 GPUs each for a total of 64 GPUs. Expected training time for 64 GPUs is 24 hours, depending on the storage solution.
**Note 2:** hyperparameters for exact replication of our training can be found on the section below. Some hyperparameters such as learning rate must be scaled linearly in proportion to the number of GPUs. The default values assume 64 GPUs.

### Single GPU 

**Note:** training on a single gpu can be extremely slow. 


```bash
python train.py --data-path=/data/kinectics400 --kinetics-version="400" --batch-size=8 --cache-dataset
```


### Additional Kinetics versions

Since the original release, additional versions of Kinetics dataset became available (Kinetics 600).
Our training scripts support these versions of dataset as well by setting the `--kinetics-version` parameter to `"600"`.

**Note:** training on Kinetics 600 requires a different set of hyperparameters for optimal performance. We do not provide Kinetics 600 pretrained models.


## Video classification models

Starting with version `0.4.0` we have introduced support for basic video tasks and video classification modelling.
For more information about the available models check [here](https://pytorch.org/docs/stable/torchvision/models.html#video-classification). 

### Video ResNet models

See reference training script [here](https://github.com/pytorch/vision/blob/main/references/video_classification/train.py):

- input space: RGB
- resize size: [128, 171]
- crop size: [112, 112]
- mean: [0.43216, 0.394666, 0.37645]
- std: [0.22803, 0.22145, 0.216989]
- number of classes: 400

Input data augmentations at training time (with optional parameters):

1. ConvertImageDtype
2. Resize (resize size value above)
3. Random horizontal flip (0.5)
4. Normalization (mean, std, see values above)
5. Random Crop (crop size value above)
6. Convert BCHW to CBHW

Input data augmentations at validation time (with optional parameters):

1. ConvertImageDtype
2. Resize (resize size value above)
3. Normalization (mean, std, see values above)
4. Center Crop (crop size value above)
5. Convert BCHW to CBHW

This translates in the following set of command-line arguments. Please note that `--batch-size` parameter controls the
batch size per GPU. Moreover, note that our default `--lr` is configured for 64 GPUs which is how many we used for the 
Video resnet models:
```
# number of frames per clip
--clip_len 16 \ 
--frame-rate 15 \
# allow for temporal jittering
--clips_per_video 5 \
--batch-size 24 \
--epochs 45 \
--lr 0.64 \
# we use 10 epochs for linear warmup
--lr-warmup-epochs 10 \
# learning rate is decayed at 20, 30, and 40 epoch by a factor of 10
--lr-milestones 20, 30, 40 \
--lr-gamma 0.1 \
--train-resize-size 128 171 \
--train-crop-size 112 112 \
--val-resize-size 128 171 \
--val-crop-size 112 112
```

### S3D

The S3D model was trained similarly to the above but with the following changes on the default configuration:
```
--batch-size=12 --lr 0.2 --clip-len 64 --clips-per-video 5 --sync-bn \
--train-resize-size 256 256 --train-crop-size 224 224 --val-resize-size 256 256 --val-crop-size 224 224
```

We used 64 GPUs to train the architecture. 

To estimate the validation statistics of the model, we run the reference script with the following configuration:
```
--batch-size=16 --test-only --clip-len 128 --clips-per-video 1 
```

### Additional video modelling resources

- [Video Model Zoo](https://github.com/facebookresearch/VMZ)
- [PySlowFast](https://github.com/facebookresearch/SlowFast)

### References

[0] _D. Tran, H. Wang, L. Torresani, J. Ray, Y. LeCun and M. Paluri_: A Closer Look at Spatiotemporal Convolutions for Action Recognition. _CVPR 2018_ ([paper](https://research.fb.com/wp-content/uploads/2018/04/a-closer-look-at-spatiotemporal-convolutions-for-action-recognition.pdf))

[1] _W. Kay, J. Carreira, K. Simonyan, B. Zhang, C. Hillier, S. Vijayanarasimhan, F. Viola, T. Green, T. Back, P. Natsev, M. Suleyman, A. Zisserman_: The Kinetics Human Action Video Dataset ([paper](https://arxiv.org/abs/1705.06950))

# Similarity Learning Using Triplet Loss #

In this reference, we use triplet loss to learn embeddings which can be used to differentiate images. This learning technique was popularized by [FaceNet: A Unified Embedding for Face Recognition and Clustering](https://arxiv.org/abs/1503.03832) and has been quite effective in learning embeddings to differentiate between faces.

This reference can be directly applied to the following use cases:

* You have an unknown number of classes and would like to train a model to learn how to differentiate between them.
* You want to train a model to learn a distance-based metric between samples. For example, learning a distance-based similarity measure between faces.

### Training ###
By default, the training script trains ResNet50 on the FashionMNIST Dataset to learn image embeddings which can be used to differentiate between images by measuring the euclidean distance between embeddings. This can be changed as per your requirements.

Image embeddings of the same class should be 'close' to each other, while image embeddings between different classes should be 'far' away.

To run the training script:

```bash
python train.py -h    # Lists all optional arguments
python train.py 			# Runs training script with default args
```

Running the training script as is should yield 97% accuracy on the FMNIST test set within 10 epochs.

### Loss ###
`TripletMarginLoss` is a loss function which takes in a triplet of samples. A valid triplet has an:

1. Anchor: a sample from the dataset
2. Positive: another sample with the same label/group as the anchor (Generally, positive != anchor)
3. Negative: a sample with a different label/group from the anchor

`TripletMarginLoss` (refer to `loss.py`) does the following:

```python
loss = max(dist(anchor, positive) - dist(anchor, negative) + margin, 0)
```
Where `dist` is a distance function. Minimizing this function effectively leads to minimizing `dist(anchor, positive)` and maximizing `dist(anchor, negative)`.

The FaceNet paper describe this loss in more detail.

### Sampler ###

In order to generate valid triplets from a batch of samples, we need to make sure that each batch has multiple samples with the same label. We do this using `PKSampler` (refer to `sampler.py`), which ensures that each batch of size `p * k` will have samples from exactly `p` classes and `k` samples per class.

### Triplet Mining ###

`TripletMarginLoss` currently supports the following mining techniques:

* `batch_all`: Generates all possible triplets from a batch and excludes the triplets which are 'easy' (which have `loss = 0`) before passing it through the loss function.
* `batch_hard`: For every anchor, `batch_hard` creates a triplet with the 'hardest' positive (farthest positive) and negative (closest negative).

These mining strategies usually speed up training.

This [webpage](https://omoindrot.github.io/triplet-loss) describes the sampling and mining strategies in more detail. 

# Optical flow reference training scripts

This folder contains reference training scripts for optical flow.
They serve as a log of how to train specific models, so as to provide baseline
training and evaluation scripts to quickly bootstrap research.


### RAFT Large

The RAFT large model was trained on Flying Chairs and then on Flying Things.
Both used 8 A100 GPUs and a batch size of 2 (so effective batch size is 16). The
rest of the hyper-parameters are exactly the same as the original RAFT training
recipe from https://github.com/princeton-vl/RAFT. The original recipe trains for
100000 updates (or steps) on each dataset - this corresponds to about 72 and 20
epochs on Chairs and Things respectively:

```
num_epochs = ceil(num_steps / number_of_steps_per_epoch)
           = ceil(num_steps / (num_samples / effective_batch_size))
```

```
torchrun --nproc_per_node 8 --nnodes 1 train.py \
    --dataset-root $dataset_root \
    --name $name_chairs \
    --model raft_large \
    --train-dataset chairs \
    --batch-size 2 \
    --lr 0.0004 \
    --weight-decay 0.0001 \
    --epochs 72 \
    --output-dir $chairs_dir
```

```
torchrun --nproc_per_node 8 --nnodes 1 train.py \
    --dataset-root $dataset_root \
    --name $name_things \
    --model raft_large \
    --train-dataset things \
    --batch-size 2 \
    --lr 0.000125 \
    --weight-decay 0.0001 \
    --epochs 20 \
    --freeze-batch-norm \
    --output-dir $things_dir\
    --resume $chairs_dir/$name_chairs.pth
```


### Evaluation

```
torchrun --nproc_per_node 1 --nnodes 1 train.py --val-dataset sintel --batch-size 1 --dataset-root $dataset_root --model raft_large --weights Raft_Large_Weights.C_T_SKHT_V2
```

This should give an epe of about 1.3822 on the clean pass and 2.7161 on the
final pass of Sintel-train. Results may vary slightly depending on the batch
size and the number of GPUs. For the most accurate results use 1 GPU and
`--batch-size 1`:

```
Sintel val clean epe: 1.3822	1px: 0.9028	3px: 0.9573	5px: 0.9697	per_image_epe: 1.3822	f1: 4.0248
Sintel val final epe: 2.7161	1px: 0.8528	3px: 0.9204	5px: 0.9392	per_image_epe: 2.7161	f1: 7.5964
```

You can also evaluate on Kitti train:

```
torchrun --nproc_per_node 1 --nnodes 1 train.py --val-dataset kitti --batch-size 1 --dataset-root $dataset_root --model raft_large --weights Raft_Large_Weights.C_T_SKHT_V2
Kitti val epe: 4.7968	1px: 0.6388	3px: 0.8197	5px: 0.8661	per_image_epe: 4.5118	f1: 16.0679
```

# Semantic segmentation reference training scripts

This folder contains reference training scripts for semantic segmentation.
They serve as a log of how to train specific models, as provide baseline
training and evaluation scripts to quickly bootstrap research.

All models have been trained on 8x V100 GPUs.

You must modify the following flags:

`--data-path=/path/to/dataset`

`--nproc_per_node=<number_of_gpus_available>`

## fcn_resnet50
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model fcn_resnet50 --aux-loss --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

## fcn_resnet101
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model fcn_resnet101 --aux-loss --weights-backbone ResNet101_Weights.IMAGENET1K_V1
```

## deeplabv3_resnet50
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model deeplabv3_resnet50 --aux-loss --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

## deeplabv3_resnet101
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model deeplabv3_resnet101 --aux-loss --weights-backbone ResNet101_Weights.IMAGENET1K_V1
```

## deeplabv3_mobilenet_v3_large
```
torchrun --nproc_per_node=8 train.py --dataset coco -b 4 --model deeplabv3_mobilenet_v3_large --aux-loss --wd 0.000001 --weights-backbone MobileNet_V3_Large_Weights.IMAGENET1K_V1
```

## lraspp_mobilenet_v3_large
```
torchrun --nproc_per_node=8 train.py --dataset coco -b 4 --model lraspp_mobilenet_v3_large --wd 0.000001 --weights-backbone MobileNet_V3_Large_Weights.IMAGENET1K_V1
```

# Image classification reference training scripts

This folder contains reference training scripts for image classification.
They serve as a log of how to train specific models, as provide baseline
training and evaluation scripts to quickly bootstrap research.

Except otherwise noted, all models have been trained on 8x V100 GPUs with 
the following parameters:

| Parameter                | value  |
| ------------------------ | ------ |
| `--batch_size`           | `32`   |
| `--epochs`               | `90`   |
| `--lr`                   | `0.1`  |
| `--momentum`             | `0.9`  |
| `--wd`, `--weight-decay` | `1e-4` |
| `--lr-step-size`         | `30`   |
| `--lr-gamma`             | `0.1`  |

### AlexNet and VGG

Since `AlexNet` and the original `VGG` architectures do not include batch 
normalization, the default initial learning rate `--lr 0.1` is too high.

```
torchrun --nproc_per_node=8 train.py\
    --model $MODEL --lr 1e-2
```

Here `$MODEL` is one of `alexnet`, `vgg11`, `vgg13`, `vgg16` or `vgg19`. Note
that `vgg11_bn`, `vgg13_bn`, `vgg16_bn`, and `vgg19_bn` include batch
normalization and thus are trained with the default parameters.

### GoogLeNet

The weights of the GoogLeNet model are ported from the original paper rather than trained from scratch.

### Inception V3

The weights of the Inception V3 model are ported from the original paper rather than trained from scratch.

Since it expects tensors with a size of N x 3 x 299 x 299, to validate the model use the following command:

```
torchrun --nproc_per_node=8 train.py --model inception_v3\
      --test-only --weights Inception_V3_Weights.IMAGENET1K_V1
```

### ResNet
```
torchrun --nproc_per_node=8 train.py --model $MODEL
```

Here `$MODEL` is one of `resnet18`, `resnet34`, `resnet50`, `resnet101` or `resnet152`.

### ResNext
```
torchrun --nproc_per_node=8 train.py\
    --model $MODEL --epochs 100
```

Here `$MODEL` is one of `resnext50_32x4d` or `resnext101_32x8d`.
Note that the above command corresponds to a single node with 8 GPUs. If you use
a different number of GPUs and/or a different batch size, then the learning rate
should be scaled accordingly. For example, the pretrained model provided by
`torchvision` was trained on 8 nodes, each with 8 GPUs (for a total of 64 GPUs),
with `--batch_size 16` and `--lr 0.4`, instead of the current defaults
which are respectively batch_size=32 and lr=0.1

### MobileNetV2
```
torchrun --nproc_per_node=8 train.py\
     --model mobilenet_v2 --epochs 300 --lr 0.045 --wd 0.00004\
     --lr-step-size 1 --lr-gamma 0.98
```


### MobileNetV3 Large & Small
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 600 --opt rmsprop --batch-size 128 --lr 0.064\ 
     --wd 0.00001 --lr-step-size 2 --lr-gamma 0.973 --auto-augment imagenet --random-erase 0.2
```

Here `$MODEL` is one of `mobilenet_v3_large` or `mobilenet_v3_small`.

Then we averaged the parameters of the last 3 checkpoints that improved the Acc@1. See [#3182](https://github.com/pytorch/vision/pull/3182) 
and [#3354](https://github.com/pytorch/vision/pull/3354) for details.


### EfficientNet-V1

The weights of the B0-B4 variants are ported from Ross Wightman's [timm repo](https://github.com/rwightman/pytorch-image-models/blob/01cb46a9a50e3ba4be167965b5764e9702f09b30/timm/models/efficientnet.py#L95-L108).

The weights of the B5-B7 variants are ported from Luke Melas' [EfficientNet-PyTorch repo](https://github.com/lukemelas/EfficientNet-PyTorch/blob/1039e009545d9329ea026c9f7541341439712b96/efficientnet_pytorch/utils.py#L562-L564).

All models were trained using Bicubic interpolation and each have custom crop and resize sizes. To validate the models use the following commands:
```
torchrun --nproc_per_node=8 train.py --model efficientnet_b0 --test-only --weights EfficientNet_B0_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b1 --test-only --weights EfficientNet_B1_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b2 --test-only --weights EfficientNet_B2_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b3 --test-only --weights EfficientNet_B3_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b4 --test-only --weights EfficientNet_B4_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b5 --test-only --weights EfficientNet_B5_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b6 --test-only --weights EfficientNet_B6_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b7 --test-only --weights EfficientNet_B7_Weights.IMAGENET1K_V1
```


### EfficientNet-V2
```
torchrun --nproc_per_node=8 train.py \
--model $MODEL --batch-size 128 --lr 0.5 --lr-scheduler cosineannealinglr \
--lr-warmup-epochs 5 --lr-warmup-method linear --auto-augment ta_wide --epochs 600 --random-erase 0.1 \
--label-smoothing 0.1 --mixup-alpha 0.2 --cutmix-alpha 1.0 --weight-decay 0.00002 --norm-weight-decay 0.0 \
--train-crop-size $TRAIN_SIZE --model-ema --val-crop-size $EVAL_SIZE --val-resize-size $EVAL_SIZE \
--ra-sampler --ra-reps 4
```
Here `$MODEL` is one of `efficientnet_v2_s` and `efficientnet_v2_m`. 
Note that the Small variant had a `$TRAIN_SIZE` of `300` and a `$EVAL_SIZE` of `384`, while the Medium `384` and `480` respectively.

Note that the above command corresponds to training on a single node with 8 GPUs.
For generatring the pre-trained weights, we trained with 4 nodes, each with 8 GPUs (for a total of 32 GPUs),
and `--batch_size 32`.

The weights of the Large variant are ported from the original paper rather than trained from scratch. See the `EfficientNet_V2_L_Weights` entry for their exact preprocessing transforms.


### RegNet

#### Small models
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 100 --batch-size 128 --wd 0.00005 --lr=0.8\
     --lr-scheduler=cosineannealinglr --lr-warmup-method=linear\
     --lr-warmup-epochs=5 --lr-warmup-decay=0.1
```
Here `$MODEL` is one of `regnet_x_400mf`, `regnet_x_800mf`, `regnet_x_1_6gf`, `regnet_y_400mf`, `regnet_y_800mf` and `regnet_y_1_6gf`. Please note we used learning rate 0.4 for `regent_y_400mf` to get the same Acc@1 as [the paper)(https://arxiv.org/abs/2003.13678).

#### Medium models
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 100 --batch-size 64 --wd 0.00005 --lr=0.4\
     --lr-scheduler=cosineannealinglr --lr-warmup-method=linear\
     --lr-warmup-epochs=5 --lr-warmup-decay=0.1
```
Here `$MODEL` is one of `regnet_x_3_2gf`, `regnet_x_8gf`, `regnet_x_16gf`, `regnet_y_3_2gf` and `regnet_y_8gf`.

#### Large models
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 100 --batch-size 32 --wd 0.00005 --lr=0.2\
     --lr-scheduler=cosineannealinglr --lr-warmup-method=linear\
     --lr-warmup-epochs=5 --lr-warmup-decay=0.1
```
Here `$MODEL` is one of `regnet_x_32gf`, `regnet_y_16gf` and `regnet_y_32gf`.

### Vision Transformer

#### vit_b_16
```
torchrun --nproc_per_node=8 train.py\
    --model vit_b_16 --epochs 300 --batch-size 512 --opt adamw --lr 0.003 --wd 0.3\
    --lr-scheduler cosineannealinglr --lr-warmup-method linear --lr-warmup-epochs 30\
    --lr-warmup-decay 0.033 --amp --label-smoothing 0.11 --mixup-alpha 0.2 --auto-augment ra\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generatring the pre-trained weights, we trained with 8 nodes, each with 8 GPUs (for a total of 64 GPUs),
and `--batch_size 64`.

#### vit_b_32
```
torchrun --nproc_per_node=8 train.py\
    --model vit_b_32 --epochs 300 --batch-size 512 --opt adamw --lr 0.003 --wd 0.3\
    --lr-scheduler cosineannealinglr --lr-warmup-method linear --lr-warmup-epochs 30\
    --lr-warmup-decay 0.033 --amp --label-smoothing 0.11 --mixup-alpha 0.2 --auto-augment imagenet\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generatring the pre-trained weights, we trained with 2 nodes, each with 8 GPUs (for a total of 16 GPUs),
and `--batch_size 256`.

#### vit_l_16
```
torchrun --nproc_per_node=8 train.py\
    --model vit_l_16 --epochs 600 --batch-size 128 --lr 0.5 --lr-scheduler cosineannealinglr\
    --lr-warmup-method linear --lr-warmup-epochs 5 --label-smoothing 0.1 --mixup-alpha 0.2\
    --auto-augment ta_wide --random-erase 0.1 --weight-decay 0.00002 --norm-weight-decay 0.0\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema --val-resize-size 232
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generatring the pre-trained weights, we trained with 2 nodes, each with 8 GPUs (for a total of 16 GPUs),
and `--batch_size 64`.

#### vit_l_32
```
torchrun --nproc_per_node=8 train.py\
    --model vit_l_32 --epochs 300 --batch-size 512 --opt adamw --lr 0.003 --wd 0.3\
    --lr-scheduler cosineannealinglr --lr-warmup-method linear --lr-warmup-epochs 30\
    --lr-warmup-decay 0.033 --amp --label-smoothing 0.11 --mixup-alpha 0.2 --auto-augment ra\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generatring the pre-trained weights, we trained with 8 nodes, each with 8 GPUs (for a total of 64 GPUs),
and `--batch_size 64`.


### ConvNeXt
```
torchrun --nproc_per_node=8 train.py\ 
--model $MODEL --batch-size 128 --opt adamw --lr 1e-3 --lr-scheduler cosineannealinglr \ 
--lr-warmup-epochs 5 --lr-warmup-method linear --auto-augment ta_wide --epochs 600 --random-erase 0.1 \ 
--label-smoothing 0.1 --mixup-alpha 0.2 --cutmix-alpha 1.0 --weight-decay 0.05 --norm-weight-decay 0.0 \
--train-crop-size 176 --model-ema --val-resize-size 232 --ra-sampler --ra-reps 4
```
Here `$MODEL` is one of `convnext_tiny`, `convnext_small`, `convnext_base` and `convnext_large`. Note that each variant had its `--val-resize-size` optimized in a post-training step, see their `Weights` entry for their exact value.

Note that the above command corresponds to training on a single node with 8 GPUs.
For generatring the pre-trained weights, we trained with 2 nodes, each with 8 GPUs (for a total of 16 GPUs),
and `--batch_size 64`.


### SwinTransformer
```
torchrun --nproc_per_node=8 train.py\ 
--model $MODEL --epochs 300 --batch-size 128 --opt adamw --lr 0.001 --weight-decay 0.05 --norm-weight-decay 0.0  --bias-weight-decay 0.0 --transformer-embedding-decay 0.0 --lr-scheduler cosineannealinglr --lr-min 0.00001 --lr-warmup-method linear  --lr-warmup-epochs 20 --lr-warmup-decay 0.01 --amp --label-smoothing 0.1 --mixup-alpha 0.8 --clip-grad-norm 5.0 --cutmix-alpha 1.0 --random-erase 0.25 --interpolation bicubic --auto-augment ta_wide --model-ema --ra-sampler --ra-reps 4  --val-resize-size 224
```
Here `$MODEL` is one of `swin_t`, `swin_s` or `swin_b`.
Note that `--val-resize-size` was optimized in a post-training step, see their `Weights` entry for the exact value.




### SwinTransformer V2
```
torchrun --nproc_per_node=8 train.py\
--model $MODEL --epochs 300 --batch-size 128 --opt adamw --lr 0.001 --weight-decay 0.05 --norm-weight-decay 0.0  --bias-weight-decay 0.0 --transformer-embedding-decay 0.0 --lr-scheduler cosineannealinglr --lr-min 0.00001 --lr-warmup-method linear  --lr-warmup-epochs 20 --lr-warmup-decay 0.01 --amp --label-smoothing 0.1 --mixup-alpha 0.8 --clip-grad-norm 5.0 --cutmix-alpha 1.0 --random-erase 0.25 --interpolation bicubic --auto-augment ta_wide --model-ema --ra-sampler --ra-reps 4  --val-resize-size 256 --val-crop-size 256 --train-crop-size 256 
```
Here `$MODEL` is one of `swin_v2_t`, `swin_v2_s` or `swin_v2_b`.
Note that `--val-resize-size` was optimized in a post-training step, see their `Weights` entry for the exact value.


### MaxViT
```
torchrun --nproc_per_node=8 --n_nodes=4 train.py\
--model $MODEL --epochs 400 --batch-size 128 --opt adamw --lr 3e-3 --weight-decay 0.05 --lr-scheduler cosineannealinglr --lr-min 1e-5 --lr-warmup-method linear  --lr-warmup-epochs 32  --label-smoothing 0.1 --mixup-alpha 0.8 --clip-grad-norm 1.0 --interpolation bicubic --auto-augment ta_wide --policy-magnitude 15 --model-ema --val-resize-size 224\
--val-crop-size 224 --train-crop-size 224 --amp  --model-ema-steps 32 --transformer-embedding-decay 0 --sync-bn
```
Here `$MODEL` is `maxvit_t`.
Note that `--val-resize-size` was not optimized in a post-training step.


### ShuffleNet V2
```
torchrun --nproc_per_node=8 train.py \
--batch-size=128 \
--lr=0.5 --lr-scheduler=cosineannealinglr --lr-warmup-epochs=5 --lr-warmup-method=linear \
--auto-augment=ta_wide --epochs=600 --random-erase=0.1 --weight-decay=0.00002 \
--norm-weight-decay=0.0 --label-smoothing=0.1 --mixup-alpha=0.2 --cutmix-alpha=1.0 \
--train-crop-size=176 --model-ema --val-resize-size=232 --ra-sampler --ra-reps=4
```
Here `$MODEL` is either `shufflenet_v2_x1_5` or `shufflenet_v2_x2_0`.

The models `shufflenet_v2_x0_5` and `shufflenet_v2_x1_0` were contributed by the community. See [PR-849](https://github.com/pytorch/vision/pull/849#issuecomment-483391686) for details.


## Mixed precision training
Automatic Mixed Precision (AMP) training on GPU for Pytorch can be enabled with the [torch.cuda.amp](https://pytorch.org/docs/stable/amp.html?highlight=amp#module-torch.cuda.amp).

Mixed precision training makes use of both FP32 and FP16 precisions where appropriate. FP16 operations can leverage the Tensor cores on NVIDIA GPUs (Volta, Turing or newer architectures) for improved throughput, generally without loss in model accuracy. Mixed precision training also often allows larger batch sizes. GPU automatic mixed precision training for Pytorch Vision can be enabled via the flag value `--amp=True`.

```
torchrun --nproc_per_node=8 train.py\
    --model resnext50_32x4d --epochs 100 --amp
```

## Quantized

### Post training quantized models

For all post training quantized models, the settings are:

1. num_calibration_batches: 32
2. num_workers: 16
3. batch_size: 32
4. eval_batch_size: 128
5. backend: 'fbgemm'

```
python train_quantization.py --device='cpu' --post-training-quantize --backend='fbgemm' --model='$MODEL'
```
Here `$MODEL` is one of `googlenet`, `inception_v3`, `resnet18`, `resnet50`, `resnext101_32x8d`, `shufflenet_v2_x0_5` and `shufflenet_v2_x1_0`.

### Quantized ShuffleNet V2

Here are commands that we use to quantize the `shufflenet_v2_x1_5` and `shufflenet_v2_x2_0` models.
```
# For shufflenet_v2_x1_5
python train_quantization.py --device='cpu' --post-training-quantize --backend='fbgemm' \
    --model=shufflenet_v2_x1_5 --weights="ShuffleNet_V2_X1_5_Weights.IMAGENET1K_V1" \
    --train-crop-size 176 --val-resize-size 232 --data-path /datasets01_ontap/imagenet_full_size/061417/

# For shufflenet_v2_x2_0
python train_quantization.py --device='cpu' --post-training-quantize --backend='fbgemm' \
    --model=shufflenet_v2_x2_0 --weights="ShuffleNet_V2_X2_0_Weights.IMAGENET1K_V1" \
    --train-crop-size 176 --val-resize-size 232 --data-path /datasets01_ontap/imagenet_full_size/061417/
```

### QAT MobileNetV2

For Mobilenet-v2, the model was trained with quantization aware training, the settings used are:
1. num_workers: 16
2. batch_size: 32
3. eval_batch_size: 128
4. backend: 'qnnpack'
5. learning-rate: 0.0001
6. num_epochs: 90
7. num_observer_update_epochs:4
8. num_batch_norm_update_epochs:3
9. momentum: 0.9
10. lr_step_size:30
11. lr_gamma: 0.1
12. weight-decay: 0.0001

```
torchrun --nproc_per_node=8 train_quantization.py --model='mobilenet_v2'
```

Training converges at about 10 epochs.

### QAT MobileNetV3

For Mobilenet-v3 Large, the model was trained with quantization aware training, the settings used are:
1. num_workers: 16
2. batch_size: 32
3. eval_batch_size: 128
4. backend: 'qnnpack'
5. learning-rate: 0.001
6. num_epochs: 90
7. num_observer_update_epochs:4
8. num_batch_norm_update_epochs:3
9. momentum: 0.9
10. lr_step_size:30
11. lr_gamma: 0.1
12. weight-decay: 0.00001

```
torchrun --nproc_per_node=8 train_quantization.py --model='mobilenet_v3_large' \
    --wd 0.00001 --lr 0.001
```

For post training quant, device is set to CPU. For training, the device is set to CUDA.

### Command to evaluate quantized models using the pre-trained weights:

```
python train_quantization.py --device='cpu' --test-only --backend='<backend>' --model='<model_name>'
```

For inception_v3 you need to pass the following extra parameters:
```
--val-resize-size 342 --val-crop-size 299 --train-crop-size 299
```

# Semantic segmentation reference training scripts

This folder contains reference training scripts for semantic segmentation.
They serve as a log of how to train specific models and provide baseline
training and evaluation scripts to quickly bootstrap research.

All models have been trained on 8x V100 GPUs.

You must modify the following flags:

`--data-path=/path/to/dataset`

`--nproc_per_node=<number_of_gpus_available>`

## fcn_resnet50
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model fcn_resnet50 --aux-loss --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

## fcn_resnet101
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model fcn_resnet101 --aux-loss --weights-backbone ResNet101_Weights.IMAGENET1K_V1
```

## deeplabv3_resnet50
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model deeplabv3_resnet50 --aux-loss --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

## deeplabv3_resnet101
```
torchrun --nproc_per_node=8 train.py --lr 0.02 --dataset coco -b 4 --model deeplabv3_resnet101 --aux-loss --weights-backbone ResNet101_Weights.IMAGENET1K_V1
```

## deeplabv3_mobilenet_v3_large
```
torchrun --nproc_per_node=8 train.py --dataset coco -b 4 --model deeplabv3_mobilenet_v3_large --aux-loss --wd 0.000001 --weights-backbone MobileNet_V3_Large_Weights.IMAGENET1K_V1
```

## lraspp_mobilenet_v3_large
```
torchrun --nproc_per_node=8 train.py --dataset coco -b 4 --model lraspp_mobilenet_v3_large --wd 0.000001 --weights-backbone MobileNet_V3_Large_Weights.IMAGENET1K_V1
```

# Object detection reference training scripts

This folder contains reference training scripts for object detection.
They serve as a log of how to train specific models, to provide baseline
training and evaluation scripts to quickly bootstrap research.

To execute the example commands below you must install the following:

```
cython
pycocotools
matplotlib
```

You must modify the following flags:

`--data-path=/path/to/coco/dataset`

`--nproc_per_node=<number_of_gpus_available>`

Except otherwise noted, all models have been trained on 8x V100 GPUs. 

### Faster R-CNN ResNet-50 FPN
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model fasterrcnn_resnet50_fpn --epochs 26\
    --lr-steps 16 22 --aspect-ratio-group-factor 3 --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

### Faster R-CNN MobileNetV3-Large FPN
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model fasterrcnn_mobilenet_v3_large_fpn --epochs 26\
    --lr-steps 16 22 --aspect-ratio-group-factor 3 --weights-backbone MobileNet_V3_Large_Weights.IMAGENET1K_V1
```

### Faster R-CNN MobileNetV3-Large 320 FPN
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model fasterrcnn_mobilenet_v3_large_320_fpn --epochs 26\
    --lr-steps 16 22 --aspect-ratio-group-factor 3 --weights-backbone MobileNet_V3_Large_Weights.IMAGENET1K_V1
```

### FCOS ResNet-50 FPN
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model fcos_resnet50_fpn --epochs 26\
    --lr-steps 16 22 --aspect-ratio-group-factor 3  --lr 0.01 --amp --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

### RetinaNet
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model retinanet_resnet50_fpn --epochs 26\
    --lr-steps 16 22 --aspect-ratio-group-factor 3 --lr 0.01 --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

### SSD300 VGG16
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model ssd300_vgg16 --epochs 120\
    --lr-steps 80 110 --aspect-ratio-group-factor 3 --lr 0.002 --batch-size 4\
    --weight-decay 0.0005 --data-augmentation ssd --weights-backbone VGG16_Weights.IMAGENET1K_FEATURES
```

### SSDlite320 MobileNetV3-Large
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model ssdlite320_mobilenet_v3_large --epochs 660\
    --aspect-ratio-group-factor 3 --lr-scheduler cosineannealinglr --lr 0.15 --batch-size 24\
    --weight-decay 0.00004 --data-augmentation ssdlite
```


### Mask R-CNN
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco --model maskrcnn_resnet50_fpn --epochs 26\
    --lr-steps 16 22 --aspect-ratio-group-factor 3 --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```


### Keypoint R-CNN
```
torchrun --nproc_per_node=8 train.py\
    --dataset coco_kp --model keypointrcnn_resnet50_fpn --epochs 46\
    --lr-steps 36 43 --aspect-ratio-group-factor 3 --weights-backbone ResNet50_Weights.IMAGENET1K_V1
```

# Image classification reference training scripts

This folder contains reference training scripts for image classification.
They serve as a log of how to train specific models, as provide baseline
training and evaluation scripts to quickly bootstrap research.

Except otherwise noted, all models have been trained on 8x V100 GPUs with 
the following parameters:

| Parameter                | value  |
| ------------------------ | ------ |
| `--batch_size`           | `32`   |
| `--epochs`               | `90`   |
| `--lr`                   | `0.1`  |
| `--momentum`             | `0.9`  |
| `--wd`, `--weight-decay` | `1e-4` |
| `--lr-step-size`         | `30`   |
| `--lr-gamma`             | `0.1`  |

### AlexNet and VGG

Since `AlexNet` and the original `VGG` architectures do not include batch 
normalization, the default initial learning rate `--lr 0.1` is too high.

```
torchrun --nproc_per_node=8 train.py\
    --model $MODEL --lr 1e-2
```

Here `$MODEL` is one of `alexnet`, `vgg11`, `vgg13`, `vgg16` or `vgg19`. Note
that `vgg11_bn`, `vgg13_bn`, `vgg16_bn`, and `vgg19_bn` include batch
normalization and thus are trained with the default parameters.

### GoogLeNet

The weights of the GoogLeNet model are ported from the original paper rather than trained from scratch.

### Inception V3

The weights of the Inception V3 model are ported from the original paper rather than trained from scratch.

Since it expects tensors with a size of N x 3 x 299 x 299, to validate the model use the following command:

```
torchrun --nproc_per_node=8 train.py --model inception_v3\
      --test-only --weights Inception_V3_Weights.IMAGENET1K_V1
```

### ResNet
```
torchrun --nproc_per_node=8 train.py --model $MODEL
```

Here `$MODEL` is one of `resnet18`, `resnet34`, `resnet50`, `resnet101` or `resnet152`.

### ResNext
```
torchrun --nproc_per_node=8 train.py\
    --model $MODEL --epochs 100
```

Here `$MODEL` is one of `resnext50_32x4d` or `resnext101_32x8d`.
Note that the above command corresponds to a single node with 8 GPUs. If you use
a different number of GPUs and/or a different batch size, then the learning rate
should be scaled accordingly. For example, the pretrained model provided by
`torchvision` was trained on 8 nodes, each with 8 GPUs (for a total of 64 GPUs),
with `--batch_size 16` and `--lr 0.4`, instead of the current defaults
which are respectively batch_size=32 and lr=0.1

### MobileNetV2
```
torchrun --nproc_per_node=8 train.py\
     --model mobilenet_v2 --epochs 300 --lr 0.045 --wd 0.00004\
     --lr-step-size 1 --lr-gamma 0.98
```


### MobileNetV3 Large & Small
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 600 --opt rmsprop --batch-size 128 --lr 0.064\ 
     --wd 0.00001 --lr-step-size 2 --lr-gamma 0.973 --auto-augment imagenet --random-erase 0.2
```

Here `$MODEL` is one of `mobilenet_v3_large` or `mobilenet_v3_small`.

Then we averaged the parameters of the last 3 checkpoints that improved the Acc@1. See [#3182](https://github.com/pytorch/vision/pull/3182) 
and [#3354](https://github.com/pytorch/vision/pull/3354) for details.


### EfficientNet-V1

The weights of the B0-B4 variants are ported from Ross Wightman's [timm repo](https://github.com/rwightman/pytorch-image-models/blob/01cb46a9a50e3ba4be167965b5764e9702f09b30/timm/models/efficientnet.py#L95-L108).

The weights of the B5-B7 variants are ported from Luke Melas' [EfficientNet-PyTorch repo](https://github.com/lukemelas/EfficientNet-PyTorch/blob/1039e009545d9329ea026c9f7541341439712b96/efficientnet_pytorch/utils.py#L562-L564).

All models were trained using Bicubic interpolation and each have custom crop and resize sizes. To validate the models use the following commands:
```
torchrun --nproc_per_node=8 train.py --model efficientnet_b0 --test-only --weights EfficientNet_B0_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b1 --test-only --weights EfficientNet_B1_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b2 --test-only --weights EfficientNet_B2_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b3 --test-only --weights EfficientNet_B3_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b4 --test-only --weights EfficientNet_B4_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b5 --test-only --weights EfficientNet_B5_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b6 --test-only --weights EfficientNet_B6_Weights.IMAGENET1K_V1
torchrun --nproc_per_node=8 train.py --model efficientnet_b7 --test-only --weights EfficientNet_B7_Weights.IMAGENET1K_V1
```


### EfficientNet-V2
```
torchrun --nproc_per_node=8 train.py \
--model $MODEL --batch-size 128 --lr 0.5 --lr-scheduler cosineannealinglr \
--lr-warmup-epochs 5 --lr-warmup-method linear --auto-augment ta_wide --epochs 600 --random-erase 0.1 \
--label-smoothing 0.1 --mixup-alpha 0.2 --cutmix-alpha 1.0 --weight-decay 0.00002 --norm-weight-decay 0.0 \
--train-crop-size $TRAIN_SIZE --model-ema --val-crop-size $EVAL_SIZE --val-resize-size $EVAL_SIZE \
--ra-sampler --ra-reps 4
```
Here `$MODEL` is one of `efficientnet_v2_s` and `efficientnet_v2_m`. 
Note that the Small variant had a `$TRAIN_SIZE` of `300` and a `$EVAL_SIZE` of `384`, while the Medium `384` and `480` respectively.

Note that the above command corresponds to training on a single node with 8 GPUs.
For generating the pre-trained weights, we trained with 4 nodes, each with 8 GPUs (for a total of 32 GPUs),
and `--batch_size 32`.

The weights of the Large variant are ported from the original paper rather than trained from scratch. See the `EfficientNet_V2_L_Weights` entry for their exact preprocessing transforms.


### RegNet

#### Small models
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 100 --batch-size 128 --wd 0.00005 --lr=0.8\
     --lr-scheduler=cosineannealinglr --lr-warmup-method=linear\
     --lr-warmup-epochs=5 --lr-warmup-decay=0.1
```
Here `$MODEL` is one of `regnet_x_400mf`, `regnet_x_800mf`, `regnet_x_1_6gf`, `regnet_y_400mf`, `regnet_y_800mf` and `regnet_y_1_6gf`. Please note we used learning rate 0.4 for `regent_y_400mf` to get the same Acc@1 as [the paper](https://arxiv.org/abs/2003.13678).

#### Medium models
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 100 --batch-size 64 --wd 0.00005 --lr=0.4\
     --lr-scheduler=cosineannealinglr --lr-warmup-method=linear\
     --lr-warmup-epochs=5 --lr-warmup-decay=0.1
```
Here `$MODEL` is one of `regnet_x_3_2gf`, `regnet_x_8gf`, `regnet_x_16gf`, `regnet_y_3_2gf` and `regnet_y_8gf`.

#### Large models
```
torchrun --nproc_per_node=8 train.py\
     --model $MODEL --epochs 100 --batch-size 32 --wd 0.00005 --lr=0.2\
     --lr-scheduler=cosineannealinglr --lr-warmup-method=linear\
     --lr-warmup-epochs=5 --lr-warmup-decay=0.1
```
Here `$MODEL` is one of `regnet_x_32gf`, `regnet_y_16gf` and `regnet_y_32gf`.

### Vision Transformer

#### vit_b_16
```
torchrun --nproc_per_node=8 train.py\
    --model vit_b_16 --epochs 300 --batch-size 512 --opt adamw --lr 0.003 --wd 0.3\
    --lr-scheduler cosineannealinglr --lr-warmup-method linear --lr-warmup-epochs 30\
    --lr-warmup-decay 0.033 --amp --label-smoothing 0.11 --mixup-alpha 0.2 --auto-augment ra\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generating the pre-trained weights, we trained with 8 nodes, each with 8 GPUs (for a total of 64 GPUs),
and `--batch_size 64`.

#### vit_b_32
```
torchrun --nproc_per_node=8 train.py\
    --model vit_b_32 --epochs 300 --batch-size 512 --opt adamw --lr 0.003 --wd 0.3\
    --lr-scheduler cosineannealinglr --lr-warmup-method linear --lr-warmup-epochs 30\
    --lr-warmup-decay 0.033 --amp --label-smoothing 0.11 --mixup-alpha 0.2 --auto-augment imagenet\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generating the pre-trained weights, we trained with 2 nodes, each with 8 GPUs (for a total of 16 GPUs),
and `--batch_size 256`.

#### vit_l_16
```
torchrun --nproc_per_node=8 train.py\
    --model vit_l_16 --epochs 600 --batch-size 128 --lr 0.5 --lr-scheduler cosineannealinglr\
    --lr-warmup-method linear --lr-warmup-epochs 5 --label-smoothing 0.1 --mixup-alpha 0.2\
    --auto-augment ta_wide --random-erase 0.1 --weight-decay 0.00002 --norm-weight-decay 0.0\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema --val-resize-size 232
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generating the pre-trained weights, we trained with 2 nodes, each with 8 GPUs (for a total of 16 GPUs),
and `--batch_size 64`.

#### vit_l_32
```
torchrun --nproc_per_node=8 train.py\
    --model vit_l_32 --epochs 300 --batch-size 512 --opt adamw --lr 0.003 --wd 0.3\
    --lr-scheduler cosineannealinglr --lr-warmup-method linear --lr-warmup-epochs 30\
    --lr-warmup-decay 0.033 --amp --label-smoothing 0.11 --mixup-alpha 0.2 --auto-augment ra\
    --clip-grad-norm 1 --ra-sampler --cutmix-alpha 1.0 --model-ema
```

Note that the above command corresponds to training on a single node with 8 GPUs.
For generating the pre-trained weights, we trained with 8 nodes, each with 8 GPUs (for a total of 64 GPUs),
and `--batch_size 64`.


### ConvNeXt
```
torchrun --nproc_per_node=8 train.py\ 
--model $MODEL --batch-size 128 --opt adamw --lr 1e-3 --lr-scheduler cosineannealinglr \ 
--lr-warmup-epochs 5 --lr-warmup-method linear --auto-augment ta_wide --epochs 600 --random-erase 0.1 \ 
--label-smoothing 0.1 --mixup-alpha 0.2 --cutmix-alpha 1.0 --weight-decay 0.05 --norm-weight-decay 0.0 \
--train-crop-size 176 --model-ema --val-resize-size 232 --ra-sampler --ra-reps 4
```
Here `$MODEL` is one of `convnext_tiny`, `convnext_small`, `convnext_base` and `convnext_large`. Note that each variant had its `--val-resize-size` optimized in a post-training step, see their `Weights` entry for their exact value.

Note that the above command corresponds to training on a single node with 8 GPUs.
For generating the pre-trained weights, we trained with 2 nodes, each with 8 GPUs (for a total of 16 GPUs),
and `--batch_size 64`.


### SwinTransformer
```
torchrun --nproc_per_node=8 train.py\ 
--model $MODEL --epochs 300 --batch-size 128 --opt adamw --lr 0.001 --weight-decay 0.05 --norm-weight-decay 0.0  --bias-weight-decay 0.0 --transformer-embedding-decay 0.0 --lr-scheduler cosineannealinglr --lr-min 0.00001 --lr-warmup-method linear  --lr-warmup-epochs 20 --lr-warmup-decay 0.01 --amp --label-smoothing 0.1 --mixup-alpha 0.8 --clip-grad-norm 5.0 --cutmix-alpha 1.0 --random-erase 0.25 --interpolation bicubic --auto-augment ta_wide --model-ema --ra-sampler --ra-reps 4  --val-resize-size 224
```
Here `$MODEL` is one of `swin_t`, `swin_s` or `swin_b`.
Note that `--val-resize-size` was optimized in a post-training step, see their `Weights` entry for the exact value.




### SwinTransformer V2
```
torchrun --nproc_per_node=8 train.py\
--model $MODEL --epochs 300 --batch-size 128 --opt adamw --lr 0.001 --weight-decay 0.05 --norm-weight-decay 0.0  --bias-weight-decay 0.0 --transformer-embedding-decay 0.0 --lr-scheduler cosineannealinglr --lr-min 0.00001 --lr-warmup-method linear  --lr-warmup-epochs 20 --lr-warmup-decay 0.01 --amp --label-smoothing 0.1 --mixup-alpha 0.8 --clip-grad-norm 5.0 --cutmix-alpha 1.0 --random-erase 0.25 --interpolation bicubic --auto-augment ta_wide --model-ema --ra-sampler --ra-reps 4  --val-resize-size 256 --val-crop-size 256 --train-crop-size 256 
```
Here `$MODEL` is one of `swin_v2_t`, `swin_v2_s` or `swin_v2_b`.
Note that `--val-resize-size` was optimized in a post-training step, see their `Weights` entry for the exact value.


### MaxViT
```
torchrun --nproc_per_node=8 --n_nodes=4 train.py\
--model $MODEL --epochs 400 --batch-size 128 --opt adamw --lr 3e-3 --weight-decay 0.05 --lr-scheduler cosineannealinglr --lr-min 1e-5 --lr-warmup-method linear  --lr-warmup-epochs 32  --label-smoothing 0.1 --mixup-alpha 0.8 --clip-grad-norm 1.0 --interpolation bicubic --auto-augment ta_wide --policy-magnitude 15 --model-ema --val-resize-size 224\
--val-crop-size 224 --train-crop-size 224 --amp  --model-ema-steps 32 --transformer-embedding-decay 0 --sync-bn
```
Here `$MODEL` is `maxvit_t`.
Note that `--val-resize-size` was not optimized in a post-training step.


### ShuffleNet V2
```
torchrun --nproc_per_node=8 train.py \
--batch-size=128 \
--lr=0.5 --lr-scheduler=cosineannealinglr --lr-warmup-epochs=5 --lr-warmup-method=linear \
--auto-augment=ta_wide --epochs=600 --random-erase=0.1 --weight-decay=0.00002 \
--norm-weight-decay=0.0 --label-smoothing=0.1 --mixup-alpha=0.2 --cutmix-alpha=1.0 \
--train-crop-size=176 --model-ema --val-resize-size=232 --ra-sampler --ra-reps=4
```
Here `$MODEL` is either `shufflenet_v2_x1_5` or `shufflenet_v2_x2_0`.

The models `shufflenet_v2_x0_5` and `shufflenet_v2_x1_0` were contributed by the community. See [PR-849](https://github.com/pytorch/vision/pull/849#issuecomment-483391686) for details.


## Mixed precision training
Automatic Mixed Precision (AMP) training on GPU for Pytorch can be enabled with the [torch.cuda.amp](https://pytorch.org/docs/stable/amp.html?highlight=amp#module-torch.cuda.amp).

Mixed precision training makes use of both FP32 and FP16 precisions where appropriate. FP16 operations can leverage the Tensor cores on NVIDIA GPUs (Volta, Turing or newer architectures) for improved throughput, generally without loss in model accuracy. Mixed precision training also often allows larger batch sizes. GPU automatic mixed precision training for Pytorch Vision can be enabled via the flag value `--amp=True`.

```
torchrun --nproc_per_node=8 train.py\
    --model resnext50_32x4d --epochs 100 --amp
```

## Quantized

### Post training quantized models

For all post training quantized models, the settings are:

1. num_calibration_batches: 32
2. num_workers: 16
3. batch_size: 32
4. eval_batch_size: 128
5. qbackend: 'fbgemm'

```
python train_quantization.py --device='cpu' --post-training-quantize --qbackend='fbgemm' --model='$MODEL'
```
Here `$MODEL` is one of `googlenet`, `inception_v3`, `resnet18`, `resnet50`, `resnext101_32x8d`, `shufflenet_v2_x0_5` and `shufflenet_v2_x1_0`.

### Quantized ShuffleNet V2

Here are commands that we use to quantize the `shufflenet_v2_x1_5` and `shufflenet_v2_x2_0` models.
```
# For shufflenet_v2_x1_5
python train_quantization.py --device='cpu' --post-training-quantize --qbackend='fbgemm' \
    --model=shufflenet_v2_x1_5 --weights="ShuffleNet_V2_X1_5_Weights.IMAGENET1K_V1" \
    --train-crop-size 176 --val-resize-size 232 --data-path /datasets01_ontap/imagenet_full_size/061417/

# For shufflenet_v2_x2_0
python train_quantization.py --device='cpu' --post-training-quantize --qbackend='fbgemm' \
    --model=shufflenet_v2_x2_0 --weights="ShuffleNet_V2_X2_0_Weights.IMAGENET1K_V1" \
    --train-crop-size 176 --val-resize-size 232 --data-path /datasets01_ontap/imagenet_full_size/061417/
```

### QAT MobileNetV2

For Mobilenet-v2, the model was trained with quantization aware training, the settings used are:
1. num_workers: 16
2. batch_size: 32
3. eval_batch_size: 128
4. qbackend: 'qnnpack'
5. learning-rate: 0.0001
6. num_epochs: 90
7. num_observer_update_epochs:4
8. num_batch_norm_update_epochs:3
9. momentum: 0.9
10. lr_step_size:30
11. lr_gamma: 0.1
12. weight-decay: 0.0001

```
torchrun --nproc_per_node=8 train_quantization.py --model='mobilenet_v2'
```

Training converges at about 10 epochs.

### QAT MobileNetV3

For Mobilenet-v3 Large, the model was trained with quantization aware training, the settings used are:
1. num_workers: 16
2. batch_size: 32
3. eval_batch_size: 128
4. qbackend: 'qnnpack'
5. learning-rate: 0.001
6. num_epochs: 90
7. num_observer_update_epochs:4
8. num_batch_norm_update_epochs:3
9. momentum: 0.9
10. lr_step_size:30
11. lr_gamma: 0.1
12. weight-decay: 0.00001

```
torchrun --nproc_per_node=8 train_quantization.py --model='mobilenet_v3_large' \
    --wd 0.00001 --lr 0.001
```

For post training quant, device is set to CPU. For training, the device is set to CUDA.

### Command to evaluate quantized models using the pre-trained weights:

```
python train_quantization.py --device='cpu' --test-only --qbackend='<qbackend>' --model='<model_name>'
```

For inception_v3 you need to pass the following extra parameters:
```
--val-resize-size 342 --val-crop-size 299 --train-crop-size 299
```

# Stereo Matching reference training scripts

This folder contains reference training scripts for Stereo Matching.
They serve as a log of how to train specific models, so as to provide baseline
training and evaluation scripts to quickly bootstrap research.


### CREStereo

The CREStereo model was trained on a dataset mixture between **CREStereo**, **ETH3D** and the additional split from **Middlebury2014**.
A ratio of **88-6-6** was used in order to train a baseline weight set. We provide multi-set variant as well.
Both used 8 A100 GPUs and a batch size of 2 (so effective batch size is 16). The
rest of the hyper-parameters loosely follow the recipe from https://github.com/megvii-research/CREStereo.
The original recipe trains for **300000** updates (or steps) on the dataset mixture. We modify the learning rate
schedule to one that starts decaying the weight much sooner. Throughout the experiments we found that this reduces 
overfitting during evaluation time and gradient clip help stabilize the loss during a pre-mature learning rate change.

```
torchrun --nproc_per_node 8 --nnodes 1 train.py \
    --dataset-root $dataset_root \
    --name $name_cre \
    --model crestereo_base \
    --train-datasets crestereo eth3d-train middlebury2014-other \
    --dataset-steps 264000 18000 18000
    --batch-size 2 \
    --lr 0.0004 \
    --min-lr 0.00002 \
    --lr-decay-method cosine \
    --warmup-steps 6000 \
    --decay-after-steps 30000 \
    --clip-grad-norm 1.0 \
```

We employ a multi-set fine-tuning stage where we uniformly sample from multiple datasets. Given hat some of these datasets have extremely large images (``2048x2048`` or more) we opt for a very aggressive scale-range ``[0.2 - 0.8]`` such that as much of the original frame composition is captured inside the ``384x512`` crop.

```
torchrun --nproc_per_node 8 --nnodes 1 train.py \
    --dataset-root $dataset_root \
    --name $name_things \
    --model crestereo_base \
    --train-datasets crestereo eth3d-train middlebury2014-other instereo2k fallingthings carla-highres sintel sceneflow-monkaa sceneflow-driving \
    --dataset-steps 12000 12000 12000 12000 12000 12000 12000 12000 12000
    --batch-size 2 \
    --scale-range 0.2 0.8 \
    --lr 0.0004 \
    --lr-decay-method cosine \
    --decay-after-steps 0 \
    --warmup-steps 0 \
    --min-lr 0.00002 \
    --resume-path $checkpoint_dir/$name_cre.pth
```


### Evaluation

Evaluating the base weights

```
torchrun --nproc_per_node 1 --nnodes 1 cascade_evaluation.py --dataset middlebury2014-train --batch-size 1 --dataset-root $dataset_root --model crestereo_base --weights CREStereo_Base_Weights.CRESTEREO_ETH_MBL_V1
```

This should give an **mae of about 1.416** on the train set of `Middlebury2014`. Results may vary slightly depending on the batch size and the number of GPUs. For the most accurate results use 1 GPU and `--batch-size 1`. The created log file should look like this, where the first key is the number of cascades and the nested key is the number of recursive iterations:

```
Dataset: middlebury2014-train @size: [384, 512]:
{
	1: {
		2: {'mae': 2.363, 'rmse': 4.352, '1px': 0.611, '3px': 0.828, '5px': 0.891, 'relepe': 0.176, 'fl-all': 64.511}
		5: {'mae': 1.618, 'rmse': 3.71, '1px': 0.761, '3px': 0.879, '5px': 0.918, 'relepe': 0.154, 'fl-all': 77.128}
		10: {'mae': 1.416, 'rmse': 3.53, '1px': 0.777, '3px': 0.896, '5px': 0.933, 'relepe': 0.148, 'fl-all': 78.388}
		20: {'mae': 1.448, 'rmse': 3.583, '1px': 0.771, '3px': 0.893, '5px': 0.931, 'relepe': 0.145, 'fl-all': 77.7}
	},
}
{
	2: {
		2: {'mae': 1.972, 'rmse': 4.125, '1px': 0.73, '3px': 0.865, '5px': 0.908, 'relepe': 0.169, 'fl-all': 74.396}
		5: {'mae': 1.403, 'rmse': 3.448, '1px': 0.793, '3px': 0.905, '5px': 0.937, 'relepe': 0.151, 'fl-all': 80.186}
		10: {'mae': 1.312, 'rmse': 3.368, '1px': 0.799, '3px': 0.912, '5px': 0.943, 'relepe': 0.148, 'fl-all': 80.379}
		20: {'mae': 1.376, 'rmse': 3.542, '1px': 0.796, '3px': 0.91, '5px': 0.942, 'relepe': 0.149, 'fl-all': 80.054}
	},
}
```

You can also evaluate the Finetuned weights:

```
torchrun --nproc_per_node 1 --nnodes 1 cascade_evaluation.py --dataset middlebury2014-train --batch-size 1 --dataset-root $dataset_root --model crestereo_base --weights CREStereo_Base_Weights.CRESTEREO_FINETUNE_MULTI_V1
```

```
Dataset: middlebury2014-train @size: [384, 512]:
{
	1: {
		2: {'mae': 1.85, 'rmse': 3.797, '1px': 0.673, '3px': 0.862, '5px': 0.917, 'relepe': 0.171, 'fl-all': 69.736}
		5: {'mae': 1.111, 'rmse': 3.166, '1px': 0.838, '3px': 0.93, '5px': 0.957, 'relepe': 0.134, 'fl-all': 84.596}
		10: {'mae': 1.02, 'rmse': 3.073, '1px': 0.854, '3px': 0.938, '5px': 0.96, 'relepe': 0.129, 'fl-all': 86.042}
		20: {'mae': 0.993, 'rmse': 3.059, '1px': 0.855, '3px': 0.942, '5px': 0.967, 'relepe': 0.126, 'fl-all': 85.784}
	},
}
{
	2: {
		2: {'mae': 1.667, 'rmse': 3.867, '1px': 0.78, '3px': 0.891, '5px': 0.922, 'relepe': 0.165, 'fl-all': 78.89}
		5: {'mae': 1.158, 'rmse': 3.278, '1px': 0.843, '3px': 0.926, '5px': 0.955, 'relepe': 0.135, 'fl-all': 84.556}
		10: {'mae': 1.046, 'rmse': 3.13, '1px': 0.85, '3px': 0.934, '5px': 0.96, 'relepe': 0.13, 'fl-all': 85.464}
		20: {'mae': 1.021, 'rmse': 3.102, '1px': 0.85, '3px': 0.935, '5px': 0.963, 'relepe': 0.129, 'fl-all': 85.417}
	},
}
```

Evaluating the author provided weights:

```
torchrun --nproc_per_node 1 --nnodes 1 cascade_evaluation.py --dataset middlebury2014-train --batch-size 1 --dataset-root $dataset_root --model crestereo_base --weights CREStereo_Base_Weights.MEGVII_V1
```

```
Dataset: middlebury2014-train @size: [384, 512]:
{
	1: {
		2: {'mae': 1.704, 'rmse': 3.738, '1px': 0.738, '3px': 0.896, '5px': 0.933, 'relepe': 0.157, 'fl-all': 76.464}
		5: {'mae': 0.956, 'rmse': 2.963, '1px': 0.88, '3px': 0.948, '5px': 0.965, 'relepe': 0.124, 'fl-all': 88.186}
		10: {'mae': 0.792, 'rmse': 2.765, '1px': 0.905, '3px': 0.958, '5px': 0.97, 'relepe': 0.114, 'fl-all': 90.429}
		20: {'mae': 0.749, 'rmse': 2.706, '1px': 0.907, '3px': 0.961, '5px': 0.972, 'relepe': 0.113, 'fl-all': 90.807}
	},
}
{
	2: {
		2: {'mae': 1.702, 'rmse': 3.784, '1px': 0.784, '3px': 0.894, '5px': 0.924, 'relepe': 0.172, 'fl-all': 80.313}
		5: {'mae': 0.932, 'rmse': 2.907, '1px': 0.877, '3px': 0.944, '5px': 0.963, 'relepe': 0.125, 'fl-all': 87.979}
		10: {'mae': 0.773, 'rmse': 2.768, '1px': 0.901, '3px': 0.958, '5px': 0.972, 'relepe': 0.117, 'fl-all': 90.43}
		20: {'mae': 0.854, 'rmse': 2.971, '1px': 0.9, '3px': 0.957, '5px': 0.97, 'relepe': 0.122, 'fl-all': 90.269}
	},
}
```

# Concerns when training

We encourage users to be aware of the **aspect-ratio** and **disparity scale** they are targeting when doing any sort of training or fine-tuning. The model is highly sensitive to these two factors, as a consequence of naive multi-set fine-tuning one can achieve `0.2 mae` relatively fast. We recommend that users pay close attention to how they **balance dataset sizing** when training such networks.

 Ideally, dataset scaling should be trated at an individual level and a thorough **EDA** of the disparity distribution in random crops at the desired training / inference size should be performed prior to any large compute investments.

### Disparity scaling

##### Sample A
 The top row contains a sample from `Sintel` whereas the bottom row one from `Middlebury`.

![Disparity1](assets/disparity-domain-drift.jpg)

From left to right (`left_image`, `right_image`, `valid_mask`, `valid_mask & ground_truth`, `prediction`). **Darker is further away, lighter is closer**. In the case of `Sintel` which is more closely aligned to the original distribution of `CREStereo` we notice that the model accurately predicts the background scale whereas in the case of `Middlebury2014` it cannot correctly estimate the continuous disparity. Notice that the frame composition is similar for both examples. The blue skybox in the `Sintel` scene behaves similarly to the `Middlebury` black background. However, because the `Middlebury` samples comes from an extremely large scene the crop size of `384x512` does not correctly capture the general training distribution.




##### Sample B

The top row contains a scene from `Sceneflow` using the `Monkaa` split whilst the bottom row is a scene from `Middlebury`. This sample exhibits the same issues when it comes to **background estimation**. Given the exaggerated size of the `Middlebury` samples the model **colapses the smooth background** of the sample to what it considers to be a mean background disparity value.

![Disparity2](assets/disparity-background-mode-collapse.jpg)


For more detail on why this behaviour occurs based on the training distribution proportions you can read more about the network at: https://github.com/pytorch/vision/pull/6629#discussion_r978160493


### Metric overfitting

##### Learning is critical in the beginning

We also advise users to make user of faster training schedules, as the performance gain over long periods time is marginal. Here we exhibit a difference between a faster decay schedule and later decay schedule.

![Loss1](assets/Loss.jpg)

In **grey** we set the lr decay to begin after `30000` steps whilst in **orange** we opt for a very late learning rate decay at around `180000` steps. Although exhibiting stronger variance, we can notice that unfreezing the learning rate earlier whilst employing `gradient-norm` out-performs the default configuration.

##### Gradient norm saves time

![Loss2](assets/gradient-norm-removal.jpg)

In **grey** we keep ``gradient norm`` enabled whilst in **orange** we do not. We can notice that remvoing the gradient norm exacerbates the performance decrease in the early stages whilst also showcasing an almost complete collapse around the `60000` steps mark where we started decaying the lr for **orange**.

Although both runs ahieve an improvement of about ``0.1`` mae after the lr decay start, the benefits of it are observable much faster when ``gradient norm`` is employed as the recovery period is no longer accounted for.

Using torchvision models in C++
===============================

This is a minimal example of getting TorchVision models to work in C++ with
Torchscript. The model is first scripted in Python and exported to a file, and
then loaded in C++. For a similar tutorial, see [this
tutorial](https://pytorch.org/tutorials/advanced/cpp_export.html).

In order to successfully compile this example, make sure you have ``LibTorch``
installed. You can either:

- Install PyTorch normally
- Or download the LibTorch C++ distribution.

In both cases refer [here](https://pytorch.org/get-started/locally/) the
corresponding install or download instructions.

Some torchvision models only depend on PyTorch operators, and can be used in C++
without depending on the torchvision lib. Other models rely on torchvision's C++
operators like NMS, RoiAlign (typically the detection models) and those need to
be linked against the torchvision lib.

We'll first see the simpler case of running a model without the torchvision lib
dependency.

Running a model that doesn't need torchvision lib
-------------------------------------------------

Create a ``build`` directory inside the current one.

```bash
mkdir build
cd build
```

Then run `python ../trace_model.py` which should create a `resnet18.pt` file in
the build directory. This is the scripted model that will be used in the C++
code.

We can now start building with CMake. We have to tell CMake where it can find
the necessary PyTorch resources. If you installed PyTorch normally, you can do:

```bash
TORCH_PATH=$(python -c "import pathlib, torch; print(pathlib.Path(torch.__path__[0]))")
Torch_DIR="${TORCH_PATH}/share/cmake/Torch"   # there should be .cmake files in there

cmake .. -DTorch_DIR=$Torch_DIR
```

If instead you downloaded the LibTorch somewhere, you can do:

```bash
cmake .. -DCMAKE_PREFIX_PATH=/path/to/libtorch
```

Then `cmake --build .` and you should now be able to run

```bash
./run_model resnet18.pt
```

If you try to run the model with a model that depends on the torchvision lib, like
`./run_model fasterrcnn_resnet50_fpn.pt`, you should get a runtime error. This is
because the executable wasn't linked against the torchvision lib.


Running a model that needs torchvision lib
------------------------------------------

First, we need to build the torchvision lib. To build the torchvision lib go to
the root of the torchvision project and run:

```bash
mkdir build
cd build
cmake .. -DCMAKE_PREFIX_PATH=/path/to/libtorch  # or -DTorch_DIR= if you installed PyTorch normally, see above
cmake --build .
cmake --install .
```

You may want to pass `-DCMAKE_INSTALL_PREFIX=/path/to/libtorchvision` for
cmake to copy/install the files to a specific location (e.g. `$CONDA_PREFIX`).

**DISCLAIMER**: the `libtorchvision` library includes the torchvision
custom ops as well as most of the C++ torchvision APIs. Those APIs do not come
with any backward-compatibility guarantees and may change from one version to
the next. Only the Python APIs are stable and with backward-compatibility
guarantees. So, if you need stability within a C++ environment, your best bet is
to export the Python APIs via torchscript.

Now that libtorchvision is built and installed we can tell our project to use
and link to it via the `-DUSE_TORCHVISION` flag. We also need to tell CMake
where to find it, just like we did with LibTorch, e.g.:

```bash
cmake .. -DTorch_DIR=$Torch_DIR -DTorchVision_DIR=path/to/libtorchvision -DUSE_TORCHVISION=ON
cmake --build .
```

Now the `run_model` executable should be able to run the
`fasterrcnn_resnet50_fpn.pt` file.

Video meta-information Notation

Video File Name
  video: codec, fps
  audio: codec, bits per sample, sample rate

Test videos are listed below.
--------------------------------

- RATRACE_wave_f_nm_np1_fr_goo_37.avi
  - source: hmdb51
  - video: DivX MPEG-4
    - fps: 30
  - audio: N/A

- SchoolRulesHowTheyHelpUs_wave_f_nm_np1_ba_med_0.avi
  - source: hmdb51
  - video: DivX MPEG-4
    - fps: 30
  - audio: N/A

- TrumanShow_wave_f_nm_np1_fr_med_26.avi
  - source: hmdb51
  - video: DivX MPEG-4
    - fps: 30
  - audio: N/A

- v_SoccerJuggling_g23_c01.avi
  - source: ucf101
  - video: Xvid MPEG-4
    - fps: 29.97
  - audio: N/A

- v_SoccerJuggling_g24_c01.avi
  - source: ucf101
  - video: Xvid MPEG-4
    - fps: 29.97
  - audio: N/A

- R6llTwEh07w.mp4
  - source: kinetics-400
  - video: H-264 - MPEG-4 AVC (part 10) (avc1)
    - fps: 30
  - audio: MPEG AAC audio (mp4a)
    - sample rate: 44.1K Hz

- SOX5yA1l24A.mp4
  - source: kinetics-400
  - video: H-264 - MPEG-4 AVC (part 10) (avc1)
    - fps: 29.97
  - audio: MPEG AAC audio (mp4a)
    - sample rate: 48K Hz

- WUzgd7C1pWA.mp4
  - source: kinetics-400
  - video: H-264 - MPEG-4 AVC (part 10) (avc1)
    - fps: 29.97
  - audio: MPEG AAC audio (mp4a)
    - sample rate: 48K Hz

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# library(torchvision)
# 
# # Search for specific topics
# search_rf100("cell")        # Find cell-related datasets
# search_rf100("solar")       # Find solar panel datasets
# search_rf100("x-ray")       # Find X-ray datasets
# 
# # List all datasets in a collection
# search_rf100(collection = "biology")
# search_rf100(collection = "medical")
# 
# # View complete catalog
# catalog <- get_rf100_catalog()
# View(catalog)

## ----eval=FALSE---------------------------------------------------------------
# # Search for solar/photovoltaic datasets
# search_rf100("solar")
# search_rf100("photovoltaic")
# 
# # Result shows:
# # - solar_panel in infrared collection
# # - solar_panel in damage collection

## ----eval=FALSE---------------------------------------------------------------
# library(torchvision)
# library(knitr)
# 
# catalog <- get_rf100_catalog()
# 
# # Display key columns
# kable(catalog[, c("collection", "dataset", "description", "total_size_mb", "estimated_images")])

## ----eval=FALSE---------------------------------------------------------------
# search_rf100(collection = "biology")

## ----eval=FALSE---------------------------------------------------------------
# search_rf100(collection = "medical")

## ----eval=FALSE---------------------------------------------------------------
# search_rf100(collection = "infrared")

## ----eval=FALSE---------------------------------------------------------------
# search_rf100(collection = "damage")

## ----eval=FALSE---------------------------------------------------------------
# search_rf100(collection = "underwater")

## ----eval=FALSE---------------------------------------------------------------
# search_rf100(collection = "document")

## ----eval=FALSE---------------------------------------------------------------
# library(torchvision)
# 
# # Search for blood cell dataset
# search_rf100("blood")
# 
# # Load the dataset
# ds <- rf100_biology_collection(
#   dataset = "blood_cell",
#   split = "train",
#   download = TRUE
# )
# 
# # Inspect a sample
# item <- ds[1]
# print(item$y$labels)  # Object classes
# print(item$y$boxes)   # Bounding boxes
# 
# # Visualize with bounding boxes
# boxed <- draw_bounding_boxes(item)
# tensor_image_browse(boxed)

## ----eval=FALSE---------------------------------------------------------------
# catalog <- get_rf100_catalog()
# 
# # Total size of all datasets
# sum(catalog$total_size_mb) / 1024  # In GB
# 
# # Datasets by size
# catalog[order(-catalog$total_size_mb), c("dataset", "collection", "total_size_mb")]
# 
# # Smallest and largest datasets
# catalog[which.min(catalog$total_size_mb), ]
# catalog[which.max(catalog$total_size_mb), ]
# 
# # Average size by collection
# aggregate(total_size_mb ~ collection, data = catalog, FUN = mean)

## ----eval=FALSE---------------------------------------------------------------
# # Find small datasets (< 20 MB total)
# subset(catalog, total_size_mb < 20)
# 
# # Find large datasets (> 200 MB total)
# subset(catalog, total_size_mb > 200)
# 
# # Find datasets with specific keywords
# subset(catalog, grepl("tumor|cancer|disease", description, ignore.case = TRUE))
# 
# # Datasets with all three splits
# subset(catalog, has_train & has_test & has_valid)

#' Caltech Datasets
#'
#' The Caltech-101 and Caltech-256 collections are **classification** datasets
#' made of color images with varying sizes. They cover 101 and 256 object
#' categories respectively and are commonly used for evaluating visual
#' recognition models.
#'
#' @inheritParams fgvc_aircraft_dataset
#' @param root Character. Root directory for dataset storage. The dataset will be stored under `root/caltech-101`.
#'
#' @return An object of class \code{caltech101_dataset}, which behaves like a torch dataset.
#' Each element is a named list with:
#' - `x`: A H x W x 3 integer array representing an RGB image.
#' - `y`: An Integer representing the label.
#'
#' @details
#' The Caltech-101 dataset contains around 9,000 images
#' spread over 101 object categories plus a background class. Images
#' have varying sizes.
#'
#' Caltech-256 extends this to about 30,000 images
#' across 256 categories.
#'
#' @examples
#' \dontrun{
#' caltech101 <- caltech101_dataset(download = TRUE)
#'
#' first_item <- caltech101[1]
#' first_item$x  # Image array
#' first_item$y  # Integer label
#' }
#'
#' @name caltech_dataset
#' @title Caltech Datasets
#' @rdname caltech_dataset
#' @family classification_dataset
#' @export
caltech101_dataset <- torch::dataset(
  name = "caltech-101",
  subname = "101_ObjectCategories",
  archive_size = "130 MB",
  resources = list(
    list(
      url = "https://data.caltech.edu/records/mzrjq-6wc02/files/caltech-101.zip",
      filename = "caltech-101.zip",
      md5 = "3138e1922a9193bfa496528edbbc45d0"
    )
  ),

  initialize = function(
    root = tempdir(),
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
    self$root <- fs::path(root, class(self)[[1]])
    self$transform <- transform
    self$target_transform <- target_transform

    if (download){
      cli_inform("Dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")

    obj_dir <- fs::path(self$root, class(self)[[1]], self$subname)
    all_dirs <- fs::dir_ls(obj_dir, type = "directory")
    self$classes <- sort(base::basename(all_dirs))
    self$classes <- self$classes[self$classes != "BACKGROUND_Google"]

    name_map <- list("Faces" = "Faces_2", "Faces_easy" = "Faces_3", "Motorbikes" = "Motorbikes_16", "airplanes" = "Airplanes_Side_2")
    self$annotation_classes <- vapply(self$classes, function(x) if (x %in% names(name_map)) name_map[[x]] else x, character(1))

    for (i in seq_along(self$classes)) {
      img_dir <- fs::path(obj_dir, self$classes[[i]])
      imgs <- sort(fs::dir_ls(img_dir, glob = "*.jpg"))
      self$img_path <- append(self$img_path, imgs)
      self$labels <- c(self$labels, rep(i, length(imgs)))
      self$image_indices <- c(self$image_indices, seq_along(imgs))
    }

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  .getitem = function(index) {
    img_path <- self$img_path[[index]]
    y <- self$labels[[index]]

    x <- jpeg::readJPEG(img_path)

    if (!is.null(self$transform))
      x <- self$transform(x)

    if (!is.null(self$target_transform))
      y <- self$target_transform(y)

    list(x = x, y = y)
  },

  .length = function() {
    length(self$img_path)
  },

  check_exists = function() {
    fs::dir_exists(fs::path(self$root, class(self)[[1]], self$subname))
  },

  download = function() {

    if (self$check_exists())
      return()

    fs::dir_create(self$root)

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    invisible(lapply(self$resources, function(res) {
      archive <- download_and_cache(res$url, prefix = class(self)[1])
      dest <- fs::path(self$root, fs::path_file(res$filename))
      fs::file_move(archive, dest)
      md5 <- tools::md5sum(dest)[[1]]
      if (md5 != res$md5)
        runtime_error("Corrupt file! Delete the file in {archive} and try again.")
      if(class(self)[1] == "caltech-101")
        utils::unzip(dest, exdir = self$root)
      else
        utils::untar(dest, exdir = self$root)

      extracted <- fs::path(self$root, class(self)[[1]])
      if (fs::file_exists(fs::path(extracted, "101_ObjectCategories.tar.gz")))
        utils::untar(fs::path(extracted, "101_ObjectCategories.tar.gz"), exdir = extracted)
    }))

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  }
)

#' Caltech-256 Dataset
#'
#' Loads the Caltech-256 Object Category Dataset for image classification. It consists of 30,607 images across 256 distinct object categories.
#' Each category has at least 80 images, with variability in image size.
#'
#' @inheritParams fgvc_aircraft_dataset
#' @param root Character. Root directory for dataset storage. The dataset will be stored under `root/caltech256`.
#'
#' @return An object of class \code{caltech256_dataset}, which behaves like a torch dataset.
#' Each element is a named list with:
#' - `x`: A H x W x 3 integer array representing an RGB image.
#' - `y`: An Integer representing the label.
#'
#' @rdname caltech_dataset
#' @export
caltech256_dataset <- torch::dataset(
  name = "caltech256",
  subname = "256_ObjectCategories",
  inherit = caltech101_dataset,
  classes = NULL,
  archive_size = "1.1 GB",
  resources = list(
    list(
      url = "https://data.caltech.edu/records/nyy15-4j048/files/256_ObjectCategories.tar",
      filename = "256_ObjectCategories.tar",
      md5 = "67b4f42ca05d46448c6bb8ecd2220f6d"
    )
  ),

  initialize = function(
    root = tempdir(),
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
  self$root <- fs::path(root, class(self)[[1]])
  self$transform <- transform
  self$target_transform <- target_transform



  if (download) {
    cli_inform("Dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already cached.")
    self$download()
  }

  if (!self$check_exists()) {
    cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
  }

  obj_dir <- fs::path(self$root, self$subname)
  all_dirs <- fs::dir_ls(obj_dir, type = "directory")
  self$classes <- sort(base::basename(all_dirs))
  self$classes <- self$classes[self$classes != "BACKGROUND_Google"]

  class_dirs <- fs::path(self$root, self$subname, self$classes)
  self$classes <- sub("^\\d+\\.", "", self$classes)
  images_per_class <- lapply(class_dirs, function(class_dir) {
    imgs <- fs::dir_ls(class_dir, glob = "*.jpg")
    sort(imgs)
  })
  self$img_path <- unlist(images_per_class, use.names = FALSE)
  self$labels <- unlist(
    mapply(function(i, imgs) {
      rep(i, length(imgs))
    }, seq_along(self$classes), images_per_class, SIMPLIFY = FALSE),
    use.names = FALSE
  )
  cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  check_exists = function() {
    fs::dir_exists(fs::path(self$root, self$subname))
  }
)

#' Caltech Class Labels
#'
#' Utilities for resolving Caltech class identifiers to their corresponding
#' human readable labels.
#'
#' @param class_id Integer vector of 1-based class identifiers.
#' @return A character vector with 257 entries representing the Caltech 257
#'   class labels.
#' @family class_resolution
#' @importFrom utils read.csv
#' @export
caltech_classes <- function(class_id = 1:257) {
  if (any(class_id > 257)) {
    cli_warn("Caltech {.var class_id} cannot be > 257")
  }
  url <- download_and_cache("https://raw.githubusercontent.com/vkinakh/stable-diffusion-imagenet/refs/heads/main/caltech256_classes.txt")
  labels_df <- read.csv(url, sep = ".", header = FALSE, col.names = c("id", "label"))
  labels_df[class_id,]$label
}

context("dataset-rf100-doc")

test_that("rf100_document_collection handles missing dataset gracefully", {
  expect_error(
    rf100_document_collection(dataset = "paragraph", split = "train", download = FALSE),
    "'arg' should be one of"
  )
})

# small datasets
dataset <- data.frame(name = c("tweeter_post", "tweeter_profile", "document_part",
                                "activity_diagram", "signature", "currency"),
                       num_classes = c(2L, 1L, 2L, 19L, 1L, 10L)
)

for (ds_name in dataset$name) {
  test_that(paste0("rf100_document_collection loads ", ds_name, " correctly"), {
    ds <- rf100_document_collection(dataset = ds_name, split = "train", download = TRUE)

    expect_s3_class(ds, "rf100_document_collection")
    expect_gt(ds$.length(), 1)
    expect_type(ds$classes, "character")
    expect_length(unique(ds$classes), dataset[dataset$name == ds_name,]$num_classes)

    item <- ds[1]

    expect_type(item$y, "list")
    expect_named(item$y, c("image_id","labels","boxes"))
    expect_type(item$y$labels, "integer")
    expect_tensor(item$y$boxes)
    expect_equal(item$y$boxes$ndim, 2)
    expect_equal(item$y$boxes$size(2), 4)
    expect_s3_class(item, "image_with_bounding_box")
  })
}

dataset <- data.frame(name = c("paper_part", "wine_label"),
                      num_classes = c(19L, 12L)
)

for (ds_name in dataset$name) {
  test_that(paste0("rf100_document_collection loads ", ds_name, " correctly"), {
    skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")
    ds <- rf100_document_collection(dataset = ds_name, split = "test", download = TRUE)

    expect_s3_class(ds, "rf100_document_collection")
    expect_gt(ds$.length(), 1)
    expect_type(ds$classes, "character")
    expect_length(unique(ds$classes), dataset[dataset$name == ds_name,]$num_classes)

    item <- ds[1]

    expect_type(item$y, "list")
    expect_named(item$y, c("image_id","labels","boxes"))
    expect_type(item$y$labels, "integer")
    expect_tensor(item$y$boxes)
    expect_equal(item$y$boxes$ndim, 2)
    expect_equal(item$y$boxes$size(2), 4)
    expect_s3_class(item, "image_with_bounding_box")
  })
}


test_that("rf100_document_collection datasets can be turned into a dataloader wo transform", {
  ds <- rf100_document_collection(dataset = "document_part", split = "test", download = TRUE)

  expect_equal(ds$.length(), 181)
  expect_type(ds$classes, "character")
  expect_length(unique(ds$classes), 2)

  items <- ds[7:9]
  # Check shape, dtype, and values on X
  expect_named(items, c("x", "y"))
  expect_length(items$x, 3)
  expect_equal(dim(items$x[[1]]), c(640, 640, 3))
  expect_lte(max(items$x[[1]]), 1)
  # Check shape, dtype and names on y
  expect_named(items$y,c("image_id","labels","boxes"))
  expect_length(items$y$image_id,6)
  expect_tensor(items$y$boxes)
  expect_tensor_shape(items$y$boxes, c(6,4))


  dl <- dataloader(ds, batch_size = 10, shuffle = TRUE)
  # 181 turns into 19 batches of 10
  expect_length(dl, 19)
  iter <- dataloader_make_iter(dl)
  expect_no_error(
    i <- dataloader_next(iter)
  )
  # Check shape, dtype, and values on X
  expect_named(i, c("x", "y"))
  expect_length(i$x, 10)
  expect_equal(dim(i$x[[1]]), c(640, 640, 3))
  expect_lte(max(i$x[[1]])$item(), 1)
  # Check shape, dtype and names on y
  expect_named(i$y, c("image_id","labels","boxes"))
  expect_tensor(i$y$image_id)
  expect_length(unique(as_array(i$y$image_id)), 10)
  expect_tensor(i$y$labels)
  expect_tensor(i$y$boxes)
  N <- i$y$boxes$shape[1]
  expect_gte(N, 10)
  box_points <- i$y$boxes$shape[2]
  expect_equal(box_points, 4)

})

test_that("rf100_document_collection datasets can be turned into a dataloader with transform", {
  ds <- rf100_document_collection(dataset = "document_part", split = "test",
                                  download = TRUE, transform = transform_to_tensor)

  items <- ds[2:4]
  expect_tensor(items$x)
  expect_tensor_shape(items$x, c(3,3,640, 640))
  expect_tensor_dtype(items$x, torch_float())
  # Check shape, dtype and names on y
  expect_named(items$y,c("image_id","labels","boxes"))
  expect_length(items$y$image_id,6)
  expect_tensor(items$y$boxes)
  expect_tensor_shape(items$y$boxes, c(6,4))


  dl <- dataloader(ds, batch_size = 10, shuffle = TRUE)
  # 181 turns into 19 batches of 10
  expect_length(dl, 19)
  iter <- dataloader_make_iter(dl)
  expect_no_error(
    i <- dataloader_next(iter)
  )
  expect_tensor(i$x)
  expect_tensor_shape(i$x, c(10, 3, 640, 640))
  # Check shape, dtype and names on y
  expect_named(i$y, c("image_id","labels","boxes"))
  expect_tensor(i$y$image_id)
  expect_length(unique(as_array(i$y$image_id)), 10)
  expect_tensor(i$y$labels)
  expect_tensor(i$y$boxes)
  N <- i$y$boxes$shape[1]
  expect_gte(N, 10)
  box_points <- i$y$boxes$shape[2]
  expect_equal(box_points, 4)

})

context("dataset-rf100-biology")

t <- withr::local_tempdir()

test_that("rf100_biology_collection handles missing files gracefully", {
  expect_error(
    rf100_biology_collection(dataset = "stomata_cell", split = "train", download = FALSE),
    class = "runtime_error"
  )
})

small_dataset <- data.frame(name = c("blood_cell",  "cell", "bacteria", "mitosis"),
                            num_classes = c(3L, 1L, 1L, 1L)
)

for (ds_name in small_dataset$name) {
  test_that(paste0("rf100_biology_collection loads ", ds_name, " correctly"), {
    ds <- rf100_biology_collection(dataset = ds_name, split = "train", download = TRUE)

    expect_s3_class(ds, "rf100_biology_collection")
    expect_gt(ds$.length(), 1)
    expect_type(ds$classes, "character")
    expect_length(unique(ds$classes),
                  small_dataset[small_dataset$name == ds_name,]$num_classes)

    item <- ds[1]

    expect_type(item$y, "list")
    expect_named(item$y, c("image_id","labels","boxes"))
    expect_type(item$y$labels, "integer")
    expect_tensor(item$y$boxes)
    expect_equal(item$y$boxes$ndim, 2)
    expect_equal(item$y$boxes$size(2), 4)
    expect_s3_class(item, "image_with_bounding_box")
  })
}

dataset <- data.frame(name = c("stomata_cell", "parasite", "cotton_disease", "phage", "liver_disease", "moth"),
                      num_classes = c(2L, 8L, 1L, 2L, 4L, 28L)
)

for (ds_name in dataset$name) {
  test_that(paste0("rf100_biology_collection loads ", ds_name, " correctly"), {
    skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
            "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")
    ds <- rf100_biology_collection(dataset = ds_name, split = "val", download = TRUE)

    expect_s3_class(ds, "rf100_biology_collection")
    expect_gt(ds$.length(), 1)
    expect_type(ds$classes, "character")
    expect_length(unique(ds$classes), dataset[dataset$name == ds_name,]$num_classes)

    item <- ds[1]

    expect_type(item$y, "list")
    expect_named(item$y, c("image_id","labels","boxes"))
    expect_type(item$y$labels, "integer")
    expect_tensor(item$y$boxes)
    expect_equal(item$y$boxes$ndim, 2)
    expect_equal(item$y$boxes$size(2), 4)
    expect_s3_class(item, "image_with_bounding_box")
  })
}

context("dataset-rf100-infrared")

t <- withr::local_tempdir()

test_that("rf100_infrared_collection handles missing files gracefully", {
  expect_error(
    rf100_infrared_collection(dataset = "thermal_dog_and_people", split = "train", download = FALSE),
    class = "runtime_error"
  )
})

dataset <- data.frame(name = c("thermal_dog_and_people", "solar_panel", "thermal_cheetah"),
                      num_classes = c(2L, 5L, 2L)
)

for (ds_name in dataset$name) {
  test_that(paste0("rf100_infrared_collection loads ", ds_name, " correctly"), {
    ds <- rf100_infrared_collection(dataset = ds_name, split = "train", download = TRUE)

    expect_s3_class(ds, "rf100_infrared_collection")
    expect_gt(ds$.length(), 1)
    expect_type(ds$classes, "character")
    expect_length(unique(ds$classes), dataset[dataset$name == ds_name,]$num_classes)

    item <- ds[2] # as 2 datasets have their first item wo bbox

    expect_type(item$y, "list")
    expect_named(item$y, c("image_id","labels","boxes"))
    expect_type(item$y$labels, "integer")
    expect_tensor(item$y$boxes)
    expect_equal(item$y$boxes$ndim, 2)
    expect_equal(item$y$boxes$size(2), 4)
    expect_s3_class(item, "image_with_bounding_box")
  })
}
test_that(paste0("rf100_infrared_collection loads 'ir_object' correctly"), {
    skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
            "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")
    ds <- rf100_infrared_collection(dataset ="ir_object", split = "train", download = TRUE)

    expect_s3_class(ds, "rf100_infrared_collection")
    expect_gt(ds$.length(), 1)
    expect_type(ds$classes, "character")
    expect_length(unique(ds$classes), 4)

    item <- ds[1]

    expect_type(item$y, "list")
    expect_named(item$y, c("image_id","labels","boxes"))
    expect_type(item$y$labels, "integer")
    expect_tensor(item$y$boxes)
    expect_equal(item$y$boxes$ndim, 2)
    expect_equal(item$y$boxes$size(2), 4)
    expect_s3_class(item, "image_with_bounding_box")
  })

#' Places365 Dataset
#'
#' Loads the MIT Places365 dataset for scene classification.
#'
#' The dataset provides three splits: `"train"`, `"val"`, and `"test"`.
#' Folder structure and image layout on disk are handled internally by the loader.
#'
#' This function downloads and prepares the smaller 256x256 image version (~30 GB).
#' For the high-resolution variant (~160 GB), use [places365_dataset_large()].
#' Note that images in the large version come in varying sizes, so resizing may be
#' needed before batching.
#'
#' @inheritParams mnist_dataset
#' @param split One of `"train"`, `"val"`, or `"test"`.
#' @param loader A function to load an image given its path. Defaults to
#'   [magick_loader()], which uses the `{magick}` package.
#'
#' @return A torch dataset of class \code{places365_dataset}. Each element is a named
#' list with:
#' - `x`: the image as loaded (or transformed if `transform` is set).
#' - `y`: the integer class label. For the `test` split, no labels are available
#'   and `y` will always be `NA`.
#'
#' @details
#' The `test` split corresponds to the \emph{private} evaluation set used in the
#' Places365 challenge. Annotation files are not publicly released, so only the
#' images are provided.
#'
#' @examples
#' \dontrun{
#' ds <- places365_dataset(
#'   split = "val",
#'   download = TRUE,
#'   transform = transform_to_tensor
#' )
#' item <- ds[1]
#' tensor_image_browse(item$x)
#'
#' # Show class index and label
#' label_idx <- item$y
#' label_name <- ds$classes[label_idx]
#' label_idx  # Class Index
#' label_name  # Name of the Label
#'
#' dl <- dataloader(ds, batch_size = 2)
#' batch <- dataloader_next(dataloader_make_iter(dl))
#' batch$x
#'
#' ds_large <- places365_dataset_large(
#'   split = "val",
#'   download = TRUE,
#'   transform = . %>% transform_to_tensor() %>% transform_resize(c(256, 256))
#' )
#' dl <- torch::dataloader(dataset = ds_large, batch_size = 2)
#' batch <- dataloader_next(dataloader_make_iter(dl))
#' batch$x
#' }
#'
#' @family classification_dataset
#' @export

places365_dataset <- torch::dataset(
  name = "places365_dataset",
  archive_size = "30 GB",
  resources = list(
    devkit = c(
      "http://data.csail.mit.edu/places/places365/filelist_places365-standard.tar",
      "35a0585fee1fa656440f3ab298f8479c"
    ),
    val = c(
      "http://data.csail.mit.edu/places/places365/val_256.tar",
      "e27b17d8d44f4af9a78502beb927f808"
    ),
    train = c(
      "http://data.csail.mit.edu/places/places365/train_256_places365standard.tar",
      "53ca1c756c3d1e7809517cc47c5561c5"
    ),
    test = c(
      "http://data.csail.mit.edu/places/places365/test_256.tar",
      "f532f6ad7b582262a2ec8009075e186b"
    )
  ),

  initialize = function(
    root = tempdir(),
    split = c("train", "val", "test"),
    transform = NULL,
    target_transform = NULL,
    download = FALSE,
    loader = magick_loader
  ) {
    self$split <- match.arg(split)
    self$root_path <- normalizePath(root, mustWork = FALSE)
    self$transform <- transform
    self$target_transform <- target_transform
    self$loader <- loader

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")

    # Only read categories if not test
    if (self$split != "test") {
      classes_df <- read.delim(self$categories_file, header = FALSE, sep = " ", col.names = c("category", "index"))
      self$classes <- sub(".*/", "", classes_df$category)
      self$class_to_idx <- setNames(seq_along(self$classes), self$classes)
    }

    if (self$split == "train") {
      ann <- read.delim(self$train_ann, header = FALSE, col.names = c("path", "label"), sep = " ")
      ann$label <- ann$label + 1L
      ann$path <- file.path(self$train_dir, ann$path)
      exist <- fs::file_exists(ann$path)
      self$items <- ann[exist, c("path", "label")]
    } else if (self$split == "val") {
      ann <- read.delim(self$val_ann, header = FALSE, col.names = c("path", "label"), sep = " ")
      ann$label <- ann$label + 1L
      ann$path <- file.path(self$val_dir, ann$path)
      exist <- fs::file_exists(ann$path)
      self$items <- ann[exist, c("path", "label")]
    } else if (self$split == "test") {
      self$files <- fs::dir_ls(self$test_dir, recurse = FALSE, type = "file", glob = "*.jpg")
    }

    if (self$split == "test") {
      cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images.")
    } else {
      cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
    }
  },

  .length = function() {
    if (self$split == "test")
      length(self$files)
    else
      nrow(self$items)
  },

  .getitem = function(index) {
    if (self$split == "test") {
      path <- self$files[[index]]
      x <- self$loader(path)
      if (!is.null(self$transform))
        x <- self$transform(x)
      list(x = x, y = NA_integer_)
    } else if (self$split %in% c("train", "val")) {
      item <- self$items[index, ]
      x <- self$loader(item$path)
      if (!is.null(self$transform))
        x <- self$transform(x)
      y <- item$label
      if (!is.null(self$target_transform))
        y <- self$target_transform(y)
      list(x = x, y = y)
    } else {
      cli_abort("Invalid split: {self$split}")
    }
  },

  download = function() {
    if (self$check_exists())
      return()

    cli_inform("Downloading {.cls {class(self)[[1]]}} split '{self$split}'...")
    fs::dir_create(self$base_dir, recurse = TRUE)

    needed <- switch(
      self$split,
      train = c("devkit", "train"),
      val   = c("devkit", "val"),
      test  = "test"
    )

    for (n in needed) {
      res <- self$resources[[n]]

      # Apply extended timeout to all downloads
      archive <- withr::with_options(list(timeout = 7200), {
        download_and_cache(res[1], prefix = "places365")
      })

      if (!is.na(res[2]) && tools::md5sum(archive) != res[2])
        cli_abort("Corrupt file! Delete the file in {archive} and try again.")

      utils::untar(archive, exdir = self$base_dir)
    }
  },

  check_exists = function() {
    if (self$split == "train")
      fs::file_exists(self$train_ann) && fs::dir_exists(self$train_dir)
    else if (self$split == "val")
      fs::file_exists(self$val_ann) && fs::dir_exists(self$val_dir)
    else if (self$split == "test")
      fs::dir_exists(self$test_dir)
    else
      FALSE
  },

  active = list(
    base_dir = function() file.path(self$root_path, "places365"),
    train_dir = function() file.path(self$base_dir, "data_256"),
    val_dir = function() file.path(self$base_dir, "val_256"),
    test_dir = function() file.path(self$base_dir, "test_256"),
    val_ann = function() file.path(self$base_dir, "places365_val.txt"),
    train_ann = function() file.path(self$base_dir, "places365_train_standard.txt"),
    categories_file = function() file.path(self$base_dir, "categories_places365.txt")
  )
)

#' @describeIn places365_dataset High resolution variant (~160 GB).
#' @export
places365_dataset_large <- torch::dataset(
  name = "places365_dataset_large",
  inherit = places365_dataset,
  archive_size = "160 GB",
  resources = list(
    devkit = c(
      "http://data.csail.mit.edu/places/places365/filelist_places365-standard.tar",
      "35a0585fee1fa656440f3ab298f8479c"
    ),
    val = c(
      "http://data.csail.mit.edu/places/places365/val_large.tar",
      "9b71c4993ad89d2d8bcbdc4aef38042f"
    ),
    train = c(
      "http://data.csail.mit.edu/places/places365/train_large_places365standard.tar",
      "67e186b496a84c929568076ed01a8aa1"
    ),
    test = c(
      "http://data.csail.mit.edu/places/places365/test_large.tar",
      "41a4b6b724b1d2cd862fb3871ed59913"
    )
  ),
  active = list(
    train_dir = function() file.path(self$base_dir, "data_large"),
    val_dir   = function() file.path(self$base_dir, "val_large"),
    test_dir  = function() file.path(self$base_dir, "test_large")
  )
)

#' Non-maximum Suppression (NMS)
#'
#' Performs non-maximum suppression  (NMS) on the boxes according
#' to their intersection-over-union  (IoU). NMS iteratively removes
#' lower scoring boxes which have an IoU greater than iou_threshold
#' with another (higher scoring) box.
#'
#' @param boxes  (Tensor\[N, 4\])): boxes to perform NMS on. They are
#' expected to be in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
#' * \eqn{0 \leq x_{min} < x_{max}} and
#' * \eqn{0 \leq y_{min} < y_{max}}.
#' @param scores (Tensor\[N\]): scores for each one of the boxes
#' @param iou_threshold  (float): discards all overlapping boxes with IoU > iou_threshold
#'
#' @details
#'    If multiple boxes have the exact same score and satisfy the IoU
#'    criterion with respect to a reference box, the selected box is
#'    not guaranteed to be the same between CPU and GPU. This is similar
#'    to the behavior of argsort in torch when repeated values are present.
#'
#'    Current algorithm has a time complexity of O(n^2) and runs in native R.
#'    It may be improve in the future by a Rcpp implementation or through alternative algorithm
#'
#' @return keep (Tensor): int64 tensor with the indices of the elements that
#'  have been kept by NMS, sorted in decreasing order of scores.
#'
#' @export
nms <- function(boxes, scores, iou_threshold) {

  # Handle empty inputs gracefully

  n_boxes <- boxes$shape[1]
  if (n_boxes == 0) {
    return(torch::torch_empty(0L, dtype = torch::torch_long(), device = boxes$device))
  }

  # Sort scores in descending order
  sort_result <- scores$sort(descending = TRUE)
  order <- sort_result[[2]]
  sorted_boxes <- boxes[order, ]

  # Pre-allocate keep list for efficiency

  keep <- vector("integer", n_boxes)
  keep[1] <- 1L
  n_keep <- 1L

  # Handle single box case
  if (n_boxes == 1) {
    return(order[1:1])
  }

  for (i in 2:n_boxes) {
    # Get current box - use unsqueeze to ensure 2D tensor [1, 4]
    current_box <- sorted_boxes[i, ]$unsqueeze(1)

    # Get all kept boxes so far
    kept_indices <- keep[1:n_keep]
    kept_boxes <- sorted_boxes[kept_indices, , drop = FALSE]

    # Compute IoU with all kept boxes
    iou <- box_iou(kept_boxes, current_box)

    # Check if the current box has IoU <= iou_threshold with all kept boxes
    if (all(as.logical(iou <= iou_threshold))) {
      n_keep <- n_keep + 1L
      keep[n_keep] <- i
    }
  }

  # Return indices in original order
  kept_indices <- keep[1:n_keep]
  return(order[kept_indices])
}


#' Batched Non-maximum Suppression (NMS)
#'
#'    Performs non-maximum suppression in a batched fashion.
#'    Each index value correspond to a category, and NMS
#'    will not be applied between elements of different categories.
#'
#' @param boxes (Tensor\[N, 4\]): boxes where NMS will be performed. They are expected to be
#'  in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
#'  *  \eqn{0 \leq x_{min} < x_{max}} and
#'  *  \eqn{0 \leq y_{min} < y_{max}}.
#' @param scores  (Tensor\[N\]): scores for each one of the boxes
#' @param idxs  (Tensor\[N\]): indices of the categories for each one of the boxes.
#' @param iou_threshold  (float): discards all overlapping boxes with IoU > `iou_threshold`
#'
#'
#' @return keep (Tensor): int64 tensor with the indices of
#' the elements that have been kept by NMS, sorted
#' in decreasing order of scores
#'
#' @export
batched_nms <- function(
  boxes,
  scores,
  idxs,
  iou_threshold
) {
  boxes_dtype <- boxes$dtype
  boxes_device <- boxes$device

  # strategy: in order to perform NMS independently per class.
  # we add an offset to all the boxes. The offset is dependent
  # only on the class idx, and is large enough so that boxes
  # from different classes do not overlap

    if(boxes$numel() == 0) {
      return(torch::torch_empty(0, dtype=torch::torch_int64(), device = boxes_device))
    } else {
      max_coordinate <- boxes$max()
      offsets <- idxs$to(device = boxes_device, dtype = boxes_dtype) * (max_coordinate + torch::torch_tensor(1)$to(device = boxes_device, dtype = boxes_dtype))
      boxes_for_nms <- boxes + offsets[, NULL]
      keep <- nms(boxes_for_nms, scores, iou_threshold)
      return(keep)
    }
}

#' Remove Small Boxes
#'
#' Remove boxes which contains at least one side smaller than min_size.
#'
#' @param boxes  (Tensor\[N, 4\]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format
#'  with
#'  * \eqn{0 \leq x_{min} < x_{max}} and
#'  * \eqn{0 \leq y_{min} < y_{max}}.
#' @param min_size  (float): minimum size
#'
#' @return keep (Tensor\[K\]): indices of the boxes that have both sides
#'  larger than min_size
#'
#' @export
remove_small_boxes <- function(boxes, min_size) {
  c(ws, hs) %<-% c(boxes[, 3] - boxes[, 1], boxes[, 4] - boxes[, 2])
  keep <- (ws >= min_size) & (hs >= min_size)
  keep <- torch::torch_where(keep)[[1]]
  return(keep)
}

#' Clip Boxes to Image
#'
#' Clip boxes so that they lie inside an image of size `size`.
#'
#' @param boxes  (Tensor\[N, 4\]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format
#' with
#' * \eqn{0 \leq x_{min} < x_{max}} and
#' * \eqn{0 \leq y_{min} < y_{max}}.
#' @param size  (Tuple\[height, width]): size of the image
#'
#' @return clipped_boxes (Tensor\[N, 4\])
#'
#' @export
clip_boxes_to_image <- function(boxes, size) {
  dim <- boxes$dim()
  boxes_x <- boxes[.., seq(1, boxes$shape[2], 2)]
  boxes_y <- boxes[.., seq(2, boxes$shape[2], 2)]
  c(height, width) %<-% size

  # if(torchvision$_is_tracing()) {
  #   boxes_x = torch::torch_max(boxes_x, other = torch::torch_tensor(0, dtype=boxes$dtype, device=boxes$device))
  #   boxes_x = torch::torch_min(boxes_x, other = torch::torch_tensor(width, dtype=boxes$dtype, device=boxes$device))
  #   boxes_y = torch::torch_max(boxes_y, other = torch::torch_tensor(0, dtype=boxes$dtype, device=boxes$device))
  #   boxes_y = torch::torch_min(boxes_y, other = torch::torch_tensor(height, dtype=boxes$dtype, device=boxes$device))
  # } else {
  boxes_x <- boxes_x$clamp(min=0, max=width)
  boxes_y <- boxes_y$clamp(min=0, max=height)

  clipped_boxes <- torch::torch_stack(c(boxes_x, boxes_y), dim=dim+1)
  return(clipped_boxes$reshape(boxes$shape))
}

#' Box Convert
#'
#'  Converts boxes from given in_fmt to out_fmt.
#'
#' @param boxes  (Tensor\[N, 4\]): boxes which will be converted.
#' @param in_fmt  (str): Input format of given boxes. Supported formats are \['xyxy', 'xywh', 'cxcywh'\].
#' @param out_fmt  (str): Output format of given boxes. Supported formats are \['xyxy', 'xywh', 'cxcywh'\]
#' @return boxes (Tensor\[N, 4]): Boxes into converted format.
#'
#' @details
#' Supported in_fmt and out_fmt are:
#' * 'xyxy': boxes are represented via corners,
#'    * \eqn{x_{min}, y_{min}} being top left and
#'    * \eqn{x_{max}, y_{max}} being bottom right.
#' * 'xywh' : boxes are represented via corner, width and height,
#'    * \eqn{x_{min}, y_{min}} being top left,
#'    * w, h being width and height.
#' * 'cxcywh' : boxes are represented via centre, width and height,
#'    * \eqn{c_x, c_y} being center of box,
#'    * w, h  being width and height.
#'
#' @export
box_convert <- function(boxes, in_fmt, out_fmt) {
  allowed_fmts <- c("xyxy", "xywh", "cxcywh")
  if((!in_fmt %in% allowed_fmts) || (!out_fmt %in% allowed_fmts))
    value_error("Unsupported Bounding Box Conversions for given in_fmt and out_fmt")

  if(in_fmt == out_fmt)
    return(boxes$clone())

  if(in_fmt != 'xyxy' && out_fmt != 'xyxy') {
    # convert to xyxy and change in_fmt xyxy
    if(in_fmt == "xywh") {
      boxes <- box_xywh_to_xyxy(boxes)
    } else if(in_fmt == "cxcywh") {
      boxes <- box_cxcywh_to_xyxy(boxes)
    }
    in_fmt <- 'xyxy'
  }
  if(in_fmt == "xyxy") {
    if(out_fmt == "xywh") {
      boxes <- box_xyxy_to_xywh(boxes)
    } else if(out_fmt == "cxcywh") {
      boxes <- box_xyxy_to_cxcywh(boxes)
    }
  } else if(out_fmt == "xyxy") {
    if(in_fmt == "xywh") {
      boxes <- box_xywh_to_xyxy(boxes)
    } else if(in_fmt == "cxcywh") {
      boxes <- box_cxcywh_to_xyxy(boxes)
    }
  }
  return(boxes)
}

upcast <- function(t) {
  t_dtype <- as.character(t$dtype)
  # Protects from numerical overflows in multiplications by upcasting to the equivalent higher type
  if(t$is_floating_point()) {
    return(if(t_dtype %in% c("Float", "Double")) t else t$to(device = torch::torch_float()))
  } else {
    return(if(t_dtype %in% c("Int", "Long")) t else t$to(device = torch::torch_int()))
  }
}

#' Box Area
#'
#' Computes the area of a set of bounding boxes, which are specified by its
#' \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} coordinates.
#'
#' @param boxes  (Tensor\[N, 4\]): boxes for which the area will be computed. They
#' are expected to be in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
#' * \eqn{0 \leq x_{min} < x_{max}} and
#' * \eqn{0 \leq y_{min} < y_{max}}.
#'
#' @return area (Tensor\[N\]): area for each box
#'
#' @export
box_area <- function(boxes) {
  boxes <- upcast(boxes)
  return((boxes[, 3] - boxes[, 1]) * (boxes[, 4] - boxes[, 2]))
}

box_inter_union <- function(boxes1, boxes2) {
  # implementation from https://github.com/kuangliu/torchcv/blob/master/torchcv/utils/box.py
  # with slight modifications
  area1 <- box_area(boxes1)
  area2 <- box_area(boxes2)

  lt <- torch::torch_max(boxes1[, NULL, 1:2], other = boxes2[, 1:2]) # [N,M,2]
  rb <- torch::torch_min(boxes1[, NULL, 3:N], other = boxes2[, 3:N]) # [N,M,2]

  wh <- upcast(rb - lt)$clamp(min=0) # [N,M,2]
  inter <- wh[, , 1] * wh[, , 2] # [N,M]

  union <- area1[, NULL] + area2 - inter

  return(list(inter, union))
}

#' Box IoU
#'
#' Return intersection-over-union  (Jaccard index) of boxes.
#' Both sets of boxes are expected to be in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
#' \eqn{0 \leq x_{min} < x_{max}} and \eqn{0 \leq y_{min} < y_{max}}.
#'
#' @param boxes1  (Tensor\[N, 4\])
#' @param boxes2  (Tensor\[M, 4\])
#'
#' @return iou (Tensor\[N, M\]): the NxM matrix containing the pairwise IoU values for every element in boxes1 and boxes2
#'
#' @export
box_iou <- function(boxes1, boxes2) {
    inter_union <- box_inter_union(boxes1, boxes2)
    iou <- inter_union[[1]] / inter_union[[2]]
    return(iou)
}

#' Generalized Box IoU
#'
#' Return generalized intersection-over-union  (Jaccard index) of boxes.
#' Both sets of boxes are expected to be in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
#' \eqn{0 \leq x_{min} < x_{max}} and \eqn{0 \leq y_{min} < y_{max}}.
#'
#' @param boxes1  (Tensor\[N, 4\])
#' @param boxes2  (Tensor\[M, 4\])
#'
#' @details
#' Implementation adapted from https://github.com/facebookresearch/detr/blob/master/util/box_ops.py
#'
#' @return generalized_iou (Tensor\[N, M\]): the NxM matrix containing the pairwise generalized_IoU values
#'        for every element in boxes1 and boxes2
#'
#' @export
generalized_box_iou <- function(boxes1, boxes2) {
  # degenerate boxes gives inf / nan results
  # so do an early check
  if(as.numeric((boxes1[, 3:N] >= boxes1[, 1:2])$all()) != 1)
    value_error("(boxes1[, 3:N] >= boxes1[, 1:2])$all() not TRUE")
  if(as.numeric((boxes2[, 3:N] >= boxes2[, 1:2])$all()) != 1)
    value_error("(boxes2[, 3:N] >= boxes2[, 1:2])$all() not TRUE")

  inter_union <- box_inter_union(boxes1, boxes2)
  iou <- inter_union[[1]] / inter_union[[2]]

  lti <- torch::torch_min(boxes1[, NULL, 1:2], other = boxes2[, 1:2])
  rbi <- torch::torch_max(boxes1[, NULL, 3:N], other = boxes2[, 3:N])

  whi <- upcast(rbi - lti)$clamp(min=0) # [N,M,2]
  areai <- whi[, , 1] * whi[, , 2]

  return(iou - (areai - inter_union[[2]]) / areai)
}

x1 <- torch::torch_ones(5)
y1 <- torch::torch_ones(5)
x2 <- x1 + 1
y2 <- y1 + 1
boxes <- torch::torch_stack(list(x1,y1,x2,y2))$transpose(2,1)

test_that("batched_nms", {
  expect_no_error(
    x <- batched_nms(
      boxes = boxes,
      scores = torch::torch_ones(5)*0.6,
      idxs = torch::torch_ones(5),
      iou_threshold = 0.5
    )
  )
  expect_tensor(x)
  expect_equal_to_r(x, 1L)

  boxes <- torch::torch_stack(list(x1 + torch_randint(0,6,5),
                                  y1 + torch_randint(0,6,5),
                                  x2 + torch_randint(6,12,5),
                                  y2 + torch_randint(6,12,5)))$transpose(2,1)
  expect_no_error(
    x <- batched_nms(
      boxes = boxes,
      scores = torch::torch_ones(5)*0.6,
      idxs = torch::torch_ones(5),
      iou_threshold = 2
    )
  )
  expect_tensor(x)
  expect_gt(x$shape, 1L)
})

test_that("remove_small_boxes", {
  expect_no_error(x <- remove_small_boxes(boxes, 1))
  expect_tensor(x)
})

test_that("clip_boxes_to_image", {
  expect_no_error(x <- clip_boxes_to_image(boxes, c(10,10)))
  expect_tensor(x)
})

test_that("box_convert", {
  xyxy <- boxes
  xywh <- box_convert(xyxy, "xyxy", "xywh")
  cxcywh <- box_convert(xyxy, "xyxy", "cxcywh")

  expect_tensor(xywh)
  expect_tensor(cxcywh)
})

test_that("box_area", {
  area_6 <- torch::torch_tensor(matrix(c(0,0,2,3), 1, 4))
  expect_no_error(x <- box_area(area_6))
  expect_equal_to_r(x, 6)
})

test_that("box_iou", {
  expect_no_error(x <- box_iou(boxes, boxes))
  expect_tensor(x)
  expect_equal_to_r(x, matrix(1, nrow = 5, ncol = 5))

  expect_no_error(x <- generalized_box_iou(boxes, boxes))
  expect_tensor(x)
  expect_equal_to_r(x, matrix(1, nrow = 5, ncol = 5))
})

#' @importFrom magrittr %>%
NULL

#' A simplified version of torchvision.utils.make_grid
#'
#' Arranges a batch B of (image) tensors in a grid, with optional padding between
#'   images. Expects a 4d mini-batch tensor of shape (B x C x H x W).
#'
#' @param tensor tensor of shape (B x C x H x W) to arrange in grid.
#' @param scale whether to normalize (min-max-scale) the input tensor.
#' @param num_rows number of rows making up the grid (default 8).
#' @param padding amount of padding between batch images (default 2).
#' @param pad_value pixel value to use for padding.
#'
#' @return  a 3d torch_tensor of shape \eqn{\approx(C , num\_rows \times H , num\_cols \times W)} of all images arranged in a grid.
#' @family image display
#' @export
vision_make_grid <- function(tensor,
                             scale = TRUE,
                             num_rows = 8,
                             padding = 2,
                             pad_value = 0) {

  min_max_scale <- function(x) {
    min <- x$min()$item()
    max <- x$max()$item()
    x$clamp_(min = min, max = max)
    x$add_(-min)$div_(max - min + 1e-5)
    x
  }
  if(scale) tensor <- min_max_scale(tensor)

  nmaps <- tensor$size(1)
  xmaps <- min(num_rows, nmaps)
  ymaps <- ceiling(nmaps / xmaps)
  height <- floor(tensor$size(3) + padding)
  width <- floor(tensor$size(4) + padding)
  num_channels <- tensor$size(2)
  grid <-
    tensor$new_full(c(num_channels, height * ymaps + padding, width * xmaps + padding),
                    pad_value)
  k <- 0

  for (y in 0:(ymaps - 1)) {
    for (x in 0:(xmaps - 1)) {
      if (k >= nmaps)
        break
      grid$narrow(
        dim = 2,
        start =  1 + torch::torch_tensor(y * height + padding, dtype = torch::torch_int64())$sum(dim = 1),
        length = height - padding
      )$narrow(
        dim = 3,
        start = 1 + torch::torch_tensor(x * width + padding, dtype = torch::torch_int64())$sum(dim = 1),
        length = width - padding
      )$copy_(tensor[k + 1, , ,])
      k <- k + 1
    }
  }

  grid
}


#' Draws bounding boxes on image.
#'
#' Draws bounding boxes on top of one image tensor
#'
#' @param x Tensor of shape (C x H x W) and dtype `uint8` or dtype `float`.
#'              In case of dtype float, values are assumed to be in range \eqn{[0, 1]}.
#'              C value for channel can only be 1 (grayscale) or 3 (RGB).
#' @param boxes Tensor of size (N, 4) containing N bounding boxes in
#'            c(\eqn{x_{min}}, \eqn{y_{min}}, \eqn{x_{max}}, \eqn{y_{max}}).
#'            format. Note that the boxes coordinates are absolute with respect
#'            to the image. In other words: \eqn{0  \leq x_{min} < x_{max} < Height } and
#'            \eqn{0  \leq y_{min} < y_{max} < Width }.
#' @param labels character vector containing the labels of bounding boxes.
#' @param colors character vector containing the colors
#'            of the boxes or single color for all boxes. The color can be represented as
#'            strings e.g. "red" or "#FF00FF". By default, viridis colors are generated for boxes.
#' @param color Deprecated alias for `colors`.
#' @param fill If `TRUE` fills the bounding box with specified color.
#' @param width  Width of text shift to the bounding box.
#' @param font NULL for the current font family, or a character vector of length 2 for Hershey vector fonts.
#' @param font_size The requested font size in points.
#' @param ... Additional arguments passed to methods.
#'
#' @return  torch_tensor of size (C, H, W) of dtype uint8: Image Tensor with bounding boxes plotted.
#'
#' @examples
#' if (torch::torch_is_installed()) {
#' \dontrun{
#' image_tensor <- torch::torch_randint(170, 250, size = c(3, 360, 360))$to(torch::torch_uint8())
#' x <- torch::torch_randint(low = 1, high = 160, size = c(12,1))
#' y <- torch::torch_randint(low = 1, high = 260, size = c(12,1))
#' boxes <- torch::torch_cat(c(x, y, x + 20, y +  10), dim = 2)
#' bboxed <- draw_bounding_boxes(image_tensor, boxes, colors = "black", fill = TRUE)
#' tensor_image_browse(bboxed)
#' }
#' }
#' @family image display
#' @export
draw_bounding_boxes <- function(x, ...) {
  UseMethod("draw_bounding_boxes")
}

#' @rdname draw_bounding_boxes
#' @export
draw_bounding_boxes.default <- function(x, ...) {
  cli_abort("The provided x class {.class {class(x)}} is not supported")
}

#' @rdname draw_bounding_boxes
#' @export
draw_bounding_boxes.torch_tensor <- function(x,
                                             boxes,
                                             labels = NULL,
                                             colors = NULL,
                                             color = NULL,
                                             fill = FALSE,
                                             width = 1,
                                             font = c("serif", "plain"),
                                             font_size = 10, ...) {
  rlang::check_installed("magick")

  if (!is.null(color)) {
    if (!is.null(colors)) {
      cli_abort("Use either {.arg colors} or {.arg color}, not both.")
    }
    cli_warn("{.arg color} is deprecated; use {.arg colors} instead.")
    colors <- color
  }

  if (x$ndim == 4 && x$size(1) == 1) x <- x$squeeze(1)
  if (x$ndim != 3) value_error("Pass an individual image as `x`, not a batch")
  if (!x$size(1) %in% c(1, 3)) value_error("Only grayscale and RGB images are supported")

  img_to_draw <- if (x$dtype == torch::torch_uint8()) {
    x$div(255)$permute(c(2, 3, 1))$to(device = "cpu") %>% as.array()
  } else if (x$dtype == torch::torch_float()) {
    x$permute(c(2, 3, 1))$to(device = "cpu") %>% as.array()
  } else type_error("`x` should be torch_uint8 or torch_float")

  if ((boxes[, 1] >= boxes[, 3])$any() %>% as.logical() ||
      (boxes[, 2] >= boxes[, 4])$any() %>% as.logical()) {
    value_error("Boxes must be in c(xmin, ymin, xmax, ymax) format")
  }
  num_boxes <- boxes$shape[1]
  if (num_boxes == 0) {
    cli_warn("boxes doesn't contain any box. No box was drawn")
    return(x)
  }
  if (!is.null(labels) && (num_boxes %% length(labels) != 0)) {
    cli_abort(
      "Number of labels {.val {length(labels)}} cannot be broadcasted on number of boxes {.val {num_boxes}}"
    )
  }
  if (is.null(colors)) {
    colors <- grDevices::hcl.colors(n = num_boxes)
  }
  if (num_boxes %% length(colors) != 0) {
    value_error("colors vector cannot be broadcasted on boxes")
  }

  if (!fill) {
    fill_col <- NA
  } else {
    fill_col <- colors
  }

  if (is.null(font)) {
    vfont <- c("serif", "plain")
  } else {
    if (is.null(font_size)) font_size <- 10
  }

  if (x$size(1) == 1) {
    x <- x$tile(c(4, 2, 2))
  }

  img_bb <- boxes$to(torch::torch_int64()) %>% as.array()

  draw <- png::writePNG(img_to_draw) %>%
    magick::image_read() %>%
    magick::image_draw()

  graphics::rect(img_bb[, 1], img_bb[, 2], img_bb[, 3], img_bb[, 4],
                 col = fill_col, border = colors, lwd = width)

  if (!is.null(labels)) {
    graphics::text(
      img_bb[, 1] + 2 * width + font_size,
      img_bb[, 2] + 2 * width,
      labels = labels,
      col = colors,
      vfont = font,
      cex = font_size / 10
    )
  }

  grDevices::dev.off()

  draw_tt <- draw %>%
    magick::image_data(channels = "rgb") %>%
    as.integer() %>%
    torch::torch_tensor(dtype = torch::torch_uint8())

  return(draw_tt$permute(c(3, 1, 2)))
}

#' @rdname draw_bounding_boxes
#' @export
draw_bounding_boxes.image_with_bounding_box <- function(x, ...) {
  draw_bounding_boxes(
    x = x$x,
    boxes = x$y$boxes,
    labels = x$y$labels,
    ...
  )
}



#' Convert COCO polygon to mask tensor (Robust Version)
#'
#' Converts a COCO-style polygon annotation (list of coordinates) into a binary mask tensor.
#'
#' @param segmentation A list of polygons from COCO annotations (e.g., \code{anns$segmentation[[i]]}).
#' @param height Height of the image
#' @param width Width of the image
#'
#' @return A torch_bool() tensor of shape (height, width)
#'
#' @keywords internal
coco_polygon_to_mask <- function(segmentation, height, width) {
  rlang::check_installed("magick")

  # Handle empty polygon list early to avoid graphics device issues
  if (length(segmentation) == 0) {
    mask_logical <- matrix(FALSE, nrow = height, ncol = width)
    mask_tensor <- torch::torch_tensor(mask_logical, dtype = torch::torch_bool())
    return(mask_tensor)
  }

  mask_img <- magick::image_blank(width = width, height = height, color = "black")
  mask_img <- magick::image_draw(mask_img)

  for (poly in segmentation) {
    flat <- unlist(poly)
    n_coords <- length(flat)

    # Ensure number of coordinates is even (x, y) pairs
    if (n_coords %% 2 != 0) {
      flat <- flat[-n_coords]  # Drop the last element if odd
    }

    if (length(flat) >= 6) {  # At least 3 points to form a polygon
      coords <- matrix(flat, ncol = 2, byrow = TRUE)
      polygon(coords[, 1], coords[, 2], col = "white", border = NA)
    }
  }

  dev.off()

  gray <- magick::image_data(mask_img, channels = "gray")

  if (length(dim(gray)) == 3) {
    mask_matrix <- t(as.matrix(gray[1, , ]))
  } else if (length(dim(gray)) == 2) {
    mask_matrix <- t(as.matrix(gray))
  } else {
    gray_vec <- as.vector(gray)
    mask_matrix <- matrix(gray_vec, nrow = height, ncol = width, byrow = TRUE)
  }

  if (nrow(mask_matrix) != height || ncol(mask_matrix) != width) {
    stop(sprintf("Mask matrix dimensions (%d x %d) don't match expected (%d x %d)",
                 nrow(mask_matrix), ncol(mask_matrix), height, width))
  }

  mask_logical <- mask_matrix > 0
  mask_tensor <- torch::torch_tensor(mask_logical, dtype = torch::torch_bool())

  return(mask_tensor)
}


#' Draw segmentation masks
#'
#' Draw segmentation masks with their respective colors on top of a given RGB tensor image
#'
#' @param x Tensor of shape (C x H x W) and dtype `uint8` or dtype `float`.
#'              In case of dtype float, values are assumed to be in range \eqn{[0, 1]}.
#'              C value for channel can only be 1 (grayscale) or 3 (RGB).
#' @param masks torch_tensor of shape (num_masks, H, W) or (H, W) and dtype bool.
#' @param alpha number between 0 and 1 denoting the transparency of the masks.
#   0 means full transparency, 1 means no transparency.
#' @param colors character vector containing the colors
#'            of the boxes or single color for all boxes. The color can be represented as
#'            strings e.g. "red" or "#FF00FF". By default, viridis colors are generated for masks
#' @param ... Additional arguments passed to methods.
#'
#' @importFrom graphics polygon
#' @importFrom grDevices dev.off
#' @importFrom torch as_array
#'
#' @return torch_tensor of shape (3, H, W) and dtype uint8 of the image with segmentation masks drawn on top.
#'
#' @examplesIf torch::torch_is_installed() && rlang::is_installed("magick")
#' image_tensor <- torch::torch_randint(170, 250, size = c(3, 360, 360))$to(torch::torch_uint8())
#' mask <- torch::torch_tril(torch::torch_ones(c(360, 360)))$to(torch::torch_bool())
#' masked_image <- draw_segmentation_masks(image_tensor, mask, alpha = 0.2)
#' tensor_image_browse(masked_image)
#' @family image display
#' @export
draw_segmentation_masks <- function(x, ...) {
  UseMethod("draw_segmentation_masks")
}

#' @rdname draw_segmentation_masks
#' @export
draw_segmentation_masks.default <- function(x, ...) {
  type_error("The provided object of class {.cls {class(x)}} is not supported by draw_segmentation_masks.")
}

#' @rdname draw_segmentation_masks
#' @export
draw_segmentation_masks.torch_tensor <- function(x,
                                                 masks,
                                                 alpha = 0.8,
                                                 colors = NULL, ...) {
  rlang::check_installed("magick")
  out_dtype <- torch::torch_uint8()

  if (x$ndim != 3) {
    value_error("Pass individual `image`, not batches")
  }
  if (!x$size(1) %in% c(1, 3)) {
    value_error("Only grayscale and RGB images are supported")
  }
  if (x$dtype == out_dtype) {
    img_to_draw <- x$detach()$clone()
  } else if (x$dtype == torch::torch_float()) {
    img_to_draw <- x$detach()$clone()$mul(255)$to(dtype = out_dtype)
  } else {
    type_error("`x` (image) should be of dtype `torch_uint8` or `torch_float`")
  }

  if (masks$ndim == 2) {
    masks <- masks$unsqueeze(1)
  }
  if (masks$ndim != 3) {
    value_error("`masks` must be of shape (H, W) or (num_masks, H, W)")
  }
  # datasets item include boolean masks, and models inference produce logits floats for masks
  if (masks$dtype != torch::torch_bool() && masks$dtype != torch::torch_float() ) {
    type_error("`masks` is expected to be of dtype torch_bool() or torch_float()")
  }
  if (any(masks$shape[-2:-1] != img_to_draw$shape[-2:-1])) {
    value_error("`masks` and `image` must have the same height and width")
  }
  # if mask is a model inference output, we need to convert float mask to boolean mask
  if (masks$dtype == torch::torch_float() ) {
    mask_id <- masks$argmax(dim = 1)
    masks_seq <- mask_id$aminmax()[[1]]$item():mask_id$aminmax()[[2]]$item()
    # turn mask_id \code{[LongType{H,W}]} into a boolean mask \code{[BoolType{num_masks,H,W}]}
    masks <- torch::torch_stack(lapply(masks_seq, function(x) mask_id$eq(x)), dim = 1)
  } else {
    masks_seq <- seq(masks$size(1))
  }
  num_masks <- length(masks_seq)
  if (num_masks == 0) {
    cli_warn("masks doesn't contain any mask. No mask was drawn")
    return(x)
  }

  if (is.null(colors)) {
    colors <- grDevices::hcl.colors(n = num_masks)
  }
  if (num_masks %% length(colors) != 0) {
    cli_abort("colors vector of size {.value {length(colors)}} cannot be broadcasted on {.value {num_masks}} masks")
  }

  color_tt <- colors %>%
    grDevices::col2rgb() %>%
    t() %>%
    torch::torch_tensor(dtype = out_dtype)

  colored_mask_stack <- torch::torch_stack(lapply(
    masks_seq,
    function(i) color_tt[i, ]$unsqueeze(2)$unsqueeze(2)$mul(masks[i:i, , ])
  ), dim = 1)

  out <- img_to_draw * (1 - alpha) + torch::torch_sum(colored_mask_stack, dim = 1) * alpha
  return(out$to(out_dtype))
}

#' @rdname draw_segmentation_masks
#' @export
draw_segmentation_masks.image_with_segmentation_mask <- function(x,
                                                                 alpha = 0.5,
                                                                 colors = NULL, ...) {
  draw_segmentation_masks(
    x = x$x,
    masks = x$y$masks,
    alpha = alpha,
    colors = colors,
    ...
  )
}


#' Draws Keypoints
#'
#' Draws Keypoints, an object describing a body part (like rightArm or leftShoulder), on given RGB tensor image.
#' @param image Tensor of shape (3 x H x W) and dtype `uint8` or dtype `float`.
#'              In case of dtype float, values are assumed to be in range \eqn{[0, 1]}.
#' @param keypoints Tensor of shape (N, K, 2) the K keypoints location for each of the N detected poses instance,
#         in the format c(x, y).
#' @param connectivity List of integer pairs `c(i, j)` specifying which
#'            keypoints to connect with a line, e.g. `list(c(1, 2), c(2, 3))`.
#'            `NULL` (default) draws no connecting lines.
#' @param colors character vector containing the colors
#'            of the keypoints or single color for all keypoints. The color can be represented as
#'            strings e.g. "red" or "#FF00FF". By default, rainbow colors are generated for keypoints
#' @param radius radius of the plotted keypoint.
#' @param width width of line connecting keypoints.
#'
#' @return Image Tensor of dtype uint8 with keypoints drawn.
#'
#' @examples
#' if (torch::torch_is_installed()) {
#' \dontrun{
#' image <- torch::torch_randint(190, 255, size = c(3, 360, 360))$to(torch::torch_uint8())
#' keypoints <- torch::torch_randint(low = 60, high = 300, size = c(4, 5, 2))
#' keypoint_image <- draw_keypoints(image, keypoints)
#' tensor_image_browse(keypoint_image)
#' }
#' }
#' @family image display
#' @export
draw_keypoints <- function(image,
                           keypoints,
                           connectivity = NULL,
                           colors = NULL,
                           radius = 2,
                           width = 3) {

  rlang::check_installed("magick")
  if (!inherits(image, "torch_tensor")) {
    type_error("`image` should be a torch_tensor")
  }
  if (image$ndim != 3) {
    value_error("Pass individual `image`, not batches")
  }
  if (!image$size(1) %in% c(1, 3)) {
    value_error("Only grayscale and RGB images are supported")
  }
  if (image$dtype == torch::torch_uint8()) {
    img_to_draw <- image$div(255)$permute(c(2, 3, 1))$to(device = "cpu") %>% as.array
  } else if (image$dtype == torch::torch_float()) {
    img_to_draw <- image$permute(c(2, 3, 1))$to(device = "cpu") %>% as.array
  } else {
    type_error("`image` should be of dtype `torch_uint8` or `torch_float`")
  }

  if (keypoints$ndim != 3) {
    cli_abort("{.var keypoints} must be of shape (num_instances, K, 2), but current shape is {.value {keypoints$shape}}")
  }

  img_kpts <- keypoints$to(torch::torch_int64()) %>% as.array

  if (is.null(colors)) {
    colors <- grDevices::hcl.colors(n = dim(img_kpts)[[2]])
  }

  draw <- png::writePNG(img_to_draw) %>%
    magick::image_read() %>%
    magick::image_draw()

  for (pose in 1:dim(img_kpts)[[1]]) {
    graphics::points(img_kpts[pose,,1], img_kpts[pose,,2], pch = ".", col = colors, cex = radius)

    if (!is.null(connectivity)) {
      for (conn in connectivity) {
        start_idx <- conn[1]
        end_idx <- conn[2]
        start_x <- img_kpts[pose, start_idx, 1]
        start_y <- img_kpts[pose, start_idx, 2]
        end_x <- img_kpts[pose, end_idx, 1]
        end_y <- img_kpts[pose, end_idx, 2]
        graphics::lines(c(start_x, end_x), c(start_y, end_y), col = colors[start_idx], lwd = width)
      }
    }
  }
  grDevices::dev.off()
  draw_tt <-
    draw %>% magick::image_data(channels = "rgb") %>% as.integer %>% torch::torch_tensor(dtype = torch::torch_uint8())

  return(draw_tt$permute(c(3, 1, 2)))
}


#' Display image tensor
#'
#' Display image tensor onto the X11 device
#' @param image `torch_tensor()` of shape (1, W, H) for grayscale image or (3, W, H) for
#'  color image to display
#' @param animate support animations in the X11 display
#'
#' @family image display
#' @export
tensor_image_display <- function(image, animate = TRUE) {
  if (image$ndim != 3) {
    value_error("Pass individual `image`, not batches")
  }
  if (!image$size(1) %in% c(1, 3)) {
    value_error("Only grayscale and RGB images are supported")
  }

  if (image$dtype == torch::torch_uint8()) {
    img_to_draw <- image$permute(c(2, 3, 1))$to(device = "cpu", dtype = torch::torch_long()) %>%
      as.array() / 255
  } else {
    img_to_draw <- image$permute(c(2, 3, 1))$to(device = "cpu") %>%
      as.array()
  }
  png::writePNG(img_to_draw) %>% magick::image_read() %>% magick::image_display(animate = animate)

  invisible(NULL)
}


#' Display image tensor
#'
#' Display image tensor into browser
#' @param image `torch_tensor()` of shape (1, W, H) for grayscale image or (3, W, H) for
#'  color image to display
#' @param browser argument passed to [browseURL]
#'
#' @family image display
#' @export
tensor_image_browse <- function(image, browser = getOption("browser")) {
  if (image$ndim != 3) {
    value_error("Pass individual `image`, not batches")
  }
  if (!image$size(1) %in% c(1, 3)) {
    value_error("Only grayscale and RGB images are supported")
  }

  if (image$dtype == torch::torch_uint8()) {
    img_to_draw <- image$permute(c(2, 3, 1))$to(device = "cpu", dtype = torch::torch_long()) %>%
      as.array() / 255
  } else {
    img_to_draw <- image$permute(c(2, 3, 1))$to(device = "cpu") %>%
      as.array()
  }

  png::writePNG(img_to_draw) %>% magick::image_read() %>% magick::image_browse(browser = browser)

  invisible(NULL)
}

context("vision-utils")

test_that("vision_make_grid", {

  images <- torch::torch_randn(c(4, 3, 16, 16))

  grid <- vision_make_grid(images, num_rows = 2, padding = 0)


  expect_equal(grid$size(), c(3, 32, 32))
  expect_equal(as.numeric(grid$max() - grid$min()), 1, tolerance = 1e-4)

})

test_that("draw_bounding_boxes works", {

  image_float <- 1 - (torch::torch_randn(c(3, 360, 360)) / 20)
  image_uint <- (255 - (torch::torch_randint(low = 1, high = 60, size = c(3, 360, 360))))$to(torch::torch_uint8())
  x <- torch::torch_randint(low = 1, high = 160, size = c(12,1))
  y <- torch::torch_randint(low = 1, high = 260, size = c(12,1))
  w <- torch::torch_randint(low = 10, high = 100, size = c(12,1))
  h <- torch::torch_randint(low = 30, high = 60, size = c(12,1))
  boxes <- torch::torch_cat(c(x, y, x + w, y +  h), dim = 2)

  expect_error(bboxed_image <- draw_bounding_boxes(image_uint$to(dtype = torch::torch_int32()), boxes),
               class = "type_error", regexp = "torch_uint8")

  expect_no_error(bboxed_image <- draw_bounding_boxes(image_float, boxes, labels = "dog", width = 5))
  expect_no_error(bboxed_image <- draw_bounding_boxes(image_uint, boxes, labels = "Leptailurus serval constantina", width = 1))
  expect_tensor_dtype(bboxed_image, torch::torch_uint8())
  expect_tensor_shape(bboxed_image, c(3, 360, 360))

  expect_no_error(bboxed_image <- draw_bounding_boxes(image_uint, boxes, colors = "black", fill = TRUE))
})

test_that("draw_bounding_boxes correctly mask a complete image", {

  image_float <- 1 - (torch::torch_randn(c(3, 360, 360)) / 20)
  image_uint <- torch::torch_randint(low = 1, high = 240, size = c(3, 360, 360))$to(torch::torch_uint8())
  boxes <- torch::torch_tensor(c(0,0,360,360))$unsqueeze(1)

  expect_no_error(bboxed_image <- draw_bounding_boxes(image_float, boxes, colors = "black", fill = TRUE))
  expect_no_error(bboxed_image <- draw_bounding_boxes(image_uint, boxes, colors = "black", fill = TRUE))
  # some invisible glitch remains
  expect_lte(bboxed_image$sum() %>% as.numeric, 3000)


})

test_that("draw_bounding_boxes supports deprecated color alias", {
  image_uint <- torch::torch_randint(low = 1, high = 240, size = c(3, 64, 64))$to(torch::torch_uint8())
  boxes <- torch::torch_tensor(c(1, 1, 40, 40))$unsqueeze(1)

  expect_warning(
    bboxed_image <- draw_bounding_boxes(image_uint, boxes, color = "black", fill = TRUE),
    "deprecated"
  )
  expect_tensor_dtype(bboxed_image, torch::torch_uint8())
  expect_tensor_shape(bboxed_image, c(3, 64, 64))
})

test_that("draw_segmentation_masks works with boolean mask", {

  image_float <- 1 - (torch::torch_randn(c(3, 360, 360)) / 20)
  image_uint <- torch::torch_randint(low = 190, high = 255, size = c(3, 360, 360))$to(torch::torch_uint8())
  lower_mask <- torch::torch_tril(torch::torch_ones(c(360, 360)), diagonal = FALSE)$to(torch::torch_bool())
  upper_mask <- torch::torch_triu(torch::torch_ones(c(360, 360)), diagonal = FALSE)$to(torch::torch_bool())
  masks <- torch::torch_stack(c(lower_mask, upper_mask), dim = 1)

  expect_no_error(masked_image <- draw_segmentation_masks(image_float, masks))
  expect_tensor_dtype(masked_image, torch::torch_uint8())
  expect_tensor_shape(masked_image, c(3, 360, 360))

  expect_no_error(masked_image <- draw_segmentation_masks(image_uint, masks))
  expect_tensor_dtype(masked_image, torch::torch_uint8())
  expect_tensor_shape(masked_image, c(3, 360, 360))

  colors <-  c("navyblue", "orange3")
  expect_no_error(masked_image <- draw_segmentation_masks(image_uint, masks, colors = colors, alpha = 0.5))
})

test_that("draw_segmentation_masks works with float mask", {

  image_float <- 1 - (torch::torch_randn(c(3, 360, 360)) / 20)
  image_uint <- torch::torch_randint(low = 190, high = 255, size = c(3, 360, 360))$to(torch::torch_uint8())
  lower_mask <- torch::torch_tril(torch::torch_ones(c(360, 360)), diagonal = FALSE)
  upper_mask <- torch::torch_triu(torch::torch_ones(c(360, 360))*2, diagonal = FALSE)
  masks <- torch::torch_stack(c(lower_mask, upper_mask), dim = 1)

  expect_no_error(masked_image <- draw_segmentation_masks(image_float, masks))
  expect_tensor_dtype(masked_image, torch::torch_uint8())
  expect_tensor_shape(masked_image, c(3, 360, 360))

  expect_no_error(masked_image <- draw_segmentation_masks(image_uint, masks))
  expect_tensor_dtype(masked_image, torch::torch_uint8())
  expect_tensor_shape(masked_image, c(3, 360, 360))

  colors <-  c("navyblue", "orange3")
  expect_no_error(masked_image <- draw_segmentation_masks(image_uint, masks, colors = colors, alpha = 0.5))
})

test_that("draw_keypoints works", {

  image_float <- 1 - (torch::torch_randn(c(3, 360, 360)) / 20)
  image_uint <- torch::torch_randint(low = 190, high = 255, size = c(3, 360, 360))$to(torch::torch_uint8())
  keypoints <- torch::torch_randint(low = 60, high = 300, size = c(4, 5, 2))
  colors <-  hcl.colors(n = 5)

  expect_no_error(keypoint_image <- draw_keypoints(image_float, keypoints))
  expect_tensor_dtype(keypoint_image, torch::torch_uint8())
  expect_tensor_shape(keypoint_image, c(3, 360, 360))
  expect_no_error(keypoint_image <- draw_keypoints(image_float, keypoints, colors = colors, radius = 7))

  expect_no_error(keypoint_image <- draw_keypoints(image_uint, keypoints))
  expect_tensor_dtype(keypoint_image, torch::torch_uint8())
  expect_tensor_shape(keypoint_image, c(3, 360, 360))
  expect_no_error(keypoint_image <- draw_keypoints(image_uint, keypoints, colors = colors, radius = 7))
})

test_that("tensor_image_browse works", {
  skip_on_cran()
  skip_on_ci()
  # uint8 color image
  image <- (255 - (torch::torch_randint(low = 1, high = 200, size = c(3, 360, 360))))$to(torch::torch_uint8())
  expect_no_error(tensor_image_browse(image))
  # uint8 grayscale image
  image <- (255 - (torch::torch_randint(low = 1, high = 200, size = c(1, 360, 360))))$to(torch::torch_uint8())
  expect_no_error(tensor_image_browse(image))

  # float color image
  image <- torch::torch_rand(size = c(3, 360, 360))
  expect_no_error(tensor_image_browse(image))
  # float grayscale image
  image <- torch::torch_rand(size = c(1, 360, 360))
  expect_no_error(tensor_image_browse(image))

  # error cases : shape
  image <- torch::torch_randint(low = 1, high = 200, size = c(4, 3, 360, 360))$to(torch::torch_uint8())
  expect_error(tensor_image_browse(image), "Pass individual `image`, not batches")
  image <- torch::torch_randint(low = 1, high = 200, size = c(4, 360, 360))$to(torch::torch_uint8())
  expect_error(tensor_image_browse(image), "Only grayscale and RGB")

})

# ==== COCO sample drawing ====

test_that("draw_bounding_boxes works with coco_detection_sample", {
  skip_if_not(torch::torch_is_installed())

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- coco_detection_dataset(root = "~/data", train = FALSE, year = "2017", download = TRUE)
  item <- ds[3]

  out <- draw_bounding_boxes(item)
  expect_tensor(out)
  expect_equal(out$ndim, 3)
  expect_equal(out$shape[1], 3)  # 3 color channels
  expect_gt(out$shape[2], 100)   # image height is reasonable
  expect_gt(out$shape[3], 100)   # image width is reasonable
})

test_that("draw_segmentation_masks works with coco_detection_sample", {
  skip_if_not(torch::torch_is_installed())

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- coco_detection_dataset(root = "~/data", train = FALSE, year = "2017", download = TRUE)
  item <- ds[3]

  if (item$y$masks$size(1) > 0) {
    out <- draw_segmentation_masks(item)
    expect_tensor(out)
    expect_equal(out$ndim, 3)
    expect_equal(out$shape[1], 3)
    expect_gt(out$shape[2], 100)
    expect_gt(out$shape[3], 100)
  } else {
    skip("No masks in this item — skipping mask drawing test.")
  }
})

context("segmentation-transforms")

# Helper to create mock COCO target
mock_coco_target <- function(h = 100, w = 100) {
  list(
    boxes = torch::torch_tensor(matrix(c(10, 20, 50, 60), nrow = 1)),
    labels = "cat",
    area = torch::torch_tensor(c(800)),
    iscrowd = torch::torch_tensor(FALSE),
    segmentation = list(list(c(10, 20, 50, 20, 50, 60, 10, 60))),
    image_height = h,
    image_width = w
  )
}

# Helper to create mock trimap
mock_trimap <- function(h = 3, w = 3) {
  torch::torch_tensor(
    array(rep(c(1, 2, 3), length.out = h * w), dim = c(h, w)),
    dtype = torch::torch_int32()
  )
}

# COCO mask transformation tests

test_that("target_transform_coco_masks() converts polygons to masks", {
  skip_if_not_installed("torch")
  skip_if_not_installed("magick")
  y <- mock_coco_target()
  result <- target_transform_coco_masks(y)
  expect_true("masks" %in% names(result))
  expect_tensor(result$masks)
  expect_tensor_dtype(result$masks, torch::torch_bool())
  expect_equal(result$masks$ndim, 3)
  expect_true("segmentation" %in% names(result))
})

test_that("target_transform_coco_masks() handles empty segmentation", {
  skip_if_not_installed("torch")
  y <- list(
    boxes = torch::torch_zeros(c(0, 4)),
    labels = character(),
    segmentation = list(),
    image_height = 100,
    image_width = 100
  )
  result <- target_transform_coco_masks(y)
  expect_tensor(result$masks)
  expect_equal(result$masks$shape[1], 0)
})

test_that("target_transform_coco_masks() handles multiple objects", {
  skip_if_not_installed("torch")
  skip_if_not_installed("magick")
  y <- list(
    boxes = torch::torch_tensor(matrix(c(10,20,50,60, 60,10,90,40), nrow = 2, byrow = TRUE)),
    labels = c("cat", "dog"),
    segmentation = list(
      list(c(10, 20, 50, 20, 50, 60, 10, 60)),
      list(c(60, 10, 90, 10, 90, 40, 60, 40))
    ),
    image_height = 100,
    image_width = 100
  )
  result <- target_transform_coco_masks(y)
  expect_equal(result$masks$shape, c(2, 100, 100))
})

# Trimap mask transformation tests

test_that("target_transform_trimap_masks() converts trimap to masks", {
  skip_if_not_installed("torch")
  y <- list(trimap = mock_trimap(), label = 1L)
  result <- target_transform_trimap_masks(y)
  expect_true("masks" %in% names(result))
  expect_tensor(result$masks)
  expect_tensor_dtype(result$masks, torch::torch_bool())
  expect_equal(result$masks$shape[1], 3)
  expect_equal(result$label, 1L)
})

test_that("target_transform_trimap_masks() creates mutually exclusive masks", {
  skip_if_not_installed("torch")
  y <- list(trimap = mock_trimap(), label = 1L)
  result <- target_transform_trimap_masks(y)
  mask_sum <- result$masks$sum(dim = 1)
  expect_true(all(as.array(mask_sum) == 1))
})

# Dataset behavior tests (skipped unless TEST_LARGE_DATASETS=1)

test_that("coco_segmentation_dataset with target_transform produces masks", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Set TEST_LARGE_DATASETS=1 to enable")
  tmp <- withr::local_tempdir()
  ds <- coco_segmentation_dataset(root = tmp, train = FALSE, year = "2017",
                               download = TRUE, target_transform = target_transform_coco_masks)
  y <- ds[1]$y
  expect_true("masks" %in% names(y))
  expect_tensor(y$masks)
  expect_equal(y$masks$ndim, 3)
  expect_tensor_dtype(y$masks, torch::torch_bool())
})

# Integration with draw_segmentation_masks()

test_that("transformed masks work with draw_segmentation_masks()", {
  skip_if_not_installed("torch")
  skip_if_not_installed("magick")
  image <- torch::torch_randint(100, 200, c(3, 50, 50), dtype = torch::torch_uint8())
  y <- list(trimap = mock_trimap(50, 50), label = 1L)
  transformed_y <- target_transform_trimap_masks(y)
  item <- list(x = image, y = transformed_y)
  class(item) <- c("image_with_segmentation_mask", class(item))
  result <- draw_segmentation_masks(item)
  expect_tensor(result)
  expect_equal(result$shape[1], 3)
})

# Existing helper tests

test_that("coco_polygon_to_mask handles single polygon", {
  skip_if_not_installed("torch")
  skip_if_not_installed("magick")
  polygon <- list(c(10, 10, 50, 10, 50, 50, 10, 50))
  mask <- coco_polygon_to_mask(polygon, 100, 100)
  expect_tensor(mask)
  expect_tensor_dtype(mask, torch::torch_bool())
  expect_tensor_shape(mask, c(100, 100))
  expect_true(as.array(mask[30, 30]))
})

test_that("coco_polygon_to_mask handles empty polygon", {
  skip_if_not_installed("torch")
  skip_if_not_installed("magick")
  mask <- coco_polygon_to_mask(list(), 100, 100)
  expect_tensor(mask)
  expect_equal_to_r(mask$sum(), 0)
})

#' Flickr8k Dataset
#'
#' The Flickr8k and Flickr30k collections are **image captionning** datasets
#' composed of 8,000 and 30,000 color images respectively, each paired with five
#' human-annotated captions. The images are in RGB format with varying spatial
#' resolutions, and these datasets are widely used for training and evaluating
#' vision-language models.
#'
#' @inheritParams fgvc_aircraft_dataset
#' @param root : Root directory for dataset storage. The dataset will be stored under `root/flickr8k`.
#' @param train : If `TRUE`, loads the training set. If `FALSE`, loads the test set. Default is `TRUE`.
#'
#' @return A torch dataset of class \code{flickr8k_caption_dataset}.
#' Each element is a named list:
#' - `x`: a H x W x 3 integer array representing an RGB image.
#' - `y`: a character vector containing all five captions associated with the image.
#'
#' @examples
#' \dontrun{
#' # Load the Flickr8k caption dataset
#' flickr8k <- flickr8k_caption_dataset(download = TRUE)
#'
#' # Access the first item
#' first_item <- flickr8k[1]
#' first_item$x  # image array with shape {3, H, W}
#' first_item$y  # character vector containing five captions.
#'
#' # Load the Flickr30k caption dataset
#' flickr30k <- flickr30k_caption_dataset(download = TRUE)
#'
#' # Access the first item
#' first_item <- flickr30k[1]
#' first_item$x  # image array with shape {3, H, W}
#' first_item$y  # character vector containing five captions.
#' }
#'
#' @name flickr_caption_dataset
#' @title Flickr Caption Datasets
#' @rdname flickr_caption_dataset
#' @family caption_dataset
#' @export
flickr8k_caption_dataset <- torch::dataset(
  name = "flickr8k",
  training_file = "train.rds",
  test_file = "test.rds",
  class_index_file = "classes.rds",
  archive_size = "1 GB",

  resources = list(
    c("https://github.com/jbrownlee/Datasets/releases/download/Flickr8k/Flickr8k_text.zip", "bf6c1abcb8e4a833b7f922104de18627"),
    c("https://github.com/jbrownlee/Datasets/releases/download/Flickr8k/Flickr8k_Dataset.zip", "f18a1e2920de5bd84dae7cf08ec78978")
  ),

  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$train <- train
    self$split <- ifelse(train, "train", "test")

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      cli_abort("Dataset not found. Use `download = TRUE` to download it.")

    if (!self$check_processed_exists()) {
      fs::dir_create(self$processed_folder)

      caption_df <- read.delim(
        file.path(self$raw_folder, "Flickr8k.token.txt"),
        sep = "\t",
        header = FALSE,
        col.names = c("file_id", "caption"),
        stringsAsFactors = FALSE
      )
      caption_df[c("file", "id")] <- read.delim(
        text = caption_df$file,
        sep = "#",
        header = FALSE
      )

      caption_df$file_id <- NULL
      caption_df$file <- trimws(caption_df$file)
      captions_map <- split(caption_df$caption, caption_df$file)
      all_ids <- sort(unique(caption_df$file))

      caption_to_index <- setNames(seq_along(all_ids), all_ids)
      unique_merged_captions <- captions_map[all_ids]

      saveRDS(unique_merged_captions, file.path(self$processed_folder, self$class_index_file))

      train_ids <- readLines(file.path(self$raw_folder, "Flickr_8k.trainImages.txt"))
      test_ids <- readLines(file.path(self$raw_folder, "Flickr_8k.testImages.txt"))

      process_split <- function(ids) {
        ids <- trimws(ids)
        img_paths <- file.path(self$raw_folder, "Flicker8k_Dataset", ids)
        img_paths <- img_paths[file.exists(img_paths)]

        missing <- setdiff(ids, names(caption_to_index))
        if (length(missing) > 0) {
          cli_abort("The following IDs are missing captions: {paste(missing, collapse=', ')}")
        }

        caption_indices <- vapply(ids, function(id) caption_to_index[[id]], integer(1))
        list(images = img_paths, captions = caption_indices)
      }

      saveRDS(process_split(train_ids), file.path(self$processed_folder, self$training_file))
      saveRDS(process_split(test_ids), file.path(self$processed_folder, self$test_file))
    }

    file <- if (train) self$training_file else self$test_file
    data <- readRDS(file.path(self$processed_folder, file))
    self$images <- data$images
    self$captions <- data$captions
    self$classes <- readRDS(file.path(self$processed_folder, self$class_index_file))

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {

    if (self$check_exists())
      return()

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    fs::dir_create(self$raw_folder)

    for (r in self$resources) {
      archive <- download_and_cache(r[1], prefix = class(self)[1])
      if (tools::md5sum(archive) != r[2]) {
        cli_abort("Corrupt file! Delete the file at {archive} and try again.")
      }
      if (tools::file_ext(archive) == "zip") {
        utils::unzip(archive, exdir = self$raw_folder)
      } else if (tools::file_ext(archive) == "gz") {
        tar_path <- sub("\\.gz$", "", archive)
        R.utils::gunzip(archive, tar_path, overwrite = TRUE)
        utils::untar(tar_path, exdir = self$raw_folder)
      }
    }

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")

  },

  check_processed_exists = function() {
    fs::file_exists(file.path(self$processed_folder, self$training_file)) &&
    fs::file_exists(file.path(self$processed_folder, self$test_file)) &&
    fs::file_exists(file.path(self$processed_folder, self$class_index_file))
  },

  check_exists = function() {
    fs::file_exists(file.path(self$raw_folder, "Flickr8k.token.txt")) &&
    fs::file_exists(file.path(self$raw_folder, "Flickr_8k.trainImages.txt")) &&
    fs::file_exists(file.path(self$raw_folder, "Flickr_8k.testImages.txt")) &&
    fs::dir_exists(file.path(self$raw_folder, "Flicker8k_Dataset"))
  },

  .getitem = function(index) {
    x <- jpeg::readJPEG(self$images[[index]])
    caption_index <- self$captions[[index]]
    y <- self$classes[[caption_index]]

    if (!is.null(self$transform))
      x <- self$transform(x)

    if (!is.null(self$target_transform))
      y <- self$target_transform(y)

    list(x = x, y = y)
  },

  .length = function() {
    length(self$images)
  },

  active = list(
    raw_folder = function() file.path(self$root, "flickr8k", "raw"),
    processed_folder = function() file.path(self$root, "flickr8k", "processed")
  )
)


#' Flickr30k Dataset
#'
#' @inheritParams flickr8k_caption_dataset
#' @param root Character. Root directory where the dataset will be stored under `root/flickr30k`.
#'
#' @return A torch dataset of class \code{flickr30k_caption_dataset}.
#' Each element is a named list:
#' - `x`: a H x W x 3 integer array representing an RGB image.
#' - `y`: a character vector containing all five captions associated with the image.
#'
#' @rdname flickr_caption_dataset
#' @export
flickr30k_caption_dataset <- torch::dataset(
  name = "flickr30k",
  inherit = flickr8k_caption_dataset,
  archive_size = "4.1 GB",
  resources = list(
    c("https://uofi.app.box.com/shared/static/1cpolrtkckn4hxr1zhmfg0ln9veo6jpl.gz", "985ac761bbb52ca49e0c474ae806c07c"),
    c("https://cs.stanford.edu/people/karpathy/deepimagesent/caption_datasets.zip", "4fa8c08369d22fe16e41dc124bd1adc2")
  ),

  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
    self$root <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$train <- train
    self$split <- ifelse(train, "train", "test")

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      cli_abort("Dataset not found. Use `download = TRUE` to download it.")

    captions_path <- file.path(self$raw_folder, "dataset_flickr30k.json")
    captions_json <- jsonlite::fromJSON(captions_path)

    imgs_df <- captions_json$images
    filtered <- imgs_df[imgs_df$split == self$split, ]
    self$filenames <- filtered$filename

    captions_map <- setNames(
      lapply(filtered$sentences, function(s) unname(vapply(s$raw, identity, character(1)))),
      filtered$filename
    )

    all_ids <- names(captions_map)
    caption_to_index <- setNames(seq_along(all_ids), all_ids)

    self$images <- file.path(self$raw_folder, "flickr30k-images", self$filenames)
    self$captions <- vapply(self$filenames, function(f) caption_to_index[[f]], integer(1))
    self$classes <- captions_map

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  check_exists = function() {
    fs::file_exists(file.path(self$raw_folder, "dataset_flickr30k.json")) &&
    fs::dir_exists(file.path(self$raw_folder, "flickr30k-images"))
  },

  active = list(
    raw_folder = function() file.path(self$root, "flickr30k", "raw")
  )
)

rpn_head <- torch::nn_module(
    "rpn_head",
    initialize = function(in_channels, num_anchors = 3) {
      self$conv <- nn_sequential(nn_sequential(nn_conv2d(in_channels, in_channels, kernel_size = 3, padding = 1)))
      self$cls_logits <- nn_conv2d(in_channels, num_anchors, kernel_size = 1)
      self$bbox_pred <- nn_conv2d(in_channels, num_anchors * 4, kernel_size = 1)
    },
    forward = function(features) {
      objectness_list <- list()
      bbox_reg_list <- list()

      for (i in seq_along(features)) {
        x <- features[[i]]
        t <- torch::nnf_relu(self$conv(x))
        objectness_list[[i]] <- self$cls_logits(t)
        bbox_reg_list[[i]] <- self$bbox_pred(t)
      }

      list(objectness = objectness_list, bbox_deltas = bbox_reg_list)
    }
  )


rpn_head_v2 <- torch::nn_module(
    "rpn_head_v2",
    initialize = function(in_channels, num_anchors = 3) {
      # PyTorch V2 uses Conv2dNormActivation with norm_layer=None, which creates
      # Conv2d with bias=True, NO batch norm, and ReLU activation.
      # This matches torchvision.ops.misc.Conv2dNormActivation behavior when norm_layer=None.
      block <- function() {
        nn_sequential(
          nn_conv2d(in_channels, in_channels, kernel_size = 3, padding = 1, bias = TRUE),
          nn_relu()
        )
      }
      self$conv <- nn_sequential(block(), block())
      self$cls_logits <- nn_conv2d(in_channels, num_anchors, kernel_size = 1, bias = TRUE)
      self$bbox_pred <- nn_conv2d(in_channels, num_anchors * 4, kernel_size = 1, bias = TRUE)
    },
    forward = function(features) {
      objectness_list <- list()
      bbox_reg_list <- list()

      for (i in seq_along(features)) {
        x <- features[[i]]
        t <- self$conv(x)
        objectness_list[[i]] <- self$cls_logits(t)
        bbox_reg_list[[i]] <- self$bbox_pred(t)
      }

      list(objectness = objectness_list, bbox_deltas = bbox_reg_list)
    }
  )


rpn_head_mobilenet <- torch::nn_module(
    initialize = function(in_channels, num_anchors = 15) {
      self$conv <- nn_sequential(nn_sequential(nn_conv2d(in_channels, in_channels, kernel_size = 3, padding = 1)))
      self$cls_logits <- nn_conv2d(in_channels, num_anchors, kernel_size = 1)
      self$bbox_pred <- nn_conv2d(in_channels, num_anchors * 4, kernel_size = 1)
    },
    forward = function(features) {
      objectness <- list()
      bbox_deltas <- list()

      for (x in features) {
        t <- nnf_relu(self$conv(x))
        objectness[[length(objectness) + 1]] <- self$cls_logits(t)
        bbox_deltas[[length(bbox_deltas) + 1]] <- self$bbox_pred(t)
      }

      list(objectness = objectness, bbox_deltas = bbox_deltas)
    }
  )


#' @importFrom torch torch_meshgrid torch_stack torch_tensor torch_stack torch_zeros_like torch_max torch_float32
generate_level_anchors <- function(h, w, stride, base_sizes = c(32, 64, 128, 256, 512), aspect_ratios = c(0.5, 1.0, 2.0), device = "cpu") {
  # Grid centers
  shift_x <- torch_arange(0.5, w - 0.5, 1.0, device = device) * stride
  shift_y <- torch_arange(0.5, h - 0.5, 1.0, device = device) * stride
  shifts <- torch_meshgrid(list(shift_x, shift_y), indexing = "xy")
  shift_grid <- torch_stack(list(shifts[[1]], shifts[[2]], torch_zeros_like(shifts[[1]]), torch_zeros_like(shifts[[2]])), dim = 3)$unsqueeze(3)  # [H, W, 1, 4]

  # Create anchors for each combination of base_size and aspect_ratio
  anchor_list <- list()

  for (base_size in base_sizes) {
    for (ratio in aspect_ratios) {
      w_anchor <- base_size * sqrt(ratio)
      h_anchor <- base_size / sqrt(ratio)
      anchor_list[[length(anchor_list) + 1]] <- c(w_anchor, h_anchor)
    }
  }

  # Stack anchors [A, 2]
  anchors_wh <- do.call(rbind, anchor_list)
  anchor_widths <- torch_tensor(anchors_wh[, 1], device = device)
  anchor_heights <- torch_tensor(anchors_wh[, 2], device = device)

  # Create base anchors [4, A] then transpose to [A, 4] (xc, yc, w, h)
  anchors <- torch_stack(list(
    torch_zeros_like(anchor_widths),
    torch_zeros_like(anchor_heights),
    anchor_widths,
    anchor_heights
  ), dim = 1)$permute(c(2, 1))

  # Expand to [H, W, A, 4]
  anchors <- anchors$reshape(c(1, 1, -1, 4)) + shift_grid  # Broadcasting

  # Convert from (xc, yc, w, h) to (x1, y1, x2, y2)
  orig_shape <- anchors$shape
  anchors <- anchors$reshape(c(-1, 4))
  anchors <- box_cxcywh_to_xyxy(anchors)
  anchors$reshape(orig_shape)
}

decode_boxes <- function(anchors, deltas, weights = c(10, 10, 5, 5)) {
  widths <- anchors[, 3] - anchors[, 1]
  heights <- anchors[, 4] - anchors[, 2]
  ctr_x <- anchors[, 1] + widths / 2
  ctr_y <- anchors[, 2] + heights / 2

  # Apply bbox regression weights (default: 10, 10, 5, 5)
  dx <- deltas[, 1] / weights[1]
  dy <- deltas[, 2] / weights[2]
  dw <- deltas[, 3] / weights[3]
  dh <- deltas[, 4] / weights[4]

  # Clamp dw and dh to prevent numerical instability in exp() to log(1000 / 16)  # ~4.135
  dw <- torch::torch_clamp(dw, max = 4.135)
  dh <- torch::torch_clamp(dh, max = 4.135)

  pred_ctr_x <- ctr_x + dx * widths
  pred_ctr_y <- ctr_y + dy * heights
  pred_w <- torch::torch_exp(dw) * widths
  pred_h <- torch::torch_exp(dh) * heights

  x1 <- pred_ctr_x - pred_w / 2
  y1 <- pred_ctr_y - pred_h / 2
  x2 <- pred_ctr_x + pred_w / 2
  y2 <- pred_ctr_y + pred_h / 2

  torch::torch_stack(list(x1, y1, x2, y2), dim = 2)
}

#' Postprocess Detections
#'
#' Processes class logits and box regression to produce final detections.
#' Keeps all class predictions before filtering, rather than taking max across
#' classes first, allowing proper per-class NMS.
#'
#' @param class_logits Tensor of shape \[N, num_classes_with_bg\] - raw classification scores including background
#' @param box_regression Tensor of shape \[N, num_classes_with_bg * 4\] - box deltas for each class
#' @param proposals Tensor of shape \[N, 4\] - proposal boxes in (x1, y1, x2, y2) format
#' @param image_size Integer vector of length 2: \[height, width\]
#' @param num_classes Integer - number of classes excluding background (e.g., 90 for COCO)
#' @param score_thresh Numeric - minimum score threshold for detections
#' @param nms_thresh Numeric - NMS IoU threshold
#' @param detections_per_img Integer - maximum detections to return
#'
#' @details
#' Box regression uses standard Faster R-CNN decoding with weights (10, 10, 5, 5).
#' Background class (index 1 in class_logits) is removed, returning labels
#' in range \[1, num_classes\] which correspond to COCO classes \[1=person, 2=bicycle, ..., 90=toothbrush\].
#'
#' @return List with boxes, labels, scores tensors
#' @noRd
postprocess_detections <- function(class_logits, box_regression, proposals,
                                   image_size, num_classes, score_thresh,
                                   nms_thresh, detections_per_img) {
  device <- class_logits$device
  num_proposals <- proposals$shape[1]

  # class_logits and box_regression include background class
  # num_classes excludes background, so total classes = num_classes + 1
  num_classes_with_bg <- num_classes + 1L

  # Decode boxes for all classes: [N, num_classes_with_bg, 4]
  pred_boxes <- box_regression$view(c(num_proposals, num_classes_with_bg, 4))

  # Decode boxes using proposals as anchors (expand proposals for all classes)
  proposals_expanded <- proposals$unsqueeze(2)$expand(c(num_proposals, num_classes_with_bg, 4))

  # Decode for each class
  decoded_boxes <- torch_zeros_like(pred_boxes)
  for (cls_idx in seq_len(num_classes_with_bg)) {
    decoded_boxes[, cls_idx, ] <- decode_boxes(proposals, pred_boxes[, cls_idx, ])
  }

  # Clip to image boundaries
  decoded_boxes <- clip_boxes_to_image(decoded_boxes$view(c(-1, 4)), image_size)
  decoded_boxes <- decoded_boxes$view(c(num_proposals, num_classes_with_bg, 4))

  # Apply softmax to get class probabilities: [N, num_classes_with_bg]
  pred_scores <- torch::nnf_softmax(class_logits, dim = 2)

  # Remove background class (class 1 in 1-based indexing)
  # Keep classes 2 through num_classes_with_bg
  pred_boxes <- decoded_boxes[, 2:num_classes_with_bg, ]   # [N, num_classes, 4]
  pred_scores <- pred_scores[, 2:num_classes_with_bg]       # [N, num_classes]

  # Create labels for each prediction: [num_classes]
  # These are the actual class indices (1, 2, 3, ... num_classes)
  # torch_arange in R is inclusive on both ends
  labels <- torch_arange(1L, num_classes, device = device, dtype = torch_long())
  # Expand to match scores: [N, num_classes]
  labels <- labels$view(c(1, -1))$expand_as(pred_scores)

  # Flatten everything: treat each (proposal, class) pair as a separate detection
  boxes <- pred_boxes$reshape(c(-1, 4))           # [N * num_classes, 4]
  scores <- pred_scores$reshape(c(-1))            # [N * num_classes]
  labels <- labels$reshape(c(-1))                 # [N * num_classes]

  # Remove low scoring boxes
  keep <- scores > score_thresh
  boxes <- boxes[keep, ]
  scores <- scores[keep]
  labels <- labels[keep]

  if (boxes$shape[1] == 0) {
    return(list(
      boxes = torch_empty(c(0, 4), device = device),
      labels = torch_empty(c(0), dtype = torch_long(), device = device),
      scores = torch_empty(c(0), device = device)
    ))
  }

  # Remove small boxes
  keep <- remove_small_boxes(boxes, min_size = 1e-2)
  boxes <- boxes[keep, ]
  scores <- scores[keep]
  labels <- labels[keep]

  if (boxes$shape[1] == 0) {
    return(list(
      boxes = torch_empty(c(0, 4), device = device),
      labels = torch_empty(c(0), dtype = torch_long(), device = device),
      scores = torch_empty(c(0), device = device)
    ))
  }

  # Apply per-class NMS
  keep <- batched_nms(boxes, scores, labels, nms_thresh)

  # Keep only top k detections
  if (keep$shape[1] > detections_per_img) {
    keep <- keep[1:detections_per_img]
  }

  boxes <- boxes[keep, ]
  scores <- scores[keep]
  labels <- labels[keep]

  list(boxes = boxes, labels = labels, scores = scores)
}

#' @importFrom torch nnf_grid_sample torch_empty
generate_proposals <- function(features, rpn_out, image_size, strides, batch_idx,
                               score_thresh = 0.05, nms_thresh = 0.7) {
  device <- rpn_out$objectness[[1]]$device
  all_proposals <- torch_empty(0L, 4L, device = device)
  all_scores <- torch_empty(0L, device = device)

  for (i in seq_along(features)) {
    objectness <- rpn_out$objectness[[i]][batch_idx, , , ]
    deltas <- rpn_out$bbox_deltas[[i]][batch_idx, , , ]

    c(a, h, w) %<-% objectness$shape

    if (a == 15) {
      base_sizes <- c(32, 64, 128, 256, 512)
    } else if (a == 3) {
      base_sizes <- c(4 * strides[[i]])
    } else {
      value_error("Unexpected number of anchors: {a}. Expected 3 or 15.")
    }
    anchors <- generate_level_anchors(h, w, strides[[i]], base_sizes = base_sizes, aspect_ratios = c(0.5, 1, 2), device = device)
    anchors <- anchors$reshape(c(-1, 4))  # [H*W*A, 4]

    objectness <- objectness$sigmoid()$flatten() ## [H*W*A]
    deltas <- deltas$permute(c(2, 3, 1))$reshape(c(-1, 4))  # [H*W*A, 4]

    proposals <- decode_boxes(anchors, deltas)
    proposals <- clip_boxes_to_image(proposals, image_size)

    all_proposals <- torch::torch_cat(list(all_proposals, proposals), dim = 1L)
    all_scores <- torch::torch_cat(list(all_scores, objectness), dim = 1L)
  }

  scores <- all_scores$flatten()
  keep <- scores > score_thresh
  proposals <- all_proposals[keep, ]
  scores <- scores[keep]

  if (proposals$shape[1] > 0) {
    keep_idx <- nms(proposals, scores, nms_thresh)
    proposals <- proposals[keep_idx, ]
  } else {
    proposals <- torch_empty(c(0, 4), device = device, dtype = torch_float32())
  }

  list(proposals = proposals)
}

roi_align <- function(feature_map, proposals, batch_idx, output_size = c(7L, 7L)) {
  # A vectorized version of roi_align_stub for feature_map: [B, C, H, W] and proposals: [N, 4] (x1, y1, x2, y2)

  num_rois <- proposals$size(1)
  if (num_rois == 0) {
    return(torch_empty(c(0, feature_map$size(2), output_size[1], output_size[2]), device = feature_map$device))
  }

  channels <- feature_map$size(2)
  h_feat <- feature_map$size(3)
  w_feat <- feature_map$size(4)

  # Normalize coordinnates to match grid_sample [-1 et 1]
  x1 <- (proposals[, 1] / (w_feat - 1) * 2) - 1
  y1 <- (proposals[, 2] / (h_feat - 1) * 2) - 1
  x2 <- (proposals[, 3] / (w_feat - 1) * 2) - 1
  y2 <- (proposals[, 4] / (h_feat - 1) * 2) - 1

  # Create a grid of output_size
  grid_y <- torch_linspace(0, 1, output_size[1], device = feature_map$device)
  grid_x <- torch_linspace(0, 1, output_size[2], device = feature_map$device)

  # Meshgrid to get relative coordiantes in [7, 7]
  grids <- torch_meshgrid(list(grid_y, grid_x), indexing = "ij")
  rel_y <- grids[[1]]
  rel_x <- grids[[2]]

  # Linear interpolation for each ROI [N, 7, 7]
  # x <- x1 + rel_x * (x2 - x1)
  sampling_x <- x1$view(c(-1, 1, 1)) + rel_x$view(c(1, output_size[1], output_size[2])) * (x2 - x1)$view(c(-1, 1, 1))
  sampling_y <- y1$view(c(-1, 1, 1)) + rel_y$view(c(1, output_size[1], output_size[2])) * (y2 - y1)$view(c(-1, 1, 1))

  # Clamp grid to [-1, 1] to avoid needing padding (especially for MPS compatibility)
  sampling_x <- sampling_x$clamp(-1, 1)
  sampling_y <- sampling_y$clamp(-1, 1)

  # Concat to get a grid of [N, 7, 7, 2]
  grid <- torch_stack(list(sampling_x, sampling_y), dim = -1)

  # bilinear sampling
  input_selected <- feature_map[batch_idx, , , ]$unsqueeze(1)$expand(c(num_rois, channels, h_feat, w_feat))

  pooled_features <- nnf_grid_sample(
    input_selected,
    grid,
    mode = "bilinear",
    padding_mode = "zeros",
    align_corners = FALSE
  )

  # Return [N, C, 7, 7]
  pooled_features
}

roi_heads_module <-  torch::nn_module(
    "region-of-interest head",
    initialize = function(num_classes = 90) {
      # num_classes excludes background, but predictor needs background + all classes
      num_classes_with_bg <- num_classes + 1L

      # Define box_head with named layers to match expected state dict structure
      self$box_head <- torch::nn_module(
        initialize = function() {
          self$fc6 <- torch::nn_linear(256 * 7 * 7, 1024, bias = TRUE)
          self$fc7 <- torch::nn_linear(1024, 1024, bias = TRUE)
        },
        forward = function(x) {
          x <- torch::nnf_relu(self$fc6(x))
          x <- torch::nnf_relu(self$fc7(x))
          x
        }
      )()

      self$box_predictor <- torch::nn_module(
        initialize = function() {
          self$cls_score <- torch::nn_linear(1024, num_classes_with_bg, bias = TRUE)
          self$bbox_pred <- torch::nn_linear(1024, num_classes_with_bg * 4, bias = TRUE)
        },
        forward = function(x) {
          list(
            scores = self$cls_score(x),
            boxes = self$bbox_pred(x)
          )
        }
      )()
    },
    forward = function(features, proposals, batch_idx) {
      feature_maps <- features[c("p2", "p3", "p4", "p5")]
      pooled <- roi_align(feature_maps[[1]], proposals, batch_idx)
      x <- self$box_head(pooled$flatten(start_dim = 2))
      self$box_predictor(x)
    }
  )

roi_heads_module_v2 <-  torch::nn_module(
    "region-of-interest head v2",
    initialize = function(num_classes = 90, in_channels = 256) {
      # num_classes excludes background, but predictor needs background + all classes
      num_classes_with_bg <- num_classes + 1L

      # V2 uses FastRCNNConvFCHead: 4 conv layers + 1 FC layer
      # PyTorch: FastRCNNConvFCHead((256, 7, 7), [256, 256, 256, 256], [1024], norm_layer=BatchNorm2d)
      self$box_head <- torch::nn_module(
        initialize = function() {
          conv_block <- function(in_ch, out_ch) {
            nn_sequential(
              nn_conv2d(in_ch, out_ch, kernel_size = 3, padding = 1, bias = TRUE),
              nn_batch_norm2d(out_ch),
              nn_relu()
            )
          }
          # State dict expects numeric indices: 0, 1, 2, 3, 4
          # Use character names that will be accessed as self[["0"]], etc.
          self[["0"]] <- conv_block(in_channels, 256)
          self[["1"]] <- conv_block(256, 256)
          self[["2"]] <- conv_block(256, 256)
          self[["3"]] <- conv_block(256, 256)
          self[["4"]] <- nn_linear(256 * 7 * 7, 1024, bias = TRUE)
        },
        forward = function(x) {
          # x is [N, 256, 7, 7]
          x <- self[["0"]](x)
          x <- self[["1"]](x)
          x <- self[["2"]](x)
          x <- self[["3"]](x)
          x <- x$flatten(start_dim = 2)  # [N, 256*7*7]
          x <- self[["4"]](x)            # [N, 1024]
          x <- nnf_relu(x)
          x
        }
      )()

      self$box_predictor <- torch::nn_module(
        initialize = function() {
          self$cls_score <- torch::nn_linear(1024, num_classes_with_bg, bias = TRUE)
          self$bbox_pred <- torch::nn_linear(1024, num_classes_with_bg * 4, bias = TRUE)
        },
        forward = function(x) {
          list(
            scores = self$cls_score(x),
            boxes = self$bbox_pred(x)
          )
        }
      )()
    },
    forward = function(features, proposals, batch_idx) {
      feature_maps <- features[c("p2", "p3", "p4", "p5")]
      pooled <- roi_align(feature_maps[[1]], proposals, batch_idx)
      x <- self$box_head(pooled)
      self$box_predictor(x)
    }
  )


fpn_module <- torch::nn_module(
    "feature pyramid network",
    initialize = function(in_channels, out_channels) {
      self$inner_blocks <- nn_module_list(lapply(in_channels, function(c) {
        nn_sequential(torch::nn_conv2d(c, out_channels, kernel_size = 1))
      }))
      self$layer_blocks <- nn_module_list(lapply(rep(out_channels, 4), function(i) {
        nn_sequential(torch::nn_conv2d(out_channels, out_channels, kernel_size = 3, padding = 1))
      }))
    },
    forward = function(inputs) {
      names(inputs) <- c("c2", "c3", "c4", "c5")

      last_inner <- self$inner_blocks[[4]](inputs$c5)
      results <- list()
      results[[4]] <- self$layer_blocks[[4]](last_inner)

      for (i in 3:1) {
        lateral <- self$inner_blocks[[i]](inputs[[i]])
        target_size <- as.integer(lateral$shape[3:4])
        upsampled <- torch::nnf_interpolate(last_inner, size = target_size, mode = "nearest")
        last_inner <- lateral + upsampled
        results[[i]] <- self$layer_blocks[[i]](last_inner)
      }

      names(results) <- c("p2", "p3", "p4", "p5")
      results
    }
  )


resnet_fpn_backbone <- function(pretrained = TRUE) {
  resnet <- model_resnet50(pretrained = pretrained)

  resnet_body <- torch::nn_module(
    initialize = function() {
      self$conv1 <- resnet$conv1
      self$bn1 <- resnet$bn1
      self$relu <- resnet$relu
      self$maxpool <- resnet$maxpool
      self$layer1 <- resnet$layer1
      self$layer2 <- resnet$layer2
      self$layer3 <- resnet$layer3
      self$layer4 <- resnet$layer4
    },
    forward = function(x) {
      c2 <- x %>%
        self$conv1() %>%
        self$bn1() %>%
        self$relu() %>%
        self$maxpool() %>%
        self$layer1()

      c3 <- self$layer2(c2)
      c4 <- self$layer3(c3)
      c5 <- self$layer4(c4)

      list(c2, c3, c4, c5)
    }
  )

  backbone <- torch::nn_module(
    initialize = function() {
      self$body <- resnet_body()
      self$fpn <- fpn_module(
        in_channels = c(256, 512, 1024, 2048),
        out_channels = 256
      )
    },
    forward = function(x) {
      c2_to_c5 <- self$body(x)
      self$fpn(c2_to_c5)
    }
  )

  backbone <- backbone()
  backbone$out_channels <- 256
  backbone
}


fasterrcnn_model <- torch::nn_module(
    initialize = function(backbone, num_classes,
                          score_thresh = 0.05,
                          nms_thresh = 0.5,
                          detections_per_img = 100) {
      self$backbone <- backbone
      self$num_classes <- num_classes

      # Store configurable detection parameters
      self$score_thresh <- score_thresh
      self$nms_thresh <- nms_thresh
      self$detections_per_img <- detections_per_img

      self$rpn <- torch::nn_module(
        initialize = function() {
          self$head <- rpn_head(in_channels = backbone$out_channels)
        },
        forward = function(features) {
          self$head(features)
        }
      )()

      # Use the roi_heads_module instead of inline definition
      self$roi_heads <- roi_heads_module(num_classes = num_classes)
    },

    forward = function(images) {
      features <- self$backbone(images)
      rpn_out <- self$rpn(features)

      batch_size <- images$shape[1]
      image_size <- images$shape[3:4]
      final_results <- list()

      for (b in seq_len(batch_size)) {
        props <- generate_proposals(features, rpn_out, image_size, c(4, 8, 16, 32),
                                    batch_idx = b, score_thresh = 0,
                                    nms_thresh = self$nms_thresh)

        if (props$proposals$shape[1] == 0) {
          empty <- list(
            boxes = torch_empty(c(0, 4)),
            labels = torch_empty(c(0), dtype = torch::torch_long()),
            scores = torch_empty(c(0))
          )
          final_results[[b]] <- empty
          next
        }

        # Get ROI head predictions
        roi_out <- self$roi_heads(features, props$proposals, batch_idx = b)

        # Postprocess detections
        result <- postprocess_detections(
          class_logits = roi_out$scores,
          box_regression = roi_out$boxes,
          proposals = props$proposals,
          image_size = image_size,
          num_classes = self$num_classes,
          score_thresh = self$score_thresh,
          nms_thresh = self$nms_thresh,
          detections_per_img = self$detections_per_img
        )

        final_results[[b]] <- result
      }
      list(features = features, detections = final_results)
    }
  )



fpn_module_v2 <- torch::nn_module(
    "feature pyramid network v2",
    initialize = function(in_channels, out_channels) {
      self$inner_blocks <- nn_module_list(lapply(in_channels, function(c) {
        nn_sequential(
          nn_conv2d(c, out_channels, kernel_size = 1, bias = FALSE),
          nn_batch_norm2d(out_channels),
          nn_relu()
        )
      }))
      self$layer_blocks <- nn_module_list(lapply(rep(out_channels, 4), function(i) {
        nn_sequential(
          nn_conv2d(out_channels, out_channels, kernel_size = 3, padding = 1, bias = FALSE),
          nn_batch_norm2d(out_channels),
          nn_relu()
        )
      }))
    },
    forward = function(inputs) {
      names(inputs) <- c("c2", "c3", "c4", "c5")

      last_inner <- self$inner_blocks[[4]](inputs$c5)
      results <- list()
      results[[4]] <- self$layer_blocks[[4]](last_inner)

      for (i in 3:1) {
        lateral <- self$inner_blocks[[i]](inputs[[i]])
        target_size <- as.integer(lateral$shape[3:4])
        upsampled <- torch::nnf_interpolate(last_inner, size = target_size, mode = "nearest")
        last_inner <- lateral + upsampled
        results[[i]] <- self$layer_blocks[[i]](last_inner)
      }

      names(results) <- c("p2", "p3", "p4", "p5")
      results
    }
  )


resnet_fpn_backbone_v2 <- function(pretrained = TRUE) {
  resnet <- model_resnet50(pretrained = pretrained)

  resnet_body <- torch::nn_module(
    initialize = function() {
      self$conv1 <- resnet$conv1
      self$bn1 <- resnet$bn1
      self$relu <- resnet$relu
      self$maxpool <- resnet$maxpool
      self$layer1 <- resnet$layer1
      self$layer2 <- resnet$layer2
      self$layer3 <- resnet$layer3
      self$layer4 <- resnet$layer4
    },
    forward = function(x) {
       c2 <- x %>%
        self$conv1() %>%
        self$bn1() %>%
        self$relu() %>%
        self$maxpool() %>%
        self$layer1()

      c3 <- self$layer2(c2)
      c4 <- self$layer3(c3)
      c5 <- self$layer4(c4)

      list(c2, c3, c4, c5)
    }
  )

  backbone <- torch::nn_module(
    initialize = function() {
      self$body <- resnet_body()
      self$fpn <- fpn_module_v2(
        in_channels = c(256, 512, 1024, 2048),
        out_channels = 256
      )
    },
    forward = function(x) {
      c2_to_c5 <- self$body(x)
      self$fpn(c2_to_c5)
    }
  )

  backbone <- backbone()
  backbone$out_channels <- 256
  backbone
}


fasterrcnn_model_v2 <- torch::nn_module(
    initialize = function(backbone, num_classes,
                          score_thresh = 0.05,
                          nms_thresh = 0.5,
                          detections_per_img = 100) {
      self$backbone <- backbone
      self$num_classes <- num_classes

      # Store configurable detection parameters
      self$score_thresh <- score_thresh
      self$nms_thresh <- nms_thresh
      self$detections_per_img <- detections_per_img

      self$rpn <- torch::nn_module(
        initialize = function() {
          self$head <- rpn_head_v2(in_channels = backbone$out_channels)
        },
        forward = function(features) {
          self$head(features)
        }
      )()
      self$roi_heads <- roi_heads_module_v2(num_classes = num_classes)
    },
    forward = function(images) {
      features <- self$backbone(images)
      rpn_out <- self$rpn(features)

      batch_size <- images$shape[1]
      image_size <- images$shape[3:4]
      final_results <- list()

      for (b in seq_len(batch_size)) {
        props <- generate_proposals(features, rpn_out, image_size, c(4, 8, 16, 32),
                                    batch_idx = b, score_thresh = 0,
                                    nms_thresh = self$nms_thresh)

        if (props$proposals$shape[1] == 0) {
          empty <- list(
            boxes = torch_empty(c(0, 4)),
            labels = torch_empty(c(0), dtype = torch::torch_long()),
            scores = torch_empty(c(0))
          )
          final_results[[b]] <- empty
          next
        }

        # Get ROI head predictions
        roi_out <- self$roi_heads(features, props$proposals, batch_idx = b)

        # Postprocess detections
        result <- postprocess_detections(
          class_logits = roi_out$scores,
          box_regression = roi_out$boxes,
          proposals = props$proposals,
          image_size = image_size,
          num_classes = self$num_classes,
          score_thresh = self$score_thresh,
          nms_thresh = self$nms_thresh,
          detections_per_img = self$detections_per_img
        )

        final_results[[b]] <- result
      }
      list(features = features, detections = final_results)
    }
  )



fpn_module_2level <- torch::nn_module(
    initialize = function(in_channels, out_channels) {
      self$inner_blocks <- nn_module_list(lapply(in_channels, function(c) {
        nn_sequential(nn_conv2d(c, out_channels, kernel_size = 1))
      }))
      self$layer_blocks <- nn_module_list(lapply(rep(out_channels, 2), function(i) {
        nn_sequential(nn_conv2d(out_channels, out_channels, kernel_size = 3, padding = 1))
      }))
    },
    forward = function(inputs) {
      last_inner <- self$inner_blocks[[2]](inputs[[2]])
      results <- list()
      results[[2]] <- self$layer_blocks[[2]](last_inner)

      lateral <- self$inner_blocks[[1]](inputs[[1]])
      upsampled <- nnf_interpolate(last_inner, size = as.integer(lateral$shape[3:4]), mode = "nearest")
      last_inner <- lateral + upsampled
      results[[1]] <- self$layer_blocks[[1]](last_inner)

      names(results) <- c("p1", "p2")
      results
    }
  )



mobilenet_v3_fpn_backbone <- function(pretrained = TRUE) {
  mobilenet <- model_mobilenet_v3_large(pretrained = pretrained)

  backbone_module <- torch::nn_module(
    initialize = function() {
      self$body <- mobilenet$features
      self$fpn <- fpn_module_2level(
        in_channels = c(160, 960),
        out_channels = 256
      )
    },
    forward = function(x) {
      all_feats <- list()

      for (i in seq_len(length(self$body))) {
        x <- self$body[[i]](x)
        all_feats[[i]] <- x
      }

      feats <- list(
        all_feats[[14]],  # 160 channels
        all_feats[[17]]   # 960 channels
      )

      self$fpn(feats)
    }
  )

  backbone <- backbone_module()
  backbone$out_channels <- 256
  backbone
}

fasterrcnn_mobilenet_model <- torch::nn_module(
    initialize = function(backbone, num_classes,
                          score_thresh = 0.05,
                          nms_thresh = 0.5,
                          detections_per_img = 100) {
      self$backbone <- backbone
      self$num_classes <- num_classes

      # Store configurable detection parameters
      self$score_thresh <- score_thresh
      self$nms_thresh <- nms_thresh
      self$detections_per_img <- detections_per_img

      self$rpn <- torch::nn_module(
        initialize = function() {
          self$head <- rpn_head_mobilenet(in_channels = backbone$out_channels)
        },
        forward = function(features) {
          self$head(features)
        }
      )()
      self$roi_heads <- roi_heads_module(num_classes = num_classes)
    },
    forward = function(images) {
      features <- self$backbone(images)
      rpn_out <- self$rpn(features)

      batch_size <- images$shape[1]
      image_size <- images$shape[3:4]
      final_results <- list()

      for (b in seq_len(batch_size)) {
        props <- generate_proposals(features, rpn_out, image_size, c(8, 16),
                                    batch_idx = b, score_thresh = 0,
                                    nms_thresh = self$nms_thresh)

        if (props$proposals$shape[1] == 0) {
          empty <- list(
            boxes = torch_empty(c(0, 4)),
            labels = torch_empty(c(0), dtype = torch::torch_long()),
            scores = torch_empty(c(0))
          )
          final_results[[b]] <- empty
          next
        }

        # Get ROI head predictions
        roi_out <- self$roi_heads(features, props$proposals, batch_idx = b)

        # Postprocess detections
        result <- postprocess_detections(
          class_logits = roi_out$scores,
          box_regression = roi_out$boxes,
          proposals = props$proposals,
          image_size = image_size,
          num_classes = self$num_classes,
          score_thresh = self$score_thresh,
          nms_thresh = self$nms_thresh,
          detections_per_img = self$detections_per_img
        )

        final_results[[b]] <- result
      }
      list(features = features, detections = final_results)
    }
  )




mobilenet_v3_320_fpn_backbone <- function(pretrained = TRUE) {
  mobilenet <- model_mobilenet_v3_large(pretrained = pretrained)

  backbone_module <- torch::nn_module(
    initialize = function() {
      self$body <- mobilenet$features
      self$fpn <- fpn_module_2level(
        in_channels = c(160, 960),  # output channels of layer 13 and 16
        out_channels = 256
      )
    },
    forward = function(x) {
      all_feats <- list()

      for (i in seq_len(length(self$body))) {
        x <- self$body[[i]](x)
        all_feats[[i]] <- x
      }

      feats <- list(
        all_feats[[14]],  # 160 channels
        all_feats[[17]]   # 960 channels
      )

      self$fpn(feats)
    }
  )

  backbone <- backbone_module()
  backbone$out_channels <- 256
  backbone
}


#' Faster R-CNN Models
#'
#' Construct Faster R-CNN model variants for object-detection task.
#'
#' @param pretrained Logical. If TRUE, loads pretrained weights from local file.
#' @param progress Logical. Show progress bar during download (unused).
#' @param num_classes Number of output classes excluding background (default: 90 for COCO).
#' @param score_thresh Numeric. Minimum score threshold for detections (default: 0.05).
#' @param nms_thresh Numeric. Non-Maximum Suppression (NMS) IoU threshold for removing overlapping boxes (default: 0.5).
#' @param detections_per_img Integer. Maximum number of detections per image (default: 100).
#' @param ... Other arguments (unused).
#' @return A `fasterrcnn_model` nn_module.
#'
#' @section Task:
#' Object detection over images with bounding boxes and class labels.
#'
#' @section Input Format:
#' Input images should be `torch_tensor`s of shape
#' \verb{(batch_size, 3, H, W)} where `H` and `W` are typically around 800.
#'
#' @section Available Models:
#' \itemize{
#' \item `model_fasterrcnn_resnet50_fpn()`
#' \item `model_fasterrcnn_resnet50_fpn_v2()`
#' \item `model_fasterrcnn_mobilenet_v3_large_fpn()`
#' \item `model_fasterrcnn_mobilenet_v3_large_320_fpn()`
#' }
#'
#' @examples
#' \dontrun{
#' library(magrittr)
#' # ImageNet normalization constants, see https://pytorch.org/vision/stable/models.html
#' norm_mean <- c(0.485, 0.456, 0.406)
#' norm_std  <- c(0.229, 0.224, 0.225)
#' # Use a publicly available image of an animal
#' url <- paste0("https://upload.wikimedia.org/wikipedia/commons/thumb/",
#'        "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg")
#' image <- magick_loader(url) %>%
#'   transform_to_tensor() %>%
#'   transform_resize(c(520, 520))
#' # ResNet backbone requires image normalization
#' input <- image  %>%
#'   transform_normalize(norm_mean, norm_std)
#' batch_normalized <- input$unsqueeze(1)    # Add batch dimension (1, 3, H, W)
#'
#' # ResNet-50 FPN V2
#' model <- model_fasterrcnn_resnet50_fpn_v2(pretrained = TRUE, , detections_per_img = 5 )
#' model$eval()
#' torch::with_no_grad({pred <- model(batch_normalized)$detections[[1]]})
#' labels <- coco_classes(as.integer(pred$labels))
#'
#' # Visualize boxes
#' labels <- coco_classes(as.integer(pred$labels))
#' boxed <- draw_bounding_boxes(image, pred$boxes, labels = labels)
#' tensor_image_browse(boxed)
#'
#' # MobileNet V3 Large 320 FPN
#' batch <- image$unsqueeze(1)    # Add batch dimension (1, 3, H, W)
#' model <- model_fasterrcnn_mobilenet_v3_large_320_fpn(
#'   pretrained = TRUE, score_thresh = 0.02, nms_thresh = 0.8, detections_per_img = 5
#' )
#' model$eval()
#' torch::with_no_grad({pred <- model(batch)$detections[[1]]})
#'
#' # Visualize boxes
#' labels <- coco_classes(as.integer(pred$labels))
#' boxed <- draw_bounding_boxes(image, pred$boxes, labels = labels)
#' tensor_image_browse(boxed)
#' }
#'
#' @family object_detection_model
#' @name model_fasterrcnn
#' @rdname model_fasterrcnn
NULL

rpn_model_urls <- list(
  fasterrcnn_resnet50 = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/fasterrcnn_resnet50.pth",
    "8c519bb0e3a1a4fd94fb7bd21d51c135", "160 MB"),
  fasterrcnn_resnet50_v2 = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/fasterrcnn_resnet50_v2.pth",
    "88b414aecf00367413650dc732aa0aba", "170 MB"),
  fasterrcnn_mobilenet_v3_large = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/fasterrcnn_mobilenet_v3_large.pth",
    "58eba0ba379ed1497da8aa1adb8b7a7e", "75 MB"),
  fasterrcnn_mobilenet_v3_large_320 = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/fasterrcnn_mobilenet_v3_large_320.pth",
    "bd711fc4bb0da7fce38ca9916fc98753", "75 MB")
)


#' @describeIn model_fasterrcnn Faster R-CNN with ResNet-50 FPN
#' @export
model_fasterrcnn_resnet50_fpn <- function(pretrained = FALSE, progress = TRUE,
                                          num_classes = 90,
                                          score_thresh = 0.05,
                                          nms_thresh = 0.5,
                                          detections_per_img = 100,
                                          ...) {
  backbone <- resnet_fpn_backbone(pretrained = pretrained)
  model <- fasterrcnn_model(backbone, num_classes = num_classes,
                            score_thresh = score_thresh,
                            nms_thresh = nms_thresh,
                            detections_per_img = detections_per_img)
  if (pretrained && num_classes != 90)
    cli_abort("Pretrained weights require num_classes = 90 (excluding background).")

  if (pretrained) {
    r <- rpn_model_urls$fasterrcnn_resnet50
    name <- "fasterrcnn_resnet50"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "fasterrcnn")
    if (!tools::md5sum(state_dict_path) == r[2]) {
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(.rename_fasterrcnn_state_dict(state_dict), strict = FALSE)
  }

  model
}


#' @describeIn model_fasterrcnn Faster R-CNN with ResNet-50 FPN V2
#' @export
model_fasterrcnn_resnet50_fpn_v2 <- function(pretrained = FALSE, progress = TRUE,
                                             num_classes = 90,
                                             score_thresh = 0.05,
                                             nms_thresh = 0.5,
                                             detections_per_img = 100,
                                             ...) {
  backbone <- resnet_fpn_backbone_v2(pretrained = pretrained)
  model <- fasterrcnn_model_v2(backbone, num_classes = num_classes,
                               score_thresh = score_thresh,
                               nms_thresh = nms_thresh,
                               detections_per_img = detections_per_img)

  if (pretrained && num_classes != 90)
    cli_abort("Pretrained weights require num_classes = 90 (excluding background).")

  if (pretrained) {
    r <- rpn_model_urls$fasterrcnn_resnet50_v2
    name <- "fasterrcnn_resnet50_v2"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "fasterrcnn")
    if (!tools::md5sum(state_dict_path) == r[2]) {
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }
    state_dict <- torch::load_state_dict(state_dict_path)

    model_state <- model$state_dict()
    # TODO remove that model scalping
    # TODO will fail due to setdiff(names(model$modules), names(state_dict)), currently 221 discrepancies
    state_dict <- state_dict[names(state_dict) %in% names(model_state)]
    for (n in names(state_dict)) {
      if (!all(state_dict[[n]]$size() == model_state[[n]]$size())) {
        state_dict[[n]] <- model_state[[n]]
      }
    }
    missing <- setdiff(names(model_state), names(state_dict))
    if (length(missing) > 0) {
      for (n in missing) {
        state_dict[[n]] <- model_state[[n]]
      }
    }

    model$load_state_dict(state_dict, strict = TRUE)
  }

  model
}


#' @describeIn model_fasterrcnn Faster R-CNN with MobileNet V3 Large FPN
#' @export
model_fasterrcnn_mobilenet_v3_large_fpn <- function(pretrained = FALSE,
                                                    progress = TRUE,
                                                    num_classes = 90,
                                                    score_thresh = 0.05,
                                                    nms_thresh = 0.5,
                                                    detections_per_img = 100,
                                                    ...) {
  backbone <- mobilenet_v3_fpn_backbone(pretrained = pretrained)
  model <- fasterrcnn_mobilenet_model(backbone, num_classes = num_classes,
                                      score_thresh = score_thresh,
                                      nms_thresh = nms_thresh,
                                      detections_per_img = detections_per_img)

  if (pretrained && num_classes != 90)
    cli_abort("Pretrained weights require num_classes = 90 (excluding background).")

  if (pretrained) {
    r <- rpn_model_urls$fasterrcnn_mobilenet_v3_large
    name <- "fasterrcnn_mobilenet_v3_large"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "fasterrcnn")
    if (!tools::md5sum(state_dict_path) == r[2]) {
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(.rename_fasterrcnn_large_state_dict(state_dict), strict = FALSE)
  }

  model
}


#' @describeIn model_fasterrcnn Faster R-CNN with MobileNet V3 Large 320 FPN
#' @export
model_fasterrcnn_mobilenet_v3_large_320_fpn <- function(pretrained = FALSE,
                                                        progress = TRUE,
                                                        num_classes = 90,
                                                        score_thresh = 0.05,
                                                        nms_thresh = 0.5,
                                                        detections_per_img = 100,
                                                        ...) {
  backbone <- mobilenet_v3_320_fpn_backbone(pretrained = pretrained)
  model <- fasterrcnn_mobilenet_model(backbone, num_classes = num_classes,
                                      score_thresh = score_thresh,
                                      nms_thresh = nms_thresh,
                                      detections_per_img = detections_per_img)

  if (pretrained && num_classes != 90)
    cli_abort("Pretrained weights require num_classes = 90 (excluding background).")

  if (pretrained) {
    r <- rpn_model_urls$fasterrcnn_mobilenet_v3_large_320
    name <- "fasterrcnn_mobilenet_v3_large_320"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "fasterrcnn")
    if (!tools::md5sum(state_dict_path) == r[2]) {
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(.rename_fasterrcnn_large_state_dict(state_dict), strict = FALSE)
  }

  model
}

#' @importFrom stats setNames
.rename_fasterrcnn_state_dict <- function(state_dict) {
  . <- NULL # Nulling strategy for no visible binding check Note
  new_names <- names(state_dict) %>%
    # add ".0" to inner_blocks + layer_blocks layer renaming
    sub(pattern = "(inner_blocks\\.[0-3]\\.)", replacement = "\\10\\.", x = .) %>%
    sub(pattern = "(layer_blocks\\.[0-3]\\.)", replacement = "\\10\\.", x = .) %>%
    # add ".0.0" to rpn.head.conv
    sub(pattern = "(rpn\\.head\\.conv\\.)", replacement = "\\10\\.0\\.", x = .)

  # Recreate a list with renamed keys
  setNames(state_dict[names(state_dict)], new_names)
}


.rename_fasterrcnn_large_state_dict <- function(state_dict) {
  . <- NULL # Nulling strategy for no visible binding check Note
  new_names <- names(.rename_fasterrcnn_state_dict(state_dict)) %>%
    # turn bn into 'O' value and conv into '1' value
    sub(pattern = "(block\\.[0-3]\\.)0\\.", replacement = "\\1conv\\.", x = .) %>%
    sub(pattern = "(block\\.[0-3]\\.)1\\.", replacement = "\\1bn\\.", x = .) %>%
    sub(pattern = "(body\\.0\\.)0\\.", replacement = "\\1conv\\.", x = .) %>%
    sub(pattern = "(body\\.0\\.)1\\.", replacement = "\\1bn\\.", x = .) %>%
    sub(pattern = "(body\\.16\\.)0\\.", replacement = "\\1conv\\.", x = .) %>%
    sub(pattern = "(body\\.16\\.)1\\.", replacement = "\\1bn\\.", x = .)

  # Recreate a list with renamed keys
  setNames(state_dict[names(state_dict)], new_names)
}

test_that("tests for non-pretrained model_fasterrcnn_resnet50_fpn", {

  model <- model_fasterrcnn_resnet50_fpn(score_thresh = 0.6, nms_thresh = 0.9, detections_per_img = 100)
  input <- base_loader("assets/class/cat/cat.0.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(200,200)) %>% torch_unsqueeze(1)
  model$eval()
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_is(out$detections, "list")
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)


  model <- model_fasterrcnn_resnet50_fpn(num_classes = 10, score_thresh = 0.5, nms_thresh = 0.8, detections_per_img = 3)
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
})

test_that("tests for non-pretrained model_fasterrcnn_resnet50_fpn_v2", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_fasterrcnn_resnet50_fpn_v2(score_thresh = 0.5, nms_thresh = 0.8, detections_per_img = 3)
  input <- base_loader("assets/class/cat/cat.1.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180,180)) %>% torch_unsqueeze(1)
  model$eval()
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)


  model <- model_fasterrcnn_resnet50_fpn_v2(num_classes = 10, score_thresh = 0.5, nms_thresh = 0.8, detections_per_img = 3)
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
})

test_that("tests for non-pretrained model_fasterrcnn_mobilenet_v3_large_fpn", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_fasterrcnn_mobilenet_v3_large_fpn(score_thresh = 0.5, nms_thresh = 0.8, detections_per_img = 3)
  input <- base_loader("assets/class/cat/cat.2.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180,180)) %>% torch_unsqueeze(1)
  model$eval()
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)


  model <- model_fasterrcnn_resnet50_fpn_v2(num_classes = 10, score_thresh = 0.5, nms_thresh = 0.8, detections_per_img = 3)
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
})

test_that("tests for non-pretrained model_fasterrcnn_mobilenet_v3_large_320_fpn", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_fasterrcnn_mobilenet_v3_large_320_fpn(score_thresh = 0.5, nms_thresh = 0.8, detections_per_img = 3)
  input <- base_loader("assets/class/cat/cat.3.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180,180)) %>% torch_unsqueeze(1)
  model$eval()
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)

  model <- model_fasterrcnn_mobilenet_v3_large_320_fpn(num_classes = 10, score_thresh = 0.5, nms_thresh = 0.8, detections_per_img = 3)
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
})

test_that("tests for pretrained model_fasterrcnn_resnet50_fpn", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  input <- base_loader("assets/class/cat/cat.4.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180,180)) %>%
    transform_normalize(mean = c(0.485, 0.456, 0.406), std = c(0.229, 0.224, 0.225)) %>% torch_unsqueeze(1)
  model <- model_fasterrcnn_resnet50_fpn(pretrained = TRUE, score_thresh = 0.35, nms_thresh = 0.9, detections_per_img = 10)
  model$eval()
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_tensor(out$detections[[1]]$boxes)
  # Verify background class is removed, labels should be COCO IDs [1, 90]
  if (out$detections[[1]]$labels$shape[1] > 0) {
    labels_vec <- as.integer(out$detections[[1]]$labels$cpu())
    expect_true(all(labels_vec >= 1 & labels_vec <= 90), info = "All labels are in range [1, 90]")
    expect_true(any(labels_vec == 17), info = "model found a cat")
  }
  # we cannot succesfully assert bbox here :
  #   expect_bbox_is_xyxy(out$detections[[1]]$boxes, 180, 180)
  #
})

test_that("tests for pretrained model_fasterrcnn_resnet50_fpn_v2", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  input <- base_loader("assets/class/cat/cat.4.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180,180)) %>%
    transform_normalize(mean = c(0.485, 0.456, 0.406), std = c(0.229, 0.224, 0.225)) %>% torch_unsqueeze(1)
  model <- model_fasterrcnn_resnet50_fpn_v2(pretrained = TRUE, score_thresh = 0.4, nms_thresh = 0.9, detections_per_img = 10)
  model$eval()
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  if (out$detections[[1]]$boxes$shape[1] > 0) {
    expect_bbox_is_xyxy(out$detections[[1]]$boxes, 180, 180)
    # Verify background class is removed, labels should be COCO IDs [1, 90]
    labels_vec <- as.integer(out$detections[[1]]$labels$cpu())
    expect_true(all(labels_vec >= 1 & labels_vec <= 90), info = "All labels are in range [1, 90]")
    expect_true(any(labels_vec == 17), info = "model found a cat")  }
})

test_that("tests for pretrained model_fasterrcnn_mobilenet_v3_large_fpn", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_fasterrcnn_mobilenet_v3_large_fpn(pretrained = TRUE, detections_per_img = 10)
  input <- base_loader("assets/class/dog/dog.0.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(240,240)) %>% torch_unsqueeze(1)
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  if (out$detections[[1]]$boxes$shape[1] > 0) {
    expect_bbox_is_xyxy(out$detections[[1]]$boxes, 240, 240)
    # Verify background class is removed, labels should be COCO IDs [1, 90]
    labels_vec <- as.integer(out$detections[[1]]$labels$cpu())
    expect_true(all(labels_vec >= 1 & labels_vec <= 90), info = "All labels are in range [1, 90]")
    expect_true(any(labels_vec == 18), info = "model found a dog")  }
})

test_that("tests for pretrained model_fasterrcnn_mobilenet_v3_large_320_fpn", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_fasterrcnn_mobilenet_v3_large_320_fpn(pretrained = TRUE, detections_per_img = 10)
  input <- base_loader("assets/class/dog/dog.1.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(360,360)) %>% torch_unsqueeze(1)
  torch::with_no_grad({out <- model(input)})
  expect_named(out, c("features","detections"))
  expect_named(out$detections[[1]], c("boxes","labels", "scores"))
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  if (out$detections[[1]]$boxes$shape[1] > 0) {
    expect_bbox_is_xyxy(out$detections[[1]]$boxes, 360, 360)
    # Verify background class is removed, labels should be COCO IDs [1, 90]
    labels_vec <- as.integer(out$detections[[1]]$labels$cpu())
    expect_true(all(labels_vec >= 1 & labels_vec <= 90), info = "All labels are in range [1, 90]")
    expect_true(any(labels_vec == 18), info = "model found a dog")  }
})

# Tests for Faster R-CNN configurable detection parameters

context("rcnn-performance")

test_that("model_fasterrcnn_resnet50_fpn accepts score_thresh parameter", {
  skip_on_cran()
  
  expect_no_error(
    model_fasterrcnn_resnet50_fpn(
      pretrained = FALSE,
      num_classes = 91,
      score_thresh = 0.5
    )
  )
})

test_that("model_fasterrcnn_resnet50_fpn accepts nms_thresh parameter", {
  skip_on_cran()
  
  expect_no_error(
    model_fasterrcnn_resnet50_fpn(
      pretrained = FALSE,
      num_classes = 91,
      nms_thresh = 0.4
    )
  )
})

test_that("model_fasterrcnn_resnet50_fpn accepts detections_per_img parameter", {
  skip_on_cran()
  
  expect_no_error(
    model_fasterrcnn_resnet50_fpn(
      pretrained = FALSE,
      num_classes = 91,
      detections_per_img = 50
    )
  )
})

test_that("model_fasterrcnn_resnet50_fpn accepts all detection parameters together", {
  skip_on_cran()
  
  expect_no_error(
    model_fasterrcnn_resnet50_fpn(
      pretrained = FALSE,
      num_classes = 91,
      score_thresh = 0.5,
      nms_thresh = 0.4,
      detections_per_img = 50
    )
  )
})

test_that("model_fasterrcnn_resnet50_fpn_v2 accepts detection parameters", {
  skip_on_cran()
  
  expect_no_error(
    model_fasterrcnn_resnet50_fpn_v2(
      pretrained = FALSE,
      num_classes = 91,
      score_thresh = 0.5,
      nms_thresh = 0.4,
      detections_per_img = 50
    )
  )
})

test_that("model_fasterrcnn_mobilenet_v3_large_fpn accepts detection parameters", {
  skip_on_cran()
  
  expect_no_error(
    model_fasterrcnn_mobilenet_v3_large_fpn(
      pretrained = FALSE,
      num_classes = 91,
      score_thresh = 0.5,
      nms_thresh = 0.4,
      detections_per_img = 50
    )
  )
})

test_that("model_fasterrcnn_mobilenet_v3_large_320_fpn accepts detection parameters", {
  skip_on_cran()
  
  expect_no_error(
    model_fasterrcnn_mobilenet_v3_large_320_fpn(
      pretrained = FALSE,
      num_classes = 91,
      score_thresh = 0.5,
      nms_thresh = 0.4,
      detections_per_img = 50
    )
  )
})

#' @include collection-rf100-doc.R
NULL

#' RoboFlow 100 Infrared dataset Collection
#'
#' Loads one of the RoboFlow 100 Infrared datasets
#' with per-dataset folders and train/valid/test splits.
#'
#' @inheritParams rf100_document_collection
#' @param dataset Dataset to select within \code{c("thermal_dog_and_people", "solar_panel", "thermal_cheetah", "ir_object")}.
#' @inherit rf100_document_collection return
#'
#' @examples
#' \dontrun{
#' ds <- rf100_infrared_collection(
#'   dataset = "thermal_dog_and_people",
#'   split = "test",
#'   transform = transform_to_tensor,
#'   download = TRUE
#' )
#' item <- ds[1]
#' boxed <- draw_bounding_boxes(item)
#' tensor_image_browse(boxed)
#' }
#'
#' @family detection_dataset
#' @export
rf100_infrared_collection <- torch::dataset(
  name = "rf100_infrared_collection",
  inherit = rf100_document_collection,

  resources = data.frame(
    dataset = rep(
      c("thermal_dog_and_people", "solar_panel","thermal_cheetah", "ir_object"), each = 3
    ),
    split   = rep(c("train", "test", "valid"), times = 4),
    url = c(
      # thermal_dog_and_people
      "https://huggingface.co/datasets/Francesco/thermal-dogs-and-people-x6ejw/resolve/main/data/train-00000-of-00001-1610caffe8805fb2.parquet",
      "https://huggingface.co/datasets/Francesco/thermal-dogs-and-people-x6ejw/resolve/main/data/test-00000-of-00001-d8b5e33cc7b3e0f0.parquet",
      "https://huggingface.co/datasets/Francesco/thermal-dogs-and-people-x6ejw/resolve/main/data/validation-00000-of-00001-7c75a093271552c9.parquet",
      # solar_panel
      "https://huggingface.co/datasets/Francesco/solar-panels-taxvb/resolve/main/data/train-00000-of-00001-70141d825b15d748.parquet",
      "https://huggingface.co/datasets/Francesco/solar-panels-taxvb/resolve/main/data/test-00000-of-00001-6828d49996604be7.parquet",
      "https://huggingface.co/datasets/Francesco/solar-panels-taxvb/resolve/main/data/validation-00000-of-00001-11cffae851c69d23.parquet",
      # thermal_cheetah
      "https://huggingface.co/datasets/Francesco/thermal-cheetah-my4dp/resolve/main/data/train-00000-of-00001-f25decaf4cece7b6.parquet",
      "https://huggingface.co/datasets/Francesco/thermal-cheetah-my4dp/resolve/main/data/test-00000-of-00001-9a9c67c1034043ae.parquet",
      "https://huggingface.co/datasets/Francesco/thermal-cheetah-my4dp/resolve/main/data/validation-00000-of-00001-80c7672b224b69ad.parquet",
      # ir_object
      "https://huggingface.co/datasets/Francesco/flir-camera-objects/resolve/main/data/train-00000-of-00001-f9046a472e12c19a.parquet",
      "https://huggingface.co/datasets/Francesco/flir-camera-objects/resolve/main/data/test-00000-of-00001-135a7244612639cc.parquet",
      "https://huggingface.co/datasets/Francesco/flir-camera-objects/resolve/main/data/validation-00000-of-00001-0e03b58b35fd7419.parquet"
    ),
    md5 = c(
      # thermal_dog_and_people
      "eac1d66da9b5e92d1e9a21f3c99563e1",      "fe6317e4930f8be1c0c15a600c38e7a4",      "952d5adf6d969844527eb94b63656ce4",
      # solar_panel
      "a6078675c6c4fe50a74ed3fdbe781195",      "ff8983ca30ab4a5eac02ba516d0a5345",      "8f02bfbab9d2c59c1d7a57089c159e56",
      # thermal_cheetah
      "cbbb821abbb4ebf802a5d21e352f68c6",      "2628129c62f797b719d5793af46b0e0f",      "0c5fe2b61ca6d3b0b519a57575182541",
      # ir_object
      "f263921d2671cad90ed13eb0d901df2c",      "29adfd94c41294823ea78754e65116ec",      "4879197eccf7ef0511e2f1c9b7e07d64"
    ),
    size = c(6, 1.7, 0.8,  8.5, 2.3, 1.5,  2.7, 0.7, 0.4,  411, 148, 74) * 1e6,
    stringsAsFactors = FALSE
  )
)

#' @importFrom zeallot %<-%
NULL

#' box_cxcywh_to_xyxy
#'
#' Converts bounding boxes from  \eqn{(c_x, c_y, w, h)} format to \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
#'  \eqn{(c_x, c_y)} refers to center of bounding box
#'  (w, h) are width and height of bounding box
#'
#' @param boxes  (Tensor\[N, 4]): boxes in \eqn{(c_x, c_y, w, h)} format which will be converted.
#'
#' @return boxes (Tensor(N, 4)): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
box_cxcywh_to_xyxy <- function(boxes) {
  # We need to change all 4 of them so some temporary variable is needed.
  c(cx, cy, w, h) %<-% boxes$unbind(-1)
  x1 <- cx - 0.5 * w
  y1 <- cy - 0.5 * h
  x2 <- cx + 0.5 * w
  y2 <- cy + 0.5 * h

  boxes <- torch::torch_stack(list(x1, y1, x2, y2), dim=-1)

  return(boxes)
}

#' box_xyxy_to_cxcywh
#'
#' Converts bounding boxes from  \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format to \eqn{(c_x, c_y, w, h)} format.
#'  (x1, y1) refer to top left of bounding box
#'  (x2, y2) refer to bottom right of bounding box
#'
#' @param boxes  (Tensor\[N, 4\]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format which will be converted.
#'
#' @return boxes (Tensor(N, 4)): boxes in \eqn{(c_x, c_y, w, h)} format.
box_xyxy_to_cxcywh <- function(boxes) {
  c(x1, y1, x2, y2) %<-% boxes$unbind(-1)
  cx <- (x1 + x2) / 2
  cy <- (y1 + y2) / 2
  w <- x2 - x1
  h <- y2 - y1

  boxes <- torch::torch_stack(list(cx, cy, w, h), dim=-1)

  return(boxes)
}

#' box_xywh_to_xyxy
#'
#' Converts bounding boxes from  (x, y, w, h) format to \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
#' (x, y) refers to top left of bouding box.
#' (w, h) refers to width and height of box.
#'
#' @param boxes  (Tensor\[N, 4\]): boxes in (x, y, w, h) which will be converted.
#'
#' @return boxes (Tensor\[N, 4\]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
box_xywh_to_xyxy <- function(boxes) {
  c(x, y, w, h) %<-% boxes$unbind(-1)
  boxes <- torch::torch_stack(list(x, y, x + w, y + h), dim=-1)
  return(boxes)
}

#' box_xyxy_to_xywh
#'
#' Converts bounding boxes from  \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format to (x, y, w, h) format.
#' (x1, y1) refer to top left of bounding box
#' (x2, y2) refer to bottom right of bounding box
#'
#' @param boxes  (Tensor\[N, 4\]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} which will be converted.
#'
#' @return boxes (Tensor\[N, 4\]): boxes in (x, y, w, h) format.
box_xyxy_to_xywh <- function(boxes) {
  c(x1, y1, x2, y2) %<-% boxes$unbind(-1)
  w <- x2 - x1 # x2 - x1
  h <- y2 - y1 # y2 - y1
  boxes <- torch::torch_stack(list(x1, y1, w, h), dim=-1)
  return(boxes)
}

x1 <- torch::torch_ones(5)
y1 <- torch::torch_ones(5)
x2 <- x1 + 1
y2 <- y1 + 1
xyxy <- torch::torch_stack(list(x1,y1,x2,y2))$transpose(2,1)

test_that("box_cxcywh_to_xyxy box_xyxy_to_cxcywh box_xywh_to_xyxy box_xyxy_to_xywh", {
  # from xyxy
  cxcywh <- box_cxcywh_to_xyxy(box_xyxy_to_cxcywh(xyxy))
  xywh <- box_xywh_to_xyxy(box_xyxy_to_xywh(xyxy))

  expect_tensor(cxcywh)
  expect_tensor(xywh)
  expect_equal(torch::as_array(cxcywh), torch::as_array(xyxy))
  expect_equal(torch::as_array(xywh), torch::as_array(xyxy))
})

#' @export
transform_to_tensor.array <- function(img) {
  # HW to HWC format for grayscale img
  if (length(dim(img)) == 2)
    dim(img) <- c(dim(img), 1)
  dims <- dim(img)
  ndim <- length(dims)

  is_3d_chw <- (ndim == 3 && dims[1] <= 4 && dims[3] > 4)

  if (ndim < 3 || ndim > 4)
    value_error("Expected a 2D (grayscale), 3D (image) or 4D (batch) array.")

  # Support both CHW / BCHW (channel-first arrays) and HWC (default image arrays) .
  if (is_3d_chw || ndim == 4) {
    res <- torch::torch_tensor(img)
  } else {
    res <- torch::torch_tensor(img)$permute(c(3, 1, 2))
  }

  if (res$dtype == torch::torch_long())
    res <- res/255

  res
}

#' @export
transform_to_tensor.matrix <- transform_to_tensor.array

#' @export
transform_to_tensor.list <- function(img) {
  if (inherits(img[[1]], "array")) {
    ndim <- length(dim(img[[1]]))
    if (ndim > 3) {
      value_error("Expected a list of 2D or 3D arrays.")
    }
    torch::torch_stack(lapply(img, transform_to_tensor))
  } else if (inherits(img[[1]], "magick-image")) {
    torch::torch_stack(lapply(img, transform_to_tensor))
  } else {
    not_implemented_for_class(img[[1]])
  }

}

resnet_model_urls <- list(
  "resnet18" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/resnet18.pth", "b36d2ea3a4dbf3fbcb27552a410054f9", "~45 MB"),
  "resnet34" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/resnet34.pth", "cfd906d8f35b0256e1066ce970f65c50","~85 MB"),
  "resnet50" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/resnet50.pth", "9ee57c8355a4ad3759c3e91b0d9f6144", "~100 MB"),
  "resnet101" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/resnet101.pth", "0898550c0f6156d18d83fc7454760827", "~170 MB"),
  "resnet152" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/resnet152.pth", "cfe6cf55d97bdd838fe7c892c30788ef", "~230 MB"),
  "resnext50_32x4d" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/resnext50_32x4d.pth", "13f1b58b0694b634e43cf55e7208abc0", "~95 MB"),
  "resnext101_32x8d" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/resnext101_32x8d.pth", "c79668937f01117ca74193a71b31557b", "~340 MB"),
  "wide_resnet50_2" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/wide_resnet50_2.pth", "b2d17fb9c5929ab96becf63b1403423d", "~130 MB"),
  "wide_resnet101_2" = c("https://torch-cdn.mlverse.org/models/vision/v2/models/wide_resnet101_2.pth","7ccd7583b94e3f5fa0f231b4c134760e", "~240 MB")
)

conv_3x3 <- function(in_planes, out_planes, stride = 1, groups = 1, dilation = 1) {
  torch::nn_conv2d(in_planes, out_planes, kernel_size = 3, stride = stride,
            padding = dilation, groups = groups, bias = FALSE,
            dilation = dilation)
}

conv_1x1 <- function(in_planes, out_planes, stride=1) {
  torch::nn_conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=FALSE)
}

basic_block <- torch::nn_module(
  "basic_block",
  expansion = 1,
  initialize = function(inplanes, planes, stride=1, downsample=NULL, groups=1,
                        base_width=64, dilation=1, norm_layer=NULL) {

    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    if (groups != 1 || base_width != 64)
      value_error("basic_block only supports groups=1 and base_width=64")

    if (dilation > 1)
      not_implemented_error("Dilation > 1 not supported in basic_block")

    self$conv1 <- conv_3x3(inplanes, planes, stride)
    self$bn1 <- norm_layer(planes)
    self$relu <- torch::nn_relu(inplace = TRUE)
    self$conv2 <- conv_3x3(planes, planes)
    self$bn2 <- norm_layer(planes)
    self$downsample <- downsample
    self$stride <- stride

  },
  forward = function(x) {

    out <- self$conv1(x)
    out <- self$bn1(out)
    out <- self$relu(out)

    out <- self$conv2(out)
    out <- self$bn2(out)

    if (!is.null(self$downsample))
      identity <- self$downsample(x)
    else
      identity <- x

    out$add_(identity)
    out <- self$relu(out)

    out
  }
)

# Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
# while original implementation places the stride at the first 1x1 convolution(self.conv1)
# according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
# This variant is also known as ResNet V1.5 and improves accuracy according to
# https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.
bottleneck <- torch::nn_module(
  "bottleneck",
  expansion = 4,
  initialize = function(inplanes, planes, stride=1, downsample=NULL, groups=1,
                        base_width=64, dilation=1, norm_layer=NULL) {

    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    width <- as.integer(planes * (base_width / 64)) * groups

    self$conv1 <- conv_1x1(inplanes, width)
    self$bn1 <- norm_layer(width)

    self$conv2 <- conv_3x3(width, width, stride, groups, dilation)
    self$bn2 <- norm_layer(width)

    self$conv3 <- conv_1x1(width, planes * self$expansion)
    self$bn3 <- norm_layer(planes * self$expansion)

    self$relu <- torch::nn_relu(inplace = TRUE)
    self$downsample <- downsample
    self$stride <- stride

  },
  forward = function(x) {

    out <- self$conv1(x)
    out <- self$bn1(out)
    out <- self$relu(out)

    out <- self$conv2(out)
    out <- self$bn2(out)
    out <- self$relu(out)

    out <- self$conv3(out)
    out <- self$bn3(out)

    if (!is.null(self$downsample))
      identity <- self$downsample(x)
    else
      identity <- x

    out$add_(identity)
    out <- self$relu(out)

    out
  }
)

resnet <- torch::nn_module(
  "resnet",
  initialize = function(block, layers, num_classes=1000, zero_init_residual=FALSE,
                        groups=1, width_per_group=64, replace_stride_with_dilation=NULL,
                        norm_layer=NULL) {

    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    self$.norm_layer <- norm_layer

    self$inplanes <- 64
    self$dilation <- 1

    # each element in the tuple indicates if we should replace
    # the 2x2 stride with a dilated convolution instead
    if (is.null(replace_stride_with_dilation))
      replace_stride_with_dilation <- rep(FALSE, 3)

    if (length(replace_stride_with_dilation) != 3)
      value_error(
        "replace_stride_with_dilation should be NULL ",
        "or a 3-element tuple, got {length(replace_stride_with_dilation)}"
      )

    self$groups <- groups
    self$base_width <- width_per_group
    self$conv1 <- torch::nn_conv2d(3, self$inplanes, kernel_size=7, stride=2, padding=3,
                           bias=FALSE)
    self$bn1 <- norm_layer(self$inplanes)
    self$relu <- torch::nn_relu(inplace=TRUE)
    self$maxpool <- torch::nn_max_pool2d(kernel_size=3, stride=2, padding=1)
    self$layer1 <- self$.make_layer(block, 64, layers[1])
    self$layer2 <- self$.make_layer(block, 128, layers[2], stride=2,
                                   dilate=replace_stride_with_dilation[1])
    self$layer3 <- self$.make_layer(block, 256, layers[3], stride=2,
                                   dilate=replace_stride_with_dilation[2])
    self$layer4 <- self$.make_layer(block, 512, layers[4], stride=2,
                                   dilate=replace_stride_with_dilation[3])
    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(1, 1))
    self$fc <- torch::nn_linear(512 * block$public_fields$expansion, num_classes)

    for (m in private$modules_) {
      if (inherits(m, "nn_conv2d")) {
        torch::nn_init_kaiming_normal_(m$weight, mode=  "fan_out",
                                       nonlinearity = "relu")
      } else if (inherits(m, "nn_batch_norm2d") || inherits(m, "nn_group_norm")) {
        torch::nn_init_constant_(m$weight, 1)
        torch::nn_init_constant_(m$bias, 0)
      }
    }

    # Zero-initialize the last BN in each residual branch,
    # so that the residual branch starts with zeros, and each residual block behaves like an identity.
    # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
    if (zero_init_residual) {
      for (m in private$modules_) {

        if (inherits(m, "bottleneck"))
          torch::nn_init_constant_(m$bn3$weight, 0)
        else if (inherits(m, "basic_block"))
          torch::nn_init_constant_(m$bn2$weight, 0)

      }
    }

  },
  .make_layer = function(block, planes, blocks, stride=1, dilate=FALSE) {

    norm_layer <- self$.norm_layer
    downsample <- NULL
    previous_dilation <- self$dilation

    if (dilate) {
      self$dilation <- self$dilation*stride
      stride <- 1
    }

    if (stride != 1 || self$inplanes != planes * block$public_fields$expansion) {
      downsample <- torch::nn_sequential(
        conv_1x1(self$inplanes, planes * block$public_fields$expansion, stride),
        norm_layer(planes * block$public_fields$expansion)
      )
    }

    layers <- list()
    layers[[1]] <- block(self$inplanes, planes, stride, downsample, self$groups,
            self$base_width, previous_dilation, norm_layer)
    self$inplanes <- planes * block$public_fields$expansion

    for (i in seq(from = 2, to = blocks)) {
      layers[[i]] <- block(self$inplanes, planes, groups=self$groups,
                           base_width=self$base_width, dilation=self$dilation,
                           norm_layer=norm_layer)
    }

    do.call(torch::nn_sequential, layers)
  },
  forward = function(x) {
    x <- self$conv1(x)
    x <- self$bn1(x)
    x <- self$relu(x)
    x <- self$maxpool(x)

    x <- self$layer1(x)
    x <- self$layer2(x)
    x <- self$layer3(x)
    x <- self$layer4(x)

    x <- self$avgpool(x)
    x <- torch::torch_flatten(x, start_dim = 2)
    x <- self$fc(x)
    x
  }
)

.resnet <- function(arch, block, layers, pretrained, progress, ...) {
  model <- resnet(block, layers, ...)

  if (pretrained) {
    r <- resnet_model_urls[[arch]]
    cli_inform("Model weights for {.cls {arch}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1])
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict)
  }

  model
}

#' ResNet implementation
#'
#' ResNet models implementation from
#'   [Deep Residual Learning for Image Recognition](https://arxiv.org/pdf/1512.03385) and later
#'   related papers (see Functions)
#'
#' @param pretrained (bool): If TRUE, returns a model pre-trained on ImageNet.
#' @param progress (bool): If TRUE, displays a progress bar of the download to
#'   stderr.
#' @param ... Other parameters passed to the resnet model.
#'
#' @family classification_model
#' @name model_resnet
#' @rdname model_resnet
NULL

#' @describeIn model_resnet ResNet 18-layer model
#' @export
model_resnet18 <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("resnet18", basic_block, c(2,2,2,2), pretrained, progress, ...)
}

#' @describeIn model_resnet ResNet 34-layer model
#' @export
model_resnet34 <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("resnet34", basic_block, c(3,4,6,3), pretrained, progress, ...)
}

#' @describeIn model_resnet ResNet 50-layer model
#' @export
model_resnet50 <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("resnet50", bottleneck, c(3,4,6,3), pretrained, progress, ...)
}

#' @describeIn model_resnet ResNet 101-layer model
#' @export
model_resnet101 <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("resnet101", bottleneck, c(3,4,23,3), pretrained, progress, ...)
}

#' @describeIn model_resnet ResNet 152-layer model
#' @export
model_resnet152 <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("resnet152", bottleneck, c(3,8,36,3), pretrained, progress, ...)
}

#' @describeIn model_resnet ResNeXt-50 32x4d model from ["Aggregated Residual Transformation for Deep Neural Networks"](https://arxiv.org/pdf/1611.05431)
#' with 32 groups having each a width of 4.
#' @export
model_resnext50_32x4d <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("resnext50_32x4d", bottleneck, c(3,4,6,3), pretrained, progress, groups=32, width_per_group=4,...)
}

#' @describeIn model_resnet ResNeXt-101 32x8d model from ["Aggregated Residual Transformation for Deep Neural Networks"](https://arxiv.org/pdf/1611.05431)
#' with 32 groups having each a width of 8.
#' @export
model_resnext101_32x8d <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("resnext101_32x8d", bottleneck, c(3,4,23,3), pretrained, progress, groups=32, width_per_group=8,...)
}

#' @describeIn model_resnet Wide ResNet-50-2 model from ["Wide Residual Networks"](https://arxiv.org/pdf/1605.07146)
#' with width per group of 128.
#' @export
model_wide_resnet50_2 <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("wide_resnet50_2", bottleneck, c(3,4,6,3), pretrained, progress, width_per_group=64*2,...)
}

#' @describeIn model_resnet Wide ResNet-101-2 model from ["Wide Residual Networks"](https://arxiv.org/pdf/1605.07146)
#' with width per group of 128.
#' @export
model_wide_resnet101_2 <- function(pretrained = FALSE, progress = TRUE, ...) {
  .resnet("wide_resnet101_2", bottleneck, c(3,4,23,3), pretrained, progress, width_per_group=64*2,...)
}

test_that("resnet18", {

  model <- model_resnet18()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  model <- model_resnet18(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("resnet34", {
  skip_on_os(c("windows", "mac"))

  model <- model_resnet34()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  model <- model_resnet34(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("resnet50", {
  skip_on_os(c("windows", "mac"))

  model <- model_resnet50()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  model <- model_resnet50(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("resnet101", {
  skip_on_os(c("windows", "mac"))

  model <- model_resnet101()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  withr::with_options(list(timeout = 360),
                      model <- model_resnet101(pretrained = TRUE))
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("resnet152", {
  skip_on_os(c("windows", "mac"))
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_resnet152()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  withr::with_options(list(timeout = 360),
                      model <- model_resnet152(pretrained = TRUE))
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("resnext50_32x4d", {
  skip_on_os(c("windows", "mac"))

  model <- model_resnext50_32x4d()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  withr::with_options(list(timeout = 360),
                      model <- model_resnext50_32x4d(pretrained = TRUE))
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("resnext50_32x4d", {
  skip_on_os(c("windows", "mac"))

  model <- model_resnext50_32x4d()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  withr::with_options(list(timeout = 360),
                      model <- model_resnext50_32x4d(pretrained = TRUE))
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("resnext101_32x8d", {
  skip_on_os(c("windows", "mac"))
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_resnext101_32x8d()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  withr::with_options(list(timeout = 360),
                      model <- model_resnext101_32x8d(pretrained = TRUE))
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("wide_resnet50_2", {
  skip_on_os(c("windows", "mac"))

  model <- model_wide_resnet50_2()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  withr::with_options(list(timeout = 360),
                      model <- model_wide_resnet50_2(pretrained = TRUE))
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("wide_resnet101_2", {
  skip_on_os(c("windows", "mac"))
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_wide_resnet101_2()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  withr::with_options(list(timeout = 360),
                      model <- model_wide_resnet101_2(pretrained = TRUE))
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

})

test_that("we can prune head of resnet34 moels", {
  resnet34 <- model_resnet34(pretrained=TRUE)

  expect_error(prune <- nn_prune_head(resnet34, 1), NA)
  # expect_true(inherits(prune, "nn_sequential"))
  expect_length(prune, 9)
  expect_s3_class(prune[[length(prune)]], "nn_adaptive_avg_pool2d")

  input <- torch::torch_randn(1, 3, 256, 256)
  out <- prune(input)
  expect_tensor_shape(out, c(1, 512, 1, 1))

})

test_that("we can prune head of resnet50 moels", {
  resnet50 <- model_resnet50(pretrained=TRUE)

  expect_error(prune <- nn_prune_head(resnet50, 1), NA)
  expect_s3_class(prune, "nn_sequential")
  expect_length(prune, 9)
  expect_s3_class(prune[[length(prune)]], "nn_adaptive_avg_pool2d")

  input <- torch::torch_randn(1, 3, 256, 256)
  out <- prune(input)
  expect_tensor_shape(out, c(1, 2048, 1, 1))


})

test_that("we can prune head of resnext101 moels", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  resnext101 <- model_resnext101_32x8d(pretrained=TRUE)

  expect_error(prune <- torch::nn_prune_head(resnext101, 1), NA)
  expect_s3_class(prune, "nn_sequential")
  expect_length(prune, 9)
  expect_s3_class(prune[[length(prune)]], "nn_adaptive_avg_pool2d")

  input <- torch::torch_randn(1, 3, 256, 256)
  out <- prune(input)
  expect_tensor_shape(out, c(1, 2048, 1, 1))

  unlink_model_file()
})

#' Tiny ImageNet dataset
#'
#' Prepares the Tiny ImageNet dataset and optionally downloads it.
#'
#' @param root directory path to download the dataset.
#' @param split dataset split, `train`, `validation` or `test`.
#' @param download whether to download or not the dataset.
#' @param ... other arguments passed to [image_folder_dataset()].
#'
#' @family classification_dataset
#'
#' @export
tiny_imagenet_dataset <- torch::dataset(
  "tiny_imagenet",
  inherit = image_folder_dataset,
  tar_name = "tiny-imagenet-200",
  url = "http://cs231n.stanford.edu/tiny-imagenet-200.zip",
  initialize = function(root, split='train', download = FALSE, ...) {

    root <- normalizePath(root, mustWork = FALSE)

    if (!fs::dir_exists(root))
      fs::dir_create(root)

    self$root_path <- root

    if (download)
      self$download()

    super$initialize(root = fs::path_join(c(root, self$tar_name, split)), ...)

  },
  download = function() {

    p <- fs::path_join(c(self$root_path, self$tar_name))

    if (fs::dir_exists(p))
      return(NULL)

    raw_path <- fs::path_join(c(self$root_path, "tiny-imagenet-200.zip"))

    cli_inform("Downloading {.cls {class(self)[[1]]}} ...")

    p <- download_and_cache(self$url)
    fs::file_copy(p, raw_path)

    cli_inform("Processing {.cls {class(self)[[1]]}} ...")

    utils::unzip(raw_path, exdir = self$root_path)

    # organize validation images
    val_path <- fs::path_join(c(self$root_path, self$tar_name, "val"))
    val_images <- read.table(fs::path_join(c(val_path, "val_annotations.txt")))

    fs::dir_create(
      fs::path(val_path, unique(val_images$V2))
    )

    fs::file_move(
      fs::path(val_path, "images", val_images$V1),
      fs::path(val_path, val_images$V2, val_images$V1)
    )

    fs::dir_delete(fs::path(val_path, "images"))

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")

  }
)

#' ImageNet Class Labels
#'
#' Utilities for resolving ImageNet-1k class identifiers to their corresponding
#' human readable labels. The labels are retrieved from the same source used by
#' PyTorch's reference implementation.
#'
#' @param class_id Integer vector of 1-based class identifiers.
#' @return A character vector with 1000 entries representing the ImageNet-1k
#'   class labels.
#' @family class_resolution
#' @export
imagenet_classes <- function(class_id = 1:1000) {
  if (any(class_id > 1000)) {
    cli_warn("Imagenet {.var class_id} cannot be > 1000")
  }
  url <- download_and_cache("https://raw.githubusercontent.com/pytorch/hub/master/imagenet_classes.txt")
  labels <- readLines(url, warn = FALSE)
  labels[nzchar(labels)][class_id]
}

#' @family class_resolution
#' @rdname imagenet_classes
#' @export
imagenet_1k_classes <- imagenet_classes

#' @param class_id Integer vector of 1-based class identifiers.
#' @return A data.frame containing columns `id` and `label` representing 
#'   the ImageNet-21k class identifiers and labels. By default, returns all 21.8K rows.
#' @family class_resolution
#' @rdname imagenet_classes
#' @export
imagenet_21k_df <- function(class_id = 1:21843) {
  if (any(class_id > 21843)) {
    cli_warn("Imagenet {.var class_id} cannot be > 21843")
  }
  url <- download_and_cache("https://raw.githubusercontent.com/ailia-ai/ailia-models/refs/heads/master/object_detection/detic/datasets_detic/imagenet21k_wordnet_ids.txt")
  ids <- readLines(url, warn = FALSE)
  url <- download_and_cache("https://raw.githubusercontent.com/vkinakh/stable-diffusion-imagenet/refs/heads/main/imagenet21k_wordnet_lemmas.txt")
  labels <- readLines(url, warn = FALSE)

  data.frame(id = ids, label = labels)[class_id,]
}

#' @param class_id Integer vector of 1-based class identifiers.
#' @return A character vector with the labels associated with `class_id`.
#' @family class_resolution
#' @rdname imagenet_classes
#' @export
imagenet_21k_classes <- function(class_id) {
  imagenet_21k_df(class_id)$label
}

#' Labelled Faces in the Wild (LFW) Datasets
#'
#' The LFW dataset collection provides facial images for evaluating face recognition systems.
#' It includes two variants:
#' - `lfw_people_dataset`: A **multi-class classification** dataset where each image is labelled by person identity.
#' - `lfw_pairs_dataset`: A **face verification** dataset containing image pairs with binary labels (same or different person).
#'
#' @inheritParams oxfordiiitpet_dataset
#' @param root Root directory for dataset storage. The dataset will be stored under `root/lfw_people` or `root/lfw_pairs`.
#' @param split Which version of the dataset to use. One of `"original"` or `"funneled"`. Defaults to `"original"`.
#' @param train For `lfw_pairs_dataset`, whether to load the training (`pairsDevTrain.txt`) or test (`pairsDevTest.txt`) split.
#'
#' @return A torch dataset object \code{lfw_people_dataset} or \code{lfw_pairs_dataset}.
#' Each element is a named list with:
#' - \code{x}:
#'   - For \code{lfw_people_dataset}: a H x W x 3 numeric array representing a single RGB image.
#'   - For \code{lfw_pairs_dataset}: a list of two H x W x 3 numeric arrays representing a pair of RGB images.
#' - \code{y}:
#'   - For \code{lfw_people_dataset}: an integer index from 1 to the number of identities in the dataset.
#'   - For \code{lfw_pairs_dataset}: 1 if the pair shows the same person, 2 if different people.
#'
#' @details
#' This R implementation of the LFW dataset is based on the `fetch_lfw_people()` and `fetch_lfw_pairs()` functions from the `scikit-learn` library,
#' but deviates in a few key aspects due to dataset availability and R API conventions:
#'
#' - The \code{color} and \code{resize} arguments from Python are not directly exposed. Instead, all images are RGB with a fixed size of 250x250.
#'
#' - The \code{split} argument in Python (e.g., ``train``, ``test``, ``10fold``) is simplified to a \code{train} boolean flag in R.
#'   The ``10fold`` split is not supported, as the original protocol files are unavailable or incompatible with clean separation of image-label pairs.
#'
#' - The \code{split} parameter in R controls which version of the dataset to use: `"original"` (unaligned) or `"funneled"` (aligned using funneling).
#'   The funneled version contains geometrically normalized face images, offering better alignment and typically improved performance for face recognition models.
#'
#' - The dataset is downloaded from Figshare,
#'   which hosts the same files referenced in `scikit-learn`'s dataset utilities.
#'
#' - `lfw_people_dataset`: 13,233 images across multiple identities (using either `"original"` or `"funneled"` splits)
#' - `lfw_pairs_dataset`:
#'   - Training split (`train = TRUE`): 2,200 image pairs
#'   - Test split (`train = FALSE`): 1,000 image pairs
#'
#' @examples
#' \dontrun{
#' # Load data for LFW People Dataset
#' lfw <- lfw_people_dataset(download = TRUE)
#' first_item <- lfw[1]
#' first_item$x  # RGB image
#' first_item$y  # Label index
#' lfw$classes[first_item$y]  # person's name (e.g., "Aaron_Eckhart")
#'
#' # Load training data for LFW Pairs Dataset
#' lfw <- lfw_pairs_dataset(download = TRUE, train = TRUE)
#' first_item <- lfw[1]
#' first_item$x  # List of 2 RGB Images
#' first_item$x[[1]]  # RGB Image
#' first_item$x[[2]]  # RGB Image
#' first_item$y  # Label index
#' lfw$classes[first_item$y]  # Class Name (e.g., "Same" or "Different")
#'
#' # Load test data for LFW Pairs Dataset
#' lfw <- lfw_pairs_dataset(download = TRUE, train = FALSE)
#' first_item <- lfw[1]
#' first_item$x  # List of 2 RGB Images
#' first_item$x[[1]]  # RGB Image
#' first_item$x[[2]]  # RGB Image
#' first_item$y  # Label index
#' lfw$classes[first_item$y]  # Class Name (e.g., "Same" or "Different")
#' }
#'
#' @name lfw_dataset
#' @title LFW Datasets
#' @rdname lfw_dataset
#' @family classification_dataset
#' @export
lfw_people_dataset <- torch::dataset(
  name = "lfw_people",
  archive_size_table = list(
    "original" = "170 MB",
    "funneled" = "230 MB"
  ),
  base_url = "https://ndownloader.figshare.com/files/",
  resources = list(
    original = c("5976018", "a17d05bd522c52d84eca14327a23d494"),
    funneled = c("5976015", "1b42dfed7d15c9b2dd63d5e5840c86ad")
  ),

  initialize = function(
    root = tempdir(),
    transform = NULL,
    split = "original",
    target_transform = NULL,
    download = FALSE
  ) {

    self$root <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$archive_size <- self$archive_size_table[[split]]

    self$split <- split
    if (split == "original"){
      self$split_name <- "lfw"
    } else {
      self$split_name <- "lfw_funneled"
    }

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }

    image_dir <- file.path(root, self$split_name)

    person_dirs <- list.dirs(image_dir, full.names = TRUE, recursive = FALSE)
    class_names <- basename(person_dirs)
    class_to_idx <- setNames(seq_along(class_names), class_names)

    df_list <- lapply(person_dirs, function(person) {
      images <- list.files(person, pattern = "\\.jpg$", full.names = TRUE)
      label <- class_to_idx[[basename(person)]]
      data.frame(img_path = images, label = label, stringsAsFactors = FALSE)
    })

    df <- do.call(rbind, df_list)

    self$img_path <- df$img_path
    self$labels <- df$label
    self$classes <- class_names
    self$class_to_idx <- class_to_idx

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {

    if (self$check_exists()) {
      return()
    }

    fs::dir_create(self$root)

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    res <- self$resources[[self$split]]

    archive <- download_and_cache(paste0(self$base_url, res[1]), prefix = class(self)[1])

    expected_md5 <- res[2]
    actual_md5 <- tools::md5sum(archive)
    if (actual_md5 != expected_md5) {
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")
    }

    untar(archive, exdir = self$root)

    if (class(self)[[1]] == "lfw_pairs") {
      for (name in c("pairsDevTrain.txt", "pairsDevTest.txt", "pairs.txt")) {
        res <- self$resources[[name]]
        archive <- download_and_cache(paste0(self$base_url, res[1]), prefix = class(self)[1])

        expected_md5 <- res[2]
        actual_md5 <- tools::md5sum(archive)
        if (actual_md5 != expected_md5) {
          runtime_error("Corrupt file! Delete the file in {archive} and try again.")
        }
        dest_path <- file.path(self$root, name)
        fs::file_move(archive, dest_path)
      }
    }

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  check_exists = function() {
    fs::dir_exists(file.path(self$root, self$split_name))
  },

  .getitem = function(index) {
    x <- jpeg::readJPEG(self$img_path[index])
    y <- self$labels[index]

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    list(x = x, y = y)
  },

  .length = function() {
    length(self$img_path)
  }
)

#' @rdname lfw_dataset
#' @export
lfw_pairs_dataset <- torch::dataset(
  inherit = lfw_people_dataset,
  name = "lfw_pairs",
  archive_size_table = list(
    "original" = "170 MB",
    "funneled" = "230 MB"
  ),
  base_url = "https://ndownloader.figshare.com/files/",
  resources = list(
    original = c("5976018", "a17d05bd522c52d84eca14327a23d494"),
    funneled = c("5976015", "1b42dfed7d15c9b2dd63d5e5840c86ad"),
    pairsDevTrain.txt = c("5976012", "4f27cbf15b2da4a85c1907eb4181ad21"),
    pairsDevTest.txt = c("5976009", "5132f7440eb68cf58910c8a45a2ac10b"),
    pairs.txt = c("5976006", "9f1ba174e4e1c508ff7cdf10ac338a7d")
  ),

  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    split = "original",
    target_transform = NULL,
    download = FALSE
  ) {
    self$root <- root
    self$train <- train
    self$transform <- transform
    self$target_transform <- target_transform
    self$classes <- c("Same", "Different")
    self$split <- split
    if (split == "original"){
      self$split_name <- "lfw"
    } else {
      self$split_name <- "lfw_funneled"
    }
    self$archive_size <- self$archive_size_table[[split]]

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }

    if (train) {
      pair_file <- file.path(root, "pairsDevTrain.txt")
    } else {
      pair_file <- file.path(root, "pairsDevTest.txt")
    }

    lines <- readLines(pair_file)
    lines <- lines[-1]
    pair_list <- lapply(lines, function(line) {
      parts <- strsplit(line, "\\s+")[[1]]
      if (length(parts) == 3) {
        name <- parts[1]
        num1 <- as.integer(parts[2])
        num2 <- as.integer(parts[3])
        img1 <- file.path(name, sprintf("%s_%04d.jpg", name, num1))
        img2 <- file.path(name, sprintf("%s_%04d.jpg", name, num2))
        label <- 1
      } else {
        name1 <- parts[1]
        num1 <- as.integer(parts[2])
        name2 <- parts[3]
        num2 <- as.integer(parts[4])
        img1 <- file.path(name1, sprintf("%s_%04d.jpg", name1, num1))
        img2 <- file.path(name2, sprintf("%s_%04d.jpg", name2, num2))
        label <- 2
      }
      data.frame(img1 = img1, img2 = img2, label = label, stringsAsFactors = FALSE)
    })

    self$pairs <- do.call(rbind, pair_list)
    self$img_path <- c(self$pairs$img1, self$pairs$img2)

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  .getitem = function(index) {
    pair <- self$pairs[index, ]

    img1_path <- file.path(self$root, self$split_name, pair$img1)
    img2_path <- file.path(self$root, self$split_name, pair$img2)

    x1 <- jpeg::readJPEG(img1_path)
    x2 <- jpeg::readJPEG(img2_path)
    x <- list(x1, x2)
    y <- as.integer(pair$label)

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    list(x = x, y = y)
  },

  .length = function() {
    nrow(self$pairs)
  },

  check_exists = function() {
    required_files <- c("pairs.txt", "pairsDevTest.txt", "pairsDevTrain.txt")
    fs::dir_exists(file.path(self$root, self$split_name)) && all(fs::file_exists(file.path(self$root, required_files)))
  }
)

#' EuroSAT datasets
#'
#' A collection of Sentinel-2 satellite images for land-use **classification**.
#' The standard version contains 27,000 RGB thumbnails (64x64) across 10
#' classes. Variants include the full 13 spectral bands and a small 100-image
#' subset useful for demos.
#'
#' @name eurosat_dataset
#' @details
#' `eurosat_dataset()` provides a total of 27,000 RGB labeled images.
#'
#' @inheritParams mnist_dataset
#' @param root (Optional) Character. The root directory where the dataset will be stored.
#'  if empty, will use the default `rappdirs::user_cache_dir("torch")`.
#' @param split One of `"train"`, `"val"`, or `"test"`. Default is `"val"`.
#'
#' @return A `torch::dataset` object. Each item is a list with:
#' * `x`: a 64x64 image tensor with 3 (RGB) or 13 (all bands) channels
#' * `y`: the class label
#'
#' @examples
#' \dontrun{
#' # Initialize the dataset
#' ds <- eurosat100_dataset(split = "train", download = TRUE)
#'
#' # Access the first item
#' head <- ds[1]
#' print(head$x) # Image
#' print(head$y) # Label
#' }
#' @family classification_dataset
#' @export
eurosat_dataset <- torch::dataset(
  name = "eurosat",
  archive_url = "https://huggingface.co/datasets/torchgeo/eurosat/resolve/main/EuroSAT.zip?download=true",
  archive_md5 = "c8fa014336c82ac7804f0398fcb19387",
  split_url = "https://huggingface.co/datasets/torchgeo/eurosat/resolve/main/eurosat-{split}.txt?download=true",
  archive_size = "90 MB",

  initialize = function(
    root = tempdir(),
    split = "val",
    download = FALSE,
    transform = NULL,
    target_transform = NULL
  ) {

    self$root <- normalizePath(root, mustWork = FALSE)
    self$split <- match.arg(split, c("train", "val", "test"))
    self$transform <- transform
    self$target_transform <- target_transform
    self$split_url <- glue::glue(self$split_url)
    self$images_dir <- file.path(self$root, class(self)[1], "images")
    self$split_file <- file.path(self$root, fs::path_ext_remove(basename(self$split_url)))

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    self$img_files <- list.files(self$images_dir, pattern = "\\.(tif|jpg)", recursive = TRUE, full.names = TRUE)

    if (!self$check_exists())
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")

    self$data <- suppressWarnings(readLines(self$split_file))
    self$load_meta()

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  load_meta = function() {
    self$classes <- unique(sub("_.*", "", self$data))
    self$class_to_idx <- setNames(seq_along(self$classes) - 1, self$classes)
  },

  download = function() {
    if (self$check_exists())
      return(NULL)

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    fs::dir_create(self$root, recurse = TRUE, showWarnings = FALSE)

    # Download and extract the dataset archive
    archive <- download_and_cache(self$archive_url, prefix = class(self)[1])
    if (!tools::md5sum(archive) == self$archive_md5)
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")


    if (!dir.exists(self$images_dir)) {
      cli_inform("Extracting {.cls {class(self)[[1]]}} archive...")
      utils::unzip(archive, exdir = self$images_dir)
      cli_inform("Extraction of {.cls {class(self)[[1]]}} is complete.")
    }

    # Download the split-specific text file
    cli_inform("Downloading {.cls {class(self)[[1]]}} split file: {self$split_url}")
    p <- download_and_cache(self$split_url, prefix = class(self)[1])
    fs::file_copy(p, self$split_file, overwrite = TRUE)
    if (file.size(self$split_file) == 0) {
      runtime_error("Downloaded split file `{self$split_file}` is empty.")
    }
    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },
  check_exists = function() {
    fs::file_exists(self$split_file) &&
      fs::file_exists(self$images_dir) &&
      length(self$img_files) > 1
  },

  .getitem = function(index) {
    filename <- fs::path_ext_remove(self$data[index])

    image_path <- grep(paste0(filename,"."), self$img_files, value = TRUE, fixed = TRUE)
    if (length(image_path) != 1) {
      value_error("Image file `{filename}` not found.")
    }
    image_ext <- fs::path_ext(image_path)
    if (image_ext == "jpg") {
      x <- jpeg::readJPEG(image_path)
    } else {
      x <- suppressWarnings(tiff::readTIFF(image_path)) %>% aperm(c(3,1,2))
    }

    # Ensure label exists in class_to_idx
    label <- sub("_.*", "", filename)  # Ensure label is a character string
    if (!label %in% names(self$class_to_idx)) {
      value_error("Label `{label}` not found in class_to_idx." )
    }

    # Convert label
    y <- self$class_to_idx[[label]] + 1

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform))
      y <- self$target_transform(y)

    list(x = x, y = y)
  },

  .length = function() {
    length(self$data)
  }
)


#' EuroSAT All Bands dataset
#'
#' Downloads and prepares the EuroSAT dataset with 13 spectral bands.
#'
#' @rdname eurosat_dataset
#' @details `eurosat_all_bands_dataset()` provides a total of 27,000 labeled images with 13 spectral channel bands.
#'
#' @export
eurosat_all_bands_dataset <- torch::dataset(
  name = "eurosat_all_bands",
  inherit = eurosat_dataset,
  archive_url = "https://huggingface.co/datasets/torchgeo/eurosat/resolve/main/EuroSATallBands.zip?download=true",
  archive_md5 = "5ac12b3b2557aa56e1826e981e8e200e",
  split_url = "https://huggingface.co/datasets/torchgeo/eurosat/resolve/main/eurosat-{split}.txt?download=true",
  archive_size = "2 GB"
)




#' EuroSAT-100 dataset
#'
#' A subset of 100 images with 13 spectral bands useful for workshops and demos.
#'
#' @rdname eurosat_dataset
#' @details `eurosat100_dataset()` provides a subset of 100 labeled images with 13 spectral channel bands.
#'
#' @export
eurosat100_dataset <- torch::dataset(
  name = "eurosat100",
  inherit = eurosat_dataset,
  archive_url = "https://huggingface.co/datasets/torchgeo/eurosat/resolve/main/EuroSAT100.zip?download=true",
  archive_md5 = "c21c649ba747e86eda813407ef17d596",
  split_url = "https://huggingface.co/datasets/torchgeo/eurosat/resolve/main/eurosat-100-{split}.txt?download=true",
  archive_size = "7 MB"
)

has_ops <- function() {
  not_implemented_error("has_ops() Not implemented yet. https://github.com/pytorch/vision/blob/b266c2f1a5c10f5caf22f5aea7418acc392a5075/torchvision/extension.py#L60")
}

assert_has_ops <- function() {
  if(!has_ops()) {
    runtime_error(
      "Couldn't load custom C++ ops. This can happen if your torch and torchvision
       versions are incompatible, or if you had errors while compiling torchvision
       from source. Please reinstall torchvision so that it matches your torch install."
    )
  }
}

alexnet <- torch::nn_module(
  "AlexNet",
  initialize = function(num_classes = 1000) {
    self$features <- torch::nn_sequential(
      torch::nn_conv2d(3, 64, kernel_size = 11, stride = 4, padding = 2),
      torch::nn_relu(inplace = TRUE),
      torch::nn_max_pool2d(kernel_size = 3, stride = 2),
      torch::nn_conv2d(64, 192, kernel_size = 5, padding = 2),
      torch::nn_relu(inplace = TRUE),
      torch::nn_max_pool2d(kernel_size = 3, stride = 2),
      torch::nn_conv2d(192, 384, kernel_size = 3, padding = 1),
      torch::nn_relu(inplace = TRUE),
      torch::nn_conv2d(384, 256, kernel_size = 3, padding = 1),
      torch::nn_relu(inplace = TRUE),
      torch::nn_conv2d(256, 256, kernel_size = 3, padding = 1),
      torch::nn_relu(inplace = TRUE),
      torch::nn_max_pool2d(kernel_size = 3, stride = 2)
    )
    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(6,6))
    self$classifier <- torch::nn_sequential(
      torch::nn_dropout(),
      torch::nn_linear(256 * 6 * 6, 4096),
      torch::nn_relu(inplace = TRUE),
      torch::nn_dropout(),
      torch::nn_linear(4096, 4096),
      torch::nn_relu(inplace = TRUE),
      torch::nn_linear(4096, num_classes)
    )
  },
  forward = function(x) {
    x <- self$features(x)
    x <- self$avgpool(x)
    x <- torch::torch_flatten(x, start_dim = 2)
    x <- self$classifier(x)
  }
)

#' AlexNet Model Architecture
#'
#' AlexNet model architecture from the
#'   [One weird trick...](https://arxiv.org/abs/1404.5997) paper.
#'
#' @param pretrained (bool): If TRUE, returns a model pre-trained on ImageNet.
#' @param progress (bool): If TRUE, displays a progress bar of the download to
#'   stderr.
#' @param ... other parameters passed to the model intializer. currently only
#'   `num_classes` is used.
#'
#' @family classification_model
#'
#' @export
model_alexnet <- function(pretrained = FALSE, progress = TRUE, ...) {
  r <- c("https://torch-cdn.mlverse.org/models/vision/v2/models/alexnet.pth", "41ac4efd60b7e72480c4b9ba75618507", "~235 MB" )
  model <- alexnet(...)

  if (pretrained) {
    cli_inform("Model weights for {.cls {class(model)[1]}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1])
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict)
  }

  model
}

test_that("alexnet", {

  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

    m <- model_alexnet()
  input <- torch::torch_randn(1, 3, 256, 256)

  out <- m(input)

  expect_tensor_shape(out, c(1, 1000))

  m <- model_alexnet(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 256, 256)

  out <- m(input)

  expect_tensor_shape(out, c(1, 1000))

})

#' @include folder-dataset.R
NULL

#' RoboFlow 100 Document dataset Collection
#'
#' Loads one of the RoboFlow 100 Document datasets with
#' bounding box annotations for object-detection task.
#'
#' @param dataset Dataset to select within \code{c("tweeter_post", "tweeter_profile", "document_part",
#'   "activity_diagram", "signature", "paper_part", "tabular_data", "paragraph", "currency", "wine_label")}.
#' @param split the subset of the dataset to choose between \code{c("train", "test", "valid")}.
#' @param download Logical. If TRUE, downloads the dataset if not present at `root`.
#' @param transform Optional transform function applied to the image.
#' @param target_transform Optional transform function applied to the target.
#'
#' @return A torch dataset. Each element is a named list with:
#' - `x`: H x W x 3 array representing the image, auto-oriented and stretched to 640 x 640.
#' - `y`: a list containing the target with:
#'     - `image_id`: numeric identifier of the x image.
#'     - `labels`: numeric identifier of the N bounding-box object class.
#'     - `boxes`: a torch_tensor of shape (N, 4) with bounding boxes, each in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
#'
#' The returned item inherits the class `image_with_bounding_box` so it can be
#' visualised with helper functions such as [draw_bounding_boxes()].
#'
#' @examples
#' \dontrun{
#' ds <- rf100_document_collection(
#'   dataset = "tweeter_post",
#'   split = "train",
#'   transform = transform_to_tensor,
#'   download = TRUE
#' )
#'
#' # Retrieve a sample and inspect annotations
#' item <- ds[1]
#' item$y$labels
#' item$y$boxes
#'
#' # Draw bounding boxes and display the image
#' boxed_img <- draw_bounding_boxes(item)
#' tensor_image_browse(boxed_img)
#' }
#'
#' @name rf100_document_collection
#' @title RF100 Document Collection Datasets
#' @family detection_dataset
#' @export
rf100_document_collection <- torch::dataset(
  name = "rf100_document_collection",

  resources = data.frame(
    dataset = rep(c("tweeter_post", "tweeter_profile", "document_part",
                    "activity_diagram", "signature", "paper_part", "currency", "wine_label"), each = 3),
    split   = rep(c("train", "test", "valid"), times = 8),
    url = c(
      # tweeter_post
      "https://huggingface.co/datasets/Francesco/tweeter-posts/resolve/main/data/train-00000-of-00001-5ca0e754c63f9a31.parquet",
      "https://huggingface.co/datasets/Francesco/tweeter-posts/resolve/main/data/test-00000-of-00001-489e49c5dfdce787.parquet",
      "https://huggingface.co/datasets/Francesco/tweeter-posts/resolve/main/data/validation-00000-of-00001-2ab456c0d1f04f82.parquet",

      # tweeter_profile
      "https://huggingface.co/datasets/Francesco/tweeter-profile/resolve/main/data/train-00000-of-00001-1c7085071a0f4972.parquet",
      "https://huggingface.co/datasets/Francesco/tweeter-profile/resolve/main/data/test-00000-of-00001-be2dd504b504117f.parquet",
      "https://huggingface.co/datasets/Francesco/tweeter-profile/resolve/main/data/validation-00000-of-00001-747ade4e297e090a.parquet",

      # document_part
      "https://huggingface.co/datasets/Francesco/document-parts/resolve/main/data/train-00000-of-00001-5503c1fa031a4929.parquet",
      "https://huggingface.co/datasets/Francesco/document-parts/resolve/main/data/test-00000-of-00001-6cb74e4a35ca2ba5.parquet",
      "https://huggingface.co/datasets/Francesco/document-parts/resolve/main/data/validation-00000-of-00001-53417fa849d940f8.parquet",

      # activity_diagram
      "https://huggingface.co/datasets/Francesco/activity-diagrams-qdobr/resolve/main/data/train-00000-of-00001-9c2ac6dd4a9e53d8.parquet",
      "https://huggingface.co/datasets/Francesco/activity-diagrams-qdobr/resolve/main/data/test-00000-of-00001-acf5b67e3c7ca657.parquet",
      "https://huggingface.co/datasets/Francesco/activity-diagrams-qdobr/resolve/main/data/validation-00000-of-00001-d6f6f66b7dc88280.parquet",

      # signature
      "https://huggingface.co/datasets/Francesco/signatures-xc8up/resolve/main/data/train-00000-of-00001-aab07332622fb759.parquet",
      "https://huggingface.co/datasets/Francesco/signatures-xc8up/resolve/main/data/test-00000-of-00001-acf5b67e3c7ca657.parquet",
      "https://huggingface.co/datasets/Francesco/signatures-xc8up/resolve/main/data/validation-00000-of-00001-2cd116e72c9571b0.parquet",

      # paper_part
      "https://huggingface.co/datasets/Francesco/paper-parts/resolve/main/data/train-00000-of-00001-0f677be56de6ff94.parquet",
      "https://huggingface.co/datasets/Francesco/paper-parts/resolve/main/data/test-00000-of-00001-94db1ab1c191f5e2.parquet",
      "https://huggingface.co/datasets/Francesco/paper-parts/resolve/main/data/validation-00000-of-00001-2ce552e0b2a0aac5.parquet",

      # currency
      "https://huggingface.co/datasets/Francesco/currency-v4f8j/resolve/main/data/train-00000-of-00001-3c80a5c3981074e8.parquet",
      "https://huggingface.co/datasets/Francesco/currency-v4f8j/resolve/main/data/test-00000-of-00001-06a9ff754a1f48d1.parquet",
      "https://huggingface.co/datasets/Francesco/currency-v4f8j/resolve/main/data/validation-00000-of-00001-6dbf43e62bbf1c54.parquet",

      # wine labels
      "https://huggingface.co/datasets/Francesco/wine-labels/resolve/main/data/train-00000-of-00001-e8550a2abd7d3aea.parquet",
      "https://huggingface.co/datasets/Francesco/wine-labels/resolve/main/data/test-00000-of-00001-8dea61f083eb4b77.parquet",
      "https://huggingface.co/datasets/Francesco/wine-labels/resolve/main/data/validation-00000-of-00001-ef878f604ee5efed.parquet"
    ),

    md5 = c(
      # tweeter_post
      "30f8ee708cdfc443bfa8f9bc1d89e3b2",
      "00d694afad4f384d37fadbc9325a16ad",
      "14b26370147438e5c7c9b9de246d6891",
      # tweeter_profile
      "9a5fe681eded1fc8a08975a5ed142b24",
      "b54028739024b34e2acda860eb6a8068",
      "1eb31069867c3f855bda0aa269bb1eda",
      # document_part
      "5181f82eb8f91d92dd225dd23387b5e5",
      "a8bb1bd010ece905acb8a3ec2851de93",
      "85a20f23b6d53fb6a5148b125dd3ec4c",
      # activity_diagram
      "340249bc764ebcd4c00c243fdd75b773",
      "04d07c012caf23e643c4e28f15d43f83",
      "e2d3050bd5fed14664e71444e8df2ab9",
      # signature
      "742e26e8ee5d4d8605da68dda4df3c62",
      "e07083621e8b1cf73f6daea8f93e8943",
      "46836f6ebe462605e37b4640c0cc336d",
      # paper_part
      "253fd2189e89eb4664e89e9ed4b08dcc",
      "23cebd238571b4d130a11a6df760a180",
      "e02e6da02d92de11283739c5b0daeb4b",
      # currency
      "bfdca87eaf49018d16b13f9e08b1d8de",
      "b3c94d97b8fc3edf25ca0eaf31b5db4f",
      "80b5c6352457d66ef1c2dccd607ec732",
      # wine_label
      "f124cc6ca0969bce4a9195fea1130775",
      "a388d8b1abfaf15d45b330982e1e85a6",
      "ca0b96eb696da512eed629895a433bad"
    ),

    size = c(rep(50, 18), 32, 9, 5, 108, 32, 23) * 1e6 # placeholder; optional
  ),

  initialize = function(
    dataset, split = c("train", "test", "valid"),
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
    if (!requireNamespace("arrow", quietly = TRUE)) install.packages("arrow")
    if (!requireNamespace("prettyunits", quietly = TRUE)) install.packages("prettyunits")

    self$dataset <- match.arg(dataset, self$resources$dataset)
    self$split   <- match.arg(split)
    self$transform <- transform
    self$target_transform <- target_transform

    sel <- self$resources$dataset == self$dataset & self$resources$split == self$split
    self$archive_url  <- self$resources$url[sel]
    self$archive_size <- prettyunits::pretty_bytes(sum(self$resources$size[sel]))
    self$archive_md5  <- self$resources$md5[sel]
    self$split_file   <- file.path(rappdirs::user_cache_dir("torch"), class(self)[1], self$dataset, basename(self$archive_url))

    if (download) {
      cli_inform("Dataset {.val {self$dataset}} split {.val {self$split}} of {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      runtime_error("Dataset not found. Use download=TRUE or check that parquet files exist at the expected paths.")

    ads <- arrow::open_dataset(self$split_file)
    self$classes <- jsonlite::parse_json(ads$metadata$huggingface, simplifyVector = TRUE)$info$features$objects$feature$category$names[-1]

    # single parquet file versus arrow dataset, and only keep bboxed images
    if (sum(sel) == 1) {
      self$.data <- arrow::read_parquet(self$split_file) %>% subset(lengths(objects$bbox) > 0)
    } else {
      self$.data <- ads$to_data_frame() %>% subset(lengths(objects$bbox) > 0)
    }

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {
    if (self$check_exists()) return(invisible(NULL))
    cli_inform("Downloading {.val {self$dataset}}...")

    archive <- sapply(self$archive_url, download_and_cache, prefix = file.path(class(self)[1], self$dataset))

    if (!all(tools::md5sum(archive) == self$archive_md5))
      runtime_error("Corrupt file! Delete the cached files and try again.")

    for (i in seq_along(archive)) fs::file_move(archive[i], self$split_file[i])
  },

  check_exists = function() all(fs::file_exists(self$split_file)),

  .getitem = function(index) {
    df <- self$.data[index, ]
    x_raw <- unlist(df$image$bytes) %>% as.raw()
    if (tolower(tools::file_ext(df$image$path)) == "jpg") {
      x <- jpeg::readJPEG(x_raw)
    } else {
      x <- png::readPNG(x_raw)
    }
    # remove alpha channel
    if (length(dim(x)) == 3 && dim(x)[3] == 4) x <- x[, , 1:3, drop = FALSE]

    bbox <- df$objects$bbox
    boxes  <- box_xywh_to_xyxy(torch::torch_tensor(unlist(bbox), dtype = torch::torch_float())$view(c(-1, 4)))
    labels <- unlist(df$objects$category)
    if (is.null(labels)) {
      labels <- unlist(df$objects$label)
    }

    y <- list(image_id = df$image_id, labels = labels, boxes = boxes)
    if (!is.null(self$transform)) x <- self$transform(x)
    if (!is.null(self$target_transform)) y <- self$target_transform(y)

    item <- list(x = x, y = y)
    class(item) <- "image_with_bounding_box"
    item
  },

  .getbatch = function(index) {
    df <- self$.data[index, ]
    # hoist bytes
    df$bytes <- df$image$bytes

    df$is_jpg <- tolower(tools::file_ext(df$image$path)) == "jpg"

    x_lst <- mapply(function(bytes, is_jpg) {
      if (is_jpg) {
        x <- jpeg::readJPEG(as.raw(bytes))
      } else {
        x <- png::readPNG(as.raw(bytes))
      }
      # remove alpha channel if any
      if (length(dim(x)) == 3 && dim(x)[3] == 4) x <- x[, , 1:3, drop = FALSE]
      x
      }, bytes = df$bytes, is_jpg = df$is_jpg, SIMPLIFY = FALSE)

    # y, step 1 unnest and select `objects`
    unnested_df <- do.call(rbind, lapply(seq_len(nrow(df)), function(row) {
      obj_df <- df$objects[row, ]
      obj_df$bbox <- list(unlist(obj_df$bbox))
      obj_df$category <- list(unlist(obj_df$category))
      return(obj_df[,c("bbox", "category")])
    }))
    # y step 2 select, and unnest_longer bbox and label
    unnested_bbox <- box_xywh_to_xyxy(torch::torch_tensor(unlist(unnested_df$bbox), dtype = torch::torch_float())$view(c(-1, 4)))
    unnested_label <- unlist(unnested_df$category)
    # y step 3 repeat image_id along each of unnested value
    image_id_rep <- rep(df$image_id, lengths(unnested_df$category))

    y <- list(image_id = image_id_rep, labels = unnested_label, boxes = unnested_bbox)

    if (!is.null(self$transform)) {
      x_lst <- self$transform(x_lst)
    }
    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    item <- list(x = x_lst, y = y)
    class(item) <- "image_with_bounding_box"
    item
  },

  .length = function() nrow(self$.data)
)

#' Quantized MobileNetV3 Large model architecture.
#'
#' This function mirrors `torchvision::quantization::mobilenet_v3_large` and
#' loads quantized weights when `pretrained` is `TRUE`.
#'
#' @inheritParams model_mobilenet_v3_large
#' @family classification_model
#' @rdname model_mobilenet_v3
#' @name model_mobilenet_v3
#' @export
model_mobilenet_v3_large_quantized <- function(pretrained = FALSE, progress = TRUE, ...) {
  # resources
  r <- c("https://torch-cdn.mlverse.org/models/vision/v2/models/mobilenet_v3_large_quantized.pth",
         "06af6062e42ad3c80e430219a6560ca0", "~13 MB")
  model <- mobilenet_v3_large_quant_impl(...)

  if (pretrained) {
    cli_inform("Model weights for {.cls {class(model)[1]}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "mobilenet")
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict, strict = FALSE)
  }

  model
}


# ORIGINAL WORKING REGULAR MODEL COMPONENTS
nn_hardswish <- torch::nn_module(
  initialize = function(inplace = TRUE) {
    self$inplace <- inplace
  },
  forward = function(x) {
    x * torch::nnf_relu6(x + 3, inplace = self$inplace) / 6
  }
)

nn_hardsigmoid <- torch::nn_module(
  initialize = function(inplace = TRUE) {
    self$inplace <- inplace
  },
  forward = function(x) {
    torch::nnf_relu6(x + 3, inplace = self$inplace) / 6
  }
)

conv_norm_act_v3 <- torch::nn_module(
  inherit = torch::nn_sequential,
  initialize = function(in_channels, out_channels, kernel_size = 3, stride = 1,
                        groups = 1, norm_layer = NULL, activation_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    if (is.null(activation_layer))
      activation_layer <- nn_hardswish
    padding <- (kernel_size - 1) %/% 2
    super$initialize(
      torch::nn_conv2d(in_channels, out_channels, kernel_size, stride, padding,
                       groups = groups, bias = FALSE),
      norm_layer(out_channels),
      activation_layer(inplace = TRUE)
    )
    self$out_channels <- out_channels
  }
)

se_block_v3 <- torch::nn_module(
  initialize = function(in_channels, squeeze_channels) {
    self$avgpool <- torch::nn_adaptive_avg_pool2d(output_size = 1)
    self$fc1 <- torch::nn_conv2d(in_channels, squeeze_channels, kernel_size = 1)
    self$activation <- torch::nn_relu(inplace = TRUE)
    self$fc2 <- torch::nn_conv2d(squeeze_channels, in_channels, kernel_size = 1)
    self$scale_activation <- nn_hardsigmoid(inplace = TRUE)
  },
  forward = function(x) {
    scale <- self$avgpool(x)
    scale <- self$fc1(scale)
    scale <- self$activation(scale)
    scale <- self$fc2(scale)
    scale <- self$scale_activation(scale)
    x * scale
  }
)

inverted_residual_v3 <- torch::nn_module(
  initialize = function(in_channels, expand_channels, out_channels, kernel_size,
                        stride, use_se = FALSE, activation_layer = nn_hardswish,
                        norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    self$use_res_connect <- stride == 1 && in_channels == out_channels
    layers <- list()

    if (expand_channels != in_channels) {
      layers[[length(layers) + 1]] <- conv_norm_act_v3(
        in_channels, expand_channels, kernel_size = 1,
        norm_layer = norm_layer, activation_layer = activation_layer
      )
    }

    layers[[length(layers) + 1]] <- conv_norm_act_v3(
      expand_channels, expand_channels, kernel_size = kernel_size, stride = stride,
      groups = expand_channels, norm_layer = norm_layer,
      activation_layer = activation_layer
    )

    if (use_se) {
      make_divisible <- function(v, divisor = 8, min_value = NULL) {
        if (is.null(min_value)) min_value <- divisor
        new_v <- max(min_value, as.integer(v + divisor / 2) %/% divisor * divisor)
        if (new_v < 0.9 * v) new_v <- new_v + divisor
        new_v
      }

      squeeze_channels <- make_divisible(expand_channels / 4)
      layers[[length(layers) + 1]] <- se_block_v3(expand_channels, squeeze_channels)
    }

    layers[[length(layers) + 1]] <- torch::nn_sequential(
      torch::nn_conv2d(expand_channels, out_channels, 1, bias = FALSE),
      norm_layer(out_channels)
    )

    self$block <- torch::nn_sequential(!!!layers)
  },
  forward = function(x) {
    out <- self$block(x)
    if (self$use_res_connect)
      out <- out + x
    out
  }
)

# ORIGINAL WORKING REGULAR MODEL IMPLEMENTATION
mobilenet_v3_large_impl <- torch::nn_module(
  "mobilenet_v3_large_impl",
  initialize = function(num_classes = 1000, dropout = 0.2, norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    activation <- nn_hardswish

    layers <- list()
    layers[[length(layers) + 1]] <- conv_norm_act_v3(3, 16, stride = 2,
                                                     norm_layer = norm_layer,
                                                     activation_layer = activation)
    input_channel <- 16

    cfgs <- list(
      list(3, 16, 16, FALSE, torch::nn_relu, 1),
      list(3, 64, 24, FALSE, torch::nn_relu, 2),
      list(3, 72, 24, FALSE, torch::nn_relu, 1),
      list(5, 72, 40, TRUE, torch::nn_relu, 2),
      list(5, 120, 40, TRUE, torch::nn_relu, 1),
      list(5, 120, 40, TRUE, torch::nn_relu, 1),
      list(3, 240, 80, FALSE, activation, 2),
      list(3, 200, 80, FALSE, activation, 1),
      list(3, 184, 80, FALSE, activation, 1),
      list(3, 184, 80, FALSE, activation, 1),
      list(3, 480, 112, TRUE, activation, 1),
      list(3, 672, 112, TRUE, activation, 1),
      list(5, 672, 160, TRUE, activation, 2),
      list(5, 960, 160, TRUE, activation, 1),
      list(5, 960, 160, TRUE, activation, 1)
    )

    for (cfg in cfgs) {
      k <- cfg[[1]]; exp <- cfg[[2]]; c <- cfg[[3]]; se <- cfg[[4]];
      nl <- cfg[[5]]; s <- cfg[[6]]
      layers[[length(layers) + 1]] <- inverted_residual_v3(
        input_channel, exp, c, kernel_size = k, stride = s, use_se = se,
        activation_layer = nl, norm_layer = norm_layer
      )
      input_channel <- c
    }

    layers[[length(layers) + 1]] <- conv_norm_act_v3(
      input_channel, 960, kernel_size = 1,
      norm_layer = norm_layer, activation_layer = activation
    )

    self$features <- torch::nn_sequential(!!!layers)
    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(1,1))
    self$classifier <- torch::nn_sequential(
      torch::nn_linear(960, 1280),
      activation(inplace = TRUE),
      torch::nn_dropout(dropout),
      torch::nn_linear(1280, num_classes)
    )
  },
  forward = function(x) {
    x <- self$features(x)
    x <- self$avgpool(x)
    x <- torch::torch_flatten(x, start_dim = 2)
    self$classifier(x)
  }
)


###########################################################################################
# QUANTIZED MODEL COMPONENTS
nn_identity <- torch::nn_module(
  "nn_identity",
  initialize = function() {},
  forward = function(x) x
)
# Fixed Quantized Conv-BN-Activation block to match .pth structure
conv_bn_act_quantized <- torch::nn_module(
  initialize = function(in_channels, out_channels, kernel_size = 3, stride = 1,
                        groups = 1, norm_layer = NULL, activation_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    if (is.null(activation_layer))
      activation_layer <- nn_hardswish

    padding <- (kernel_size - 1) %/% 2

    # Create nested structure to match .pth: features.X.0.0.weight, features.X.0.bn.weight, etc.
    self$`0` <- torch::nn_module()
    self$`0`$`0` <- torch::nn_conv2d(in_channels, out_channels, kernel_size, stride, padding,
                                     groups = groups, bias = FALSE)
    self$`0`$bn <- norm_layer(out_channels)
    self$`0`$`2` <- activation_layer(inplace = TRUE)

    # Quantization stubs
    self$`0`$activation_post_process <- nn_identity()

    self$out_channels <- out_channels
  },
  forward = function(x) {
    x <- self$`0`$`0`(x)
    x <- self$`0`$bn(x)
    x <- self$`0`$`2`(x)
    x
  }
)

# Updated Quantized Inverted Residual Block to match structure
inverted_residual_quantized <- torch::nn_module(
  initialize = function(in_channels, expand_channels, out_channels, kernel_size,
                        stride, use_se = FALSE, activation_layer = nn_hardswish,
                        norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    self$use_res_connect <- stride == 1 && in_channels == out_channels

    # Create block structure to match expected .pth keys
    block_idx <- 0

    # Expansion layer (if needed)
    if (expand_channels != in_channels) {
      self[[as.character(block_idx)]] <- conv_bn_act_quantized(
        in_channels, expand_channels, kernel_size = 1,
        norm_layer = norm_layer, activation_layer = activation_layer
      )
      block_idx <- block_idx + 1
    }

    # Depthwise layer
    self[[as.character(block_idx)]] <- conv_bn_act_quantized(
      expand_channels, expand_channels, kernel_size = kernel_size, stride = stride,
      groups = expand_channels, norm_layer = norm_layer,
      activation_layer = activation_layer
    )
    block_idx <- block_idx + 1

    # SE block (if used)
    if (use_se) {
      make_divisible <- function(v, divisor = 8, min_value = NULL) {
        if (is.null(min_value)) min_value <- divisor
        new_v <- max(min_value, as.integer(v + divisor / 2) %/% divisor * divisor)
        if (new_v < 0.9 * v) new_v <- new_v + divisor
        new_v
      }

      squeeze_channels <- make_divisible(expand_channels / 4)
      self[[as.character(block_idx)]] <- se_block_quantized(expand_channels, squeeze_channels)
      block_idx <- block_idx + 1
    }

    # Project layer (no activation) - direct conv and bn
    self[[as.character(block_idx)]] <- torch::nn_module()
    self[[as.character(block_idx)]]$`0` <- torch::nn_conv2d(expand_channels, out_channels, 1, bias = FALSE)
    self[[as.character(block_idx)]]$bn <- norm_layer(out_channels)

    # Store the number of blocks for forward pass
    self$num_blocks <- block_idx + 1

    # Add quantization stubs
    self$activation_post_process <- nn_identity()
  },
  forward = function(x) {
    out <- x
    for (i in 0:(self$num_blocks - 1)) {
      layer <- self[[as.character(i)]]
      if (!is.null(layer)) {
        out <- layer(out)
      }
    }

    if (self$use_res_connect)
      out <- out + x
    out
  }
)

# Quantized SE Block
se_block_quantized <- torch::nn_module(
  initialize = function(in_channels, squeeze_channels) {
    self$avgpool <- torch::nn_adaptive_avg_pool2d(output_size = 1)
    self$fc1 <- torch::nn_conv2d(in_channels, squeeze_channels, kernel_size = 1)
    self$activation <- torch::nn_relu(inplace = TRUE)
    self$fc2 <- torch::nn_conv2d(squeeze_channels, in_channels, kernel_size = 1)
    self$scale_activation <- nn_hardsigmoid(inplace = TRUE)

    # Quantization stubs
    self$activation_post_process <- nn_identity()
  },
  forward = function(x) {
    scale <- self$avgpool(x)
    scale <- self$fc1(scale)
    scale <- self$activation(scale)
    scale <- self$fc2(scale)
    scale <- self$scale_activation(scale)
    x * scale
  }
)

# QUANTIZED MODEL IMPLEMENTATION
mobilenet_v3_large_quant_impl <- torch::nn_module(
  "mobilenet_v3_large_quant_impl",
  initialize = function(num_classes = 1000, dropout = 0.2, norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    # Quantization stubs
    self$quant <- nn_identity()
    self$dequant <- nn_identity()

    activation <- nn_hardswish

    # Create features as a module list to match the quantized structure
    self$features <- torch::nn_module_list()

    # First layer - create the nested structure manually to match features.0.0.bn.weight
    self$features[["0"]] <- torch::nn_module()
    self$features[["0"]]$`0` <- torch::nn_module()
    self$features[["0"]]$`0`$`0` <- torch::nn_conv2d(3, 16, 3, stride = 2, padding = 1, bias = FALSE)
    self$features[["0"]]$`0`$bn <- norm_layer(16)
    self$features[["0"]]$`0`$`2` <- activation(inplace = TRUE)
    self$features[["0"]]$`0`$activation_post_process <- nn_identity()

    input_channel <- 16

    cfgs <- list(
      list(3, 16, 16, FALSE, torch::nn_relu, 1),
      list(3, 64, 24, FALSE, torch::nn_relu, 2),
      list(3, 72, 24, FALSE, torch::nn_relu, 1),
      list(5, 72, 40, TRUE, torch::nn_relu, 2),
      list(5, 120, 40, TRUE, torch::nn_relu, 1),
      list(5, 120, 40, TRUE, torch::nn_relu, 1),
      list(3, 240, 80, FALSE, activation, 2),
      list(3, 200, 80, FALSE, activation, 1),
      list(3, 184, 80, FALSE, activation, 1),
      list(3, 184, 80, FALSE, activation, 1),
      list(3, 480, 112, TRUE, activation, 1),
      list(3, 672, 112, TRUE, activation, 1),
      list(5, 672, 160, TRUE, activation, 2),
      list(5, 960, 160, TRUE, activation, 1),
      list(5, 960, 160, TRUE, activation, 1)
    )

    layer_idx <- 1
    for (cfg in cfgs) {
      k <- cfg[[1]]; exp <- cfg[[2]]; c <- cfg[[3]]; se <- cfg[[4]];
      nl <- cfg[[5]]; s <- cfg[[6]]

      self$features[[as.character(layer_idx)]] <- inverted_residual_quantized(
        input_channel, exp, c, kernel_size = k, stride = s, use_se = se,
        activation_layer = nl, norm_layer = norm_layer
      )
      input_channel <- c
      layer_idx <- layer_idx + 1
    }

    # Final conv layer
    self$features[[as.character(layer_idx)]] <- conv_bn_act_quantized(
      input_channel, 960, kernel_size = 1,
      norm_layer = norm_layer, activation_layer = activation
    )

    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(1,1))

    # Classifier with quantization stubs
    self$classifier <- torch::nn_module()
    self$classifier$`0` <- torch::nn_linear(960, 1280)
    self$classifier$`1` <- activation(inplace = TRUE)
    self$classifier$`2` <- torch::nn_dropout(dropout)
    self$classifier$`3` <- torch::nn_linear(1280, num_classes)
  },
  forward = function(x) {
    x <- self$quant(x)

    # Forward through first layer manually
    x <- self$features[["0"]]$`0`$`0`(x)
    x <- self$features[["0"]]$`0`$bn(x)
    x <- self$features[["0"]]$`0`$`2`(x)

    # Forward through remaining features
    for (i in 1:(length(self$features) - 1)) {
      x <- self$features[[as.character(i)]](x)
    }

    x <- self$avgpool(x)
    x <- torch::torch_flatten(x, start_dim = 2)

    # Forward through classifier
    x <- self$classifier$`0`(x)
    x <- self$classifier$`1`(x)
    x <- self$classifier$`2`(x)
    x <- self$classifier$`3`(x)

    self$dequant(x)
  }
)

#' ConvNeXt Segmentation Models
#'
#' @description
#' Semantic segmentation models that use a ConvNeXt backbone with either
#' an FCN (Fully Convolutional Network) head or a UPerNet (Unified Perceptual
#' Parsing Network) head.
#'
#' These models follow the architecture patterns from mmsegmentation and
#' can be used for semantic segmentation tasks.
#'
#' @section Available FCN Models:
#' \itemize{
#'   \item `model_convnext_tiny_fcn()`
#'   \item `model_convnext_small_fcn()`
#'   \item `model_convnext_base_fcn()`
#' }
#'
#' @section Available UPerNet Models:
#' \itemize{
#'   \item `model_convnext_tiny_upernet()`
#'   \item `model_convnext_small_upernet()`
#'   \item `model_convnext_base_upernet()`
#' }
#'
#' @param num_classes Number of output segmentation classes. Default: 21 (PASCAL VOC).
#' @param aux_loss If TRUE, includes an auxiliary classifier branch. Default: FALSE.
#' @param pretrained If TRUE, loads convnext pretrained weights of backbone and segmentation heads.
#' @param pretrained_backbone If TRUE, loads ImageNet pretrained
#'   weights for the ConvNeXt backbone. Default: FALSE.
#' @param pool_scales Numeric vector. Pooling scales used in the Pyramid Pooling
#'   Module for UPerNet models. Default: c(1, 2, 3, 6).
#' @param ... Additional arguments passed to the backbone.
#'
#' @return An `nn_module` representing the segmentation model.
#'
#' @examples
#' \dontrun{
#' library(magrittr)
#' norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants
#' norm_std  <- c(0.229, 0.224, 0.225)
#'
#' # Use a publicly available image
#' wmc <- "https://upload.wikimedia.org/wikipedia/commons/thumb/"
#' url <- "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg"
#' img <- base_loader(paste0(wmc, url))
#'
#' input <- img %>%
#'   transform_to_tensor() %>%
#'   transform_resize(c(520, 520)) %>%
#'   transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)
#'
#' # ConvNeXt Tiny FCN segmentation
#' model <- model_convnext_tiny_fcn(num_classes = 21, pretrained_backbone = TRUE)
#' model$eval()
#' output <- model(batch)
#'
#' # Visualize result
#' segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
#' tensor_image_display(segmented)
#'
#' # ConvNeXt Tiny UPerNet segmentation
#' model <- model_convnext_tiny_upernet(num_classes = 21, pretrained_backbone = TRUE)
#' model$eval()
#' output <- model(batch)
#'
#' # Visualize result
#' segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
#' tensor_image_display(segmented)
#' }
#'
#' @family semantic_segmentation_model
#' @name model_convnext_segmentation
NULL


# ------------------------------------------------------------------------------
# ConvNeXt Backbone for Segmentation (extracts multi-scale features)
# ------------------------------------------------------------------------------

# Backbone wrapper for FCN-style models (returns out/aux keys)
convnext_fcn_backbone <- torch::nn_module(
  "convnext_fcn_backbone",
  initialize = function(convnext_model, out_channels, aux_channels) {
    self$model <- convnext_model
    self$out_channels <- out_channels
    self$aux_channels <- aux_channels
  },
  forward = function(x) {
    # Extract features at different scales
    c2 <- x %>%
      self$model$downsample_layers[[1]]() %>%
      self$model$stages[[1]]()

    c3 <- c2 %>%
      self$model$downsample_layers[[2]]() %>%
      self$model$stages[[2]]()

    c4 <- c3 %>%
      self$model$downsample_layers[[3]]() %>%
      self$model$stages[[3]]()

    c5 <- c4 %>%
      self$model$downsample_layers[[4]]() %>%
      self$model$stages[[4]]()

    # Return in FCN-compatible format (out = deepest, aux = intermediate)
    list(out = c5, aux = c4)
  }
)

# Backbone wrapper for UPerNet models (returns c2/c3/c4/c5 keys)
convnext_upernet_backbone <- torch::nn_module(
  "convnext_upernet_backbone",
  initialize = function(convnext_model) {
    self$model <- convnext_model
  },
  forward = function(x) {
    # Extract features at different scales
    c2 <- x %>%
      self$model$downsample_layers[[1]]() %>%
      self$model$stages[[1]]()

    c3 <- c2 %>%
      self$model$downsample_layers[[2]]() %>%
      self$model$stages[[2]]()

    c4 <- c3 %>%
      self$model$downsample_layers[[3]]() %>%
      self$model$stages[[3]]()

    c5 <- c4 %>%
      self$model$downsample_layers[[4]]() %>%
      self$model$stages[[4]]()

    list(c2 = c2, c3 = c3, c4 = c4, c5 = c5)
  }
)


# ------------------------------------------------------------------------------
# Pyramid Pooling Module (PPM) for UPerNet
# ------------------------------------------------------------------------------

ppm_module <- torch::nn_module(
  "PPM",
  initialize = function(in_channels, channels, pool_scales = c(1, 2, 3, 6)) {
    self$pool_scales <- pool_scales
    self$stages <- torch::nn_module_list()

    for (scale in pool_scales) {
      self$stages$append(
        torch::nn_sequential(
          torch::nn_adaptive_avg_pool2d(scale),
          torch::nn_conv2d(in_channels, channels, kernel_size = 1, bias = FALSE),
          torch::nn_batch_norm2d(channels),
          torch::nn_relu(inplace = TRUE)
        )
      )
    }
  },
  forward = function(x) {
    target_size <- x$shape[3:4]
    ppm_outs <- list()

    for (i in seq_along(self$stages)) {
      ppm_out <- self$stages[[i]](x)
      ppm_out <- torch::nnf_interpolate(
        ppm_out,
        size = as.integer(target_size),
        mode = "bilinear",
        align_corners = FALSE
      )
      ppm_outs[[i]] <- ppm_out
    }

    ppm_outs
  }
)


# ------------------------------------------------------------------------------
# UPerNet Head
# ------------------------------------------------------------------------------

upernet_head <- torch::nn_module(
  "UPerNetHead",
  initialize = function(in_channels, channels, num_classes, pool_scales = c(1, 2, 3, 6)) {
    # in_channels is a vector: c(c2_channels, c3_channels, c4_channels, c5_channels)
    self$in_channels <- in_channels
    self$channels <- channels
    self$num_classes <- num_classes

    # PSP module on deepest features (c5)
    self$psp_modules <- ppm_module(in_channels[4], channels, pool_scales)
    self$bottleneck <- torch::nn_sequential(
      torch::nn_conv2d(in_channels[4] + length(pool_scales) * channels, channels, kernel_size = 3, padding = 1, bias = FALSE),
      torch::nn_batch_norm2d(channels),
      torch::nn_relu(inplace = TRUE)
    )

    # Lateral convolutions for FPN
    self$lateral_convs <- torch::nn_module_list()
    self$fpn_convs <- torch::nn_module_list()

    for (i in 1:3) {  # c2, c3, c4
      l_conv <- torch::nn_sequential(
        torch::nn_conv2d(in_channels[i], channels, kernel_size = 1, bias = FALSE),
        torch::nn_batch_norm2d(channels),
        torch::nn_relu(inplace = TRUE)
      )
      fpn_conv <- torch::nn_sequential(
        torch::nn_conv2d(channels, channels, kernel_size = 3, padding = 1, bias = FALSE),
        torch::nn_batch_norm2d(channels),
        torch::nn_relu(inplace = TRUE)
      )
      self$lateral_convs$append(l_conv)
      self$fpn_convs$append(fpn_conv)
    }

    # Final bottleneck
    self$fpn_bottleneck <- torch::nn_sequential(
      torch::nn_conv2d(4 * channels, channels, kernel_size = 3, padding = 1, bias = FALSE),
      torch::nn_batch_norm2d(channels),
      torch::nn_relu(inplace = TRUE)
    )

    # Classification head
    self$cls_seg <- torch::nn_conv2d(channels, num_classes, kernel_size = 1)
  },

  psp_forward = function(x) {
    psp_outs <- list(x)
    psp_outs <- c(psp_outs, self$psp_modules(x))
    psp_out <- torch::torch_cat(psp_outs, dim = 2)
    self$bottleneck(psp_out)
  },

  forward = function(features) {
    # features is a list: list(c2, c3, c4, c5)
    inputs <- list(features$c2, features$c3, features$c4, features$c5)

    # Build laterals
    laterals <- list()
    for (i in 1:3) {
      laterals[[i]] <- self$lateral_convs[[i]](inputs[[i]])
    }
    # Add PSP output as the last lateral
    laterals[[4]] <- self$psp_forward(inputs[[4]])

    # Build top-down path
    for (i in 4:2) {
      prev_shape <- laterals[[i - 1]]$shape[3:4]
      laterals[[i - 1]] <- laterals[[i - 1]] + torch::nnf_interpolate(
        laterals[[i]],
        size = as.integer(prev_shape),
        mode = "bilinear",
        align_corners = FALSE
      )
    }

    # Build FPN outputs
    fpn_outs <- list()
    for (i in 1:3) {
      fpn_outs[[i]] <- self$fpn_convs[[i]](laterals[[i]])
    }
    fpn_outs[[4]] <- laterals[[4]]

    # Resize all to the largest feature map size
    target_size <- fpn_outs[[1]]$shape[3:4]
    for (i in 2:4) {
      fpn_outs[[i]] <- torch::nnf_interpolate(
        fpn_outs[[i]],
        size = as.integer(target_size),
        mode = "bilinear",
        align_corners = FALSE
      )
    }

    # Concatenate and apply bottleneck
    fpn_cat <- torch::torch_cat(fpn_outs, dim = 2)
    feats <- self$fpn_bottleneck(fpn_cat)

    # Classification
    self$cls_seg(feats)
  }
)


# UPerNet Segmentation Model
convnext_upernet <- torch::nn_module(
  "convnext_upernet",
  initialize = function(backbone, decode_head, aux_classifier = NULL) {
    self$backbone <- backbone
    self$decode_head <- decode_head
    self$aux_classifier <- aux_classifier
  },
  forward = function(x) {
    input_shape <- x$shape[3:4]
    features <- self$backbone(x)

    # Main decoder
    out <- self$decode_head(features)
    out <- torch::nnf_interpolate(out, size = input_shape, mode = "bilinear", align_corners = FALSE)

    result <- list(out = out)

    # Auxiliary classifier uses c4
    if (!is.null(self$aux_classifier)) {
      aux_out <- self$aux_classifier(features$c4)
      aux_out <- torch::nnf_interpolate(aux_out, size = input_shape, mode = "bilinear", align_corners = FALSE)
      result$aux <- aux_out
    }

    result
  }
)


# ------------------------------------------------------------------------------
# ConvNeXt FCN Model Factories
# ------------------------------------------------------------------------------

#' @describeIn model_convnext_segmentation ConvNeXt Tiny with FCN head
#' @export
model_convnext_tiny_fcn <- function(num_classes = 21,
                                    aux_loss = FALSE,
                                    pretrained_backbone = FALSE,
                                    ...) {
  validate_num_classes(num_classes, pretrained=FALSE)

  convnext <- model_convnext_tiny_1k(pretrained = pretrained_backbone, ...)
  # ConvNeXt Tiny dims: c(96, 192, 384, 768)
  backbone <- convnext_fcn_backbone(convnext, out_channels = 768, aux_channels = 384)

  classifier <- fcn_head(768, 512, num_classes)
  aux_classifier <- if (aux_loss) fcn_head(384, 256, num_classes) else NULL

  fcn(backbone, classifier, aux_classifier)
}


#' @describeIn model_convnext_segmentation ConvNeXt Small with FCN head
#' @export
model_convnext_small_fcn <- function(num_classes = 21,
                                     aux_loss = FALSE,
                                     pretrained_backbone = FALSE,
                                     ...) {
  validate_num_classes(num_classes, pretrained=FALSE)

  convnext <- model_convnext_small_22k(pretrained = pretrained_backbone, ...)
  # ConvNeXt Small dims: c(96, 192, 384, 768)
  backbone <- convnext_fcn_backbone(convnext, out_channels = 768, aux_channels = 384)

  classifier <- fcn_head(768, 512, num_classes)
  aux_classifier <- if (aux_loss) fcn_head(384, 256, num_classes) else NULL

  fcn(backbone, classifier, aux_classifier)
}


#' @describeIn model_convnext_segmentation ConvNeXt Base with FCN head
#' @export
model_convnext_base_fcn <- function(num_classes = 21,
                                    aux_loss = FALSE,
                                    pretrained_backbone = FALSE,
                                    ...) {
  validate_num_classes(num_classes, pretrained=FALSE)

  convnext <- model_convnext_base_1k(pretrained = pretrained_backbone, ...)
  # ConvNeXt Base dims: c(128, 256, 512, 1024)
  backbone <- convnext_fcn_backbone(convnext, out_channels = 1024, aux_channels = 512)

  classifier <- fcn_head(1024, 512, num_classes)
  aux_classifier <- if (aux_loss) fcn_head(512, 256, num_classes) else NULL

  fcn(backbone, classifier, aux_classifier)
}


# ------------------------------------------------------------------------------
# ConvNeXt UPerNet Model Factories
# ------------------------------------------------------------------------------
convnext_upernet_model_urls <- list(
  convnext_tiny_upernet_512_ade20k = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_tiny_upernet_512.pth",
    "6a8321d6e804aaa92a55d8b91b7d6f39",
    "230 MB"
  ),
  convnext_small_upernet_512_ade20k = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_small_upernet_512.pth",
    "7040ac0cb06a748d25e4d26e7d27eedc",
    "312 MB"
  ),
  convnext_base_upernet_512_ade20k = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_base_upernet_512.pth",
    "df98aa4c6bca23a3732dae19680eceb9",
    "466 MB"
  ),
  # not implemented yet
  convnext_base_upernet_640_ade20k = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_base_upernet_640.pth",
    "acc1a1e8cf3afbba62967b352bb9639a",
    "466 MB"
  ),
  convnext_large_upernet_640_ade20k = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_large_upernet_640.pth",
    "5f0405c1cb14b6a1e4e576fc787ac829",
    "896 MB"
  ),
  convnext_xlarge_upernet_640_ade20k = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_xlarge_upernet_640.pth",
    "c19d2cf463e2c08213f8e676dc554fa7",
    "1492 MB"
  )
)

#' @describeIn model_convnext_segmentation ConvNeXt Tiny with UPerNet head
#' @export
model_convnext_tiny_upernet <- function(num_classes = 21,
                                        aux_loss = FALSE,
                                        pretrained = FALSE,
                                        pretrained_backbone = FALSE,
                                        pool_scales = c(1, 2, 3, 6),
                                        ...) {
  validate_num_classes(num_classes, pretrained, label = "ade20k")

  if (pretrained && pretrained_backbone)
    cli_warn("`pretrained_backbone` ignored when `pretrained = TRUE`." )

  convnext <- model_convnext_tiny_1k(pretrained = pretrained_backbone, ...)
  backbone <- convnext_upernet_backbone(torch::nn_prune_head(convnext, 1))

  # ConvNeXt Tiny dims: c(96, 192, 384, 768)
  decode_head <- upernet_head(
    in_channels = c(96, 192, 384, 768),
    channels = 512,
    num_classes = num_classes,
    pool_scales = pool_scales
  )
  aux_classifier <- if (aux_loss || pretrained) fcn_head(384, 256, num_classes) else NULL

  model <- convnext_upernet(backbone, decode_head, aux_classifier)

  if (pretrained) {
    arch <- "convnext_tiny_upernet_512"
    info <- convnext_upernet_model_urls[[paste0(arch, "_ade20k")]]
    cli_inform("Downloading {.cls {arch}} pretrained weights (~{.emph {info[3]}}) ...")
    state_dict_path <- download_and_cache(info[1], prefix = "convnext")
    if (tools::md5sum(state_dict_path) != info[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(.rename_convnext_state_dict(state_dict), strict = FALSE)
  }
  model
}


#' @describeIn model_convnext_segmentation ConvNeXt Small with UPerNet head
#' @export
model_convnext_small_upernet <- function(num_classes = 21,
                                         aux_loss = FALSE,
                                         pretrained = FALSE,
                                         pretrained_backbone = FALSE,
                                         pool_scales = c(1, 2, 3, 6),
                                         ...) {
  validate_num_classes(num_classes, pretrained, label = "ade20k")

  if (pretrained && pretrained_backbone)
    cli_warn("`pretrained_backbone` ignored when `pretrained = TRUE`." )

  convnext <- model_convnext_small_22k(pretrained = pretrained_backbone, ...)
  backbone <- convnext_upernet_backbone(torch::nn_prune_head(convnext, 1))

  # ConvNeXt Small dims: c(96, 192, 384, 768)
  decode_head <- upernet_head(
    in_channels = c(96, 192, 384, 768),
    channels = 512,
    num_classes = num_classes,
    pool_scales = pool_scales
  )
  aux_classifier <- if (aux_loss || pretrained) fcn_head(384, 256, num_classes) else NULL

  model <- convnext_upernet(backbone, decode_head, aux_classifier)

  if (pretrained) {
    arch <- "convnext_small_upernet_512"
    info <- convnext_upernet_model_urls[[paste0(arch, "_ade20k")]]
    cli_inform("Downloading {.cls {arch}} pretrained weights (~{.emph {info[3]}}) ...")
    state_dict_path <- download_and_cache(info[1], prefix = "convnext")
    if (tools::md5sum(state_dict_path) != info[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(.rename_convnext_state_dict(state_dict), strict = FALSE)
  }
  model
}


#' @describeIn model_convnext_segmentation ConvNeXt Base with UPerNet head
#' @export
model_convnext_base_upernet <- function(num_classes = 21,
                                        aux_loss = FALSE,
                                        pretrained = FALSE,
                                        pretrained_backbone = FALSE,
                                        pool_scales = c(1, 2, 3, 6),
                                        ...) {
  validate_num_classes(num_classes, pretrained, label = "ade20k")

  if (pretrained && pretrained_backbone)
    cli_warn("`pretrained_backbone` ignored when `pretrained = TRUE`." )

  convnext <- model_convnext_base_1k(pretrained = pretrained_backbone, ...)
  backbone <- convnext_upernet_backbone(torch::nn_prune_head(convnext, 1))

  # ConvNeXt Base dims: c(128, 256, 512, 1024)
  decode_head <- upernet_head(
    in_channels = c(128, 256, 512, 1024),
    channels = 512,
    num_classes = num_classes,
    pool_scales = pool_scales
  )
  aux_classifier <- if (aux_loss || pretrained) fcn_head(512, 256, num_classes) else NULL

  model <- convnext_upernet(backbone, decode_head, aux_classifier)

  if (pretrained) {
    arch <- "convnext_base_upernet_512"
    info <- convnext_upernet_model_urls[[paste0(arch, "_ade20k")]]
    cli_inform("Downloading {.cls {arch}} pretrained weights (~{.emph {info[3]}}) ...")
    state_dict_path <- download_and_cache(info[1], prefix = "convnext")
    if (tools::md5sum(state_dict_path) != info[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(.rename_convnext_state_dict(state_dict), strict = FALSE)
  }
  model
}

#' @importFrom stats setNames
.rename_convnext_state_dict <- function(state_dict) {
  . <- NULL # Nulling strategy for no visible binding check Note
  new_names <- names(state_dict) %>%
    sub(pattern = "backbone\\.", replacement = "backbone.model.", x = .) %>%
    sub(pattern = "model\\.norm3", replacement = "model.norm", x = .) %>%
    sub(pattern = "psp_modules\\.", replacement = "psp_modules.stages.", x = .) %>%
    sub(pattern = "depthwise_", replacement = "dw", x = .) %>%
    sub(pattern = "pointwise_", replacement = "pw", x = .) %>%
    sub(pattern = "(\\d\\.\\d)\\.conv\\.weight", replacement = "\\1.weight", x = .) %>%
    sub(pattern = "\\.conv\\.weight", replacement = ".0.weight", x = .) %>%
    sub(pattern = "\\.(\\d)\\.1\\.bn\\.", replacement = ".\\1.2.", x = .) %>%
    sub(pattern = "\\.bn\\.", replacement = ".1.", x = .) %>%
    sub(pattern = "decode_head\\.conv_seg", replacement = "decode_head.cls_seg", x = .) %>%
    sub(pattern = "auxiliary_head\\.convs\\.0\\.", replacement = "aux_classifier.", x = .) %>%
    sub(pattern = "auxiliary_head\\.conv_seg\\.", replacement = "aux_classifier.4.", x = .)

  # Recreate a list with renamed keys
  setNames(state_dict[names(state_dict)], new_names)
}

# Tests for ConvNeXt Segmentation Models
# These tests cover FCN and UPerNet heads with ConvNeXt backbone

# FCN Head Tests

test_that("model_convnext_tiny_fcn returns correct output shape", {
  model <- model_convnext_tiny_fcn(num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_true(is.list(output))
  expect_named(output, "out")
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_small_fcn returns correct output shape", {
  model <- model_convnext_small_fcn(num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_true(is.list(output))
  expect_named(output, "out")
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_base_fcn returns correct output shape", {
  model <- model_convnext_base_fcn(num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_true(is.list(output))
  expect_named(output, "out")
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_tiny_fcn works with custom num_classes", {
  model <- model_convnext_tiny_fcn(num_classes = 3)
  model$eval()

  input <- torch::torch_randn(2, 3, 224, 224)
  output <- model(input)

  expect_tensor_shape(output$out, c(2, 3, 224, 224))
})

test_that("model_convnext_tiny_fcn with aux_loss returns aux output", {
  model <- model_convnext_tiny_fcn(num_classes = 21, aux_loss = TRUE)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_named(output, c("out", "aux"))
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
  expect_tensor_shape(output$aux, c(1, 21, 224, 224))
})

test_that("model_convnext_tiny_fcn works with different input sizes", {
  model <- model_convnext_tiny_fcn(num_classes = 21)
  model$eval()

  # Test with 256x256
  input <- torch::torch_randn(1, 3, 256, 256)
  output <- model(input)
  expect_tensor_shape(output$out, c(1, 21, 256, 256))

  # Test with 512x512
  input <- torch::torch_randn(1, 3, 512, 512)
  output <- model(input)
  expect_tensor_shape(output$out, c(1, 21, 512, 512))
})

test_that("model_convnext_tiny_fcn validates num_classes", {
  expect_error(
    model_convnext_tiny_fcn(num_classes = 0),
    "num_classes"
  )
  expect_error(
    model_convnext_tiny_fcn(num_classes = -1),
    "num_classes"
  )
})

# UPerNet Head Tests

test_that("model_convnext_tiny_upernet returns correct output shape", {
  model <- model_convnext_tiny_upernet(num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_true(is.list(output))
  expect_named(output, "out")
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_small_upernet returns correct output shape", {
  model <- model_convnext_small_upernet(num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_true(is.list(output))
  expect_named(output, "out")
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_base_upernet returns correct output shape", {
  model <- model_convnext_base_upernet(num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_true(is.list(output))
  expect_named(output, "out")
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_tiny_upernet works with custom num_classes", {
  model <- model_convnext_tiny_upernet(num_classes = 3)
  model$eval()

  input <- torch::torch_randn(2, 3, 224, 224)
  output <- model(input)

  expect_tensor_shape(output$out, c(2, 3, 224, 224))
})

test_that("model_convnext_tiny_upernet with aux_loss returns aux output", {
  model <- model_convnext_tiny_upernet(num_classes = 21, aux_loss = TRUE)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_named(output, c("out", "aux"))
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
  expect_tensor_shape(output$aux, c(1, 21, 224, 224))
})

test_that("model_convnext_tiny_upernet works with different input sizes", {
  model <- model_convnext_tiny_upernet(num_classes = 21)
  model$eval()

  # Test with 256x256
  input <- torch::torch_randn(1, 3, 256, 256)
  output <- model(input)
  expect_tensor_shape(output$out, c(1, 21, 256, 256))

  # Test with 512x512
  input <- torch::torch_randn(1, 3, 512, 512)
  output <- model(input)
  expect_tensor_shape(output$out, c(1, 21, 512, 512))
})

test_that("model_convnext_tiny_upernet validates num_classes", {
  expect_error(
    model_convnext_tiny_upernet(num_classes = 0),
    "num_classes"
  )
  expect_error(
    model_convnext_tiny_upernet(num_classes = -1),
    "num_classes"
  )
  expect_error(
    model_convnext_tiny_upernet(num_classes = 21, pretrained = TRUE),
    "num_classes"
  )
})

# Pretrained Backbone Tests

test_that("model_convnext_tiny_fcn works with pretrained backbone", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_convnext_tiny_fcn(num_classes = 21, pretrained_backbone = TRUE)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_tiny_upernet works with pretrained backbone", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_convnext_tiny_upernet(num_classes = 21, pretrained_backbone = TRUE)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_convnext_tiny_upernet works with pretrained model", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  expect_no_error(
    model <- model_convnext_tiny_upernet(num_classes = 150, pretrained = TRUE)
  )
  model$eval()

  input <- torch::torch_randn(1, 3, 512, 512)
  output <- model(input)

  expect_tensor_shape(output$out, c(1, 150, 512, 512))
})

test_that("model_convnext_small_upernet works with pretrained model", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  expect_no_error(
    model <- model_convnext_small_upernet(num_classes = 150, pretrained = TRUE)
  )
  model$eval()

  input <- torch::torch_randn(1, 3, 512, 512)
  output <- model(input)

  expect_tensor_shape(output$out, c(1, 150, 512, 512))
})

test_that("model_convnext_base_upernet works with pretrained model", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  expect_no_error(
    model <- model_convnext_base_upernet(num_classes = 150, pretrained = TRUE)
  )
  model$eval()

  input <- torch::torch_randn(1, 3, 512, 512)
  output <- model(input)

  expect_tensor_shape(output$out, c(1, 150, 512, 512))
})

# Pool Scales Tests (UPerNet specific)

test_that("model_convnext_tiny_upernet works with custom pool_scales", {
  model <- model_convnext_tiny_upernet(num_classes = 21, pool_scales = c(1, 2, 3))
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

#' @export
transform_convert_image_dtype.torch_tensor <- function(img, dtype = torch::torch_float()) {

  if (img$dtype == dtype)
    return(img)

  if (img$is_floating_point()) {

    # float to float
    if (dtype$is_floating_point)
      return(img$to(dtype = dtype))

    # float to int
    if ((img$dtype == torch::torch_float32() &&
         (dtype == torch::torch_float32() || dtype == torch::torch_float64())) ||
        (img$dtype == torch::torch_float64() && dtype == torch::torch_int64())
    )
      runtime_error("The cast from {img$dtype} to {dtype} cannot be performed safely.")

    # For data in the range 0-1, (float * 255).to(uint) is only 255
    # when float is exactly 1.0.
    # `max + 1 - epsilon` provides more evenly distributed mapping of
    # ranges of floats to ints.
    eps <- 1e-3
    result <- img$mul(torch::torch_iinfo(dtype)$max + 1 - eps)
    result <- result$to(dtype = dtype)

    return(result)
  } else {
    # int to float

    if (dtype$is_floating_point) {
      max <- torch::torch_iinfo(img$dtype)$max
      img <- img$to(dtype)

      return(img/max)
    }


    # int to int
    input_max <- torch::torch_iinfo(img$dtype)$max
    output_max <- torch::torch_iinfo(dtype)$max

    if (input_max > output_max) {
      factor <- (input_max + 1) %/% (output_max + 1)
      img <- img %/% factor
      return(img$to(dtype = dtype))
    } else {

      factor <- (output_max + 1) %/% (input_max + 1)
      img <- img$to(dtype = dtype)

      return(img * factor)
    }
  }

}

#' @export
transform_normalize.torch_tensor <- function(img, mean, std, inplace = FALSE) {

  check_img(img)

  if (!inplace)
    img <- img$clone()

  dtype <- img$dtype
  mean <- torch::torch_tensor(mean, dtype=dtype, device=img$device)
  std <- torch::torch_tensor(std, dtype=dtype, device=img$device)

  if ((std == 0)$any()$item()) {
    value_error("std evaluated to zero after conversion to {dtype}, leading to division by zero.")
  }

  unsqueezed <- if (img$ndim == 3) c(-1, 1, 1) else c(1, -1, 1, 1)
  if (mean$dim() == 1)
    mean <- mean$view(unsqueezed)

  if (std$dim() == 1)
    std <- std$view(unsqueezed)

  img$sub_(mean)$div_(std)
}

#' @export
transform_resize.torch_tensor <- function(img, size, interpolation = 2) {

  check_img(img)

  interpolation_modes <- c(
    "0" = "nearest",
    "2" = "bilinear",
    "3" = "bicubic"
  )

  if (!interpolation %in% names(interpolation_modes))
    value_error("This interpolation mode is unsupported with Tensor input")

  if (!length(size) %in% c(1,2) || !is.numeric(size))
    value_error("Size must be a numeric vector of length 1 or 2.")

  wh <- get_image_size(img)
  w <- wh[1]
  h <- wh[2]

  if (length(size) == 1) {

    if (w <= h) {
      short <- w
      long <- h
    } else {
      short <- h
      long <- w
    }

    requested_new_short <- size

    if (short == requested_new_short)
      return(img)

    new_short <- requested_new_short
    new_long <- as.integer(requested_new_short * long / short)

    if (w <= h) {
      new_w <- new_short
      new_h <- new_long
    } else {
      new_w <- new_long
      new_h <- new_short
    }

  } else {
    new_w <- size[2]
    new_h <- size[1]
  }

  # make NCHW
  need_squeeze <- FALSE
  if (img$dim() < 4) {
    img <- img$unsqueeze(1)
    need_squeeze <- TRUE
  }

  mode <- interpolation_modes[interpolation == names(interpolation_modes)]
  mode <- unname(mode)

  out_dtype <- img$dtype
  need_cast <- FALSE
  if (!img$dtype == torch::torch_float32() &&
      !img$dtype == torch::torch_float64()) {
    need_cast <- TRUE
    img <- img$to(dtype = torch::torch_float32())
  }

  if (mode %in% c("bilinear", "bicubic"))
    align_corners <- FALSE
  else
    align_corners <- NULL


  img <- torch::nnf_interpolate(img, size = c(new_h, new_w), mode = mode,
                                align_corners = align_corners)

  if (need_squeeze)
    img <- img$squeeze(1)

  if (need_cast) {
    if (mode == "bicubic")
      img <- img$clamp(min = 0, max = 255)

    img <- img$to(out_dtype)
  }

  img
}

#' @export
transform_pad.torch_tensor <- function(img, padding, fill = 0, padding_mode = "constant") {

  check_img(img)

  if (!length(padding) %in% c(1,2,4) || !is.numeric(padding))
    value_error("Padding must be an int or a 1, 2, or 4 element numeric vector")

  if (!padding_mode %in% c("constant", "edge", "reflect", "symmetric"))
    value_error("Padding mode should be either constant, edge, reflect or symmetric")


  if (length(padding) == 1)
    pad_left <- pad_right <- pad_top <- pad_bottom <- padding
  else if (length(padding) == 2) {
    pad_left <- pad_right <- padding[1]
    pad_top <- pad_bottom <- padding[2]
  } else if (length(padding) == 4) {
    pad_left <- padding[1]
    pad_right <- padding[2]
    pad_top <- padding[3]
    pad_bottom <- padding[4]
  }

  p <- c(pad_left, pad_right, pad_top, pad_bottom)

  if (padding_mode == "edge")
    padding_mode <- "replicate"
  else if (padding_mode == "symmetric") {
    if (any(p < 0))
      value_error("Padding can not be negative for symmetric padding_mode")

    return(pad_symmetric(img, p))
  }

  need_squeeze <- FALSE

  if (img$dim() < 4) {
    img <- img$unsqueeze(1)
    need_squeeze <- TRUE
  }

  out_dtype <- img$dtype
  need_cast <- FALSE

  if (padding_mode != "constant" &&
      (img$dtype == torch::torch_float32() || img$dtype == torch::torch_float64())) {
    need_cast <- TRUE
    img <- img$to(torch::torch_float32())
  }

  img <- torch::nnf_pad(img, p, mode = padding_mode, value = as.numeric(fill))

  if (need_squeeze)
    img <- img$squeeze(dim = 1)

  if (need_cast)
    img <- img$to(dtype = out_dtype)

  img
}

#' @export
transform_five_crop.torch_tensor <- function(img, size) {

  check_img(img)

  if (!length(size) == 2)
    value_error("Please provide only 2 dimensions (h, w) for size.")

  image_width <- img$size(-2)
  image_height <- img$size(-1)

  crop_height <- size[1]
  crop_width <- size[2]

  if (crop_width > image_width || crop_height > image_height)
    value_error("Requested crop size is bigger than input size.")

  tl <- transform_crop(img, 1, 1, crop_width, crop_height)
  tr <- transform_crop(img, image_width - crop_width + 1, 1, image_width, crop_height)
  bl <- transform_crop(img, 1, image_height - crop_height + 1, crop_width, image_height)
  br <- transform_crop(img, image_width - crop_width + 1, image_height - crop_height + 1, image_width, image_height)
  center <- transform_center_crop(img, c(crop_height, crop_width))

  list(tl, tr, bl, br, center)
}

#' @export
transform_ten_crop.torch_tensor <- function(img, size, vertical_flip = FALSE) {

  check_img(img)

  if (!length(size) == 2)
    value_error("Please provide only 2 dimensions (h, w) for size.")

  first_five <- transform_five_crop(img, size)

  if (vertical_flip)
    img <- transform_vflip(img)
  else
    img <- transform_hflip(img)

  second_five <- transform_five_crop(img, size)

  c(first_five, second_five)
}

#' @export
transform_linear_transformation.torch_tensor <- function(img, transformation_matrix,
                                                         mean_vector) {
  flat_dim <- if (img$ndim == 3) c(1, -1) else c(img$size(1), -1)
  flat_tensor <- img$view(flat_dim) - mean_vector
  transformed_tensor <- torch::torch_mm(flat_tensor, transformation_matrix)

  transformed_tensor$view(img$size())
}


# Other methods -----------------------------------------------------------

#' @export
transform_random_grayscale.torch_tensor <- function(img, p = 0.1) {
  check_img(img)

  if (stats::runif(1) > p) {
    img
  } else {
    transform_grayscale(img, num_output_channels = 3)
  }
}

#' @export
transform_grayscale.torch_tensor <- function(img, num_output_channels) {
  check_img(img)
  repeat_on_channel <- if(img$ndim == 3) c(num_output_channels, 1, 1) else c(1, num_output_channels, 1, 1)
  transform_rgb_to_grayscale(img)$`repeat`(repeat_on_channel)
}

#' @export
transform_crop.torch_tensor <- function(img, top, left, height, width) {
  check_img(img)

  img[.., top:(top + height - 1), left:(left + width - 1)]
}

#' @export
transform_hflip.torch_tensor <- function(img) {
  check_img(img)
  img$flip(-1)
}

#' @export
transform_vflip.torch_tensor <- function(img) {
  check_img(img)
  img$flip(-2)
}

#' @export
transform_adjust_brightness.torch_tensor <- function(img, brightness_factor) {
  if (brightness_factor < 0)
    value_error("brightness factor is negative")

  check_img(img)

  blend(img, torch::torch_zeros_like(img), brightness_factor)
}

#' @export
transform_adjust_contrast.torch_tensor <- function(img, contrast_factor) {

  if (contrast_factor < 0)
    value_error("contrast must be positive")

  check_img(img)

  mean <- torch::torch_mean(transform_rgb_to_grayscale(img)$to(torch::torch_float()))

  blend(img, mean, contrast_factor)
}

#' @export
transform_adjust_hue.torch_tensor <- function(img, hue_factor) {

  if (hue_factor < -0.5 || hue_factor > 0.5)
    value_error("hue_factor must be between -0.5 and 0.5.")

  check_img(img)

  orig_dtype <- img$dtype
  if (img$dtype == torch::torch_uint8())
    img <- img$to(dtype = torch::torch_float32())/255

  img <-rgb2hsv(img)

  channel_dim <- if (img$ndim == 3) 1 else 2
  hsv <- img$unbind(channel_dim)
  hsv[[1]] <- hsv[[1]] + hue_factor
  hsv[[1]] <- hsv[[1]] %% 1
  img <- torch::torch_stack(hsv, dim=channel_dim)
  img_hue_adj <- hsv2rgb(img)

  if (orig_dtype == torch::torch_uint8())
    img_hue_adj <- (img_hue_adj * 255.0)$to(dtype=orig_dtype)

  img_hue_adj
}

#' @export
transform_adjust_saturation.torch_tensor <- function(img, saturation_factor) {

  if (saturation_factor < 0)
    value_error("saturation factor must be positive.")

  check_img(img)


  blend(img, transform_rgb_to_grayscale(img), saturation_factor)
}

#' @export
transform_rotate.torch_tensor <- function(img, angle, interpolation = 0, expand = FALSE,
                                          center = NULL, fill = NULL, resample) {
  if (!missing(resample)) {
    interpolation <- resample
  }

  center_f <- c(0.0, 0.0)
  if (!is.null(center)) {
    img_size <- get_image_size(img)
    # Center values should be in pixel coordinates but translated such that (0, 0) corresponds to image center.
    center_f <- mapply(function(c, s) 1 * (c - s * 0.5), center, img_size)
  }


  # due to current incoherence of rotation angle direction between affine and
  # rotate implementations we need to set -angle.
  matrix <- get_inverse_affine_matrix(center_f, -angle, c(0.0, 0.0), 1.0, c(0.0, 0.0))

  rotate_impl(img, matrix=matrix, interpolation=interpolation, expand=expand, fill=fill)
}

#' @export
transform_affine.torch_tensor <- function(img, angle, translate, scale, shear,
                                          interpolation = 0, fill = NULL,
                                          resample, fillcolor, center = NULL) {
  if (!missing(resample)) {
    interpolation <- resample
  }
  if (!missing(fillcolor)) {
    fill <- fillcolor
  }

  if (!is.numeric(angle))
    value_error("`angle` should be int or float")

  if (!length(translate) == 2)
    value_error("`translate` should be length 2")

  if (scale <= 0)
    value_error("`scale` should be positive")

  if (!is.numeric(shear))
    type_error("`shear` should be either a single value or a sequence of 2 values")

  if (length(shear) == 1)
    shear <- c(shear, 0)

  center_f <- c(0.0, 0.0)
  if (!is.null(center)) {
    img_size <- get_image_size(img)
    # Center values should be in pixel coordinates but translated such that (0, 0) corresponds to image center.
    center_f <- mapply(function(c, s) 1 * (c - s * 0.5), center, img_size)
  }

  matrix <- get_inverse_affine_matrix(center_f, angle, translate, scale, shear)
  affine_impl(img, matrix=matrix, interpolation=interpolation, fill=fill)
}

#' @export
transform_perspective.torch_tensor <- function(img, startpoints, endpoints, interpolation = 2,
                                               fill = NULL) {
  not_implemented_error("`transform_perspective` is not implemented yet.")
}

#' @export
transform_rgb_to_grayscale.torch_tensor <- function(img) {
  check_img(img)
  (0.2989 * img[..,1,,] + 0.5870 * img[..,2,,] + 0.1140 * img[..,3,,])$to(img$dtype)
}

# Helpers -----------------------------------------------------------------

is_tensor_image <- function(x) {
  x$dim() >= 2
}

check_img <- function(x) {
  if (!is_tensor_image(x))
    type_error("tensor is not a torch image.")
}

#' @importFrom utils tail
get_image_size.torch_tensor <- function(img) {
  check_img(img)

  rev(tail(img$size(), 2))
}

blend <- function(img1, img2, ratio) {

  if (img1$is_floating_point())
    bound <- 1
  else
    bound <- 255

  (ratio * img1 + (1 - ratio) * img2)$clamp(0, bound)$to(img1$dtype)
}

rgb2hsv <- function(img) {
  channel_dim <- if (img$ndim == 3) 1 else 2

  rgb <- img$unbind(channel_dim)
  r <- rgb[[1]]; g <- rgb[[2]]; b <- rgb[[3]]

  maxc <- torch::torch_max(img, dim=channel_dim)[[1]]
  minc <- torch::torch_min(img, dim=channel_dim)[[1]]

  # The algorithm erases S and H channel where `maxc = minc`. This avoids NaN
  # from happening in the results, because
  #   + S channel has division by `maxc`, which is zero only if `maxc = minc`
  #   + H channel has division by `(maxc - minc)`.
  #
  # Instead of overwriting NaN afterwards, we just prevent it from occuring so
  # we don't need to deal with it in case we save the NaN in a buffer in
  # backprop, if it is ever supported, but it doesn't hurt to do so.
  eqc <- maxc == minc

  cr <- maxc - minc

  # Since `eqc => cr = 0`, replacing denominator with 1 when `eqc` is fine.
  s <- cr / torch::torch_where(eqc, maxc$new_full(list(), 1), maxc)
  # Note that `eqc => maxc = minc = r = g = b`. So the following calculation
  # of `h` would reduce to `bc - gc + 2 + rc - bc + 4 + rc - bc = 6` so it
  # would not matter what values `rc`, `gc`, and `bc` have here, and thus
  # replacing denominator with 1 when `eqc` is fine.
  cr_divisor <- torch::torch_where(eqc, maxc$new_full(list(), 1), cr)
  rc <- (maxc - r) / cr_divisor
  gc <- (maxc - g) / cr_divisor
  bc <- (maxc - b) / cr_divisor

  hr <- (maxc == r) * (bc - gc)
  hg <- ((maxc == g) & (maxc != r)) * (2.0 + rc - bc)
  hb <- ((maxc != g) & (maxc != r)) * (4.0 + gc - rc)
  h <- (hr + hg + hb)
  h <- torch::torch_fmod((h / 6.0 + 1.0), 1.0)
  torch::torch_stack(list(h, s, maxc), dim=channel_dim)
}

hsv2rgb <- function(img) {

  channel_dim <- if (img$ndim == 3) 1 else 2

  hsv <- img$unbind(channel_dim)
  i <- torch::torch_floor(hsv[[1]] * 6)
  f <- (hsv[[1]] * 6) - 1
  i <- i$to(dtype = torch::torch_int32())

  p <- torch::torch_clamp((hsv[[3]] * (1 - hsv[[2]])), 0, 1)
  q <- torch::torch_clamp((hsv[[3]] * (1 - hsv[[2]] * f)), 0, 1)
  t <- torch::torch_clamp((hsv[[3]] * (1 - f)), 0, 1)
  i <- i %% 6

  mask <- if (img$ndim == 3)
    i == torch::torch_arange(start= 0, end = 5)[,NULL,NULL]
  else
    i == torch::torch_arange(start= 0, end = 5)[,NULL,NULL,NULL]

  a1 <- torch::torch_stack(list(hsv[[3]], q, p, p, t, hsv[[3]]))
  a2 <- torch::torch_stack(list(t, hsv[[3]], hsv[[3]], q, p, p))
  a3 <- torch::torch_stack(list(p, p, t, hsv[[3]], hsv[[3]], q))
  a4 <- torch::torch_stack(list(a1, a2, a3))

  if (img$ndim == 3)
    torch::torch_einsum("ijk, xijk -> xjk", list(mask$to(dtype = img$dtype), a4))
  else
    torch::torch_einsum("ibjk, xibjk -> bxjk", list(mask$to(dtype = img$dtype), a4))
}

# https://stackoverflow.com/questions/32370485/convert-radians-to-degree-degree-to-radians
rad2deg <- function(rad) {(rad * 180) / (pi)}
deg2rad <- function(deg) {(deg * pi) / (180)}

# Helper method to compute inverse matrix for affine transformation
# As it is explained in PIL.Image.rotate
# We need compute INVERSE of affine transformation matrix: M = T * C * RSS * C^-1
# where T is translation matrix: [1, 0, tx | 0, 1, ty | 0, 0, 1]
#       C is translation matrix to keep center: [1, 0, cx | 0, 1, cy | 0, 0, 1]
#       RSS is rotation with scale and shear matrix
#       RSS(a, s, (sx, sy)) =
#       = R(a) * S(s) * SHy(sy) * SHx(sx)
#       = [ s*cos(a - sy)/cos(sy), s*(-cos(a - sy)*tan(x)/cos(y) - sin(a)), 0 ]
#         [ s*sin(a + sy)/cos(sy), s*(-sin(a - sy)*tan(x)/cos(y) + cos(a)), 0 ]
#         [ 0                    , 0                                      , 1 ]
#
# where R is a rotation matrix, S is a scaling matrix, and SHx and SHy are the shears:
# SHx(s) = [1, -tan(s)] and SHy(s) = [1      , 0]
#          [0, 1      ]              [-tan(s), 1]
#
# Thus, the inverse is M^-1 = C * RSS^-1 * C^-1 * T^-1
get_inverse_affine_matrix <- function(center, angle, translate, scale, shear) {

  rot <- deg2rad(angle)
  sx <- deg2rad(shear[1])
  sy <- deg2rad(shear[2])

  cx <- center[1]
  cy <- center[2]

  tx <- translate[1]
  ty <- translate[2]

  # RSS without scaling
  a <- cos(rot - sy) / cos(sy)
  b <- -cos(rot - sy) * tan(sx) / cos(sy) - sin(rot)
  c <- sin(rot - sy) / cos(sy)
  d <- -sin(rot - sy) * tan(sx) / cos(sy) + cos(rot)

  # Inverted rotation matrix with scale and shear
  # det([[a, b], [c, d]]) == 1, since det(rotation) = 1 and det(shear) = 1
  matrix <- c(d, -b, 0.0, -c, a, 0.0) / scale

  # Apply inverse of translation and of center translation: RSS^-1 * C^-1 * T^-1
  matrix[3] = matrix[3] + matrix[1] * (-cx - tx) + matrix[2] * (-cy - ty)
  matrix[6] = matrix[6] + matrix[4] * (-cx - tx) + matrix[5] * (-cy - ty)

  # Apply center translation: C * RSS^-1 * C^-1 * T^-1
  matrix[3] = matrix[3] + cx
  matrix[6] = matrix[6] + cy

  matrix
}

assert_grid_transform_inputs <- function(img, matrix, interpolation, fill,
                                         interpolation_modes, coeffs) {
  check_img(img)
}

apply_grid_transform <- function(img, grid, mode) {

  need_squeeze <- FALSE
  if (img$dim() < 4) {
    img <- img$unsqueeze(dim=1)
    need_squeeze <- TRUE
  }

  out_dtype <- img$dtype
  need_cast <- FALSE
  if (!img$dtype == torch::torch_float32() && !img$dtype == torch::torch_float64()) {
    need_cast <- TRUE
    img <- img$to(dtype = torch::torch_float32())
  }

  img <- torch::nnf_grid_sample(img, grid, mode=mode, padding_mode="zeros",
                                align_corners=FALSE)

  if (need_squeeze)
    img <- img$squeeze(dim=1)

  if (need_cast) {
    # it is better to round before cast
    img <- torch::torch_round(img)$to(dtype = out_dtype)
  }

  img
}

# https://github.com/pytorch/pytorch/blob/74b65c32be68b15dc7c9e8bb62459efbfbde33d8/aten/src/ATen/native/
# AffineGridGenerator.cpp#L18
# Difference with AffineGridGenerator is that:
# 1) we normalize grid values after applying theta
# 2) we can normalize by other image size, such that it covers "extend" option like in PIL.Image.rotate
gen_affine_grid <- function(theta, w, h, ow, oh) {

  d <- 0.5
  base_grid <- torch::torch_empty(1, oh, ow, 3)
  base_grid[.., 1]$copy_(torch::torch_linspace(start = -ow * 0.5 + d, end = ow * 0.5 + d - 1,
                                               steps=ow))

  base_grid[.., 2]$copy_(torch::torch_linspace(start = -oh * 0.5 + d, end = oh * 0.5 + d - 1,
                                               steps=oh)$unsqueeze_(-1))
  base_grid[.., 3]$fill_(1)

  output_grid <- base_grid$view(c(1, oh * ow, 3))$bmm(
    theta$transpose(2, 3) / torch::torch_tensor(c(0.5 * w, 0.5 * h))
  )

  output_grid$view(c(1, oh, ow, 2))
}

rotate_compute_output_size <- function(theta, w, h) {
  # Inspired of PIL implementation:
  # https://github.com/python-pillow/Pillow/blob/11de3318867e4398057373ee9f12dcb33db7335c/src/PIL/Image.py#L2054

  # pts are Top-Left, Top-Right, Bottom-Left, Bottom-Right points.
  pts <- torch::torch_tensor(rbind(
    c(-0.5 * w, -0.5 * h, 1.0),
    c(-0.5 * w, 0.5 * h, 1.0),
    c(0.5 * w, 0.5 * h, 1.0),
    c(0.5 * w, -0.5 * h, 1.0)
  ))
  new_pts <- pts$view(c(1, 4, 3))$bmm(theta$transpose(2, 3))$view(c(4, 2))
  min_vals <- new_pts$min(dim=1)[[1]]
  max_vals <- new_pts$max(dim=1)[[1]]

  # Truncate precision to 1e-4 to avoid ceil of Xe-15 to 1.0
  tol <- 1e-4
  cmax <- torch::torch_ceil((max_vals / tol)$trunc_() * tol)
  cmin <- torch::torch_floor((min_vals / tol)$trunc_() * tol)
  size <- cmax - cmin

  as.integer(c(size[1]$item(), size[2]$item()))
}

rotate_impl <- function(img, matrix, interpolation = 0, expand = FALSE, fill = NULL) {

  interpolation_modes <- c(
    "0" = "nearest",
    "2" = "bilinear",
    "nearest" = "nearest",
    "nearest-exact" = "nearest",
    "bilinear" = "bilinear"
  )

  assert_grid_transform_inputs(img, matrix, interpolation, fill, interpolation_modes)
  theta <- torch::torch_tensor(matrix)$reshape(c(1, 2, 3))
  w <- tail(img$shape, 2)[1]
  h <- tail(img$shape, 2)[2]

  if (expand) {
    o_shape <- rotate_compute_output_size(theta, w, h)
    ow <- o_shape[1]
    oh <- o_shape[2]
  } else {
    ow <- w
    oh <- h
  }

  grid <- gen_affine_grid(theta, w=w, h=h, ow=ow, oh=oh)
  mode <- normalize_interpolation_mode(interpolation, interpolation_modes)

  apply_grid_transform(img, grid, mode)
}

affine_impl <- function(img, matrix, interpolation = 0, fill = NULL) {

  interpolation_modes <- c(
    "0" = "nearest",
    "2" = "bilinear",
    "nearest" = "nearest",
    "nearest-exact" = "nearest",
    "bilinear" = "bilinear"
  )

  assert_grid_transform_inputs(img, matrix, interpolation, fill, interpolation_modes)

  theta <- torch::torch_tensor(matrix, dtype=torch::torch_float())$reshape(c(1, 2, 3))
  shape <- img$shape
  grid <- gen_affine_grid(theta, w=rev(shape)[1], h=rev(shape)[2],
                         ow=rev(shape)[1], oh=rev(shape)[2])
  mode <- normalize_interpolation_mode(interpolation, interpolation_modes)
  apply_grid_transform(img, grid, mode)
}

normalize_interpolation_mode <- function(interpolation, interpolation_modes) {
  if (is.numeric(interpolation) && length(interpolation) == 1) {
    key <- as.character(as.integer(interpolation))
  } else if (is.character(interpolation) && length(interpolation) == 1) {
    key <- tolower(interpolation)
    key <- sub("^interpolationmode\\.", "", key)
    key <- gsub("_", "-", key, fixed = TRUE)
  } else {
    value_error("`interpolation` should be an integer code or a single character mode")
  }

  if (!key %in% names(interpolation_modes)) {
    value_error("Unsupported interpolation mode")
  }

  interpolation_modes[[key]]
}

pad_symmetric <- function(img, padding) {

  in_sizes <- img$size()

  x_indices <- seq_len(tail(in_sizes, 1))
  left_indices <- rev(seq_len(padding[1]))
  right_indices <- -seq_len(padding[2])
  x_indices <- torch::torch_tensor(c(left_indices, x_indices, right_indices),
                                   dtype = torch::torch_long())


  y_indices <- seq_len(rev(in_sizes)[2])
  top_indices <- rev(seq_len(padding[3]))
  bottom_indices <- -seq_len(padding[4])
  y_indices <- torch::torch_tensor(c(top_indices, y_indices, bottom_indices),
                                   dtype = torch::torch_long())


  ndim <- length(dim(img))
  if (ndim == 3)
    img[, y_indices[, NULL], x_indices[, NULL]]
  else if (ndim == 4)
    img[,,y_indices[, NULL], x_indices[, NULL]]
  else
    runtime_error("Symmetric padding of N-D tensors are not supported yet")
}

#' @export
transform_adjust_gamma.torch_tensor <- function(img, gamma, gain = 1) {

  check_img(img)

  if (gamma < 0)
    value_error("gamma must be non-negative")

  result <- img
  dtype <- img$dtype

  if (!torch::torch_is_floating_point(img))
    result <- result/255.0

  result <- (gain * result ^ gamma)$clamp(0, 1)

  if (!result$dtype == dtype) {
    eps <- 1e-3
    result <- (255 + 1.0 - eps) * result
  }

  result <- result$to(dtype = dtype)
  result
}

get_perspective_coeffs <- function(startpoints, endpoints) {

  a_matrix <- torch::torch_zeros(2 * length(startpoints), 8,
                                 dtype = torch::torch_float())

  for (i in seq_along(startpoints)) {

    p1 <- endpoints[[i]]
    p2 <- startpoints[[i]]

    a_matrix[1 + 2*(i-1), ] <- torch::torch_tensor(c(p1[1], p1[2], 2, 1, 1, 1, -p2[1] * p1[1], -p2[1] * p1[2]))
    a_matrix[2*i,] <- torch::torch_tensor(c(1, 1, 1, p1[1], p1[2], 2, -p2[2] * p1[1], -p2[2] * p1[1]))

  }

  b_matrix <- torch::torch_tensor(unlist(startpoints), dtype = torch::torch_float())$view(8)
  res <- torch::linalg_lstsq(b_matrix, a_matrix)[[1]]

  output <- torch::as_array(res$squeeze(1))
  output
}

#' Convert an image to a tensor
#'
#' Converts a Magick Image or array (H x W x C) in the range `[0, 255]` to a
#'   `torch_tensor` of shape (C x H x W) in the range `[0.0, 1.0]`. In the
#'   other cases, tensors are returned without scaling.
#'
#' @note
#' Because the input image is scaled to `[0.0, 1.0]`, this transformation
#'   should not be used when transforming target image masks.
#'
#' @param img A `magick-image`, `array` or `torch_tensor`.
#'
#' @family unitary_transforms
#'
#' @export
transform_to_tensor <- function(img) {
  UseMethod("transform_to_tensor", img)
}

#' Convert a tensor image to the given `dtype` and scale the values accordingly
#'
#' @inheritParams transform_to_tensor
#' @param dtype (torch.dtype): Desired data type of the output.
#'
#' @note
#' When converting from a smaller to a larger integer `dtype` the maximum
#'   values are **not** mapped exactly. If converted back and forth, this
#'   mismatch has no effect.
#'
#' @family unitary_transforms
#'
#' @export
transform_convert_image_dtype <- function(img, dtype = torch::torch_float()) {
  UseMethod("transform_convert_image_dtype", img)
}

#' Normalize a tensor image with mean and standard deviation
#'
#' Given mean: `(mean[1],...,mean[n])` and std: `(std[1],..,std[n])` for `n`
#'   channels, this transform will normalize each channel of the input
#'   `torch_tensor` i.e.,
#'   `output[channel] = (input[channel] - mean[channel]) / std[channel]`
#'
#' @note
#' This transform acts out of place, i.e., it does not mutate the input tensor.
#'
#' @inheritParams transform_to_tensor
#' @param mean (sequence): Sequence of means for each channel.
#' @param std (sequence): Sequence of standard deviations for each channel.
#' @param inplace (bool,optional): Bool to make this operation in-place.
#'
#' @family unitary_transforms
#'
#' @export
transform_normalize <- function(img, mean, std, inplace = FALSE) {
  UseMethod("transform_normalize", img)
}

#' Resize the input image to the given size
#'
#' The image can be a Magic Image or a torch Tensor, in which case it is
#'   expected to have `[..., H, W]` shape, where ... means an arbitrary number
#'   of leading dimensions
#'
#' @inheritParams transform_to_tensor
#' @param size (sequence or int): Desired output size. If size is a sequence
#'   like c(h, w), output size will be matched to this. If size is an int,
#'   smaller edge of the image will be matched to this number.
#'   i.e, if height > width, then image will be rescaled to
#'   (size * height / width, size).
#' @param interpolation (int, optional) Desired interpolation. An integer
#'   `0 = nearest`, `2 = bilinear`, and `3 = bicubic` or a name from
#'   [magick::filter_types()].
#'
#' @family unitary_transforms
#'
#' @export
transform_resize <- function(img, size, interpolation = 2) {
  UseMethod("transform_resize", img)
}

#' Crops the given image at the center
#'
#' The image can be a Magick Image or a torch Tensor, in which case it is
#'   expected to have `[..., H, W]` shape, where ... means an arbitrary number
#'   of leading dimensions.
#'
#' @inheritParams transform_to_tensor
#' @param size (sequence or int): Desired output size of the crop. If size is
#'   an int instead of sequence like c(h, w), a square crop (size, size) is
#'   made. If provided a tuple or list of length 1, it will be interpreted as
#'   `c(size, size)`.
#'
#' @family unitary_transforms
#'
#' @export
transform_center_crop <- function(img, size) {
  UseMethod("transform_center_crop", img)
}

#' Pad the given image on all sides with the given "pad" value
#'
#' The image can be a Magick Image or a torch Tensor, in which case it is
#'   expected to have `[..., H, W]` shape, where ... means an arbitrary number
#'   of leading dimensions.
#'
#' @inheritParams transform_to_tensor
#' @param padding (int or tuple or list): Padding on each border. If a single
#'   int is provided this is used to pad all borders. If tuple of length 2 is
#'   provided this is the padding on left/right and top/bottom respectively.
#'   If a tuple of length 4 is provided this is the padding for the left, right,
#'   top and bottom borders respectively.
#' @param fill (int or str or tuple): Pixel fill value for constant fill.
#'   Default is 0. If a tuple of length 3, it is used to fill R, G, B channels
#'   respectively. This value is only used when the padding_mode is constant.
#'   Only int value is supported for Tensors.
#' @param padding_mode Type of padding. Should be: constant, edge, reflect or
#'   symmetric. Default is constant. Mode symmetric is not yet supported for
#'   Tensor inputs.
#'
#'  - constant: pads with a constant value, this value is specified with fill
#'
#'  - edge: pads with the last value on the edge of the image
#'
#'  - reflect: pads with reflection of image (without repeating the last
#'    value on the edge) padding `[1, 2, 3, 4]` with 2 elements on both sides
#'    in reflect mode will result in `[3, 2, 1, 2, 3, 4, 3, 2]`
#'
#'  - symmetric: pads with reflection of image (repeating the last value on
#'    the edge) padding `[1, 2, 3, 4]` with 2 elements on both sides in
#'    symmetric mode will result in `[2, 1, 1, 2, 3, 4, 4, 3]`
#'
#' @family unitary_transforms
#'
#' @export
transform_pad <- function(img, padding, fill = 0, padding_mode = "constant") {
  UseMethod("transform_pad", img)
}

#' Apply a list of transformations randomly with a given probability
#'
#' @inheritParams transform_to_tensor
#' @param transforms (list or tuple): list of transformations.
#' @param p (float): probability.
#'
#' @family combining_transforms
#'
#' @export
transform_random_apply <- function(img, transforms, p = 0.5) {
  UseMethod("transform_random_apply", img)
}

#' Apply single transformation randomly picked from a list
#'
#' @inheritParams transform_random_apply
#'
#' @family combining_transforms
#'
#' @export
transform_random_choice <- function(img, transforms) {
  UseMethod("transform_random_choice", img)
}

#' Apply a list of transformations in a random order
#'
#' @inheritParams transform_random_apply
#' @family combining_transforms
#' @export
transform_random_order <- function(img, transforms) {
  UseMethod("transform_random_order", img)
}

#' Crop the given image at a random location
#'
#' The image can be a Magick Image or a Tensor, in which case it is expected
#'   to have `[..., H, W]` shape, where ... means an arbitrary number of leading
#'   dimensions.
#'
#' @inheritParams transform_resize
#' @inheritParams transform_pad
#' @param pad_if_needed (boolean): It will pad the image if smaller than the
#'   desired size to avoid raising an exception. Since cropping is done
#'   after padding, the padding seems to be done at a random offset.
#'
#' @family random_transforms
#'
#' @export
transform_random_crop <- function(img, size, padding=NULL, pad_if_needed=FALSE,
                                  fill=0, padding_mode="constant") {
  UseMethod("transform_random_crop", img)
}

#' Horizontally flip an image randomly with a given probability
#'
#' Horizontally flip an image randomly with a given probability. The image can
#'   be a Magick Image or a torch Tensor, in which case it is expected to have
#'   `[..., H, W]` shape, where ... means an arbitrary number of leading
#'   dimensions
#'
#' @inheritParams transform_to_tensor
#' @param p (float): probability of the image being flipped.
#'   Default value is 0.5
#'
#' @family random_transforms
#' @export
transform_random_horizontal_flip <- function(img, p = 0.5) {
  UseMethod("transform_random_horizontal_flip", img)
}

#' Vertically flip an image randomly with a given probability
#'
#' The image can be a PIL Image or a torch Tensor, in which case it is expected
#'   to have `[..., H, W]` shape, where `...` means an arbitrary number of
#'   leading dimensions
#'
#' @inheritParams transform_random_horizontal_flip
#'
#' @family random_transforms
#' @export
transform_random_vertical_flip <- function(img, p = 0.5) {
  UseMethod("transform_random_vertical_flip", img)
}

#' Crop image to random size and aspect ratio
#'
#' Crop the given image to a random size and aspect ratio. The image can be a
#'   Magick Image or a Tensor, in which case it is expected to have
#'   `[..., H, W]` shape, where ... means an arbitrary number of leading
#'   dimensions
#'
#' @details A crop of random size (default: of 0.08 to 1.0) of the original size
#'   and a random aspect ratio (default: of 3/4 to 4/3) of the original aspect
#'   ratio is made. This crop is finally resized to given size. This is
#'   popularly used to train the Inception networks.
#'
#' @inheritParams transform_resize
#' @inheritParams transform_pad
#' @param scale (tuple of float): range of size of the origin size cropped
#' @param ratio (tuple of float): range of aspect ratio of the origin aspect
#'   ratio cropped.
#'
#' @family random_transforms
#' @export
transform_random_resized_crop <- function(img, size, scale=c(0.08, 1.0),
                                          ratio=c(3. / 4., 4. / 3.),
                                          interpolation=2) {
  UseMethod("transform_random_resized_crop", img)
}

#' Crop image into four corners and a central crop
#'
#' Crop the given image into four corners and the central crop. This transform
#'   returns a tuple of images and there may be a mismatch in the number of
#'   inputs and targets your Dataset returns.
#'
#' @inheritParams transform_resize
#'
#' @family combining_transforms
#' @export
transform_five_crop <- function(img, size) {
  UseMethod("transform_five_crop", img)
}

#' Crop an image and the flipped image each into four corners and a central crop
#'
#' Crop the given image into four corners and the central crop, plus the flipped
#'   version of these (horizontal flipping is used by default). This transform
#'   returns a tuple of images and there may be a mismatch in the number of
#'   inputs and targets your Dataset returns.
#'
#' @inheritParams transform_five_crop
#' @param vertical_flip (bool): Use vertical flipping instead of horizontal
#'
#' @family combining_transforms
#' @export
transform_ten_crop <- function(img, size, vertical_flip = FALSE) {
  UseMethod("transform_ten_crop", img)
}

#' Transform a tensor image with a square transformation matrix and a
#'   mean_vector computed offline
#'
#' Given `transformation_matrix` and `mean_vector`, will flatten the
#'   `torch_tensor` and subtract `mean_vector` from it which is then followed by
#'   computing the dot product with the transformation matrix and then reshaping
#'   the tensor to its original shape.
#'
#' @section Applications:
#'   whitening transformation: Suppose X is a column vector zero-centered data.
#'   Then compute the data covariance matrix `[D x D]` with torch.mm(X.t(), X),
#'   perform SVD on this matrix and pass it as `transformation_matrix`.
#'
#' @inheritParams transform_to_tensor
#' @param transformation_matrix (Tensor): tensor `[D x D]`, D = C x H x W.
#' @param mean_vector (Tensor): tensor [D], D = C x H x W.
#'
#' @family unitary_transforms
#' @export
transform_linear_transformation <- function(img, transformation_matrix, mean_vector) {
  UseMethod("transform_linear_transformation", img)
}

#' Randomly change the brightness, contrast and saturation of an image
#'
#' @param brightness (float or tuple of float (min, max)): How much to jitter
#'   brightness. `brightness_factor` is chosen uniformly from
#'   `[max(0, 1 - brightness), 1 + brightness]` or the given `[min, max]`.
#'   Should be non negative numbers.
#' @param contrast (float or tuple of float (min, max)): How much to jitter
#'   contrast. `contrast_factor` is chosen uniformly from
#'   `[max(0, 1 - contrast), 1 + contrast]` or the given `[min, max]`. Should
#'   be non negative numbers.
#' @param saturation (float or tuple of float (min, max)): How much to jitter
#'   saturation. `saturation_factor` is chosen uniformly from
#'   `[max(0, 1 - saturation), 1 + saturation]` or the given `[min, max]`.
#'   Should be non negative numbers.
#' @param hue (float or tuple of float (min, max)): How much to jitter hue.
#'   `hue_factor` is chosen uniformly from `[-hue, hue]` or the given
#'   `[min, max]`. Should have 0<= hue <= 0.5 or -0.5 <= min <= max <= 0.5.
#' @inheritParams transform_to_tensor
#'
#' @family random_transforms
#' @export
transform_color_jitter <- function(img, brightness=0, contrast=0, saturation=0, hue=0) {
  UseMethod("transform_color_jitter", img)
}

#' Rotate the image by angle
#'
#' @param degrees (sequence or float or int): Range of degrees to select from.
#'   If degrees is a number instead of sequence like c(min, max), the range of
#'   degrees will be (-degrees, +degrees).
#' @param interpolation (int, optional): Interpolation mode. 0 for nearest, 2 for bilinear.
#'   Default is 0 (nearest).
#' @param resample Deprecated. Use interpolation instead.
#' @param expand (bool, optional): Optional expansion flag. If true, expands the
#'   output to make it large enough to hold the entire rotated image. If false
#'   or omitted, make the output image the same size as the input image. Note
#'   that the expand flag assumes rotation around the center and no translation.
#' @param center (list or tuple, optional): Optional center of rotation, c(x, y).
#'   Origin is the upper left corner. Default is the center of the image.
#' @param fill (n-tuple or int or float): Pixel fill value for area outside the
#'   rotated image. If int or float, the value is used for all bands
#'   respectively. Defaults to 0 for all bands. This option is only available
#'   for Pillow>=5.2.0. This option is not supported for Tensor input. Fill
#'   value for the area outside the transform in the output image is always 0.
#' @inheritParams transform_to_tensor
#'
#' @family random_transforms
#' @export
transform_random_rotation <- function(img, degrees, interpolation = 0, expand=FALSE,
                                      center=NULL, fill=NULL, resample) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }
  UseMethod("transform_random_rotation", img)
}

#' Random affine transformation of the image keeping center invariant
#'
#' @inheritParams transform_random_rotation
#' @param translate (tuple, optional): tuple of maximum absolute fraction for
#'   horizontal and vertical translations. For example `translate=c(a, b)`, then
#'   horizontal shift is randomly sampled in the range
#'   -img_width * a < dx < img_width * a and vertical shift is randomly sampled
#'   in the range -img_height * b < dy < img_height * b. Will not translate by
#'   default.
#' @param scale (tuple, optional): scaling factor interval, e.g c(a, b), then
#'   scale is randomly sampled from the range a <= scale <= b. Will keep
#'   original scale by default.
#' @param shear (sequence or float or int, optional): Range of degrees to select
#'   from. If shear is a number, a shear parallel to the x axis in the range
#'   (-shear, +shear) will be applied. Else if shear is a tuple or list of 2
#'   values a shear parallel to the x axis in the range `(shear[1], shear[2])`
#'   will be applied. Else if shear is a tuple or list of 4 values, a x-axis
#'   shear in `(shear[1], shear[2])` and y-axis shear in `(shear[3], shear[4])`
#'   will be applied. Will not apply shear by default.
#' @param fill (tuple or int): Fill color for the area outside the transform.
#'   Default is 0. This option is not supported for Tensor input.
#' @param interpolation (int or character, optional): Interpolation mode.
#'   Supported values are 0 / "nearest" and 2 / "bilinear". Default is 0.
#' @param fillcolor Deprecated. Use fill instead.
#' @param resample Deprecated. Use interpolation instead.
#'
#' @family random_transforms
#' @export
transform_random_affine <- function(img, degrees, translate=NULL, scale=NULL,
                                    shear=NULL, interpolation=0, fill=0,
                                    resample, fillcolor) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }
  if (!missing(fillcolor)) {
    deprecated("'fillcolor' is deprecated, use 'fill' instead")
    fill <- fillcolor
  }
  UseMethod("transform_random_affine", img)
}

#' Convert image to grayscale
#'
#' @inheritParams transform_to_tensor
#' @param num_output_channels (int): (1 or 3) number of channels desired for
#'   output image
#'
#' @family unitary_transforms
#' @export
transform_grayscale <- function(img, num_output_channels) {
  UseMethod("transform_grayscale", img)
}

#' Randomly convert image to grayscale with a given probability
#'
#' Convert image to grayscale with a probability of `p`.
#'
#' @inheritParams transform_to_tensor
#' @param p (float): probability that image should be converted to grayscale
#'   (default 0.1).
#'
#' @family random_transforms
#' @export
transform_random_grayscale <- function(img, p = 0.1) {
  UseMethod("transform_random_grayscale", img)
}


#' Random perspective transformation of an image with a given probability
#'
#' Performs a random perspective transformation of the given image with a given
#'   probability
#'
#' @param distortion_scale (float): argument to control the degree of distortion
#'   and ranges from 0 to 1. Default is 0.5.
#' @param p (float): probability of the image being transformed. Default is 0.5.
#' @inheritParams transform_resize
#' @inheritParams transform_pad
#'
#' @family random_transforms
#' @export
transform_random_perspective <- function(img, distortion_scale=0.5, p=0.5,
                                         interpolation=2, fill=0) {
  UseMethod("transform_random_perspective", img)
}

#' Randomly selects a rectangular region in an image and erases its pixel values
#'
#' 'Random Erasing Data Augmentation' by Zhong _et al._
#'   See <https://arxiv.org/pdf/1708.04896>
#'
#' @inheritParams transform_to_tensor
#' @param p probability that the random erasing operation will be performed.
#' @param scale range of proportion of erased area against input image.
#' @param ratio range of aspect ratio of erased area.
#' @param value erasing value. Default is 0. If a single int, it is used to
#'   erase all pixels. If a tuple of length 3, it is used to erase
#'   R, G, B channels respectively.
#'   If a str of 'random', erasing each pixel with random values.
#' @param inplace boolean to make this transform inplace. Default set to FALSE.
#'
#' @family random_transforms
#' @export
transform_random_erasing <- function(img, p=0.5, scale=c(0.02, 0.33), ratio=c(0.3, 3.3),
                                     value=0, inplace=FALSE) {
  UseMethod("transform_random_erasing", img)
}

# Other generics ----------------------------------------------------------

#' Crop the given image at specified location and output size
#'
#' @inheritParams transform_to_tensor
#' @param top (int): Vertical component of the top left corner of the crop box.
#' @param left (int): Horizontal component of the top left corner of the crop
#'   box.
#' @param height (int): Height of the crop box.
#' @param width (int): Width of the crop box.
#'
#' @family unitary_transforms
#' @export
transform_crop <- function(img, top, left, height, width) {
  UseMethod("transform_crop", img)
}

#' Horizontally flip a PIL Image or Tensor
#'
#' @inheritParams transform_to_tensor
#'
#' @family unitary_transforms
#' @export
transform_hflip <- function(img) {
  UseMethod("transform_hflip", img)
}

#' Vertically flip a PIL Image or Tensor
#'
#' @inheritParams transform_to_tensor
#'
#' @family unitary_transforms
#' @export
transform_vflip <- function(img) {
  UseMethod("transform_vflip", img)
}

#' Crop an image and resize it to a desired size
#'
#' @param top (int): Vertical component of the top left corner of the crop box.
#' @param left (int): Horizontal component of the top left corner of the crop
#'   box.
#' @param height (int): Height of the crop box.
#' @param width (int): Width of the crop box.
#' @inheritParams transform_resize
#'
#' @family combining_transforms
#' @export
transform_resized_crop <- function(img, top, left, height, width, size,
                                   interpolation = 2) {
  UseMethod("transform_resized_crop", img)
}


#' Adjust the brightness of an image
#'
#' @param brightness_factor (float):  How much to adjust the brightness. Can be
#'   any non negative number. 0 gives a black image, 1 gives the
#'   original image while 2 increases the brightness by a factor of 2.
#' @inheritParams transform_resize
#'
#' @family unitary_transforms
#' @export
transform_adjust_brightness <- function(img, brightness_factor) {
  UseMethod("transform_adjust_brightness", img)
}

#' Adjust the contrast of an image
#'
#' @param contrast_factor (float): How much to adjust the contrast. Can be any
#'   non negative number. 0 gives a solid gray image, 1 gives the
#'   original image while 2 increases the contrast by a factor of 2.
#'
#' @inheritParams transform_resize
#'
#' @family unitary_transforms
#' @export
transform_adjust_contrast <- function(img, contrast_factor) {
  UseMethod("transform_adjust_contrast", img)
}

#' Adjust the color saturation of an image
#'
#' @param saturation_factor (float):  How much to adjust the saturation. 0 will
#'   give a black and white image, 1 will give the original image while
#'   2 will enhance the saturation by a factor of 2.
#'
#' @inheritParams transform_resize
#'
#' @family unitary_transforms
#' @export
transform_adjust_saturation <- function(img, saturation_factor) {
  UseMethod("transform_adjust_saturation", img)
}

#' Adjust the hue of an image
#'
#' The image hue is adjusted by converting the image to HSV and
#'   cyclically shifting the intensities in the hue channel (H).
#'   The image is then converted back to original image mode.
#'
#' `hue_factor` is the amount of shift in H channel and must be in the
#'   interval `[-0.5, 0.5]`.
#'
#' Search for Hue for more details.
#'
#' @param hue_factor (float):  How much to shift the hue channel. Should be in
#'   `[-0.5, 0.5]`. 0.5 and -0.5 give complete reversal of hue channel in
#'   HSV space in positive and negative direction respectively.
#'   0 means no shift. Therefore, both -0.5 and 0.5 will give an image
#'   with complementary colors while 0 gives the original image.
#'
#' @inheritParams transform_resize
#'
#' @family unitary_transforms
#' @export
transform_adjust_hue <- function(img, hue_factor) {
  UseMethod("transform_adjust_hue", img)
}

#' Angular rotation of an image
#'
#' @inheritParams transform_to_tensor
#' @inheritParams transform_random_rotation
#' @param angle (float or int): rotation angle value in degrees,
#'   counter-clockwise.
#'
#' @family unitary_transforms
#' @export
transform_rotate <- function(img, angle, interpolation = 0, expand = FALSE,
                             center = NULL, fill = NULL, resample) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }
  UseMethod("transform_rotate", img)
}

#' Apply affine transformation on an image keeping image center invariant
#'
#' @inheritParams transform_rotate
#' @inheritParams transform_random_affine
#' @param translate (sequence of int) – horizontal and vertical translations
#'  (post-rotation translation)
#' @param scale (float) – overall scale
#' @param shear (float or sequence) – shear angle value in degrees between -180 to 180,
#'  clockwise direction. If a sequence is specified, the first value corresponds
#'  to a shear parallel to the x-axis, while the second value corresponds to a
#'  shear parallel to the y-axis.
#' @param interpolation (int or character): Interpolation mode.
#'   Supported values are 0 / "nearest" and 2 / "bilinear". Default is 0.
#' @param fill Fill color for area outside the transform. Default is NULL.
#' @param resample Deprecated. Use interpolation instead.
#' @param fillcolor Deprecated. Use fill instead.
#' @param center Optional center of rotation, c(x, y). Default is image center.
#'
#' @family unitary_transforms
#' @export
transform_affine <- function(img, angle, translate, scale, shear,
                             interpolation = 0, fill = NULL,
                             resample, fillcolor, center = NULL) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }
  if (!missing(fillcolor)) {
    deprecated("'fillcolor' is deprecated, use 'fill' instead")
    fill <- fillcolor
  }
  UseMethod("transform_affine", img)
}

#' Perspective transformation of an image
#'
#' @param startpoints (list of list of ints): List containing four lists of two
#'   integers corresponding to four corners
#'   `[top-left, top-right, bottom-right, bottom-left]` of the original image.
#' @param endpoints (list of list of ints): List containing four lists of two
#'   integers corresponding to four corners
#'   `[top-left, top-right, bottom-right, bottom-left]` of the transformed
#'   image.
#' @inheritParams transform_resize
#' @inheritParams transform_pad
#'
#' @family unitary_transforms
#' @export
transform_perspective <- function(img, startpoints, endpoints, interpolation = 2,
                                  fill = NULL) {
  UseMethod("transform_perspective", img)
}

#' Adjust the gamma of an RGB image
#'
#' Also known as Power Law Transform. Intensities in RGB mode are adjusted
#'   based on the following equation:
#'   \deqn{
#'     I_{\mbox{out}} = 255 \times \mbox{gain} \times \left
#'     (\frac{I_{\mbox{in}}}{255}\right)^{\gamma}
#'     }
#'
#' @details Search for Gamma Correction for more details.
#'
#' @param gamma (float): Non negative real number, same as \eqn{\gamma} in the
#'   equation. gamma larger than 1 make the shadows darker, while gamma smaller
#'   than 1 make dark regions lighter.
#' @param gain (float): The constant multiplier.
#' @inheritParams transform_to_tensor
#'
#' @family unitary_transforms
#' @export
transform_adjust_gamma <- function(img, gamma, gain = 1) {
  UseMethod("transform_adjust_gamma", img)
}

#' Convert RGB Image Tensor to Grayscale
#'
#' For RGB to Grayscale conversion, ITU-R 601-2 luma transform is performed
#'   which is L = R * 0.2989 + G * 0.5870 + B * 0.1140
#'
#' @inheritParams transform_to_tensor
#'
#' @family unitary_transforms
#' @export
transform_rgb_to_grayscale <- function(img) {
  UseMethod("transform_rgb_to_grayscale")
}


# Utils -------------------------------------------------------------------

get_image_size <- function(img) {
  UseMethod("get_image_size")
}

test_that("model_fcn_resnet50 returns correct output shape", {
  model <- model_fcn_resnet50(pretrained = FALSE, num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_named(output, c("out"))
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_fcn_resnet50 works with custom num_classes", {
  model <- model_fcn_resnet50(pretrained = FALSE, num_classes = 3)
  model$eval()

  input <- torch::torch_randn(2, 3, 224, 224)
  output <- model(input)

  expect_tensor_shape(output$out, c(2, 3, 224, 224))
})

test_that("model_fcn_resnet50 with aux classifier returns aux output", {
  model <- model_fcn_resnet50(pretrained = FALSE, num_classes = 21, aux_loss = TRUE)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_named(output, c("out", "aux"))
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
  expect_tensor_shape(output$aux, c(1, 21, 224, 224))
})

test_that("model_fcn_resnet50 loads pretrained weights with 21 classes", {
  model <- model_fcn_resnet50(pretrained = TRUE, num_classes = 21)
  expect_s3_class(model, "fcn")
  model$eval()

  input <- torch::torch_randn(2, 3, 224, 224)
  output <- model(input)

  expect_named(output, c("out", "aux"))
  expect_tensor_shape(output$out, c(2, 21, 224, 224))
  expect_tensor_shape(output$aux, c(2, 21, 224, 224))
})

test_that("model_fcn_resnet50 can segment a cat", {
  voc_idx <- setNames(seq_along(pascal_voc_classes()), pascal_voc_classes())

  model <- model_fcn_resnet50(pretrained = TRUE)
  img <- torch::torch_tensor(jpeg::readJPEG("assets/class/cat/cat.1.jpg"))$permute(c(3,1,2))
  norm_mean <- c(0.485, 0.456, 0.406) #ImageNet normalization constants
  norm_std  <- c(0.229, 0.224, 0.225)
  input <- img %>% transform_resize(c(520, 520)) %>% transform_normalize(norm_mean, norm_std)

  output <- model(input$unsqueeze(1))
  mask_id <- output$out$argmax(dim = 2)
  mask_table <- factor(mask_id %>% torch::as_array(), levels = voc_idx, labels = names(voc_idx)) %>% table()

  expect_gt(mask_table[["cat"]], 0)
  expect_gt(mask_table[["cat"]], mask_table[["dog"]])
  expect_gt(mask_table[["cat"]], mask_table[["horse"]])

  expect_gt(mask_table[["background"]], 0)

})


test_that("model_fcn_resnet101 returns correct output shape", {
  model <- model_fcn_resnet101(pretrained = FALSE, num_classes = 21)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_named(output, c("out"))
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
})

test_that("model_fcn_resnet101 works with custom num_classes", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")
  model <- model_fcn_resnet101(pretrained = FALSE, num_classes = 3)
  model$eval()

  input <- torch::torch_randn(2, 3, 224, 224)
  output <- model(input)

  expect_tensor_shape(output$out, c(2, 3, 224, 224))
})

test_that("model_fcn_resnet101 with aux classifier returns aux output", {
  model <- model_fcn_resnet101(pretrained = FALSE, num_classes = 21, aux_loss = TRUE)
  model$eval()

  input <- torch::torch_randn(1, 3, 224, 224)
  output <- model(input)

  expect_named(output, c("out", "aux"))
  expect_tensor_shape(output$out, c(1, 21, 224, 224))
  expect_tensor_shape(output$aux, c(1, 21, 224, 224))
})

test_that("model_fcn_resnet101 loads pretrained weights", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")
  model <- model_fcn_resnet101(pretrained = TRUE)
  expect_s3_class(model, "fcn")
})

context("dataset-flowers")

t <- withr::local_tempdir()

test_that("tests for the Flowers102 dataset for train split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_error(
    flowers102_dataset(root = tempfile(), download = FALSE),
    class = "rlang_error"
  )

  flowers <- flowers102_dataset(root = t, split = "train", download = TRUE)
  expect_length(flowers, 1020)
  first_item <- flowers[1]
  expect_length(first_item$x,1131000)
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the Flowers102 dataset for test split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  flowers <- flowers102_dataset(root = t, split = "test", download = TRUE)
  expect_length(flowers, 6149)
  first_item <- flowers[1]
  expect_length(first_item$x,784500)
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the Flowers102 dataset for validation split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  flowers <- flowers102_dataset(root = t, split = "val", download = TRUE)
  expect_length(flowers, 1020)
  first_item <- flowers[1]
  expect_length(first_item$x,909000)
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the Flowers102 dataset for dataloader", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  flowers <- flowers102_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    }
  )
  dl <- dataloader(
    dataset = flowers,
    batch_size = 4
  )
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor(batch$y)
  expect_tensor_shape(batch$y,4)
  expect_tensor_dtype(batch$y,torch_long())
  expect_length(batch$y, 4)
  expect_equal_to_r(batch$y[1],1)
  expect_equal_to_r(batch$y[2],1)
  expect_equal_to_r(batch$y[3],1)
  expect_equal_to_r(batch$y[4],1)
})

context("dataset-flickr")

t <- withr::local_tempdir()

test_that("tests for the flickr8k dataset for train split", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")


  flickr8k <- flickr8k_caption_dataset(root = t, train = TRUE, download = TRUE)
  expect_length(flickr8k, 6000)
  first_item <- flickr8k[1]
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"character")
  expect_length(first_item$x,598500)
  expect_equal(first_item$y[1],"A black dog is running after a white dog in the snow .")
  expect_equal(first_item$y[2],"Black dog chasing brown dog through snow")
  expect_equal(first_item$y[3],"Two dogs chase each other across the snowy ground .")
  expect_equal(first_item$y[4],"Two dogs play together in the snow .")
  expect_equal(first_item$y[5],"Two dogs running through a low lying body of water .")
})

test_that("tests for the flickr8k dataset for test split", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  flickr8k <- flickr8k_caption_dataset(root = t, train = FALSE)
  expect_length(flickr8k, 1000)
  first_item <- flickr8k[1]
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"character")
  expect_length(first_item$x,502500)
  expect_equal(first_item$y[1], "The dogs are in the snow in front of a fence .")
  expect_equal(first_item$y[2], "The dogs play on the snow .")
  expect_equal(first_item$y[3], "Two brown dogs playfully fight in the snow .")
  expect_equal(first_item$y[4], "Two brown dogs wrestle in the snow .")
  expect_equal(first_item$y[5], "Two dogs playing in the snow .")
})

test_that("tests for the flickr8k dataset for dataloader", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")
  
  flickr8k <- flickr8k_caption_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    },
    target_transform = function(y) glue::glue_collapse(y, sep = " ")
  )
  dl <- dataloader(flickr8k, batch_size = 4)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_type(batch$y,"character")
  expect_length(batch$y,4)
  expect_equal(batch$y[1],"A black dog is running after a white dog in the snow . Black dog chasing brown dog through snow Two dogs chase each other across the snowy ground . Two dogs play together in the snow . Two dogs running through a low lying body of water .")
  expect_equal(batch$y[2],"A little baby plays croquet . A little girl plays croquet next to a truck . The child is playing croquette by the truck . The kid is in front of a car with a put and a ball . The little boy is playing with a croquet hammer and ball beside the car .")
  expect_equal(batch$y[3],"A brown dog in the snow has something hot pink in its mouth . A brown dog in the snow holding a pink hat . A brown dog is holding a pink shirt in the snow . A dog is carrying something pink in its mouth while walking through the snow . A dog with something pink in its mouth is looking forward .")
  expect_equal(batch$y[4],"A brown dog is running along a beach . A brown dog wearing a black collar running across the beach . A dog walks on the sand near the water . Brown dog running on the beach . The large brown dog is running on the beach by the ocean .")
})

test_that("tests for the flickr30k dataset for train split", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")


  flickr30k <- flickr30k_caption_dataset(root = t, train = TRUE, download = TRUE)
  expect_length(flickr30k, 29000)
  first_item <- flickr30k[1]
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"character")
  expect_length(first_item$x,499500)
  expect_equal(first_item$y[1], "Two young guys with shaggy hair look at their hands while hanging out in the yard.")
  expect_equal(first_item$y[2], "Two young, White males are outside near many bushes.")
  expect_equal(first_item$y[3], "Two men in green shirts are standing in a yard.")
  expect_equal(first_item$y[4], "A man in a blue shirt standing in a garden.")
  expect_equal(first_item$y[5], "Two friends enjoy time spent together.")

})

test_that("tests for the flickr30k dataset for test split", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  flickr30k <- flickr30k_caption_dataset(root = t, train = FALSE)
  expect_length(flickr30k, 1000)
  first_item <- flickr30k[1]
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"character")
  expect_length(first_item$x,691500)
  expect_equal(first_item$y[1], "The man with pierced ears is wearing glasses and an orange hat.")
  expect_equal(first_item$y[2], "A man with glasses is wearing a beer can crocheted hat.")
  expect_equal(first_item$y[3], "A man with gauges and glasses is wearing a Blitz hat.")
  expect_equal(first_item$y[4], "A man in an orange hat starring at something.")
  expect_equal(first_item$y[5], "A man wears an orange hat and glasses.")
})

test_that("tests for the flickr30k dataset for dataloader", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  flickr30k <- flickr30k_caption_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    },
    target_transform = function(y) glue::glue_collapse(y, sep = " ")
  )
  dl <- dataloader(flickr30k, batch_size = 4)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_type(batch$y, "character")
  expect_equal(batch$y[1],"Two young guys with shaggy hair look at their hands while hanging out in the yard. Two young, White males are outside near many bushes. Two men in green shirts are standing in a yard. A man in a blue shirt standing in a garden. Two friends enjoy time spent together.")
  expect_equal(batch$y[2],"Several men in hard hats are operating a giant pulley system. Workers look down from up above on a piece of equipment. Two men working on a machine wearing hard hats. Four men on top of a tall structure. Three men on a large rig.")
  expect_equal(batch$y[3],"A child in a pink dress is climbing up a set of stairs in an entry way. A little girl in a pink dress going into a wooden cabin. A little girl climbing the stairs to her playhouse. A little girl climbing into a wooden playhouse. A girl going into a wooden building.")
  expect_equal(batch$y[4],"Someone in a blue shirt and hat is standing on stair and leaning against a window. A man in a blue shirt is standing on a ladder cleaning a window. A man on a ladder cleans the window of a tall building. Man in blue shirt and jeans on ladder cleaning windows A man on a ladder cleans a window")
})

test_that("random_resized_crop", {

  img <- torch::torch_randn(3, 224, 224)
  o <- transform_random_resized_crop(img, size = c(32, 32))
  expect_tensor_shape(o, c(3, 32,32))

  im <- magick::image_read("torch.png")
  o <- transform_random_resized_crop(im, size = c(32, 32))
  expect_tensor_shape(transform_to_tensor(o), c(3, 32, 32))

})

test_that("transform works for magick images", {
  im <- magick::image_read("assets/class/horse/horse-2.tif")

  im <- transform_random_resized_crop(im, size = c(224, 224))
  ii <- magick::image_info(im)
  expect_equal(c(ii$width, ii$height), c(224,224))
})

test_that("transform_to_tensor works with magick image lists", {
  im <- magick::image_read("assets/class/horse/horse-2.tif")
  o <- transform_to_tensor(list(im, im, im))
  expect_tensor_shape(o, c(3, 3, 142, 180))
})

test_that("deeplabv3_resnet50 works with default aux_loss=NULL", {
  model <- model_deeplabv3_resnet50(num_classes = 21)
  input <- torch::torch_randn(1, 3, 32, 32)
  out <- model(input)

  expect_true(is.list(out))
  expect_named(out, "out")
  expect_tensor_shape(out$out, c(1, 21, 32, 32))

  withr::with_options(list(timeout = 360), {
    model <- model_deeplabv3_resnet50(pretrained = TRUE)
  })
  out <- model(input)
  expect_named(out, "out")
  expect_tensor_shape(out$out, c(1, 21, 32, 32))
})

test_that("deeplabv3_resnet101 works with default aux_loss=NULL", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_deeplabv3_resnet101(num_classes = 21)
  input <- torch::torch_randn(1, 3, 32, 32)
  out <- model(input)

  expect_true(is.list(out))
  expect_named(out, "out")
  expect_tensor_shape(out$out, c(1, 21, 32, 32))

  withr::with_options(list(timeout = 360), {
    model <- model_deeplabv3_resnet101(pretrained = TRUE)
  })
  out <- model(input)
  expect_named(out, "out")
  expect_tensor_shape(out$out, c(1, 21, 32, 32))
})

test_that("deeplabv3_resnet50 works with aux_loss = TRUE", {
  model <- model_deeplabv3_resnet50(aux_loss = TRUE, num_classes = 21)
  input <- torch::torch_randn(1, 3, 32, 32)
  out <- model(input)

  expect_named(out, c("out", "aux"))
  expect_tensor_shape(out$out, c(1, 21, 32, 32))
  expect_tensor_shape(out$aux, c(1, 21, 32, 32))
})

test_that("deeplabv3_resnet50 works with aux_loss = FALSE", {
  model <- model_deeplabv3_resnet50(aux_loss = FALSE, num_classes = 21)
  input <- torch::torch_randn(1, 3, 32, 32)
  out <- model(input)

  expect_named(out, "out")
  expect_tensor_shape(out$out, c(1, 21, 32, 32))
})

test_that("custom num_classes works with aux_loss = TRUE", {
  model <- model_deeplabv3_resnet50(num_classes = 3, aux_loss = TRUE)
  input <- torch::torch_randn(1, 3, 32, 32)
  out <- model(input)

  expect_named(out, c("out", "aux"))
  expect_tensor_shape(out$out, c(1, 3, 32, 32))
  expect_tensor_shape(out$aux, c(1, 3, 32, 32))
})

test_that("pretrained requires num_classes to be \'21\'", {
  expect_error(
    model_deeplabv3_resnet50(pretrained = TRUE, num_classes = 3),
    "Pretrained weights on COCO require"
  )
})

test_that("model_deeplabv3_resnet50 detects cat in Wikipedia image", {

  img_url <- "https://upload.wikimedia.org/wikipedia/commons/3/36/United_Airlines_Boeing_777-200_Meulemans.jpg"
  img <- magick::image_read(img_url)

  norm_mean <- c(0.485, 0.456, 0.406)
  norm_std <- c(0.229, 0.224, 0.225)

  input <- transform_to_tensor(img)
  input <- transform_resize(input, c(520, 520))
  input <- transform_normalize(input, mean = norm_mean, std = norm_std)
  input <- input$unsqueeze(1)

  model <- model_deeplabv3_resnet50(pretrained = TRUE)
  model$eval()

  output <- model(input)
  mask <- output$out$argmax(dim = 2)  # shape (1, H, W)

  label_array <- mask %>% torch::as_array()  # convert to R array
  label_table <- table(factor(label_array, levels = 0:20, labels = pascal_voc_classes()))

  expect_gt(label_table[["aeroplane"]], 0)
  expect_gt(label_table[["aeroplane"]], label_table[["dog"]])
  expect_gt(label_table[["aeroplane"]], label_table[["person"]])
})

context("dataset-lfw")

t <- withr::local_tempdir()

test_that("tests for the LFW People dataset for original image_set", {
  # MacOS / Windows runner fails with `cannot open URL 'https://ndownloader.figshare.com/files/5976015'` or `timeout`
  skip_on_os("mac")
  skip_on_os("windows")
  skip_on_os("linux")
  lfw <- lfw_people_dataset(root = t, download = TRUE, split = "original")
  expect_length(lfw, 13233)
  first_item <- lfw[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x, 187500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the LFW People dataset for funneled image_set", {
  # MacOS / Windows runner fails with `cannot open URL 'https://ndownloader.figshare.com/files/5976015'` or `timeout`
  skip_on_os("mac")
  skip_on_os("windows")
  skip_on_os("linux")
  lfw <- lfw_people_dataset(root = t, download = TRUE, split = "funneled" )
  expect_length(lfw, 13233)
  first_item <- lfw[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x, 187500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the LFW People dataset for dataloader", {
  # MacOS / Windows runner fails with `cannot open URL 'https://ndownloader.figshare.com/files/5976015'` or `timeout`
  skip_on_os("mac")
  skip_on_os("windows")
  skip_on_os("linux")
  lfw <- lfw_people_dataset(root = t, transform = transform_to_tensor)
  dl <- dataloader(lfw, batch_size = 32)
  batch <- dataloader_next(dataloader_make_iter(dl))
  expect_named(batch, c("x", "y"))
  expect_length(batch$x, 6000000)
  expect_tensor(batch$x)
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor_dtype(batch$y,torch_long())
  expect_tensor_shape(batch$x,c(32,3,250,250))
  expect_tensor_shape(batch$y,32)
  expect_tensor(batch$y)
  expect_equal_to_r(batch$y[1], 1)
  expect_equal_to_r(batch$y[2], 2)
})

test_that("tests for the LFW Pairs dataset for original image_set train split", {
  # MacOS runner fails with cannot open URL 'https://ndownloader.figshare.com/files/5976015'
  skip_on_os("mac")
  skip_on_os("windows")
  skip_on_os("linux")
  lfw <- lfw_pairs_dataset(root = t, download = TRUE, split = "original", train = TRUE)
  expect_length(lfw, 2200)
  first_item <- lfw[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x, 2)
  expect_length(first_item$x[[1]], 187500)
  expect_length(first_item$x[[2]], 187500)
  expect_type(first_item$x, "list")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the LFW Pairs dataset for funneled image_set train split", {
  # MacOS runner fails with cannot open URL 'https://ndownloader.figshare.com/files/5976015'
  skip_on_os("mac")
  skip_on_os("windows")
  skip_on_os("linux")
  lfw <- lfw_pairs_dataset(root = t, train = TRUE, split = "funneled", download = TRUE)
  expect_length(lfw, 2200)
  first_item <- lfw[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x, 2)
  expect_length(first_item$x[[1]], 187500)
  expect_length(first_item$x[[2]], 187500)
  expect_type(first_item$x, "list")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the LFW Pairs dataset for original image_set test split", {
  # MacOS runner fails with cannot open URL 'https://ndownloader.figshare.com/files/5976015'
  skip_on_os("mac")
  skip_on_os("windows")
  skip_on_os("linux")
  lfw <- lfw_pairs_dataset(root = t, split = "original", train = FALSE)
  expect_length(lfw, 1000)
  first_item <- lfw[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x, 2)
  expect_length(first_item$x[[1]], 187500)
  expect_length(first_item$x[[2]], 187500)
  expect_type(first_item$x, "list")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the LFW Pairs dataset for funneled image_set test split", {
  # MacOS runner fails with cannot open URL 'https://ndownloader.figshare.com/files/5976015'
  skip_on_os("mac")
  skip_on_os("windows")
  skip_on_os("linux")
  lfw <- lfw_pairs_dataset(root = t, train = FALSE, split = "funneled")
  expect_length(lfw, 1000)
  first_item <- lfw[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x, 2)
  expect_length(first_item$x[[1]], 187500)
  expect_length(first_item$x[[2]], 187500)
  expect_type(first_item$x, "list")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the LFW People dataset for dataloader", {
  # MacOS runner fails with cannot open URL 'https://ndownloader.figshare.com/files/5976015'
  skip_on_os("mac")
  skip_on_os("windows")
  lfw <- lfw_pairs_dataset(
    root = t,
    transform = function(pair) {
      pair[[1]] %>% transform_to_tensor()
      pair[[2]] %>% transform_to_tensor()
      pair
    },
    download = TRUE)
  dl <- dataloader(lfw, batch_size = 32)
  batch <- dataloader_next(dataloader_make_iter(dl))
  expect_named(batch, c("x", "y"))
  expect_length(batch$x, 2)
  expect_type(batch$x, "list")
  expect_tensor(batch$y)
  expect_tensor_shape(batch$y, 32)
  expect_tensor_dtype(batch$y, torch_long())
  expect_equal_to_r(batch$y[1], 1)
  expect_equal_to_r(batch$y[2], 1)
  expect_equal_to_r(batch$y[32], 1)
})

test_that("R-level type_error messages are correctly translated in FR", {
  withr::with_language(lang = "fr",
                       expect_error(
                         transform_adjust_gamma(torch::torch_rand_like(c(3, 5, 5)), gamma = 0.5),
                        regexp = "Le tenseur n'est pas une image torch",
                        fixed = TRUE
                      )
  )
})

test_that("R-level cli_warning messages are correctly translated in FR", {
  withr::with_language(lang = "fr",
                       expect_warning(
                         torchvision:::Inception3(),
                        regexp = "L'initialisation des poids par défaut de inception_v3",
                        fixed = TRUE
                      )
  )
})

test_that("R-level value_error messages are glued and correctly translated in FR", {
  withr::with_language(lang = "fr",
                       expect_error(
                         transform_normalize(torch::torch_rand(c(3,5,5)), 3, 0),
                        regexp = "Après conversion en Float,",
                        fixed = TRUE
                      )
  )
})

context("dataset-caltech")

t <- withr::local_tempdir()

test_that("Caltech101 dataset works correctly", {

  expect_error(
    caltech101_dataset(root = tempfile(), download = FALSE),
    class = "rlang_error"
  )

  ds <- caltech101_dataset(root = t, download = TRUE)
  expect_length(ds, 8677)
  first_item <- ds[1]
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("Caltech101 dataset works correctly (dataloader)", {

  caltech101 <- caltech101_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    }
  )
  dl <- dataloader(caltech101, batch_size = 4)
  batch <- dataloader_next(dataloader_make_iter(dl))
  expect_length(batch, 2)
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor(batch$y)
  expect_tensor_dtype(batch$y, torch_long())
  expect_tensor_shape(batch$y, 4)
  expect_tensor(batch$y[1])
  expect_tensor_dtype(batch$y[1],torch_long())
  expect_equal_to_r(batch$y[1],1)
  expect_equal_to_r(batch$y[2],1)
  expect_equal_to_r(batch$y[3],1)
  expect_equal_to_r(batch$y[4],1)

})

test_that("Caltech256 dataset works correctly", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_error(
    caltech256_dataset(root = tempfile(), download = FALSE),
    class = "rlang_error"
  )

  caltech256 <- caltech256_dataset(root = t, download = TRUE)
  expect_length(caltech256, 30607)
  first_item <- caltech256[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,416166)
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)

  expect_equal(caltech256$classes, caltech_classes())
})

test_that("Caltech256 dataset works correctly (dataloader)", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  caltech256 <- caltech256_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    }
  )
  dl <- dataloader(caltech256, batch_size = 4)
  batch <- dataloader_next(dataloader_make_iter(dl))
  expect_length(batch, 2)
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor(batch$y)
  expect_tensor_dtype(batch$y, torch_long())
  expect_tensor_shape(batch$y, 4)
  expect_tensor(batch$y[1])
  expect_tensor_dtype(batch$y[1],torch_long())
  expect_equal_to_r(batch$y[1],1)
  expect_equal_to_r(batch$y[2],1)
  expect_equal_to_r(batch$y[3],1)
  expect_equal_to_r(batch$y[4],1)

})

test_that("transform_to_tensor works for 2D, 3D, 4D array", {
  x <- array(1L:120L, dim = c(10,12))
  o <- transform_to_tensor(x)
  expect_tensor_shape(o, c(1,10,12))
  expect_tensor_dtype(o, torch_float())
  expect_equal_to_r(o$max(), 120/255)

  x <- array(1L:120L, dim = c(1, 10,12))
  o <- transform_to_tensor(x)
  expect_tensor_shape(o, c(1,10,12))
  expect_tensor_dtype(o, torch_float())
  expect_equal_to_r(o$max(), 120/255)

  x <- array(1L:120L, dim = c(3, 5, 8))
  o <- transform_to_tensor(x)
  expect_tensor_shape(o, c(3, 5, 8))
  expect_tensor_dtype(o, torch_float())
  expect_equal_to_r(o$max(), 120/255)


  x <- array(1L:120L, dim = c(2, 1, 4,5))
  ob <- transform_to_tensor(x)
  expect_tensor_shape(ob, c(2, 1, 4, 5))
  expect_tensor_dtype(ob, torch_float())

  x <- array(1L:120L, dim = c(2, 3, 10,12))
  ob <- transform_to_tensor(x)
  expect_tensor_shape(ob, c(2, 3,10,12))
  expect_tensor_dtype(ob, torch_float())

})

test_that("transform_to_tensor works for list of arrays", {
  x <- array(1L:120L, dim = c(10,12))
  o <- transform_to_tensor(list(x, x))
  expect_tensor_shape(o, c(2,1,10,12))
  expect_tensor_dtype(o, torch_float())
  expect_equal_to_r(o$max(), 120/255)

  x <- array(1L:120L, dim = c(1, 10,12))
  o <- transform_to_tensor(list(x, x))
  expect_tensor_shape(o, c(2,1,10,12))
  expect_tensor_dtype(o, torch_float())
  expect_equal_to_r(o$max(), 120/255)

  x <- array(1L:120L, dim = c(3, 5, 8))
  o <- transform_to_tensor(list(x, x, x))
  expect_tensor_shape(o, c(3, 3, 5, 8))
  expect_tensor_dtype(o, torch_float())
  expect_equal_to_r(o$max(), 120/255)


  x <- array(1L:120L, dim = c(2, 1, 4,5))
  expect_error(transform_to_tensor(list(x,x)),
               "3D arrays")

})

input <- base_loader("assets/class/cat/cat.0.jpg") %>%
  transform_to_tensor() %>%
  transform_normalize(c(0.485, 0.456, 0.406), c(0.229, 0.224, 0.225)) %>%
  transform_resize(c(200,200)) %>%
  torch_unsqueeze(1)


test_that("maskrcnn_resnet50_fpn loads without pretrained weights", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_maskrcnn_resnet50_fpn(pretrained = FALSE, num_classes = 90)
  expect_s3_class(model, "nn_module")
  expect_true(!is.null(model$backbone))
  expect_true(!is.null(model$rpn))
  expect_true(!is.null(model$roi_heads))
  expect_true(!is.null(model$mask_head))
})

test_that("maskrcnn_resnet50_fpn_v2 loads without pretrained weights", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_maskrcnn_resnet50_fpn_v2(pretrained = FALSE, num_classes = 91)
  expect_s3_class(model, "nn_module")
  expect_true(!is.null(model$backbone))
  expect_true(!is.null(model$rpn))
  expect_true(!is.null(model$roi_heads))
  expect_true(!is.null(model$mask_head))
})

test_that("maskrcnn_resnet50_fpn inference works", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  # We expect 0 detection here, we keep the compute budget for pretrained model
  model <- model_maskrcnn_resnet50_fpn(pretrained = FALSE, num_classes = 91, score_thresh = 0.6, nms_thresh = 0.1, detections_per_img = 10)
  model$eval()

  # Test forward pass
  torch::with_no_grad({output <- model(input)})

  # Check output structure
  expect_named(output, c("features" ,"detections"))
  expect_named(output$detections[[1]], c("boxes" ,"labels","scores","masks"))

  # Verify tensor types
  expect_tensor(output$detections[[1]]$boxes)
  expect_tensor(output$detections[[1]]$labels)
  expect_tensor(output$detections[[1]]$scores)
  expect_tensor(output$detections[[1]]$masks)

  # Verify dimensions are consistent
  n_masks <- output$detections[[1]]$masks$shape[1]
  expect_tensor_shape(output$detections[[1]]$masks, c(n_masks, 28, 28))
  expect_tensor_shape(output$detections[[1]]$labels, n_masks)
  expect_tensor_shape(output$detections[[1]]$scores, n_masks)


  # Check mask dimensions (should be N x 28 x 28)
  N <- output$detections[[1]]$masks$shape[1]
  expect_tensor_shape(output$detections[[1]]$masks, c(N, 28, 28))

})

test_that("maskrcnn_resnet50_fpn_v2 inference works", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")


  # We expect 0 detection here, we keep the compute budget for pretrained model
  model <- model_maskrcnn_resnet50_fpn_v2(pretrained = FALSE, num_classes = 91, score_thresh = 0.6, nms_thresh = 0.9, detections_per_img = 10)
  model$eval()

  # Test forward pass
  torch::with_no_grad({output <- model(input)})

  # Check output structure
  expect_named(output, c("features" ,"detections"))
  expect_named(output$detections[[1]], c("boxes" ,"labels","scores","masks"))

  # Check masks are present
  expect_tensor(output$detections[[1]]$masks)

  # Check mask dimensions (should be N x 28 x 28)
  n_masks <- output$detections[[1]]$masks$shape[1]
  expect_tensor_shape(output$detections[[1]]$masks, c(n_masks, 28, 28))
})

test_that("maskrcnn handles empty detections correctly", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")


  # Use very high score threshold to force empty detections[[1]]
  model <- model_maskrcnn_resnet50_fpn(
    pretrained = FALSE,
    num_classes = 91,
    score_thresh = 0.99, nms_thresh = 0.9, detections_per_img = 10
  )
  model$eval()

  torch::with_no_grad({output <- model(input)})

  # Should return empty tensors with correct shapes
  expect_tensor_shape(output$detections[[1]]$boxes, c(0,4))
  expect_tensor_shape(output$detections[[1]]$labels, 0)
  expect_tensor_shape(output$detections[[1]]$scores, 0)

  # Mask shape should be (0, 28, 28)
  expect_tensor_shape(output$detections[[1]]$masks, c(0,28,28))
})

test_that("maskrcnn respects detections_per_img parameter", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  max_detections <- 5
  model <- model_maskrcnn_resnet50_fpn(
    pretrained = FALSE,
    num_classes = 91,
    score_thresh = 0.4,  # Low threshold to get more detections
    nms_thresh = 0.8,
    detections_per_img = max_detections
  )
  model$eval()

  output <- model(input)

  # Number of boxes should not exceed max_detections
  n_boxes <- output$detections[[1]]$boxes$shape[1]
  expect_lte(n_boxes, max_detections)
})

test_that("maskrcnn with different num_classes works", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  # Test with custom number of classes
  model <- model_maskrcnn_resnet50_fpn(pretrained = FALSE, num_classes = 10, score_thresh = 0.6, nms_thresh = 0.9, detections_per_img = 100)
  model$eval()

  input <- torch::torch_randn(1, 3, 800, 800)
  torch::with_no_grad({output <- model(input)})

  # Should work without errors
  expect_tensor(output$detections[[1]]$masks)
})

test_that("maskrcnn pretrained requires num_classes = 90", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  # Should error when pretrained=TRUE with num_classes != 91
  expect_error(
    model_maskrcnn_resnet50_fpn(pretrained = TRUE, num_classes = 10, score_thresh = 0.6, nms_thresh = 0.9, detections_per_img = 100),
    "Pretrained weights require num_classes = 90"
  )
})

test_that("mask_rcnn pretrained infer correctly", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_maskrcnn_resnet50_fpn(pretrained = TRUE, score_thresh = 0.5, nms_thresh = 0.7, detections_per_img = 10)
  model$eval()

  torch::with_no_grad({output <- model(input)})
  # Masks should be expanded back to image size
  if (output$detections[[1]]$boxes$shape[1] > 0) {
    expect_bbox_is_xyxy(output$detections[[1]]$boxes, 200, 200)
    # Verify background class is removed, labels should be COCO IDs [1, 90]
    labels_vec <- as.integer(output$detections[[1]]$labels$cpu())
    expect_true(all(labels_vec >= 1 & labels_vec <= 90), info = "All labels are in range [1, 90]")
    expect_true(any(labels_vec == 17), info = "model found a cat")
    # Verify masks match number of detections
    expect_equal(output$detections[[1]]$masks$shape[1], length(labels_vec),
                 info = "Number of masks should match number of detections")
  }

})

test_that("mask_rcnn_v2 pretrained infer correctly", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_maskrcnn_resnet50_fpn_v2(pretrained = TRUE, score_thresh = 0.5, nms_thresh = 0.7, detections_per_img = 10)
  model$eval()

  torch::with_no_grad({output <- model(input)})
  # Masks should be expanded back to image size
  if (output$detections[[1]]$boxes$shape[1] > 0) {
    expect_bbox_is_xyxy(output$detections[[1]]$boxes, 200, 200)
    # Verify background class is removed, labels should be COCO IDs [1, 90]
    labels_vec <- as.integer(output$detections[[1]]$labels$cpu())
    expect_true(all(labels_vec >= 1 & labels_vec <= 90), info = "All labels are in range [1, 90]")
    expect_true(any(labels_vec == 17), info = "model found a cat")
    # Verify masks match number of detections
    expect_equal(output$detections[[1]]$masks$shape[1], length(labels_vec),
                 info = "Number of masks should match number of detections")
  }

})

test_that("mask_head_module initializes correctly", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  mask_head <- mask_head_module()
  mask_rcnn_pred <- mask_rcnn_predictor(num_classes = 90)
  expect_is(mask_head, "nn_module")

  # Test forward pass
  input <- torch::torch_randn(2, 256, 14, 14)
  output <- input %>% mask_head() %>% mask_rcnn_pred()

  # Output should be (2, 91, 28, 28) - 90 classes + 1 background
  expect_tensor_shape(output, c(2, 91, 28, 28))
})

test_that("mask_head_module_v2 initializes correctly", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  mask_head <- mask_head_module_v2(num_classes = 90)
  expect_s3_class(mask_head, "nn_module")

  # Test forward pass
  input <- torch::torch_randn(2, 256, 14, 14)
  output <- mask_head(input)

  # Output should be (2, 91, 28, 28) - 90 classes + 1 background
  expect_tensor_shape(output, c(2, 91, 28, 28))
})

test_that("roi_align_masks produces correct output shape", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  # Create dummy feature map (1, 256, 100, 100)
  feature_map <- torch::torch_randn(1, 256, 100, 100)

  # Create dummy proposals (5 boxes)
  proposals <- torch::torch_tensor(rbind(
    c(10, 10, 50, 50),
    c(20, 20, 60, 60),
    c(30, 30, 70, 70),
    c(40, 40, 80, 80),
    c(50, 50, 90, 90)
  ))

  # Apply ROI align
  output <- roi_align_masks(feature_map, proposals, output_size = c(14L, 14L))

  # Output should be (5, 256, 14, 14)
  expect_tensor_shape(output, c(5, 256, 14, 14))
})

context("dataset-oxfordiiitpet")

t <- withr::local_tempdir()

test_that("tests for the Oxford-IIIT Pet Segmentation dataset for train split with target type category", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_segmentation_dataset(root = t, target_type = "category", train = TRUE, download = TRUE)
  expect_length(oxfordiiitpet, 3680)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,591000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(3,500,394))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$label, "integer")
  expect_equal(first_item$y$label, 1)
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Oxford-IIIT Pet Segmentation dataset for train split with target type binary category", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_segmentation_dataset(root = t, target_type = "binary-category", train = TRUE)
  expect_length(oxfordiiitpet, 3680)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,591000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(3,500,394))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$label,"integer")
  expect_equal(first_item$y$label, 1)
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Oxford-IIIT Pet Segmentation dataset for test split with target type category", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_segmentation_dataset(root = t, target_type = "category", train = FALSE)
  expect_length(oxfordiiitpet, 3669)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,202500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(3,225,300))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$label,"integer")
  expect_equal(first_item$y$label, 1)
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Oxford-IIIT Pet Segmentation dataset for test split with target type binary category", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_segmentation_dataset(root = t, target_type = "binary-category", train = FALSE)
  expect_length(oxfordiiitpet, 3669)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,202500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(3,225,300))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$label,"integer")
  expect_equal(first_item$y$label, 1)
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Oxford-IIIT Pet Segmentation dataset for dataloader", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_segmentation_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    },
    target_transform = function(y) {
      y$masks <- y$masks %>% transform_resize(c(224, 224))
      y
    }
  )
  expect_s3_class(oxfordiiitpet[1], "image_with_segmentation_mask")
  dl <- dataloader(oxfordiiitpet, batch_size = 4)
  batch <- dataloader_next(dataloader_make_iter(dl))
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_type(batch$y,"list")
  expect_length(batch$y, 2)
  expect_type(batch$y[1], "list")
  expect_tensor(batch$y$mask)
  expect_tensor_shape(batch$y$mask, c(4,3,224,224))
  expect_tensor_dtype(batch$y$mask, torch_bool())
  expect_tensor(batch$y$label)
  expect_tensor_shape(batch$y$label, 4)
  expect_tensor_dtype(batch$y$label, torch_long())
  expect_equal_to_r(batch$y$label[1], 1)
  expect_equal_to_r(batch$y$label[2], 1)
  expect_equal_to_r(batch$y$label[3], 1)
  expect_equal_to_r(batch$y$label[4], 1)
})

test_that("tests for the Oxford-IIIT Pet dataset for train split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_dataset(root = t, train = TRUE, download = TRUE)
  expect_length(oxfordiiitpet, 3680)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,591000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the Oxford-IIIT Pet dataset for test split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_dataset(root = t, train = FALSE)
  expect_length(oxfordiiitpet, 3669)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,202500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the Oxford-IIIT Pet dataset for dataloader", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    }
  )
  dl <- dataloader(oxfordiiitpet, batch_size = 4)
  batch <- dataloader_next(dataloader_make_iter(dl))
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor(batch$y)
  expect_tensor_dtype(batch$y,torch_long())
  expect_tensor_shape(batch$y,4)
  expect_equal_to_r(batch$y[1], 1)
  expect_equal_to_r(batch$y[2], 1)
  expect_equal_to_r(batch$y[3], 1)
  expect_equal_to_r(batch$y[4], 1)
})

test_that("tests for the Oxford-IIIT Pet binary dataset for train split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_binary_dataset(root = t, train = TRUE)
  expect_length(oxfordiiitpet, 3680)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,591000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the Oxford-IIIT Pet binary dataset for test split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_binary_dataset(root = t, train = FALSE)
  expect_length(oxfordiiitpet, 3669)
  first_item <- oxfordiiitpet[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,202500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)
})

test_that("tests for the Oxford-IIIT Pet binary dataset for dataloader", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  oxfordiiitpet <- oxfordiiitpet_binary_dataset(
    root = t,
    transform = function(x) {
      x %>% transform_to_tensor() %>% transform_resize(c(224, 224))
    }
  )
  dl <- dataloader(oxfordiiitpet, batch_size = 4)
  batch <- dataloader_next(dataloader_make_iter(dl))
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_length(batch$x,602112)
  expect_tensor_shape(batch$x,c(4,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor(batch$y)
  expect_tensor_dtype(batch$y,torch_long())
  expect_tensor_shape(batch$y,4)
  expect_equal_to_r(batch$y[1], 1)
  expect_equal_to_r(batch$y[2], 1)
  expect_equal_to_r(batch$y[3], 1)
  expect_equal_to_r(batch$y[4], 1)
})

test_that("image_folder dataset", {

  ds <- image_folder_dataset(
    root = "assets/class",
    transform = . %>% transform_to_tensor %>%
      transform_resize(c(32,32))
  )
  expect_length(ds[1], 2)

  dl <- torch::dataloader(ds, batch_size = 2, drop_last = TRUE)
  coro::loop(for(batch in dl) {
    expect_tensor_shape(batch[[1]], c(2, 3, 32, 32))
    expect_tensor_shape(batch[[2]], 2)
    expect_tensor_shape(batch$x, c(2, 3, 32, 32))
    expect_tensor_shape(batch$y, 2)
  })

  expect_length(ds, 15)

})

test_that("default_loader works as expected", {
  # rvb jpeg
  cat1 <- base_loader("assets/class/cat/cat.1.jpg")
  expect_equal(dim(cat1)[3], 3L)
  # rvb png
  horse1 <- base_loader("assets/class/horse/horse-1.png")
  expect_equal(dim(horse1)[3], 3L)
  # rvb tiff
  horse2 <- base_loader("assets/class/horse/horse-2.tif")
  expect_equal(dim(horse2)[3], 3L)
  # grayscale jpeg
  dog5 <- base_loader("assets/class/dog/dog.5.jpg")
  expect_equal(dim(dog5)[3], 3L)
})

test_that("default_loader gives clear error for missing file", {
  missing_path <- file.path(tempdir(), "missing-image.jpg")
  expect_false(file.exists(missing_path))
  expect_error(base_loader(missing_path), "File not found")
})

# each test runs on single tensor image (dim = 3) and with the `b` suffix on batch tensor images (dim = 4)
test_that("convert_image_dtype", {

  x <- torch::torch_rand(1, 2, 2, dtype = torch_float32())
  o <- transform_convert_image_dtype(x, torch_int16())
  y <- transform_convert_image_dtype(o, torch_float32())

  expect_equal(round(as_array(x),1), round(as_array(y),1))

  ob <- transform_convert_image_dtype(x$unsqueeze(1), torch_int16())
  yb <- transform_convert_image_dtype(ob, torch_float32())

  expect_equal(round(as_array(x$unsqueeze(1)),1), round(as_array(yb),1))

})

test_that("normalize", {

  x <- torch_randn(3, 10, 10)
  o <- transform_normalize(x, 1, 2)

  expect_equal_to_r(o, as_array((x - 1)/2))

  ob <- transform_normalize(x$unsqueeze(1), 1, 2)

  expect_equal_to_r(ob, as_array((x$unsqueeze(1) - 1)/2))

})

test_that("normalize error is glued", {

  x <- torch_randn(3, 10, 10)

  expect_error(transform_normalize(x, 1, 0), "evaluated to zero after conversion to Float")
  expect_error(transform_normalize(x$unsqueeze(1), 1, 0), "evaluated to zero after conversion to Float")

})

test_that("resize", {
  x <- torch_randn(3, 10, 10)
  o <- transform_resize(x, c(20, 20))
  expect_tensor_shape(o, c(3, 20, 20))
  ob <- transform_resize(x$unsqueeze(1), c(20, 20))
  expect_tensor_shape(ob, c(1, 3, 20, 20))

  x <- torch_randn(3, 10, 20)
  o <- transform_resize(x, c(10, 10))
  expect_tensor_shape(o, c(3, 10, 10))
  ob <- transform_resize(x$unsqueeze(1), c(10, 10))
  expect_tensor_shape(ob, c(1, 3, 10, 10))

  x <- torch_randn(3, 10, 20)
  o <- transform_resize(x, c(10))
  expect_tensor_shape(o, c(3, 10, 20))
  ob <- transform_resize(x$unsqueeze(1), c(10))
  expect_tensor_shape(ob, c(1, 3, 10, 20))

  x <- torch_randn(3, 20, 10)
  o <- transform_resize(x, c(10))
  expect_tensor_shape(o, c(3, 20, 10))
  ob <- transform_resize(x$unsqueeze(1), c(10))
  expect_tensor_shape(ob, c(1, 3, 20, 10))

  x <- torch_randn(3, 10, 5)
  o <- transform_resize(x, 10)
  expect_tensor_shape(o, c(3, 20, 10))
  ob <- transform_resize(x$unsqueeze(1), 10)
  expect_tensor_shape(ob, c(1, 3, 20, 10))
})

test_that("pad", {

  x <- torch_randn(3, 10, 10)
  o <- transform_pad(x, c(1,2))
  expect_tensor_shape(o, c(3, 14, 12))

  ob <- transform_pad(x$unsqueeze(1), c(1,2))
  expect_tensor_shape(ob, c(1, 3, 14, 12))
})

test_that("crop", {

  x <- torch_randn(3, 10, 10)
  o <- transform_crop(x, 1, 1, 2, 2)

  expect_tensor_shape(o, c(3,2,2))
  expect_equal(as_array(x[,1,1]), as_array(o[,1,1]))

  ob <- transform_crop(x$unsqueeze(1), 1, 1, 2, 2)

  expect_tensor_shape(ob, c(1,3,2,2))
  expect_equal(as_array(x$unsqueeze(1)[,,1,1]), as_array(ob[,,1,1]))

})

test_that("center_crop", {

  x <- torch_randn(3, 10, 10)
  o <- transform_center_crop(x, c(2,2))

  expect_tensor_shape(o, c(3,2,2))

  ob <- transform_center_crop(x$unsqueeze(1), c(2,2))

  expect_tensor_shape(ob, c(1,3,2,2))

})

test_that("resized_crop", {

  x <- torch_randn(3, 10, 10)
  o <- transform_resized_crop(x, 1, 1, 2, 2, size = c(6, 6))

  expect_tensor_shape(o, c(3,6,6))

  ob <- transform_resized_crop(x$unsqueeze(1), 1, 1, 2, 2, size = c(6, 6))

  expect_tensor_shape(ob, c(1,3,6,6))

})

test_that("hflip", {

  x <- torch_randn(3, 10, 10)
  o <- transform_hflip(x)

  expect_equal_to_r(o[,,1], as_array(x[,,10]))

  ob <- transform_hflip(x$unsqueeze(1))

  expect_equal_to_r(ob[..,1], as_array(x$unsqueeze(1)[..,10]))

})

test_that("perspective", {

  skip("not implemented")

  x <- torch_randn(3, 50, 50)
  o <- transform_perspective(x, startpoints = list(c(2,2), c(2,3), c(3,2), c(3,3)),
                             endpoints = list(c(4,4), c(4,5), c(5,4), c(5,5)))

  ob <- transform_perspective(x$unsqueeze(1), startpoints = list(c(2,2), c(2,3), c(3,2), c(3,3)),
                             endpoints = list(c(4,4), c(4,5), c(5,4), c(5,5)))

})

test_that("vflip", {

  x <- torch_randn(3, 10, 10)
  o <- transform_vflip(x)

  expect_equal_to_r(o[,1,], as_array(x[,10,]))

  ob <- transform_vflip(x$unsqueeze(1))

  expect_equal_to_r(ob[,,1,], as_array(x$unsqueeze(1)[,,10,]))

})

test_that("five_crop", {

  x <- torch_randn(3, 10, 12)
  o <- transform_five_crop(x, c(3, 3))

  expect_length(o, 5)
  expect_tensor_shape(o[[1]], c(3,3,3))

  ob <- transform_five_crop(x$unsqueeze(1), c(3, 3))

  expect_length(ob, 5)
  expect_tensor_shape(ob[[1]], c(1,3,3,3))

})

test_that("ten_crop", {

  x <- torch_randn(3, 10, 10)
  o <- transform_ten_crop(x, c(3, 3))

  expect_length(o, 10)
  expect_tensor_shape(o[[1]], c(3,3,3))

  ob <- transform_ten_crop(x$unsqueeze(1), c(3, 3))

  expect_length(ob, 10)
  expect_tensor_shape(ob[[1]], c(1,3,3,3))
})

test_that("rotate", {

  img <- torch::torch_tensor(matrix(1:16))$view(c(1, 4, 4))
  output <- transform_rotate(img, 90)

  expect_tensor_shape(output, c(1,4,4))
  expect_equal_to_r(output[1,,1], c(4,3,2,1))

  output <- transform_rotate(img, 45, expand = TRUE)
  expect_equal_to_r(output[1,,2], c(0,0, 2, 5, 0, 0))
  expect_equal_to_r(output[1,,3], c(0,3, 7, 10, 9, 0))

  outputb <- transform_rotate(img$unsqueeze(1), 90)

  expect_tensor_shape(outputb, c(1,1,4,4))
  expect_equal_to_r(outputb[,1,,1]$squeeze(1), c(4,3,2,1))

  outputb <- transform_rotate(img$unsqueeze(1), 45, expand = TRUE)
  expect_equal_to_r(outputb[,1,,2]$squeeze(1), c(0,0, 2, 5, 0, 0))
  expect_equal_to_r(outputb[,1,,3]$squeeze(1), c(0,3, 7, 10, 9, 0))

})

test_that("rotate a rectangle image", {

  img <- torch::torch_tensor(matrix(1:20))$view(c(1, 4, 5))
  output <- transform_rotate(img, 90)

  expect_tensor_shape(output, c(1,5,4))
  expect_equal_to_r(output[1,,1], c(5,4,3,2,1))
  expect_equal_to_r(output[1,,4], c(20,19,18,17,16))

  outputb <- transform_rotate(img$unsqueeze(1), 90)

  expect_tensor_shape(outputb, c(1,1,5,4))
  expect_equal_to_r(outputb[,1,,1]$squeeze(1), c(5,4,3,2,1))
  expect_equal_to_r(outputb[,1,,4]$squeeze(1), c(20,19,18,17,16))

})

test_that("random_affine", {

  x <- torch_eye(8)$view(c(1, 8, 8))

  # no translation
  o <- transform_random_affine(x, 0, c(0, 0))
  expect_equal(as.numeric(torch_sum(x)), as.numeric(torch_sum(o)))

  ob <- transform_random_affine(x$unsqueeze(1), 0, c(0, 0))
  expect_equal(as.numeric(torch_sum(x)), as.numeric(torch_sum(ob)))

  # probabilistic transformation with p = 0.1 should not result in sum deviating by > 1
  o <- transform_random_affine(x, 0, c(0.1, 0))
  expect_lte(as.numeric(torch_sum(x) - 1), as.numeric(torch_sum(o)))
  expect_gte(as.numeric(torch_sum(x)), as.numeric(torch_sum(o)))

  o <- transform_random_affine(x, 0, c(0, 0.1))
  expect_lte(as.numeric(torch_sum(x) - 1), as.numeric(torch_sum(o)))
  expect_gte(as.numeric(torch_sum(x)), as.numeric(torch_sum(o)))

  ob <- transform_random_affine(x$unsqueeze(1), 0, c(0.1, 0))
  expect_lte(as.numeric(torch_sum(x) - 1), as.numeric(torch_sum(ob)))
  expect_gte(as.numeric(torch_sum(x)), as.numeric(torch_sum(ob)))

  ob <- transform_random_affine(x$unsqueeze(1), 0, c(0, 0.1))
  expect_lte(as.numeric(torch_sum(x) - 1), as.numeric(torch_sum(ob)))
  expect_gte(as.numeric(torch_sum(x)), as.numeric(torch_sum(ob)))

})

test_that("affine", {

  x <- torch_eye(8)$view(c(1, 8, 8))

  # translate by 1 pixel horizontally
  # should result in sum smaller by 1
  o <- transform_affine(x, 0, c(0, 1), 1, 0)
  expect_equal(as.numeric(torch_sum(x)) - 1, as.numeric(torch_sum(o)))

  ob <- transform_affine(x$unsqueeze(1), 0, c(0, 1), 1, 0)
  expect_equal(as.numeric(torch_sum(x)) - 1, as.numeric(torch_sum(ob)))

  # translate by 1 pixel vertically
  # should result in sum smaller by 1
  o <- transform_affine(x, 0, c(1, 0), 1, 0)
  expect_equal(as.numeric(torch_sum(x) - 1), as.numeric(torch_sum(o)))

  # interpolation accepts enum-style string values
  o_int <- transform_affine(x, 0, c(0, 1), 1, 0, interpolation = 0)
  o_chr <- transform_affine(x, 0, c(0, 1), 1, 0, interpolation = "nearest")
  expect_equal_to_r(o_chr, as_array(o_int))
})

test_that("affine deprecated arguments still work", {

  x <- torch_eye(8)$view(c(1, 1, 8, 8))

  old_resample <- expect_warning(
    transform_affine(x, 0, c(0, 1), 1, 0, resample = 0, fill = 0),
    "resample"
  )

  old_fillcolor <- expect_warning(
    transform_affine(x, 0, c(0, 1), 1, 0, interpolation = 0, fillcolor = 0),
    "fillcolor"
  )

  new <- transform_affine(x, 0, c(0, 1), 1, 0, interpolation = 0, fill = 0)

  expect_equal_to_r(old_resample, as_array(new))
  expect_equal_to_r(old_fillcolor, as_array(new))

})

test_that("affine validates positive scale", {

  x <- torch_eye(8)$view(c(1, 1, 8, 8))
  expect_error(transform_affine(x, 0, c(0, 0), 0, 0), "positive")
  expect_error(transform_affine(x, 0, c(0, 0), 1, 0, interpolation = "bicubic"), "Unsupported interpolation mode")

})

test_that("linear transformation", {

  c <- 3
  h <- 24
  w <- 32

  tensor <- torch::torch_randn(c, h, w)
  matrix <- torch::torch_rand(c * h * w, c * h * w)
  mean_vector <- torch::torch_rand(c * h * w)

  out <- transform_linear_transformation(tensor, matrix, mean_vector)

  expect_equal(dim(out), c(3, 24, 32))

  outb <- transform_linear_transformation(tensor$unsqueeze(1), matrix, mean_vector)

  expect_equal(dim(outb), c(1, 3, 24, 32))
})

test_that("adjust hue", {

  hue_factor <- c(-0.45, -0.25, 0.0, 0.25, 0.45)
  x <- torch::torch_rand(3, 24, 32)

  for (f in hue_factor) {
    out <- transform_adjust_hue(x, f)
    expect_equal(dim(out), dim(x))
  }

  for (f in hue_factor) {
    out <- transform_adjust_hue(x$unsqueeze(1), f)
    expect_equal(dim(out), dim(x$unsqueeze(1)))
  }

})

test_that("grayscale", {

  x <- torch::torch_rand(3, 24, 32)
  out <- transform_grayscale(x, 3)
  expect_equal(dim(out), dim(x))
  expect_equal(dim(out)[1], 3)

  out <- transform_grayscale(x, 1)
  expect_equal(dim(out)[2:3], dim(x)[2:3])
  expect_equal(dim(out)[1], 1)

  outb <- transform_grayscale(x$unsqueeze(1), 3)
  expect_equal(dim(outb), dim(x$unsqueeze(1)))
  expect_equal(dim(outb)[2], 3)

  outb <- transform_grayscale(x$unsqueeze(1), 1)
  expect_equal(dim(outb)[3:4], dim(x)[2:3])
  expect_equal(dim(outb)[2], 1)

})

test_that("random grayscale", {

  tensor <- torch::torch_rand(3, 24, 32)
  for (p in seq(0, 1, length.out = 10)) {
    out <- transform_random_grayscale(tensor, p)
    expect_equal(dim(out), dim(tensor))
  }

  for (p in seq(0, 1, length.out = 10)) {
    outb <- transform_random_grayscale(tensor$unsqueeze(1), p)
    expect_equal(dim(outb), dim(tensor$unsqueeze(1)))
  }

})


test_that("random vertical flip", {

  tensor <- torch::torch_randn(3, 24, 32)

  for (i in 1:10) {
    out <- transform_random_vertical_flip(tensor)
    expect_equal(dim(out), dim(tensor))
  }
  for (p in seq(0, 1, length.out = 10)) {
    out <- transform_random_vertical_flip(tensor, p)
    expect_equal(dim(out), dim(tensor))
  }

  for (i in 1:10) {
    outb <- transform_random_vertical_flip(tensor$unsqueeze(1))
    expect_equal(dim(outb), dim(tensor$unsqueeze(1)))
  }
  for (p in seq(0, 1, length.out = 10)) {
    outb <- transform_random_vertical_flip(tensor$unsqueeze(1), p)
    expect_equal(dim(outb), dim(tensor$unsqueeze(1)))
  }
})


test_that("random rotation works", {

  x <- torch::torch_tensor(array(1, dim = c(3, 200, 200)))

  # Transforms
  rotate <- function(img) transform_random_rotation(img, 20)

  expect_error(rotate(x), regexp = NA)
  expect_error(rotate(x$unsqueeze(1)), regexp = NA)


})

test_that("random choice transform works", {

  # Example Image
  x <- torch_ones(c(3, 200, 200))

  # Transforms
  color_transform <- function(img) transform_color_jitter(
    img, brightness = 0.5, contrast = 0.5, saturation = 0.5, hue = 0.5
  )
  resize_crop <- function(img) transform_random_resized_crop(img, size = c(200, 200))
  hflip <- function(img) transform_random_horizontal_flip(img)
  vflip <- function(img) transform_random_vertical_flip(img)
  rotate <- function(img) transform_random_rotation(img, 20)
  identity <- function(img) img

  # Select a Random Transform to Apply
  expect_error(regexp = NA, {
    transform_random_choice(
      x,
      list(
        color_transform,
        resize_crop,
        hflip,
        vflip,
        rotate,
        identity
      )
    )
  })

  expect_error(regexp = NA, {
    transform_random_choice(
      x$unsqueeze(1),
      list(
        color_transform,
        resize_crop,
        hflip,
        vflip,
        rotate,
        identity
      )
    )
  })

})

context("dataset-fgvc")

t <- withr::local_tempdir()

test_that("tests for the FGVC-Aircraft dataset", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_error(
    fgvc_aircraft_dataset(root = tempfile(), split = "train", annotation_level = "variant", download = FALSE),
    class = "runtime_error"
  )

  resize_collate_fn <- function(batch) {
    xs <- lapply(batch, function(item) {
      torchvision::transform_resize(item$x, c(224, 224))
    })
    xs <- torch::torch_stack(xs)
    ys <- torch::torch_tensor(sapply(batch, function(item) item$y), dtype = torch::torch_long())
    list(x = xs, y = ys)
  }
  fgvc <- fgvc_aircraft_dataset(root = t, transform = transform_to_tensor, download = TRUE)
  dl <- torch::dataloader(dataset = fgvc,batch_size = 2,collate_fn = resize_collate_fn)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_tensor_shape(batch$x,c(2,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor(batch$y)
  expect_tensor_shape(batch$y,2)
  expect_tensor_dtype(batch$y,torch_long())
  expect_equal_to_r(batch$y[1],86)
  expect_equal_to_r(batch$y[2],42)

  fgvc <- fgvc_aircraft_dataset(root = t, split = "trainval", annotation_level = "variant")
  expect_length(fgvc, 6667)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 1854900)
  expect_type(item$y, "integer")
  expect_equal(item$y, 56)

  fgvc <- fgvc_aircraft_dataset(root = t, split = "test", annotation_level = "variant")
  expect_length(fgvc, 3333)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 2155620)
  expect_type(item$y, "integer")
  expect_equal(item$y, 3)

  fgvc <- fgvc_aircraft_dataset(root = t, split = "trainval", annotation_level = "family")
  expect_length(fgvc, 6667)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 1854900)
  expect_type(item$y, "integer")
  expect_equal(item$y, 32)

  fgvc <- fgvc_aircraft_dataset(root = t, split = "test", annotation_level = "family")
  expect_length(fgvc, 3333)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 2155620)
  expect_type(item$y, "integer")
  expect_equal(item$y, 16)

  fgvc <- fgvc_aircraft_dataset(root = t, split = "trainval", annotation_level = "manufacturer")
  expect_length(fgvc, 6667)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 1854900)
  expect_type(item$y, "integer")
  expect_equal(item$y, 13)

  fgvc <- fgvc_aircraft_dataset(root = t, split = "test", annotation_level = "manufacturer")
  expect_length(fgvc, 3333)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 2155620)
  expect_type(item$y, "integer")
  expect_equal(item$y, 5)

  fgvc <- fgvc_aircraft_dataset(root = t, split = "test", annotation_level = "all")
  expect_length(fgvc, 3333)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 2155620)
  expect_type(item$y, "integer")
  expect_equal(fgvc$classes$manufacturer[item$y[1]],"Boeing")
  expect_equal(fgvc$classes$family[item$y[2]],"Boeing 737")
  expect_equal(fgvc$classes$variant[item$y[3]],"737-200")

  fgvc <- fgvc_aircraft_dataset(root = t, split = "trainval", annotation_level = "all")
  expect_length(fgvc, 6667)
  item <- fgvc[1]
  expect_named(item, c("x", "y"))
  expect_type(item$x, "double")
  expect_length(item$x, 1854900)
  expect_type(item$y, "integer")
  expect_equal(fgvc$classes$manufacturer[item$y[1]],"Douglas Aircraft Company")
  expect_equal(fgvc$classes$family[item$y[2]],"DC-8")
  expect_equal(fgvc$classes$variant[item$y[3]],"DC-8")

    resize_collate_fn <- function(batch) {
    xs <- lapply(batch, function(item) {
      torchvision::transform_resize(item$x, c(224, 224))
    })
    xs <- torch::torch_stack(xs)
    ys <- torch::torch_tensor(sapply(batch, function(item) item$y), dtype = torch::torch_long())
    list(x = xs, y = ys)
  }
  fgvc <- fgvc_aircraft_dataset(root = t, annotation_level = "all", transform = transform_to_tensor)
  dl <- torch::dataloader(dataset = fgvc,batch_size = 2,collate_fn = resize_collate_fn)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_tensor_shape(batch$x,c(2,3,224,224))
  expect_tensor_dtype(batch$x,torch_float())
  expect_tensor(batch$y)
  expect_tensor_shape(batch$y,c(3,2))
  expect_tensor_dtype(batch$y,torch_long())
  expect_equal_to_r(batch$y[1],c(22,5))
  expect_equal_to_r(batch$y[2],c(58,14))
  expect_equal_to_r(batch$y[3],c(86,42))

})

context("dataset-plankton")


test_that("whoi_small_plankton_dataset val downloads correctly", {
  skip_on_cran()
  skip_if_not_installed("torch")

  expect_error(
    whoi_small_plankton_dataset(split = "test", download = FALSE),
    "Dataset not found. You can use `download = TRUE`",
    label = "Dataset should fail if not previously downloaded"
  )

  expect_no_error(
    val_ds <- whoi_small_plankton_dataset(split = "val", download = TRUE, transform = transform_to_tensor)
  )
  # Validation dataset should have exactly 5799 samples
  expect_equal(val_ds$.length(), 5799)

  first_item <- val_ds[1]
  expect_tensor_shape(first_item$x, c(1,145, 230))
  # classification of the first item is "47: Leegaardiella_ovalis"
  expect_equal(first_item$y, 47L)
  expect_equal(val_ds$classes[first_item$y], "Leegaardiella_ovalis")

})


test_that("whoi_small_plankton_dataset downloads correctly whatever the split", {
  skip_on_cran()
  skip_if_not_installed("torch")
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_error(
    whoi_small_plankton_dataset(split = "test", download = FALSE),
    "Dataset not found. You can use `download = TRUE`",
    label = "Dataset should fail if not previously downloaded"
  )

  expect_no_error(
    train_ds <- whoi_small_plankton_dataset(split = "train", download = TRUE)
  )

  expect_is(train_ds, "dataset", "train should be a dataset")
  # Train dataset should have exactly 40599 samples
  expect_equal(train_ds$.length(), 40599)

  expect_no_error(
    test_ds <- whoi_small_plankton_dataset(split = "test", download = TRUE)
  )
  # Test dataset should have exactly 11601 samples
  expect_equal(test_ds$.length(), 11601)

})


test_that("whoi_small_coralnet_dataset downloads correctly whatever the split", {
  skip_on_cran()
  skip_if_not_installed("torch")
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_error(
    whoi_small_coralnet_dataset(split = "test", download = FALSE),
    "Dataset not found. You can use `download = TRUE`",
    label = "Dataset should fail if not previously downloaded"
  )

  expect_no_error(
    train_ds <- whoi_small_coralnet_dataset(split = "train", download = TRUE)
  )

  expect_is(train_ds, "dataset", "train should be a dataset")
  # Train dataset should have exactly 314 samples
  expect_equal(train_ds$.length(), 314)

  expect_no_error(
    val_ds <- whoi_small_coralnet_dataset(split = "val", download = TRUE, transform = transform_to_tensor)
  )
  # Validation dataset should have exactly 45 samples
  expect_equal(val_ds$.length(), 45)

  first_item <- val_ds[1]
  expect_tensor_shape(first_item$x, c(3, 3000, 4000))
  # classification of the first item is "1: diploria_labrinthyformis"
  expect_equal(first_item$y, 1L)
  expect_equal(val_ds$classes[first_item$y], "diploria_labrinthyformis")

  expect_no_error(
    test_ds <- whoi_small_coralnet_dataset(split = "test", download = TRUE)
  )
  # Test dataset should have exactly 91 samples
  expect_equal(test_ds$.length(), 91)

})

test_that("whoi_small_plankton_dataset derivatives download and prepare correctly", {
  skip_on_cran()
  skip_if_not_installed("torch")

  val_ds <- whoi_small_plankton_dataset(
    split = "val", download = TRUE,
    transform = . %>% transform_to_tensor() %>% transform_resize(size = c(150, 300))
    )


  dl <- torch::dataloader(val_ds, batch_size = 10)
  # 5799 turns into 580 batches of 10
  expect_length(dl, 580)
  iter <- dataloader_make_iter(dl)
  expect_no_error(
    i <- dataloader_next(iter)
  )

  # Check shape, dtype, and values on X
  expect_tensor_shape(i[[1]], c(10, 1, 150, 300))
  expect_tensor_dtype(i[[1]], torch_float())
  expect_true((torch_max(i[[1]]) <= 1)$item())
  # Check shape, dtype and names on y
  expect_length(i[[2]], 10)
  expect_named(i, c("x", "y"))})




test_that("whoi_plankton_dataset downloads correctly whatever the split", {
  skip_on_cran()
  skip_if_not_installed("torch")
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")


  expect_error(
    whoi_plankton_dataset(split = "test", download = FALSE),
    "Dataset not found. You can use `download = TRUE`",
    label = "Dataset should fail if not previously downloaded"
  )

  expect_no_error(
    train_ds <- whoi_plankton_dataset(split = "train", download = TRUE)
  )

  expect_is(train_ds, "dataset", "train should be a dataset")
  # Train dataset should have exactly 669806 samples
  expect_equal(train_ds$.length(), 669806)

  val_ds <- whoi_plankton_dataset(split = "val", download = TRUE, transform = transform_to_tensor)
  # Validation dataset should have exactly 95686 samples
  expect_equal(val_ds$.length(), 95686)

  first_item <- val_ds[1]
  expect_tensor_shape(first_item$x, c(1,45, 388))
  # classification of the first item is "48: Leptocylindrus"
  expect_equal(first_item$y, 48L)
  expect_equal(val_ds$classes[first_item$y], "Leptocylindrus")

  test_ds <- whoi_plankton_dataset(split = "test", download = TRUE)
  # Test dataset should have exactly 191375 samples
  expect_equal(test_ds$.length(), 191375)

})

test_that("dataloader from whoi_plankton_dataset gets torch tensors", {
  skip_on_cran()
  skip_if_not_installed("torch")
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- whoi_plankton_dataset(
    split = "val", download = TRUE,
    transform = . %>% transform_to_tensor() %>% transform_resize(size = c(150, 300))
  )
  dl <- torch::dataloader(ds, batch_size = 10)
  # 956,9k turns into 9569 batches of 10
  expect_length(dl, 9569)
  iter <- dataloader_make_iter(dl)
  expect_no_error(
    i <- dataloader_next(iter)
  )
  # Check shape, dtype, and values on X
  expect_tensor_shape(i[[1]], c(10, 1, 150, 300))
  expect_tensor_dtype(i[[1]], torch_float())
  expect_true((torch_max(i[[1]]) <= 1)$item())
  # Check shape, dtype and names on y
  expect_length(i[[2]],10)
  expect_named(i, c("x", "y"))})

#' Fully Convolutional Network for Semantic Segmentation
#'
#' Constructs an FCN (Fully Convolutional Network) model for semantic image segmentation,
#' based on a ResNet backbone as described in
#' \href{https://arxiv.org/abs/1411.4038}{Fully Convolutional Networks for Semantic Segmentation}.
#'
#' The 21 output classes follow the PASCAL VOC convention:
#' `background`, `aeroplane`, `bicycle`, `bird`, `boat`,
#' `bottle`, `bus`, `car`, `cat`, `chair`,
#' `cow`, `dining table`, `dog`, `horse`, `motorbike`,
#' `person`, `potted plant`, `sheep`, `sofa`, `train`,
#' `tv/monitor`.
#'
#' @inheritParams model_resnet18
#' @param num_classes Number of output classes. Default: 21.
#' @param aux_loss If TRUE, includes the auxiliary classifier. If NULL, defaults to TRUE when `pretrained = TRUE`.
#' @param pretrained_backbone If TRUE, uses a backbone pre-trained on ImageNet.
#' @param ... Additional arguments passed to the backbone implementation.
#'
#' @details Pretrained weights require \code{num_classes = 21}.
#'
#' @return An `nn_module` representing the FCN model.
#'
#' @family semantic_segmentation_model
#' @name model_fcn_resnet
#' @rdname model_fcn_resnet
#'
#' @examples
#' \dontrun{
#' library(magrittr)
#' norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants, see
#' # https://pytorch.org/vision/stable/models.html
#' norm_std  <- c(0.229, 0.224, 0.225)
#' img_url <- "https://en.wikipedia.org/wiki/Special:FilePath/Felis_catus-cat_on_snow.jpg"
#' img <- base_loader(img_url)
#'
#' input <- img %>%
#'  transform_to_tensor() %>%
#'  transform_resize(c(520, 520)) %>%
#'  transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)
#'
#' model <- model_fcn_resnet50(pretrained = TRUE)
#' model$eval()
#' output <- model(batch)
#'
#' # visualize the result
#' # `draw_segmentation_masks()` turns the torch_float output into a boolean mask internaly:
#' segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
#' tensor_image_display(segmented)
#'
#'
#' model <- model_fcn_resnet101(pretrained = TRUE)
#' model$eval()
#' output <- model(batch)
#'
#' # visualize the result
#' segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
#' tensor_image_display(segmented)
#' }
NULL
fcn_model_urls <- list(
  fcn_resnet50_coco = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/fcn_resnet50_coco.pth",
    "c79a7e2675d73817cfe6ba383be6ca7d", "135 MB"),
  fcn_resnet101_coco = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/fcn_resnet101_coco.pth",
    "369109597fa68546df1231ae2fe0f66f", "207 MB")
)

fcn_head <- function(in_channels, channels, num_classes) {
  torch::nn_sequential(
    torch::nn_conv2d(in_channels, channels, kernel_size = 3, padding = 1, bias = FALSE),
    torch::nn_batch_norm2d(channels),
    torch::nn_relu(inplace = TRUE),
    torch::nn_dropout(0.1),
    torch::nn_conv2d(channels, num_classes, kernel_size = 1)
  )
}

fcn_backbone <- torch::nn_module(
  "fcn_backbone",
  initialize = function(block, layers, replace_stride_with_dilation = c(FALSE, FALSE, FALSE), ...) {
    args <- list(...)
    args$pretrained <- NULL

    resnet_model <- do.call(resnet, c(list(block = block, layers = layers,
                                           replace_stride_with_dilation = replace_stride_with_dilation),
                                      args))

    self$conv1 <- resnet_model$conv1
    self$bn1 <- resnet_model$bn1
    self$relu <- resnet_model$relu
    self$maxpool <- resnet_model$maxpool

    self$layer1 <- resnet_model$layer1
    self$layer2 <- resnet_model$layer2
    self$layer3 <- resnet_model$layer3
    self$layer4 <- resnet_model$layer4

    self$out_channels <- 512 * block$public_fields$expansion
    self$aux_channels <- 256 * block$public_fields$expansion
  },

  forward = function(x) {
    x <- self$conv1(x)
    x <- self$bn1(x)
    x <- self$relu(x)
    x <- self$maxpool(x)

    x <- self$layer1(x)
    x <- self$layer2(x)
    x <- self$layer3(x)
    aux <- x
    x <- self$layer4(x)
    list(out = x, aux = aux)
  }
)

fcn <- torch::nn_module(
  "fcn",
  initialize = function(backbone, classifier, aux_classifier = NULL) {
    self$backbone <- backbone
    self$classifier <- classifier
    self$aux_classifier <- aux_classifier
  },

  forward = function(x) {
    input_shape <- x$shape[3:4]
    features <- self$backbone(x)
    x <- features$out
    x <- self$classifier(x)
    x <- torch::nnf_interpolate(x, size = input_shape, mode = "bilinear", align_corners = FALSE)
    result <- list(out = x)
    if (!is.null(self$aux_classifier)) {
      x_aux <- features$aux
      x_aux <- self$aux_classifier(x_aux)
      x_aux <- torch::nnf_interpolate(x_aux, size = input_shape, mode = "bilinear", align_corners = FALSE)
      result$aux <- x_aux
    }
    result
  }
)

#' @rdname model_fcn_resnet
#' @export
model_fcn_resnet50 <- function(pretrained = FALSE, progress = TRUE, num_classes = 21,
                               aux_loss = NULL, pretrained_backbone = TRUE, ...) {
  if (is.null(aux_loss)) aux_loss <- pretrained

  if (pretrained && num_classes != 21)
    cli_abort("Pretrained weights require num_classes = 21.")

  backbone <- fcn_backbone(bottleneck, c(3, 4, 6, 3),
                           replace_stride_with_dilation = c(FALSE, FALSE, FALSE),
                           ...)

  if (pretrained_backbone) {
    state_dict_path <- download_and_cache(resnet_model_urls[["resnet50"]][1], prefix = "fcn")
    state_dict <- torch::load_state_dict(state_dict_path)
    backbone$load_state_dict(state_dict, strict = FALSE)
  }

  classifier <- fcn_head(backbone$out_channels, 512, num_classes)
  aux_classifier <- if (aux_loss) fcn_head(backbone$aux_channels, 256, num_classes) else NULL

  model <- fcn(backbone, classifier, aux_classifier)

  if (pretrained) {
    r <- fcn_model_urls$fcn_resnet50_coco
    name <- "model_fcn_resnet50"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "fcn")
    if (!tools::md5sum(state_dict_path) == r[2]) {
        runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }
    state_dict <- torch::load_state_dict(state_dict_path)


    if (num_classes != 21) {
      state_dict <- state_dict[!grepl("^classifier\\.4\\.", names(state_dict))]
      state_dict <- state_dict[!grepl("^aux_classifier\\.4\\.", names(state_dict))]
    }

    strict_loading <- num_classes == 21 && aux_loss
    model$load_state_dict(state_dict, strict = strict_loading)
  }

  model
}


#' @rdname model_fcn_resnet
#' @export
model_fcn_resnet101 <- function(pretrained = FALSE, progress = TRUE, num_classes = 21,
                                aux_loss = NULL, pretrained_backbone = TRUE, ...) {
  if (is.null(aux_loss)) aux_loss <- pretrained

  if (pretrained && num_classes != 21)
    cli_abort("Pretrained weights require num_classes = 21.")

  backbone <- fcn_backbone(bottleneck, c(3, 4, 23, 3),
                           replace_stride_with_dilation = c(FALSE, FALSE, FALSE),
                           ...)

  if (pretrained_backbone) {
    state_dict_path <- download_and_cache(resnet_model_urls[["resnet101"]][1], prefix = "fcn")
    state_dict <- torch::load_state_dict(state_dict_path)
    backbone$load_state_dict(state_dict, strict = FALSE)
  }

  classifier <- fcn_head(backbone$out_channels, 512, num_classes)
  aux_classifier <- if (aux_loss) fcn_head(backbone$aux_channels, 256, num_classes) else NULL

  model <- fcn(backbone, classifier, aux_classifier)

  if (pretrained) {
    r <- fcn_model_urls$fcn_resnet101_coco
    name <- "model_fcn_resnet101"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "fcn")
    if (!tools::md5sum(state_dict_path) == r[2]) {
        runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }
    state_dict <- torch::load_state_dict(state_dict_path)


    if (num_classes != 21) {
      state_dict <- state_dict[!grepl("^classifier\\.4\\.", names(state_dict))]
      state_dict <- state_dict[!grepl("^aux_classifier\\.4\\.", names(state_dict))]
    }

    strict_loading <- num_classes == 21 && aux_loss
    model$load_state_dict(state_dict, strict = strict_loading)
  }

  model
}

#' @export
`transform_to_tensor.magick-image` <- function(img) {
  img <- as.integer(magick::image_data(img, channels = "rgb"))
  img <- torch::torch_tensor(img)$permute(c(3,1,2))
  img <- img$to(dtype = torch::torch_float32())
  img <- img$contiguous()
  img <- img$div(255)

  img
}

#' @export
`transform_resize.magick-image` <- function(img, size, interpolation = 2) {

  interpolation_modes <- c(
    "0" = "Pint", # nearest,
    "2" = "Triangle", # bilinear
    "3" = "Catrom" # bicubic
  )

  if (is.numeric(interpolation))
    interpolation <- interpolation_modes[names(interpolation_modes) == interpolation]

  if (length(size) == 1) {

    w <- magick::image_info(img)$width
    h <- magick::image_info(img)$height

    if (w < h)
      size <- paste0(size, "x")
    else
      size <- paste0("x", size)

  } else {
    size <- paste0(paste0(size, collapse = "x"), "!")
  }

  magick::image_resize(img, geometry = size, filter = interpolation)
}

#' @export
`transform_crop.magick-image` <- function(img, top, left, height, width) {
  magick::image_crop(
    img,
    paste0(height, "x", width, "+", left, "+", top)
  )
}

#' @export
`transform_hflip.magick-image` <- function(img) {
  magick::image_flip(img)
}

# Utils -------------------------------------------------------------------

#' @method get_image_size magick-image
#' @export
`get_image_size.magick-image` <- function(img) {
  info <- magick::image_info(img)
  c(info$width, info$height)
}

#' @include models-faster_rcnn.R
#' @importFrom torch nn_conv_transpose2d torch_int32 torch_float torch_log2 torch_nonzero
NULL

# Mask R-CNN Implementation
# Instance Segmentation model extending Faster R-CNN with mask prediction

# Mask Head Module - Predicts segmentation masks for detected objects
mask_head_module <- torch::nn_module(
    "mask_head",
    initialize = function() {
      # 4 convolutional layers for feature extraction
      self$mask_fcn1 <- nn_conv2d(256, 256, kernel_size = 3, padding = 1)
      self$mask_fcn2 <- nn_conv2d(256, 256, kernel_size = 3, padding = 1)
      self$mask_fcn3 <- nn_conv2d(256, 256, kernel_size = 3, padding = 1)
      self$mask_fcn4 <- nn_conv2d(256, 256, kernel_size = 3, padding = 1)

    },
    forward = function(x) {
      x <- nnf_relu(self$mask_fcn1(x))
      x <- nnf_relu(self$mask_fcn2(x))
      x <- nnf_relu(self$mask_fcn3(x))
      nnf_relu(self$mask_fcn4(x))
    }
)

# Mask RCNN predictor - Predicts segmentation masks for detected objects
mask_rcnn_predictor <- torch::nn_module(
    "mask_rcnn_predictor",
    initialize = function(num_classes = 90) {
      # num_classes excludes background, but predictor needs background + all classes
      num_classes_with_bg <- num_classes + 1L

      # Deconvolution layer to upsample from 14x14 to 28x28
      self$conv5_mask <- nn_conv_transpose2d(256, 256, kernel_size = 2, stride = 2)

      # Final 1x1 conv for class-specific mask logits
      self$mask_fcn_logits <- nn_conv2d(256, num_classes_with_bg, kernel_size = 1)
    },
    forward = function(x) {
      x <- nnf_relu(self$conv5_mask(x))
      self$mask_fcn_logits(x)
    }
)

# Mask Head Module V2 - With batch normalization
mask_head_module_v2 <- torch::nn_module(
    "mask_head_v2",
    initialize = function(num_classes = 90) {
      # num_classes excludes background, but predictor needs background + all classes
      num_classes_with_bg <- num_classes + 1L

      # Convolutional blocks with batch normalization
      conv_block <- function() {
        nn_sequential(
          nn_conv2d(256, 256, kernel_size = 3, padding = 1, bias = FALSE),
          nn_batch_norm2d(256),
          nn_relu()
        )
      }

      self$mask_head.0 <- conv_block()
      self$mask_head.1 <- conv_block()
      self$mask_head.2 <- conv_block()
      self$mask_head.3 <- conv_block()

      # Deconvolution without batch norm
      self$mask_predictor.conv5_mask <- nn_sequential(
        nn_conv_transpose2d(256, 256, kernel_size = 2, stride = 2),
        nn_relu()
      )

      # Final 1x1 conv for class-specific mask logits
      self$mask_predictor.mask_fcn_logits <- nn_conv2d(256, num_classes_with_bg, kernel_size = 1)
    },
    forward = function(x) {
      x <- self$mask_head.0(x)
      x <- self$mask_head.1(x)
      x <- self$mask_head.2(x)
      x <- self$mask_head.3(x)
      x <- self$mask_predictor.conv5_mask(x)
      self$mask_predictor.mask_fcn_logits(x)
    }
)


#' ROI Align for Mask Prediction
#'
#' Extracts fixed-size feature maps from regions of interest for mask prediction.
#' Returns a 4D tensor suitable for the mask head (unlike roi_align which flattens).
#'
#' ROI Align for mask prediction – single FPN level
#'
#' This uses grid_sample for bilinear interpolation.
#' It **expects** the proposals in *image* coordinates
#' and applies the supplied `spatial_scale` before the ROI‑Align.
#'
#' @param feature_map   Tensor of shape (1, C, H, W) – one FPN level.
#' @param proposals     Tensor of shape (N, 4) with boxes (x1, y1, x2, y2) in
#'                       the original image coordinate system.
#' @param output_size   integer vector of length 2, e.g. c(14L, 14L).
#' @param spatial_scale Scaling factor between the original image and the
#'                       feature map.  For the first FPN level (stride 4) use
#'                       1/4, for the second level (stride 8) use 1/8, …
#' @param sampling_ratio Integer – number of sampling points per output
#'                       element (default = 2, the same as Torchvision).
#' @param aligned       Logical – if TRUE uses the “aligned” mode (default FALSE).
#'
#' @return Tensor of shape (N, C, output_size\[1\], output_size\[2\]).
#' @noRd
roi_align_masks <- function(feature_map,
                            proposals,
                            output_size = c(14L, 14L),
                            spatial_scale = 1.0,
                            sampling_ratio = 2L,
                            aligned = FALSE) {

  # Scale the boxes to the feature map resolution
  boxes_scaled <- proposals$to(dtype = torch_float())$mul(spatial_scale)

  num_rois <- boxes_scaled$size(1)
  if (num_rois == 0) {
    return(torch_empty(c(0, feature_map$size(2), output_size[1], output_size[2]),
                       device = feature_map$device))
  }

  channels <- feature_map$size(2)
  h_feat <- feature_map$size(3)
  w_feat <- feature_map$size(4)

  # Normalize coordinates to match grid_sample [-1 to 1]
  x1 <- (boxes_scaled[, 1] / (w_feat - 1) * 2) - 1
  y1 <- (boxes_scaled[, 2] / (h_feat - 1) * 2) - 1
  x2 <- (boxes_scaled[, 3] / (w_feat - 1) * 2) - 1
  y2 <- (boxes_scaled[, 4] / (h_feat - 1) * 2) - 1

  # Create a grid of output_size
  grid_y <- torch_linspace(0, 1, output_size[1], device = feature_map$device)
  grid_x <- torch_linspace(0, 1, output_size[2], device = feature_map$device)

  # Meshgrid to get relative coordinates
  grids <- torch_meshgrid(list(grid_y, grid_x), indexing = "ij")
  rel_y <- grids[[1]]
  rel_x <- grids[[2]]

  # Linear interpolation for each ROI [N, output_size[1], output_size[2]]
  sampling_x <- x1$view(c(-1, 1, 1)) + rel_x$view(c(1, output_size[1], output_size[2])) * (x2 - x1)$view(c(-1, 1, 1))
  sampling_y <- y1$view(c(-1, 1, 1)) + rel_y$view(c(1, output_size[1], output_size[2])) * (y2 - y1)$view(c(-1, 1, 1))

  # Clamp grid to [-1, 1] to avoid needing padding (especially for MPS compatibility)
  sampling_x <- sampling_x$clamp(-1, 1)
  sampling_y <- sampling_y$clamp(-1, 1)

  # Concat to get a grid of [N, output_size[1], output_size[2], 2]
  grid <- torch_stack(list(sampling_x, sampling_y), dim = -1)

  # Bilinear sampling
  # Feature map is (1, C, H, W), expand to (N, C, H, W) for all ROIs
  input_expanded <- feature_map$squeeze(1)$unsqueeze(1)$expand(c(num_rois, channels, h_feat, w_feat))

  pooled_features <- nnf_grid_sample(
    input_expanded,
    grid,
    mode = "bilinear",
    padding_mode = "zeros",
    align_corners = FALSE
  )

  # Return [N, C, output_size[1], output_size[2]]
  pooled_features
}


#' Multi‑scale ROI Align for Mask‑RCNN (FPN)
#'
#'
#' @param feature_maps   List of 4 tensors, each of shape (1, C, H_l, W_l)
#'                       – the four FPN levels P2, P3, P4, P5.
#' @param proposals      Tensor (N, 4) – boxes in image coordinates.
#' @param output_size    Integer vector, default c(14L, 14L).
#' @param sampling_ratio Sampling ratio passed to the low‑level op (default 2L).
#' @param aligned        Whether to use the aligned ROI‑Align mode (default FALSE).
#'
#' @return Tensor (N, C, out_h, out_w) – pooled mask features.
#' @noRd
roi_align_masks_fpn <- function(feature_maps,
                                proposals,
                                output_size = c(14L, 14L),
                                sampling_ratio = 2L,
                                aligned = FALSE) {
  #  Compute the level for each ROI (FPN paper Eq. (1))
  # width and height in original image space
  widths  <- proposals[, 3] - proposals[, 1]
  heights <- proposals[, 4] - proposals[, 2]

  # Guard against degenerate boxes (avoid log2(0))
  widths[widths <= 0]   <- 1
  heights[heights <= 0] <- 1

  # sqrt(area) / 224  (224 is the reference size in the paper)
  area_sqrt <- (widths * heights)$sqrt()
  target_level <- (torch_log2(area_sqrt / 224) + 4)$floor()   # level in {2,3,4,5}
  target_level <- torch_clamp(target_level, min = 2, max = 5) # clamp to existing levels

  # Convert from the level {2,3,4,5} to an index in feature_maps list (1‑based)
  level_idx <- (target_level - 2L)$to(dtype = torch_long()) + 1L   # 1,2,3,4

  #  Allocate output tensor
  n_boxes   <- proposals$size(1)
  c_channels <- feature_maps[[1]]$size(2)    # same number of channels for all levels
  out       <- torch_zeros(c(n_boxes, c_channels,
                             output_size[1], output_size[2]),
                           dtype = feature_maps[[1]]$dtype,
                           device = feature_maps[[1]]$device)

  #  Loop over the four levels and pool the ROIs belonging to it
  spatial_scales <- list(1/4, 1/8, 1/16, 1/32)   # stride of each level

  for (lvl in seq_along(feature_maps)) {
    # mask of boxes that belong to the current level
    lvl_mask <- level_idx$eq(lvl)$squeeze()
    idx      <- torch_nonzero(lvl_mask)$squeeze()   # indices

    # Check if there are any boxes at this level
    if (idx$numel() == 0) next

    # Ensure idx is 1D for consistent indexing
    if (idx$dim() == 0) idx <- idx$unsqueeze(1)  # scalar -> (1,)

    # pick the subset of proposals that belong to this level
    rois_lvl <- proposals[idx, , drop = FALSE]  # idx is already 1-based

    # call the *single‑scale* wrapper (it adds the batch column internally)
    pooled_lvl <- roi_align_masks(
      feature_map   = feature_maps[[lvl]],
      proposals     = rois_lvl,
      output_size   = output_size,
      spatial_scale = spatial_scales[[lvl]],
      sampling_ratio = sampling_ratio,
      aligned        = aligned
    )   # shape (M, C, out_h, out_w)

    # write the result back into the pre‑allocated output tensor
    out[idx, , , ] <- pooled_lvl  # idx is already 1-based
  }

  out
}

# Mask R-CNN Model - Extends Faster R-CNN with mask prediction
maskrcnn_model <- torch::nn_module(
  "maskrcnn_model",
    initialize = function(backbone, num_classes,
                          score_thresh = 0.05,
                          nms_thresh = 0.5,
                          detections_per_img = 100) {
      self$backbone <- backbone
      self$num_classes <- num_classes
      # Store configurable detection parameters
      self$score_thresh <- score_thresh
      self$nms_thresh <- nms_thresh
      self$detections_per_img <- detections_per_img

      # RPN (Region Proposal Network)
      self$rpn <- torch::nn_module(
        initialize = function() {
          self$head <- rpn_head(in_channels = backbone$out_channels)
        },
        forward = function(features) {
          self$head(features)
        }
      )()

      # ROI heads for box prediction
      self$roi_heads <- roi_heads_module(num_classes = num_classes)

      # Mask head without mask predictor
      self$mask_head <- mask_head_module()

      # Mask predictor for mask prediction
      self$mask_predictor <- mask_rcnn_predictor(num_classes = num_classes)
    },

    forward = function(images) {
      features <- self$backbone(images)
      rpn_out <- self$rpn(features)

      batch_size <- images$shape[1]
      image_size <- images$shape[3:4]
      final_results <- list()

      for (b in seq_len(batch_size)) {

        props <- generate_proposals(features, rpn_out, image_size, c(4, 8, 16, 32),
                                    batch_idx = b, score_thresh = 0,
                                    nms_thresh = self$nms_thresh)

        if (props$proposals$shape[1] == 0) {
          empty <- list(
            boxes = torch::torch_empty(c(0, 4)),
            labels = torch::torch_empty(c(0), dtype = torch::torch_long()),
            scores = torch::torch_empty(c(0)),
            masks = torch::torch_empty(c(0, 28, 28))
          )
          final_results[[b]] <- empty
          next
        }

        # Limit proposals to avoid slow ROI pooling
        max_proposals <- 1000
        if (props$proposals$shape[1] > max_proposals) {
          props$proposals <- props$proposals[1:max_proposals, ]
        }

        # Get ROI head predictions
        roi_out <- self$roi_heads(features, props$proposals, batch_idx = b)

        # Postprocess detections
        det_result <- postprocess_detections(
          class_logits = roi_out$scores,
          box_regression = roi_out$boxes,
          proposals = props$proposals,
          image_size = image_size,
          num_classes = self$num_classes,
          score_thresh = self$score_thresh,
          nms_thresh = self$nms_thresh,
          detections_per_img = self$detections_per_img
        )

        final_boxes <- det_result$boxes
        final_labels <- det_result$labels
        final_scores <- det_result$scores

        # Predict masks for detected objects
        if (final_boxes$shape[1] > 0) {
          # For mask prediction, we need to map final boxes back to proposals
          # Since postprocess_detections decodes boxes, we need the original proposals
          # that correspond to the kept detections. We'll use the final_boxes directly.

          mask_features <- roi_align_masks_fpn(
            feature_maps = features,
            proposals    = final_boxes,
            output_size  = c(14L, 14L),
            sampling_ratio = 2L,
            aligned        = FALSE
          )
          mask_conv <- self$mask_head(mask_features)
          mask_logits <- self$mask_predictor(mask_conv)  # Shape: (N, num_classes, 28, 28)

          # Extract masks for predicted classes
          n_kept <- final_labels$shape[1]
          final_masks <- torch::torch_zeros(c(n_kept, 28, 28), device = final_boxes$device)

          for (i in seq_len(n_kept)) {
            class_idx <- as.integer(final_labels[i]$item())
            # mask_logits has channels [background, class1, ..., class90]
            # labels are [1, 2, ..., 90], so we need to skip channel 1 (background)
            # In 1-based R indexing: label 1 should access channel 2 (class 1)
            final_masks[i, , ] <- mask_logits[i, class_idx + 1L, , ]
          }

          # Apply sigmoid to get probabilities
          final_masks <- torch::torch_sigmoid(final_masks)

        } else {
          final_masks <- torch::torch_empty(c(0, 28, 28))
        }

        final_results[[b]] <- list(
          boxes = final_boxes,
          labels = final_labels,
          scores = final_scores,
          masks = final_masks
        )
      }
      list(features = features, detections = final_results)
    }
)


# Mask R-CNN Model V2 - With batch normalization
maskrcnn_model_v2 <- torch::nn_module(
  "maskrcnn_model_v2",
    initialize = function(backbone, num_classes,
                          score_thresh = 0.05,
                          nms_thresh = 0.5,
                          detections_per_img = 100) {
      self$backbone <- backbone
      self$num_classes <- num_classes

      # Store configurable detection parameters
      self$score_thresh <- score_thresh
      self$nms_thresh <- nms_thresh
      self$detections_per_img <- detections_per_img

      # RPN with V2 head
      self$rpn <- torch::nn_module(
        initialize = function() {
          self$head <- rpn_head_v2(in_channels = backbone$out_channels)
        },
        forward = function(features) {
          self$head(features)
        }
      )()

      # ROI heads V2 for box prediction
      self$roi_heads <- roi_heads_module_v2(num_classes = num_classes)

      # Mask head V2 for mask prediction
      self$mask_head <- mask_head_module_v2(num_classes = num_classes)

    },

    forward = function(images) {
      features <- self$backbone(images)
      rpn_out <- self$rpn(features)

      batch_size <- images$shape[1]
      image_size <- images$shape[3:4]
      final_results <- list()

      for (b in seq_len(batch_size)) {
        props <- generate_proposals(features, rpn_out, image_size, c(4, 8, 16, 32),
                                    batch_idx = b, score_thresh = 0,
                                    nms_thresh = self$nms_thresh)

        if (props$proposals$shape[1] == 0L) {
          empty <- list(
            boxes = torch::torch_empty(c(0, 4)),
            labels = torch::torch_empty(c(0), dtype = torch::torch_long()),
            scores = torch::torch_empty(c(0)),
            masks = torch::torch_empty(c(0, 28, 28))
          )
          final_results[[b]] <- empty
          next
        }

        # Limit proposals to avoid slow ROI pooling
        max_proposals <- 1000
        if (props$proposals$shape[1] > max_proposals) {
          props$proposals <- props$proposals[1:max_proposals, ]
        }

        # Get ROI head predictions
        roi_out <- self$roi_heads(features, props$proposals, batch_idx = b)

        # Postprocess detections
        det_result <- postprocess_detections(
          class_logits = roi_out$scores,
          box_regression = roi_out$boxes,
          proposals = props$proposals,
          image_size = image_size,
          num_classes = self$num_classes,
          score_thresh = self$score_thresh,
          nms_thresh = self$nms_thresh,
          detections_per_img = self$detections_per_img
        )

        final_boxes <- det_result$boxes
        final_labels <- det_result$labels
        final_scores <- det_result$scores

        # Predict masks for detected objects
        if (final_boxes$shape[1] > 0) {
          mask_features <- roi_align_masks_fpn(
            feature_maps = features,
            proposals    = final_boxes,
            output_size  = c(14L, 14L),
            sampling_ratio = 2L,
            aligned        = FALSE
          )
          mask_logits <- self$mask_head(mask_features)  # (N, num_classes, 28, 28)

          # Extract masks for predicted classes
          n_kept <- final_labels$shape[1]
          final_masks <- torch::torch_zeros(c(n_kept, 28, 28), device = final_boxes$device)

          for (i in seq_len(n_kept)) {
            class_idx <- as.integer(final_labels[i]$item())
            # mask_logits has channels [background, class1, ..., class90]
            # labels are [1, 2, ..., 90], so we need to skip channel 1 (background)
            # In 1-based R indexing: label 1 should access channel 2 (class 1)
            final_masks[i, , ] <- mask_logits[i, class_idx + 1L, , ]
          }

          # Apply sigmoid to get probabilities
          final_masks <- torch::torch_sigmoid(final_masks)

        } else {
          final_masks <- torch::torch_empty(c(0, 28, 28))
        }

        final_results[[b]] <- list(
          boxes = final_boxes,
          labels = final_labels,
          scores = final_scores,
          masks = final_masks
        )
    }
  list(features = features, detections = final_results)
  }
)


#' Mask R-CNN Models
#'
#' Construct Mask R-CNN model variants for instance segmentation task.
#' Mask R-CNN extends Faster R-CNN by adding a mask prediction branch that
#' outputs segmentation masks for each detected object.
#'
#' @param pretrained Logical. If TRUE, loads pretrained weights from local file.
#' @param progress Logical. Show progress bar during download (unused).
#' @param num_classes Number of output classes excluding background (default: 90 for COCO).
#' @param score_thresh Numeric. Minimum score threshold for detections (default: 0.05).
#' @param nms_thresh Numeric. Non-Maximum Suppression (NMS) IoU threshold for removing overlapping boxes (default: 0.5).
#' @param detections_per_img Integer. Maximum number of detections per image (default: 100).
#' @param ... Other arguments (unused).
#' @return A `maskrcnn_model` nn_module.
#'
#' @section Task:
#' Instance segmentation over images with bounding boxes, class labels, and segmentation masks.
#'
#' @section Input Format:
#' Input images should be `torch_tensor`s of shape
#' \verb{(batch_size, 3, H, W)} where `H` and `W` are typically around 800.
#'
#' @section Output Format:
#' Returns a list with:
#' \itemize{
#'   \item `features`: Feature maps from the backbone
#'   \item `detections`: List containing:
#'     \itemize{
#'       \item `boxes`: Bounding boxes (N, 4)
#'       \item `labels`: Class labels (N)
#'       \item `scores`: Confidence scores (N)
#'       \item `masks`: Segmentation masks (N, 28, 28)
#'     }
#' }
#'
#' @section Available Models:
#' \itemize{
#' \item `model_maskrcnn_resnet50_fpn()`
#' \item `model_maskrcnn_resnet50_fpn_v2()`
#' }
#'
#' @examples
#' \dontrun{
#' library(magrittr)
#' # ImageNet normalization constants, see https://pytorch.org/vision/stable/models.html
#' norm_mean <- c(0.485, 0.456, 0.406)
#' norm_std  <- c(0.229, 0.224, 0.225)
#'
#' # Load an image
#' url <- paste0("https://upload.wikimedia.org/wikipedia/commons/thumb/",
#'               "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg")
#' img <- base_loader(url)
#'
#' input <- img %>%
#'   transform_to_tensor() %>%
#'   transform_resize(c(800, 800)) %>%
#'   transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)
#'
#' # Mask R-CNN ResNet-50 FPN
#' model <- model_maskrcnn_resnet50_fpn(pretrained = TRUE, , detections_per_img = 5)
#' model$eval()
#'
#' torch::with_no_grad({pred <- model(batch)$detections[[1]]})
#'
#' # Visualize boxes
#' labels <- coco_classes(as.integer(pred$labels))
#' boxed <- draw_bounding_boxes(image, pred$boxes, labels = labels)
#' tensor_image_browse(boxed)
#'}
#'
#' @family object_detection_model
#' @name model_maskrcnn
#' @rdname model_maskrcnn
NULL

# Model URLs for pretrained weights
mask_rcnn_model_urls <- list(
  maskrcnn_resnet50 = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/maskrcnn_resnet50.pth",
    "8bbfb4cf0d3fafff09739b15647fd123",
    "170 MB"
  ),
  maskrcnn_resnet50_v2 = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/maskrcnn_resnet50_v2.pth",
    "50aa7c34a52e9a9f16d899db3c56b8e5",
    "178 MB"
  )
)

#' @describeIn model_maskrcnn Mask R-CNN with ResNet-50 FPN
#' @export
model_maskrcnn_resnet50_fpn <- function(pretrained = FALSE, progress = TRUE,
                                        num_classes = 90,
                                        score_thresh = 0.05,
                                        nms_thresh = 0.5,
                                        detections_per_img = 100,
                                        ...) {
  backbone <- resnet_fpn_backbone(pretrained = pretrained)
  model <- maskrcnn_model(backbone, num_classes = num_classes,
                         score_thresh = score_thresh,
                         nms_thresh = nms_thresh,
                         detections_per_img = detections_per_img)

  if (pretrained && num_classes != 90)
    cli_abort("Pretrained weights require num_classes = 90 (excluding background).")

  if (pretrained) {
    r <- mask_rcnn_model_urls$maskrcnn_resnet50
    name <- "maskrcnn_resnet50_fpn"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "maskrcnn")

    if (!tools::md5sum(state_dict_path) == r[2]) {
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(.rename_maskrcnn_state_dict(state_dict), strict = FALSE)
  }

  model
}

#' @describeIn model_maskrcnn Mask R-CNN with ResNet-50 FPN V2
#' @export
model_maskrcnn_resnet50_fpn_v2 <- function(pretrained = FALSE, progress = TRUE,
                                           num_classes = 90,
                                           score_thresh = 0.05,
                                           nms_thresh = 0.5,
                                           detections_per_img = 100,
                                           ...) {
  backbone <- resnet_fpn_backbone_v2(pretrained = pretrained)
  model <- maskrcnn_model_v2(backbone, num_classes = num_classes,
                            score_thresh = score_thresh,
                            nms_thresh = nms_thresh,
                            detections_per_img = detections_per_img)

  if (pretrained && num_classes != 90)
    cli_abort("Pretrained weights require num_classes = 90 (excluding background).")

  if (pretrained) {
    r <- mask_rcnn_model_urls$maskrcnn_resnet50_v2
    name <- "maskrcnn_resnet50_fpn_v2"
    cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "maskrcnn")

    if (!tools::md5sum(state_dict_path) == r[2]) {
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    }

    state_dict <- torch::load_state_dict(state_dict_path)

    # Rename state dict keys to match model structure
    state_dict <- .rename_maskrcnn_state_dict_v2(state_dict)

    # Load with flexible matching (similar to fasterrcnn_v2)
    model_state <- model$state_dict()
    state_dict <- state_dict[names(state_dict) %in% names(model_state)]
    for (n in names(state_dict)) {
      if (!all(state_dict[[n]]$size() == model_state[[n]]$size())) {
        state_dict[[n]] <- model_state[[n]]
      }
    }
    missing <- setdiff(names(model_state), names(state_dict))
    if (length(missing) > 0) {
      for (n in missing) {
        state_dict[[n]] <- model_state[[n]]
      }
    }

    model$load_state_dict(state_dict, strict = TRUE)
  }

  model
}

#' @importFrom stats setNames
.rename_maskrcnn_state_dict <- function(state_dict) {
  . <- NULL # Nulling strategy for no visible binding check Note
  new_names <- names(state_dict) %>%
    # add ".0" to inner_blocks + layer_blocks layer renaming
    sub(pattern = "(inner_blocks\\.[0-3]\\.)", replacement = "\\10\\.", x = .) %>%
    sub(pattern = "(layer_blocks\\.[0-3]\\.)", replacement = "\\10\\.", x = .) %>%
    # add ".0.0" to rpn.head.conv
    sub(pattern = "(rpn\\.head\\.conv\\.)", replacement = "\\10\\.0\\.", x = .) %>%
    # remove roi_head prefix to mask_head
    sub(pattern = "roi_heads\\.mask", replacement = "mask", x = .)

  # Recreate a list with renamed keys
  setNames(state_dict[names(state_dict)], new_names)
}

.rename_maskrcnn_state_dict_v2 <- function(state_dict) {
  . <- NULL # Nulling strategy for no visible binding check Note
  new_names <- names(state_dict) %>%
    # change roi_head prefix into mask_head.mask_
    sub(pattern = "roi_heads\\.mask_", replacement = "mask_head\\.mask_", x = .) %>%
    sub(pattern = "roi_heads\\.box_head\\.5\\.", replacement = "roi_heads\\.box_head\\.4\\.", x = .) %>%
    sub(pattern = "(mask_head\\.mask_predictor\\.conv5_mask)", replacement = "\\1\\.0", x = .)

  # Recreate a list with renamed keys
  setNames(state_dict[names(state_dict)], new_names)
}

type_error <- function(msg, env = rlang::caller_env()) {
  rlang::abort(glue::glue(gettext(msg), .envir = env), class = "type_error")
}

value_error <- function(..., env = rlang::caller_env()) {
  rlang::abort(glue::glue(gettext(..., domain = "R-torchvision"), .envir = env), class = "value_error")
}

runtime_error <- function(msg, env = rlang::caller_env()) {
  rlang::abort(glue::glue(gettext(msg), .envir = env), class = "runtime_error")
}

not_implemented_error <- function(msg, env = rlang::caller_env()) {
  rlang::abort(glue::glue(gettext(msg), .envir = env), class = "not_implemented_error")
}

warn <- function(..., env = rlang::caller_env()) {
  rlang::warn(glue::glue(gettext(..., domain = "R-torchvision")[[1]], .envir = env), class = "warning")
}

cli_abort <- function(..., env = rlang::caller_env()) {
  cli::cli_abort(gettext(..., domain = "R-torchvision")[[1]], .envir = env)
}

cli_warn <- function(..., env = rlang::caller_env()) {
  cli::cli_warn(gettext(..., domain = "R-torchvision")[[1]], .envir = env)
}

cli_inform <- function(..., env = rlang::caller_env()) {
  cli::cli_inform(gettext(..., domain = "R-torchvision")[[1]], .envir = env)
}


stop_iteration_error <- function(..., env = rlang::caller_env()) {
  rlang::abort(glue::glue(gettext(..., domain = "R-torchvision")[[1]], .envir = env), class = "stop_iteration_error")
}

inform <- rlang::inform

deprecated <- function(..., env = rlang::caller_env()) {
  rlang::warn(gettext(..., domain = "R-torchvision")[[1]], class = "deprecated")
}

download_and_cache <- function(url, redownload = FALSE, prefix = NULL) {

  cache_path <- rappdirs::user_cache_dir("torch")

  fs::dir_create(cache_path)
  if (!is.null(prefix)) {
    cache_path <- file.path(cache_path, prefix)
  }
  try(fs::dir_create(cache_path, recurse = TRUE), silent = TRUE)
  path <- file.path(cache_path, fs::path_sanitize(fs::path_file(url)))

  if (!file.exists(path) || redownload) {
    # we should first download to a temporary file because
    # download probalems could cause hard to debug errors.
    tmp <- tempfile(fileext = fs::path_ext(path))
    on.exit({try({fs::file_delete(tmp)}, silent = TRUE)}, add = TRUE)

    withr::with_options(
      list(timeout = max(600, getOption("timeout", default = 0))),
      utils::download.file(url, tmp, mode = "wb")
    )
    fs::file_move(tmp, path)
  }

  path
}


#' Validate num_classes parameter
#'
#' @param num_classes Number of classes to validate
#' @param pretrained shall the classes match a pretrained model
#' @param label the label family for the classification head. Shall be "coco" or "ade20k"
#' @keywords internal
#' @noRd
validate_num_classes <- function(num_classes, pretrained, label = "coco") {
  match.arg(label, c("coco","ade20k"))
  if (num_classes <= 0) {
    cli_abort("{.var num_classes} must be positive")
  }
  if (label == "coco" && pretrained && num_classes != 21) {
    cli_abort("Pretrained weights on COCO require {.var num_classes} to be {.val 21}.")
  }
  if (label == "ade20k" && pretrained && num_classes != 150) {
    cli_abort("Pretrained weights on ADE20K require {.var num_classes} to be {.val 150}.")
  }
}


# add additional checks to release issues created with usethis::use_release_issue()
# https://usethis.r-lib.org/reference/use_release_issue.html
release_bullets <- function() {
  c("Update `po/R-torchvision.pot` file with `potools::po_update()`",
    "Contact translators to collect their translation `.po` files",
    "Compile the translations with `potools::po_compile()`"
  )
}

# R/model-maxvit.R - torch compatible MaxViT implementation
#' @importFrom torch nn_module nn_conv2d nn_gelu nn_batch_norm2d nn_sequential
maxvit_conv_norm_act <- function(in_channels, mid_channels = 64, out_channels = 64, ...) {
  nn_sequential(
    "stem.0" = nn_sequential(
      "0" = nn_conv2d(in_channels, mid_channels, ..., bias = FALSE),
      "1" = nn_batch_norm2d(mid_channels, track_running_stats = TRUE),
      "2" = nn_gelu()
    ),
    "stem.1" = nn_sequential(
      "0" = nn_conv2d(mid_channels, out_channels, kernel_size = 3, stride = 1, padding = 1, bias = TRUE)
    )
  )
}

#' @importFrom torch nn_module nn_conv2d nnf_gelu torch_sigmoid
squeeze_excitation <- nn_module(
  initialize = function(in_channels, squeeze_factor = 16) {
    squeeze_channels <- max(1L, as.integer(in_channels / squeeze_factor))
    self$fc1 <- nn_conv2d(in_channels, squeeze_channels, 1)
    self$fc2 <- nn_conv2d(squeeze_channels, in_channels, 1)
  },
  forward = function(x) {
    scale <- x$mean(dim = c(3, 4), keepdim = TRUE)
    scale <- self$fc1(scale)
    scale <- nnf_gelu(scale)
    scale <- self$fc2(scale)
    scale <- torch_sigmoid(scale)
    x * scale
  }
)

# Simplified relative position implementation
#' @importFrom torch torch_arange torch_zeros torch_long
create_relative_position_params <- function(window_size, heads) {
  # Create a simple relative position index
  coords <- torch_arange(window_size * window_size)
  relative_position_index <- torch_zeros(window_size * window_size, window_size * window_size, dtype = torch_long())

  # Create bias table
  num_relative_distance <- (2 * window_size - 1) * (2 * window_size - 1)
  relative_position_bias_table <- torch_zeros(num_relative_distance, heads)

  list(
    relative_position_bias_table = relative_position_bias_table,
    relative_position_index = relative_position_index
  )
}

#' @importFrom torch nn_module nn_module_dict nn_layer_norm nn_linear nn_gelu nn_parameter torch_chunk
#' @importFrom torch torch_matmul nnf_softmax
window_attention <- nn_module(
  initialize = function(dim, heads = 2, window_size = 7) {
    self$dim <- dim
    self$heads <- heads
    self$window_size <- window_size

    # Layer structure to match expected keys
    self$attn_layer <- nn_module_dict(list(
      "0" = nn_layer_norm(dim),  # norm1
      "1" = nn_module_dict(list(
        # Custom attention implementation
        "to_qkv" = nn_linear(dim, dim * 3, bias = TRUE),
        "merge" = nn_linear(dim, dim, bias = TRUE)
      ))
    ))

    # Add relative position parameters as buffers/parameters
    rel_pos_params <- create_relative_position_params(window_size, heads)
    self$attn_layer[["1"]][["relative_position_bias_table"]] <- nn_parameter(rel_pos_params$relative_position_bias_table)
    self$register_buffer("attn_layer.1.relative_position_index", rel_pos_params$relative_position_index)

    self$mlp_layer <- nn_module_dict(list(
      "0" = nn_layer_norm(dim),  # norm2
      "1" = nn_linear(dim, dim * 4),
      "2" = nn_gelu(),
      "3" = nn_linear(dim * 4, dim)
    ))
  },

  forward = function(x) {
    b <- x$size(1); c <- x$size(2); h <- x$size(3); w <- x$size(4)
    x <- x$permute(c(1, 3, 4, 2))$reshape(c(b, h * w, c))

    # Attention path
    normed <- self$attn_layer[["0"]](x)
    qkv <- self$attn_layer[["1"]][["to_qkv"]](normed)
    qkv_chunks <- torch_chunk(qkv, 3, dim = -1)
    q <- qkv_chunks[[1]]
    k <- qkv_chunks[[2]]
    v <- qkv_chunks[[3]]

    # Simplified attention computation
    scale <- 1.0 / sqrt(c / self$heads)
    attn <- torch_matmul(q, k$transpose(-2, -1)) * scale
    attn <- nnf_softmax(attn, dim = -1)
    out <- torch_matmul(attn, v)
    out <- self$attn_layer[["1"]][["merge"]](out)

    x <- x + out

    # MLP path
    normed <- self$mlp_layer[["0"]](x)
    mlp_out <- self$mlp_layer[["1"]](normed)
    mlp_out <- self$mlp_layer[["2"]](mlp_out)
    mlp_out <- self$mlp_layer[["3"]](mlp_out)
    x <- x + mlp_out

    x <- x$reshape(c(b, h, w, c))$permute(c(1, 4, 2, 3))
    x
  }
)

#' @importFrom torch nn_module nn_module_dict nn_layer_norm nn_linear nn_gelu nn_parameter torch_chunk
grid_attention <- nn_module(
  initialize = function(dim, heads = 2, grid_size = 7) {
    self$dim <- dim
    self$heads <- heads
    self$grid_size <- grid_size

    # Layer structure to match expected keys
    self$attn_layer <- nn_module_dict(list(
      "0" = nn_layer_norm(dim),  # norm1
      "1" = nn_module_dict(list(
        # Custom attention implementation
        "to_qkv" = nn_linear(dim, dim * 3, bias = TRUE),
        "merge" = nn_linear(dim, dim, bias = TRUE)
      ))
    ))

    # Add relative position parameters as buffers/parameters
    rel_pos_params <- create_relative_position_params(grid_size, heads)
    self$attn_layer[["1"]][["relative_position_bias_table"]] <- nn_parameter(rel_pos_params$relative_position_bias_table)
    self$register_buffer("attn_layer.1.relative_position_index", rel_pos_params$relative_position_index)

    self$mlp_layer <- nn_module_dict(list(
      "0" = nn_layer_norm(dim),  # norm2
      "1" = nn_linear(dim, dim * 4),
      "2" = nn_gelu(),
      "3" = nn_linear(dim * 4, dim)
    ))
  },

  forward = function(x) {
    b <- x$size(1); c <- x$size(2); h <- x$size(3); w <- x$size(4)
    gs <- self$grid_size
    gh <- h %/% gs; gw <- w %/% gs
    x <- x$reshape(c(b, c, gh, gs, gw, gs))$permute(c(1, 3, 5, 4, 6, 2))$reshape(c(b * gh * gw, gs * gs, c))

    # Attention path
    normed <- self$attn_layer[["0"]](x)
    qkv <- self$attn_layer[["1"]][["to_qkv"]](normed)
    qkv_chunks <- torch_chunk(qkv, 3, dim = -1)
    q <- qkv_chunks[[1]]
    k <- qkv_chunks[[2]]
    v <- qkv_chunks[[3]]

    # Simplified attention computation
    scale <- 1.0 / sqrt(c / self$heads)
    attn <- torch_matmul(q, k$transpose(-2, -1)) * scale
    attn <- nnf_softmax(attn, dim = -1)
    out <- torch_matmul(attn, v)
    out <- self$attn_layer[["1"]][["merge"]](out)

    x <- x + out

    # MLP path
    normed <- self$mlp_layer[["0"]](x)
    mlp_out <- self$mlp_layer[["1"]](normed)
    mlp_out <- self$mlp_layer[["2"]](mlp_out)
    mlp_out <- self$mlp_layer[["3"]](mlp_out)
    x <- x + mlp_out

    x <- x$reshape(c(b, gh, gw, gs, gs, c))$permute(c(1, 6, 2, 4, 3, 5))$reshape(c(b, c, h, w))
    x
  }
)

#' @importFrom torch nn_module nn_module_dict nn_batch_norm2d nn_conv2d nn_identity nn_sequential
mbconv <- nn_module(
  initialize = function(in_channels, out_channels, expansion = 4, stride = 1) {
    hidden_dim <- in_channels * expansion

    self$layers <- nn_module_dict(list(
      pre_norm = nn_batch_norm2d(in_channels, track_running_stats = TRUE),
      conv_a = nn_sequential(
        "0" = nn_conv2d(in_channels, hidden_dim, 1, bias = FALSE),
        "1" = nn_batch_norm2d(hidden_dim, track_running_stats = TRUE),
        "2" = nn_gelu()
      ),
      conv_b = nn_sequential(
        "0" = nn_conv2d(hidden_dim, hidden_dim, 3, stride, 1, groups = hidden_dim, bias = FALSE),
        "1" = nn_batch_norm2d(hidden_dim, track_running_stats = TRUE),
        "2" = nn_gelu()
      ),
      squeeze_excitation = squeeze_excitation(hidden_dim),
      conv_c = nn_conv2d(hidden_dim, out_channels, 1, bias = TRUE)
    ))

    # Only add projection layer for blocks that change dimensions
    self$has_proj <- (stride != 1 || in_channels != out_channels)
    if (self$has_proj) {
      self$proj <- nn_sequential(
        "0" = nn_identity(),
        "1" = nn_conv2d(in_channels, out_channels, 1, bias = TRUE)
      )
    } else {
      self$proj <- nn_sequential(
        "0" = nn_identity()
      )
    }

    self$use_res_connect <- stride == 1 && in_channels == out_channels
  },

  forward = function(x) {
    identity <- x
    out <- self$layers$pre_norm(x)
    out <- self$layers$conv_a(out)
    out <- self$layers$conv_b(out)
    out <- self$layers$squeeze_excitation(out)
    out <- self$layers$conv_c(out)

    if (self$use_res_connect) {
      out <- out + identity
    } else if (self$has_proj) {
      identity <- self$proj(identity)
      if (identity$size(3) != out$size(3) || identity$size(4) != out$size(4)) {
        identity <- nnf_avg_pool2d(identity, kernel_size = 2, stride = 2)
      }
      out <- out + identity
    }

    out
  }
)

#' @importFrom torch nn_module nn_module_dict nn_sequential
maxvit_block <- nn_module(
  initialize = function(in_channels, out_channels, expansion = 4, stride = 1, num_heads = 2) {
    self$layers <- nn_module_dict(list(
      MBconv = mbconv(in_channels, out_channels, expansion, stride),
      window_attention = window_attention(out_channels, heads = num_heads),
      grid_attention = grid_attention(out_channels, heads = num_heads)
    ))
  },

  forward = function(x) {
    x <- self$layers$MBconv(x)
    x <- self$layers$window_attention(x)
    x <- self$layers$grid_attention(x)
    x
  }
)

#' @importFrom torch nn_module nn_module_dict nn_sequential
maxvit_stage <- nn_module(
  initialize = function(...) {
    self$layers <- nn_sequential(...)
  },
  forward = function(x) {
    self$layers(x)
  }
)

#' @importFrom torch nn_module nn_module_list nn_adaptive_avg_pool2d nn_identity nn_layer_norm nn_linear
maxvit_impl <- nn_module(
  initialize = function(num_classes = 1000) {
    self$stem <- maxvit_conv_norm_act(
      3, mid_channels = 64, out_channels = 64,
      kernel_size = 3, stride = 2, padding = 1
    )

    self$blocks <- nn_module_list(list(
      # stage 0 - 2 heads (default)
      maxvit_stage(
        "0" = maxvit_block(64, 64, expansion = 4, stride = 2, num_heads = 2),
        "1" = maxvit_block(64, 64, expansion = 4, stride = 1, num_heads = 2)
      ),
      # stage 1 - 4 heads
      maxvit_stage(
        "0" = maxvit_block(64, 128, expansion = 8, stride = 2, num_heads = 4),
        "1" = maxvit_block(128, 128, expansion = 4, stride = 1, num_heads = 4)
      ),
      # stage 2 - 8 heads
      maxvit_stage(
        "0" = maxvit_block(128, 256, expansion = 8, stride = 2, num_heads = 8),
        "1" = maxvit_block(256, 256, expansion = 4, stride = 1, num_heads = 8),
        "2" = maxvit_block(256, 256, expansion = 4, stride = 1, num_heads = 8),
        "3" = maxvit_block(256, 256, expansion = 4, stride = 1, num_heads = 8),
        "4" = maxvit_block(256, 256, expansion = 4, stride = 1, num_heads = 8)
      ),
      # stage 3 - 16 heads
      maxvit_stage(
        "0" = maxvit_block(256, 512, expansion = 8, stride = 2, num_heads = 16),
        "1" = maxvit_block(512, 512, expansion = 4, stride = 1, num_heads = 16)
      )
    ))

    self$pool <- nn_adaptive_avg_pool2d(c(1, 1))

    # Fixed classifier structure to match original naming
    self$classifier <- nn_sequential(
      "0" = nn_identity(),
      "1" = nn_identity(),
      "2" = nn_layer_norm(512),
      "3" = nn_linear(512, 512, bias = TRUE),
      "4" = nn_identity(),
      "5" = nn_linear(512, num_classes, bias = FALSE)  # Additional classifier layer
    )
  },

  forward = function(x) {
    x <- self$stem(x)

    for (i in seq_along(self$blocks)) {
      stage <- self$blocks[[i]]
      x <- stage(x)
    }

    x <- self$pool(x)
    x <- x$flatten(start_dim = 2)
    self$classifier(x)
  }
)

.rename_maxvit_state_dict <- function(state_dict) {
  renamed <- list()

  for (nm in names(state_dict)) {

    new_nm <- nm

    # MBConv projection layers - fix the renaming
    if (grepl("\\.proj\\.0\\.(weight|bias)$", new_nm)) {
      # Skip identity layer renaming
      next
    }

    # Don't rename the relative position bias and index - keep original names
    if (grepl("relative_position_bias_table|relative_position_index", new_nm)) {
      renamed[[new_nm]] <- state_dict[[nm]]
      next
    }

    # Attention + MLP layer renaming
    new_nm <- sub("attn_layer\\.0\\.", "attn_layer.0.", new_nm)
    new_nm <- sub("attn_layer\\.1\\.to_qkv\\.", "attn_layer.1.to_qkv.", new_nm)
    new_nm <- sub("attn_layer\\.1\\.merge\\.", "attn_layer.1.merge.", new_nm)
    new_nm <- sub("mlp_layer\\.0\\.", "mlp_layer.0.", new_nm)
    new_nm <- sub("mlp_layer\\.1\\.", "mlp_layer.1.", new_nm)
    new_nm <- sub("mlp_layer\\.3\\.", "mlp_layer.3.", new_nm)

    # Keep classifier names exactly as they are in the .pth file
    # Don't rename classifier.2, classifier.3, classifier.5

    renamed[[new_nm]] <- state_dict[[nm]]
  }

  renamed
}

#' MaxViT Model
#'
#' Implementation of the MaxViT architecture described in
#' [MaxViT: Multi-Axis Vision Transformer](https://arxiv.org/abs/2204.01697).
#' The model performs image classification and by default returns logits for
#' 1000 ImageNet classes.
#'
#' @inheritParams model_resnet18
#' @param num_classes (integer) Number of output classes.
#' @param ... Additional parameters passed to the model initializer.
#'
#' @family classification_model
#'
#' @examples
#' \dontrun{
#' library(magrittr)
#' # 1. Load the basketball image
#' img_url <- "https://upload.wikimedia.org/wikipedia/commons/7/7a/Basketball.png"
#' img <- base_loader(img_url)
#'
#' # 2. Define normalization (ImageNet)
#' norm_mean <- c(0.485, 0.456, 0.406)
#' norm_std <- c(0.229, 0.224, 0.225)
#'
#' # 3. Preprocess: convert to tensor, resize, Normalize
#' input <- img %>%
#'   transform_to_tensor() %>%
#'   transform_resize(c(400, 400)) %>%
#'   transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)    # Add batch dimension (1, 3, H, W)
#'
#' # 4. Display the image before normalization
#' tensor_image_browse(input)
#'
#' # 5. Load MaxViT model
#' model <- model_maxvit(pretrained = TRUE)
#' model$eval()
#'
#' # 6. Run inference
#' output <- model(batch)
#' topk <- output$topk(k = 5, dim = 2)
#' indices <- as.integer(topk[[2]][1, ])
#' scores <- as.numeric(topk[[1]][1, ])
#'
#' # 7. Show Top-5 predictions
#' glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}%)")
#' }
#'
#' @export
model_maxvit <- function(pretrained = FALSE, progress = TRUE, num_classes = 1000, ...) {
  model <- maxvit_impl(num_classes = num_classes)

  if (pretrained) {
    path <- download_and_cache("https://torch-cdn.mlverse.org/models/vision/v2/models/maxvit.pth")
    state_dict <- torch::load_state_dict(path)
    state_dict <- .rename_maxvit_state_dict(state_dict)
    state_dict <- state_dict[!grepl("num_batches_tracked$", names(state_dict))]

    model_state <- model$state_dict()
    model_state <- model_state[!grepl("num_batches_tracked$", names(model_state))]

    model$load_state_dict(state_dict, strict = FALSE)
  }

  model
}

test_that("tests for non-pretrained model_maxvit", {
  model <- model_maxvit()
  input <- torch::torch_randn(1, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  model <- model_maxvit(num_classes = 10)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))
})

test_that("tests for pretrained model_maxvit", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_maxvit(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 448, 448)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))
})

#' FGVC Aircraft Dataset
#'
#' The FGVC-Aircraft dataset supports the following official splits:
#' - `"train"`: training subset with labels.
#' - `"val"`: validation subset with labels.
#' - `"trainval"`: combined training and validation set with labels.
#' - `"test"`: test set with labels (used for evaluation).
#'
#' @param root Character. Root directory for dataset storage. The dataset will be stored under `root/fgvc-aircraft-2013b`.
#' @param split Character. One of `"train"`, `"val"`, `"trainval"`, or `"test"`. Default is `"train"`.
#' @param annotation_level Character. Level of annotation to use for classification. Default is `"variant"`.
#' One of `"variant"`, `"family"`, `"manufacturer"`, or `"all"`. See *Details*.
#' @param transform Optional function to transform input images after loading. Default is `NULL`.
#' @param target_transform Optional function to transform labels. Default is `NULL`.
#' @param download Logical. Whether to download the dataset if not found locally. Default is `FALSE`.
#'
#' @details
#' The `annotation_level` determines the granularity of labels used for classification and supports four values:
#' - `"variant"`: the most fine-grained level, e.g., `"Boeing 737-700"`. There are 100 visually distinguishable variants.
#' - `"family"`: a mid-level grouping, e.g., `"Boeing 737"`, which includes multiple variants. There are 70 distinct families.
#' - `"manufacturer"`: the coarsest level, e.g., `"Boeing"`, grouping multiple families under a single manufacturer. There are 30 manufacturers.
#' - `"all"`: multi-label format that returns all three levels as a vector of class indices `c(manufacturer_idx, family_idx, variant_idx)`.
#'
#' These levels form a strict hierarchy: each `"manufacturer"` consists of multiple `"families"`, and each `"family"` contains several `"variants"`.
#' Not all combinations of levels are valid — for example, a `"variant"` always belongs to exactly one `"family"`, and a `"family"` to exactly one `"manufacturer"`.
#'
#' When `annotation_level = "all"` is used, the `$classes` field is a named list with three components:
#' - `classes$manufacturer`: a character vector of manufacturer names
#' - `classes$family`: a character vector of family names
#' - `classes$variant`: a character vector of variant names
#'
#' @return An object of class \code{fgvc_aircraft_dataset}, which behaves like a torch-style dataset.
#' Each element is a named list with:
#' - `x`: an array of shape (H, W, C) with pixel values in the range (0, 255). Please note that images have varying sizes.
#' - `y`: for single-level annotation (`"variant"`, `"family"`, `"manufacturer"`): an integer class label.
#'        for multi-level annotation (`"all"`): a vector of three integers `c(manufacturer_idx, family_idx, variant_idx)`.
#'
#' @examples
#' \dontrun{
#' # Single-label classification
#' fgvc <- fgvc_aircraft_dataset(transform = transform_to_tensor, download = TRUE)
#'
#' # Create a custom collate function to resize images and prepare batches
#' resize_collate_fn <- function(batch) {
#'   xs <- lapply(batch, function(item) {
#'     torchvision::transform_resize(item$x, c(768, 1024))
#'   })
#'   xs <- torch::torch_stack(xs)
#'   ys <- torch::torch_tensor(sapply(batch, function(item) item$y), dtype = torch::torch_long())
#'   list(x = xs, y = ys)
#' }
#' dl <- torch::dataloader(dataset = fgvc, batch_size = 2, collate_fn = resize_collate_fn)
#' batch <- dataloader_next(dataloader_make_iter(dl))
#' batch$x  # batched image tensors with shape (2, 3, 768, 1024)
#' batch$y  # class labels as integer tensor of shape 2
#'
#' # Multi-label classification
#' fgvc <- fgvc_aircraft_dataset(split = "test", annotation_level = "all")
#' item <- fgvc[1]
#' item$x  # a double vector representing the image
#' item$y  # an integer vector of length 3: manufacturer, family, and variant indices
#' fgvc$classes$manufacturer[item$y[1]]  # e.g., "Boeing"
#' fgvc$classes$family[item$y[2]]        # e.g., "Boeing 707"
#' fgvc$classes$variant[item$y[3]]       # e.g., "707-320"
#' }
#'
#' @family classification_dataset
#' @export
fgvc_aircraft_dataset <- dataset(
  name = "fgvc_aircraft",
  archive_size = "2.8 GB",

  initialize = function(
    root = tempdir(),
    split = "train",
    annotation_level = "variant",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root <- root
    self$split <- split
    self$annotation_level <- annotation_level
    self$transform <- transform
    self$target_transform <- target_transform
    self$base_dir <- file.path(root, "fgvc-aircraft-2013b")
    self$data_dir <- file.path(self$base_dir, "data")

    if (download){
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      runtime_error("Dataset not found. Use `download = TRUE` to download it.")
    }

    self$classes <- list(
      manufacturer = readLines(file.path(self$data_dir, "manufacturers.txt")),
      family = readLines(file.path(self$data_dir, "families.txt")),
      variant = readLines(file.path(self$data_dir, "variants.txt"))
    )

    levels <- names(self$classes)
    label_data <- lapply(levels, function(level) {
      read.fwf(
        file.path(self$data_dir, glue::glue("images_{level}_{self$split}.txt")),
        widths = c(7, 100),
        colClasses = "character",
        stringsAsFactors = FALSE,
        col.names = c("img_id", level),
        strip.white = TRUE
      )
    })
    names(label_data) <- levels

    merged_df <- Reduce(function(x, y) merge(x, y, by = "img_id"), label_data)
    merged_df[levels] <- lapply(levels, function(level) {
      as.integer(factor(merged_df[[level]], levels = self$classes[[level]]))
    })

    merged_df <- merged_df[complete.cases(merged_df), ]
    self$image_paths <- file.path(self$data_dir, "images", glue::glue("{merged_df$img_id}.jpg"))
    self$labels_df <- merged_df[, levels]

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes[[annotation_level]])} classes.")
  },

  .getitem = function(index) {
    x <- jpeg::readJPEG(self$image_paths[index]) * 255

    if (self$annotation_level == "all") {
      y <- as.integer(self$labels_df[index, ])
    } else {
      y <- self$labels_df[[self$annotation_level]][index]
    }

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    list(x = x, y = y)
  },

  .length = function() {
    length(self$image_paths)
  },

  check_exists = function() {
    fs::dir_exists(self$base_dir)
  },

  download = function() {
    if (self$check_exists()){
      return()
    }

    fs::dir_create(self$root)
    url <- "https://www.robots.ox.ac.uk/~vgg/data/fgvc-aircraft/archives/fgvc-aircraft-2013b.tar.gz"
    md5 <- "d4acdd33327262359767eeaa97a4f732"

    cli_inform("Downloading {.cls {class(self)[[1]]}} ...")

    archive <- withr::with_options(list(timeout = 1200), download_and_cache(url))
    if (!tools::md5sum(archive) == md5) {
      runtime_error("Corrupt file! Delete the file at {archive} and try again.")
    }

    untar(archive, exdir = self$root)

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  }
)

#' MobileNetV3 Model
#'
#' MobileNetV3 is a state-of-the-art lightweight convolutional neural network architecture
#' designed for mobile and embedded vision applications. This implementation follows the
#' design and optimizations presented in the original paper:[MobileNetV3: Searching for MobileNetV3](https://arxiv.org/abs/1905.02244)
#'
#' The model includes two variants:
#' - `model_mobilenet_v3_large()`
#' - `model_mobilenet_v3_small()`
#'
#' Both variants utilize efficient blocks such as inverted residuals, squeeze-and-excitation (SE) modules,
#' and hard-swish activations for improved accuracy and efficiency.
#'
#' ## Model Summary and Performance for pretrained weights
#' ```
#' | Model                  | Top-1 Acc | Top-5 Acc | Params  | GFLOPS | File Size | Notes                               |
#' |------------------------|-----------|-----------|---------|--------|-----------|-------------------------------------|
#' | MobileNetV3 Large      | 74.04%    | 91.34%    | 5.48M   | 0.22   | 21.1 MB   | Trained from scratch, simple recipe |
#' | MobileNetV3 Small      | 67.67%    | 87.40%    | 2.54M   | 0.06   | 9.8 MB    | Improved recipe over original paper |
#' ```
#'
#' @examples
#' \dontrun{
#' # 1. Download sample image (dog)
#' norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants, see
#' # https://pytorch.org/vision/stable/models.html
#' norm_std  <- c(0.229, 0.224, 0.225)
#' img_url <- "https://en.wikipedia.org/wiki/Special:FilePath/Felis_catus-cat_on_snow.jpg"
#' img <- base_loader(img_url)
#'
#' # 2. Convert to tensor (RGB only), resize and normalize
#' input <- img %>%
#'  transform_to_tensor() %>%
#'  transform_resize(c(224, 224)) %>%
#'  transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)
#'
#' # 3. Load pretrained models
#' model_small <- model_mobilenet_v3_small(pretrained = TRUE)
#' model_small$eval()
#'
#' # 4. Forward pass
#' output_s <- model_small(batch)
#'
#' # 5. Top-5 printing helper
#' topk <- output_s$topk(k = 5, dim = 2)
#' indices <- as.integer(topk[[2]][1, ])
#' scores <- as.numeric(topk[[1]][1, ])
#'
#' # 6. Show Top-5 predictions
#' glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}%)")
#'
#' # 7. Same with large model
#' model_large <- model_mobilenet_v3_large(pretrained = TRUE)
#' model_large$eval()
#' output_l <- model_large(input)
#' topk <- output_l$topk(k = 5, dim = 2)
#' indices <- as.integer(topk[[2]][1, ])
#' scores <- as.numeric(topk[[1]][1, ])
#' glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}%)")
#' }
#'
#' @importFrom torch nn_module nn_conv2d nn_batch_norm2d nn_relu nn_hardswish nn_hardsigmoid nn_identity nn_sequential
#' @importFrom torch nn_adaptive_avg_pool2d nn_linear nn_dropout torch_clamp torch_flatten load_state_dict
#'
#' @inheritParams model_mobilenet_v2
#' @param num_classes number of output classes (default: 1000).
#' @param width_mult width multiplier for model scaling (default: 1.0).
#'
#' @family classification_model
#' @rdname model_mobilenet_v3
#' @name model_mobilenet_v3
NULL

make_divisible <- function(v, divisor = 8, min_value = NULL) {
  if (is.null(min_value)) min_value <- divisor
  new_v <- max(min_value, floor((v + divisor / 2) / divisor) * divisor)
  if (new_v < 0.9 * v) new_v <- new_v + divisor
  new_v
}

HardSwish <- nn_module(
  "HardSwish",
  forward = function(x) {
    x * torch_clamp(x + 3, min = 0, max = 6) / 6
  }
)

HardSigmoid <- nn_module(
  "HardSigmoid",
  forward = function(x) {
    torch_clamp(x + 3, min = 0, max = 6) / 6
  }
)

Conv2dNormActivation <- nn_module(
  "Conv2dNormActivation",
  initialize = function(in_channels, out_channels, kernel_size, stride = 1, groups = 1,
                        norm_layer = nn_batch_norm2d, activation_layer = nn_relu,
                        dilation = 1) {
    padding <- floor(((kernel_size - 1) / 2) * dilation)
    self$conv <- nn_conv2d(
      in_channels, out_channels, kernel_size,
      stride = stride, padding = padding, groups = groups,
      dilation = dilation, bias = FALSE
    )
    self$bn <- norm_layer(out_channels)
    if (is.character(activation_layer) && activation_layer == "hardswish") {
      self$activation <- HardSwish()
    } else if (identical(activation_layer, nn_hardswish)) {
      self$activation <- HardSwish()
    } else if (identical(activation_layer, nn_hardsigmoid)) {
      self$activation <- HardSigmoid()
    } else {
      self$activation <- activation_layer()
    }
  },
  forward = function(x) {
    x %>% self$conv() %>% self$bn() %>% self$activation()
  }
)

SELayer <- nn_module(
  "SELayer",
  initialize = function(input_channels, squeeze_channels) {
    self$avg_pool <- nn_adaptive_avg_pool2d(1)
    self$fc1 <- nn_conv2d(input_channels, squeeze_channels, 1)
    self$relu <- nn_relu()
    self$fc2 <- nn_conv2d(squeeze_channels, input_channels, 1)
    self$hsigmoid <- nn_hardsigmoid()
  },
  forward = function(x) {
    scale <- x %>%
      self$avg_pool() %>%
      self$fc1() %>%
      self$relu() %>%
      self$fc2() %>%
      self$hsigmoid()
    x * scale
  }
)

InvertedResidual <- nn_module(
  "InvertedResidual",
  initialize = function(input_channels, expanded_channels, out_channels, kernel, stride,
                        use_se, use_hs, dilation = 1, norm_layer = nn_batch_norm2d) {

    if (!(stride %in% c(1, 2))) stop("illegal stride value")
    self$use_res_connect <- (stride == 1) && (input_channels == out_channels)

    layers <- list()
    activation_layer <- if (use_hs) nn_hardswish else nn_relu

    if (expanded_channels != input_channels) {
      layers <- c(layers, list(
        Conv2dNormActivation(
          input_channels, expanded_channels, kernel_size = 1,
          norm_layer = norm_layer, activation_layer = activation_layer
        )
      ))
    }

    stride_ <- if (dilation > 1) 1 else stride
    layers <- c(layers, list(
      Conv2dNormActivation(
        expanded_channels, expanded_channels, kernel_size = kernel,
        stride = stride_, groups = expanded_channels, dilation = dilation,
        norm_layer = norm_layer, activation_layer = activation_layer
      )
    ))

    if (use_se) {
      squeeze_channels <- make_divisible(expanded_channels / 4)
      layers <- c(layers, list(SELayer(expanded_channels, squeeze_channels)))
    }

    layers <- c(layers, list(
      Conv2dNormActivation(
        expanded_channels, out_channels, kernel_size = 1,
        norm_layer = norm_layer, activation_layer = nn_identity
      )
    ))

    self$block <- nn_sequential(!!!layers)
    self$out_channels <- out_channels
  },
  forward = function(x) {
    out <- self$block(x)
    if (self$use_res_connect) {
      out + x
    } else {
      out
    }
  }
)

InvertedResidualConfig <- function(input_c, kernel, expanded_c, out_c, use_se, use_hs, stride, dilation = 1, width_mult = 1.0) {
  input_channels <- make_divisible(input_c * width_mult)
  expanded_channels <- make_divisible(expanded_c * width_mult)
  out_channels <- make_divisible(out_c * width_mult)

  list(
    input_channels = input_channels,
    kernel = kernel,
    expanded_channels = expanded_channels,
    out_channels = out_channels,
    use_se = use_se,
    use_hs = use_hs,
    stride = stride,
    dilation = dilation
  )
}

MobileNetV3 <- nn_module(
  "MobileNetV3",
  initialize = function(inverted_residual_setting, last_channel, num_classes = 1000,
                        dropout = 0.2, norm_layer = nn_batch_norm2d) {

    layers <- list()

    firstconv_out <- inverted_residual_setting[[1]]$input_channels
    layers <- c(layers, list(
      Conv2dNormActivation(
        3, firstconv_out, kernel_size = 3, stride = 2,
        norm_layer = norm_layer, activation_layer = nn_hardswish
      )
    ))

    for (conf in inverted_residual_setting) {
      layers <- c(layers, list(
        InvertedResidual(
          input_channels = conf$input_channels,
          expanded_channels = conf$expanded_channels,
          out_channels = conf$out_channels,
          kernel = conf$kernel,
          stride = conf$stride,
          use_se = conf$use_se,
          use_hs = conf$use_hs,
          dilation = conf$dilation,
          norm_layer = norm_layer
        )
      ))
    }

    lastconv_in <- inverted_residual_setting[[length(inverted_residual_setting)]]$out_channels
    lastconv_out <- 6 * lastconv_in
    layers <- c(layers, list(
      Conv2dNormActivation(
        lastconv_in, lastconv_out, kernel_size = 1,
        norm_layer = norm_layer, activation_layer = nn_hardswish
      )
    ))

    self$features <- nn_sequential(!!!layers)
    self$avgpool <- nn_adaptive_avg_pool2d(1)

    self$classifier <- nn_sequential(
      nn_linear(lastconv_out, last_channel),
      HardSwish(),
      nn_dropout(p = dropout, inplace = TRUE),
      nn_linear(last_channel, num_classes)
    )
  },
  forward = function(x) {
    x %>%
      self$features() %>%
      self$avgpool() %>%
      torch_flatten(start_dim = 2) %>%
      self$classifier()
  }
)

mobilenet_v3_large_config <- function(width_mult = 1.0, reduced_tail = FALSE, dilated = FALSE) {
  reduce_divider <- ifelse(reduced_tail, 2, 1)
  dilation <- ifelse(dilated, 2, 1)
  b <- function(in_c, k, exp_c, out_c, se, act, s, d = 1) {
    InvertedResidualConfig(in_c, k, exp_c, out_c, se, act == "HS", s, d, width_mult)
  }

  list(
    b(16, 3, 16, 16, FALSE, "RE", 1, 1),
    b(16, 3, 64, 24, FALSE, "RE", 2, 1),
    b(24, 3, 72, 24, FALSE, "RE", 1, 1),
    b(24, 5, 72, 40, TRUE, "RE", 2, 1),
    b(40, 5, 120, 40, TRUE, "RE", 1, 1),
    b(40, 5, 120, 40, TRUE, "RE", 1, 1),
    b(40, 3, 240, 80, FALSE, "HS", 2, 1),
    b(80, 3, 200, 80, FALSE, "HS", 1, 1),
    b(80, 3, 184, 80, FALSE, "HS", 1, 1),
    b(80, 3, 184, 80, FALSE, "HS", 1, 1),
    b(80, 3, 480, 112, TRUE, "HS", 1, 1),
    b(112, 3, 672, 112, TRUE, "HS", 1, 1),
    b(112, 5, 672, 160 / reduce_divider, TRUE, "HS", 2, dilation),
    b(160 / reduce_divider, 5, 960 / reduce_divider, 160 / reduce_divider, TRUE, "HS", 1, dilation),
    b(160 / reduce_divider, 5, 960 / reduce_divider, 160 / reduce_divider, TRUE, "HS", 1, dilation)
  )
}

mobilenet_v3_small_config <- function(width_mult = 1.0, reduced_tail = FALSE, dilated = FALSE) {
  reduce_divider <- ifelse(reduced_tail, 2, 1)
  dilation <- ifelse(dilated, 2, 1)
  b <- function(in_c, k, exp_c, out_c, se, act, s, d = 1) {
    InvertedResidualConfig(in_c, k, exp_c, out_c, se, act == "HS", s, d, width_mult)
  }

  list(
    b(16, 3, 16, 16, TRUE, "RE", 2, 1),
    b(16, 3, 72, 24, FALSE, "RE", 2, 1),
    b(24, 3, 88, 24, FALSE, "RE", 1, 1),
    b(24, 5, 96, 40, TRUE, "HS", 2, 1),
    b(40, 5, 240, 40, TRUE, "HS", 1, 1),
    b(40, 5, 240, 40, TRUE, "HS", 1, 1),
    b(40, 5, 120, 48, TRUE, "HS", 1, 1),
    b(48, 5, 144, 48, TRUE, "HS", 1, 1),
    b(48, 5, 288, 96 / reduce_divider, TRUE, "HS", 2, dilation),
    b(96 / reduce_divider, 5, 576 / reduce_divider, 96 / reduce_divider, TRUE, "HS", 1, dilation),
    b(96 / reduce_divider, 5, 576 / reduce_divider, 96 / reduce_divider, TRUE, "HS", 1, dilation)
  )
}

#' @describeIn model_mobilenet_v3 MobileNetV3 Large model with about 5.5 million parameters.
#' @export
model_mobilenet_v3_large <- function(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 1000,
  width_mult = 1.0
) {

  config <- mobilenet_v3_large_config(width_mult)
  last_channel <- make_divisible(1280 * width_mult)
  model <- MobileNetV3(config, last_channel, num_classes = num_classes)
  if (pretrained) {
    r <- c("https://torch-cdn.mlverse.org/models/vision/v2/models/mobilenet_v3_large.pth",
           "71625955bc3be9516032a6d5bab49199", "~21 MB")
    state_dict_path <- download_and_cache(r[1], prefix = "mobilenet")
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    state_dict <- load_state_dict(state_dict_path)
    renamed_state_dict <- .rename_mobilenet_v3_state_dict(state_dict)
    model$load_state_dict(renamed_state_dict)
  }
  model$eval()
  model
}

#' @describeIn model_mobilenet_v3 MobileNetV3 Small model with about 2.5 million parameters.
#' @export
model_mobilenet_v3_small <- function(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 1000,
  width_mult = 1.0
) {

  config <- mobilenet_v3_small_config(width_mult)
  last_channel <- make_divisible(1024 * width_mult)
  model <- MobileNetV3(config, last_channel, num_classes = num_classes)
  if (pretrained) {
    state_dict_path <- download_and_cache("https://torch-cdn.mlverse.org/models/vision/v2/models/mobilenet_v3_small.pth", prefix = "mobilenet")
    state_dict <- load_state_dict(state_dict_path)
    renamed_state_dict <- .rename_mobilenet_v3_state_dict(state_dict)
    model$load_state_dict(renamed_state_dict)
  }
  model$eval()
  model
}


#' Rename state_dict keys to match current MobileNetV3 model structure
#' @param state_dict Named list of model parameters
#' @return Named list with updated key names
#' @keywords internal
.rename_mobilenet_v3_state_dict <- function(state_dict) {
  . <- NULL # Nulling strategy for no visible binding check Note
  new_names <- names(state_dict) %>%
    # features.N.0 -> features.N.conv
    gsub(pattern = "^features\\.([0-9]+)\\.0\\.", replacement = "features.\\1.conv.", x = .) %>%
    # features.N.1 -> features.N.bn
    gsub(pattern = "^features\\.([0-9]+)\\.1\\.", replacement = "features.\\1.bn.", x = .) %>%
    # features.N.block.M.0 -> features.N.block.M.conv
    gsub(pattern = "(features\\.[0-9]+\\.block\\.[0-9]+)\\.0\\.", replacement = "\\1.conv.", x = .) %>%
    # features.N.block.M.1 -> features.N.block.M.bn
    gsub(pattern = "(features\\.[0-9]+\\.block\\.[0-9]+)\\.1\\.", replacement = "\\1.bn.", x = .) %>%
    # features.N.block.M.fc1. -> fc1. (no change needed, but ensure correct)
    gsub(pattern = "(features\\.[0-9]+\\.block\\.[0-9]+\\.fc1)\\.", replacement = "\\1.", x = .) %>%
    # features.N.block.M.fc2. -> fc2.
    gsub(pattern = "(features\\.[0-9]+\\.block\\.[0-9]+\\.fc2)\\.", replacement = "\\1.", x = .)

  # Recreate a list with renamed keys
  setNames(state_dict[names(state_dict)], new_names)
}

context("models-mobilenetv3")

test_that("tests for non-pretrained model_mobilenet_v3_large", {
  model_large <- model_mobilenet_v3_large(pretrained = FALSE)
  input <- torch_randn(1, 3, 224, 224)
  model_large$eval()
  out <- model_large(input)
  expect_tensor_shape(out, c(1, 1000))

  model <- model_mobilenet_v3_large(pretrained = FALSE,num_classes = 10)
  input <- torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model_large)
  rm(model)
  gc()
})

test_that("tests for pretrained model_mobilenet_v3_large", {
  model_large <- model_mobilenet_v3_large(pretrained = TRUE)
  input <- torch_randn(1, 3, 224, 224)
  model_large$eval()
  out <- model_large(input)
  expect_tensor_shape(out, c(1, 1000))

  rm(model_large)
  gc()
})

test_that("tests for non-pretrained model_mobilenet_v3_small", {
  model_small <- model_mobilenet_v3_small(pretrained = FALSE)
  input <- torch_randn(1, 3, 224, 224)
  model_small$eval()
  out <- model_small(input)
  expect_tensor_shape(out, c(1, 1000))

  rm(model_small)
  gc()
})

test_that("tests for pretrained model_mobilenet_v3_small", {
  model_small <- model_mobilenet_v3_small(pretrained = TRUE)
  input <- torch_randn(1, 3, 224, 224)
  model_small$eval()
  out <- model_small(input)
  expect_tensor_shape(out, c(1, 1000))

  model <- model_mobilenet_v3_small(pretrained = FALSE,num_classes = 10)
  input <- torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model_small)
  rm(model)
  gc()
})

test_that("tests for model_mobilenet_v3_large with non-divisible input shapes", {

  model <- model_mobilenet_v3_large(pretrained = FALSE)
  input <- torch_randn(2, 3, 223, 225)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(2, 1000))
  rm(model)
  gc()
})

test_that("tests for model_mobilenet_v3_small with non-divisible input shapes", {

  model <- model_mobilenet_v3_small(pretrained = FALSE)
  input <- torch_randn(1, 3, 223, 225)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))
  rm(model)
  gc()
})

test_that("tests for model_mobilenet_v3_small with varied width_mult", {

  model <- model_mobilenet_v3_small(pretrained = FALSE, width_mult = 0.5)
  input <- torch_randn(2, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(2, 1000))
  rm(model)
  gc()
})

test_that("tests for model_mobilenet_v3_large with varied width_mult", {

  model <- model_mobilenet_v3_large(pretrained = FALSE, width_mult = 0.5)
  input <- torch_randn(1, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))
  rm(model)
  gc()
})


test_that("we can prune head of mobilenetv3 models", {
  mobilenet <- model_mobilenet_v3_large(pretrained=TRUE)

  expect_no_error(prune <- nn_prune_head(mobilenet, 1))
  expect_s3_class(prune, "nn_sequential")
  expect_length(prune, 2)
  expect_s3_class(prune[[1]][1], "nn_sequential")

  input <- torch::torch_randn(1, 3, 256, 256)
  out <- prune(input)
  expect_tensor_shape(out, c(1, 960, 1, 1))
})

#' CIFAR datasets
#'
#' The CIFAR datasets are benchmark **classification** datasets composed of
#' 60,000 RGB thumbnail images of size 32x32 pixels. The
#' [CIFAR10](https://www.cs.toronto.edu/~kriz/cifar.html) variant contains
#' 10 classes while CIFAR100 provides 100 classes. Images are split into
#' 50,000 training samples and 10,000 test samples.
#'
#' Downloads and prepares the CIFAR archives.
#'
#' @rdname cifar_datasets
#'
#' @inheritParams mnist_dataset
#' @param root (string): Root directory of dataset where directory
#'   `cifar-10-batches-bin` exists or will be saved to if download is set to TRUE.
#' @return A torch::dataset object. Each item is a list with:
#' * `x`: a 32x32x3 integer array
#' * `y`: the class label
#'
#' @examples
#' \dontrun{
#' ds <- cifar10_dataset(root = tempdir(), download = TRUE)
#' item <- ds[1]
#' item$x
#' item$y
#' }
#' @family classification_dataset
#'
#' @export
cifar10_dataset <- torch::dataset(
  name = "cifar10_dataset",
  url = "https://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz",
  md5 = "c32a1d4ab5d03f1284b67883e8d87530",
  fname = "cifar-10-batches-bin",
  type = 10,
  label_fname = "batches.meta.txt",
  archive_size = "160 MB",
  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$train <- train
    self$split <- ifelse(train, "train", "test")

    if (download){
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    check <- self$check_files()

    if (!check)
      runtime_error("Files not found. Use download = TRUE")

    if (train) {
      files <- self$get_files()$train
    } else {
      files <- self$get_files()$test
    }

    batches <- lapply(files, function(x) read_batch(x, self$type))

    if (self$type == 10)
      data <- combine_batches(batches)
    else
      data <- batches[[1]]

    self$.load_meta()

    self$x <- data$imgs
    self$y <- data$labels + 1L

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },
  .load_meta = function() {
    cl <- readLines(fs::path(self$root, self$fname, self$label_fname))
    self$class_to_idx <- setNames(seq_along(cl), cl)
    self$classes <- cl
  },
  .getitem = function(i) {
    x <- self$x[i,,,]
    y <- self$y[i]

    if (!is.null(self$transform))
      x <- self$transform(x)

    if (!is.null(self$target_transform))
      y <- self$target_transform(y)

    list(x = x, y = y)
  },
  .length = function() {
    length(self$y)
  },
  download = function() {

    if(self$check_files())
      return()

    cli_inform("Downloading {.cls {class(self)[[1]]}} ...")

    archive <- download_and_cache(self$url)

    if (!tools::md5sum(archive) == self$md5)
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")

    utils::untar(archive, exdir = self$root)

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },
  check_files = function() {

    if (!fs::dir_exists(self$root))
      return(FALSE)

    p <- fs::path(self$root, self$fname)
    if (!fs::dir_exists(p))
      return(FALSE)

    f <- self$get_files()

    if (!length(f$train) == 5 && self$type == 10)
      return(FALSE)

    if (!length(f$train) == 1 && self$type == 100)
      return(FALSE)

    if (!length(f$test) == 1)
      return(FALSE)

    return(TRUE)
  },
  get_files = function() {
    p <- fs::path(self$root, self$fname)

    if (self$type == 10) {
      list(
        train = fs::dir_ls(p, regexp = "data_batch"),
        test = fs::dir_ls(p, regexp = "test_batch")
      )
    } else {
      list(
        train = fs::dir_ls(p, regexp = "train"),
        test = fs::dir_ls(p, regexp = "test")
      )
    }
  }
)

#' CIFAR100 dataset
#'
#' Downloads and prepares the
#' [CIFAR100](https://www.cs.toronto.edu/~kriz/cifar.html) dataset.
#'
#' @rdname cifar_datasets
#' @export
cifar100_dataset <- torch::dataset(
  name = "cifar100_dataset",
  inherit = cifar10_dataset,
  url = "https://www.cs.toronto.edu/~kriz/cifar-100-binary.tar.gz",
  md5 = "03b5dce01913d631647c71ecec9e9cb8",
  fname = "cifar-100-binary",
  type = 100,
  label_fname = "fine_label_names.txt",
  archive_size = "160 MB"
)

read_batch <- function(path, type = 10) {

  if (type == 10)
    n <- 10000
  else if (type == 100 && grepl("test", path))
    n <- 10000
  else
    n <- 50000

  imgs <- array(dim = c(n, 32, 32, 3))
  labels <- integer(length = n)
  if (type == 100)
    fine_labels <- integer(length = n)

  con <- file(path, open = "rb")
  on.exit({close(con)}, add = TRUE)

  for (i in seq_len(n)) {

    labels[i] <- readBin(con, integer(), size=1, n=1, endian="big")

    if (type == 100) {
      fine_labels[i] <- readBin(con, integer(), size=1, n=1, endian="big")
    }

    r <- as.integer(readBin(con, raw(), size=1, n=1024, endian="big"))
    g <- as.integer(readBin(con, raw(), size=1, n=1024, endian="big"))
    b <- as.integer(readBin(con, raw(), size=1, n=1024, endian="big"))

    imgs[i,,,1] <- matrix(r, ncol = 32, byrow = TRUE)
    imgs[i,,,2] <- matrix(g, ncol = 32, byrow = TRUE)
    imgs[i,,,3] <- matrix(b, ncol = 32, byrow = TRUE)
  }

  if (type == 100)
    list(imgs = imgs, labels = fine_labels)
  else
    list(imgs = imgs, labels = labels)
}

combine_batches <- function(batches) {

  n <- 10000

  imgs <- array(dim = c(length(batches)* n, 32, 32, 3))
  labels <- integer(length = length(batches)* n)

  for (i in seq_along(batches)) {
    imgs[((i-1)*n + 1):(i*n),,,] <- batches[[i]]$imgs
    labels[((i-1)*n + 1):(i*n)] <- batches[[i]]$labels
  }

  list(imgs = imgs, labels = labels)
}

#' EfficientNet Models
#'
#' Constructs EfficientNet model architectures as described in
#' \emph{EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks}.
#' These models are designed for image classification tasks and provide a balance
#' between accuracy and computational efficiency through compound scaling.
#'
#' @section Task:
#' Image classification with 1000 output classes by default (ImageNet).
#'
#' @section Input Format:
#' The models expect input tensors of shape \code{(batch_size, 3, H, W)}, where H and W
#' should typically be 224 for B0 and scaled versions for B1–B7 (e.g., B7 uses 600x600).
#'
#' @section Variants and Scaling:
#' \tabular{lllllll}{
#'   \strong{Model} \tab \strong{Width} \tab \strong{Depth} \tab
#'   \strong{Resolution} \tab \strong{Params (M)} \tab \strong{GFLOPs} \tab
#'   \strong{Top-1 Acc.} \cr
#'   B0 \tab 1.0 \tab 1.0 \tab 224 \tab 5.3 \tab 0.39 \tab 77.1 \cr
#'   B1 \tab 1.0 \tab 1.1 \tab 240 \tab 7.8 \tab 0.70 \tab 79.1 \cr
#'   B2 \tab 1.1 \tab 1.2 \tab 260 \tab 9.2 \tab 1.00 \tab 80.1 \cr
#'   B3 \tab 1.2 \tab 1.4 \tab 300 \tab 12.0 \tab 1.80 \tab 81.6 \cr
#'   B4 \tab 1.4 \tab 1.8 \tab 380 \tab 19.0 \tab 4.20 \tab 82.9 \cr
#'   B5 \tab 1.6 \tab 2.2 \tab 456 \tab 30.0 \tab 9.90 \tab 83.6 \cr
#'   B6 \tab 1.8 \tab 2.6 \tab 528 \tab 43.0 \tab 19.0 \tab 84.0 \cr
#'   B7 \tab 2.0 \tab 3.1 \tab 600 \tab 66.0 \tab 37.0 \tab 84.3
#' }
#'
#' @inheritParams model_resnet18
#' @param ... Other parameters passed to the model implementation, such as
#' \code{num_classes} to change the output dimension.
#'
#' @family classification_model
#'
#' @examples
#' \dontrun{
#' model <- model_efficientnet_b0()
#' image_batch <- torch::torch_randn(1, 3, 224, 224)
#' output <- model(image_batch)
#' imagenet_classes(which.max(as.numeric(output)))
#' }
#'
#' \dontrun{
#' # Example of using EfficientNet-B5 with its native image size
#' model <- model_efficientnet_b5()
#' image_batch <- torch::torch_randn(1, 3, 456, 456)
#' output <- model(image_batch)
#' imagenet_classes(which.max(as.numeric(output)))
#' }
#'
#' @name model_efficientnet
NULL

efficientnet_model_urls <- list(
  efficientnet_b0 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b0.pth", "5cb997d28f1a30cdf1732dd1e69a6647", "~22 MB"),
  efficientnet_b1 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b1.pth", "1f1fdcc560d5a91875bbec02c0a105e8", "~32 MB"),
  efficientnet_b2 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b2.pth", "c5440198bb37adc12d8d88f58863b2d3", "~37 MB"),
  efficientnet_b3 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b3.pth", "74967d21f6d437845ab5f2fd87f31df1", "~50 MB"),
  efficientnet_b4 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b4.pth", "a044c7adf3bf4c457581669eae41f713", "~80 MB"),
  efficientnet_b5 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b5.pth", "8bea08f18d29d5becce388f363247a26", "~122 MB"),
  efficientnet_b6 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b6.pth", "96844282790aff48afe3c56b434f4863", "~175 MB"),
  efficientnet_b7 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_b7.pth", "c3ddfc4df9851a8fe55312fdbe8f8c7e", "~270 MB")
)

conv_norm_act <- torch::nn_module(
  inherit = torch::nn_sequential,
  initialize = function(in_channels, out_channels, kernel_size = 3, stride = 1,
                        groups = 1, norm_layer = NULL, activation_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    if (is.null(activation_layer))
      activation_layer <- torch::nn_silu
    padding <- (kernel_size - 1) %/% 2
    super$initialize(
      torch::nn_conv2d(in_channels, out_channels, kernel_size, stride, padding,
                       groups = groups, bias = FALSE),
      norm_layer(out_channels),
      activation_layer(inplace = TRUE)
    )
    self$out_channels <- out_channels
  }
)

se_block <- torch::nn_module(
  initialize = function(in_channels, squeeze_channels) {
    self$avgpool <- torch::nn_adaptive_avg_pool2d(output_size = 1)
    self$fc1 <- torch::nn_conv2d(in_channels, squeeze_channels, kernel_size = 1)
    self$activation <- torch::nn_relu()
    self$fc2 <- torch::nn_conv2d(squeeze_channels, in_channels, kernel_size = 1)
    self$scale_activation <- torch::nn_sigmoid()
  },
  forward = function(x) {
    scale <- self$avgpool(x)
    scale <- self$fc1(scale)
    scale <- self$activation(scale)
    scale <- self$fc2(scale)
    scale <- self$scale_activation(scale)
    x * scale
  }
)


mbconv_block <- torch::nn_module(
  initialize = function(in_channels, out_channels, kernel_size, stride,
                        expand_ratio, se_ratio = 0.25, norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    hidden_dim <- in_channels * expand_ratio
    self$use_res_connect <- stride == 1 && in_channels == out_channels
    layers <- list()

    # Expand
    if (expand_ratio != 1) {
      layers[[length(layers) + 1]] <- conv_norm_act(
        in_channels, hidden_dim, kernel_size = 1,
        norm_layer = norm_layer, activation_layer = torch::nn_silu
      )
    }

    # Depthwise
    layers[[length(layers) + 1]] <- conv_norm_act(
      hidden_dim, hidden_dim, kernel_size = kernel_size, stride = stride,
      groups = hidden_dim, norm_layer = norm_layer,
      activation_layer = torch::nn_silu
    )

    # Correct SE block with proper squeeze logic
    if (!is.null(se_ratio) && se_ratio > 0) {
      squeeze_channels <- max(1, as.integer(in_channels * se_ratio))
      layers[[length(layers) + 1]] <- se_block(
        hidden_dim, squeeze_channels = squeeze_channels
      )
    }

    # Projection
    layers[[length(layers) + 1]] <- torch::nn_sequential(
      torch::nn_conv2d(hidden_dim, out_channels, 1, bias = FALSE),
      norm_layer(out_channels)
    )

    self$block <- torch::nn_sequential(!!!layers)
  },
  forward = function(x) {
    out <- self$block(x)
    if (self$use_res_connect)
      out <- out + x
    out
  }
)


efficientnet <- torch::nn_module(
  "efficientnet",
  initialize = function(width_coefficient = 1, depth_coefficient = 1,
                        dropout = 0.2, num_classes = 1000, norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    b0_cfg <- list(
      list(expand = 1, channels = 16, repeats = 1, stride = 1, kernel = 3),
      list(expand = 6, channels = 24, repeats = 2, stride = 2, kernel = 3),
      list(expand = 6, channels = 40, repeats = 2, stride = 2, kernel = 5),
      list(expand = 6, channels = 80, repeats = 3, stride = 2, kernel = 3),
      list(expand = 6, channels = 112, repeats = 3, stride = 1, kernel = 5),
      list(expand = 6, channels = 192, repeats = 4, stride = 2, kernel = 5),
      list(expand = 6, channels = 320, repeats = 1, stride = 1, kernel = 3)
    )
    round_filters <- function(filters) {
      divisor <- 8
      filters <- filters * width_coefficient
      new_filters <- max(divisor, as.integer(filters + divisor/2) %/% divisor * divisor)
      if (new_filters < 0.9 * filters)
        new_filters <- new_filters + divisor
      as.integer(new_filters)
    }
    round_repeats <- function(repeats) {
      as.integer(ceiling(depth_coefficient * repeats))
    }
    out_channels <- round_filters(32)
    features <- list(conv_norm_act(3, out_channels, stride = 2, norm_layer = norm_layer, activation_layer = torch::nn_silu))
    in_channels <- out_channels

    for (cfg in b0_cfg) {
      oc <- round_filters(cfg$channels)
      r <- round_repeats(cfg$repeats)
      stage_blocks <- list()
      for (i in 1:r) {
        s <- ifelse(i == 1, cfg$stride, 1)
        stage_blocks[[i]] <- mbconv_block(in_channels, oc,
                                          kernel_size = cfg$kernel, stride = s, expand_ratio = cfg$expand,
                                          se_ratio = 0.25, norm_layer = norm_layer)
        in_channels <- oc
      }
      features[[length(features) + 1]] <- torch::nn_sequential(!!!stage_blocks)
    }

    final_channels <- round_filters(1280)
    features[[length(features) + 1]] <- conv_norm_act(in_channels, final_channels,
                                                      kernel_size = 1, norm_layer = norm_layer, activation_layer = torch::nn_silu)
    self$features <- torch::nn_sequential(!!!features)
    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(1, 1))
    self$classifier <- torch::nn_sequential(
      torch::nn_dropout(dropout),
      torch::nn_linear(final_channels, num_classes)
    )
  },
  forward = function(x) {
    x <- self$features(x)
    x <- self$avgpool(x)
    x <- torch::torch_flatten(x, start_dim = 2)
    self$classifier(x)
  }
)

effnet <- function(arch, width, depth, dropout, pretrained, progress, ...) {
  args <- rlang::list2(...)
  model <- do.call(efficientnet, append(args, list(
    width_coefficient = width,
    depth_coefficient = depth,
    dropout = dropout
  )))

  if (pretrained) {
    r <- efficientnet_model_urls[[arch]]
    cli_inform("Model weights for {.cls {arch}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "efficientnet")
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    state_dict <- state_dict[!grepl("num_batches_tracked", names(state_dict))]
    model$load_state_dict(state_dict, strict = FALSE)
  }

  model
}

# Individual model variants

#' @describeIn model_efficientnet EfficientNet B0 model
#' @export
model_efficientnet_b0 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b0", 1.0, 1.0, 0.2, pretrained, progress, ...)
}


#' @describeIn model_efficientnet EfficientNet B1 model
#' @export
model_efficientnet_b1 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b1", 1.0, 1.1, 0.2, pretrained, progress, ...)
}

#' @describeIn model_efficientnet EfficientNet B2 model
#' @export
model_efficientnet_b2 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b2", 1.1, 1.2, 0.3, pretrained, progress, ...)
}

#' @describeIn model_efficientnet EfficientNet B3 model
#' @export
model_efficientnet_b3 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b3", 1.2, 1.4, 0.3, pretrained, progress, ...)
}

#' @describeIn model_efficientnet EfficientNet B4 model
#' @export
model_efficientnet_b4 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b4", 1.4, 1.8, 0.4, pretrained, progress, ...)
}

#' @describeIn model_efficientnet EfficientNet B5 model
#' @export
model_efficientnet_b5 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b5", 1.6, 2.2, 0.4, pretrained, progress, ...)
}

#' @describeIn model_efficientnet EfficientNet B6 model
#' @export
model_efficientnet_b6 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b6", 1.8, 2.6, 0.5, pretrained, progress, ...)
}

#' @describeIn model_efficientnet EfficientNet B7 model
#' @export
model_efficientnet_b7 <- function(pretrained = FALSE, progress = TRUE, ...) {
  effnet("efficientnet_b7", 2.0, 3.1, 0.5, pretrained, progress, ...)
}

test_that("efficientnet models produce correct output shapes", {


  variants <- list(
    b0 = model_efficientnet_b0,
    b1 = model_efficientnet_b1,
    b2 = model_efficientnet_b2,
    b3 = model_efficientnet_b3,
    b4 = model_efficientnet_b4,
    b5 = model_efficientnet_b5
  )

  sizes <- c(b0 = 224, b1 = 240, b2 = 260, b3 = 300,
             b4 = 380, b5 = 456)

  for (name in names(variants)) {
    fn <- variants[[name]]
    size <- sizes[[name]]

    # without pretrained
    model <- fn(pretrained = FALSE)
    input <- torch::torch_randn(1, 3, size, size)
    out <- model(input)
    expect_tensor_shape(out, c(1, 1000))

    # with pretrained
    withr::with_options(list(timeout = 360), {
      model <- fn(pretrained = TRUE)
    })
    input <- torch::torch_randn(1, 3, size, size)
    out <- model(input)
    expect_tensor_shape(out, c(1, 1000))
  }
  unlink_model_file()
})

test_that("large efficientnet models produce correct output shapes", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  variants <- list(
    b6 = model_efficientnet_b6,
    b7 = model_efficientnet_b7
  )

  sizes <- c(b6 = 528, b7 = 600)

  for (name in names(variants)) {
    fn <- variants[[name]]
    size <- sizes[[name]]

    # without pretrained
    model <- fn(pretrained = FALSE)
    input <- torch::torch_randn(1, 3, size, size)
    out <- model(input)
    expect_tensor_shape(out, c(1, 1000))

    # with pretrained
    withr::with_options(list(timeout = 360), {
      model <- fn(pretrained = TRUE)
    })
    input <- torch::torch_randn(1, 3, size, size)
    out <- model(input)
    expect_tensor_shape(out, c(1, 1000))
  }
  unlink_model_file()
})

test_that("efficientnet b7 model produce correct output shapes", {
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  # without pretrained
  model <- model_efficientnet_b7(pretrained = FALSE)
  input <- torch::torch_randn(1, 3, 600, 600)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  # with pretrained
  withr::with_options(list(timeout = 360), {
    model <- model_efficientnet_b7(pretrained = TRUE)
  })
  input <- torch::torch_randn(1, 3, 600, 600)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  unlink_model_file()
})

test_that("efficientnet allows custom num_classes", {
  model <- model_efficientnet_b0(num_classes = 42)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 42))
})

#' Dataset Collection Catalog
#'
#' A comprehensive catalog of all collections RF100 (RoboFlow 100) and EMNIST datasets available in torchvision.
#' This data frame contains metadata about each dataset including descriptions, sizes,
#' available splits, and collection information.
#'
#' @format A data frame with datasets as rows and 17 columns:
#' \describe{
#'   \item{collection}{Collection name (biology, medical, infrared, damage, underwater, document, mnist)}
#'   \item{dataset}{Dataset identifier used in collection functions}
#'   \item{description}{Brief description of the dataset and its purpose}
#'   \item{task}{Machine learning task type (currently all "object_detection")}
#'   \item{num_classes}{Number of different object classes}
#'   \item{num_images}{Total images across all splits}
#'   \item{image_width}{Typical image width in pixels}
#'   \item{image_height}{Typical image height in pixels}
#'   \item{train_size_mb}{Size of training split in megabytes}
#'   \item{test_size_mb}{Size of test split in megabytes}
#'   \item{valid_size_mb}{Size of validation split in megabytes}
#'   \item{total_size_mb}{Total size across all splits in megabytes}
#'   \item{has_train}{Is training split available}
#'   \item{has_test}{Is test split available}
#'   \item{has_valid}{Is validation split available}
#'   \item{function_name}{R function name to load this dataset's collection}
#'   \item{roboflow_url}{URL to the collection on RoboFlow Universe}
#' }
#'
#' @examples
#' \dontrun{
#' # View the complete catalog
#' data(collection_catalog)
#' View(collection_catalog)
#'
#' # See all biology datasets
#' subset(collection_catalog, collection == "biology")
#'
#' # Find large datasets (> 100 MB)
#' subset(collection_catalog, total_size_mb > 100)
#' }
#'
#' @seealso [search_collection()], [get_collection_catalog()]
"collection_catalog"

#' Search Collection Catalog
#'
#' Search through all Collection datasets by keywords in name or description,
#' or filter by collection. This makes it easy to discover datasets relevant
#' to your task without browsing each collection individually.
#'
#' @param keyword Character string to search for (case-insensitive). Searches
#'   in both dataset names and descriptions. If NULL, returns all datasets
#'   (optionally filtered by collection).
#' @param collection Filter by collection name. One of: "biology", "medical",
#'   "infrared", "damage", "underwater", "document". If NULL, searches all collections.
#'
#' @return A data frame with matching datasets and their metadata. Returns NULL
#'   invisibly if no matches are found.
#'
#' @examples
#' \dontrun{
#' # Find all medical datasets
#' search_collection(collection = "medical")
#'
#' # Find datasets about cells
#' search_collection("cell")
#'
#' # Find photovoltaic/solar datasets
#' search_collection("solar")
#' search_collection("photovoltaic")
#'
#' # Find all biology datasets with "cell" in name/description
#' search_collection("cell", collection = "biology")
#'
#' # List all available datasets
#' search_collection()
#' }
#'
#' @seealso [get_collection_catalog()], [collection_catalog]
#' @export
search_collection <- function(keyword = NULL, collection = NULL) {
  # Load the catalog
  utils::data("collection_catalog", package = "torchvision", envir = environment())

  result <- collection_catalog

  # Filter by collection first
  if (!is.null(collection)) {
    valid_collections <- c("rf100_biology", "rf100_medical", "rf100_infrared", "rf100_damage", "rf100_underwater", "rf100_document", "mnist")
    if (!collection %in% valid_collections) {
      stop("Invalid collection. Must be one of: ", paste(valid_collections, collapse = ", "))
    }
    result <- result[result$collection == collection, ]
  }

  # Filter by keyword in dataset name or description
  if (!is.null(keyword)) {
    pattern <- tolower(keyword)
    matches <- grepl(pattern, tolower(result$dataset)) |
               grepl(pattern, tolower(result$description))
    result <- result[matches, ]
  }

  if (nrow(result) == 0) {
    message("No datasets found matching criteria")
    return(invisible(NULL))
  }

  # Reset row names for cleaner display
  rownames(result) <- NULL

  result
}

#' Get Complete Collection Catalog
#'
#' Returns the complete catalog of datasets in collections with their metadata.
#' This is a convenience function that loads and returns the collection_catalog data.
#'
#' @return A data frame with all datasets and their metadata.
#'
#' @examples
#' \dontrun{
#' # Get complete catalog
#' catalog <- get_collection_catalog()
#'
#' # View in RStudio
#' View(catalog)
#'
#' # Summary statistics
#' summary(catalog$total_size_mb)
#' table(catalog$collection)
#'
#' # Find smallest dataset
#' catalog[which.min(catalog$total_size_mb), ]
#'
#' # Find largest dataset
#' catalog[which.max(catalog$total_size_mb), ]
#' }
#'
#' @seealso [search_collection()], [collection_catalog]
#' @export
get_collection_catalog <- function() {
  utils::data("collection_catalog", package = "torchvision", envir = environment())
  collection_catalog
}

#' List Datasets in a Collection
#'
#' List all available datasets within a specific RF100 collection.
#'
#' @param collection Collection name. One of: "biology", "medical", "infrared",
#'   "damage", "underwater", "document".
#'
#' @return Character vector of dataset names in the collection.
#'
#' @examples
#' \dontrun{
#' # List all biology datasets
#' list_collection_datasets("biology")
#'
#' # List all medical datasets
#' list_collection_datasets("medical")
#' }
#'
#' @seealso [search_collection()], [get_collection_catalog()]
#' @export
list_collection_datasets <- function(collection) {
  valid_collections <- c("rf100_biology", "rf100_medical", "rf100_infrared", "rf100_damage", "rf100_underwater", "rf100_document", "mnist")
  if (!collection %in% valid_collections) {
    stop("Invalid collection. Must be one of: ", paste(valid_collections, collapse = ", "))
  }

  utils::data("collection_catalog", package = "torchvision", envir = environment())
  datasets <- collection_catalog[collection_catalog$collection == collection, "dataset"]
  sort(datasets)
}

test_that("RF100 catalog data exists and has correct structure", {
  skip_on_cran()

  # Load catalog
  catalog <- get_collection_catalog()

  # Should be a data frame
  expect_s3_class(catalog, "data.frame")

  # Should have expected columns
  expected_cols <- c("collection", "dataset", "description", "task", "num_classes",
                     "num_images", "image_width", "image_height",
                     "train_size_mb", "test_size_mb", "valid_size_mb", "total_size_mb",
                     "has_train", "has_test", "has_valid",
                     "function_name", "roboflow_url")
  expect_length(setdiff(expected_cols, names(catalog)), 0)

  # Should have datasets (allow for future additions)
  expect_gte(nrow(catalog), 34)

  # All datasets should have descriptions
  expect_true(all(nchar(catalog$description) > 5))

  # All datasets should have positive sizes
  expect_true(all(catalog$total_size_mb > 0))

  # All datasets should have positive image counts
  expect_true(all(catalog$num_images > 0))

  # All datasets should have positive dimensions
  expect_true(all(catalog$image_width > 0))
  expect_true(all(catalog$image_height > 0))

  # Check collections are valid
  valid_collections <- c("rf100_biology", "rf100_medical", "rf100_infrared", "rf100_damage",
                         "rf100_underwater", "rf100_document", "emnist")
  expect_length(setdiff(catalog$collection, valid_collections), 0)

  # All datasets should have all three splits
  expect_true(all(catalog$has_train))
  expect_true(all(catalog$has_test))
})

test_that("search_collection works with keyword search", {
  skip_on_cran()

  # Search for cell-related datasets
  cell_results <- search_collection("cell")
  expect_s3_class(cell_results, "data.frame")
  expect_gt(nrow(cell_results), 0)

  # All results should contain "cell" in name or description
  has_cell <- sapply(seq_len(nrow(cell_results)), function(i) {
    grepl("cell", cell_results$dataset[i], ignore.case = TRUE) ||
    grepl("cell", cell_results$description[i], ignore.case = TRUE)
  })
  expect_true(all(has_cell))

  # Search for solar datasets
  solar_results <- search_collection("solar")
  expect_s3_class(solar_results, "data.frame")
  expect_gt(nrow(solar_results), 0)

  # Search for non-existent keyword
  expect_message(
    result <- search_collection("nonexistent_keyword_xyz"),
    "No datasets found"
  )
  expect_null(result)
})

test_that("search_collection works with collection filter", {
  skip_on_cran()

  # Search biology collection
  biology_results <- search_collection(collection = "rf100_biology")
  expect_s3_class(biology_results, "data.frame")
  expect_gte(nrow(biology_results), 10)
  expect_true(all(biology_results$collection == "rf100_biology"))

  # Search medical collection
  medical_results <- search_collection(collection = "rf100_medical")
  expect_s3_class(medical_results, "data.frame")
  expect_gte(nrow(medical_results), 8)
  expect_true(all(medical_results$collection == "rf100_medical"))

  # Invalid collection should error
  expect_error(
    search_collection(collection = "invalid_collection"),
    "Invalid collection"
  )
})

test_that("search_collection works with combined keyword and collection", {
  skip_on_cran()

  # Search for "cell" in biology collection
  results <- search_collection("cell", collection = "rf100_biology")
  expect_s3_class(results, "data.frame")
  expect_true(all(results$collection == "rf100_biology"))

  # All results should contain "cell"
  has_cell <- sapply(seq_len(nrow(results)), function(i) {
    grepl("cell", results$dataset[i], ignore.case = TRUE) ||
    grepl("cell", results$description[i], ignore.case = TRUE)
  })
  expect_true(all(has_cell))
})

test_that("search_collection with no arguments returns all datasets", {
  skip_on_cran()

  all_results <- search_collection()
  expect_s3_class(all_results, "data.frame")
  expect_gte(nrow(all_results), 37)
})

test_that("list_collection_datasets works correctly", {
  skip_on_cran()

  # List biology datasets
  biology_datasets <- list_collection_datasets("rf100_biology")
  expect_type(biology_datasets, "character")
  expect_gte(length(biology_datasets), 10)
  expect_true("blood_cell" %in% biology_datasets)

  # List medical datasets
  medical_datasets <- list_collection_datasets("rf100_medical")
  expect_type(medical_datasets, "character")
  expect_length(medical_datasets, 8)

  # Invalid collection should error
  expect_error(
    list_collection_datasets("invalid_collection"),
    "Invalid collection"
  )
})

test_that("catalog has correct dataset counts per collection", {
  skip_on_cran()

  catalog <- get_collection_catalog()

  # Check counts
  counts <- table(catalog$collection)
  expect_gte(as.numeric(counts["rf100_biology"]), 9)
  expect_gte(as.numeric(counts["rf100_medical"]), 8)
  expect_gte(as.numeric(counts["rf100_infrared"]), 4)
  expect_gte(as.numeric(counts["rf100_damage"]), 3)
  expect_gte(as.numeric(counts["rf100_underwater"]), 4)
  expect_gte(as.numeric(counts["rf100_document"]), 8)
})

test_that("catalog includes known datasets", {
  skip_on_cran()

  catalog <- get_collection_catalog()

  # Check for some known datasets
  expect_true("blood_cell" %in% catalog$dataset)
  expect_true("stomata_cell" %in% catalog$dataset)
  expect_true("solar_panel" %in% catalog$dataset)
  expect_true("brain_tumor" %in% catalog$dataset)
  expect_true("coral" %in% catalog$dataset)
})

test_that("CSV catalog file exists and is readable", {
  skip_on_cran()

  csv_path <- system.file("extdata", "collection_catalog.csv", package = "torchvision")

  # File should exist
  expect_true(file.exists(csv_path))

  # Should be readable as CSV
  csv_data <- read.csv(csv_path, stringsAsFactors = FALSE)
  expect_s3_class(csv_data, "data.frame")
  expect_gte(nrow(csv_data), 37)
})

#' Vision Transformer Implementation
#'
#' Vision Transformer (ViT) models implement the architecture proposed in the paper
#' [An Image is Worth 16x16 Words](https://arxiv.org/abs/2010.11929).
#' These models are designed for image classification tasks and operate by treating
#' image patches as tokens in a Transformer model.
#'
#' ## Model Variants and Performance (ImageNet-1k)
#' ```
#' | Model     | Top-1 Acc | Top-5 Acc | Params  | GFLOPS | File Size | Weights Used              | Notes                  |
#' |-----------|-----------|-----------|---------|--------|-----------|---------------------------|------------------------|
#' | vit_b_16  | 81.1%     | 95.3%     | 86.6M   | 17.56  | 346 MB    | IMAGENET1K_V1             | Base, 16x16 patches    |
#' | vit_b_32  | 75.9%     | 92.5%     | 88.2M   | 4.41   | 353 MB    | IMAGENET1K_V1             | Base, 32x32 patches    |
#' | vit_l_16  | 79.7%     | 94.6%     | 304.3M  | 61.55  | 1.22 GB   | IMAGENET1K_V1             | Large, 16x16 patches   |
#' | vit_l_32  | 77.0%     | 93.1%     | 306.5M  | 15.38  | 1.23 GB   | IMAGENET1K_V1             | Large, 32x32 patches   |
#' | vit_h_14  | 88.6%     | 98.7%     | 633.5M  | 1016.7 | 2.53 GB   | IMAGENET1K_SWAG_E2E_V1    | Huge, 14x14 patches    |
#' ```
#' - **TorchVision Recipe**: <https://github.com/pytorch/vision/tree/main/references/classification>
#' - **SWAG Recipe**: <https://github.com/facebookresearch/SWAG>
#'
#' **Weights Selection**:
#' - All models use the default `IMAGENET1K_V1` weights for consistency, stability, and official support from TorchVision.
#' - These are supervised weights trained on ImageNet-1k.
#' - For `vit_h_14`, the default weight is `IMAGENET1K_SWAG_E2E_V1`, pretrained on SWAG and fine-tuned on ImageNet.
#'
#' @inheritParams model_mobilenet_v2
#'
#' @family classification_model
#' @rdname model_vit
#' @name model_vit
NULL
vit_torchscript_urls <- list(
  vit_b_16 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/vit_b_16.pth", "dc04e8807c138d7ed4ae2754f59aec00", "330 MB"),
  vit_b_32 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/vit_b_32.pth", "6ea9578eb4e1cf9ffe1eec998ffebe14", "330 MB"),
  vit_l_16 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/vit_l_16.pth", "bc28a00cbcb48fb25318c7d542100915", "1.2 GB"),
  vit_l_32 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/vit_l_32.pth", "fc44a3ff3f6905a6868f55070d6106af", "1.2 GB"),
  vit_h_14 = c("https://torch-cdn.mlverse.org/models/vision/v2/models/vit_h_14.pth", "1f8d098982a80ccbe8dab3a5ff45752b", "2.4 GB")
)

load_vit_torchscript_model <- function(name, ...) {

  r <- vit_torchscript_urls[[name]]

  cli_inform("Model weights for {.cls {name}} (~{.emph {r[3]}}) will be downloaded and processed if not already available.")

  archive <- download_and_cache(r[1], prefix = name)

  if (tools::md5sum(archive) != r[2])
    runtime_error("Corrupt file! Delete the file in {archive} and try again.")

  model <- model_vit_base(name = name, pretrained = FALSE, progress = FALSE, ...)
  state_dict <- torch::load_state_dict(archive)

  names(state_dict) <- gsub("^conv_proj\\.", "patch_embed.proj.", names(state_dict))
  names(state_dict) <- gsub("^encoder\\.layers\\.encoder_layer_(\\d+)\\.ln_1\\.", "blocks.\\1.norm1.", names(state_dict))
  names(state_dict) <- gsub("^encoder\\.layers\\.encoder_layer_(\\d+)\\.self_attention\\.", "blocks.\\1.attn.", names(state_dict))
  names(state_dict) <- gsub("^encoder\\.layers\\.encoder_layer_(\\d+)\\.ln_2\\.", "blocks.\\1.norm2.", names(state_dict))
  names(state_dict) <- gsub("^encoder\\.layers\\.encoder_layer_(\\d+)\\.mlp\\.linear_1\\.", "blocks.\\1.mlp.0.", names(state_dict))
  names(state_dict) <- gsub("^encoder\\.layers\\.encoder_layer_(\\d+)\\.mlp\\.linear_2\\.", "blocks.\\1.mlp.3.", names(state_dict))
  names(state_dict) <- gsub("^encoder\\.pos_embedding", "encoder.pos_embedding", names(state_dict))
  names(state_dict) <- gsub("^encoder\\.ln\\.", "norm.", names(state_dict))
  names(state_dict) <- gsub("^heads\\.head\\.", "head.", names(state_dict))

  model$load_state_dict(state_dict)
  model
}

model_vit_base <- function(name, pretrained, progress, ...) {

  if (pretrained) {
    return(load_vit_torchscript_model(name,...))
  } else {
    args <- list(...)
    config <- switch(name,
      vit_b_16 = list(img_size = 224, patch_size = 16, embed_dim = 768, depth = 12, num_heads = 12, mlp_ratio = 4, num_classes = 1000),
      vit_b_32 = list(img_size = 224, patch_size = 32, embed_dim = 768, depth = 12, num_heads = 12, mlp_ratio = 4, num_classes = 1000),
      vit_l_16 = list(img_size = 224, patch_size = 16, embed_dim = 1024, depth = 24, num_heads = 16, mlp_ratio = 4, num_classes = 1000),
      vit_l_32 = list(img_size = 224, patch_size = 32, embed_dim = 1024, depth = 24, num_heads = 16, mlp_ratio = 4, num_classes = 1000),
      vit_h_14 = list(img_size = 518, patch_size = 14, embed_dim = 1280, depth = 32, num_heads = 16, mlp_ratio = 4, num_classes = 1000),
    )

    config[names(args)] <- args
    do.call(vit_model, config)
  }
}

#' @describeIn model_vit ViT-B/16 model (Base, 16×16 patch size)
#' @export
model_vit_b_16 <- function(pretrained = FALSE, progress = TRUE, ...) {
  model_vit_base("vit_b_16", pretrained, progress, ...)
}

#' @describeIn model_vit ViT-B/32 model (Base, 32×32 patch size)
#' @export
model_vit_b_32 <- function(pretrained = FALSE, progress = TRUE, ...) {
  model_vit_base("vit_b_32", pretrained, progress, ...)
}

#' @describeIn model_vit ViT-L/16 model (Base, 16×16 patch size)
#' @export
model_vit_l_16 <- function(pretrained = FALSE, progress = TRUE, ...) {
  model_vit_base("vit_l_16", pretrained, progress, ...)
}

#' @describeIn model_vit ViT-L/32 model (Base, 32×32 patch size)
#' @export
model_vit_l_32 <- function(pretrained = FALSE, progress = TRUE, ...) {
  model_vit_base("vit_l_32", pretrained, progress, ...)
}

#' @describeIn model_vit ViT-H/14 model (Base, 14×14 patch size)
#' @export
model_vit_h_14 <- function(pretrained = FALSE, progress = TRUE, ...) {
  model_vit_base("vit_h_14", pretrained, progress, ...)
}

vit_model <- torch::nn_module(
  classname = "vit_model",
  initialize = function(
    img_size = 224,
    patch_size = 16,
    in_chans = 3,
    num_classes = 1000,
    embed_dim = 768,
    depth = 12,
    num_heads = 12,
    mlp_ratio = 4,
    qkv_bias = TRUE,
    representation_size = NULL,
    dropout = 0.0
  ) {
    self$patch_embed <- patch_embed(img_size, patch_size, in_chans, embed_dim)
    num_patches <- self$patch_embed$num_patches

    self$class_token <- nn_parameter(torch_zeros(1, 1, embed_dim))
    self$encoder$pos_embedding <- nn_parameter(torch_zeros(1, num_patches + 1, embed_dim))
    self$pos_drop <- nn_dropout(p = dropout)

    self$blocks <- torch::nn_module_list(
      lapply(1:depth, function(i) {
        encoder_block(embed_dim, num_heads, mlp_ratio, qkv_bias, dropout)
      })
    )

    self$norm <- nn_layer_norm(embed_dim)

    self$head <- if (is.null(num_classes) || num_classes == 0) {
      nn_identity()
    } else {
      nn_linear(embed_dim, num_classes)
    }

    self$init_weights()
  },

  init_weights = function() {
    nn_init_trunc_normal_(self$encoder$pos_embedding, std = 0.02)
    nn_init_trunc_normal_(self$class_token, std = 0.02)
    for (m in self$modules) {
      if (inherits(m, "nn_linear")) {
        nn_init_trunc_normal_(m$weight, std = 0.02)
        if (!is.null(m$bias))
          nn_init_zeros_(m$bias)
      } else if (inherits(m, "nn_layer_norm")) {
        nn_init_zeros_(m$bias)
        nn_init_ones_(m$weight)
      }
      else if (inherits(m, "nn_multihead_attention")) {
        nn_init_xavier_uniform_(m$in_proj_weight)
        if (!is.null(m$in_proj_bias))
          nn_init_zeros_(m$in_proj_bias)
      }
    }
  },

  forward = function(x) {
    B <- x$size(1)
    x <- self$patch_embed(x)

    cls_tokens <- self$class_token$expand(c(B, -1, -1))
    x <- torch_cat(list(cls_tokens, x), dim = 2)
    x <- x + self$encoder$pos_embedding
    x <- self$pos_drop(x)

    for (i in seq_along(self$blocks)) {
      blk <- self$blocks[[i]]
      x <- blk(x)
    }

    x <- self$norm(x)
    return(self$head(x[ ,1, ]))
  }
)

patch_embed <- torch::nn_module(
  classname = "patch_embed",
  initialize = function(img_size, patch_size, in_chans, embed_dim) {
    self$img_size <- img_size
    self$patch_size <- patch_size
    self$grid_size <- as.integer(img_size / patch_size)
    self$num_patches <- self$grid_size^2

    self$proj <- nn_conv2d(
      in_channels = in_chans,
      out_channels = embed_dim,
      kernel_size = patch_size,
      stride = patch_size
    )
  },

  forward = function(x) {
    x <- self$proj(x)
    x <- torch_flatten(x, start_dim = 3, end_dim = 4)
    x <- x$transpose(2, 3)
    x
  }
)

encoder_block <- torch::nn_module(
  classname = "encoder_block",
  initialize = function(
    embed_dim = 768,
    num_heads = 12,
    mlp_ratio = 4,
    qkv_bias = TRUE,
    dropout = 0.0
  ) {

    self$norm1 <- nn_layer_norm(embed_dim)
    self$attn <- nn_multihead_attention(embed_dim, num_heads, bias = qkv_bias, dropout = dropout)
    self$norm2 <- nn_layer_norm(embed_dim)
    self$mlp <- nn_sequential(
      nn_linear(embed_dim, as.integer(embed_dim * mlp_ratio)),
      nn_gelu(),
      nn_dropout(p = dropout),
      nn_linear(as.integer(embed_dim * mlp_ratio), embed_dim),
      nn_dropout(p = dropout)
    )
  },

  forward = function(x) {
    x <- x + self$attn(self$norm1(x), self$norm1(x), self$norm1(x))[[1]]
    x <- x + self$mlp(self$norm2(x))
    return(x)
  }
)

test_that("tests for non-pretrained model_vit_b_16", {
  model <- model_vit_b_16()
  input <- torch::torch_randn(1, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  model <- model_vit_b_16(num_classes = 10)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model)
  gc()
})

test_that("tests for pretrained model_vit_b_16", {

  model <- model_vit_b_16(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  rm(model)
  gc()
})

test_that("tests for non-pretrained model_vit_b_32", {
  model <- model_vit_b_32()
  input <- torch::torch_randn(1, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  model <- model_vit_b_32(num_classes = 10)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model)
  gc()
})

test_that("tests for pretrained model_vit_b_32", {

  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_vit_b_32(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  rm(model)
  gc()
})

test_that("tests for non-pretrained model_vit_l_16", {
  model <- model_vit_l_16()
  input <- torch::torch_randn(1, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  model <- model_vit_l_16(num_classes = 10)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model)
  gc()
})

test_that("tests for pretrained model_vit_l_16", {

  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_vit_l_16(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  rm(model)
  gc()
})

test_that("tests for non-pretrained model_vit_l_32", {
  model <- model_vit_l_32()
  input <- torch::torch_randn(1, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  model <- model_vit_l_32(num_classes = 10)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model)
  gc()
})

test_that("tests for pretrained model_vit_l_32", {

  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_vit_l_32(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  rm(model)
  gc()
})

test_that("tests for model_vit_h_14", {
  model <- model_vit_h_14()
  input <- torch::torch_randn(1, 3, 518, 518)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  skip_if(Sys.info()[["sysname"]] == "Linux", "Skipping on Ubuntu CI")
  model <- model_vit_h_14(num_classes = 10)
  input <- torch::torch_randn(1, 3, 518, 518)
  out <- model(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model)
  gc()
})

test_that("tests for model_vit_h_14", {

  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
      "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  model <- model_vit_h_14(pretrained = TRUE)
  input <- torch::torch_randn(1, 3, 518, 518)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  rm(model)
  gc()
})

#' @include collection-rf100-doc.R
NULL

#' RoboFlow 100 Medical dataset Collection
#'
#' Loads one of the RoboFlow 100 Medical datasets
#' with per-dataset folders and train/valid/test splits.
#'
#' @inheritParams rf100_document_collection
#' @param dataset Dataset to select within \code{c("radio_signal",
#'  "rheumatology", "knee", "abdomen_mri", "brain_axial_mri",
#'  "gynecology_mri", "brain_tumor", "fracture")}.
#' @inherit rf100_document_collection return
#'
#' @examples
#' \dontrun{
#' ds <- rf100_medical_collection(
#'   dataset = "rheumatology",
#'   split = "test",
#'   transform = transform_to_tensor,
#'   download = TRUE
#' )
#' item <- ds[1]
#' boxed <- draw_bounding_boxes(item)
#' tensor_image_browse(boxed)
#' }
#'
#' @family detection_dataset
#' @export
rf100_medical_collection <- torch::dataset(
  name = "rf100_medical_collection",
  inherit = rf100_document_collection,

  resources = data.frame(
    dataset = rep(
      c("radio_signal", "rheumatology", "knee", "abdomen_mri", "brain_axial_mri",
        "gynecology_mri", "brain_tumor", "fracture"), each = 3
    ),
    split   = rep(c("train", "test", "valid"), times = 8),
    url = c(
      # radio_signal
      "https://huggingface.co/datasets/Francesco/radio-signal/resolve/main/data/train-00000-of-00001-0e1bb7466d6c3ca3.parquet",
      "https://huggingface.co/datasets/Francesco/radio-signal/resolve/main/data/test-00000-of-00001-cea246c736448c0f.parquet",
      "https://huggingface.co/datasets/Francesco/radio-signal/resolve/main/data/validation-00000-of-00001-0e172ea211bc5c25.parquet",
      # rheumatology
      "https://huggingface.co/datasets/Francesco/x-ray-rheumatology/resolve/main/data/train-00000-of-00001-3ca059ad007a94ea.parquet",
      "https://huggingface.co/datasets/Francesco/x-ray-rheumatology/resolve/main/data/test-00000-of-00001-8ce1659e856c6713.parquet",
      "https://huggingface.co/datasets/Francesco/x-ray-rheumatology/resolve/main/data/validation-00000-of-00001-0a5d512b2f7219ad.parquet",
      # knee
      "https://huggingface.co/datasets/Francesco/acl-x-ray/resolve/main/data/train-00000-of-00001-297d76f4f8e3f0d1.parquet",
      "https://huggingface.co/datasets/Francesco/acl-x-ray/resolve/main/data/test-00000-of-00001-771fc5699ba6259e.parquet",
      "https://huggingface.co/datasets/Francesco/acl-x-ray/resolve/main/data/validation-00000-of-00001-d64bcf3a8b32ec7d.parquet",
      # abdomen_mri
      "https://huggingface.co/datasets/Francesco/abdomen-mri/resolve/main/data/train-00000-of-00001-b5aa979424bb4685.parquet",
      "https://huggingface.co/datasets/Francesco/abdomen-mri/resolve/main/data/test-00000-of-00001-8b677ef1cabf7f16.parquet",
      "https://huggingface.co/datasets/Francesco/abdomen-mri/resolve/main/data/validation-00000-of-00001-73d9615650a3749b.parquet",
      # brain_axial_mri
      "https://huggingface.co/datasets/Francesco/axial-mri/resolve/main/data/train-00000-of-00001-62cf6bf015fef032.parquet",
      "https://huggingface.co/datasets/Francesco/axial-mri/resolve/main/data/test-00000-of-00001-7780878af8cf3e7b.parquet",
      "https://huggingface.co/datasets/Francesco/axial-mri/resolve/main/data/validation-00000-of-00001-bcd8291312ff472b.parquet",
      # gynecology_mri
      "https://huggingface.co/datasets/Francesco/gynecology-mri/resolve/main/data/train-00000-of-00001-ff598b0e3b7eb2c0.parquet",
      "https://huggingface.co/datasets/Francesco/gynecology-mri/resolve/main/data/test-00000-of-00001-ef538477a1b308ab.parquet",
      "https://huggingface.co/datasets/Francesco/gynecology-mri/resolve/main/data/validation-00000-of-00001-1d8f153a588ec0a9.parquet",
      # brain_tumor
      "https://huggingface.co/datasets/Francesco/brain-tumor-m2pbp/resolve/main/data/train-00000-of-00001-92b37a681420e786.parquet",
      "https://huggingface.co/datasets/Francesco/brain-tumor-m2pbp/resolve/main/data/test-00000-of-00001-bc5b44853d12ccfc.parquet",
      "https://huggingface.co/datasets/Francesco/brain-tumor-m2pbp/resolve/main/data/validation-00000-of-00001-4b3513c4ec0a5e8f.parquet",
      # fracture
      "https://huggingface.co/datasets/Francesco/bone-fracture-7fylg/resolve/main/data/train-00000-of-00001-26e4f0e80e263728.parquet",
      "https://huggingface.co/datasets/Francesco/bone-fracture-7fylg/resolve/main/data/test-00000-of-00001-9fe0a8c08ab79a8b.parquet",
      "https://huggingface.co/datasets/Francesco/bone-fracture-7fylg/resolve/main/data/validation-00000-of-00001-647af68e048aed16.parquet"
    ),
    md5 = c(
      # radio_signal
      "0e92cde0cae78019cdf736a0ec09cb6a",      "696458e584fb090a79790c46d8f0621d",      "08195336ff727c222fcd011d47164ec1",
     # rheumatology
      "3cdb356519def48577f4fbbd075c7328",      "0047b0c762e1434635a4ef78ff979e3d",      "bd6cbaa11f9d0e822881158e5f936e99",
      # knee
      "0e85f93fe793c25b8da3f245cd1968f6",      "7d0da6a924a3aabc21234da9423638a8",      "6afc61c17ad6da97f614ef558afca522",
      # abdomen_mri
      "6790fe6e214d5cf8ffef20bbd9a145bd",      "7b2985dbd628aaa2d6b226bbc13002ee",      "eab70beb5373212ba127c9dc9032214e",
      # brain_axial_mri
      "34d9fa3f0b86b75b98ddb3e111a78222",      "10df9ece0cfd4fd870cd6e77934f66c7",      "5e6a425dede86157a753cb8f62bb302f",
      # gynecology_mri
      "4705c18355707e921780c8c5c66f9233",      "a839a9242eeeed991623f376f83011b7",      "163cf422080c43db4a7d37fe7585d4f6",
      # brain_tumor
      "ac40f245c7c45f0eb9e4491e3452380f",      "0f1b174a26706f35d377dff81293f99b",      "f32abc88b31b06d77479f2aafc2a2062",
      # fracture
      "419080e7f1f400eb97f518c660083a32",      "5a2ee2edede004350478fc8feaa1458f",      "4a6c11900ff3f5b64fac71dba4d1f7f2"),
    size = c(51, 15, 7,  2.5, 0.6, 0.3,  46, 13, 6.5,  62, 15, 9,  3.4, 0.8, 0.5,
             50, 13, 7,  142, 40, 20,  9, 2, 1) * 1e6,
    stringsAsFactors = FALSE
  )
)

#' @include collection-rf100-doc.R
NULL

#' RF100 Peixos Segmentation Dataset
#'
#' Loads the Roboflow 100 "peixos" dataset for semantic segmentation.
#' "peixos" contains 3 splits of respectively 821 / 118 / 251 color images of size 640 x 640.
#' Segmentation masks are generated on-the-fly from polygon
#' annotations of the unique "fish" category.
#'
#' @inheritParams rf100_document_collection
#' @inheritParams tiny_imagenet_dataset
#' @return A torch dataset. Each element is a named list with:
#' - `x`: H × W × 3 array (use `transform_to_tensor()` in `transform` to get
#'   C × H × W tensor).
#' - `y`: a list with:
#'     - `masks`: boolean tensor of shape (1, H, W).
#'     - `labels`: integer vector with the class index (always 1 for "fish").
#'
#' The returned item is given class `image_with_segmentation_mask` so it can be
#' visualised with helpers like [draw_segmentation_masks()].
#'
#' @examples
#' \dontrun{
#' ds <- rf100_peixos_segmentation_dataset(
#'   split = "train",
#'   transform = transform_to_tensor,
#'   download = TRUE
#' )
#' item <- ds[1]
#' overlay <- draw_segmentation_masks(item)
#' tensor_image_browse(overlay)
#' }
#'
#' @family segmentation_dataset
#' @export
rf100_peixos_segmentation_dataset <- torch::dataset(
  name = "rf100_peixos_segmentation_dataset",
  inherit = coco_detection_dataset,
  resources = data.frame(
    dataset = "peixos",
    url = "https://huggingface.co/datasets/Francesco/peixos-fish/resolve/main/dataset.tar.gz",
    md5 = "0eb13ea40677178aed2fd47f153fabe2",
    size = "119 MB",
    stringsAsFactors = FALSE
  ),
  classes = c("fish"),

  initialize = function(
    split = c("train", "test", "valid"),
    root = tempdir(),
    download = FALSE,
    transform = NULL,
    target_transform = NULL
  ) {
    self$dataset <- "peixos"
    self$split <- match.arg(split)
    self$root <- fs::path_expand(root)
    self$transform <- transform
    self$target_transform <- target_transform

    self$data_dir <- fs::path(self$root, class(self)[[1]])
    self$image_dir <- fs::path(self$data_dir, self$split)
    fs::dir_create(self$image_dir, recurse = TRUE)
    self$annotation_file <- fs::path(self$image_dir, "_annotations.coco.json")
    self$archive_size <- self$resources$size

    if (download) {
      cli_inform("Dataset {.val {self$dataset}} split {.val {self$split}} of {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")
    }

    self$load_annotations()

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {
    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    archive <- download_and_cache(self$resources$url, prefix = class(self)[[1]])

    if (tools::md5sum(archive) != self$resources$md5) {
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")
    }
    archive_gz <- fs::path(self$data_dir, basename(archive))
    fs::file_copy(archive, archive_gz, overwrite = TRUE)
    utils::untar(archive_gz, exdir = self$data_dir)
    # workaround the `--strip-compoents = 8` not widely available
    fs::file_move(
      fs::path_filter(
        list.files(self$data_dir, recursive = TRUE, full.names = TRUE, pattern = "jpg|png|json"),
        regexp = self$split),
      self$image_dir)

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  load_annotations = function() {
    data <- jsonlite::fromJSON(self$annotation_file)
    # Select relevant cols in each listed dataframe
    ann  <- data$annotations[, c("image_id", "category_id", "segmentation")]
    imgs <- data$images[, c("id", "file_name")]
    cats <- data$categories[, c("id", "name")]

    # Left join annotations with images
    ann_df <- merge(ann, imgs,
                 by.x = "image_id", by.y = "id",
                 all.x = TRUE)

    # shift image index and category index to be 1-indexed
    ann_df$image_id <- ann_df$image_id + 1
    ann_df$category_id <- ann_df$category_id + 1
    ann_df$file_name <- fs::path(self$data_dir, self$split, ann_df$file_name)

    # Filter on existing images and drop _id columns
    has_image <- fs::file_exists(ann_df$file_name)
    self$annotation <- ann_df[has_image, c("image_id", "segmentation", "file_name", "category_id")]
  },

  .length = function() {
    nlevels(as.factor(self$annotation$image_id))
  },

  .getitem = function(index) {
    index_annotation <- self$annotation[self$annotation$image_id == index, ]
    x <- base_loader(index_annotation$file_name[1])

    H <- dim(x)[1]
    W <- dim(x)[2]
    mask <- torch::torch_zeros(c(H, W), dtype = torch::torch_bool())

    mask_lst <- lapply(index_annotation$segmentation, function(seg) {
        if (is.list(seg) && length(seg) > 0) {
          coco_polygon_to_mask(seg, height = H, width = W)
        } else if (is.numeric(seg) && length(seg) > 0) {
          coco_polygon_to_mask(list(seg), height = H, width = W)
        }
      })

    y <- list(masks = torch_stack(mask_lst)$sum(dim = 1, keepdim = TRUE), labels = 1L)

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }
    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    item <- list(x = x, y = y)
    class(item) <- c("image_with_segmentation_mask", class(item))
    item
  }
)

VGG <- torch::nn_module(
  "VGG",
  initialize = function(features, num_classes=1000, init_weights=TRUE) {
    self$features <- features
    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(7,7))
    self$classifier <- torch::nn_sequential(
      torch::nn_linear(512 * 7 * 7, 4096),
      torch::nn_relu(TRUE),
      torch::nn_dropout(),
      torch::nn_linear(4096, 4096),
      torch::nn_relu(TRUE),
      torch::nn_dropout(),
      torch::nn_linear(4096, num_classes),
    )

    if (init_weights)
      self$.initialize_weights()

  },
  forward = function(x) {
    x <- self$features(x)
    x <- self$avgpool(x)
    x <- torch::torch_flatten(x,start_dim = 2)
    x <- self$classifier(x)
    x
  },
  .initialize_weights = function() {

    for (m in self$modules) {

      if (inherits(m, "nn_conv2d")) {
        torch::nn_init_kaiming_normal_(m$weight, mode = "fan_out", nonlinearity="relu")
        if (!is.null(m$bias))
          torch::nn_init_constant_(m$bias, 0)
      } else if (inherits(m, "nn_batch_norm2d")) {
        torch::nn_init_constant_(m$weight, 1)
        torch::nn_init_constant_(m$bias, 0)
      } else if (inherits(m, "nn_linear")) {
        torch::nn_init_normal_(m$weight, 0, 0.01)
        torch::nn_init_constant_(m$bias, 0)
      }

    }

  }
)

vgg_make_layers <- function(cfg, batch_norm) {
  layers <- list()
  in_channels <- 3
  for (v in cfg) {

    if (v == "M") {

      layers[[length(layers) + 1]] <- torch::nn_max_pool2d(
        kernel_size = 2,
        stride = 2
      )

    } else {

      v <- as.integer(v)
      layers[[length(layers) + 1]] <- torch::nn_conv2d(
        in_channels = in_channels, out_channels = v,
        kernel_size = 3, padding = 1
      )

      if (batch_norm)
        layers[[length(layers) + 1]] <- torch::nn_batch_norm2d(num_features = v)

      layers[[length(layers) + 1]] <- torch::nn_relu(inplace = TRUE)
      in_channels <- v

    }

  }
  torch::nn_sequential(!!!layers)
}

vgg_cfgs <- list(
  "A" = list(64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"),
  "B" = list(64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"),
  "D" = list(64, 64, "M", 128, 128, "M", 256, 256, 256, "M", 512, 512, 512, "M", 512, 512, 512, "M"),
  "E" = list(64, 64, "M", 128, 128, "M", 256, 256, 256, 256, "M", 512, 512, 512, 512, "M", 512, 512, 512, 512, "M")
)

vgg_model_urls <- list(
  "vgg11"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg11.pth", "3afa0c0e9eecb80c0f4ebc4303e19f68", "~510 MB"),
  "vgg13"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg13.pth", "1733a72a8b22e37a57fb9cb9d4db3a74", "~510 MB"),
  "vgg16"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg16.pth", "2329d5702a1d78713179f6127a280fcd", "~530 MB"),
  "vgg19"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg19.pth", "6acea77a847bb1549171504b872efb96", "~575 MB"),
  "vgg11_bn"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg11_bn.pth", "76ca955d0aa3e61152bdd726949d8fed", "~510 MB"),
  "vgg13_bn"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg13_bn.pth", "612e11f29432ea79ae94a66b93db9109", "~510 MB"),
  "vgg16_bn"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg16_bn.pth", "5ee7dc31de40677452eec2fdffce7e17", "~530 MB"),
  "vgg19_bn"= c("https://torch-cdn.mlverse.org/models/vision/v2/models/vgg19_bn.pth", "7112c457070af40fdddda8e75f9b727b", "~575 MB")
)

vgg <- function(arch, cfg, batch_norm, pretrained, progress, ...) {

  args <- rlang::list2(...)

  if (pretrained)
    args$init_weights <- FALSE

  layers <- vgg_make_layers(cfg = vgg_cfgs[[cfg]], batch_norm = batch_norm)
  args <- append(args, list(features = layers))
  model <- do.call(VGG, args)

  if (pretrained) {
    r <- vgg_model_urls[[arch]]
    cli_inform("Model weights for {.cls {arch}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1])
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict)

  }

  model
}

#' VGG implementation
#'
#'
#' VGG models implementations based on
#' [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556)
#'
#' @param pretrained (bool): If TRUE, returns a model pre-trained on ImageNet
#' @param progress (bool): If TRUE, displays a progress bar of the download
#'   to stderr
#' @param ... other parameters passed to the VGG model implementation.
#'
#' @name model_vgg
#' @rdname model_vgg
NULL

#' @describeIn model_vgg VGG 11-layer model (configuration "A")
#'
#' @family classification_model
#' @export
model_vgg11 <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg('vgg11', 'A', FALSE, pretrained, progress, ...)
}

#' @describeIn model_vgg VGG 11-layer model (configuration "A") with batch normalization
#' @export
model_vgg11_bn <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg("vgg11_bn", "A", TRUE, pretrained, progress, ...)
}

#' @describeIn model_vgg VGG 13-layer model (configuration "B")
#' @export
model_vgg13 <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg('vgg13', 'B', FALSE, pretrained, progress, ...)
}

#' @describeIn model_vgg VGG 13-layer model (configuration "B") with batch normalization
#' @export
model_vgg13_bn <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg("vgg13_bn", "B", TRUE, pretrained, progress, ...)
}

#' @describeIn model_vgg VGG 13-layer model (configuration "D")
#' @export
model_vgg16 <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg('vgg16', 'D', FALSE, pretrained, progress, ...)
}

#' @describeIn model_vgg VGG 13-layer model (configuration "D") with batch normalization
#' @export
model_vgg16_bn <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg("vgg16_bn", "D", TRUE, pretrained, progress, ...)
}

#' @describeIn model_vgg VGG 19-layer model (configuration "E")
#' @export
model_vgg19 <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg('vgg19', 'E', FALSE, pretrained, progress, ...)
}

#' @describeIn model_vgg VGG 19-layer model (configuration "E") with batch normalization
#' @export
model_vgg19_bn <- function(pretrained = FALSE, progress = TRUE, ...) {
  vgg("vgg19_bn", "E", TRUE, pretrained, progress, ...)
}

library(roxygen2)


# Patch examples ----------------------------------------------------------
# we path examples so we add the `if` conditional.

.S3method("roxy_tag_parse", "roxy_tag_examples", function(x) {
  roxygen2::tag_examples(x)
})

.S3method("roxy_tag_rd", "roxy_tag_examples", function(x, base_path, env) {
  rd_section("examples", x$val)
})

.S3method("format", "rd_section_examples", function (x, ...) {
  value <- paste0(x$value, collapse = "\n")
  roxygen2:::rd_macro("examples",
                      c("if (torch_is_installed()) {", value, "}"),
                      space = TRUE
  )
})

# Subsection --------------------------------------------------------------

.S3method("roxy_tag_parse", "roxy_tag_subsection", function(x) {
  roxygen2::tag_markdown(x)
})

.S3method("roxy_tag_rd", "roxy_tag_subsection", function(x, base_path, env) {
  pieces <- stringr::str_split(x$val, ":", n = 2)[[1]]
  title <- stringr::str_split(pieces[1], "\n")[[1]]

  if (length(title) > 1) {
    roxygen2:::roxy_tag_warning(x, paste0(
      "Section title spans multiple lines.\n",
      "Did you forget a colon (:) at the end of the title?"
    ))
    return()
  }

  roxygen2:::rd_section_section(pieces[1], pieces[2])
})

.S3method("format", "rd_section_subsection", function(x, ...) {
  paste0(
    "\\subsection{", x$value$title, "}{\n", x$value$content, "\n}\n",
    collapse = "\n"
  )
})

# Params ------------------------------------------------------------------
# avoids using the \arguments{} so we can have two sections with parameters
# for different signatures.




devtools::document()

#' @include collection-rf100-doc.R
NULL

#' RoboFlow 100 Damages dataset Collection
#'
#' Loads one of the RoboFlow 100 Damage & Risk assesment datasets with
#' bounding box annotations for object-detection task.
#'
#' @inheritParams rf100_document_collection
#' @param dataset Dataset to select within \code{c("liquid_crystals", "solar_panel", "asbestos")}.
#' @inherit rf100_document_collection return
#'
#' @examples
#' \dontrun{
#' ds <- rf100_damage_collection(
#'   dataset = "solar_panel",
#'   split = "test",
#'   transform = transform_to_tensor,
#'   download = TRUE
#' )
#' item <- ds[1]
#' boxed <- draw_bounding_boxes(item)
#' tensor_image_browse(boxed)
#' }
#'
#' @family detection_dataset
#' @export
rf100_damage_collection <- torch::dataset(
  name = "rf100_damage_collection",
  inherit = rf100_document_collection,

  resources = data.frame(
    dataset = rep(c("liquid_crystals", "solar_panel", "asbestos"),each = 3),
    split   = rep(c("train", "test", "valid"), times = 3),
    url = c(
      # liquid_crystals
      "https://huggingface.co/datasets/Francesco/4-fold-defect/resolve/main/data/train-00000-of-00001-b438dc70da70c5f3.parquet",
      "https://huggingface.co/datasets/Francesco/4-fold-defect/resolve/main/data/test-00000-of-00001-0c18bd0b95b38167.parquet",
      "https://huggingface.co/datasets/Francesco/4-fold-defect/resolve/main/data/validation-00000-of-00001-58a56f521b5aaffb.parquet",
      # solar_panel
      "https://huggingface.co/datasets/Francesco/solar-panels-taxvb/resolve/main/data/train-00000-of-00001-70141d825b15d748.parquet",
      "https://huggingface.co/datasets/Francesco/solar-panels-taxvb/resolve/main/data/test-00000-of-00001-6828d49996604be7.parquet",
      "https://huggingface.co/datasets/Francesco/solar-panels-taxvb/resolve/main/data/validation-00000-of-00001-11cffae851c69d23.parquet",
      # asbestos
      "https://huggingface.co/datasets/Francesco/asbestos/resolve/main/data/train-00000-of-00001-4bdd9b41e9daaf91.parquet",
      "https://huggingface.co/datasets/Francesco/asbestos/resolve/main/data/test-00000-of-00001-283c6964a2943199.parquet",
      "https://huggingface.co/datasets/Francesco/asbestos/resolve/main/data/validation-00000-of-00001-6e5e04a76562dda8.parquet"
    ),
    md5 = c(
      # liquid_crystals
      "e25e75da04c0f8199a9b47215b6e3ac2",      "b9729b2776cee8704f1ddeff0597378d",      "5cbf6f189b2449969f47dfbb746ecca3",
      # solar_panel
      "a6078675c6c4fe50a74ed3fdbe781195",      "ff8983ca30ab4a5eac02ba516d0a5345",      "8f02bfbab9d2c59c1d7a57089c159e56",
      # asbestos
      "f53af703c4ce594c3950d7e732003f2d",      "8e99904ce49e7f0e830735fb22986868",      "4fef507d057690d1a55fa043696248cc"
    ),
    size = c(21.5, 5.7, 1.4,   8.5, 2.3, 1.5,  28, 7.8, 4) * 1e6
  )
)

#' DeepLabV3 Semantic Segmentation Models
#'
#' @description
#' Semantic segmentation models implementing the DeepLabV3 architecture from
#' [Rethinking Atrous Convolution for Semantic Image Segmentation](https://arxiv.org/abs/1706.05587).
#' These models use Atrous Spatial Pyramid Pooling (ASPP) to capture multi-scale
#' context, and are available with ResNet-50 and ResNet-101 backbones.
#'
#' ## Available Models
#' - `model_deeplabv3_resnet50()`
#' - `model_deeplabv3_resnet101()`
#'
#' ## Model Variants and Performance (COCO val2017, VOC labels)
#'
#' All models are trained on a 20-class subset of COCO that corresponds to
#' Pascal VOC categories, plus background (21 classes total).
#'
#' ```
#' | Model                     | mIoU  | Pixel Acc | Params | GFLOPS | File Size | Weights Used              |
#' |---------------------------|-------|-----------|--------|--------|-----------|---------------------------|
#' | model_deeplabv3_resnet50  | 66.4% | 92.4%     | 42.0M  | 178.72 | 161 MB    | COCO_WITH_VOC_LABELS_V1   |
#' | model_deeplabv3_resnet101 | 67.4% | 92.4%     | 61.0M  | 258.74 | 233 MB    | COCO_WITH_VOC_LABELS_V1   |
#' ```
#'
#' ## Weights Selection
#' - All models use `COCO_WITH_VOC_LABELS_V1` weights, trained on COCO with the
#'   20 Pascal VOC categories (+ background = 21 classes).
#' - Backbone weights default to `IMAGENET1K_V1` (supervised ImageNet-1k) when
#'   `pretrained = FALSE` and `pretrained_backbone = TRUE`.
#' - When `pretrained = TRUE`, backbone weights are overridden by the full
#'   segmentation model weights and `pretrained_backbone` is ignored.
#' - The auxiliary classifier branch (`aux_loss`) is automatically enabled when
#'   loading pretrained weights; set explicitly when training from scratch.
#'
#' ## Input Format
#' Models expect input tensors of shape `(batch_size, 3, H, W)`, normalized
#' with ImageNet mean `c(0.485, 0.456, 0.406)` and std `c(0.229, 0.224, 0.225)`.
#' Training resolution is 520x520.
#'
#' ## Output Format
#' Returns a named list with:
#' - `$out` — main segmentation logits, shape `(batch, num_classes, H, W)`
#' - `$aux` — auxiliary logits from an intermediate backbone layer (only when `aux_loss = TRUE`)
#'
#' @inheritParams model_resnet18
#' @param num_classes Integer. Number of output segmentation classes including
#'   background. Default: `21` (Pascal VOC). Set to `NULL` to infer from
#'   pretrained weights.
#' @param aux_loss Logical or `NULL`. If `TRUE`, adds an auxiliary FCN classifier
#'   head at an intermediate backbone layer, used as a secondary loss during
#'   training. If `NULL` (default), inferred from pretrained weights.
#' @param pretrained_backbone Logical. If `TRUE` and `pretrained = FALSE`, loads
#'   `IMAGENET1K_V1` weights for the backbone only. Ignored when `pretrained = TRUE`.
#'   Default: `TRUE`.
#'
#' @examples
#' \dontrun{
#' library(magrittr)
#' norm_mean <- c(0.485, 0.456, 0.406)
#' norm_std  <- c(0.229, 0.224, 0.225)
#'
#' url <- paste0("https://upload.wikimedia.org/wikipedia/commons/thumb/",
#'            "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg")
#' img <- base_loader(url)
#'
#' input <- img %>%
#'   transform_to_tensor() %>%
#'   transform_resize(c(520, 520)) %>%
#'   transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)    # Add batch dimension: (1, 3, H, W)
#'
#' # --- ResNet-50 backbone ---
#' model <- model_deeplabv3_resnet50(pretrained = TRUE)
#' model$eval()
#' output <- model(batch)
#'
#' segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
#' tensor_image_browse(segmented)
#'
#' # Show most frequent class
#' mask_id <- output$out$argmax(dim = 2)  # (1, H, W)
#' class_contingency_with_background <- mask_id$view(-1)$bincount()
#' class_contingency_with_background[1] <- 0L # we clean the counter for background class id 1
#' top_class_index <- class_contingency_with_background$argmax()$item()
#' cli::cli_inform("Majority class {.pkg ResNet-50}: {.emph {pascal_voc_classes(top_class_index)}}")
#'
#' # --- ResNet-101 backbone ---
#' model <- model_deeplabv3_resnet101(pretrained = TRUE)
#' model$eval()
#' output <- model(batch)
#'
#' segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
#' tensor_image_browse(segmented)
#'
#' # Show most frequent class
#' mask_id <- output$out$argmax(dim = 2)  # (1, H, W)
#' class_contingency_with_background <- mask_id$view(-1)$bincount()
#' class_contingency_with_background[1] <- 0L # we clean the counter for background class id 1
#' top_class_index <- class_contingency_with_background$argmax()$item()
#' cli::cli_inform("Majority class {.pkg ResNet-50}: {.emph {pascal_voc_classes(top_class_index)}}")
#' }
#'
#' @importFrom torch nn_module
#' @family semantic_segmentation_model
#' @rdname model_deeplabv3
#' @name model_deeplabv3
NULL

deeplabv3_model_urls <- list(
  deeplabv3_resnet50_coco = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/deeplabv3_resnet50_coco.pth",
    "9bc0953c76291fff296211b0eba04164",
    "168 MB"
  ),
  deeplabv3_resnet101_coco = c(
    "https://torch-cdn.mlverse.org/models/vision/v2/models/deeplabv3_resnet101_coco.pth",
    "bb73d96704937621d782af73125e6282",
    "172 MB"
  )
)

deeplabv3_meta <- list(
  classes = pascal_voc_classes(),
  class_to_idx = setNames(seq_along(pascal_voc_classes()) - 1, pascal_voc_classes())
)


# ASPP module matching PyTorch
aspp_module <- torch::nn_module(
  "ASPP",
  initialize = function(in_channels, out_channels, atrous_rates) {
    self$convs <- torch::nn_module_list()

    self$convs$append(torch::nn_sequential(
      torch::nn_conv2d(in_channels, out_channels, kernel_size = 1, bias = FALSE),
      torch::nn_batch_norm2d(out_channels),
      torch::nn_relu()
    ))

    for (rate in atrous_rates) {
      self$convs$append(torch::nn_sequential(
        torch::nn_conv2d(in_channels, out_channels, kernel_size = 3, padding = rate, dilation = rate, bias = FALSE),
        torch::nn_batch_norm2d(out_channels),
        torch::nn_relu()
      ))
    }

    self$convs$append(torch::nn_sequential(
      torch::nn_adaptive_avg_pool2d(output_size = c(1, 1)),
      torch::nn_conv2d(in_channels, out_channels, kernel_size = 1, bias = FALSE),
      torch::nn_batch_norm2d(out_channels),
      torch::nn_relu()
    ))

    self$project <- torch::nn_sequential(
      torch::nn_conv2d((length(atrous_rates) + 2) * out_channels, out_channels, kernel_size = 1, bias = FALSE),
      torch::nn_batch_norm2d(out_channels),
      torch::nn_relu(),
      torch::nn_dropout(0.5)
    )
  },
  forward = function(x) {
    input_size <- x$shape[3:4]  # Get height and width dimensions
    res <- list()

    for (i in 1:(length(self$convs) - 1)) {
      res[[i]] <- self$convs[[i]](x)
    }

    global_feat <- self$convs[[length(self$convs)]](x)
    target_size <- as.integer(input_size)
    global_feat <- nnf_interpolate(global_feat, size = target_size, mode = "bilinear", align_corners = FALSE)
    res[[length(res) + 1]] <- global_feat

    x <- torch_cat(res, dim = 2)
    self$project(x)
  }
)

# Main classifier head
deeplab_head <- function(in_channels, num_classes) {
  torch::nn_sequential(
    aspp_module(in_channels, 256, atrous_rates = c(12, 24, 36)),
    torch::nn_conv2d(256, 256, kernel_size = 3, padding = 1, bias = FALSE),
    torch::nn_batch_norm2d(256),
    torch::nn_relu(),
    torch::nn_conv2d(256, num_classes, kernel_size = 1)
  )
}

# Auxiliary classifier head (simpler, operates on layer3 features)
aux_classifier_head <- function(in_channels, num_classes) {
  torch::nn_sequential(
    torch::nn_conv2d(in_channels, 256, kernel_size = 3, padding = 1, bias = FALSE),
    torch::nn_batch_norm2d(256),
    torch::nn_relu(),
    torch::nn_dropout(0.1),
    torch::nn_conv2d(256, num_classes, kernel_size = 1)
  )
}

# Updated DeepLabV3 wrapper with auxiliary classifier
DeepLabV3 <- torch::nn_module(
  "DeepLabV3",
  initialize = function(backbone, classifier, aux_classifier = NULL) {
    self$backbone <- backbone
    self$classifier <- classifier
    self$aux_classifier <- aux_classifier
  },
  forward = function(x) {
    input_shape <- x$shape[3:4]  # Get height and width dimensions

    x <- self$backbone$conv1(x)
    x <- self$backbone$bn1(x)
    x <- self$backbone$relu(x)
    x <- self$backbone$maxpool(x)
    x <- self$backbone$layer1(x)
    x <- self$backbone$layer2(x)
    x <- self$backbone$layer3(x)

    aux_x <- x

    x <- self$backbone$layer4(x)

    # Main classifier
    x <- self$classifier(x)
    target_size <- as.integer(input_shape)
    x <- nnf_interpolate(x, size = target_size, mode = "bilinear", align_corners = FALSE)

    result <- list(out = x)

    # Auxiliary classifier (if present)
    if (!is.null(self$aux_classifier)) {
      aux_out <- self$aux_classifier(aux_x)
      aux_out <- nnf_interpolate(aux_out, size = target_size, mode = "bilinear", align_corners = FALSE)
      result$aux <- aux_out
    }

    result
  }
)

# Model constructor
deeplabv3_resnet_factory <- function(arch, block, layers, pretrained, progress,
                                     num_classes, aux_loss = NULL,
                                     pretrained_backbone = FALSE, ...) {

  validate_num_classes(num_classes, pretrained)

  if (is.null(aux_loss))
    aux_loss <- FALSE

  if (pretrained && pretrained_backbone)
    cli_warn("`pretrained_backbone` ignored when `pretrained = TRUE`." )

  backbone_arch <- sub("deeplabv3_", "", arch)

  if (pretrained_backbone && !pretrained) {
    backbone <- .resnet(backbone_arch, block, layers, pretrained = TRUE,
                        progress = progress,
                        replace_stride_with_dilation = c(FALSE, TRUE, TRUE), ...)
  } else {
    backbone <- resnet(block, layers,
                       replace_stride_with_dilation = c(FALSE, TRUE, TRUE), ...)
  }
  backbone$fc <- torch::nn_identity()
  backbone$avgpool <- torch::nn_identity()

  classifier <- deeplab_head(2048, num_classes)

  aux_classifier <- NULL
  if (aux_loss) {
    aux_classifier <- aux_classifier_head(1024, num_classes)
  }

  model <- DeepLabV3(backbone, classifier, aux_classifier)
  model$classes <- deeplabv3_meta$classes
  model$class_to_idx <- deeplabv3_meta$class_to_idx

  if (pretrained) {
    info <- deeplabv3_model_urls[[paste0(arch, "_coco")]]
    cli_inform("Downloading {.cls {arch}} pretrained weights (~{.emph {info[3]}}) ...")
    state_dict_path <- download_and_cache(info[1], prefix = "deeplabv3")
    if (tools::md5sum(state_dict_path) != info[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")
    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict)
  }
  model
}

#' @describeIn model_deeplabv3 DeepLabV3 with ResNet-50 backbone
#' @export
model_deeplabv3_resnet50 <- function(pretrained = FALSE, progress = TRUE,
                                     num_classes = 21, aux_loss = NULL,
                                     pretrained_backbone = FALSE, ...) {
  deeplabv3_resnet_factory(
    "deeplabv3_resnet50", bottleneck, c(3, 4, 6, 3),
    pretrained, progress, num_classes, aux_loss, pretrained_backbone, ...
  )
}

#' @describeIn model_deeplabv3 DeepLabV3 with ResNet-101 backbone
#' @export
model_deeplabv3_resnet101 <- function(pretrained = FALSE, progress = TRUE,
                                      num_classes = 21, aux_loss = NULL,
                                      pretrained_backbone = FALSE, ...) {
  deeplabv3_resnet_factory(
    "deeplabv3_resnet101", bottleneck, c(3, 4, 23, 3),
    pretrained, progress, num_classes, aux_loss, pretrained_backbone, ...
  )
}

#' MobileNetV2 Model
#'
#' Constructs a MobileNetV2 architecture from
#' [MobileNetV2: Inverted Residuals and Linear Bottlenecks](https://arxiv.org/abs/1801.04381).
#'
#' @inheritParams model_resnet18
#' @param ... Other parameters passed to the model implementation.
#'
#' @family classification_model
#' @export
model_mobilenet_v2 <- function(pretrained = FALSE, progress = TRUE, ...) {
  # resources
  r <- c("https://torch-cdn.mlverse.org/models/vision/v2/models/mobilenet_v2.pth", "06af6062e42ad3c80e430219a6560ca0", "~13 MB")
  model <- mobilenet_v2(...)

  if (pretrained) {
    cli_inform("Model weights for {.cls {class(model)[1]}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "mobilenet")
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict)
  }

  model
}


mobilenet_v2 <- torch::nn_module(
  "mobilenet_v2",
  initialize = function(
    num_classes = 1000,
    width_mult = 1.0,
    inverted_residual_setting = NULL,
    round_nearest = 8,
    block = NULL,
    norm_layer = NULL
  ) {

    if (is.null(block))
      block <- inverted_residual

    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    input_channel <- 32
    last_channel <- 1280

    if (is.null(inverted_residual_setting)) {
      inverted_residual_setting <- list(
        # t, c, n, s
        c(1, 16, 1, 1),
        c(6, 24, 2, 2),
        c(6, 32, 3, 2),
        c(6, 64, 4, 2),
        c(6, 96, 3, 1),
        c(6, 160, 3, 2),
        c(6, 320, 1, 1)
      )
    }

    # only check the first element, assuming user knows t,c,n,s are required
    if (length(inverted_residual_setting) == 0 || length(inverted_residual_setting[[1]]) != 4)
      cli_abort("{.var inverted_residual_setting} should be non-empty or a 4-element list, got {.val {inverted_residual_setting}}")


    # building first layer
    input_channel <- mobilenetv2.make_divisible(input_channel * width_mult, round_nearest)
    self$last_channel <- mobilenetv2.make_divisible(last_channel * max(1.0, width_mult), round_nearest)

    features <- list(conv_bn_activation(3, input_channel, stride = 2, norm_layer = norm_layer))

    # building inverted residual blocks
    for (i in inverted_residual_setting) {
      names(i) <- c("t", "c", "n", "s")
      i <- as.list(i)
      output_channel <- mobilenetv2.make_divisible(i$c * width_mult, round_nearest)
      for (k in 0:(i$n -1)) {
        stride <-  if (k == 0) i$s else 1
        features[[length(features) + 1]] <- block(
          input_channel,
          output_channel,
          stride,
          expand_ratio=i$t,
          norm_layer=norm_layer
        )

        input_channel <- output_channel
      }
    }

    # building last several layers
    features[[length(features) + 1]] <- conv_bn_activation(
      input_channel,
      self$last_channel,
      kernel_size=1,
      norm_layer=norm_layer
    )

    # make it nn.Sequential
    self$features = torch::nn_sequential(!!!features)

    # building classifier
    self$classifier = torch::nn_sequential(
      torch::nn_dropout(0.2),
      torch::nn_linear(self$last_channel, num_classes)
    )

    # weight initialization
    for (m in self$modules) {
      if (inherits(m, "nn_conv2d")) {
        torch::nn_init_kaiming_normal_(m$weight, mode='fan_out')
        if (!is.null(m$bias)) {
          torch::nn_init_zeros_(m$bias)
        }
      } else if (inherits(m, c("nn_batch_norm2d", "nn_group_norm"))) {
        torch::nn_init_ones_(m$weight)
        torch::nn_init_zeros_(m$bias)
      } else if (inherits(m, c("nn_linear"))) {
        torch::nn_init_normal_(m$weight, 0, 0.01)
        torch::nn_init_zeros_(m$bias)
      }
    }

  },
  forward = function(x) {
    x <- self$features(x)
    # Cannot use "squeeze" as batch-size can be 1
    x <- torch::nnf_adaptive_avg_pool2d(x, c(1, 1))
    x <- torch::torch_flatten(x, start_dim = 2L)
    x <- self$classifier(x)
    x
  }
)

inverted_residual <- torch::nn_module(
  "inverted_residual",
  initialize = function(inp, oup, stride, expand_ratio, norm_layer = NULL) {

    self$stride <- stride

    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    hidden_dim <- as.integer(round(inp * expand_ratio))
    self$use_res_connect = self$stride == 1 && inp == oup

    layers <- list()

    if (expand_ratio != 1) {
      layers[[length(layers) + 1]] <- conv_bn_activation(
        inp,
        hidden_dim,
        kernel_size=1,
        norm_layer=norm_layer
      )
    }

    layers <- append(layers, list(
      conv_bn_activation(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim, norm_layer=norm_layer),
      # pw-linear
      torch::nn_conv2d(hidden_dim, oup, 1, 1, 0, bias=FALSE),
      norm_layer(oup)
    ))

    self$conv <- torch::nn_sequential(!!!layers)
    self$out_channels <- oup
    self$.is_cn <- stride > 1
  },
  forward = function(x) {
    if (self$use_res_connect)
      x + self$conv(x)
    else
      self$conv(x)
  }
)

sequential <- torch::nn_module(
  classname = "nn_sequential",
  initialize = function(...) {
    modules <- rlang::list2(...)
    for (i in seq_along(modules)) {
      self$add_module(name = i - 1, module = modules[[i]])
    }
  }, forward = function(input) {
    for (module in private$modules_) {
      input <- module(input)
    }
    input
  }
)

conv_bn_activation <- torch::nn_module(
  "conv_bn_activation",
  inherit = sequential,
  initialize = function(
    in_planes,
    out_planes,
    kernel_size = 3,
    stride = 1,
    groups = 1,
    norm_layer = NULL,
    activation_layer = NULL,
    dilation = 1
  ) {

    padding <- (kernel_size - 1) %/% 2 * dilation

    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    if (is.null(activation_layer))
      activation_layer <- torch::nn_relu6

    super$initialize(
      torch::nn_conv2d(
        in_planes, out_planes, kernel_size, stride, padding,
        dilation=dilation,
        groups=groups,
        bias=FALSE
      ),
      norm_layer(out_planes),
      activation_layer(inplace=TRUE)
    )

    self$out_channels <- out_planes
  }
)

#' This function is taken from the original tf repo.
#' It ensures that all layers have a channel number that is divisible by 8
#' It can be seen here:
#' https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
#' @noRd
mobilenetv2.make_divisible <- function(v, divisor, min_value = NULL) {

  if (is.null(min_value))
    min_value <- divisor

  new_v <- max(min_value, as.integer(v + divisor/2) %/% divisor * divisor)

  # Make sure that round down does not go down by more than 10%.
  if (new_v < 0.9 * v)
    new_v <- new_v + divisor

  new_v
}

test_that("mobilenetv2 works", {
  model <- model_mobilenet_v2()
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))

  model <- model_mobilenet_v2(pretrained = TRUE)
  torch::torch_manual_seed(1)
  input <- torch::torch_randn(1, 3, 256, 256)
  out <- model(input)

  expect_tensor_shape(out, c(1, 1000))
  # expect_equal_to_r(out[1,1], -1.1959798336029053, tolerance = 1e-5) # value taken from pytorch
})

test_that("we can prune head of mobilenetv2 moels", {
  mobilenet <- model_mobilenet_v2(pretrained=TRUE)

  expect_no_error(prune <- nn_prune_head(mobilenet, 1))
  expect_s3_class(prune, "nn_sequential")
  expect_length(prune, 1)
  expect_s3_class(prune[[length(prune)]], "nn_sequential")

  input <- torch::torch_randn(1, 3, 256, 256)
  out <- prune(input)
  expect_tensor_shape(out, c(1, 1280, 8, 8))
})

Inception3 <- torch::nn_module(
  "Inception3",
  initialize = function(num_classes = 1000,
                        aux_logits = TRUE,
                        transform_input = FALSE,
                        inception_blocks = NULL,
                        init_weights = NULL,
                        dropout = 0.5) {
    if (is.null(inception_blocks))
      inception_blocks <- list(
        BasicConv2d, InceptionA, InceptionB, InceptionC, InceptionD, InceptionE, InceptionAux
      )

    if (is.null(init_weights)) {
      cli::cli_warn(paste(collapse = "", gettext(c(
        "The default weight initialization of inception_v3 will be changed in future releases of ",
        "torchvision. If you wish to keep the old behavior (which leads to long initialization times",
        " due to scipy/scipy#11299), please set init_weights={.val TRUE}."
      ))))
      init_weights <- TRUE
    }

    if (length(inception_blocks) != 7) {
      cli::cli_abort(gettext("length of {.arg inception_blocks} should be 7 instead of {length(inception_blocks)}"))
    }

    conv_block <- inception_blocks[[0+1]]
    inception_a <- inception_blocks[[1+1]]
    inception_b <- inception_blocks[[2+1]]
    inception_c <- inception_blocks[[3+1]]
    inception_d <- inception_blocks[[4+1]]
    inception_e <- inception_blocks[[5+1]]
    inception_aux <- inception_blocks[[6+1]]

    self$aux_logits <- aux_logits
    self$transform_input <- transform_input
    self$Conv2d_1a_3x3 <- conv_block(3, 32, kernel_size=3, stride=2)
    self$Conv2d_2a_3x3 <- conv_block(32, 32, kernel_size=3)
    self$Conv2d_2b_3x3 <- conv_block(32, 64, kernel_size=3, padding=1)
    self$maxpool1 <- torch::nn_max_pool2d(kernel_size=3, stride=2)
    self$Conv2d_3b_1x1 <- conv_block(64, 80, kernel_size=1)
    self$Conv2d_4a_3x3 <- conv_block(80, 192, kernel_size=3)
    self$maxpool2 <- torch::nn_max_pool2d(kernel_size=3, stride=2)
    self$Mixed_5b <- inception_a(192, pool_features=32)
    self$Mixed_5c <- inception_a(256, pool_features=64)
    self$Mixed_5d <- inception_a(288, pool_features=64)
    self$Mixed_6a <- inception_b(288)
    self$Mixed_6b <- inception_c(768, channels_7x7=128)
    self$Mixed_6c <- inception_c(768, channels_7x7=160)
    self$Mixed_6d <- inception_c(768, channels_7x7=160)
    self$Mixed_6e <- inception_c(768, channels_7x7=192)

    self$AuxLogits <- NULL

    if (aux_logits)
      self$AuxLogits <- inception_aux(768, num_classes)

    self$Mixed_7a <- inception_d(768)
    self$Mixed_7b <- inception_e(1280)
    self$Mixed_7c <- inception_e(2048)
    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(1, 1))
    self$dropout <- torch::nn_dropout(p=dropout)
    self$fc <- torch::nn_linear(2048, num_classes)

    if (init_weights) {
      for (m in self$modules) {
        if (inherits(m, "nn_conv2d") || inherits(m, "nn_linear")) {
          stddev <- if (!is.null(m$stddev)) m$stddev else 0.1
          torch::nn_init_trunc_normal_(m$weight, mean = 0, std = stddev, a = -2, b = -2)
        } else if (inherits(m, "nn_batch_norm2d")) {
          torch::nn_init_constant_(m$weight, 1)
          torch::nn_init_constant_(m$bias, 0)
        }
      }
    }

  },
  .transform_input = function(x) {
    if (self$transform_input) {
      x_ch0 <- torch::torch_unsqueeze(x[,1], 2) * (0.229 / 0.5) + (0.485 - 0.5) / 0.5
      x_ch1 <- torch::torch_unsqueeze(x[,2], 2) * (0.224 / 0.5) + (0.456 - 0.5) / 0.5
      x_ch2 <- torch::torch_unsqueeze(x[,3], 2) * (0.225 / 0.5) + (0.406 - 0.5) / 0.5
      x <- torch_cat(list(x_ch0, x_ch1, x_ch2), 2)
    }
    x
  },
  .forward = function(x) {
    # N x 3 x 299 x 299
    x <- self$Conv2d_1a_3x3(x)
    # N x 32 x 149 x 149
    x <- self$Conv2d_2a_3x3(x)
    # N x 32 x 147 x 147
    x <- self$Conv2d_2b_3x3(x)
    # N x 64 x 147 x 147
    x <- self$maxpool1(x)
    # N x 64 x 73 x 73
    x <- self$Conv2d_3b_1x1(x)
    # N x 80 x 73 x 73
    x <- self$Conv2d_4a_3x3(x)
    # N x 192 x 71 x 71
    x <- self$maxpool2(x)
    # N x 192 x 35 x 35
    x <- self$Mixed_5b(x)
    # N x 256 x 35 x 35
    x <- self$Mixed_5c(x)
    # N x 288 x 35 x 35
    x <- self$Mixed_5d(x)
    # N x 288 x 35 x 35
    x <- self$Mixed_6a(x)
    # N x 768 x 17 x 17
    x <- self$Mixed_6b(x)
    # N x 768 x 17 x 17
    x <- self$Mixed_6c(x)
    # N x 768 x 17 x 17
    x <- self$Mixed_6d(x)
    # N x 768 x 17 x 17
    x <- self$Mixed_6e(x)
    # N x 768 x 17 x 17
    if (!is.null(self$AuxLogits) && self$training) {
      aux <- self$AuxLogits(x)
    } else {
      aux <- NULL
    }
    # N x 768 x 17 x 17
    x <- self$Mixed_7a(x)
    # N x 1280 x 8 x 8
    x <- self$Mixed_7b(x)
    # N x 2048 x 8 x 8
    x <- self$Mixed_7c(x)
    # N x 2048 x 8 x 8
    # Adaptive average pooling
    x <- self$avgpool(x)
    # N x 2048 x 1 x 1
    x <- self$dropout(x)
    # N x 2048 x 1 x 1
    x <- torch::torch_flatten(x, start_dim = 2)
    # N x 2048
    x <- self$fc(x)
    # N x 1000 (num_classes)

    list(logits = x, aux_logits = aux)
  },
  forward = function(x) {
    x <- self$.transform_input(x)
    out <- self$.forward(x)

    if (self$training && self$aux_logits) {
      out
    } else {
      out$logits
    }
  }
)

InceptionA <- torch::nn_module(
  "InceptionA",
  initialize = function(in_channels, pool_features, conv_block = NULL) {
    if (is.null(conv_block)) {
      conv_block <- BasicConv2d
    }

    self$branch1x1 <- conv_block(in_channels, 64, kernel_size=1)
    self$branch5x5_1 <- conv_block(in_channels, 48, kernel_size=1)
    self$branch5x5_2 <- conv_block(48, 64, kernel_size=5, padding=2)
    self$branch3x3dbl_1 <- conv_block(in_channels, 64, kernel_size=1)
    self$branch3x3dbl_2 <- conv_block(64, 96, kernel_size=3, padding=1)
    self$branch3x3dbl_3 <- conv_block(96, 96, kernel_size=3, padding=1)
    self$branch_pool <- conv_block(in_channels, pool_features, kernel_size=1)
  },
  .forward = function(x) {
    branch1x1 <- self$branch1x1(x)

    branch5x5 <- self$branch5x5_1(x)
    branch5x5 <- self$branch5x5_2(branch5x5)

    branch3x3dbl <- self$branch3x3dbl_1(x)
    branch3x3dbl <- self$branch3x3dbl_2(branch3x3dbl)
    branch3x3dbl <- self$branch3x3dbl_3(branch3x3dbl)

    branch_pool <- torch::nnf_avg_pool2d(x, kernel_size=3, stride=1, padding=1)
    branch_pool <- self$branch_pool(branch_pool)

    outputs <- list(branch1x1, branch5x5, branch3x3dbl, branch_pool)
    outputs
  },
  forward = function(x) {
    outputs <- self$.forward(x)
    torch::torch_cat(outputs, 2)
  }
)

InceptionB <- torch::nn_module(
  "InceptionB",
  initialize = function(in_channels, conv_block = NULL) {
    if (is.null(conv_block)) {
      conv_block <- BasicConv2d
    }

    self$branch3x3 <- conv_block(in_channels, 384, kernel_size=3, stride=2)
    self$branch3x3dbl_1 <- conv_block(in_channels, 64, kernel_size=1)
    self$branch3x3dbl_2 <- conv_block(64, 96, kernel_size=3, padding=1)
    self$branch3x3dbl_3 <- conv_block(96, 96, kernel_size=3, stride=2)
  },
  .forward = function(x) {
    branch3x3 <- self$branch3x3(x)

    branch3x3dbl <- self$branch3x3dbl_1(x)
    branch3x3dbl <- self$branch3x3dbl_2(branch3x3dbl)
    branch3x3dbl <- self$branch3x3dbl_3(branch3x3dbl)

    branch_pool <- torch::nnf_max_pool2d(x, kernel_size=3, stride=2)

    list(branch3x3, branch3x3dbl, branch_pool)
  },
  forward = function(x) {
    outputs <- self$.forward(x)
    torch::torch_cat(outputs, 2)
  }
)

InceptionC <- torch::nn_module(
  "InceptionC",
  initialize = function(in_channels, channels_7x7, conv_block = NULL) {
    if (is.null(conv_block)) {
      conv_block <- BasicConv2d
    }

    self$branch1x1 <- conv_block(in_channels, 192, kernel_size=1)
    c7 <- channels_7x7
    self$branch7x7_1 <- conv_block(in_channels, c7, kernel_size=1)
    self$branch7x7_2 <- conv_block(c7, c7, kernel_size=c(1, 7), padding=c(0, 3))
    self$branch7x7_3 <- conv_block(c7, 192, kernel_size=c(7, 1), padding=c(3, 0))

    self$branch7x7dbl_1 <- conv_block(in_channels, c7, kernel_size=1)
    self$branch7x7dbl_2 <- conv_block(c7, c7, kernel_size=c(7, 1), padding=c(3, 0))
    self$branch7x7dbl_3 <- conv_block(c7, c7, kernel_size=c(1, 7), padding=c(0, 3))
    self$branch7x7dbl_4 <- conv_block(c7, c7, kernel_size=c(7, 1), padding=c(3, 0))
    self$branch7x7dbl_5 <- conv_block(c7, 192, kernel_size=c(1, 7), padding=c(0, 3))

    self$branch_pool = conv_block(in_channels, 192, kernel_size=1)
  },
  .forward = function(x) {
    branch1x1 <- self$branch1x1(x)

    branch7x7 <- self$branch7x7_1(x)
    branch7x7 <- self$branch7x7_2(branch7x7)
    branch7x7 <- self$branch7x7_3(branch7x7)

    branch7x7dbl <- self$branch7x7dbl_1(x)
    branch7x7dbl <- self$branch7x7dbl_2(branch7x7dbl)
    branch7x7dbl <- self$branch7x7dbl_3(branch7x7dbl)
    branch7x7dbl <- self$branch7x7dbl_4(branch7x7dbl)
    branch7x7dbl <- self$branch7x7dbl_5(branch7x7dbl)

    branch_pool <- torch::nnf_avg_pool2d(x, kernel_size=3, stride=1, padding=1)
    branch_pool <- self$branch_pool(branch_pool)

    list(branch1x1, branch7x7, branch7x7dbl, branch_pool)
  },
  forward = function(x) {
    outputs <- self$.forward(x)
    torch::torch_cat(outputs, 2)
  }
)

InceptionD <- torch::nn_module(
  "InceptionD",
  initialize = function(in_channels, conv_block = NULL) {

    if (is.null(conv_block)) {
      conv_block <- BasicConv2d
    }

    self$branch3x3_1 <- conv_block(in_channels, 192, kernel_size=1)
    self$branch3x3_2 <- conv_block(192, 320, kernel_size=3, stride=2)

    self$branch7x7x3_1 <- conv_block(in_channels, 192, kernel_size=1)
    self$branch7x7x3_2 <- conv_block(192, 192, kernel_size=c(1, 7), padding=c(0, 3))
    self$branch7x7x3_3 <- conv_block(192, 192, kernel_size=c(7, 1), padding=c(3, 0))
    self$branch7x7x3_4 <- conv_block(192, 192, kernel_size=3, stride=2)
  },
  .forward = function(x) {
    branch3x3 <- self$branch3x3_1(x)
    branch3x3 <- self$branch3x3_2(branch3x3)

    branch7x7x3 <- self$branch7x7x3_1(x)
    branch7x7x3 <- self$branch7x7x3_2(branch7x7x3)
    branch7x7x3 <- self$branch7x7x3_3(branch7x7x3)
    branch7x7x3 <- self$branch7x7x3_4(branch7x7x3)

    branch_pool <- torch::nnf_max_pool2d(x, kernel_size=3, stride=2)
    list(branch3x3, branch7x7x3, branch_pool)
  },
  forward = function(x) {
    outputs <- self$.forward(x)
    torch::torch_cat(outputs, 2)
  }
)

InceptionE <- torch::nn_module(
  "InceptionE",
  initialize = function(in_channels, conv_block = NULL) {
    if (is.null(conv_block)) {
      conv_block <- BasicConv2d
    }

    self$branch1x1 <- conv_block(in_channels, 320, kernel_size=1)

    self$branch3x3_1 <- conv_block(in_channels, 384, kernel_size=1)
    self$branch3x3_2a <- conv_block(384, 384, kernel_size=c(1, 3), padding=c(0, 1))
    self$branch3x3_2b <- conv_block(384, 384, kernel_size=c(3, 1), padding=c(1, 0))

    self$branch3x3dbl_1 <- conv_block(in_channels, 448, kernel_size=1)
    self$branch3x3dbl_2 <- conv_block(448, 384, kernel_size=3, padding=1)
    self$branch3x3dbl_3a <- conv_block(384, 384, kernel_size=c(1, 3), padding=c(0, 1))
    self$branch3x3dbl_3b <- conv_block(384, 384, kernel_size=c(3, 1), padding=c(1, 0))

    self$branch_pool <- conv_block(in_channels, 192, kernel_size=1)
  },
  .forward = function(x) {
    branch1x1 <- self$branch1x1(x)

    branch3x3 <- self$branch3x3_1(x)
    branch3x3 <- list(
      self$branch3x3_2a(branch3x3),
      self$branch3x3_2b(branch3x3)
    )
    branch3x3 <- torch::torch_cat(branch3x3, 2)

    branch3x3dbl <- self$branch3x3dbl_1(x)
    branch3x3dbl <- self$branch3x3dbl_2(branch3x3dbl)
    branch3x3dbl <- list(
      self$branch3x3dbl_3a(branch3x3dbl),
      self$branch3x3dbl_3b(branch3x3dbl)
    )
    branch3x3dbl <- torch::torch_cat(branch3x3dbl, 2)

    branch_pool <- torch::nnf_avg_pool2d(x, kernel_size=3, stride=1, padding=1)
    branch_pool <- self$branch_pool(branch_pool)

    list(branch1x1, branch3x3, branch3x3dbl, branch_pool)
  },
  forward = function(x) {
    outputs <- self$.forward(x)
    torch::torch_cat(outputs, 2)
  }
)

InceptionAux <- torch::nn_module(
  "InceptionAux",
  initialize = function(in_channels, num_classes, conv_block = NULL) {
    if (is.null(conv_block)) {
      conv_block <- BasicConv2d
    }

    self$conv0 <- conv_block(in_channels, 128, kernel_size=1)
    self$conv1 <- conv_block(128, 768, kernel_size=5)
    self$conv1$stddev = 0.01
    self$fc <- torch::nn_linear(768, num_classes)
    self$fc$stddev <- 0.001
  },
  forward = function(x) {
    # N x 768 x 17 x 17
    x <- torch::nnf_avg_pool2d(x, kernel_size=5, stride=3)
    # N x 768 x 5 x 5
    x <- self$conv0(x)
    # N x 128 x 5 x 5
    x <- self$conv1(x)
    # N x 768 x 1 x 1
    # Adaptive average pooling
    x <- torch::nnf_adaptive_avg_pool2d(x, c(1, 1))
    # N x 768 x 1 x 1
    x <- torch::torch_flatten(x, start_dim = 2)
    # N x 768
    x <- self$fc(x)
    # N x 1000
    x
  }
)

BasicConv2d <- torch::nn_module(
  "BasicConv2d",
  initialize = function(in_channels, out_channels, ...) {
    self$conv <- torch::nn_conv2d(in_channels, out_channels, bias=FALSE, ...)
    self$bn <- torch::nn_batch_norm2d(out_channels, eps=0.001)
  },
  forward = function(x) {
    x <- self$conv(x)
    x <- self$bn(x)
    torch::nnf_relu(x, inplace=TRUE)
  }
)

inception_model_urls <- list(
  inception_v3_google =c("https://torch-cdn.mlverse.org/models/vision/v2/models/inception_v3_google.pth", "8d60b4fcf263f2a8d2ed21f0f9690e3b", "~110 MB")
)

#' Inception v3 model
#'
#' Architecture from [Rethinking the Inception Architecture for Computer Vision](https://arxiv.org/abs/1512.00567)
#' The required minimum input size of the model is 75x75.
#' @note
#' **Important**: In contrast to the other models the inception_v3 expects tensors with a size of
#' N x 3 x 299 x 299, so ensure your images are sized accordingly.
#' @param pretrained (bool): If `TRUE`, returns a model pre-trained on ImageNet
#' @param progress (bool): If `TRUE`, displays a progress bar of the download to stderr
#' @param ... Used to pass keyword arguments to the Inception module:
#'  - aux_logits (bool): If `TRUE`, add an auxiliary branch that can improve training.
#'  Default: *TRUE*
#'  - transform_input (bool): If `TRUE`, preprocess the input according to the method with which it
#'  was trained on ImageNet. Default: *FALSE*
#'
#' @family classification_model
#' @export
model_inception_v3 <-function(pretrained = FALSE, progress = TRUE, ...) {
  args <- rlang::list2(...)

  if (pretrained) {

    if (is.null(args$transform_input)) {
      args$transform_input <- TRUE
    }
    if (!is.null(args$aux_logits)) {
      original_aux_logits <- args$aux_logits
      args$aux_logits <- TRUE
    } else {
      original_aux_logits <- TRUE
    }

    args$init_weights <- FALSE  # we are loading weights from a pretrained model

    model <- do.call(Inception3, args)
    r <- inception_model_urls[['inception_v3_google']]
    cli_inform("Model weights for {.cls {class(model)[1]}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1])
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict)
    return(model)
  }

  Inception3(...)
}

test_that("inception_v3 pretrained", {
  model <- model_inception_v3(pretrained = TRUE)
  model$eval()
  x <- model(torch_ones(2, 3, 299, 299))
  # the value has been copied from running the same model on pytorch.
  expect_equal(as.numeric(x[1,1]), 0.18005196750164032, tol = 5e-6)
  expect_tensor_shape(x, c(2, 1000))
})

#' Target Transform: COCO Polygon Segmentation to Masks
#'
#' Converts COCO-style polygon segmentation annotations from target `$segmentation` variable
#' into boolean mask tensors as target `$masks` variable in order to ease later-on visualisation
#' via `draw_segmentation_mask()`.
#' Use as `target_transform` in `coco_detection_dataset()`.
#'
#' @param y list being COCO dataset target variable, with names `segmentation`, `image_height`, `image_width`.
#'
#' @return Modified `y` list with added `masks` field (N, H, W) boolean tensor, N being the number of
#' classes.
#'
#' @examples
#' \dontrun{
#' ds <- coco_detection_dataset(
#'   root = "data",
#'   target_transform = target_transform_coco_masks
#' )
#' item <- ds[1]
#' draw_segmentation_masks(item)
#' }
#'
#' @family target_transforms
#' @export
target_transform_coco_masks <- function(y) {

  if (!"segmentation" %in% names(y)) {
    cli::cli_abort("Target must contain 'segmentation' field")
  }
  if (!all(c("image_height", "image_width") %in% names(y))) {
    cli::cli_abort("Target must contain 'image_height' and 'image_width' fields")
  }

  masks_list <- lapply(y$segmentation, function(seg) {
    if (is.list(seg) && length(seg) > 0) {
      mask <- coco_polygon_to_mask(seg, y$image_height, y$image_width)
      if (inherits(mask, "torch_tensor") && mask$ndim == 2) return(mask)
    }
    NULL
  })

  valid_masks <- Filter(function(m) !is.null(m), masks_list)

  if (length(valid_masks) > 0) {
    y$masks <- torch::torch_stack(valid_masks)
  } else {
    y$masks <- torch::torch_zeros(c(0, y$image_height, y$image_width), dtype = torch::torch_bool())
  }

  y
}


#' Target Transform: Trimap to Boolean Masks
#'
#' Converts Oxford-IIIT Pet dataset target `$trimap` variable (values 1,2,3) into
#' 3-channel boolean masks tensors as target `$masks` variable in order to ease later-on visualisation
#' via `draw_segmentation_mask()`.
#' Use as `target_transform` in `oxfordiiitpet_segmentation_dataset()`.
#'
#' @param y List containing `trimap` field (H, W) tensor with values 1, 2, 3
#'
#' @return Modified y list with added `masks` field (3, H, W) boolean tensor
#'
#' @details
#' Creates three mutually exclusive masks:
#' \itemize{
#'   \item Channel 1: Pet pixels (trimap == 1)
#'   \item Channel 2: Background pixels (trimap == 2)
#'   \item Channel 3: Outline pixels (trimap == 3)
#' }
#'
#' @examples
#' \dontrun{
#' ds <- oxfordiiitpet_segmentation_dataset(
#'   root = "data",
#'   target_transform = target_transform_trimap_masks
#' )
#' item <- ds[1]
#' draw_segmentation_masks(item)
#' }
#'
#' @family target_transforms
#' @export
target_transform_trimap_masks <- function(y) {
  if (!"trimap" %in% names(y)) {
    cli::cli_abort("Target must contain 'trimap' field")
  }

  trimap <- y$trimap

  if (!inherits(trimap, "torch_tensor")) {
    trimap <- torch::torch_tensor(trimap, dtype = torch::torch_int32())
  }

  if (trimap$ndim != 2) {
    cli::cli_abort("Trimap must be a 2D tensor")
  }

  mask1 <- (trimap == 1)
  mask2 <- (trimap == 2)
  mask3 <- (trimap == 3)

  y$masks <- torch::torch_stack(list(mask1, mask2, mask3))$to(dtype = torch::torch_bool())

  y
}

#' @include collection-rf100-doc.R
NULL

#' RoboFlow 100 Underwater dataset Collection
#'
#' Loads one of the underwater related RoboFlow 100 Environmental
#' datasets: "pipes", "aquarium", "objects", or "coral". Images are provided with
#' bounding box annotations for object-detection task.
#'
#' @inheritParams rf100_document_collection
#' @param dataset Dataset to select within \code{c("pipes", "aquarium", "objects", "coral")}.
#' @inherit rf100_document_collection return
#'
#' @examples
#' \dontrun{
#' ds <- rf100_underwater_collection(
#'   dataset = "aquarium",
#'   split = "train",
#'   transform = transform_to_tensor,
#'   download = TRUE
#' )
#'
#' item <- ds[24]
#' # map label ids into their class names
#' item$y$labels <- ds$classes[item$y$labels]
#' boxed <- draw_bounding_boxes(item)
#' tensor_image_browse(boxed)
#' }
#'
#' @family detection_dataset
#' @export
rf100_underwater_collection <- torch::dataset(
  name = "rf100_underwater_collection",
  inherit = rf100_document_collection,

  resources = data.frame(
    dataset = rep(c("pipes", "aquarium", "objects", "coral"),each = 3),
    split   = rep(c("train", "test", "valid"), times = 4),
    url = c(
      # pipes
      "https://huggingface.co/datasets/Francesco/underwater-pipes-4ng4t/resolve/main/data/train-00000-of-00001-229f86f63d040d1c.parquet",
      "https://huggingface.co/datasets/Francesco/underwater-pipes-4ng4t/resolve/main/data/test-00000-of-00001-6f4640f0fece4cfe.parquet",
      "https://huggingface.co/datasets/Francesco/underwater-pipes-4ng4t/resolve/main/data/validation-00000-of-00001-be63221743f11a18.parquet",
      # aquarium
      "https://huggingface.co/datasets/Francesco/aquarium-qlnqy/resolve/main/data/train-00000-of-00001-f7b9135b8ae646d4.parquet",
      "https://huggingface.co/datasets/Francesco/aquarium-qlnqy/resolve/main/data/test-00000-of-00001-a436c40b7f9121c8.parquet",
      "https://huggingface.co/datasets/Francesco/aquarium-qlnqy/resolve/main/data/validation-00000-of-00001-8a7f5f5ee0144783.parquet",
      # objects
      "https://huggingface.co/datasets/Francesco/underwater-objects-5v7p8/resolve/main/data/train-00000-of-00001-2c9335408bd5503d.parquet",
      "https://huggingface.co/datasets/Francesco/underwater-objects-5v7p8/resolve/main/data/test-00000-of-00001-7670918d21b7b1b6.parquet",
      "https://huggingface.co/datasets/Francesco/underwater-objects-5v7p8/resolve/main/data/validation-00000-of-00001-51a32a9b21a2097c.parquet",
      # coral
      "https://huggingface.co/datasets/Francesco/coral-lwptl/resolve/main/data/train-00000-of-00001-05c58869565f0bfb.parquet",
      "https://huggingface.co/datasets/Francesco/coral-lwptl/resolve/main/data/test-00000-of-00001-6d38ab2deab119d4.parquet",
      "https://huggingface.co/datasets/Francesco/coral-lwptl/resolve/main/data/validation-00000-of-00001-52a8f5261f53bb7c.parquet"
    ),
    md5 = c(
      # pipes
      "521c3781e2e078d6ff2a1d47800bade1",      "8ed57e5b2c81e4cf881f0d6ac1e8ce78",      "3ed01d2935747e8f4f36bb6f9c77da68",
      # aquarium
      "f271749dff54a654c6c337f255744b50",      "961b0588671b682bad4a55b18bf0264f",      "bddf0ef97f146fb7f2411e833caa713c",
      # objects
      "2c97febd495f514d96b8acd2ac293b90",      "95e3713caf397aa4182c2ed72651032a",      "c7c50c24fb1c11abc8da96eda21d8055",
      # coral
      "97bbf0dfdf902952619940493d29f9de",      "09ce58473adf04f1c38071ea86fbfab7",      "3097061e42de960c8746bcf188c5eeb0"
    ),
    size = c(223, 60, 30, 26, 7.5, 3.5 ,290,85,42, 33,7,5) *1e6,
    stringsAsFactors = FALSE
  )
)

#' EfficientNetV2 Models
#'
#' Constructs EfficientNetV2 model architectures as described in
#' \emph{EfficientNetV2: Smaller Models and Faster Training}.
#'
#' @section Task:
#' Image classification with 1000 output classes by default (ImageNet).
#'
#' @section Input Format:
#' The models expect input tensors of shape \code{(batch_size, 3, H, W)}.
#' Typical values for \code{H} and \code{W} are 384 for V2-S, 480 for V2-M,
#' and 512 for V2-L.
#'
#' @section Variants:
#' \tabular{lllll}{
#'   \strong{Model} \tab \strong{Resolution} \tab \strong{Params (M)} \tab
#'   \strong{GFLOPs} \tab \strong{Top-1 Acc.} \cr
#'   V2-S \tab 384 \tab 24 \tab 8.4 \tab 83.9 \cr
#'   V2-M \tab 480 \tab 55 \tab 24 \tab 85.1 \cr
#'   V2-L \tab 512 \tab 119 \tab 55 \tab 85.7
#' }
#'
#' @inheritParams model_resnet18
#' @param ... Other parameters passed to the model implementation, such as
#'   \code{num_classes} to change the output dimension.
#'
#' @family classification_model
#' @name model_efficientnet_v2
#' @seealso \code{\link{model_efficientnet}}
#' @examples
#' \dontrun{
#' model <- model_efficientnet_v2_s()
#' input <- torch::torch_randn(1, 3, 224, 224)
#' output <- model(input)
#'
#' # Show Top-5 predictions
#' topk <- output$topk(k = 5, dim = 2)
#' indices <- as.integer(topk[[2]][1, ])
#' scores <- as.numeric(topk[[1]][1, ])
#' glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}%)")
#' }
NULL

fused_mbconv_block <- torch::nn_module(
  initialize = function(in_channels, out_channels, kernel_size, stride,
                        expand_ratio, norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    hidden_dim <- as.integer(in_channels * expand_ratio)
    self$use_res_connect <- stride == 1 && in_channels == out_channels
    layers <- list()

    if (expand_ratio != 1) {
      # Fused: expansion + spatial convolution in one step
      layers[[length(layers) + 1]] <- conv_norm_act(
        in_channels, hidden_dim, kernel_size = kernel_size, stride = stride,
        norm_layer = norm_layer, activation_layer = torch::nn_silu
      )
      # Project back to output channels
      layers[[length(layers) + 1]] <- torch::nn_sequential(
        torch::nn_conv2d(hidden_dim, out_channels, kernel_size = 1, stride = 1, bias = FALSE),
        norm_layer(out_channels)
      )
    } else {
      # When expand_ratio == 1, just do a single convolution
      layers[[length(layers) + 1]] <- torch::nn_sequential(
        torch::nn_conv2d(in_channels, out_channels, kernel_size = kernel_size,
                         stride = stride, padding = (kernel_size - 1) %/% 2, bias = FALSE),
        norm_layer(out_channels),
        torch::nn_silu(inplace = TRUE)
      )
    }

    self$block <- torch::nn_sequential(!!!layers)
  },
  forward = function(x) {
    out <- self$block(x)
    if (self$use_res_connect)
      out <- out + x
    out
  }
)

efficientnet_v2 <- torch::nn_module(
  "efficientnet_v2",
  initialize = function(cfgs, dropout = 0.2, num_classes = 1000,
                        norm_layer = NULL, firstconv_out = 24) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d

    features <- list(
      conv_norm_act(3, firstconv_out, stride = 2,
                    norm_layer = norm_layer, activation_layer = torch::nn_silu)
    )
    in_channels <- firstconv_out

    for (cfg in cfgs) {
      oc <- cfg$channels
      r <- cfg$repeats
      block_fn <- if (identical(cfg$block, "fused")) fused_mbconv_block else mbconv_block
      stage_blocks <- list()
      for (i in seq_len(r)) {
        s <- ifelse(i == 1, cfg$stride, 1)
        if (identical(cfg$block, "fused")) {
          stage_blocks[[i]] <- block_fn(
            in_channels, oc, kernel_size = cfg$kernel, stride = s,
            expand_ratio = cfg$expand, norm_layer = norm_layer
          )
        } else {
          # For mbconv blocks, use the v2 block with corrected SE calculation
          stage_blocks[[i]] <- mbconv_block_v2(
            in_channels, oc, kernel_size = cfg$kernel, stride = s,
            expand_ratio = cfg$expand, se_ratio = 0.25, norm_layer = norm_layer
          )
        }
        in_channels <- oc
      }
      features[[length(features) + 1]] <- torch::nn_sequential(!!!stage_blocks)
    }

    final_channels <- 1280
    features[[length(features) + 1]] <- conv_norm_act(
      in_channels, final_channels, kernel_size = 1,
      norm_layer = norm_layer, activation_layer = torch::nn_silu
    )
    self$features <- torch::nn_sequential(!!!features)
    self$avgpool <- torch::nn_adaptive_avg_pool2d(c(1, 1))
    self$classifier <- torch::nn_sequential(
      torch::nn_dropout(dropout),
      torch::nn_linear(final_channels, num_classes)
    )
  },
  forward = function(x) {
    x <- self$features(x)
    x <- self$avgpool(x)
    x <- torch::torch_flatten(x, start_dim = 2)
    self$classifier(x)
  }
)

# mbconv_block for EfficientNetV2
mbconv_block_v2 <- torch::nn_module(
  initialize = function(in_channels, out_channels, kernel_size, stride,
                        expand_ratio, se_ratio = 0.25, norm_layer = NULL) {
    if (is.null(norm_layer))
      norm_layer <- torch::nn_batch_norm2d
    hidden_dim <- in_channels * expand_ratio
    self$use_res_connect <- stride == 1 && in_channels == out_channels
    layers <- list()

    # Expand
    if (expand_ratio != 1) {
      layers[[length(layers) + 1]] <- conv_norm_act(
        in_channels, hidden_dim, kernel_size = 1,
        norm_layer = norm_layer, activation_layer = torch::nn_silu
      )
    }

    # Depthwise
    layers[[length(layers) + 1]] <- conv_norm_act(
      hidden_dim, hidden_dim, kernel_size = kernel_size, stride = stride,
      groups = hidden_dim, norm_layer = norm_layer,
      activation_layer = torch::nn_silu
    )

    # SE block
    if (!is.null(se_ratio) && se_ratio > 0) {
      squeeze_channels <- max(1, as.integer(in_channels * se_ratio))
      layers[[length(layers) + 1]] <- se_block(
        hidden_dim, squeeze_channels = squeeze_channels
      )
    }

    # Projection
    layers[[length(layers) + 1]] <- torch::nn_sequential(
      torch::nn_conv2d(hidden_dim, out_channels, 1, bias = FALSE),
      norm_layer(out_channels)
    )

    self$block <- torch::nn_sequential(!!!layers)
  },
  forward = function(x) {
    out <- self$block(x)
    if (self$use_res_connect)
      out <- out + x
    out
  }
)

effnetv2 <- function(arch, cfgs, dropout, firstconv_out, pretrained, progress, ...) {
  args <- rlang::list2(...)

  model <- do.call(efficientnet_v2, append(args, list(
    cfgs = cfgs,
    dropout = dropout,
    firstconv_out = firstconv_out
  )))

  if (pretrained) {
    r <- efficientnet_v2_model_urls[[arch]]
    cli_inform("Model weights for {.cls {arch}} ({.emph {r[3]}}) will be downloaded and processed if not already available.")
    state_dict_path <- download_and_cache(r[1], prefix = "efficientnet")
    if (!tools::md5sum(state_dict_path) == r[2])
      runtime_error("Corrupt file! Delete the file in {state_dict_path} and try again.")

    state_dict <- torch::load_state_dict(state_dict_path)
    model$load_state_dict(state_dict)
  }

  model
}

#' @describeIn model_efficientnet_v2 EfficientNetV2-S model
#' @export
model_efficientnet_v2_s <- function(pretrained = FALSE, progress = TRUE, ...) {
  cfgs <- list(
    list(block = "fused", expand = 1, channels = 24, repeats = 2, stride = 1, kernel = 3),
    list(block = "fused", expand = 4, channels = 48, repeats = 4, stride = 2, kernel = 3),
    list(block = "fused", expand = 4, channels = 64, repeats = 4, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 4, channels = 128, repeats = 6, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 160, repeats = 9, stride = 1, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 256, repeats = 15, stride = 2, kernel = 3)
  )
  effnetv2("efficientnet_v2_s", cfgs, 0.2, 24, pretrained, progress, ...)
}

#' @describeIn model_efficientnet_v2 EfficientNetV2-M model
#' @export
model_efficientnet_v2_m <- function(pretrained = FALSE, progress = TRUE, ...) {
  cfgs <- list(
    list(block = "fused", expand = 1, channels = 24, repeats = 3, stride = 1, kernel = 3),
    list(block = "fused", expand = 4, channels = 48, repeats = 5, stride = 2, kernel = 3),
    list(block = "fused", expand = 4, channels = 80, repeats = 5, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 4, channels = 160, repeats = 7, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 176, repeats = 14, stride = 1, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 304, repeats = 18, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 512, repeats = 5, stride = 1, kernel = 3)
  )
  effnetv2("efficientnet_v2_m", cfgs, 0.3, 24, pretrained, progress, ...)
}

#' @describeIn model_efficientnet_v2 EfficientNetV2-L model
#' @export
model_efficientnet_v2_l <- function(pretrained = FALSE, progress = TRUE, ...) {
  cfgs <- list(
    list(block = "fused", expand = 1, channels = 32, repeats = 4, stride = 1, kernel = 3),
    list(block = "fused", expand = 4, channels = 64, repeats = 7, stride = 2, kernel = 3),
    list(block = "fused", expand = 4, channels = 96, repeats = 7, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 4, channels = 192, repeats = 10, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 224, repeats = 19, stride = 1, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 384, repeats = 25, stride = 2, kernel = 3),
    list(block = "mbconv", expand = 6, channels = 640, repeats = 7, stride = 1, kernel = 3)
  )
  effnetv2("efficientnet_v2_l", cfgs, 0.4, 32, pretrained, progress, ...)
}

efficientnet_v2_model_urls <- list(
  efficientnet_v2_s = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_v2_s.pth", "9f0f813046f802fb7151fbf0e169ae13", "~87 MB"),
  efficientnet_v2_m = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_v2_m.pth", "0a08118483687a61c6d8e38f9e24dee6", "~220 MB"),
  efficientnet_v2_l = c("https://torch-cdn.mlverse.org/models/vision/v2/models/efficientnet_v2_l.pth", "c7bca278771eb3ca70a00ff6a29e5208", "~476 MB")
)

test_that("efficientnet v2 small model produce correct output shapes", {

  # without pretrained
  model <- model_efficientnet_v2_s(pretrained = FALSE)
  input <- torch::torch_randn(1, 3, 384, 384)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  # with pretrained
  withr::with_options(list(timeout = 360), {
    model <- model_efficientnet_v2_s(pretrained = TRUE)
  })
  input <- torch::torch_randn(1, 3, 384, 384)
  out <- model(input)
  expect_tensor_shape(out, c(1, 1000))

  unlink_model_file()
})


test_that("efficientnet v2 models produce correct output shapes", {
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = "0") != "1",
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  variants <- list(
    m = model_efficientnet_v2_m,
    l = model_efficientnet_v2_l
  )

  sizes <- c(m = 480, l = 512)

  for (name in names(variants)) {
    fn <- variants[[name]]
    size <- sizes[[name]]

    # without pretrained
    model <- fn(pretrained = FALSE)
    input <- torch::torch_randn(1, 3, size, size)
    out <- model(input)
    expect_tensor_shape(out, c(1, 1000))

    # with pretrained
    withr::with_options(list(timeout = 360), {
      model <- fn(pretrained = TRUE)
    })
    input <- torch::torch_randn(1, 3, size, size)
    out <- model(input)
    expect_tensor_shape(out, c(1, 1000))
  }
})

test_that("efficientnet v2 allows custom num_classes", {
  model <- model_efficientnet_v2_s(num_classes = 17)
  input <- torch::torch_randn(1, 3, 384, 384)
  out <- model(input)
  expect_tensor_shape(out, c(1, 17))
})

#' COCO Detection Dataset
#'
#' Loads the MS COCO dataset for object detection tasks only.
#'
#' @rdname coco_detection_dataset
#' @param root Root directory where the dataset is stored or will be downloaded to.
#' @param train Logical. If TRUE, loads the training split; otherwise, loads the validation split.
#' @param year Character. Dataset version year. One of \code{"2014"} or \code{"2017"}.
#' @param download Logical. If TRUE, downloads the dataset if it's not already present in the \code{root} directory.
#' @param transform Optional transform function applied to the image.
#' @param target_transform Optional transform function applied to the target (labels, boxes, etc.).
#'
#' @return An object of class `coco_detection_dataset`. Each item is a list:
#' - `x`: a `(C, H, W)` array representing the image.
#' - `y$boxes`: a `(N, 4)` `torch_tensor` of bounding boxes in the format  \eqn{(x_{min}, y_{min}, x_{max}, y_{max})}.
#' - `y$labels`: an integer `torch_tensor` with the class label for each object.
#' - `y$area`: a float `torch_tensor` indicating the area of each object.
#' - `y$iscrowd`: a boolean `torch_tensor`, where `TRUE` marks the object as part of a crowd.
#'
#' The returned object has S3 class \code{"image_with_bounding_box"}
#' to enable automatic dispatch by visualization functions such as \code{draw_bounding_boxes()}.
#'
#' For instance segmentation tasks, use \code{\link{coco_segmentation_dataset}} instead.
#'
#' @details
#' The returned image `x` is in CHW format (channels, height, width), matching the torch convention.
#' The dataset `y` offers object detection annotations such as bounding boxes, labels,
#' areas, and crowd indicators from the official COCO annotations.
#'
#' Files are downloaded to a \code{coco} subdirectory in the torch cache directory for better organization.
#'
#' @examples
#' \dontrun{
#' # Load dataset for object detection
#' ds <- coco_detection_dataset(
#'   train = FALSE,
#'   year = "2017",
#'   download = TRUE
#' )
#'
#' item <- ds[1]
#'
#' # Visualize bounding boxes
#' boxed <- draw_bounding_boxes(item)
#' tensor_image_browse(boxed)
#' }
#' @family detection_dataset
#' @seealso \code{\link{coco_segmentation_dataset}} for instance segmentation tasks
#' @importFrom jsonlite fromJSON
#' @export
coco_detection_dataset <- torch::dataset(
  name = "coco_detection_dataset",
  resources = data.frame(
    year = rep(c(2017, 2014), each = 4 ),
    content = rep(c("image", "annotation"), time = 2, each = 2),
    split = rep(c("train", "val"), time = 4),
    url = c("http://images.cocodataset.org/zips/train2017.zip", "http://images.cocodataset.org/zips/val2017.zip",
            rep("http://images.cocodataset.org/annotations/annotations_trainval2017.zip", time = 2),
            "http://images.cocodataset.org/zips/train2014.zip", "http://images.cocodataset.org/zips/val2014.zip",
            rep("http://images.cocodataset.org/annotations/annotations_trainval2014.zip", time = 2)),
    size = c("18.4 GB", "800 MB", rep("770 MB", time = 2), "6.33 GB", "6.33 GB", rep("242 MB", time = 2)),
    md5 = c(c("cced6f7f71b7629ddf16f17bbcfab6b2", "442b8da7639aecaf257c1dceb8ba8c80"),
            rep("f4bbac642086de4f52a3fdda2de5fa2c", time = 2),
            c("0da8cfa0e090c266b78f30e2d2874f1a", "a3d79f5ed8d289b7a7554ce06a5782b3"),
            rep("0a379cfc70b0e71301e0f377548639bd", time = 2)),
    stringsAsFactors = FALSE
  ),

  initialize = function(
    root = tempdir(),
    train = TRUE,
    year = c("2017", "2014"),
    download = FALSE,
    transform = NULL,
    target_transform = NULL
  ) {

    year <- match.arg(year)
    split <- ifelse(train, "train", "val")

    root <- fs::path_expand(root)
    self$root <- root
    self$year <- year
    self$split <- split
    self$transform <- transform
    self$target_transform <- target_transform
    self$archive_size <- self$resources[self$resources$year == year & self$resources$split == split & self$resources$content == "image", ]$size

    self$data_dir <- fs::path(root, glue::glue("coco{year}"))

    image_year <- ifelse(year == "2016", "2014", year)
    self$image_dir <- fs::path(self$data_dir, glue::glue("{split}{image_year}"))
    self$annotation_file <- fs::path(self$data_dir, "annotations",
                                     glue::glue("instances_{split}{year}.json"))

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")
    }

    self$load_annotations()

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  check_exists = function() {
    fs::file_exists(self$annotation_file) && fs::dir_exists(self$image_dir)
  },

  .getitem = function(index) {
    image_id <- self$image_ids[index]
    image_info <- self$image_metadata[[as.character(image_id)]]

    img_path <- fs::path(self$image_dir, image_info$file_name)

    x <- base_loader(img_path)

    height <- dim(x)[1]
    width <- dim(x)[2]

    anns <- self$annotations[self$annotations$image_id == image_id, ]

    if (nrow(anns) > 0) {
      boxes_wh <- torch::torch_tensor(do.call(rbind, anns$bbox), dtype = torch::torch_float())
      boxes <- box_xywh_to_xyxy(boxes_wh)

      label_ids <- anns$category_id
      labels <- self$category_names[as.character(label_ids)]

      area <- torch::torch_tensor(anns$area, dtype = torch::torch_float())
      iscrowd <- torch::torch_tensor(as.logical(anns$iscrowd), dtype = torch::torch_bool())

    } else {
      # empty annotation
      boxes <- torch::torch_zeros(c(0, 4), dtype = torch::torch_float())
      labels <- character()
      area <- torch::torch_empty(0, dtype = torch::torch_float())
      iscrowd <- torch::torch_empty(0, dtype = torch::torch_bool())
      anns$segmentation <- list()
    }

    y <- list(
      boxes = boxes,
      labels = labels,
      area = area,
      iscrowd = iscrowd
    )

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y$image_height <- height
      y$image_width <- width
      y <- self$target_transform(y)
    }

    result <- list(x = x, y = y)
    class(result) <- c("image_with_bounding_box", class(result))

    result
  },

  .length = function() {
    length(self$image_ids)
  },

  download = function() {
    annotation_filter <- self$resources$year == self$year & self$resources$split == self$split & self$resources$content == "annotation"
    image_filter <- self$resources$year == self$year & self$resources$split == self$split & self$resources$content == "image"

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    ann_zip <- download_and_cache(self$resources[annotation_filter, ]$url, prefix = "coco")
    archive <- download_and_cache(self$resources[image_filter, ]$url, prefix = "coco")

    if (tools::md5sum(archive) != self$resources[image_filter, ]$md5) {
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")
    }

    utils::unzip(ann_zip, exdir = self$data_dir)
    utils::unzip(archive, exdir = self$data_dir)

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  load_annotations = function() {
    data <- jsonlite::fromJSON(self$annotation_file)

    self$image_metadata <- setNames(
      split(data$images, seq_len(nrow(data$images))),
      as.character(data$images$id)
    )

    self$annotations <- data$annotations
    self$classes <- data$categories$name
    self$category_ids <- data$categories$id
    self$category_names <- setNames(self$classes, self$category_ids)

    ids <- as.numeric(names(self$image_metadata))
    image_paths <- fs::path(self$image_dir,
                            sapply(ids, function(id) self$image_metadata[[as.character(id)]]$file_name))
    exist <- fs::file_exists(image_paths)
    self$image_ids <- ids[exist]
  },

  active = list(
    categories = function() {
      .Deprecated(
        msg = "The `$categories` attribute is deprecated. Please use `$classes` instead for class names."
      )
      data.frame(id = self$category_ids, name = self$classes)
    }
  )
)


#' COCO Segmentation Dataset
#'
#' Loads the MS COCO dataset for instance segmentation tasks.
#'
#' @rdname coco_segmentation_dataset
#' @inheritParams coco_detection_dataset
#' @param target_transform Optional transform function applied to the target.
#'   Use \code{target_transform_coco_masks} to convert polygon annotations to binary masks.
#'
#' @return An object of class `coco_segmentation_dataset`. Each item is a list:
#' - `x`: a `(C, H, W)` array representing the image.
#' - `y$labels`: an integer `torch_tensor` with the class label for each object.
#' - `y$iscrowd`: a boolean `torch_tensor`, where `TRUE` marks the object as part of a crowd.
#' - `y$segmentation`: a list of segmentation polygons for each object.
#' - `y$masks`: a `(N, H, W)` boolean `torch_tensor` containing binary segmentation masks (when using target_transform_coco_masks).
#'
#' The returned object has S3 class \code{"image_with_segmentation_mask"}
#' to enable automatic dispatch by visualization functions such as \code{draw_segmentation_masks()}.
#'
#' For object detection tasks without segmentation, use \code{\link{coco_detection_dataset}} instead.
#'
#' @details
#' The returned image `x` is in CHW format (channels, height, width), matching the torch convention.
#' The dataset `y` offers instance segmentation annotations including labels,
#' crowd indicators, and segmentation masks from the official COCO annotations.
#'
#' Files are downloaded to a \code{coco} subdirectory in the torch cache directory for better organization.
#'
#' @examples
#' \dontrun{
#' # Load dataset for instance segmentation
#' ds <- coco_segmentation_dataset(
#'   train = FALSE,
#'   year = "2017",
#'   download = TRUE,
#'   target_transform = target_transform_coco_masks
#' )
#'
#' item <- ds[1]
#'
#' # Visualize segmentation masks
#' masked <- draw_segmentation_masks(item)
#' tensor_image_browse(masked)
#' }
#' @family segmentation_dataset
#' @seealso \code{\link{coco_detection_dataset}} for object detection tasks
#' @export
coco_segmentation_dataset <- torch::dataset(
  name = "coco_segmentation_dataset",
  inherit = coco_detection_dataset,

  .getitem = function(index) {
    image_id <- self$image_ids[index]
    image_info <- self$image_metadata[[as.character(image_id)]]

    img_path <- fs::path(self$image_dir, image_info$file_name)

    x <- base_loader(img_path)

    height <- dim(x)[1]
    width <- dim(x)[2]

    anns <- self$annotations[self$annotations$image_id == image_id, ]

    if (nrow(anns) > 0) {
      label_ids <- anns$category_id
      labels <- self$category_names[as.character(label_ids)]
      iscrowd <- torch::torch_tensor(as.logical(anns$iscrowd), dtype = torch::torch_bool())
    } else {
      # empty annotation
      labels <- character()
      iscrowd <- torch::torch_empty(0, dtype = torch::torch_bool())
      anns$segmentation <- list()
    }

    y <- list(
      labels = labels,
      iscrowd = iscrowd,
      segmentation = anns$segmentation
    )

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y$image_height <- height
      y$image_width <- width
      y <- self$target_transform(y)
    }

    result <- list(x = x, y = y)
    class(result) <- c("image_with_bounding_box", class(result))

    if (!is.null(y$masks)) {
      class(result) <- c("image_with_segmentation_mask", class(result))
    }
    result
  }
)


#' COCO Caption Dataset
#'
#' Loads the MS COCO dataset for image captioning.
#'
#' @rdname coco_caption_dataset
#' @inheritParams coco_detection_dataset
#' @param year Character. Dataset version year. One of \code{"2014"}.
#'
#' @return An object of class `coco_caption_dataset`. Each item is a list:
#' - `x`: an `(H, W, C)` numeric array containing the RGB image.
#' - `y`: a character string with the image caption.
#'
#' @examples
#' \dontrun{
#' ds <- coco_caption_dataset(
#'   train = FALSE,
#'   download = TRUE
#' )
#' example <- ds[1]
#'
#' # Access image and caption
#' x <- example$x
#' y <- example$y
#'
#' # Prepare image for plotting
#' image_array <- as.numeric(x)
#' dim(image_array) <- dim(x)
#'
#' plot(as.raster(image_array))
#' title(main = y, col.main = "black")
#' }
#' @family caption_dataset
#' @export
coco_caption_dataset <- torch::dataset(
  name = "coco_caption_dataset",
  inherit = coco_detection_dataset,

  initialize = function(
    root = tempdir(),
    train = TRUE,
    year = c("2014"),
    download = FALSE,
    transform = NULL,
    target_transform = NULL
  ) {

    year <- match.arg(year)
    split <- ifelse(train, "train", "val")

    root <- fs::path_expand(root)
    self$root <- root
    self$split <- split
    self$year <- year
    self$transform <- transform
    self$target_transform <- target_transform
    self$data_dir <- fs::path(root, glue::glue("coco{year}"))
    self$image_dir <- fs::path(self$data_dir, glue::glue("{split}{year}"))
    self$annotation_file <- fs::path(self$data_dir, "annotations", glue::glue("captions_{split}{year}.json"))
    self$archive_size <- self$resources[self$resources$year == year & self$resources$split == split & self$resources$content == "image", ]$size

    if (download){
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")
    }

    self$load_annotations()

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images.")
  },

  load_annotations = function() {
    data <- jsonlite::fromJSON(self$annotation_file)
    self$annotations <- data$annotations
    self$image_ids <- unique(self$annotations$image_id)
  },

  .getitem = function(index) {
    if (index < 1 || index > length(self))
      rlang::abort("Index out of bounds")

    ann <- self$annotations[index, ]
    image_id <- ann$image_id
    y <- ann$caption

    prefix <- ifelse(self$split == "train", "COCO_train2014_", "COCO_val2014_")
    filename <- paste0(prefix, sprintf("%012d", image_id), ".jpg")
    image_path <- fs::path(self$image_dir, filename)

    x <- jpeg::readJPEG(image_path)

    if (!is.null(self$transform))
      x <- self$transform(x)

    if (!is.null(self$target_transform))
      y <- self$target_transform(y)

    list(x = x, y = y)
  }
)

#' MS COCO Class Labels
#'
#' Utilities for resolving COCO 90 class identifiers to their corresponding
#' human readable labels. The labels are retrieved from pytorch/vision source to be compliant
#' with torchvision pretrained models.
#'
#' @param class_id Integer vector of 1-based class identifiers.
#' @return A character vector with the COCO class names
#' @family class_resolution
#' @importFrom utils read.delim
#' @export
coco_classes <- function(class_id = 1:90) {
  if (any(class_id > 90)) {
    cli_warn("MS COCO {.var class_id} cannot be > 90")
  }

  labels <- c("person",
    "bicycle",
    "car",
    "motorcycle",
    "airplane",
    "bus",
    "train",
    "truck",
    "boat",
    "traffic light",
    "fire hydrant",
    "N/A",
    "stop sign",
    "parking meter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "N/A",
    "backpack",
    "umbrella",
    "N/A",
    "N/A",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sports ball",
    "kite",
    "baseball bat",
    "baseball glove",
    "skateboard",
    "surfboard",
    "tennis racket",
    "bottle",
    "N/A",
    "wine glass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hot dog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "couch",
    "potted plant",
    "bed",
    "N/A",
    "dining table",
    "N/A",
    "N/A",
    "toilet",
    "N/A",
    "tv",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cell phone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "N/A",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddy bear",
    "hair drier",
    "toothbrush"
  )
  labels[nzchar(labels)][class_id]
}

#' MNIST and Derived Datasets
#'
#' Prepares various MNIST-style image classification datasets and optionally downloads them.
#' Images are thumbnails images of 28 x 28 pixels of grayscale values encoded as integer.
#'
#' - **MNIST**: Original handwritten digit dataset.
#' - **Fashion-MNIST**: Clothing item images for classification.
#' - **Kuzushiji-MNIST**: Japanese cursive character dataset.
#' - **QMNIST**: Extended MNIST with high-precision NIST data.
#' - **EMNIST**: A collection of letters and digits with multiple datasets and  splits.
#'
#' @param root Root directory for dataset storage. The dataset will be stored under `root/<dataset-name>`. Defaults to `tempdir()`.
#' @param train Logical. If TRUE, use the training set; otherwise, use the test set. Not applicable to all datasets.
#' @param split Character. Used in `emnist_dataset()` and `qmnist_dataset()` to specify the subset. See individual descriptions for valid values.
#' @param download Logical. If TRUE, downloads the dataset to `root/`. If the dataset is already present, download is skipped.
#' @param transform Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).
#' @param target_transform Optional. A function that transforms the label.
#'
#' @return A torch dataset object, where each items is a list of `x` (image) and `y` (label).
#'
#' @section Supported `dataset`s for `emnist_collection()`:
#' - `"byclass"`: 62 classes (digits + uppercase + lowercase)
#' - `"bymerge"`: 47 classes (merged uppercase and lowercase)
#' - `"balanced"`: 47 classes, balanced digits and letters
#' - `"letters"`: 26 uppercase letters
#' - `"digits"`: 10 digit classes
#' - `"mnist"`: Standard MNIST digit classes
#'
#' @section Supported `split`s for `qmnist_dataset()`:
#' - `"train"`: 60,000 training samples (MNIST-compatible)
#' - `"test"`: Extended test set
#' - `"nist"`: Full NIST digit set
#'
#' @examples
#' \dontrun{
#' ds <- mnist_dataset(download = TRUE)
#' item <- ds[1]
#' item$x  # image
#' item$y  # label
#'
#' qmnist <- qmnist_dataset(split = "train", download = TRUE)
#' item <- qmnist[1]
#' item$x
#' item$y
#'
#' emnist <- emnist_collection(dataset = "balanced", split = "test", download = TRUE)
#' item <- emnist[1]
#' item$x
#' item$y
#'
#' kmnist <- kmnist_dataset(download = TRUE, train = FALSE)
#' fmnist <- fashion_mnist_dataset(download = TRUE, train = TRUE)
#' }
#'
#' @family classification_dataset
#' @name mnist_dataset
#' @rdname mnist_dataset
#' @export
mnist_dataset <- dataset(
  name = "mnist",
  archive_size = "12 MB",
  resources = list(
    c("https://torch-cdn.mlverse.org/datasets/mnist/train-images-idx3-ubyte.gz", "f68b3c2dcbeaaa9fbdd348bbdeb94873"),
    c("https://torch-cdn.mlverse.org/datasets/mnist/train-labels-idx1-ubyte.gz", "d53e105ee54ea40749a09fcbcd1e9432"),
    c("https://torch-cdn.mlverse.org/datasets/mnist/t10k-images-idx3-ubyte.gz", "9fb629c4189551a2d022fa330f9573f3"),
    c("https://torch-cdn.mlverse.org/datasets/mnist/t10k-labels-idx1-ubyte.gz", "ec29112dd5afa0611ce80d1b7f02629c")
  ),
  training_file = 'training.rds',
  test_file = 'test.rds',
  classes = c('0 - zero', '1 - one', '2 - two', '3 - three', '4 - four',
             '5 - five', '6 - six', '7 - seven', '8 - eight', '9 - nine'),
  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root_path <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$train <- train
    self$split <- ifelse(train, "train", "test")

    if (download){
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")

    if (self$train)
      data_file <- self$training_file
    else
      data_file <- self$test_file

    data <- readRDS(file.path(self$processed_folder, data_file))
    self$data <- data[[1]]
    self$targets <- data[[2]] + 1L

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {

    if (self$check_exists())
      return(NULL)

    fs::dir_create(self$raw_folder)
    fs::dir_create(self$processed_folder)

    for (r in self$resources) {
      filename <- tail(strsplit(r[1], "/")[[1]], 1)
      destpath <- file.path(self$raw_folder, filename)

      archive <- download_and_cache(r[1], prefix = class(self)[1])
      fs::file_copy(archive, destpath)

      if (tools::md5sum(destpath) != r[2])
        runtime_error("Corrupt file! Delete the file in {archive} and try again.")

    }

    cli_inform("Downloading {.cls {class(self)[[1]]}} ...")
    cli_inform("Processing {.cls {class(self)[[1]]}}...")

    training_set <- list(
      read_sn3_pascalvincent(file.path(self$raw_folder, 'train-images-idx3-ubyte.gz')),
      read_sn3_pascalvincent(file.path(self$raw_folder, 'train-labels-idx1-ubyte.gz'))
    )

    test_set <- list(
      read_sn3_pascalvincent(file.path(self$raw_folder, 't10k-images-idx3-ubyte.gz')),
      read_sn3_pascalvincent(file.path(self$raw_folder, 't10k-labels-idx1-ubyte.gz'))
    )

    saveRDS(training_set, file.path(self$processed_folder, self$training_file))
    saveRDS(test_set, file.path(self$processed_folder, self$test_file))

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")

  },

  check_exists = function() {
    fs::file_exists(file.path(self$processed_folder, self$training_file)) &&
      fs::file_exists(file.path(self$processed_folder, self$test_file))
  },

  .getitem = function(index) {
    x <- self$data[index, , , drop = FALSE]
    y <- self$targets[index]

    if (!is.null(self$transform)) {
        x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
        y <- self$target_transform(y)
    }

    list(x = x, y = y)
  },


  # .getbatch = function(index) {
  #   x <- self$data[index, , , drop = FALSE]
  #   # unsqueeze(2) on array
  #   dim(x) = c(length(index), 1, dim(x)[2:3])
  #   y <- self$targets[index]
  #
  #   if (!is.null(self$transform)) {
  #     x <- self$transform(x)
  #   }
  #
  #   if (!is.null(self$target_transform)) {
  #     y <- self$target_transform(y)
  #   }
  #
  #   list(x = x, y = y)
  # },

  .length = function() {
    dim(self$data)[1]
  },

  active = list(
    raw_folder = function() {
      file.path(self$root_path, "mnist", "raw")
    },

    processed_folder = function() {
      file.path(self$root_path, "mnist", "processed")
    }
  )
)

#' @describeIn mnist_dataset Kuzushiji-MNIST cursive Japanese character dataset.
#' @export
kmnist_dataset <- dataset(
  name = "kmnist_dataset",
  inherit = mnist_dataset,
  archive_size = "21 MB",
  resources = list(
    c("http://codh.rois.ac.jp/kmnist/dataset/kmnist/train-images-idx3-ubyte.gz", "bdb82020997e1d708af4cf47b453dcf7"),
    c("http://codh.rois.ac.jp/kmnist/dataset/kmnist/train-labels-idx1-ubyte.gz", "e144d726b3acfaa3e44228e80efcd344"),
    c("http://codh.rois.ac.jp/kmnist/dataset/kmnist/t10k-images-idx3-ubyte.gz", "5c965bf0a639b31b8f53240b1b52f4d7"),
    c("http://codh.rois.ac.jp/kmnist/dataset/kmnist/t10k-labels-idx1-ubyte.gz", "7320c461ea6c1c855c0b718fb2a4b134")
  ),
  classes = c('o', 'ki', 'su', 'tsu', 'na', 'ha', 'ma', 'ya', 're', 'wo')
)

#' @describeIn mnist_dataset Extended MNIST dataset with high-precision test data (QMNIST).
#' @export
qmnist_dataset <- dataset(
  name = "qmnist_dataset",
  inherit = mnist_dataset,
  archive_size = "70 MB",

  resources = list(
    train = list(
      c("https://raw.githubusercontent.com/facebookresearch/qmnist/master/qmnist-train-images-idx3-ubyte.gz", "ed72d4157d28c017586c42bc6afe6370"),
      c("https://raw.githubusercontent.com/facebookresearch/qmnist/master/qmnist-train-labels-idx2-int.gz", "0058f8dd561b90ffdd0f734c6a30e5e4")
    ),
    test = list(
      c("https://raw.githubusercontent.com/facebookresearch/qmnist/master/qmnist-test-images-idx3-ubyte.gz", "1394631089c404de565df7b7aeaf9412"),
      c("https://raw.githubusercontent.com/facebookresearch/qmnist/master/qmnist-test-labels-idx2-int.gz", "5b5b05890a5e13444e108efe57b788aa")
    ),
    nist = list(
      c("https://raw.githubusercontent.com/facebookresearch/qmnist/master/xnist-images-idx3-ubyte.xz", "7f124b3b8ab81486c9d8c2749c17f834"),
      c("https://raw.githubusercontent.com/facebookresearch/qmnist/master/xnist-labels-idx2-int.xz", "5ed0e788978e45d4a8bd4b7caec3d79d")
    )
  ),
  files = list(
    train = "training.rds",
    test = "test.rds",
    nist = "nist.rds"
  ),
  classes = c(
    '0 - zero', '1 - one', '2 - two', '3 - three', '4 - four',
    '5 - five', '6 - six', '7 - seven', '8 - eight', '9 - nine'
  ),

  initialize = function(
    root = tempdir(),
    split = "train",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    split <- match.arg(split, c("train", "test", "nist"))
    self$split <- split
    self$root_path <- root
    self$transform <- transform
    self$target_transform <- target_transform

    if (download){
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")

    data_file <- self$files[[split]]
    data <- readRDS(file.path(self$processed_folder, data_file))

    self$data <- data[[1]]
    self$targets <- data[[2]][, 1] + 1L
  },

  download = function() {
    fs::dir_create(self$raw_folder)
    fs::dir_create(self$processed_folder)

    split_exists <- function(split_name) {
      fs::file_exists(file.path(self$processed_folder, self$files[[split_name]]))
    }

    if (all(vapply(names(self$resources), split_exists, logical(1))))
      return(NULL)

    cli_inform("Downloading {.cls {class(self)[[1]]}} ...")
    for (split_name in names(self$resources)) {
      if (split_exists(split_name))
        next

      for (r in self$resources[[split_name]]) {
        filename <- basename(r[1])
        destpath <- file.path(self$raw_folder, filename)

        archive <- download_and_cache(r[1], prefix = glue::glue("qmnist-{split_name}"))
        fs::file_copy(archive, destpath, overwrite = TRUE)

        if (tools::md5sum(destpath) != r[2])
          runtime_error("Corrupt file! Delete the file in {archive} and try again.")
      }

      cli_inform("Processing {.val {split_name}} split for {.cls {class(self)[[1]]}} ...")

      if (split_name == "train") {
        saveRDS(
          list(
            read_sn3_pascalvincent(file.path(self$raw_folder, "qmnist-train-images-idx3-ubyte.gz")),
            read_sn3_pascalvincent(file.path(self$raw_folder, "qmnist-train-labels-idx2-int.gz"))
          ),
          file.path(self$processed_folder, self$files$train)
        )
      }

      if (split_name == "test") {
        saveRDS(
          list(
            read_sn3_pascalvincent(file.path(self$raw_folder, "qmnist-test-images-idx3-ubyte.gz")),
            read_sn3_pascalvincent(file.path(self$raw_folder, "qmnist-test-labels-idx2-int.gz"))
          ),
          file.path(self$processed_folder, self$files$test)
        )
      }

      if (split_name == "nist") {
        saveRDS(
          list(
            read_sn3_pascalvincent(file.path(self$raw_folder, "xnist-images-idx3-ubyte.xz")),
            read_sn3_pascalvincent(file.path(self$raw_folder, "xnist-labels-idx2-int.xz"))
          ),
          file.path(self$processed_folder, self$files$nist)
        )
      }
    }

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  check_exists = function() {
    fs::file_exists(file.path(self$processed_folder, self$files[[self$split]]))
  },

  active = list(
    raw_folder = function() {
      file.path(self$root_path, "qmnist", "raw")
    },
    processed_folder = function() {
      file.path(self$root_path, "qmnist", "processed")
    }
  )
)

#' @describeIn mnist_dataset Fashion-MNIST clothing image dataset.
#' @export
fashion_mnist_dataset <- dataset(
  name = "fashion_mnist_dataset",
  inherit = mnist_dataset,
  archive_size = "30 MB",
  resources = list(
    c("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz", "8d4fb7e6c68d591d4c3dfef9ec88bf0d"),
    c("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz", "25c81989df183df01b3e8a0aad5dffbe"),
    c("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz", "bef4ecab320f06d8554ea6380940ec79"),
    c("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz", "bb300cfdad3c16e7a12a480ee83cd310")
  ),
  classes = c(
    "T-shirt/top", "Trouser", "Pullover", "Dress", "Coat",
    "Sandal", "Shirt", "Sneaker", "Bag", "Ankle boot"
  )
)

#' @describeIn mnist_dataset EMNIST collection with digits and letters arranged in multiple datasets.
#' @param dataset to select within \code{c("byclass", "bymerge", "balanced", "letters", "digits", "mnist")}
#' representing the subset of emnist collection made of a set of classes.
#' You can look at dataset attribute `$classes` to see the actual classes.
#' @export
emnist_collection <- dataset(
  name = "emnist_collection",
  inherit = mnist_dataset,
  archive_size = "540 MB",

  resources = list(
    c("https://biometrics.nist.gov/cs_links/EMNIST/gzip.zip", "58c8d27c78d21e728a6bc7b3cc06412e")
  ),
  rds_file = function(split, kind) paste0(split,"-",kind,".rds"),
  classes_all_dataset = list(
    byclass = c(
      "0","1","2","3","4","5","6","7","8","9",
      LETTERS,
      letters[c(1:9,11,12,14,17,20)]
    ),
    bymerge = c(
      "0","1","2","3","4","5","6","7","8","9",
      LETTERS[1:26],
      letters[c(1:4,6,7,10,13,16,19,21,22,24,25,26)]
    ),
    balanced = c(
      "0","1","2","3","4","5","6","7","8","9",
      LETTERS,
      letters[c(1:9,11,12,14,17,20)]
    ),
    letters = letters,
    digits = as.character(0:9),
    mnist = as.character(0:9)
  ),

  initialize = function(
    root = tempdir(),
    split = "test",
    dataset = "balanced",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$split <- match.arg(split, choices = c("train", "test"))
    self$dataset <- match.arg(dataset,  choices = names(self$classes_all_dataset))
    self$root_path <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$classes <- self$classes_all_dataset[[self$dataset]]
    self$class <- self$classes

    if (download) {
      cli_inform("Dataset {.val {self$dataset}} split {.val {self$split}} of {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")

    dataset_lst <- readRDS(file.path(self$processed_folder, self$rds_file(self$split, self$dataset)))
    self$data <- dataset_lst[[1]]
    self$targets <- dataset_lst[[2]] + 1L

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$class)} classes.")
  },

  download = function() {
    if (self$check_exists()) return(NULL)

    fs::dir_create(self$raw_folder)
    fs::dir_create(self$processed_folder)

    url <- self$resources[[1]][1]
    archive <- download_and_cache(url, prefix = class(self)[1])

    if (tools::md5sum(archive) != self$resources[[1]][2])
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")

    unzip_dir <- file.path(self$raw_folder, "unzipped")
    fs::dir_create(unzip_dir)
    unzip(archive, exdir = unzip_dir)

    # unzip second level of archives
    unzipped_root <- fs::dir_ls(unzip_dir, type = "directory", recurse = FALSE)[1]

    # only manage extraction of the 2 ubyte.gz under interest
    img <- file.path(unzipped_root, glue::glue("emnist-{self$dataset}-{self$split}-images-idx3-ubyte.gz"))
    lbl <- file.path(unzipped_root, glue::glue("emnist-{self$dataset}-{self$split}-labels-idx1-ubyte.gz"))
    dataset_set <- list(read_sn3_pascalvincent(img), read_sn3_pascalvincent(lbl))
    saveRDS(dataset_set, file.path(self$processed_folder, self$rds_file(self$split, self$dataset)))

  },
  # only manage existence of the rds file under interest
  check_exists = function() {
    fs::file_exists(file.path(self$processed_folder, self$rds_file(self$split, self$dataset)))
  },

  active = list(
    raw_folder = function() {
      file.path(self$root_path, "emnist_collection", "raw")
    },

    processed_folder = function() {
      file.path(self$root_path, "emnist_collection", "processed")
    }
  )
)

read_sn3_pascalvincent <- function(path) {
  x <- if (grepl("\\\\.xz$", path, ignore.case = TRUE)) {
    xzfile(path, open = "rb")
  } else {
    gzfile(path, open = "rb")
  }
  on.exit({close(x)})

  magic <- readBin(x, endian = "big", what = integer(), n = 1)
  n_dimensions <- magic %% 256
  ty <- magic %/% 256

  dim <- readBin(x, what = integer(), size = 4, endian = "big",
          n = n_dimensions)

  a <- readBin(
    x,
    what = "int", endian = "big", n = prod(dim),
    size = 1, signed = FALSE
  )

  a <- array(a, dim = rev(dim))
  a <- aperm(a, perm = rev(seq_along(dim)))
  a
}

#' @describeIn mnist_dataset Deprecated. Please use emnist_collection.
#' @param kind the `dataset` in `emnist_collection`.
#' @param ... the other `emnist_collection` parameters.
#' @export
emnist_dataset <- function(kind, ...){
  .Deprecated("emnist_collection")
  emnist_collection(dataset = kind, ...)
}

context("dataset-mnist")

dir <- withr::local_tempdir()

test_that("tests for the mnist dataset", {

  expect_error(
    ds <- mnist_dataset(tempfile())
  )

  ds <- mnist_dataset(dir, download = TRUE)

  i <- ds[1]
  expect_equal(dim(i[[1]]), c(1, 28, 28))
  expect_equal(i[[2]], 6)
  expect_length(ds, 60000)

  raw_item <- ds$.getitem(5)
  expect_length(raw_item, 2)
  expect_named(raw_item, c("x", "y"))
  expect_equal(dim(raw_item$x), c(1, 28, 28))
  expect_equal(length(raw_item$y), 1)

  ds <- mnist_dataset(dir, transform = transform_to_tensor)

  dl <- torch::dataloader(ds, batch_size = 32)
  expect_length(dl, 1875)
  iter <- dataloader_make_iter(dl)
  i <- dataloader_next(iter)
  expect_tensor_shape(i[[1]], c(32, 1, 28, 28))
  expect_tensor_shape(i[[2]], 32)
  expect_true((torch_max(i[[1]]) <= 1)$item())
  expect_named(i, c("x", "y"))

  # Regression: grayscale channel must be preserved for larger batches.
  dl_128 <- torch::dataloader(ds, batch_size = 128)
  i_128 <- dataloader_next(dataloader_make_iter(dl_128))
  expect_tensor_shape(i_128$x, c(128, 1, 28, 28))
  expect_tensor_shape(i_128$y, 128)

})

test_that("tests for the kmnist dataset", {

  expect_error(
    ds <- kmnist_dataset(tempfile())
  )

  ds <- kmnist_dataset(dir, download = TRUE)

  i <- ds[1]
  expect_equal(dim(i[[1]]), c(1, 28, 28))
  expect_equal(i[[2]], 6)
  expect_length(ds, 60000)

  ds <- kmnist_dataset(dir, transform = transform_to_tensor)
  dl <- torch::dataloader(ds, batch_size = 32)
  expect_length(dl, 1875)
  iter <- dataloader_make_iter(dl)
  i <- dataloader_next(iter)
  expect_tensor_shape(i[[1]], c(32, 1, 28, 28))
  expect_tensor_shape(i[[2]], 32)
  expect_true((torch_max(i[[1]]) <= 1)$item())
  expect_named(i, c("x", "y"))
})

test_that("fashion_mnist_dataset loads correctly", {

  ds <- fashion_mnist_dataset(
    root = dir,
    train = TRUE,
    download = TRUE
  )

  expect_s3_class(ds, "fashion_mnist_dataset")
  expect_type(ds$.getitem(1), "list")
  expect_named(ds$.getitem(1), c("x", "y"))
  expect_equal(dim(ds$.getitem(1)$x), c(1, 28, 28))
  expect_true(ds$.getitem(1)$y >= 1 && ds$.getitem(1)$y <= 10)

  ds2 <- fashion_mnist_dataset(dir, transform = transform_to_tensor)
  dl <- torch::dataloader(ds2, batch_size = 32)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)
  expect_tensor_shape(batch$x, c(32, 1, 28, 28))
  expect_tensor_shape(batch$y, 32)
  expect_named(batch, c("x", "y"))
})

test_that("tests for the emnist dataset", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_error(
    ds <- emnist_collection(root = tempfile())
  )

  emnist <- emnist_collection(dir, dataset = "balanced", download = TRUE)
  expect_length(emnist, 18800)
  first_item <- emnist[1]
  expect_named(first_item, c("x", "y"))
  expect_true(is.array(first_item$x))
  expect_equal((first_item[[2]]), 42)

  emnist <- emnist_collection(dir, dataset = "byclass", split = "test", download = TRUE)
  expect_length(emnist, 116323)
  first_item <- emnist[1]
  expect_named(first_item, c("x", "y"))
  expect_true(is.array(first_item$x))
  expect_equal(dim(first_item$x), c(1, 1, 28, 28))
  expect_equal((first_item[[2]]), 19)

  emnist <- emnist_collection(dir, dataset = "bymerge", download = TRUE)
  expect_length(emnist, 116323)
  first_item <- emnist[1]
  expect_named(first_item, c("x", "y"))
  expect_true(is.array(first_item$x))
  expect_equal((first_item[[2]]), 25)

  emnist <- emnist_collection(dir, dataset = "letters", split = "train", download = TRUE,
                           transform = transform_to_tensor)
  expect_length(emnist, 124800)
  first_item <- emnist[1]
  expect_named(first_item, c("x", "y"))
  expect_tensor(first_item$x)
  expect_tensor_shape(first_item$x, c(1, 1,28,28))
  expect_equal((first_item[[2]]), 24)

  emnist <- emnist_collection(dir, dataset = "digits", download = TRUE)
  expect_length(emnist, 40000)
  first_item <- emnist[1]
  expect_named(first_item, c("x", "y"))
  expect_true(is.array(first_item$x))
  expect_equal((first_item[[2]]), 1)

  emnist <- emnist_collection(dir, dataset = "mnist", split = "train", download = TRUE)
  expect_length(emnist, 60000)
  first_item <- emnist[1]
  expect_named(first_item, c("x", "y"))
  expect_true(is.array(first_item$x))
  expect_equal((first_item[[2]]), 5)

  ds2 <- emnist_collection(
    root = dir,
    dataset = "balanced",
    split = "test",
    transform = transform_to_tensor,
    download = TRUE
  )
  dl <- torch::dataloader(ds2, batch_size = 32)
  iter <- torch::dataloader_make_iter(dl)
  batch <- torch::dataloader_next(iter)
  expect_tensor_shape(batch$x, c(32, 1, 28, 28))
  expect_tensor_shape(batch$y, 32)
  expect_named(batch, c("x", "y"))
})


test_that("tests for the qmnist dataset", {

  expect_error(
      ds <- qmnist_dataset(tempfile()),
      "Dataset not found."
  )

  for (split in c("train", "test", "nist")) {

    ds <- qmnist_dataset(dir, split = split, download = TRUE)

    i <- ds[1]
    expect_equal(dim(i[[1]]), c(1, 28, 28))
    expect_true(i[[2]] %in% 1:10)

    expect_gt(length(ds), 0)

    ds <- qmnist_dataset(dir, split = split, transform = transform_to_tensor)
    dl <- torch::dataloader(ds, batch_size = 32)
    iter <- dataloader_make_iter(dl)
    i <- dataloader_next(iter)
    expect_tensor_shape(i[[1]], c(32, 1, 28, 28))
    expect_tensor_shape(i[[2]], 32)
    expect_true((torch_max(i[[1]]) <= 1)$item())
    expect_named(i, c("x", "y"))
  }
})

test_that("tests for the emnist_dataset is deprecated", {
  skip_on_cran()

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_warning(
    emnist_dataset(kind = "digits", download = TRUE),
    "'emnist_dataset' is deprecated."
  )
})

#' WHOI-Plankton Dataset
#'
#' The WHOI-Plankton and WHOI-Plankton small are **image classification** datasets
#' from the Woods Hole Oceanographic Institution (WHOI) of microscopic marine plankton.
#' https://hdl.handle.net/10.1575/1912/7341
#' Images were collected in situ by automated submersible imaging-in-flow cytometry
#' with an instrument called Imaging FlowCytobot (IFCB). They are small grayscale images
#' of varying size.
#' Images are classified into 100 classes, with an overview  available in
#' [project Wiki page](https://whoigit.github.io/whoi-plankton/)
#' Dataset size is 957k and 58k respectively, and each provides a train / val / test split.
#'
#' @inheritParams eurosat_dataset
#'
#' @return A torch dataset with a
#' - `classes` attribute providing the vector of class names.
#'
#' Each element is a named list:
#' - `x`: a H x W x 1 integer array representing an grayscale image.
#' - `y`: the class id of the image.
#'
#' @examples
#' \dontrun{
#' # Load the small plankton dataset and turn images into tensor images
#' plankton <- whoi_small_plankton_dataset(download = TRUE, transform = transform_to_tensor)
#'
#' # Access the first item
#' first_item <- plankton[1]
#' first_item$x  # a tensor grayscale image with shape {1, H, W}
#' first_item$y  # id of the plankton class.
#' plankton$classes[first_item$y] # name of the plankton class
#'
#' # Load the full plankton dataset
#' plankton <- whoi_plankton_dataset(download = TRUE)
#'
#' # Access the first item
#' first_item <- plankton[1]
#' first_item$x  # grayscale image array with shape {H, W}
#' first_item$y  # id of the plankton class.
#' }
#'
#' @name whoi_plankton_dataset
#' @title WHOI Plankton Datasets
#' @family classification_dataset
#' @export
whoi_small_plankton_dataset <- torch::dataset(
  name = "whoi_small_plankton",
  resources = data.frame(
    split = c("test", "train", "train", "val"),
    url = c("https://huggingface.co/datasets/nf-whoi/whoi-plankton-small/resolve/main/data/test-00000-of-00001.parquet?download=true",
            "https://huggingface.co/datasets/nf-whoi/whoi-plankton-small/resolve/main/data/train-00000-of-00002.parquet?download=true",
            "https://huggingface.co/datasets/nf-whoi/whoi-plankton-small/resolve/main/data/train-00001-of-00002.parquet?download=true",
            "https://huggingface.co/datasets/nf-whoi/whoi-plankton-small/resolve/main/data/validation-00000-of-00001.parquet?download=true"),
    md5 = c("91b7794e4286c138bda667121ce91d2c",
            "170921030aee26f9676725a0c55a4420",
            "332ebc8b822058cd98f778099927c50e",
            "f0747ae16fc7cd6946ea54c3fe1f30b4"),
    size = c(217e6, 396e6, 383e6, 112e6)
  ),

  initialize = function(
    split = "val",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
    if (!requireNamespace("arrow", quietly = TRUE)) {
      install.packages("arrow")
    }
    if (!requireNamespace("prettyunits", quietly = TRUE)) {
      install.packages("prettyunits")
    }

    self$split <- match.arg(split, c("train", "val", "test"))
    self$transform <- transform
    self$target_transform <- target_transform
    self$archive_url <- self$resources[self$resources$split == split,]$url
    self$archive_size <- prettyunits::pretty_bytes(sum(self$resources[self$resources$split == split,]$size))
    self$archive_md5 <- self$resources[self$resources$split == split,]$md5
    self$split_file <- sapply(self$archive_url, function(x) file.path(rappdirs::user_cache_dir("torch"), class(self)[1], sub("\\?download=.*", "", basename(x))))

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists())
      runtime_error("Dataset not found. You can use `download = TRUE` to download it.")

    self$.data <- arrow::open_dataset(self$split_file)
    self$classes <- jsonlite::parse_json(self$.data$metadata$huggingface, simplifyVector = TRUE)$info$features$label$names
    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {
    if (self$check_exists())
      return(NULL)

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    # Download the dataset parquet
    archive <- sapply(self$archive_url, download_and_cache, prefix = class(self)[1])
    if (!all(tools::md5sum(archive) == self$archive_md5))
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")
    for (file_id in seq_along(archive)) {
      fs::file_move(archive[file_id], self$split_file[file_id])
    }
  },
  check_exists = function() {
    all(fs::file_exists(self$split_file))
  },

  .getitem = function(index) {
    df <- self$.data[index,]$to_data_frame()
    x_raw <- df$image$bytes %>% unlist() %>% as.raw()
    if (tolower(tools::file_ext(df$image$path)) == "jpg") {
      x <- jpeg::readJPEG(x_raw)
    } else {
      x <- png::readPNG(x_raw)
    }
    y <- df$label + 1L

    if (!is.null(self$transform))
      x <- self$transform(x)

    if (!is.null(self$target_transform))
      y <- self$target_transform(y)

    list(x = x, y = y)
  },

  .length = function() {
    self$.data$num_rows
  }

)


#' WHOI-Plankton Dataset
#'
#' @inheritParams whoi_plankton_dataset
#' @rdname whoi_plankton_dataset
#' @export
whoi_plankton_dataset <- torch::dataset(
  name = "whoi_plankton",
  inherit = whoi_small_plankton_dataset,
  archive_size = "9.1 GB",
  resources = data.frame(
    split = c(rep("test", 4), rep("train", 13), rep("val", 2)),
    url = c(paste0("https://huggingface.co/datasets/nf-whoi/whoi-plankton/resolve/main/data/test-0000",0:3,"-of-00004.parquet?download=true"),
            paste0("https://huggingface.co/datasets/nf-whoi/whoi-plankton/resolve/main/data/train-0000",0:9,"-of-00013.parquet?download=true"),
            paste0("https://huggingface.co/datasets/nf-whoi/whoi-plankton/resolve/main/data/train-000",10:12,"-of-00013.parquet?download=true"),
            paste0("https://huggingface.co/datasets/nf-whoi/whoi-plankton/resolve/main/data/validation-0000",0:1,"-of-00002.parquet?download=true")),
    md5 = c("cd41b344ec4b6af83e39c38e19f09190",
            "aa0965c0e59f7b1cddcb3c565d80edf3",
            "b2a75513f1a084724e100678d8ee7180",
            "a03c4d52758078bfb0799894926d60f6",
            "07eaff140f39868a8bcb1d3c02ebe60f",
            "87c927b9fbe0c327b7b9ae18388b4fcf",
            "456efd91901571a41c2157732880f6b8",
            "dc929fde45e3b2e38bdd61a15566cf32",
            "f92ab6cfb4a3dd7e0866f3fdf8dbc33c",
            "61c555bba39b6b3ccb4f02a5cf07e762",
            "57e03cecf2b5d97912ed37e1b8fc6263",
            "56081cc99e61c36e89db1566dbbf06c1",
            "60b7998630468cb18880660c81d1004a",
            "1fa94ceb54d4e53643a0d8cf323af901",
            "7a7be4e3dfdc39a50c8ca086a4d9a8de",
            "07194caf75805e956986cba68e6b398e",
            "0f4d47f240cd9c30a7dd786171fa40ca",
            "db827a7de8790cdcae67b174c7b8ea5e",
            "d3181d9ffaed43d0c01f59455924edca"),
    size = c(rep(450e6, 4), rep(490e6, 13), rep(450e6, 2))
  )
)


#' Coralnet Dataset
#'
#' Small Coralnet dataset is an image **classification dataset**
#' of very large submarine coral reef images annotated into 3 classes
#' and produced by  [CoralNet](https://coralnet.ucsd.edu),
#' a resource for benthic images classification.
#'
#' @inheritParams whoi_plankton_dataset
#' @family classification_dataset
#' @export
whoi_small_coralnet_dataset <- torch::dataset(
  name = "whoi_small_coralnet",
  inherit = whoi_small_plankton_dataset,
  archive_size = "2.1 GB",
  resources = data.frame(
    split = c("test", rep("train", 4), "val"),
    url = c("https://huggingface.co/datasets/nf-whoi/coralnet-small/resolve/main/data/test-00000-of-00001.parquet?download=true",
            paste0("https://huggingface.co/datasets/nf-whoi/coralnet-small/resolve/main/data/train-0000",0:3,"-of-00004.parquet?download=true"),
            "https://huggingface.co/datasets/nf-whoi/coralnet-small/resolve/main/data/validation-00000-of-00001.parquet?download=true"),
    md5 = c("f9a3ce864fdbeb5f1f3d243fe1285186",
            "82269e2251db22ef213e438126198afd",
            "82d2cafbad7740e476310565a2bcd44e",
            "f4dd2d2effc1f9c02918e3ee614b85d3",
            "d66ec691a4c5c63878a9cfff164a6aaf",
            "7ea146b9b2f7b6cee99092bd44182d06"),
    size = c(430e6, rep(380e6, 4), 192e6)
  )
)

#' ConvNeXt Detection Models (Faster R-CNN style)
#'
#' @description
#' Object detection models combining a ConvNeXt backbone with a Feature Pyramid
#' Network (FPN) and the Faster R-CNN detection head. The architecture mirrors
#' [model_fasterrcnn_resnet50_fpn()], with the ResNet backbone replaced by
#' ConvNeXt variants. The design follows the paper
#' [A ConvNet for the 2020s](https://arxiv.org/abs/2201.03545).
#'
#' ## Available Models
#' - `model_convnext_tiny_detection()`
#' - `model_convnext_small_detection()`
#' - `model_convnext_base_detection()`
#'
#' ## Backbone Performance (ImageNet-1k)
#'
#' Accuracy metrics reflect backbone classification performance only.
#' Detection head weights are randomly initialized and must be fine-tuned
#' on task-specific labelled data before meaningful predictions are produced.
#'
#' ```
#' | Model                             | Top-1 Acc | Top-5 Acc | Params  | GFLOPS | File Size | Backbone Weights              | Notes                    |
#' |-----------------------------------|-----------|-----------|---------|--------|-----------|-------------------------------|--------------------------|
#' | model_convnext_tiny_detection     | 82.5%     | 96.1%     | 28.6M   | 4.46   | 109 MB    | IMAGENET1K_V1                 | Tiny backbone, FPN head  |
#' | model_convnext_small_detection    | 83.6%     | 96.7%     | 50.2M   | 8.68   | 192 MB    | IMAGENET1K_V1 (22k pretrain)  | Small backbone, FPN head |
#' | model_convnext_base_detection     | 84.1%     | 96.9%     | 88.6M   | 15.36  | 338 MB    | IMAGENET1K_V1                 | Base backbone, FPN head  |
#' ```
#'
#' ## FPN Channel Configuration
#'
#' Each ConvNeXt variant produces four feature maps (C2–C5) fed into the FPN.
#' Channel widths differ between Tiny/Small and Base:
#'
#' ```
#' | Variant | FPN in_channels          | FPN out_channels |
#' |---------|--------------------------|------------------|
#' | Tiny    | c(96, 192, 384, 768)     | 256              |
#' | Small   | c(96, 192, 384, 768)     | 256              |
#' | Base    | c(128, 256, 512, 1024)   | 256              |
#' ```
#'
#' ## Weights Selection
#' - All variants use `IMAGENET1K_V1` backbone weights by default (supervised ImageNet-1k).
#' - The Small variant backbone (`model_convnext_small_22k`) was additionally
#'   pretrained on ImageNet-22k prior to fine-tuning on ImageNet-1k.
#' - Detection head weights are **randomly initialized** — bounding-box
#'   predictions are meaningless without fine-tuning on labelled detection data.
#' - Set `pretrained_backbone = TRUE` to load ImageNet backbone weights.
#'
#' @inheritParams model_fasterrcnn_resnet50_fpn
#' @param pretrained_backbone Logical. If `TRUE`, loads ImageNet-pretrained
#'   ConvNeXt backbone weights. Default: `FALSE`.
#'
#' @note
#' Detection head weights are randomly initialized. Predicted bounding boxes
#' will be arbitrary until the detection head is trained on labelled data.
#' Only the backbone benefits from `pretrained_backbone = TRUE`.
#'
#' @examples
#' \dontrun{
#' library(magrittr)
#' norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants
#' norm_std  <- c(0.229, 0.224, 0.225)
#'
#' url <- paste0("https://upload.wikimedia.org/wikipedia/commons/thumb/",
#'               "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg")
#' img <- base_loader(url) %>%
#'   transform_to_tensor() %>%
#'   transform_resize(c(520, 520))
#'
#' input <- img %>% transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)    # Add batch dimension: (1, 3, H, W)
#'
#' # ConvNeXt Tiny detection
#' model <- model_convnext_tiny_detection(pretrained_backbone = TRUE)
#' model$eval()
#' # Please wait 2 mins + on CPU
#' pred     <- model(batch)$detections[[1]]
#' num_boxes <- as.integer(pred$boxes$size()[1])
#' topk     <- pred$scores$topk(k = 5)[[2]]
#' boxes    <- pred$boxes[topk, ]
#' labels   <- imagenet_classes(as.integer(pred$labels[topk]))
#'
#' # `draw_bounding_box()` may fail if bbox values are not consistent.
#' if (num_boxes > 0) {
#'   boxed <- draw_bounding_boxes(img, boxes, labels = labels)
#'   tensor_image_browse(boxed)
#' }
#' }
#'
#' @importFrom torch nn_module
#' @family object_detection_model
#' @rdname model_convnext_detection
#' @name model_convnext_detection
NULL


convnext_fpn_backbone_tiny <- function(pretrained_backbone = FALSE, ...) {
  convnext <- model_convnext_tiny_1k(pretrained = pretrained_backbone, ...)

  convnext_body <- torch::nn_module(
    initialize = function() {
      self$model <- convnext
    },
    forward = function(x) {
      c2 <- x %>%
        self$model$downsample_layers[[1]]() %>%
        self$model$stages[[1]]()

      c3 <- c2 %>%
        self$model$downsample_layers[[2]]() %>%
        self$model$stages[[2]]()

      c4 <- c3 %>%
        self$model$downsample_layers[[3]]() %>%
        self$model$stages[[3]]()

      c5 <- c4 %>%
        self$model$downsample_layers[[4]]() %>%
        self$model$stages[[4]]()

      list(c2, c3, c4, c5)
    }
  )

  backbone_module <- torch::nn_module(
    initialize = function() {
      self$body <- convnext_body()
      self$fpn <- fpn_module(
        in_channels = c(96, 192, 384, 768),
        out_channels = 256
      )
    },
    forward = function(x) {
      c2_to_c5 <- self$body(x)
      self$fpn(c2_to_c5)
    }
  )

  backbone <- backbone_module()
  backbone$out_channels <- 256
  backbone
}


convnext_fpn_backbone_small <- function(pretrained_backbone = FALSE, ...) {
  convnext <- model_convnext_small_22k(pretrained = pretrained_backbone, ...)

  convnext_body <- torch::nn_module(
    initialize = function() {
      self$model <- convnext
    },
    forward = function(x) {
      c2 <- x %>%
        self$model$downsample_layers[[1]]() %>%
        self$model$stages[[1]]()

      c3 <- c2 %>%
        self$model$downsample_layers[[2]]() %>%
        self$model$stages[[2]]()

      c4 <- c3 %>%
        self$model$downsample_layers[[3]]() %>%
        self$model$stages[[3]]()

      c5 <- c4 %>%
        self$model$downsample_layers[[4]]() %>%
        self$model$stages[[4]]()

      list(c2, c3, c4, c5)
    }
  )

  backbone_module <- torch::nn_module(
    initialize = function() {
      self$body <- convnext_body()
      self$fpn <- fpn_module(
        in_channels = c(96, 192, 384, 768),
        out_channels = 256
      )
    },
    forward = function(x) {
      c2_to_c5 <- self$body(x)
      self$fpn(c2_to_c5)
    }
  )

  backbone <- backbone_module()
  backbone$out_channels <- 256
  backbone
}


convnext_fpn_backbone_base <- function(pretrained_backbone = FALSE, ...) {
  convnext <- model_convnext_base_1k(pretrained = pretrained_backbone, ...)

  convnext_body <- torch::nn_module(
    initialize = function() {
      self$model <- convnext
    },
    forward = function(x) {
      c2 <- x %>%
        self$model$downsample_layers[[1]]() %>%
        self$model$stages[[1]]()

      c3 <- c2 %>%
        self$model$downsample_layers[[2]]() %>%
        self$model$stages[[2]]()

      c4 <- c3 %>%
        self$model$downsample_layers[[3]]() %>%
        self$model$stages[[3]]()

      c5 <- c4 %>%
        self$model$downsample_layers[[4]]() %>%
        self$model$stages[[4]]()

      list(c2, c3, c4, c5)
    }
  )

  backbone_module <- torch::nn_module(
    initialize = function() {
      self$body <- convnext_body()
      self$fpn <- fpn_module(
        in_channels = c(128, 256, 512, 1024),
        out_channels = 256
      )
    },
    forward = function(x) {
      c2_to_c5 <- self$body(x)
      self$fpn(c2_to_c5)
    }
  )

  backbone <- backbone_module()
  backbone$out_channels <- 256
  backbone
}


validate_convnext_num_classes <- function(num_classes) {
  if (num_classes <= 0) {
    cli_abort("{.var num_classes} must be positive")
  }
}


#' @describeIn model_convnext_detection ConvNeXt Tiny with FPN detection head
#' @export
model_convnext_tiny_detection <- function(num_classes = 91,
                                          pretrained_backbone = FALSE,
                                          ...) {
  validate_convnext_num_classes(num_classes)

  backbone <- convnext_fpn_backbone_tiny(
    pretrained_backbone = pretrained_backbone,
    ...
  )

  model <- fasterrcnn_model(backbone, num_classes = num_classes)
  model
}


#' @describeIn model_convnext_detection ConvNeXt Small with FPN detection head
#' @export
model_convnext_small_detection <- function(num_classes = 91,
                                           pretrained_backbone = FALSE,
                                           ...) {
  validate_convnext_num_classes(num_classes)

  backbone <- convnext_fpn_backbone_small(
    pretrained_backbone = pretrained_backbone,
    ...
  )

  model <- fasterrcnn_model(backbone, num_classes = num_classes)
  model
}


#' @describeIn model_convnext_detection ConvNeXt Base with FPN detection head
#' @export
model_convnext_base_detection <- function(num_classes = 91,
                                          pretrained_backbone = FALSE,
                                          ...) {
  validate_convnext_num_classes(num_classes)

  backbone <- convnext_fpn_backbone_base(
    pretrained_backbone = pretrained_backbone,
    ...
  )

  model <- fasterrcnn_model(backbone, num_classes = num_classes)
  model
}

context("models-convnext-detection")

test_that("tests for non-pretrained model_convnext_tiny_detection works with batch", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_convnext_tiny_detection(pretrained_backbone = TRUE)
  input <- base_loader("assets/class/cat/cat.0.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(200, 200)) %>% torch_unsqueeze(1)
  model$eval()
  out <- model(input)
  expect_named(out, c("features", "detections"))
  expect_is(out$detections, "list")
  expect_named(out$detections[[1]], c("boxes", "labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)


  batch <- torch_stack(list(base_loader("assets/class/cat/cat.0.jpg") %>% transform_to_tensor() %>% transform_resize(c(200, 200)),
                            base_loader("assets/class/cat/cat.1.jpg") %>% transform_to_tensor() %>% transform_resize(c(200, 200))),
                       dim = 1)
  model <- model_convnext_tiny_detection(num_classes = 10)
  out <- model(batch)
  expect_named(out, c("features", "detections"))
  expect_is(out$detections, "list")
  expect_length(out$detections, 2)
  expect_named(out$detections[[1]], c("boxes", "labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
  expect_named(out$detections[[2]], c("boxes", "labels", "scores"))
  expect_tensor(out$detections[[2]]$boxes)
  expect_tensor(out$detections[[2]]$labels)
  expect_tensor(out$detections[[2]]$scores)
  expect_equal(out$detections[[2]]$boxes$shape[2], 4L)
})

test_that("tests for pretrained / non-pretrained model_convnext_small_detection", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_convnext_small_detection(pretrained_backbone = TRUE)
  input <- base_loader("assets/class/cat/cat.1.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180, 180)) %>% torch_unsqueeze(1)
  model$eval()
  out <- model(input)
  expect_named(out, c("features", "detections"))
  expect_is(out$detections, "list")
  expect_named(out$detections[[1]], c("boxes", "labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
  # we cannot succesfully assert bbox here :
  #   expect_bbox_is_xyxy(out$detections[[1]]$boxes, 180, 180)
  # }

  model <- model_convnext_small_detection(num_classes = 10)
  out <- model(input)
  expect_named(out, c("features", "detections"))
  expect_is(out$detections, "list")
  expect_named(out$detections[[1]], c("boxes", "labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
})

test_that("tests for pretrained / non-pretrained model_convnext_base_detection", {
  skip_if(Sys.getenv("TEST_HUGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_HUGE_MODELS=1 to enable tests requiring large downloads.")
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_convnext_base_detection(pretrained_backbone = TRUE)
  input <- base_loader("assets/class/cat/cat.2.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180, 180)) %>% torch_unsqueeze(1)
  model$eval()
  out <- model(input)
  expect_named(out, c("features", "detections"))
  expect_is(out$detections, "list")
  expect_named(out$detections[[1]], c("boxes", "labels", "scores"))
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
  if (out$detections[[1]]$boxes$shape[1] > 0) {
    expect_bbox_is_xyxy(out$detections[[1]]$boxes, 180, 180)
  }
})

test_that("tests for non-pretrained model_convnext_base_detection", {
  skip_if(Sys.getenv("TEST_HUGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_HUGE_MODELS=1 to enable tests requiring large downloads.")
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_convnext_base_detection(num_classes = 10)
  input <- base_loader("assets/class/cat/cat.2.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(180, 180)) %>% torch_unsqueeze(1)
  model$eval()
  out <- model(input)
  expect_is(out$detections, "list")
  expect_named(out, c("features", "detections"))
  expect_named(out$detections[[1]], c("boxes", "labels", "scores"))
  expect_tensor(out$detections[[1]]$boxes)
  expect_tensor(out$detections[[1]]$labels)
  expect_tensor(out$detections[[1]]$scores)
  expect_equal(out$detections[[1]]$boxes$shape[2], 4L)
})


test_that("model_convnext_detection handles different image sizes", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  model <- model_convnext_tiny_detection(num_classes = 10)
  model$eval()

  input_224 <- base_loader("assets/class/dog/dog.0.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(224, 224)) %>% torch_unsqueeze(1)
  out_224 <- model(input_224)
  expect_named(out_224, c("features", "detections"))
  expect_named(out_224$detections[[1]], c("boxes", "labels", "scores"))

  input_320 <- base_loader("assets/class/dog/dog.1.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(320, 320)) %>% torch_unsqueeze(1)
  out_320 <- model(input_320)
  expect_named(out_320, c("features", "detections"))
  expect_named(out_320$detections[[1]], c("boxes", "labels", "scores"))

  input_512 <- base_loader("assets/class/dog/dog.2.jpg") %>%
    transform_to_tensor() %>% transform_resize(c(512, 512)) %>% torch_unsqueeze(1)
  out_512 <- model(input_512)
  expect_named(out_512, c("features", "detections"))
  expect_named(out_512$detections[[1]], c("boxes", "labels", "scores"))
})

test_that("model_convnext_detection validates num_classes parameter", {
  skip_on_cran()
  skip_if_not(torch::torch_is_installed())

  expect_no_error(model_convnext_tiny_detection(num_classes = 10, pretrained_backbone = FALSE))
  expect_no_error(model_convnext_tiny_detection(num_classes = 91, pretrained_backbone = FALSE))
  expect_error(model_convnext_tiny_detection(num_classes = 0), "must be positive")
  expect_error(model_convnext_tiny_detection(num_classes = -1), "must be positive")
})

facenet_torchscript_urls <- list(
  ONet = c("https://torch-cdn.mlverse.org/models/vision/v2/models/facenet_onet.pth", "4833e21e3a610b4064b5a3d20683a55f", "2 MB"),
  PNet = c("https://torch-cdn.mlverse.org/models/vision/v2/models/facenet_pnet.pth", "7f59c98ccf07c4ed51caf68fde86373e", "30 KB"),
  RNet = c("https://torch-cdn.mlverse.org/models/vision/v2/models/facenet_rnet.pth", "c19b2f0df8f448455dd7ddbb47dcfa19", "400 KB")
)

#' MTCNN Face Detection Networks
#'
#' These models implement the three-stage Multi-task Cascaded Convolutional Networks (MTCNN)
#' architecture from the paper
#' [Joint Face Detection and Alignment using Multi-task Cascaded Convolutional Networks](https://arxiv.org/abs/1604.02878).
#'
#' MTCNN detects faces and facial landmarks in an image through a coarse-to-fine pipeline:
#' - **PNet** (Proposal Network): Generates candidate face bounding boxes at multiple scales.
#' - **RNet** (Refine Network): Refines candidate boxes, rejecting false positives.
#' - **ONet** (Output Network): Produces final bounding boxes and 5-point facial landmarks.
#'
#' ## Model Variants
#' ```
#' | Model | Input Size     | Parameters | File Size | Outputs                       | Notes                             |
#' |-------|----------------|------------|-----------|-------------------------------|-----------------------------------|
#' | PNet  | ~12×12+        | ~3k        | 30 kB     | 2-class face prob + bbox reg  | Fully conv, sliding window stage  |
#' | RNet  | 24×24          | ~30k       | 400 kB    | 2-class face prob + bbox reg  | Dense layers, higher recall       |
#' | ONet  | 48×48          | ~100k      | 2 MB      | 2-class prob + bbox + 5-point | Landmark detection stage          |
#' ```
#' Inception-ResNet-v1 is a convolutional neural network architecture combining Inception modules
#' with residual connections, designed for face recognition tasks. The model achieves high accuracy
#' on standard face verification benchmarks such as LFW (Labeled Faces in the Wild).
#'
#' ## Model Variants and Performance (LFW accuracy)
#' ```
#' |    Weights     | LFW Accuracy | File Size |
#' |----------------|--------------|-----------|
#' | CASIA-Webface  | 99.05%       | 111 MB    |
#' | VGGFace2       | 99.65%       | 107 MB    |
#' ```
#'
#' - The CASIA-Webface pretrained weights provide strong baseline accuracy.
#' - The VGGFace2 pretrained weights achieve higher accuracy, benefiting from a larger, more diverse dataset.
#'
#' @examples
#' \dontrun{
#' # Example usage of PNet
#' model_pnet <- model_facenet_pnet(pretrained = TRUE)
#' model_pnet$eval()
#' input_pnet <- torch_randn(1, 3, 224, 224)
#' output_pnet <- model_pnet(input_pnet)
#' output_pnet
#'
#' # Example usage of RNet
#' model_rnet <- model_facenet_rnet(pretrained = TRUE)
#' model_rnet$eval()
#' input_rnet <- torch_randn(1, 3, 24, 24)
#' output_rnet <- model_rnet(input_rnet)
#' output_rnet
#'
#' # Example usage of ONet
#' model_onet <- model_facenet_onet(pretrained = TRUE)
#' model_onet$eval()
#' input_onet <- torch_randn(1, 3, 48, 48)
#' output_onet <- model_onet(input_onet)
#' output_onet
#'
#' # Example usage of MTCNN
#' mtcnn <- model_mtcnn(pretrained = TRUE)
#' mtcnn$eval()
#' image_tensor <- torch_randn(c(1, 3, 224, 224))
#' out <- mtcnn(image_tensor)
#' out
#'
#' # Load an image from the web
#' url <- paste0("https://upload.wikimedia.org/wikipedia/commons",
#'               "/b/b4/Catherine_Bell_200101233d_hr_%28cropped%29.jpg")
#' tmp_file <- tempfile(fileext = ".jpg")
#' download.file(url, tmp_file, mode = "wb")
#' img <- jpeg::readJPEG(tmp_file)
#'
#' # Convert to torch tensor [C, H, W] normalized
#' input <- transform_to_tensor(img)  # [C, H, W]
#' batch <- input$unsqueeze(1)   # [1, C, H, W]
#'
#' # Load pretrained model
#' model <- model_facenet_inception_resnet_v1(pretrained = "vggface2")
#' model$eval()
#' output <- model(batch)
#' output
#'
#' # Example usage of Inception-ResNet-v1 with CASIA-Webface Weights
#' model <- model_facenet_inception_resnet_v1(pretrained = "casia-webface")
#' model$eval()
#' output <- model(batch)
#' output
#' }
#'
#' @importFrom torch nn_module nn_conv2d nn_prelu nn_max_pool2d nn_softmax nn_linear nn_batch_norm2d nn_batch_norm1d
#' @importFrom torch nn_relu nn_dropout nn_adaptive_avg_pool2d nn_sequential torch_randn torch_cat
#' @importFrom torch nnf_interpolate nnf_relu nnf_normalize load_state_dict
#'
#' @inheritParams model_mobilenet_v2
#' @param classify Logical, whether to include the classification head. Default is FALSE.
#' @param num_classes Integer, number of output classes for classification. Default is 10.
#' @param dropout_prob Numeric, dropout probability applied before classification. Default is 0.6.
#'
#' @family object_detection_model
#' @rdname model_facenet
#' @name model_facenet
NULL

#' @describeIn model_facenet PNet (Proposal Network) — small fully-convolutional network for candidate face box generation.
#' @export
model_facenet_pnet <- nn_module(
  classname = "PNet",
  initialize = function(pretrained=TRUE,progress=FALSE,...) {
    self$conv1 <- nn_conv2d(3, 10, kernel_size=3)
    self$prelu1 <- nn_prelu(10)
    self$pool1 <- nn_max_pool2d(2, 2, ceil_mode=TRUE)
    self$conv2 <- nn_conv2d(10, 16, kernel_size=3)
    self$prelu2 <- nn_prelu(16)
    self$conv3 <- nn_conv2d(16, 32, kernel_size=3)
    self$prelu3 <- nn_prelu(32)
    self$conv4_1 <- nn_conv2d(32, 2, kernel_size=1)
    self$softmax4_1 <- nn_softmax(dim=1)
    self$conv4_2 <- nn_conv2d(32, 4, kernel_size=1)

    self$training <- FALSE

    if (pretrained) {
      archive <- download_and_cache(facenet_torchscript_urls$PNet[1], prefix = "facenet")
      if (tools::md5sum(archive) != facenet_torchscript_urls$PNet[2]){
        runtime_error("Corrupt file! Delete the file in {archive} and try again.")
      }
      state_dict <- load_state_dict(archive)
      self$load_state_dict(state_dict)
    }
  },

  forward = function(x) {
    x <- self$conv1(x)
    x <- self$prelu1(x)
    x <- self$pool1(x)
    x <- self$conv2(x)
    x <- self$prelu2(x)
    x <- self$conv3(x)
    x <- self$prelu3(x)
    a <- self$conv4_1(x)
    a <- self$softmax4_1(a)
    b <- self$conv4_2(x)
    list(boxes = b, cls = a)
  }
)

#' @describeIn model_facenet RNet (Refine Network) — medium CNN with dense layers for refining and rejecting false positives.
#' @export
model_facenet_rnet <- nn_module(
  classname = "RNet",
  initialize = function(pretrained=TRUE,progress=FALSE,...) {
    self$conv1 <- nn_conv2d(3, 28, kernel_size=3)
    self$prelu1 <- nn_prelu(28)
    self$pool1 <- nn_max_pool2d(3, 2, ceil_mode=TRUE)
    self$conv2 <- nn_conv2d(28, 48, kernel_size=3)
    self$prelu2 <- nn_prelu(48)
    self$pool2 <- nn_max_pool2d(3, 2, ceil_mode=TRUE)
    self$conv3 <- nn_conv2d(48, 64, kernel_size=2)
    self$prelu3 <- nn_prelu(64)
    self$dense4 <- nn_linear(576, 128)
    self$prelu4 <- nn_prelu(128)
    self$dense5_1 <- nn_linear(128, 2)
    self$softmax5_1 <- nn_softmax(dim=1)
    self$dense5_2 <- nn_linear(128, 4)

    self$training <- FALSE

    if (pretrained) {
      archive <- download_and_cache(facenet_torchscript_urls$RNet[1], prefix = "facenet")
      if (tools::md5sum(archive) != facenet_torchscript_urls$RNet[2]){
        runtime_error("Corrupt file! Delete the file in {archive} and try again.")
      }
      state_dict <- load_state_dict(archive)
      self$load_state_dict(state_dict)
    }
  },

  forward = function(x) {
    x <- self$conv1(x)
    x <- self$prelu1(x)
    x <- self$pool1(x)
    x <- self$conv2(x)
    x <- self$prelu2(x)
    x <- self$pool2(x)
    x <- self$conv3(x)
    x <- self$prelu3(x)
    x <- x$permute(c(1,4,3,2))$contiguous()
    x <- self$dense4(x$view(c(x$size(1), -1)))
    x <- self$prelu4(x)
    a <- self$dense5_1(x)
    a <- self$softmax5_1(a)
    b <- self$dense5_2(x)
    list(boxes = b, cls = a)
  }
)

#' @describeIn model_facenet ONet (Output Network) — deeper CNN that outputs final bounding boxes and 5 facial landmark points.
#' @export
model_facenet_onet <- nn_module(
  classname = "ONet",
  initialize = function(pretrained=TRUE,progress=FALSE,...) {
    self$conv1 <- nn_conv2d(3, 32, kernel_size=3)
    self$prelu1 <- nn_prelu(32)
    self$pool1 <- nn_max_pool2d(3, 2, ceil_mode=TRUE)
    self$conv2 <- nn_conv2d(32, 64, kernel_size=3)
    self$prelu2 <- nn_prelu(64)
    self$pool2 <- nn_max_pool2d(3, 2, ceil_mode=TRUE)
    self$conv3 <- nn_conv2d(64, 64, kernel_size=3)
    self$prelu3 <- nn_prelu(64)
    self$pool3 <- nn_max_pool2d(2, 2, ceil_mode=TRUE)
    self$conv4 <- nn_conv2d(64, 128, kernel_size=2)
    self$prelu4 <- nn_prelu(128)
    self$dense5 <- nn_linear(1152, 256)
    self$prelu5 <- nn_prelu(256)
    self$dense6_1 <- nn_linear(256, 2)
    self$softmax6_1 <- nn_softmax(dim=1)
    self$dense6_2 <- nn_linear(256, 4)
    self$dense6_3 <- nn_linear(256, 10)

    self$training <- FALSE

    if (pretrained) {
      archive <- download_and_cache(facenet_torchscript_urls$ONet[1], prefix = "facenet")
      if (tools::md5sum(archive) != facenet_torchscript_urls$ONet[2]){
        runtime_error("Corrupt file! Delete the file in {archive} and try again.")
      }
      state_dict <- load_state_dict(archive)
      self$load_state_dict(state_dict)
    }
  },

  forward = function(x) {
    x <- self$conv1(x)
    x <- self$prelu1(x)
    x <- self$pool1(x)
    x <- self$conv2(x)
    x <- self$prelu2(x)
    x <- self$pool2(x)
    x <- self$conv3(x)
    x <- self$prelu3(x)
    x <- self$pool3(x)
    x <- self$conv4(x)
    x <- self$prelu4(x)
    x <- x$permute(c(1,4,3,2))$contiguous()
    x <- self$dense5(x$view(c(x$size(1), -1)))
    x <- self$prelu5(x)
    a <- self$dense6_1(x)
    a <- self$softmax6_1(a)
    b <- self$dense6_2(x)
    c <- self$dense6_3(x)
    list(boxes = b, landmarks = c, cls = a)
  }
)

#' @describeIn model_facenet MTCNN (Multi-task Cascaded Convolutional Networks) — face detection and alignment using a cascade of three neural networks
#'
#' @return `model_mtcnn()` returns a named list with three elements:
#' \itemize{
#'   \item \code{boxes}: A tensor of shape \code{(N, 4)} with bounding box coordinates \code{[x1, y1, x2, y2]}.
#'   \item \code{landmarks}: A tensor of shape \code{(N, 10)} with (x, y) coordinates of 5 facial landmarks:
#'         left eye, right eye, nose, left mouth corner, right mouth corner.
#'   \item \code{cls}: A tensor of shape \code{(N, 2)} with face classification probabilities
#'         (face / non-face). The \code{cls} head has two classes:
#'         \itemize{
#'           \item \code{1}: Non-face probability (background)
#'           \item \code{2}: Face probability — use this value for thresholding detections
#'         }
#' }
#' (Here, \code{N} is the number of detected faces in the input image.)
#' @family object_detection_model
#' @export
model_mtcnn <- nn_module(
  classname = "MTCNN",
  initialize = function(
    pretrained = TRUE,
    progress = TRUE,
    ...
  ) {


    self$pnet <- model_facenet_pnet(pretrained=pretrained,...)
    self$rnet <- model_facenet_rnet(pretrained=pretrained,...)
    self$onet <- model_facenet_onet(pretrained=pretrained,...)
  },

  forward = function(x) {
    pnet_out <- self$pnet(x)
    x_rnet <- nnf_interpolate(x, size = c(24, 24), mode = "bilinear", align_corners = FALSE)
    rnet_out <- self$rnet(x_rnet)
    x_onet <- nnf_interpolate(x_rnet, size = c(48, 48), mode = "bilinear", align_corners = FALSE)
    onet_out <- self$onet(x_onet)

    list(boxes = onet_out$boxes, landmarks = onet_out$landmarks, cls = onet_out$cls)
  }
)


load_inception_weights <- function(model, name) {
  if (name == "vggface2") {
    url <- "https://torch-cdn.mlverse.org/models/vision/v2/models/vggface2.pth"
    md5 <- "c446a04f0b22763858226717ba1f7410"
  } else if (name == "casia-webface") {
    url <- "https://torch-cdn.mlverse.org/models/vision/v2/models/casia-webface.pth"
    md5 <- "ff4aff482f6c1941784abba5131bae20"
  }

  archive <- download_and_cache(url,prefix = "facenet")
  if (tools::md5sum(archive) != md5){
    runtime_error("Corrupt file! Delete the file in {archive} and try again.")
  }

  state_dict <- torch::load_state_dict(archive)
  model$load_state_dict(state_dict)
  model
}

BasicConv2d <- nn_module(
  "BasicConv2d",
  initialize = function(in_channels, out_channels, kernel_size, stride, padding = 0) {
    self$conv <- nn_conv2d(in_channels, out_channels, kernel_size, stride, padding, bias = FALSE)
    self$bn <- nn_batch_norm2d(out_channels, eps = 0.001, momentum = 0.1)
  },
  forward = function(x) {
    x %>% self$conv() %>% self$bn() %>% nnf_relu(inplace = TRUE)
  }
)

Block35 <- nn_module(
  "Block35",
  initialize = function(scale = 1.0) {
    self$scale <- scale
    self$branch0 <- BasicConv2d(256, 32, kernel_size = 1, stride = 1)
    self$branch1 <- nn_sequential(
      BasicConv2d(256, 32, kernel_size = 1, stride = 1),
      BasicConv2d(32, 32, kernel_size = 3, stride = 1, padding = 1)
    )
    self$branch2 <- nn_sequential(
      BasicConv2d(256, 32, kernel_size = 1, stride = 1),
      BasicConv2d(32, 48, kernel_size = 3, stride = 1, padding = 1),
      BasicConv2d(48, 64, kernel_size = 3, stride = 1, padding = 1)
    )
    self$conv2d <- nn_conv2d(128, 256, kernel_size = 1, stride = 1)
  },
  forward = function(x) {
    branch0 <- self$branch0(x)
    branch1 <- self$branch1(x)
    branch2 <- self$branch2(x)
    print(branch0$shape)
    print(branch1$shape)
    print(branch2$shape)
    mixed <- torch_cat(list(branch0, branch1, branch2), dim = 2)
    up <- self$conv2d(mixed)
    x + self$scale * up %>% nnf_relu(inplace = TRUE)
  }
)

Block17 <- nn_module(
  initialize = function(scale = 1.0) {
    self$scale <- scale
    self$branch0 <- BasicConv2d(896, 128, kernel_size = 1, stride = 1)
    self$branch1 <- nn_sequential(
      BasicConv2d(896, 128, kernel_size = 1, stride = 1),
      BasicConv2d(128, 128, kernel_size = c(1,7), stride = 1, padding = c(0,3)),
      BasicConv2d(128, 128, kernel_size = c(7,1), stride = 1, padding = c(3,0))
    )
    self$conv2d <- nn_conv2d(256, 896, kernel_size = 1, stride = 1)
    self$relu <- nn_relu(inplace = FALSE)
  },
  forward = function(x) {
    x0 <- self$branch0(x)
    x1 <- self$branch1(x)
    out <- torch_cat(list(x0, x1), dim = 2)
    out <- self$conv2d(out)
    out <- out * self$scale + x
    out %>% self$relu()
  }
)

Block8 <- nn_module(
  initialize = function(scale = 1.0, noReLU = FALSE) {
    self$scale <- scale
    self$noReLU <- noReLU
    self$branch0 <- BasicConv2d(1792, 192, kernel_size = 1, stride = 1)
    self$branch1 <- nn_sequential(
      BasicConv2d(1792, 192, kernel_size = 1, stride = 1),
      BasicConv2d(192, 192, kernel_size = c(1,3), stride = 1, padding = c(0,1)),
      BasicConv2d(192, 192, kernel_size = c(3,1), stride = 1, padding = c(1,0))
    )
    self$conv2d <- nn_conv2d(384, 1792, kernel_size = 1, stride = 1)
    if (!noReLU) {
      self$relu <- nn_relu(inplace = FALSE)
    }
  },
  forward = function(x) {
    x0 <- self$branch0(x)
    x1 <- self$branch1(x)
    out <- torch_cat(list(x0, x1), dim = 2)
    out <- self$conv2d(out)
    out <- out * self$scale + x
    if (!self$noReLU) {
      out <- self$relu(out)
    }
    out
  }
)

Mixed_6a <- nn_module(
  initialize = function() {
    self$branch0 <- BasicConv2d(256, 384, kernel_size = 3, stride = 2, padding = 0)
    self$branch1 <- nn_sequential(
      BasicConv2d(256, 192, kernel_size = 1, stride = 1, padding = 0),
      BasicConv2d(192, 192, kernel_size = 3, stride = 1, padding = 0),
      BasicConv2d(192, 256, kernel_size = 3, stride = 2, padding = 1)
    )
    self$branch2 <- nn_max_pool2d(kernel_size = 3, stride = 2)
  },
  forward = function(x) {
    x0 <- self$branch0(x)
    x1 <- self$branch1(x)
    x2 <- self$branch2(x)

    torch_cat(list(x0, x1, x2), dim = 2)
  }
)

Mixed_7a <- nn_module(
  initialize = function() {
    self$branch0 <- nn_sequential(
      BasicConv2d(896, 256, kernel_size = 1, stride = 1),
      BasicConv2d(256, 384, kernel_size = 3, stride = 2)
    )
    self$branch1 <- nn_sequential(
      BasicConv2d(896, 256, kernel_size = 1, stride = 1),
      BasicConv2d(256, 256, kernel_size = 3, stride = 2)
    )
    self$branch2 <- nn_sequential(
      BasicConv2d(896, 256, kernel_size = 1, stride = 1),
      BasicConv2d(256, 256, kernel_size = 3, stride = 1, padding = 1),
      BasicConv2d(256, 256, kernel_size = 3, stride = 2)
    )
    self$branch3 <- nn_max_pool2d(kernel_size = 3, stride = 2)
  },
  forward = function(x) {
    x0 <- self$branch0(x)
    x1 <- self$branch1(x)
    x2 <- self$branch2(x)
    x3 <- self$branch3(x)

    torch_cat(list(x0, x1, x2, x3), dim = 2)
  }
)

#' @describeIn model_facenet Inception-ResNet-v1 — high-accuracy face recognition model combining Inception modules with residual connections, pretrained on VGGFace2 and CASIA-Webface datasets
#'
#' @return
#' `model_facenet_inception_resnet_v1()` returns a tensor output depending on the \code{classify} argument:
#' \itemize{
#'   \item When \code{classify = FALSE} (default):
#'         A tensor of shape \code{(N, 512)}, where each row is a normalized embedding
#'         vector (L2 norm = 1).
#'         These 512-dimensional FaceNet embeddings can be compared using cosine
#'         similarity or Euclidean distance for face verification and clustering.
#'
#'   \item When \code{classify = TRUE}:
#'         A tensor of shape \code{(N, num_classes)} containing class logits.
#' }
#'
#' @family classification_model
#' @export
model_facenet_inception_resnet_v1 <- nn_module(
  initialize = function(
    pretrained = NULL,
    classify = FALSE,
    num_classes = 10,
    dropout_prob = 0.6,
    ...
  ) {

    if (!is.null(pretrained)) {
      if (pretrained == "vggface2") {
        tmp_classes <- 8631
      } else if (pretrained == "casia-webface") {
        tmp_classes <- 10575
      } else {
        pretrained <- NULL
      }
    }

    self$conv2d_1a <- BasicConv2d(3, 32, kernel_size = 3, stride = 2)
    self$conv2d_2a <- BasicConv2d(32, 32, kernel_size = 3, stride = 1)
    self$conv2d_2b <- BasicConv2d(32, 64, kernel_size = 3, stride = 1, padding = 1)
    self$maxpool_3a <- nn_max_pool2d(kernel_size = 3, stride = 2)
    self$conv2d_3b <- BasicConv2d(64, 80, kernel_size = 1, stride = 1)
    self$conv2d_4a <- BasicConv2d(80, 192, kernel_size = 3, stride = 1)
    self$conv2d_4b <- BasicConv2d(192, 256, kernel_size = 3, stride = 2)

    self$mixed_6a <- Mixed_6a()
    self$repeat_2 <- nn_sequential(!!!lapply(1:10, function(i) Block17(0.10)))
    self$mixed_7a <- Mixed_7a()
    self$repeat_3 <- nn_sequential(!!!lapply(1:5, function(i) Block8(0.20)))
    self$block8 <- Block8(noReLU = TRUE)

    self$avgpool_1a <- nn_adaptive_avg_pool2d(output_size = 1)
    self$dropout <- nn_dropout(p = dropout_prob)
    self$last_linear <- nn_linear(1792, 512, bias = FALSE)
    self$last_bn <- nn_batch_norm1d(512, eps = 0.001, momentum = 0.1, affine = TRUE)

    self$classify <- classify
    if (!is.null(pretrained)) {
      self$logits <- nn_linear(512, tmp_classes)
      load_inception_weights(self, pretrained)
    }
    if (classify && !is.null(num_classes)) {
      self$logits <- nn_linear(512, num_classes)
    }
  },
  forward = function(x) {
    x <- self$conv2d_1a(x)
    x <- self$conv2d_2a(x)
    x <- self$conv2d_2b(x)
    x <- self$maxpool_3a(x)
    x <- self$conv2d_3b(x)
    x <- self$conv2d_4a(x)
    x <- self$conv2d_4b(x)
    x <- self$mixed_6a(x)
    x <- self$repeat_2(x)
    x <- self$mixed_7a(x)
    x <- self$repeat_3(x)
    x <- self$block8(x)
    x <- self$avgpool_1a(x)
    x <- self$dropout(x)
    x <- self$last_linear(x$view(c(x$shape[1], -1)))
    x <- self$last_bn(x)

    if (self$classify) {
      x <- self$logits(x)
    } else {
      x <- nnf_normalize(x, p = 2, dim = 2)
    }
    x
  }
)

#' Oxford Flowers 102 Dataset
#'
#' Loads the Oxford 102 Category Flower Dataset. This dataset consists of 102 flower categories,
#' with between 40 and 258 images per class. Images in this dataset are of variable sizes.
#'
#' This is a **classification** dataset where the goal is to assign each image to one of the 102 flower categories.
#'
#' The dataset is split into:
#' - `"train"`: training subset with labels.
#' - `"val"`: validation subset with labels.
#' - `"test"`: test subset with labels (used for evaluation).
#'
#' @inheritParams fgvc_aircraft_dataset
#' @param root Root directory for dataset storage. The dataset will be stored under `root/flowers102`.
#' @param split One of `"train"`, `"val"`, or `"test"`. Default is `"train"`.
#'
#' @return An object of class \code{flowers102_dataset}, which behaves like a torch dataset.
#' Each element is a named list:
#' - `x`: a W x H x 3 numeric array representing an RGB image.
#' - `y`: an integer label indicating the class index.
#'
#' @examples
#' \dontrun{
#' # Load the dataset with inline transforms
#' flowers <- flowers102_dataset(
#'   split = "train",
#'   download = TRUE,
#'   transform = . %>% transform_to_tensor() %>% transform_resize(c(224, 224))
#' )
#'
#' # Create a dataloader
#' dl <- dataloader(
#'   dataset = flowers,
#'   batch_size = 4
#' )
#'
#' # Access a batch
#' batch <- dataloader_next(dataloader_make_iter(dl))
#' batch$x  # Tensor of shape (4, 3, 224, 224)
#' batch$y  # Tensor of shape (4,) with numeric class labels
#' }
#'
#' @family classification_dataset
#' @export
flowers102_dataset <- dataset(
  name = "flowers102",
  archive_size = "330 MB",
  classes = c(
    "pink primrose", "hard-leaved pocket orchid", "canterbury bells", "sweet pea", "english marigold",
    "tiger lily", "moon orchid", "bird of paradise", "monkshood", "globe thistle",
    "snapdragon", "colt's foot", "king protea", "spear thistle", "yellow iris",
    "globe-flower", "purple coneflower", "peruvian lily", "balloon flower", "giant white arum lily",
    "fire lily", "pincushion flower", "fritillary", "red ginger", "grape hyacinth",
    "corn poppy", "prince of wales feathers", "stemless gentian", "artichoke", "sweet william",
    "carnation", "garden phlox", "love in the mist", "mexican aster", "alpine sea holly",
    "ruby-lipped cattleya", "cape flower", "great masterwort", "siam tulip", "lenten rose",
    "barbeton daisy", "daffodil", "sword lily", "poinsettia", "bolero deep blue",
    "wallflower", "marigold", "buttercup", "oxeye daisy", "common dandelion",
    "petunia", "wild pansy", "primula", "sunflower", "pelargonium",
    "bishop of llandaff", "gaura", "geranium", "orange dahlia", "pink-yellow dahlia?",
    "cautleya spicata", "japanese anemone", "black-eyed susan", "silverbush", "californian poppy",
    "osteospermum", "spring crocus", "bearded iris", "windflower", "tree poppy",
    "gazania", "azalea", "water lily", "rose", "thorn apple",
    "morning glory", "passion flower", "lotus", "toad lily", "anthurium",
    "frangipani", "clematis", "hibiscus", "columbine", "desert-rose",
    "tree mallow", "magnolia", "cyclamen", "watercress", "canna lily",
    "hippeastrum", "bee balm", "ball moss", "foxglove", "bougainvillea",
    "camellia", "mallow", "mexican petunia", "bromelia", "blanket flower",
    "trumpet creeper", "blackberry lily"
  ),
  resources = list(
    c("https://www.robots.ox.ac.uk/~vgg/data/flowers/102/102flowers.tgz", "52808999861908f626f3c1f4e79d11fa"),
    c("https://www.robots.ox.ac.uk/~vgg/data/flowers/102/imagelabels.mat", "e0620be6f572b9609742df49c70aed4d"),
    c("https://www.robots.ox.ac.uk/~vgg/data/flowers/102/setid.mat", "a5357ecc9cb78c4bef273ce3793fc85c")
  ),

  initialize = function(
    root = tempdir(),
    split = "train",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root_path <- root
    self$split <- match.arg(split, c("train", "val", "test"))
    self$transform <- transform
    self$target_transform <- target_transform
    self$classes <- self$classes

    if (download){
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists(self$split))
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")

    meta <- readRDS(file.path(self$processed_folder, glue::glue("{self$split}.rds")))
    self$img_path <- meta$img_path
    self$labels <- meta$labels
    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  .getitem = function(index) {
    img_path <- self$img_path[[index]]
    y <- self$labels[[index]]

    x <- jpeg::readJPEG(img_path)

    if (!is.null(self$transform))
      x <- self$transform(x)

    if (!is.null(self$target_transform))
      y <- self$target_transform(y)

    list(x = x, y = y)
  },

  .length = function() {
    length(self$img_path)
  },

  download = function() {
    if (self$check_exists(self$split)) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} is already processed and cached.")
    }
    fs::dir_create(self$raw_folder)
    fs::dir_create(self$processed_folder)

    cli_inform("Downloading {.cls {class(self)[[1]]}} ...")

    archives <- lapply(self$resources, function(r) {
      archive <- download_and_cache(r[1], prefix = class(self)[1])
      if (!tools::md5sum(archive) == r[2])
        cli_abort("Corrupt file! Delete the file in {archive} and try again.")
      archive
    })

    cli_inform("{.cls {class(self)[[1]]}} Extracting images and processing dataset...")

    untar(archives[[1]], exdir = self$raw_folder)
    labels <- R.matlab::readMat(archives[[2]])$labels
    setids <- R.matlab::readMat(archives[[3]])

    set_map <- list(
      train = as.integer(setids$trnid),
      val = as.integer(setids$valid),
      test = as.integer(setids$tstid)
    )

    split_name <- self$split
    idxs <- set_map[[split_name]]
    jpg_dir <- file.path(self$raw_folder, "jpg")
    paths <- file.path(jpg_dir, glue::glue("image_{sprintf('%05d', idxs)}.jpg"))
    lbls <- as.integer(labels[idxs])
    saveRDS(data.frame(img_path = paths, labels = lbls), file.path(self$processed_folder, glue::glue("{split_name}.rds")))
    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  check_exists = function(split) {
    fs::file_exists(file.path(self$processed_folder, glue::glue("{split}.rds")))
  },

  active = list(
    raw_folder = function() file.path(self$root_path, class(self)[1], "raw"),
    processed_folder = function() file.path(self$root_path, class(self)[1], "processed")
  )
)

#' @export
transform_to_tensor.default <- function(img) {
  not_implemented_for_class(img)
}

#' @export
transform_convert_image_dtype.default <- function(img, dtype) {
  not_implemented_for_class(img)
}


#' @export
transform_normalize.default <- function(img, mean, std, inplace = FALSE) {
  not_implemented_for_class(img)
}

#' @export
transform_resize.default <- function(img, size, interpolation) {
  not_implemented_for_class(img)
}

#' @export
transform_center_crop.default <- function(img, size) {

  output_size <- size

  if (length(size) == 1)
    output_size <- rep(size, 2)

  output_size <- as.integer(output_size)

  size <- get_image_size(img)

  image_height <- size[2]
  image_width <- size[1]

  crop_height <- output_size[1]
  crop_width <- output_size[2]

  if (crop_width > image_width || crop_height > image_height) {

    padding_ltrb <- c(
      if (crop_width > image_width) (crop_width - image_width) %/% 2  else 0,
      if (crop_width > image_width) (crop_width - image_width + 1) %/% 2  else 0,
      if (crop_height > image_height) (crop_height - image_height) %/% 2  else 0,
      if (crop_height > image_height) (crop_height - image_height + 1) %/% 2  else 0
    )

    img <- transform_pad(img, padding_ltrb, fill = 0)  # PIL uses fill value 0

    size <- get_image_size(img)
    image_height <- size[1]
    image_width <- size[2]

    if (crop_width == image_width && crop_height == image_height) return(img)
  }

  crop_top <- as.integer((image_height - crop_height) / 2)
  crop_left <- as.integer((image_width - crop_width) / 2)

  # if either of these is 0, we will lose a pixel in transform_crop
  if (crop_top == 0) crop_top <- 1
  if (crop_left == 0) crop_left <- 1

  transform_crop(img, crop_top, crop_left, crop_height, crop_width)
}

#' @export
transform_pad.default <- function(img, padding, fill = 0, padding_mode = "constant") {
  not_implemented_for_class(img)
}

#' @export
transform_random_apply.default <- function(img, transforms, p = 0.5) {

  if (p < stats::runif(1))
    return(img)

  for (tf in transforms) {
    img <- tf(img)
  }

  img
}

#' @export
transform_random_choice.default <- function(img, transforms) {
  i <- sample.int(length(transforms), 1)
  transforms[[i]](img)
}

#' @export
transform_random_order.default <- function(img, transforms) {
  i <- sample.int(length(transforms), size = length(transforms))
  transforms <- transforms[i]
  for (tf in transforms) {
    img <- tf(img)
  }
  img
}

get_random_crop_params <- function(img, output_size) {

  img_size <- get_image_size(img)
  w <- img_size[1]; h <- img_size[2]
  th <- output_size[1]; tw <- output_size[2]

  if (w == tw && h == th)
    return(c(0, 0, h, w))

  i <- as.integer(torch::torch_randint(1, h - th + 1, size=1))
  j <- as.integer(torch::torch_randint(1, w - tw + 1, size=1))

  c(i, j, th, tw)
}

#' @export
transform_random_crop.default <- function(img, size, padding=NULL, pad_if_needed=FALSE,
                                          fill=0, padding_mode="constant") {


  if (length(size) == 1)
    size <- c(size, size)

  size <- as.integer(size)


  if (!is.null(padding))
    img <- transform_pad(img, padding, fill, padding_mode)

  img_size <- get_image_size(img)
  width <- img_size[1]; height <- img_size[2]

  # pad the width if needed
  if (pad_if_needed && width < size[2]) {
    padding <- c(size[2] - width, 0)
    img <- transform_pad(img, padding, fill, padding_mode)
  }

  # pad the height if needed
  if (pad_if_needed && height < size[1]) {
    padding <- c(0, size[1] - height)
    img <- transform_pad(img, padding, fill, padding_mode)
  }

  params <- get_random_crop_params(img, size)

  transform_crop(img, params[1], params[2], params[3], params[4])
}

#' @export
transform_random_horizontal_flip.default <- function(img, p = 0.5) {

  if (stats::runif(1) < p)
    img <- transform_hflip(img)

  img
}


#' @export
transform_random_vertical_flip.default <- function(img, p = 0.5) {

  if (stats::runif(1) < p)
    img <- transform_vflip(img)

  img

}

get_random_resized_crop_params <- function(img, scale, ratio) {

  img_size <- get_image_size(img)
  width <- img_size[1]; height <- img_size[2]

  area <- height * width

  for (i in 1:10) {
    target_area <- as.numeric(area * torch::torch_empty(1)$uniform_(scale[1], scale[2]))
    log_ratio <- torch::torch_log(torch::torch_tensor(ratio))
    aspect_ratio <-  as.numeric(torch::torch_exp(
      torch::torch_empty(1)$uniform_(as.numeric(log_ratio[1]), as.numeric(log_ratio[2]))
    ))

    w <- as.integer(round(sqrt(target_area * aspect_ratio)))
    h <- as.integer(round(sqrt(target_area / aspect_ratio)))

    if (1 < w && w <= width && 1 < h && h <= height) {

      i <- sample.int(height - h + 1, size=1)
      j <- sample.int(width - w + 1, size=1)

      return(c(i, j, h, w))
    }

  }

  # Fallback to central crop
  in_ratio <- width / height
  if (in_ratio < min(ratio)) {
    w <- width
    h <- as.integer(round(w / min(ratio)))
  } else if (in_ratio > max(ratio)) {
    h <-  height
    w <- as.integer(round(h * max(ratio)))
  } else {
    w <- width
    h <- height
  }
  i <- (height - h) %/% 2
  j <- (width - w) %/% 2

  c(i, j, h, w)
}

#' @export
transform_random_resized_crop.default <- function(img, size, scale=c(0.08, 1.0),
                                                  ratio=c(3. / 4., 4. / 3.),
                                                  interpolation=2) {


  params <- get_random_resized_crop_params(img, scale, ratio)

  transform_resized_crop(img, params[1], params[2], params[3], params[4], size,
                        interpolation)
}

#' @export
transform_five_crop.default <- function(img, size) {
  not_implemented_for_class(img)
}

#' @export
transform_ten_crop.default <- function(img, size, vertical_flip = FALSE) {
  not_implemented_for_class(img)
}

#' @export
transform_linear_transformation.default <- function(img, transformation_matrix,
                                                    mean_vector) {
  not_implemented_for_class(img)
}

get_color_jitter_params <- function(brightness, contrast, saturation, hue) {

  transforms <- list()

  if (!is.null(brightness)) {
    brightness_factor <- stats::runif(1, min = brightness[1], max = brightness[2])
    transforms <- append(
      transforms,
      list(function(img) transform_adjust_brightness(img, brightness_factor))
    )
  }

  if (!is.null(contrast)) {
    contrast_factor <- stats::runif(1, contrast[1], contrast[2])
    transforms <- append(
      transforms,
      list(function(img) transform_adjust_contrast(img, contrast_factor))
    )
  }

  if (!is.null(saturation)) {
    saturation_factor <- stats::runif(1, saturation[1], saturation[2])
    transforms <- append(
      transforms,
      list(function(img) transform_adjust_saturation(img, saturation_factor))
    )
  }

  if (!is.null(hue)) {
    hue_factor <- stats::runif(1, hue[1], hue[2])
    transforms <- append(
      transforms,
      list(function(img) transform_adjust_hue(img, hue_factor))
    )
  }

  # shuffle
  i <- sample.int(length(transforms), length(transforms))
  transforms <- transforms[i]

  function(img) {
    for (tf in transforms) {
      img <- tf(img)
    }
    img
  }
}

check_color_jitter_input <- function(value, center = 1, bound = c(0, Inf),
                                     clip_first_on_zero = TRUE) {

  if (length(value) == 1) {

    if (value < 0)
      value_error("must be positive if a single number")

    value <- c(center - value, center + value)

    if (clip_first_on_zero)
      value[1] <- max(value[1], 0.0)

  } else if (length(value) == 2) {

    if (value[1] < bound[1] || value[2] > bound[2])
      value_error("out of bounds.")

  }

  # if value is 0 or (1., 1.) for brightness/contrast/saturation
  # or (0., 0.) for hue, do nothing
  if (value[1] == value[2] && value[2] == center) {
    value <- NULL
  }

  value
}

#' @export
transform_color_jitter.default <- function(img, brightness=0, contrast=0,
                                           saturation=0, hue=0) {

  brightness <- check_color_jitter_input(brightness)
  contrast <- check_color_jitter_input(contrast)
  saturation <- check_color_jitter_input(saturation)
  hue <- check_color_jitter_input(hue, center=0, bound=c(-0.5, 0.5),
                               clip_first_on_zero=FALSE)

  tf <- get_color_jitter_params(brightness, contrast, saturation, hue)
  tf(img)
}

get_random_rotation_params <- function(degrees) {
  as.numeric(torch::torch_empty(1)$uniform_(degrees[1], degrees[2]))
}

#' @export
transform_random_rotation.default <- function(img, degrees, interpolation=0,
                                              expand=FALSE, center=NULL, fill=NULL, resample) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }

  if (length(degrees) == 1) {

    if (degrees < 0)
      value_error("degrees must be positive if it's a single value")

    degrees <- c(-degrees, degrees)

  } else if (length(degrees) != 2) {
    value_error("degrees must be length 1 or 2")
  }

  angle <- get_random_rotation_params(degrees)
  transform_rotate(img, angle, interpolation, expand, center, fill)
}


get_random_affine_params <- function(degrees,
                                     translate,
                                     scale_ranges,
                                     shears,
                                     img_size) {

  angle <- as.numeric(torch::torch_empty(1)$uniform_(degrees[1], degrees[2]))
  if (!is.null(translate)) {
    max_dx <- as.numeric(translate[1] * img_size[1])
    max_dy <- as.numeric(translate[2] * img_size[2])
    tx <- as.integer(round(as.numeric(torch::torch_empty(1)$uniform_(-max_dx, max_dx))))
    ty <- as.integer(round(as.numeric(torch::torch_empty(1)$uniform_(-max_dy, max_dy))))
    translations <- c(tx, ty)
  } else {
    translations <- c(0,0)
  }

  if (!is.null(scale_ranges)) {
    scale <- as.numeric(torch::torch_empty(1)$uniform_(scale_ranges[1], scale_ranges[2]))
  } else {
    scale <- 1
  }

  shear_x <- shear_y <- 0.0

  if (!is.null(shears))  {
    shear_x <- as.numeric(torch::torch_empty(1)$uniform_(shears[1], shears[2]))
    if (length(shears) == 4)
      shear_y <- as.numeric(torch::torch_empty(1)$uniform_(shears[3], shears[4]))
  }

  shear <- c(shear_x, shear_y)

  list(angle, translations, scale, shear)
}

#' @export
transform_random_affine.default <- function(img, degrees, translate=NULL, scale=NULL,
                                            shear=NULL, interpolation=0, fill=0,
                                            resample, fillcolor) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }
  if (!missing(fillcolor)) {
    deprecated("'fillcolor' is deprecated, use 'fill' instead")
    fill <- fillcolor
  }

  if (length(degrees) == 1) {

    if (degrees < 0)
      value_error("degrees must be positive if it's a single value")

    degrees <- c(-degrees, degrees)

  } else if (length(degrees) != 2) {
    value_error("degrees must be length 1 or 2")
  }


  if (!is.null(translate)) {

    if (length(translate) != 2)
      value_error("translate must be length 2")

    if (any(translate > 1) || any(translate < 0))
      value_error("translate must be between 0 and 1")

  }

  if (!is.null(scale)) {

    if (length(scale) != 2)
      value_error("scale must be length 2")

    if (any(scale < 0))
      value_error("scale must be positive")

  }

  if (!is.null(shear)) {

    if (length(shear) == 1) {

      if (shear < 0)
        value_error("shear must be positive if it's a single value")

      degrees <- c(-degrees, degrees)

    } else if (!length(shear) %in% c(2, 4)) {
      value_error("shear's length must be 1, 2, or 4")
    }

  }


  img_size <- get_image_size(img)

  ret <- get_random_affine_params(degrees, translate, scale, shear, img_size)

  transform_affine(img, ret[[1]], ret[[2]], ret[[3]], ret[[4]],
                   interpolation=interpolation, fill=fill)
}

#' @export
transform_grayscale.default <- function(img, num_output_channels) {
  not_implemented_for_class(img)
}

#' @export
transform_random_grayscale.default <- function(img, p = 0.1) {
  not_implemented_for_class(img)
}

get_random_perspective_params <- function(width, height, distortion_scale) {
  half_height <- height %/% 2
  half_width <- width %/% 2
  topleft <- c(
    as.integer(torch::torch_randint(1 + 0,as.integer(distortion_scale * half_width) + 1, size=1)),
    as.integer(torch::torch_randint(1 + 0,as.integer(distortion_scale * half_height) + 1, size=1))
  )
  topright <- c(
    as.integer(torch::torch_randint(1 + width -as.integer(distortion_scale * half_width) - 1, width, size=1)),
    as.integer(torch::torch_randint(1 + 0,as.integer(distortion_scale * half_height) + 1, size=1))
  )
  botright <- c(
    as.integer(torch::torch_randint(1 + width -as.integer(distortion_scale * half_width) - 1, width, size=1)),
    as.integer(torch::torch_randint(1 + height -as.integer(distortion_scale * half_height) - 1, height, size=1))
  )
  botleft <- c(
    as.integer(torch::torch_randint(1 + 0,as.integer(distortion_scale * half_width) + 1, size=1)),
    as.integer(torch::torch_randint(1 + height -as.integer(distortion_scale * half_height) - 1, height, size=1))
  )
  startpoints <- list(c(1 + 0, 1 + 0), c(1 + width - 1, 1 + 0), c(1 + width - 1, 1 + height - 1), c(1 + 0, 1 + height - 1))
  endpoints <- list(topleft, topright, botright, botleft)
  list(startpoints, endpoints)
}

#' @export
transform_random_perspective.default <- function(img, distortion_scale=0.5, p=0.5,
                                                 interpolation=2, fill=0) {
  if (stats::runif(1) < p) {
    img_size <- get_image_size(img)
    params <- get_random_perspective_params(img_size[1], img_size[2], distortion_scale)
    img <- transform_perspective(img, params[[1]], params[[2]], interpolation, fill)
  }

  img
}

#' @export
transform_random_erasing.default <- function(img, p=0.5, scale=c(0.02, 0.33),
                                             ratio=c(0.3, 3.3), value=0, inplace=FALSE) {
  not_implemented_for_class(img)
}

# Other methods -----------------------------------------------------------

#' @export
transform_crop.default <- function(img, top, left, height, width) {
  not_implemented_for_class(img)
}

#' @export
transform_hflip.default <- function(img) {
  not_implemented_for_class(img)
}

#' @export
transform_vflip.default <- function(img) {
  not_implemented_for_class(img)
}

#' @export
transform_resized_crop.default <- function(img, top, left, height, width, size,
                                   interpolation = 2) {
  img <- transform_crop(img, top, left, height, width)
  img <- transform_resize(img, size, interpolation)
  img
}

#' @export
transform_adjust_brightness.default <- function(img, brightness_factor) {
  not_implemented_for_class(img)
}

#' @export
transform_adjust_contrast.default <- function(img, contrast_factor) {
  not_implemented_for_class(img)
}

#' @export
transform_adjust_hue.default <- function(img, hue_factor) {
  not_implemented_for_class(img)
}

#' @export
transform_adjust_saturation.default <- function(img, saturation_factor) {
  not_implemented_for_class(img)
}

#' @export
transform_rotate.default <- function(img, angle, interpolation = 0, expand = FALSE,
                                     center = NULL, fill = NULL, resample) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }
  not_implemented_for_class(img)
}

#' @export
transform_affine.default <- function(img, angle, translate, scale, shear,
                                     interpolation = 0, fill = NULL,
                                     resample, fillcolor, center = NULL) {
  if (!missing(resample)) {
    deprecated("'resample' is deprecated, use 'interpolation' instead")
    interpolation <- resample
  }
  if (!missing(fillcolor)) {
    deprecated("'fillcolor' is deprecated, use 'fill' instead")
    fill <- fillcolor
  }
  not_implemented_for_class(img)
}

#' @export
transform_perspective.default <- function(img, startpoints, endpoints, interpolation = 2,
                                  fill = NULL) {
  not_implemented_for_class(img)
}

#' @export
transform_adjust_gamma.default <- function(img, gamma, gain = 1) {
  not_implemented_for_class(img)
}


# Helpers -----------------------------------------------------------------

not_implemented_for_class <- function(x) {
  not_implemented_error("not implemented for {class(x)}")
}

#' Oxford-IIIT Pet Segmentation Dataset
#'
#' The Oxford-IIIT Pet Dataset is a **segmentation** dataset consisting of color images
#' of 37 pet breeds (cats and dogs). Each image is annotated with a pixel-level
#' trimap segmentation mask, identifying pet, background, and outline regions.
#' It is commonly used for evaluating models on object segmentation tasks.
#'
#' @inheritParams mnist_dataset
#' @param root Character. Root directory where the dataset is stored or will be downloaded to. Files are placed under `root/oxfordiiitpet`.
#' @param target_type Character. One of \code{"category"} or \code{"binary-category"} (default: \code{"category"}).
#'
#' @return A torch dataset object \code{oxfordiiitpet_dataset}. Each item is a named list:
#' - \code{x}: a H x W x 3 integer array representing an RGB image.
#' - \code{y$masks}: a boolean tensor of shape (3, H, W), representing the segmentation trimap as one-hot masks.
#' - \code{y$label}: an integer representing the class label, depending on the \code{target_type}:
#'   - \code{"category"}: an integer in 1–37 indicating the pet breed.
#'   - \code{"binary-category"}: 1 for Cat, 2 for Dog.
#'
#' @examples
#' \dontrun{
#' # Load the Oxford-IIIT Pet dataset with basic tensor transform
#' oxfordiiitpet <- oxfordiiitpet_segmentation_dataset(
#'    transform = transform_to_tensor,
#'    download = TRUE
#' )
#'
#' # Retrieve the image tensor and label (trimap in raw format)
#' first_item <- oxfordiiitpet[1]
#' first_item$x  # RGB image tensor of shape (3, H, W)
#' first_item$y$trimap  # (H, W) integer tensor: 1=pet, 2=background, 3=outline
#' first_item$y$label  # Integer label (1–37 or 1–2 depending on target_type)
#' oxfordiiitpet$classes[first_item$y$label] # Class name of the label
#'
#' # Load dataset with explicit segmentation mask transformation
#' oxfordiiitpet_masked <- oxfordiiitpet_segmentation_dataset(
#'    transform = transform_to_tensor,
#'    target_transform = target_transform_trimap_masks,
#'    download = TRUE
#' )
#'
#' masked_item <- oxfordiiitpet_masked[1]
#' masked_item$y$masks  # (3, H, W) bool tensor: pet, background, outline
#'
#' # Visualize segmentation masks
#' overlay <- draw_segmentation_masks(masked_item)
#' tensor_image_browse(overlay)
#' }
#'
#' @family segmentation_dataset
#' @export
oxfordiiitpet_segmentation_dataset <- torch::dataset(
  name = "oxfordiiitpet segmentation",
  resources = list(
    c("https://www.robots.ox.ac.uk/~vgg/data/pets/data/images.tar.gz", "5c4f3ee8e5d25df40f4fd59a7f44e54c"),
    c("https://www.robots.ox.ac.uk/~vgg/data/pets/data/annotations.tar.gz", "95a8c909bbe2e81eed6a22bccdf3f68f")
  ),
  training_file = "trainval.rds",
  test_file = "test.rds",
  archive_size = "750 MB",
  initialize = function(
    root = tempdir(),
    train = TRUE,
    target_type = "category",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root_path <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$train <- train
    self$target_type <- target_type
    if (train) {
      self$split <- "train"
    } else {
      self$split <- "test"
    }

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }

    if (train) {
      data_file <- self$training_file
    } else {
      data_file <- self$test_file
    }

    data <- readRDS(file.path(self$processed_folder, data_file))
    self$img_path <- data$img_path
    self$labels <- data$labels
    self$class_to_idx <- data$class_to_idx
    if (self$target_type == "category") {
      self$classes <- names(self$class_to_idx)
    } else {
      self$classes <- c("Cat", "Dog")
    }

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {

    if (self$check_exists()) {
      return()
    }

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    for (r in self$resources) {
      url <- r[1]
      archive <- download_and_cache(url, prefix = class(self)[1])
      actual_md5 <- tools::md5sum(archive)

      if (actual_md5 != r[2]) {
        runtime_error("Corrupt file! Delete the file in {archive} and try again.")
      }

      utils::untar(archive, exdir = self$raw_folder)
    }

    for (split in c("trainval", "test")) {
      ann_file <- file.path(self$raw_folder, "annotations", glue::glue("{split}.txt"))

      ann <- read.delim(
        ann_file,
        sep = " ",
        header = FALSE,
        col.names = c("img_id", "label", "bin_label", "other")
      )

      img_paths <- file.path(self$raw_folder, "images", glue::glue("{ann$img_id}.jpg"))
      seg_paths <- file.path(self$raw_folder, "annotations", "trimaps", glue::glue("{ann$img_id}.png"))

      valid <- file.exists(img_paths) & file.exists(seg_paths)

      img_path <- img_paths[valid]
      labels <- ann$label[valid]

      raw_classes <- sub("_\\d+$", "", ann$img_id[valid])
      self$classes <- unique(raw_classes)
      self$classes <- gsub("_", " ", self$classes, fixed = TRUE)
      class_to_idx <- setNames(seq_along(self$classes), self$classes)

      saveRDS(
        list(
          img_path = img_path,
          labels = labels,
          class_to_idx = class_to_idx
        ),
        file.path(self$processed_folder, glue::glue("{split}.rds"))
      )
    }

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  check_exists = function() {
    fs::file_exists(file.path(self$processed_folder, self$training_file)) && fs::file_exists(file.path(self$processed_folder, self$test_file))
  },

  .getitem = function(index) {
    x <- jpeg::readJPEG(self$img_path[index])

    label <- self$labels[index]

    seg_name <- basename(self$img_path[index])
    mask_path <- file.path(self$raw_folder, "annotations", "trimaps", sub("\\.jpg$", ".png", seg_name))
    mask_int <- torch_tensor(png::readPNG(mask_path) * 255)
    masks <- torch_stack(list(mask_int == 1, mask_int == 2, mask_int == 3))

    if (self$target_type == "binary-category") {
      class_name <- names(self$class_to_idx)[label]
      if (substr(class_name, 1, 1) %in% LETTERS) {
        label <- 1L
      } else {
        label <- 2L
      }

      self$classes <- c("Cat","Dog")
    }

    y <- list(
      masks = masks,
      label = label
    )

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    result <- list(x = x, y = y)
    class(result) <- c("image_with_segmentation_mask", class(result))
    result
  },

  .length = function() {
    length(self$img_path)
  },

  active = list(
    raw_folder = function() {
      fs::dir_create(self$root_path, "oxfordiiitpet", "raw")
      file.path(self$root_path, "oxfordiiitpet", "raw")
    },
    processed_folder = function() {
      fs::dir_create(self$root_path, "oxfordiiitpet", "processed")
      file.path(self$root_path, "oxfordiiitpet", "processed")
    }
  )
)

#' Oxford-IIIT Pet Datasets
#'
#' The Oxford-IIIT Pet collection is a **classification** dataset consisting of high-quality
#' images of 37 cat and dog breeds. It includes two variants:
#' - `oxfordiiitpet_dataset`: Multi-class classification across 37 pet breeds.
#' - `oxfordiiitpet_binary_dataset`: Binary classification distinguishing cats vs dogs.
#'
#' @inheritParams oxfordiiitpet_segmentation_dataset
#'
#' @return A torch dataset object \code{oxfordiiitpet_dataset} or \code{oxfordiiitpet_binary_dataset}.
#' Each element is a named list with:
#' - `x`: A H x W x 3 integer array representing an RGB image.
#' - `y`: An integer label:
#'   - For `oxfordiiitpet_dataset`: a value from 1–37 representing the breed.
#'   - For `oxfordiiitpet_binary_dataset`: 1 for Cat, 2 for Dog.
#'
#' @details
#' The Oxford-IIIT Pet dataset contains over 7,000 images across 37 categories,
#' with roughly 200 images per class. Each image is labeled with its breed and species (cat/dog).
#'
#' @examples
#' \dontrun{
#' # Multi-class version
#' oxford <- oxfordiiitpet_dataset(download = TRUE)
#' first_item <- oxford[1]
#' first_item$x  # RGB image
#' first_item$y  # Label in 1–37
#' oxford$classes[first_item$y]  # Breed name
#'
#' # Binary version
#' oxford_bin <- oxfordiiitpet_binary_dataset(download = TRUE)
#' first_item <- oxford_bin[1]
#' first_item$x  # RGB image
#' first_item$y  # 1 for Cat, 2 for Dog
#' oxford_bin$classes[first_item$y]  # "Cat" or "Dog"
#' }
#'
#' @name oxfordiiitpet_dataset
#' @title Oxford-IIIT Pet Classification Datasets
#' @rdname oxfordiiitpet_dataset
#' @family classification_dataset
#' @export
oxfordiiitpet_dataset <- dataset(
  inherit = oxfordiiitpet_segmentation_dataset,
  name = "oxfordiiitpet",

  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root_path <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$train <- train
    if (train) {
      self$split <- "train"
    } else {
      self$split <- "test"
    }

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }

    if (train) {
      data_file <- self$training_file
    } else {
      data_file <- self$test_file
    }
    data <- readRDS(file.path(self$processed_folder, data_file))
    self$img_path <- data$img_path
    self$labels <- data$labels
    self$class_to_idx <- data$class_to_idx
    self$classes <- names(self$class_to_idx)

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  .getitem = function(index) {
    x <- jpeg::readJPEG(self$img_path[index])

    y <- self$labels[index]

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    list(x = x, y = y)
  }
)

#' @rdname oxfordiiitpet_dataset
#' @export
oxfordiiitpet_binary_dataset <- dataset(
  inherit = oxfordiiitpet_segmentation_dataset,
  name = "oxfordiiitpet binary",

  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root_path <- root
    self$transform <- transform
    self$target_transform <- target_transform
    self$train <- train
    if (train) {
      self$split <- "train"
    } else {
      self$split <- "test"
    }

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }

    if (train) {
      data_file <- self$training_file
    } else {
      data_file <- self$test_file
    }
    data <- readRDS(file.path(self$processed_folder, data_file))
    self$img_path <- data$img_path
    self$labels <- data$labels
    self$class_to_idx <- data$class_to_idx
    self$classes <- c("Cat", "Dog")

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  .getitem = function(index) {
    x <- jpeg::readJPEG(self$img_path[index])

    label <- self$labels[index]
    class_name <- names(self$class_to_idx)[label]
    if (substr(class_name, 1, 1) %in% LETTERS) {
      y <- 1L
    } else {
      y <- 2L
    }

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    list(x = x, y = y)
  }
)

#' FER-2013 Facial Expression Dataset
#'
#' Loads the FER-2013 dataset for facial expression recognition. The dataset contains grayscale images
#' (48x48) of human faces, each labeled with one of seven emotion categories:
#' `"Angry"`, `"Disgust"`, `"Fear"`, `"Happy"`, `"Sad"`, `"Surprise"`, and `"Neutral"`.
#'
#' The dataset is split into:
#' - `"Train"`: training images labeled as `"Training"` in the original CSV.
#' - `"Test"`: includes both `"PublicTest"` and `"PrivateTest"` entries.
#'
#' @inheritParams mnist_dataset
#' @param root (string, optional): Root directory for dataset storage,
#'   the dataset will be stored under `root/fer2013`.
#'
#' @return A torch dataset of class \code{fer_dataset}.
#' Each element is a named list:
#' - `x`: a 48x48 grayscale array
#' - `y`: an integer from 1 to 7 indicating the class index
#'
#' @examples
#' \dontrun{
#' fer <- fer_dataset(train = TRUE, download = TRUE)
#' first_item <- fer[1]
#' first_item$x  # 48x48 grayscale array
#' first_item$y  # 4
#' fer$classes[first_item$y]  # "Happy"
#' }
#'
#' @family classification_dataset
#' @export
fer_dataset <- dataset(
  name = "fer_dataset",
  archive_size = "90 MB",
  url = "https://huggingface.co/datasets/DerrickUnleashed/FER-2013/resolve/main/fer2013.tar.gz",
  md5 = "ca95d94fe42f6ce65aaae694d18c628a",
  classes = c(
    "Angry",
    "Disgust",
    "Fear",
    "Happy",
    "Sad",
    "Surprise",
    "Neutral"
  ),

  initialize = function(
    root = tempdir(),
    train = TRUE,
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
    self$root <- root
    self$train <- train
    self$transform <- transform
    self$target_transform <- target_transform
    self$split <- ifelse(train, "Train", "Test")
    self$folder_name <- "fer2013"
    self$class_to_idx <- setNames(seq_along(self$classes), self$classes)

    if (download) {
      cli_inform(
        "Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available."
      )
      self$download()
    }

    if (!self$check_files()) {
      runtime_error(
        "Dataset not found. You can use `download = TRUE` to download it."
      )
    }

    csv_file <- file.path(self$root, self$folder_name, "fer2013.csv")
    parsed <- read.csv(csv_file, stringsAsFactors = FALSE)

    if (self$train) {
      parsed <- parsed[parsed$Usage == "Training", ]
    } else {
      parsed <- parsed[parsed$Usage %in% c("PublicTest", "PrivateTest"), ]
    }

    cli_inform("Processing {.cls {class(self)[[1]]}} ...")
    self$x <- lapply(strsplit(parsed$pixels, " "), as.integer)

    self$y <- parsed$emotion + 1L

    file_size <- fs::file_info(csv_file)$size
    readable <- fs::fs_bytes(file_size)

    cli_inform("Processing {.cls {class(self)[[1]]}} ...")
  },

  .getitem = function(i) {
    raw_vec <- self$x[[i]]
    x <- matrix(raw_vec, nrow = 48, ncol = 48, byrow = TRUE)
    x <- array(x, dim = c(48, 48))

    y <- self$y[i]

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }

    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    list(x = x, y = y)
  },

  .length = function() {
    length(self$y)
  },

  download = function() {
    if (self$check_files()) {
      return()
    }

    dest_dir <- file.path(self$root, self$folder_name)
    fs::dir_create(dest_dir)

    cli_inform("Downloading {.cls {class(self)[[1]]}} ...")

    archive <- download_and_cache(self$url)

    if (!tools::md5sum(archive) == self$md5) {
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")
    }

    untar(archive, exdir = self$root)
    cli_inform(
      "Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully."
    )
  },

  check_files = function() {
    file.exists(file.path(self$root, self$folder_name, "fer2013.csv"))
  }
)

#' @include collection-rf100-doc.R
NULL

#' RoboFlow 100  Biology dataset Collection
#'
#' Loads one of the RoboFlow 100 Biology datasets with
#' bounding box annotations for object-detection task.
#'
#' @inheritParams rf100_document_collection
#' @param dataset Dataset to select within \code{c("stomata_cell", "blood_cell", "parasite", "cell",
#' "bacteria", "cotton_disease","mitosis", "phage", "liver_disease", "moth")}.
#' @inherit rf100_document_collection return
#'
#' @examples
#' \dontrun{
#' ds <- rf100_biology_collection(
#'   dataset = "stomata_cell",
#'   split = "test",
#'   transform = transform_to_tensor,
#'   download = TRUE
#' )
#' item <- ds[1]
#' boxed <- draw_bounding_boxes(item)
#' tensor_image_browse(boxed)
#' }
#'
#' @family detection_dataset
#' @export
rf100_biology_collection <- torch::dataset(
  name = "rf100_biology_collection",
  inherit = rf100_document_collection,

  resources = data.frame(
    dataset = rep(c(
      "stomata_cell", "blood_cell", "parasite", "cell",
      "bacteria", "cotton_disease",
      "mitosis", "phage", "liver_disease", "moth"
    ),each = 3),
    split   = rep(c("train", "test", "valid"), times = 10),
    url = c(
      # stomata_cell
      "https://huggingface.co/datasets/Francesco/stomata-cells/resolve/main/data/train-00000-of-00001-f3d24fcd68c928f1.parquet",
      "https://huggingface.co/datasets/Francesco/stomata-cells/resolve/main/data/test-00000-of-00001-32bdd663798cbfdc.parquet",
      "https://huggingface.co/datasets/Francesco/stomata-cells/resolve/main/data/validation-00000-of-00001-e91f7b8ad813e041.parquet",
      # blood_cell
      "https://huggingface.co/datasets/Francesco/bccd-ouzjz/resolve/main/data/train-00000-of-00001-f7bd2ef2fd6e29bf.parquet",
      "https://huggingface.co/datasets/Francesco/bccd-ouzjz/resolve/main/data/test-00000-of-00001-8ea65561716eb43e.parquet",
      "https://huggingface.co/datasets/Francesco/bccd-ouzjz/resolve/main/data/validation-00000-of-00001-135e61fc2bee97fd.parquet",
      # parasite
      "https://huggingface.co/datasets/Francesco/parasites-1s07h/resolve/main/data/train-00000-of-00001-7e2c85bbcc1a45a2.parquet",
      "https://huggingface.co/datasets/Francesco/parasites-1s07h/resolve/main/data/test-00000-of-00001-829beee109b299d9.parquet",
      "https://huggingface.co/datasets/Francesco/parasites-1s07h/resolve/main/data/validation-00000-of-00001-e3aee1a46235b438.parquet",
      # cell
      "https://huggingface.co/datasets/Francesco/cells-uyemf/resolve/main/data/train-00000-of-00001-ddb451e11ab01b6e.parquet",
      "https://huggingface.co/datasets/Francesco/cells-uyemf/resolve/main/data/test-00000-of-00001-93af1cfd45e4b7aa.parquet",
      "https://huggingface.co/datasets/Francesco/cells-uyemf/resolve/main/data/validation-00000-of-00001-948f71851dc45fa4.parquet",
      # bacteria
      "https://huggingface.co/datasets/Francesco/bacteria-ptywi/resolve/main/data/train-00000-of-00001-4874c525c9b5291f.parquet",
      "https://huggingface.co/datasets/Francesco/bacteria-ptywi/resolve/main/data/test-00000-of-00001-515eb8a6a2a9bb07.parquet",
      "https://huggingface.co/datasets/Francesco/bacteria-ptywi/resolve/main/data/validation-00000-of-00001-2d3417d16e44ab71.parquet",
      # cotton_disease
      "https://huggingface.co/datasets/Francesco/cotton-plant-disease/resolve/main/data/train-00000-of-00001-fb83220158d0bab1.parquet",
      "https://huggingface.co/datasets/Francesco/cotton-plant-disease/resolve/main/data/test-00000-of-00001-cb05eef9488873d3.parquet",
      "https://huggingface.co/datasets/Francesco/cotton-plant-disease/resolve/main/data/validation-00000-of-00001-d9b9a7655deebe71.parquet",
      # mitosis
      "https://huggingface.co/datasets/Francesco/mitosis-gjs3g/resolve/main/data/train-00000-of-00001-97a883514c06adf1.parquet",
      "https://huggingface.co/datasets/Francesco/mitosis-gjs3g/resolve/main/data/test-00000-of-00001-e5d76ced0a07539a.parquet",
      "https://huggingface.co/datasets/Francesco/mitosis-gjs3g/resolve/main/data/validation-00000-of-00001-54fb9dda67697025.parquet",
      # phage
      "https://huggingface.co/datasets/Francesco/phages/resolve/main/data/train-00000-of-00001-b9eccf55a37a9c14.parquet",
      "https://huggingface.co/datasets/Francesco/phages/resolve/main/data/test-00000-of-00001-4fee5dee9a0c8f17.parquet",
      "https://huggingface.co/datasets/Francesco/phages/resolve/main/data/validation-00000-of-00001-5d783205edb8aef3.parquet",
      # liver_disease
      "https://huggingface.co/datasets/Francesco/liver-disease/resolve/main/data/train-00000-of-00001-075b34404316815c.parquet",
      "https://huggingface.co/datasets/Francesco/liver-disease/resolve/main/data/test-00000-of-00001-d6b1dd29852bde4e.parquet",
      "https://huggingface.co/datasets/Francesco/liver-disease/resolve/main/data/validation-00000-of-00001-ba8e36e9bd143c60.parquet",
      # moth
      "https://huggingface.co/datasets/Francesco/pests-2xlvx/resolve/main/data/train-00000-of-00001-a5be297b7b534376.parquet",
      "https://huggingface.co/datasets/Francesco/pests-2xlvx/resolve/main/data/test-00000-of-00001-5c1caa13c9b39012.parquet",
      "https://huggingface.co/datasets/Francesco/pests-2xlvx/resolve/main/data/validation-00000-of-00001-290a53c18d2f7a52.parquet"
    ),
    md5 = c(
      # stomata_cell
      "ba27c9cee8476a814a2869a622b9c2e3",      "95e3713caf397aa4182c2ed72651032a",      "c7c50c24fb1c11abc8da96eda21d8055",
      # blood_cell
      "40a57c143f70aa7b9898011669d5c500",      "601b0b184be5ff7b134edae5619ba852",      "4429c295861719418d038819241b40b9",
      # parasite
      "83d03d6eeee66f9b0bb32e59b9168a2f",      "c25500e4e96d60d5f6ae9ec4c40cb479",      "d74f6367de2cf8c9ef7c1d2199881362",
      # cell
      "6b23b8be679a0eff434147adec24c30a",      "24340cf1332e32241ca453f7277e4b43",      "2d9d19a6f8280bc72a7377f94a0d745d",
      # bacteria
      "c6641ee845c254e398d746c4286ae9af",      "7468dc6ae120a82d2f750a0bd051d338",      "ed9d0b920cb96cfa707b8f648675c07f",
      # cotton_disease
      "475743a07c9ffb2ad106c112f83fd110",      "8773a3dba89c3bf86879d12c1bb93007",      "45082d626c19fd3e5c3d3d955ae83f7b",
      # mitosis
      "6d7a45051cbb8bda5203004e04bb6639",      "64c4e2470e47556d3736dd2279f6457e",      "9f915191ff92498d1c9250bd1994afdf",
      # phage
      "c7622fed8a37b697b9dd1c3ab0d0708f",      "4623461f1526e50c6badae08e28b1692",      "9610c62631166e65636d25d541b6911b",
      # liver_disease
      "9b5839ae524277eb1702a0db33030e9a",      "c97b80332d710b1378a9f3ceb5abc197",      "2f255ba7ed3c1ce0948d2fc06305ce54",
      # moth
      "53c4c82727cd36a3093be388f27c3b3e",      "79a9fb92f219ce24412cbb2836e79e0f",      "2d8cd490d91bb9f25c17ff2ac332c010"
    ),
    size = c(81, 24, 12, 6.4, 1.8, 0.9 , 65.1,17.9,9,  0.3, 0.1, 0.05,
             1.4, 2.5, 0.8,   62, 16.8, 9,
             19, 5.3, 2.7,   69,9.0, 5.7,   192,55.6, 28,
             45, 13, 4.5) * 1e6
  )
)

#' VGGFace2 Dataset
#'
#' The VGGFace2 dataset is a large-scale face recognition dataset containing images
#' of celebrities from a wide range of ethnicities, professions, and ages.
#' Each identity has multiple images with variations in context, pose, age, and illumination.
#'
#' @inheritParams eurosat_dataset
#' @param root Character. Root directory where the dataset will be stored under `root/vggface2`.
#'
#' @return A torch dataset object `vggface2_dataset`:
#' - `x`: RGB image array.
#' - `y`: Integer label (1…N) for the identity.
#'
#' `ds$classes` is a named list mapping integer labels to a list with:
#' - `name`: Character name of the person.
#' - `gender`: "Male" or "Female".
#'
#' @examples
#' \dontrun{
#' #Load the training set
#' ds <- vggface2_dataset(download = TRUE)
#' item <- ds[1]
#' item$x      # image array RGB
#' item$y      # integer label
#' ds$classes[item$y]  # list(name=..., gender=...)
#'
#' #Load the test set
#' ds <- vggface2_dataset(download = TRUE, train = FALSE)
#' item <- ds[1]
#' item$x      # image array RGB
#' item$y      # integer label
#' ds$classes[item$y]  # list(name=..., gender=...)
#' }
#'
#' @family classification_dataset
#' @export
vggface2_dataset <- torch::dataset(
  name = "vggface2",
  resources = data.frame(
    split = c("train", "train", "test", "test", "identity"),
    kind = c("image", "label", "image", "label", "label"),
    url = c("https://huggingface.co/datasets/ProgramComputer/VGGFace2/resolve/main/data/vggface2_train.tar.gz",
            "https://huggingface.co/datasets/ProgramComputer/VGGFace2/resolve/main/meta/train_list.txt",
            "https://huggingface.co/datasets/ProgramComputer/VGGFace2/resolve/main/data/vggface2_test.tar.gz",
            "https://huggingface.co/datasets/ProgramComputer/VGGFace2/resolve/main/meta/test_list.txt",
            "https://huggingface.co/datasets/ProgramComputer/VGGFace2/raw/main/meta/identity_meta.csv"),
    md5 = c("88813c6b15de58afc8fa75ea83361d7f",
            "4cfbab4a839163f454d7ecef28b68669",
            "bb7a323824d1004e14e00c23974facd3",
            "d08b10f12bc9889509364ef56d73c621",
            "d315386c7e8e166c4f60e27d9cc61acc"),
    size = c(37.9e9, 62e6, 2.03e9, 3e6, 335e3)
  ),
  initialize = function(
    root = tempdir(),
    split = "val",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
    self$root_path <- root
    self$split <- match.arg(split, c("train", "test"))
    self$transform <- transform
    self$target_transform <- target_transform
    self$archive_url <- self$resources[self$resources$split %in% c(split, "identity"),]$url
    self$archive_size <- prettyunits::pretty_bytes(sum(self$resources[self$resources$split %in% c(split, "identity"),]$size))
    self$archive_md5 <- self$resources[self$resources$split %in% c(split, "identity"),]$md5
    self$split_file <- sapply(self$archive_url, function(x) file.path(rappdirs::user_cache_dir("torch"), class(self)[1], basename(x)))

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }


    data <- readRDS(file.path(self$processed_folder, glue::glue("{self$split}.rds")))

    self$img_path <- data$img_path
    self$labels <- data$labels
    self$class_to_idx <- data$class_to_idx
    self$identity_df <- data$identity_df

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$class_to_idx)} classes.")
  },

  download = function() {
    if (self$check_exists())
      return(NULL)

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    # Download and extract archives
    archive <- sapply(self$archive_url, function(x)  download_and_cache(x, prefix = class(self)[1]))
    if (!all(tools::md5sum(archive) == self$archive_md5))
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")
    for (file_id in seq_along(archive)) {
      fs::file_move(archive[file_id], self$split_file[file_id])
    }
    utils::untar(archive[[1]], exdir = self$raw_folder)

    identity_df <- read.csv(archive[[3]], sep = ",", stringsAsFactors = FALSE, strip.white = TRUE)
    identity_df$Class_ID <- trimws(identity_df$Class_ID)
    identity_df$Gender <- factor(identity_df$Gender, labels = c("Female", "Male"))

    files <- read.delim(archive[[2]], sep = "/", col.names = c("Class_ID", "img_path"), header = FALSE)
    files$img_path <- file.path(self$raw_folder, self$split, files$Class_ID, files$img_path)
    class_to_idx <- setNames(seq_len(nlevels(as.factor(files$Class_ID))), levels(as.factor(files$Class_ID)))
    labels <- names(class_to_idx)
    saveRDS(list(img_path = files,identity_df = identity_df, labels = labels,
                   class_to_idx = class_to_idx),
              file.path(self$processed_folder, glue::glue("{self$split}.rds"))
      )

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  check_exists = function() {
    all(fs::file_exists(self$split_file)) &&
      fs::file_exists(file.path(self$processed_folder, glue::glue("{self$split}.rds")))
  },

  .getitem = function(index) {
    x <- jpeg::readJPEG(self$img_path$img_path[index])
    y <- self$class_to_idx[self$img_path$Class_ID[index]]

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }
    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }
    list(x = x, y = y)
  },

  .length = function() {
    nrow(self$img_path)
  },

  active = list(
    raw_folder = function() {
      fs::dir_create(file.path(self$root_path, class(self)[1], "raw"))
      file.path(self$root_path, class(self)[1], "raw")
    },
    processed_folder = function() {
      fs::dir_create(file.path(self$root_path, class(self)[1], "processed"))
      file.path(self$root_path, class(self)[1], "processed")
    }
  )
)

#' Pascal VOC Segmentation Dataset
#'
#' The Pascal Visual Object Classes (VOC) dataset is a widely used benchmark for object detection and semantic segmentation tasks in computer vision.
#'
#' This dataset provides RGB images along with per-pixel class segmentation masks for 20 object categories, plus a background class.
#' Each pixel in the mask is labeled with a class index corresponding to one of the predefined semantic categories.
#'
#' The VOC dataset was released in yearly editions (2007 to 2012), with slight variations in data splits and annotation formats.
#' Notably, only the 2007 edition includes a separate `test` split; all other years (2008–2012) provide only the `train`, `val`, and `trainval` splits.
#'
#' The dataset defines 21 semantic classes: `"background"`, `"aeroplane"`, `"bicycle"`, `"bird"`, `"boat"`, `"bottle"`, `"bus"`, `"car"`, `"cat"`, `"chair"`,
#' `"cow"`, `"dining table"`, `"dog"`, `"horse"`, `"motorbike"`, `"person"`, `"potted plant"`, `"sheep"`, `"sofa"`, `"train"`, and `"tv/monitor"`.
#' They are available through the `classes` variable of the dataset object.
#'
#' @param class_id Integer vector of 1-based class identifiers. Must be within \[1, 21\].
#'
#' @family class_resolution
#' @export
pascal_voc_classes <- function(class_id = 1:21) {
  if (any(class_id > 21)) {
    cli_warn("PASCAL VOC {.var class_id} cannot be > 21")
  }
  c(
    "background", "aeroplane", "bicycle", "bird", "boat", "bottle",
    "bus", "car", "cat", "chair", "cow", "dining table", "dog", "horse",
    "motorbike", "person", "potted plant", "sheep", "sofa", "train", "tv/monitor"
  )[class_id]
}

#'
#' This dataset is frequently used for training and evaluating semantic segmentation models, and supports tasks requiring dense, per-pixel annotations.
#'
#' @inheritParams oxfordiiitpet_dataset
#' @param root Character. Root directory where the dataset will be stored under `root/pascal_voc_<year>`.
#' @param year Character. VOC dataset version to use. One of `"2007"`, `"2008"`, `"2009"`, `"2010"`, `"2011"`, or `"2012"`. Default is `"2012"`.
#' @param split Character. One of `"train"`, `"val"`, `"trainval"`, or `"test"`. Determines the dataset split. Default is `"train"`.
#'
#' @return A torch dataset of class \code{pascal_segmentation_dataset}.
#'
#' The returned list inherits class \code{image_with_segmentation_mask}, which allows generic visualization
#' utilities to be applied.
#'
#' Each element is a named list with the following structure:
#' - `x`: a H x W x 3 array representing the RGB image.
#' - `y`: A named list containing:
#'     - `masks`: A `torch_tensor` of dtype `bool` and shape `(21, H, W)`, representing a multi-channel segmentation mask.
#'       Each of the 21 channels corresponds to a Pascal VOC classes
#'     - `labels`: An integer vector indicating the indices of the classes present in the mask.
#'
#' @examples
#' \dontrun{
#' # Load Pascal VOC segmentation dataset (2007 train split)
#' pascal_seg <- pascal_segmentation_dataset(
#'  transform = transform_to_tensor,
#'  download = TRUE,
#'  year = "2007"
#' )
#'
#' # Access the first image and its mask
#' first_item <- pascal_seg[1]
#' first_item$x  # Image
#' first_item$y$masks  # Segmentation mask
#' first_item$y$labels  # Unique class labels in the mask
#' pascal_voc_classes(first_item$y$labels)  # Class names
#'
#' # Visualise the first image and its mask
#' masked_img <- draw_segmentation_masks(first_item)
#' tensor_image_browse(masked_img)
#'
#' # Load Pascal VOC detection dataset (2007 train split)
#' pascal_det <- pascal_detection_dataset(
#'  transform = transform_to_tensor,
#'  download = TRUE,
#'  year = "2007"
#' )
#'
#' # Access the first image and its bounding boxes
#' first_item <- pascal_det[1]
#' first_item$x  # Image
#' first_item$y$labels  # Object labels
#' first_item$y$boxes  # Bounding box tensor
#'
#' # Visualise the first image with bounding boxes
#' boxed_img <- draw_bounding_boxes(first_item)
#' tensor_image_browse(boxed_img)
#' }
#'
#' @name pascal_voc_datasets
#' @title Pascal VOC Datasets
#' @rdname pascal_voc_datasets
#' @family segmentation_dataset
#' @export
pascal_segmentation_dataset <- torch::dataset(
  name = "pascal_segmentation_dataset",

  resources = data.frame(
    year = c("2007", "2007", "2008", "2009", "2010", "2011", "2012"),
    type = c("trainval", "test", "trainval", "trainval", "trainval", "trainval", "trainval"),
    url = c("https://huggingface.co/datasets/DerrickUnleashed/Pascal_VOC/resolve/main/VOCtrainval_06-Nov-2007.tar",
            "https://huggingface.co/datasets/DerrickUnleashed/Pascal_VOC/resolve/main/VOCtest_06-Nov-2007.tar",
            "https://huggingface.co/datasets/DerrickUnleashed/Pascal_VOC/resolve/main/VOCtrainval_14-Jul-2008.tar",
            "https://huggingface.co/datasets/DerrickUnleashed/Pascal_VOC/resolve/main/VOCtrainval_11-May-2009.tar",
            "https://huggingface.co/datasets/DerrickUnleashed/Pascal_VOC/resolve/main/VOCtrainval_03-May-2010.tar",
            "https://huggingface.co/datasets/DerrickUnleashed/Pascal_VOC/resolve/main/VOCtrainval_25-May-2011.tar",
            "https://huggingface.co/datasets/DerrickUnleashed/Pascal_VOC/resolve/main/VOCtrainval_11-May-2012.tar"),
    md5 = c("c52e279531787c972589f7e41ab4ae64",
            "b6e924de25625d8de591ea690078ad9f",
            "2629fa636546599198acfcfbfcf1904a",
            "59065e4b188729180974ef6572f6a212",
            "da459979d0c395079b5c75ee67908abb",
            "6c3384ef61512963050cb5d687e5bf1e",
            "6cd6e144f989b92b3379bac3b3de84fd"),
    size = c("440 MB", "440 MB", "550 MB", "890 MB", "1.3 GB", "1.7 GB", "1.9 GB")
  ),
  classes = pascal_voc_classes(),
  voc_colormap = c(
    c(0, 0, 0), c(128, 0, 0), c(0, 128, 0), c(128, 128, 0),
    c(0, 0, 128), c(128, 0, 128), c(0, 128, 128), c(128, 128, 128),
    c(64, 0, 0), c(192, 0, 0), c(64, 128, 0), c(192, 128, 0),
    c(64, 0, 128), c(192, 0, 128), c(64, 128, 128), c(192, 128, 128),
    c(0, 64, 0), c(128, 64, 0), c(0, 192, 0), c(128, 192, 0),
    c(0, 64, 128)
  ),

  initialize = function(
    root = tempdir(),
    year = "2012",
    split = "train",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {
    self$root_path <- root
    self$year <- match.arg(year, choices = unique(self$resources$year))
    self$split <-  match.arg(split, choices = c("train", "val", "trainval", "test"))
    self$transform <- transform
    self$target_transform <- target_transform
    if (self$split == "test"){
        self$archive_key <- "test"
    } else {
        self$archive_key <- "trainval"
    }
    self$archive_size <- self$resources[self$resources$year == self$year & self$resources$type == self$archive_key,]$size

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }

    data_file <- file.path(self$processed_folder, paste0(self$split, ".rds"))
    data <- readRDS(data_file)
    self$img_path <- data$img_path
    self$mask_paths <- data$mask_paths

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  download = function() {

    if (self$check_exists()) {
      return()
    }

    cli_inform("Downloading {.cls {class(self)[[1]]}}...")

    fs::dir_create(self$raw_folder)
    fs::dir_create(self$processed_folder)

    resource <- self$resources[self$resources$year == self$year & self$resources$type == self$archive_key,]
    archive <- download_and_cache(resource$url, prefix = "pascal_dataset")
    actual_md5 <- tools::md5sum(archive)

    if (actual_md5 != resource$md5) {
      runtime_error("Corrupt file! Delete the file in {archive} and try again.")
    }

    utils::untar(archive, exdir = self$raw_folder)

    voc_dir <- file.path(self$raw_folder, "VOCdevkit", paste0("VOC", self$year))
    voc_root <- self$raw_folder
    if (self$year == "2011") {
    voc_root <- file.path(voc_root, "TrainVal")
    }
    voc_dir <- file.path(voc_root, "VOCdevkit", paste0("VOC", self$year))

    split_file <- file.path(voc_dir, "ImageSets", "Segmentation", paste0(self$split, ".txt"))

    ids <- readLines(split_file)
    img_path <- file.path(voc_dir, "JPEGImages", paste0(ids, ".jpg"))
    mask_paths <- file.path(voc_dir, "SegmentationClass", paste0(ids, ".png"))

    saveRDS(list(
      img_path = img_path,
      mask_paths = mask_paths
    ), file.path(self$processed_folder, paste0(self$split, ".rds")))

    cli_inform("Dataset {.cls {class(self)[[1]]}} downloaded and extracted successfully.")
  },

  check_exists = function() {
    fs::file_exists(file.path(self$processed_folder, paste0(self$split, ".rds")))
  },

  .getitem = function(index) {

    img_path <- self$img_path[index]
    mask_path <- self$mask_paths[index]

    x <- jpeg::readJPEG(img_path)
    mask_data <- png::readPNG(mask_path) * 255

    flat_mask <- matrix(
      c(as.vector(mask_data[, , 1]),
        as.vector(mask_data[, , 2]),
        as.vector(mask_data[, , 3])),
      ncol = 3
    )
    colormap_mat <- matrix(self$voc_colormap, ncol = 3, byrow = TRUE)
    rgb_to_int <- function(mat) {
      as.integer(mat[, 1]) * 256^2 + as.integer(mat[, 2]) * 256 + as.integer(mat[, 3])
    }
    match_indices <- match(rgb_to_int(flat_mask), rgb_to_int(colormap_mat)) - 1
    class_idx <- matrix(match_indices, nrow = dim(mask_data)[1], ncol = dim(mask_data)[2])
    class_idx_tensor <- torch_tensor(class_idx, dtype = torch_long())
    class_ids <- torch_arange(0, 20, dtype = torch_long())$view(c(21, 1, 1))
    masks <- (class_ids == class_idx_tensor$unsqueeze(1))$to(dtype = torch_bool())
    labels <- which(as_array(masks$any(dim = c(2, 3))))

    y <- list(labels = labels, masks = masks)

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }
    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    result <- list(x = x, y = y)
    class(result) <- c("image_with_segmentation_mask", class(result))
    result
  },

  .length = function() {
    length(self$img_path)
  },

  active = list(
    raw_folder = function() {
      file.path(self$root_path, paste0("pascal_voc_", self$year), "raw")
    },
    processed_folder = function() {
      file.path(self$root_path, paste0("pascal_voc_", self$year), "processed")
    }
  )
)

#' Pascal VOC Detection Dataset
#'
#' @inheritParams pascal_segmentation_dataset
#'
#' @return A torch dataset of class \code{pascal_detection_dataset}.
#'
#' The returned list inherits class \code{image_with_bounding_box}, which allows generic visualization
#' utilities to be applied.
#'
#' Each element is a named list:
#' - `x`: a H x W x 3 array representing the RGB image.
#' - `y`: a list with:
#'     - `labels`: a character vector with object class names.
#'     - `boxes`: a tensor of shape (N, 4) with bounding box coordinates in `(xmin, ymin, xmax, ymax)` format.
#'
#' @rdname pascal_voc_datasets
#' @family detection_dataset
#' @export
pascal_detection_dataset <- torch::dataset(
  name = "pascal_detection_dataset",

  inherit = pascal_segmentation_dataset,

  initialize = function(
    root = tempdir(),
    year = "2012",
    split = "train",
    transform = NULL,
    target_transform = NULL,
    download = FALSE
  ) {

    self$root_path <- root
    self$year <- match.arg(year, choices = unique(self$resources$year))
    self$split <- match.arg(split, choices = c("train", "val", "trainval", "test"))
    self$transform <- transform
    self$target_transform <- target_transform
    if (self$split == "test") {
      self$archive_key <- "test"
    } else {
      self$archive_key <- "trainval"
    }
    self$archive_size <- self$resources[self$resources$year == self$year & self$resources$type == self$archive_key,]$size

    if (download) {
      cli_inform("Split {.val {self$split}} of dataset {.cls {class(self)[[1]]}} (~{.emph {self$archive_size}}) will be downloaded and processed if not already available.")
      self$download()
    }

    if (!self$check_exists()) {
      cli_abort("Dataset not found. You can use `download = TRUE` to download it.")
    }

    voc_dir <- file.path(self$raw_folder, "VOCdevkit", paste0("VOC", self$year))
    if (self$year == "2011") {
      voc_dir <- file.path(self$raw_folder, "TrainVal", "VOCdevkit", paste0("VOC", self$year))
    }

    ids_file <- file.path(voc_dir, "ImageSets", "Main", paste0(self$split, ".txt"))
    ids <- readLines(ids_file)

    self$img_path <- file.path(voc_dir, "JPEGImages", paste0(ids, ".jpg"))
    self$annotation_paths <- file.path(voc_dir, "Annotations", paste0(ids, ".xml"))

    if (!requireNamespace("xml2", quietly = TRUE)) {
      install.packages("xml2")
    }

    cli_inform("{.cls {class(self)[[1]]}} dataset loaded with {self$.length()} images across {length(self$classes)} classes.")
  },

  .getitem = function(index) {

    x <- jpeg::readJPEG(self$img_path[index])
    ann_path <- self$annotation_paths[index]
    y <- self$parse_voc_xml(xml2::read_xml(ann_path))

    if (!is.null(self$transform)) {
      x <- self$transform(x)
    }
    if (!is.null(self$target_transform)) {
      y <- self$target_transform(y)
    }

    result <- list(x = x, y = y)
    class(result) <- c("image_with_bounding_box", class(result))
    result
  },

  parse_voc_xml = function(xml) {
    objects <- xml2::xml_find_all(xml, ".//object")

    labels <- xml2::xml_text(xml2::xml_find_all(objects, "name"))

    bboxes <- xml2::xml_find_all(objects, "bndbox")

    xmin <- xml2::xml_integer(xml2::xml_find_all(bboxes, "xmin"))
    ymin <- xml2::xml_integer(xml2::xml_find_all(bboxes, "ymin"))
    xmax <- xml2::xml_integer(xml2::xml_find_all(bboxes, "xmax"))
    ymax <- xml2::xml_integer(xml2::xml_find_all(bboxes, "ymax"))

    boxes <- torch_tensor(data.frame(xmin, ymin, xmax, ymax) %>% as.matrix(), dtype = torch_long())

    list(
      labels = labels,
      boxes = boxes
    )
  }
)

#' ConvNeXt Implementation
#'
#' Implements the ConvNeXt architecture from [ConvNeXt: A ConvNet for the 2020s](https://arxiv.org/pdf/2201.03545)
#'
#' @inheritParams model_mobilenet_v3
#' @param channels The number of channels in the input image.  Default: 3.
#' @param ... Other parameters passed to the model implementation.
#'
#' @section Variants:
#' ## Model Summary and Performance for pretrained weights
#' ```
#' | Model                | Top-1 Acc| Params | GFLOPS | File Size | `num_classes`| image size |
#' |----------------------|----------|--------|--------|-----------|--------------|------------|
#' | convnext_tiny_1k     | 82.1%    | 28M    | 4.5    | 109 MB    |         1000 | 224 x 224  |
#' | convnext_tiny_22k    | 82.9%    | 29M    | 4.5    | 170 MB    |        21841 | 224 x 224  |
#' | convnext_small_22k   | 84.6%    | 50M    | 8.7    | 252 MB    |        21841 | 224 x 224  |
#' | convnext_small_22k1k | 84.6%    | 50M    | 8.7    | 252 MB    |        21841 | 224 x 224  |
#' | convnext_base_1k     | 85.1%    | 89M    | 15.4   | 338 MB    |         1000 | 224 x 224  |
#' | convnext_base_22k    | 85.8%    | 89M    | 15.4   | 420 MB    |        21841 | 224 x 224  |
#' | convnext_large_1k    | 84.3%    | 198M   | 34.4   | 750 MB    |         1000 | 224 x 224  |
#' | convnext_large_22k   | 86.6%    | 198M   | 34.4   | 880 MB    |        21841 | 224 x 224  |
#' ```
#'
#' @examples
#' \dontrun{
#' # 1. Download sample image (dog)
#' norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants, see
#' # https://pytorch.org/vision/stable/models.html
#' norm_std  <- c(0.229, 0.224, 0.225)
#' img_url <- "https://en.wikipedia.org/wiki/Special:FilePath/Felis_catus-cat_on_snow.jpg"
#' img <- base_loader(img_url)
#'
#' # 2. Convert to tensor (RGB only), resize and normalize
#' input <- img %>%
#'  transform_to_tensor() %>%
#'  transform_resize(c(224, 224)) %>%
#'  transform_normalize(norm_mean, norm_std)
#' batch <- input$unsqueeze(1)
#'
#' # 3. Load pretrained models
#' model_small <- convnext_tiny_1k(pretrained = TRUE, root = tempdir())
#' model_small$eval()
#'
#' # 4. Forward pass
#' output_s <- model_small(batch)
#'
#' # 5. Show Top-5 predictions
#' topk <- output_s$topk(k = 5, dim = 2)
#' indices <- as.integer(topk[[2]][1, ])
#' scores <- as.numeric(topk[[1]][1, ])
#' glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}%)")
#' }
#'
#' @family classification_model
#' @name model_convnext
NULL

#' @importFrom torch nn_module nn_parameter torch_ones torch_zeros nnf_layer_norm
LayerNorm <- nn_module(
  "LayerNorm",
  initialize = function(normalized_shape,
                        eps = 1e-6,
                        data_format = "channels_last") {
    self$weight <- nn_parameter(torch_ones(normalized_shape))
    self$bias <- nn_parameter(torch_zeros(normalized_shape))
    self$eps <- eps
    self$data_format <- data_format
    self$normalized_shape <- normalized_shape
  },
  forward = function(x) {
    if (self$data_format == "channels_last") {
      nnf_layer_norm(x, self$normalized_shape, self$weight, self$bias, self$eps)
    } else if (self$data_format == "channels_first") {
      u <- x$mean(dim = 2, keepdim = TRUE)
      s <- ((x - u)$pow(2))$mean(dim = 2, keepdim = TRUE)
      x <- (x - u) / (s + self$eps)$sqrt()
      x <- self$weight$unsqueeze(2)$unsqueeze(3) * x + self$bias$unsqueeze(2)$unsqueeze(3)
      x
    } else {
      stop("Unsupported data format")
    }
  }
)

#' @importFrom torch nn_conv2d nn_linear torch_ones nn_gelu nn_identity
Block <- nn_module(
  "Block",
  initialize = function(dim,
                        drop_path = 0,
                        layer_scale_init_value = 1e-6) {
    self$dwconv <- nn_conv2d(dim, dim, kernel_size = 7, padding = 3, groups = dim)
    self$norm <- LayerNorm(dim, eps = 1e-6)
    self$pwconv1 <- nn_linear(dim, 4 * dim)
    self$act <- nn_gelu()
    self$pwconv2 <- nn_linear(4 * dim, dim)
    self$gamma <- if (layer_scale_init_value > 0) {
      nn_parameter(layer_scale_init_value * torch_ones(dim))
    } else {
      NULL
    }
    self$drop_path <- nn_identity()
  },
  forward = function(x) {
    input <- x
    x <- self$dwconv(x)
    x <- x$permute(c(1, 3, 4, 2))
    x <- self$norm(x)
    x <- self$pwconv1(x)
    x <- self$act(x)
    x <- self$pwconv2(x)
    if (!is.null(self$gamma)) {
      x <- self$gamma * x
    }
    x <- x$permute(c(1, 4, 2, 3))
    x <- input + self$drop_path(x)
    x
  }
)


#' @importFrom torch nn_conv2d nn_linear nn_sequential nn_module_list torch_linspace nn_layer_norm
ConvNeXt <- nn_module(
  "ConvNeXt",
  initialize = function(in_chans = 3,
                        num_classes = 1000,
                        depths = c(3, 3, 9, 3),
                        dims = c(96, 192, 384, 768),
                        drop_path_rate = 0.,
                        layer_scale_init_value = 1e-6,
                        head_init_scale = 1.) {
    self$downsample_layers <- nn_module_list()

    # Stem
    stem <- nn_sequential(
      nn_conv2d(in_chans, dims[1], kernel_size = 4, stride = 4),
      LayerNorm(dims[1], eps = 1e-6, data_format = "channels_first")
    )
    self$downsample_layers$append(stem)

    for (i in 1:3) {
      self$downsample_layers$append(nn_sequential(
        LayerNorm(dims[i], eps = 1e-6, data_format = "channels_first"),
        nn_conv2d(dims[i], dims[i + 1], kernel_size = 2, stride = 2)
      ))
    }

    self$stages <- nn_module_list()
    dp_rates <- as.numeric(torch_linspace(0, drop_path_rate, sum(depths)))
    cur <- 1
    for (i in 1:4) {
      blocks <- lapply(1:depths[i], function(j) {
        Block(dims[i],
              drop_path = dp_rates[cur],
              layer_scale_init_value = layer_scale_init_value)
      })
      self$stages$append(nn_sequential(!!!blocks))
      cur <- cur + depths[i]
    }

    self$norm <- nn_layer_norm(dims[4], eps = 1e-6)
    self$head <- nn_linear(dims[4], num_classes)
  },

  forward_features = function(x) {
    for (i in 1:4) {
      x <- self$downsample_layers[[i]](x)
      x <- self$stages[[i]](x)
    }
    x$mean(dim = c(3, 4)) %>% self$norm()
  },

  forward = function(x) {
    x <- self$forward_features(x)
    self$head(x)
  }
)


convnext_model_urls <- c(
  "convnext_tiny_1k" = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_tiny_1k.pth",
  "convnext_tiny_22k" = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_tiny_22k.pth",
  "convnext_small_22k" = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_small_22k.pth",
  'convnext_small_22k1k' = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_small_22k1k.pth",
  'convnext_base_1k' = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_base_1k.pth",
  'convnext_base_22k' = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_base_22k.pth",
  'convnext_large_1k' = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_large_1k.pth",
  'convnext_large_22k' = "https://torch-cdn.mlverse.org/models/vision/v2/models/convnext_large_22k.pth"
)



.convnext <- function(arch, channels, depths, dims, num_classes, pretrained, progress, ...) {
  if (!is.character(arch) || length(arch) != 1) {
    stop("arch must be a single character string.")
  }
  depths <- as.integer(depths)
  dims <- as.integer(dims)
  channels <- as.integer(channels)
  num_classes <- as.integer(num_classes)

  if (length(depths) != 4 || length(dims) != 4) {
    stop("depths and dims must be vectors of length 4.")
  }
  model <- ConvNeXt(
    in_chans = channels,
    depths = depths,
    dims = dims,
    num_classes = num_classes
  )

  if (pretrained) {
    if (!arch %in% names(convnext_model_urls)) {
      stop(paste("Pretrained model for", arch, "is not available."))
    }
    state_dict_path <- download_and_cache(convnext_model_urls[arch], prefix = "convnext")
    state_dict <- torch::load_state_dict(state_dict_path)

    # Interpolate stem weights if input channels differ.sample use cases - satellite images
    conv1_weight <- state_dict[["downsample_layers.0.0.weight"]]

    if (dim(conv1_weight)[2] != channels) {
      old_in_channels <- dim(conv1_weight)[2]
      mean_weight <- conv1_weight$mean(dim = 2, keepdim = TRUE)

      # Repeat manually using torch_cat
      new_weight_list <- rep(list(mean_weight), channels)
      new_weight <- torch_cat(new_weight_list, dim = 2)

      new_weight <- new_weight * (old_in_channels / channels)
      state_dict[["downsample_layers.0.0.weight"]] <- new_weight
    }
    model$load_state_dict(state_dict, strict = FALSE)
  }
  model
}


#' @describeIn model_convnext ConvNeXt Tiny model trained on Imagenet 1k.
#' @export
model_convnext_tiny_1k <- function(pretrained = FALSE,
                                progress = TRUE,
                                channels = 3,
                                num_classes = 1000,
                                ...) {
  .convnext(
    arch = "convnext_tiny_1k" ,
    channels = channels,
    depths = c(3, 3, 9, 3),
    dims = c(96, 192, 384, 768),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}


#' @describeIn model_convnext ConvNeXt Tiny model trained on Imagenet 22k.
#' @export
model_convnext_tiny_22k <- function(pretrained = FALSE, progress = TRUE, channels = 3, num_classes = 21841, ...) {
  .convnext(
    "convnext_tiny_22k",
    channels = channels,
    depths = c(3, 3, 9, 3),
    dims = c(96, 192, 384, 768),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}


#' @describeIn model_convnext ConvNeXt Small model trained on Imagenet 22k.
#' @export
model_convnext_small_22k <- function(pretrained = FALSE, progress = TRUE, channels = 3, num_classes = 21841, ...) {
  .convnext(
    arch = "convnext_small_22k" ,
    channels = channels,
    depths = c(3, 3, 27, 3),
    dims = c(96, 192, 384, 768),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}


#' @describeIn model_convnext ConvNeXt Small model pretrained on Imagenet 1k
#'  and fine-tuned on Imagenet 22k classes.
#' @export
model_convnext_small_22k1k <- function(pretrained = FALSE, progress = TRUE, channels = 3, num_classes = 21841, ...) {
  .convnext(
    "convnext_small_22k1k",
    channels = channels,
    depths = c(3, 3, 27, 3),
    dims = c(96, 192, 384, 768),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}


#' @describeIn model_convnext ConvNeXt Base model trained on Imagenet 1k.
#' @export
model_convnext_base_1k <- function(pretrained = FALSE,
                                   progress = TRUE,
                                   channels = 3,
                                   num_classes = 1000,
                                   ...) {
  .convnext(
    arch = "convnext_base_1k" ,
    channels = channels,
    depths = c(3, 3, 27, 3),
    dims = c(128, 256, 512, 1024),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}


#' @describeIn model_convnext ConvNeXt Base model trained on Imagenet 22k.
#' @export
model_convnext_base_22k <- function(pretrained = FALSE,
                                   progress = TRUE,
                                   channels = 3,
                                   num_classes = 21841,
                                   ...) {
  .convnext(
    arch = "convnext_base_22k" ,
    channels = channels,
    depths = c(3, 3, 27, 3),
    dims = c(128, 256, 512, 1024),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}


#' @describeIn model_convnext ConvNeXt Large model trained on Imagenet 1k.
#' @export
model_convnext_large_1k <- function(pretrained = FALSE,
                                   progress = TRUE,
                                   channels = 3,
                                   num_classes = 1000,
                                   ...) {
  .convnext(
    arch = "convnext_large_1k" ,
    channels = channels,
    depths = c(3, 3, 27, 3),
    dims = c(192, 384, 768, 1536),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}


#' @describeIn model_convnext ConvNeXt Large model trained on Imagenet 22k.
#' @export
model_convnext_large_22k <- function(pretrained = FALSE,
                                   progress = TRUE,
                                   channels = 3,
                                   num_classes = 21841,
                                   ...) {
  .convnext(
    arch = "convnext_large_22k" ,
    channels = channels,
    depths = c(3, 3, 27, 3),
    dims = c(192, 384, 768, 1536),
    num_classes = num_classes,
    pretrained,
    progress,
    ...
  )
}

context("models-convnext")

test_that("non-pretrained model_convnext_*_1k work, with or wo a changed classification layer", {
  expect_no_error(
    model_1k <- model_convnext_tiny_1k(pretrained = FALSE)
  )
  input <- torch_randn(3, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(3, 1000))

  expect_no_error(
    model_1k <- model_convnext_base_1k(pretrained = FALSE)
  )
  input <- torch_randn(1, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(1, 1000))

  expect_no_error(
    model_1k <- model_convnext_large_1k(pretrained = FALSE)
  )
  input <- torch_randn(1, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(1, 1000))

  expect_no_error(
    model <- model_convnext_tiny_1k(pretrained = FALSE, num_classes = 130)
  )
  input <- torch_randn(4, 3, 224, 224)
  out <- model(input)
  expect_tensor_shape(out, c(4, 130))

  expect_no_error(
    model_1k <- model_convnext_small_22k1k(pretrained = FALSE, num_classes = 37)
  )
  input <- torch_randn(2, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(2, 37))

  expect_no_error(
    model_1k <- model_convnext_base_1k(pretrained = FALSE, num_classes = 10)
  )
  input <- torch_randn(1, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(1, 10))

  expect_no_error(
    model_1k <- model_convnext_large_1k(pretrained = FALSE, num_classes = 10)
  )
  input <- torch_randn(1, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(1, 10))

  rm(model_1k)
  rm(model)
  gc()
})

test_that("pretrained model_convnext_*_1k works", {
  expect_no_error(
    model_1k <- model_convnext_tiny_1k(pretrained = TRUE)
  )
  input <- torch_randn(5, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(5, 1000))

  expect_no_error(
    model_1k <- model_convnext_base_1k(pretrained = TRUE)
  )
  input <- torch_randn(4, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(4, 1000))

  rm(model_1k)
  gc()
})

test_that("pretrained model_convnext_*_22k works", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  expect_no_error(
    model_22k <- model_convnext_tiny_22k(pretrained = TRUE)
  )
  input <- torch_randn(5, 3, 224, 224)
  model_22k$eval()
  out <- model_22k(input)
  expect_tensor_shape(out, c(5, 21841))

  expect_no_error(
    model_22k <- model_convnext_small_22k(pretrained = TRUE)
  )
  input <- torch_randn(2, 3, 224, 224)
  model_22k$eval()
  out <- model_22k(input)
  expect_tensor_shape(out, c(2, 21841))

  expect_no_error(
    model_1k <- model_convnext_small_22k1k(pretrained = FALSE)
  )
  input <- torch_randn(2, 3, 224, 224)
  model_1k$eval()
  out <- model_1k(input)
  expect_tensor_shape(out, c(2, 21841))

  expect_no_error(
    model_22k <- model_convnext_base_22k(pretrained = TRUE)
  )
  input <- torch_randn(1, 3, 224, 224)
  model_22k$eval()
  out <- model_22k(input)
  expect_tensor_shape(out, c(1, 21841))

  rm(model_22k)
  gc()
})

test_that("pretrained model_convnext_large_* works", {
  skip_if(Sys.getenv("TEST_LARGE_MODELS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_MODELS=1 to enable tests requiring large downloads.")

  expect_no_error(
    model_22k <- model_convnext_large_1k(pretrained = TRUE)
  )
  input <- torch_randn(1, 3, 224, 224)
  model_22k$eval()
  out <- model_22k(input)
  expect_tensor_shape(out, c(1, 1000))

  expect_no_error(
    model_22k <- model_convnext_large_22k(pretrained = TRUE)
  )
  input <- torch_randn(1, 3, 224, 224)
  model_22k$eval()
  out <- model_22k(input)
  expect_tensor_shape(out, c(1, 21841))

  rm(model_22k)
  gc()
})

IMG_EXTENSIONS <-  c('jpg', 'jpeg', 'png', 'tif', 'tiff') # 'ppm', 'bmp', 'pgm', 'webp'

has_file_allowed_extension <- function(filename, extensions) {
  tolower(fs::path_ext(filename)) %in% tolower(extensions )
}

is_image_file <- function(filename) {
  has_file_allowed_extension(filename, IMG_EXTENSIONS)
}

is_url <- function(path) {
  grepl("^((http|ftp)s?|sftp)://", path, ignore.case = TRUE)
}

folder_make_dataset <- function(directory, class_to_idx, extensions = NULL, is_valid_file = NULL) {
  directory <- normalizePath(directory)

  both_none <- is.null(extensions) && is.null(is_valid_file)
  both_something <- !is.null(extensions) && ! is.null(is_valid_file)

  if (both_none || both_something)
    value_error("Both extensions and is_valid_file cannot be None or not None at the same time")

  if (!is.null(extensions)) {
    is_valid_file <- function(filename) {
      has_file_allowed_extension(filename, extensions)
    }
  }

  paths <- c()
  indexes <- c()

  for (target_class in sort(names(class_to_idx))) {

    class_index <- class_to_idx[target_class]
    target_dir <- fs::path_join(c(directory, target_class))

    if (!fs::is_dir(target_dir))
      next

    fnames <- fs::dir_ls(target_dir, recurse = TRUE)
    fnames <- fnames[is_valid_file(fnames)]

    paths <- c(paths, fnames)
    indexes <- c(indexes, rep(class_index, length(fnames)))
  }

  list(
    paths,
    indexes
  )
}

folder_dataset <- torch::dataset(
  name = "folder",
  initialize = function(root, loader, extensions = NULL, transform = NULL,
                        target_transform = NULL, is_valid_file = NULL) {

    self$root <- root
    self$transform <- transform
    self$target_transform <- target_transform

    class_to_idx <- self$.find_classes(root)
    samples <- folder_make_dataset(self$root, class_to_idx, extensions, is_valid_file)

    if (length(samples[[1]]) == 0) {

      msg <- glue::glue("Found 0 files in subfolders of {self$root}")
      if (!is.null(extensions)) {
        msg <- paste0(msg, glue::glue("\nSupported extensions are {paste(extensions, collapse=',')}"))
      }

      runtime_error(msg)
    }

    self$loader <- loader
    self$extensions <- extensions

    self$classes <- names(class_to_idx)
    self$class_to_idx <- class_to_idx
    self$samples <- samples
    self$targets <- samples[[2]]

  },
  .find_classes = function(dir) {
    dirs <- fs::dir_ls(dir, recurse = FALSE, type = "directory")
    dirs <- sapply(fs::path_split(dirs), function(x) tail(x, 1))
    class_to_idx <- seq_along(dirs)
    names(class_to_idx) <- sort(dirs)
    class_to_idx
  },
  .getitem = function(index) {

    path <- self$samples[[1]][index]
    target <- self$samples[[2]][index]

    sample <- self$loader(path)

    if (!is.null(self$transform))
      sample <- self$transform(sample)

    if (!is.null(self$target_transform))
      target <- self$target_transform(target)

    list(x = sample, y = target)
  },
  .length = function() {
    length(self$samples[[1]])
  }
)

#' Load an Image using ImageMagick
#'
#' Load an image located at `path` using the `{magick}` package.
#'
#' @param path path or URL to load the image from.
#'
#' @returns an magick-image object as result of `image_read()`
#'
#' @export
magick_loader <- function(path) {
  magick::image_read(path)
}


#' Base loader
#'
#' Loads an image using `jpeg`, `png` or `tiff` packages depending on the
#' file extension.
#'
#' @inheritParams magick_loader
#'
#' @returns an channel-last array of image values with dim Height x Width x 3
#' @export
base_loader <- function(path) {

  if (is_url(path)) {
    path <- download_and_cache(path)
  }

  if (!fs::file_exists(path)) {
    runtime_error("File not found: {path}")
  }

  ext <- tolower(fs::path_ext(path))

  if (ext %in% c("jpg", "jpeg"))
    img <- jpeg::readJPEG(path)
  else if (ext %in% c("png"))
    img <- png::readPNG(path)
  else if (ext %in% c("tif", "tiff"))
    img <- tiff::readTIFF(path)
  else
    runtime_error("unknown extension {ext} in path {path}")

  if (length(dim(img)) == 2)
    img <- abind::abind(img, img, img, along = 3)
  else if (length(dim(img)) == 3 && dim(img)[1] == 1)
    img <- abind::abind(img, img, img, along = 1) %>% aperm(c(3,1,2))

  # all readers default to channel last
  img
}


#' Create an image folder dataset
#'
#' A generic data loader for images stored in folders.
#'   See `Details` for more information.
#'
#' @details This function assumes that the images for each class are contained
#'   in subdirectories of `root`. The names of these subdirectories are stored
#'   in the `classes` attribute of the returned object.
#'
#' An example folder structure might look as follows:
#'
#' ```
#' root/dog/xxx.png
#' root/dog/xxy.png
#' root/dog/xxz.png
#'
#' root/cat/123.png
#' root/cat/nsdf3.png
#' root/cat/asd932_.png
#' ```
#'
#' @param root Root directory path.
#' @param loader A function to load an image given its path.
#' @param transform  A function/transform that takes in an PIL image and returns
#'   a transformed version. E.g, [transform_random_crop()].
#' @param target_transform A function/transform that takes in the target and
#'   transforms it.
#' @param is_valid_file A function that takes path of an Image file and check if
#'   the file is a valid file (used to check of corrupt files)
#'
#' @family classification_dataset
#'
#' @importFrom torch dataset
#' @export
image_folder_dataset <- dataset(
  "image_folder",
  inherit = folder_dataset,
  initialize = function(root, transform=NULL, target_transform=NULL,
                        loader=NULL, is_valid_file=NULL) {

    if (is.null(loader))
      loader <- base_loader

    if (!is.null(is_valid_file))
      extensions <- NULL
    else
      extensions <- IMG_EXTENSIONS

    super$initialize(root, loader, extensions, transform=transform,
                     target_transform=target_transform,
                     is_valid_file=is_valid_file)
    self$imgs <- self$samples
  }
)

context('dataset-pascal')

t <- withr::local_tempdir()

test_that("tests for the Pascal VOC Segmentation dataset for train split for year 2007", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2007', split = 'train', download = TRUE)
  expect_length(pascal, 209)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for test split for year 2007", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2007', split = 'test', download = TRUE)
  expect_length(pascal, 210)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,375,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 2)
  expect_equal(first_item$y$labels, c(1, 4))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for trainval split for year 2007", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2007', split = 'trainval', download = TRUE)
  expect_length(pascal, 422)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for val split for year 2007", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2007', split = 'val', download = TRUE)
  expect_length(pascal, 213)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,375,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 2)
  expect_equal(first_item$y$labels, c(1, 21))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for train split for year 2008", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2008', split = 'train', download = TRUE)
  expect_length(pascal, 511)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for trainval split for year 2008", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2008', split = 'trainval', download = TRUE)
  expect_length(pascal, 1023)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for val split for year 2008", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2008', split = 'val', download = TRUE)
  expect_length(pascal, 512)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,549000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,366,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 2)
  expect_equal(first_item$y$labels, c(1, 2))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for train split for year 2009", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2009', split = 'train', download = TRUE)
  expect_length(pascal, 749)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for trainval split for year 2009", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2009', split = 'trainval', download = TRUE)
  expect_length(pascal, 1499)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for val split for year 2009", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2009', split = 'val', download = TRUE)
  expect_length(pascal, 750)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,549000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,366,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 2)
  expect_equal(first_item$y$labels, c(1, 2))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for train split for year 2010", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2010', split = 'train', download = TRUE)
  expect_length(pascal, 964)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for trainval split for year 2010", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2010', split = 'trainval', download = TRUE)
  expect_length(pascal, 1928)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for val split for year 2010", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2010', split = 'val', download = TRUE)
  expect_length(pascal, 964)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,549000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,366,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 2)
  expect_equal(first_item$y$labels, c(1, 2))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for train split for year 2011", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2011', split = 'train', download = TRUE)
  expect_length(pascal, 1112)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for trainval split for year 2011", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2011', split = 'trainval', download = TRUE)
  expect_length(pascal, 2223)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for val split for year 2011", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2011', split = 'val', download = TRUE)
  expect_length(pascal, 1111)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,549000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,366,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 2)
  expect_equal(first_item$y$labels, c(1, 2))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for train split for year 2012", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2012', split = 'train', download = TRUE)
  expect_length(pascal, 1464)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for trainval split for year 2012", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2012', split = 'trainval', download = TRUE)
  expect_length(pascal, 2913)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,421500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,281,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 3)
  expect_equal(first_item$y$labels, c(1, 2, 16))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC Segmentation dataset for val split for year 2012", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_segmentation_dataset(root = t, year = '2012', split = 'val', download = TRUE)
  expect_length(pascal, 1449)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,549000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$mask)
  expect_tensor_shape(first_item$y$mask,c(21,366,500))
  expect_tensor_dtype(first_item$y$mask,torch_bool())
  expect_type(first_item$y$labels, "integer")
  expect_length(first_item$y$labels, 2)
  expect_equal(first_item$y$labels, c(1, 2))
  expect_s3_class(first_item, "image_with_segmentation_mask")
})

test_that("tests for the Pascal VOC detection dataset for train split for year 2007", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2007', split = 'train', download = TRUE)
  expect_length(pascal, 2501)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,499500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for test split for year 2007", {

    skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2007', split = 'test', download = TRUE)
  expect_length(pascal, 4952)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,529500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(2,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 2)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for trainval split for year 2007", {

    skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2007', split = 'trainval', download = TRUE)
  expect_length(pascal, 5011)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(5,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 5)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for val split for year 2007", {

    skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2007', split = 'val', download = TRUE)
  expect_length(pascal, 2510)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(5,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 5)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for train split for year 2008", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2008', split = 'train', download = TRUE)
  expect_length(pascal, 2111)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,663000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(2,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 2)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for trainval split for year 2008", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2008', split = 'trainval', download = TRUE)
  expect_length(pascal, 4332)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for val split for year 2008", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2008', split = 'val', download = TRUE)
  expect_length(pascal, 2221)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for train split for year 2009", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2009', split = 'train', download = TRUE)
  expect_length(pascal, 3473)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,663000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(2,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 2)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for trainval split for year 2009", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2009', split = 'trainval', download = TRUE)
  expect_length(pascal, 7054)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for val split for year 2009", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2009', split = 'val', download = TRUE)
  expect_length(pascal, 3581)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for train split for year 2010", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2010', split = 'train', download = TRUE)
  expect_length(pascal, 4998)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,663000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(2,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 2)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for trainval split for year 2010", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2010', split = 'trainval', download = TRUE)
  expect_length(pascal, 10103)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for val split for year 2010", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2010', split = 'val', download = TRUE)
  expect_length(pascal, 5105)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for train split for year 2011", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2011', split = 'train', download = TRUE)
  expect_length(pascal, 5717)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,663000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(2,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 2)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for trainval split for year 2011", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2011', split = 'trainval', download = TRUE)
  expect_length(pascal, 11540)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for val split for year 2011", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2011', split = 'val', download = TRUE)
  expect_length(pascal, 5823)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for train split for year 2012", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2012', split = 'train', download = TRUE)
  expect_length(pascal, 5717)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,663000)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(2,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 2)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for trainval split for year 2012", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2012', split = 'trainval', download = TRUE)
  expect_length(pascal, 11540)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

test_that("tests for the Pascal VOC detection dataset for val split for year 2012", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  pascal <- pascal_detection_dataset(root = t, year = '2012', split = 'val', download = TRUE)
  expect_length(pascal, 5823)
  first_item <- pascal[1]
  expect_named(first_item, c("x", "y"))
  expect_length(first_item$x,562500)
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "list")
  expect_tensor(first_item$y$boxes)
  expect_tensor_shape(first_item$y$boxes,c(1,4))
  expect_tensor_dtype(first_item$y$boxes,torch_int64())
  expect_type(first_item$y$labels,"character")
  expect_length(first_item$y$labels, 1)
  expect_s3_class(first_item, "image_with_bounding_box")
})

context("dataset-coco")

tmp <- withr::local_tempdir()

collate_fn <- function(batch) {
  x <- lapply(batch, function(x) x$x)
  y <- lapply(batch, function(x) x$y)
  list(x = x, y = y)
}

test_that("coco_detection_dataset handles missing files gracefully", {
  expect_error(
    coco_detection_dataset(root = tempfile(), train = TRUE, year = "2017", download = FALSE),
    class = "runtime_error"
  )
})

test_that("coco_detection_dataset loads a single example correctly", {
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- coco_detection_dataset(root = tmp, train = FALSE, year = "2017", download = TRUE)

  expect_s3_class(ds, "coco_detection_dataset")
  expect_gt(length(ds), 0)

  item <- ds[1]
  expect_s3_class(item, "image_with_bounding_box")
  y <- item$y

  expect_is(item$x, "array")
  expect_length(dim(item$x), 3)

  expect_type(y, "list")
  expect_named(y, c("boxes", "labels", "area", "iscrowd"))

  expect_tensor(y$boxes)
  expect_equal(y$boxes$ndim, 2)
  expect_equal(y$boxes$size(2), 4)

  expect_type(y$labels, "character")
  expect_gt(length(y$labels), 0)

  expect_tensor(y$area)
  expect_tensor(y$iscrowd)
  expect_false("segmentation" %in% names(y))
})

test_that("coco_segmentation_dataset loads a single segmentation example correctly", {
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- coco_segmentation_dataset(root = tmp, train = FALSE, year = "2017", download = TRUE,
                               target_transform = target_transform_coco_masks)

  expect_s3_class(ds, "coco_segmentation_dataset")
  expect_gt(length(ds), 0)

  item <- ds[15]
  expect_s3_class(item, "image_with_segmentation_mask")
  y <- item$y

  expect_is(item$x, "array")
  expect_length(dim(item$x), 3)

  expect_type(y, "list")
  expect_named(y, c("labels", "iscrowd", "segmentation", "image_height", "image_width", "masks"))

  expect_tensor(y$masks)
  expect_equal(y$masks$ndim, 3)
})

test_that("coco_detection_dataset batches correctly using dataloader", {
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")


  ds <- coco_detection_dataset(root = tmp, train = FALSE, year = "2017", download = TRUE, transform = transform_to_tensor)

  dl <- dataloader(ds, batch_size = 2, collate_fn = collate_fn)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)

  expect_type(batch$x, "list")
  expect_true(all(vapply(batch$x, is_torch_tensor, logical(1))))

  expect_type(batch$y, "list")
  expect_named(batch$y[[1]], c("boxes", "labels", "area", "iscrowd"))
  expect_tensor(batch$y[[1]]$boxes)
  expect_equal(batch$y[[1]]$boxes$ndim, 2)
  expect_equal(batch$y[[1]]$boxes$size(2), 4)
})

test_that("coco_caption_dataset handles missing files gracefully", {
  expect_error(
    coco_caption_dataset(root = tempfile(), train = TRUE, download = FALSE),
    class = "rlang_error"
  )
})

test_that("coco_caption_dataset loads a single example correctly", {
  skip_if(Sys.getenv("TEST_HUGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_HUGE_DATASETS=1 to enable tests requiring huge downloads.")

  ds <- coco_caption_dataset(root = tmp, train = FALSE, download = TRUE)

  expect_s3_class(ds, "coco_caption_dataset")
  expect_gt(length(ds), 0)

  item <- ds[1]
  x <- item$x
  y <- item$y

  expect_is(x, "array")
  expect_length(dim(x), 3)
  expect_equal(dim(x)[3], 3)

  expect_type(y, "character")
  expect_gt(nchar(y), 0)
})

test_that("coco_caption_dataset batches correctly using dataloader", {
  skip_if(Sys.getenv("TEST_HUGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_HUGE_DATASETS=1 to enable tests requiring huge downloads.")

  ds <- coco_caption_dataset(root = tmp, train = FALSE, download = TRUE)

  dl <- dataloader(ds, batch_size = 2, collate_fn = collate_fn)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)

  expect_type(batch$x, "list")
  expect_true(all(vapply(batch$x, is.array, logical(1))))
  expect_true(all(vapply(batch$x, function(x) length(dim(x)) == 3, logical(1))))

  expect_type(batch$y, "list")
  expect_true(all(vapply(batch$y, is.character, logical(1))))
})

context("models-facenet")

test_that("tests for pretrained model_mtcnn", {
  model <- model_mtcnn(pretrained = TRUE)
  input <- torch_randn(1, 3, 192, 192)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out$boxes, c(1, 4))
  expect_tensor_shape(out$landmarks, c(1, 10))
  expect_tensor_shape(out$cls, c(1, 2))

  rm(model)
  gc()
})

test_that("tests for non-pretrained model_mtcnn", {
  model <- model_mtcnn(pretrained = FALSE)
  input <- torch_randn(1, 3, 224, 224)
  model$eval()
  out <- model(input)
  expect_tensor_shape(out$boxes, c(1, 4))
  expect_tensor_shape(out$landmarks, c(1, 10))
  expect_tensor_shape(out$cls, c(1, 2))

  rm(model)
  gc()
})

test_that("tests for pretrained model_facenet_pnet", {

  modelpnet <- model_facenet_pnet(pretrained = TRUE)
  modelpnet$eval()
  input <- torch_randn(1,3,160,160)
  out <- modelpnet(input)
  expect_tensor_shape(out$boxes, c(1,4,75,75))
  expect_tensor_shape(out$cls, c(1,2,75,75))

  rm(modelpnet)
  gc()
})

test_that("tests for non-pretrained model_facenet_pnet", {

  modelpnet <- model_facenet_pnet(pretrained = FALSE)
  modelpnet$eval()
  input <- torch_randn(1,3,160,160)
  out <- modelpnet(input)
  expect_tensor_shape(out$boxes, c(1,4,75,75))
  expect_tensor_shape(out$cls, c(1,2,75,75))

  rm(modelpnet)
  gc()
})

test_that("tests for pretrained model_facenet_rnet", {

  modelrnet <- model_facenet_rnet(pretrained = TRUE)
  modelrnet$eval()
  input <- torch_randn(1,3,24,24)
  out <- modelrnet(input)
  expect_tensor_shape(out$boxes, c(1,4))
  expect_tensor_shape(out$cls, c(1,2))

  rm(modelrnet)
  gc()
})

test_that("tests for non-pretrained model_facenet_rnet", {

  modelrnet <- model_facenet_rnet(pretrained = FALSE)
  modelrnet$eval()
  input <- torch_randn(1,3,24,24)
  out <- modelrnet(input)
  expect_tensor_shape(out$boxes, c(1,4))
  expect_tensor_shape(out$cls, c(1,2))

  rm(modelrnet)
  gc()
})

test_that("tests for pretrained model_facenet_onet", {

  modelonet <- model_facenet_onet(pretrained = TRUE)
  modelonet$eval()
  input <- torch_randn(1,3,48,48)
  out <- modelonet(input)
  expect_tensor_shape(out$boxes, c(1,4))
  expect_tensor_shape(out$cls, c(1,2))
  expect_tensor_shape(out$landmarks, c(1,10))

  rm(modelonet)
  gc()
})

test_that("tests for non-pretrained model_facenet_onet", {

  modelonet <- model_facenet_onet(pretrained = FALSE)
  modelonet$eval()
  input <- torch_randn(1,3,48,48)
  out <- modelonet(input)
  expect_tensor_shape(out$boxes, c(1,4))
  expect_tensor_shape(out$cls, c(1,2))
  expect_tensor_shape(out$landmarks, c(1,10))

  rm(modelonet)
  gc()
})

test_that("tests for pretrained model_facenet_inception_resnet_v1 with vgg2face weights", {

  model_vgg <- model_facenet_inception_resnet_v1(pretrained = 'vggface2')
  model_vgg$eval()
  input <- torch_randn(1,3,224,224)
  out <- model_vgg(input)
  expect_tensor_shape(out, c(1,512))

  rm(model_vgg)
  gc()
})

test_that("tests for pretrained model_facenet_inception_resnet_v1 with casia-webface weights", {

  model_casia <- model_facenet_inception_resnet_v1(pretrained = 'casia-webface')
  model_casia$eval()
  input <- torch_randn(1,3,320,260)
  out <- model_casia(input)
  expect_tensor_shape(out, c(1,512))

  rm(model_casia)
  gc()
})

test_that("tests for non-pretrained model_facenet_inception_resnet_v1", {

  model <- model_facenet_inception_resnet_v1(pretrained = NULL)
  model$eval()
  input <- torch_randn(1,3,224,224)
  out <- model(input)
  expect_tensor_shape(out, c(1,512))

  rm(model)
  gc()
})

test_that("tests for model_facenet_inception_resnet_v1 with classify=TRUE and default num_classes", {
  model <- model_facenet_inception_resnet_v1(pretrained = NULL, classify = TRUE)
  model$eval()
  input <- torch_randn(1,3,224,224)
  out <- model(input)
  expect_tensor_shape(out, c(1,10))  # Default num_classes is 10

  rm(model)
  gc()
})

test_that("tests for model_facenet_inception_resnet_v1 with classify=TRUE and custom num_classes", {
  model <- model_facenet_inception_resnet_v1(pretrained = NULL, classify = TRUE, num_classes = 100)
  model$eval()
  input <- torch_randn(1,3,224,224)
  out <- model(input)
  expect_tensor_shape(out, c(1,100))  # Custom num_classes is 100

  rm(model)
  gc()
})

test_that("tests for model_facenet_inception_resnet_v1 with batch size", {
  model <- model_facenet_inception_resnet_v1(pretrained = NULL)
  model$eval()
  input <- torch_randn(4,3,224,224)  # Batch size of 4
  out <- model(input)
  expect_tensor_shape(out, c(4,512))

  rm(model)
  gc()
})

test_that("error test for model_mtcnn with error input size", {
  model <- model_mtcnn(pretrained = FALSE)
  input <- torch_randn(1, 3, 10, 10)
  model$eval()

  expect_error(
    model(input),
    regexp = "size|dimension|shape"
  )

  rm(model)
  gc()
})

context('dataset-vggface2')

t <- withr::local_tempdir()
options(timeout = 60000)


test_that("VGGFace2 dataset works correctly for test split", {

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  vgg <- vggface2_dataset(root = t, split = "test", download = TRUE)
  expect_length(vgg, 169396)
  expect_all_true(c("img_path", "labels", "class_to_idx", "identity_df") %in% names(vgg))
  expect_is(vgg$img_path, "data.frame")
  expect_is(vgg$identity_df, "data.frame")

  first_item <- vgg[1000]
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 3)
  expect_equal(names(first_item$y), "n000029")
})
test_that("VGGFace2 dataset works correctly for train split", {

  skip_if(Sys.getenv("TEST_HUGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_HUGE_DATASETS=1 to enable tests requiring large downloads.")

  vgg <- vggface2_dataset(root = t, split = "train", download = TRUE)
  expect_length(vgg, 3141890)
  expect_all_true(c("img_path", "labels", "class_to_idx", "identity_df") %in% names(vgg))
  expect_is(vgg$img_path, "data.frame")
  expect_is(vgg$identity_df, "data.frame")

  first_item <- vgg[1]
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$x, "double")
  expect_type(first_item$y, "integer")
  expect_equal(first_item$y, 1)
})

test_that("vgg models works", {

  vggs <- list(
    model_vgg11,
    model_vgg11_bn,
    model_vgg13,
    model_vgg13_bn,
    model_vgg16,
    model_vgg16_bn,
    model_vgg19,
    model_vgg19_bn
  )

  for (m in vggs) {

    model <- m()
    expect_tensor_shape(model(torch_ones(5, 3, 224, 224)), c(5, 1000))

  }

  skip_on_ci() # unfortunatelly we don't have enough RAM on CI for that.
  #skip_on_os(os = "mac") # not downloading a bunch of files locally.
  #skip_on_os(os = "windows") # not downloading a bunch of files locally.

  for (m in vggs) {
    model <- m(pretrained = TRUE)
    expect_tensor_shape(model(torch_ones(1, 3, 224, 224)), c(1, 1000))
    gc()
  }
  unlink_model_file()
})

context("dataset-cifar")

t <- withr::local_tempdir()

test_that("cifar10", {

  expect_error(
    ds <- cifar10_dataset(root = tempfile(), train = TRUE),
    class = "runtime_error"
  )

  ds <- cifar10_dataset(root = t, train = TRUE, download = TRUE)
  expect_length(ds, 50000)
  el <- ds[1]
  expect_equal(dim(el[[1]]), c(32, 32, 3))
  expect_length(el[[2]], 1)
  expect_named(el, c("x", "y"))
  expect_equal(ds$classes[el[[2]]], "frog")

  ds <- cifar10_dataset(root = t, train = FALSE, download = TRUE)
  expect_length(ds, 10000)
  el <- ds[1]
  expect_equal(dim(el[[1]]), c(32, 32, 3))
  expect_length(el[[2]], 1)
  expect_named(el, c("x", "y"))

})

test_that("cifar100", {

  expect_error(
    ds <- cifar100_dataset(root = tempfile(), train = TRUE),
    class = "runtime_error"
  )

  ds <- cifar100_dataset(root = t, train = TRUE, download = TRUE)
  expect_length(ds, 50000)
  el <- ds[2500]
  expect_equal(dim(el[[1]]), c(32, 32, 3))
  expect_length(el[[2]], 1)
  expect_named(el, c("x", "y"))
  expect_equal(ds$classes[el[[2]]], "motorcycle")

  ds <- cifar100_dataset(root = t, train = FALSE, download = TRUE)
  expect_length(ds, 10000)
  el <- ds[502]
  expect_equal(dim(el[[1]]), c(32, 32, 3))
  expect_length(el[[2]], 1)
  expect_named(el, c("x", "y"))
  expect_equal(ds$classes[el[[2]]], "mouse")

})

context("dataset-fer")

t <- withr::local_tempdir()

test_that("tests for the FER-2013 dataset", {

  expect_error(
    ds <- fer_dataset(root = tempfile()),
    class = "runtime_error"
  )

  ds <- fer_dataset(root = t, train = TRUE, download = TRUE)
  expect_length(ds, 28709)
  first_item <- ds[1]
  expect_type(first_item$x,"integer")
  expect_length(first_item$x ,2304)
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)

  ds <- fer_dataset(root = t, train = FALSE)
  expect_length(ds, 7178)
  first_item <- ds[1]
  expect_type(first_item$x,"integer")
  expect_length(first_item$x ,2304)
  expect_named(first_item, c("x", "y"))
  expect_type(first_item$y,"integer")
  expect_equal(first_item$y, 1)

  ds2 <- fer_dataset(root = t, train = TRUE)
  dl <- torch::dataloader(ds2, batch_size = 32)
  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)
  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_length(batch$x , 73728)
  expect_tensor_shape(batch$x, c(32, 48, 48))
  expect_tensor_dtype(batch$x , torch_long())
  expect_tensor(batch$y)
  expect_tensor_shape(batch$y, 32)
  expect_tensor_dtype(batch$y , torch_long())
  expect_equal_to_r(batch$y[1], 1)
  expect_equal_to_r(batch$y[3], 3)
  expect_equal_to_r(batch$y[32], 7)

})

context("dataset-eurosat")

temp_root <- withr::local_tempdir()

test_that("eurosat_dataset downloads correctly whatever the split", {
  skip_on_cran()
  skip_if_not_installed("torch")


  expect_error(
    eurosat_dataset(root = tempfile(), split = "test", download = FALSE),
    "Dataset not found. You can use `download = TRUE`",
    label = "Dataset should fail if not previously downloaded"
  )

  expect_no_error(
    ds <- eurosat_dataset(root = temp_root, split = "train", download = TRUE)
  )

  expect_is(ds, "dataset", "train should be a dataset")

  extracted_dir <- file.path(temp_root, "eurosat", "images", "2750")
  extracted_dir <- list.dirs(extracted_dir, recursive = FALSE, full.names = TRUE)
  # Extracted data folder have one folder per category
  expect_length(extracted_dir, 10)

  image_files <- list.files(extracted_dir, pattern = "\\.jpg$", recursive = TRUE, full.names = TRUE)
  expect_gte(length(image_files), 16200, "Image files should be present in the extracted directory")

  train_ds <- eurosat_dataset(root = temp_root, split = "train", download = TRUE)
  # Train dataset should have exactly 16200 samples and reuse existing folder
  expect_length(train_ds, 16200)

  val_ds <- eurosat_dataset(root = temp_root, split = "val", download = TRUE)
  # Validation dataset should have exactly 5400 samples
  expect_length(val_ds, 5400)

  test_ds <- eurosat_dataset(root = temp_root, split = "test", download = TRUE)
  # Test dataset should have exactly 5400 samples
  expect_length(test_ds, 5400)

})

test_that("dataloader from eurosat_dataset gets torch tensors", {
  skip_on_cran()
  skip_if_not_installed("torch")

  ds <- eurosat_dataset(root = temp_root, split = "train", download = FALSE, transform = transform_to_tensor)
  dl <- torch::dataloader(ds, batch_size = 10)
  # 16,2k turns into 1620 batches of 10
  expect_length(dl, 1620)
  iter <- dataloader_make_iter(dl)
  i <- dataloader_next(iter)
  # Check shape, dtype, and values on X
  expect_tensor_shape(i[[1]], c(10, 3, 64, 64))
  expect_tensor_dtype(i[[1]], torch_float())
  expect_true((torch_max(i[[1]]) <= 1)$item())
  # Check shape, dtype and names on y
  expect_length(i[[2]],10)
  expect_named(i, c("x", "y"))})


test_that("eurosat100_dataset derivatives download and prepare correctly", {
  skip_on_cran()
  skip_if_not_installed("torch")


  expect_error(
    eurosat100_dataset(root = tempfile(), split = "test", download = FALSE),
    "Dataset not found. You can use `download = TRUE`",
    label = "Dataset should fail if not previously downloaded"
  )

  expect_no_error(
    ds_100 <- eurosat100_dataset(root = temp_root, split = "val", download = TRUE)
  )

  dl <- torch::dataloader(ds_100, batch_size = 10)
  # 20 turns into 2 batches of 10
  expect_length(dl, 2)
  iter <- dataloader_make_iter(dl)
  i <- dataloader_next(iter)

  # Check shape, dtype, and values on X
  expect_tensor_shape(i[[1]], c(10, 13, 64, 64))
  expect_tensor_dtype(i[[1]], torch_float())
  expect_true((torch_max(i[[1]]) <= 1)$item())
  # Check shape, dtype and names on y
  expect_length(i[[2]], 10)
  expect_named(i, c("x", "y"))})

test_that("eurosat_all_bands_dataset derivatives download and prepare correctly", {
  skip_on_cran()
  skip_if_not_installed("torch")

  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
        "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  expect_error(
    eurosat_all_bands_dataset(root = tempfile(), split = "test", download = FALSE),
    "Dataset not found. You can use `download = TRUE`",
    label = "Dataset should fail if not previously downloaded"
  )

  expect_no_error(
    ds_all <- eurosat_all_bands_dataset(root = temp_root, split = "val", download = TRUE),
    label = "Dataset should load successfully"
  )
  dl <- torch::dataloader(ds_all, batch_size = 10)
  # 5400 turns into 540 batches of 10
  expect_length(dl, 540)
  iter <- dataloader_make_iter(dl)
  i <- dataloader_next(iter)
  # Check shape, dtype, and values on X
  expect_tensor_shape(i[[1]], c(10, 13, 64, 64))
  expect_tensor_dtype(i[[1]], torch_float())
  expect_true((torch_max(i[[1]]) <= 1)$item())
  # Check shape, dtype and names on y
  expect_length(i[[2]], 10)
  expect_named(i, c("x", "y"))})

library(torch)

is_torch_tensor <- function(x) {
  inherits(x, "torch_tensor")
}

expect_no_error <- function(object, ...) {
  expect_error(object, NA, ...)
}

expect_tensor_shape <- function(object, expected) {
  expect_tensor(object)
  expect_equal(object$shape, expected)
}

expect_tensor_dtype <- function(object, expected_dtype) {
  expect_tensor(object)
  expect_true(object$dtype == expected_dtype)
}

expect_tensor <- function(object) {
  expect_true(is_torch_tensor(object))
  expect_no_error(torch::as_array(object))
}

expect_equal_to_r <- function(object, expected, ...) {
  expect_equal(torch::as_array(object), expected, ...)
}

unlink_model_file <- function() {
  cache_path <- rappdirs::user_cache_dir("torch")
  model_file <- list.files(cache_path, pattern = "*.pth", full.names = TRUE)
  unlink(model_file)
}

expect_bbox_is_xyxy <- function(object, width, height) {
  expect_tensor(object)
  N <- object$shape[1]
  expect_tensor_shape(object, c(N, 4))

  x_min <- object[, 1]
  y_min <- object[, 2]
  x_max <- object[, 3]
  y_max <- object[, 4]

  ## bbox range checks
  expect_true((x_min >= 0)$all()$item(),
              info = "All x_min values must be >= 0.")
  expect_true((y_min >= 0)$all()$item(),
              info = "All y_min values must be >= 0.")
  expect_true((x_max <= torch_tensor(width))$all()$item(),
              info = sprintf("All x_max values must be <= width (%s).", width))
  expect_true((y_max <= torch_tensor(height))$all()$item(),
              info = sprintf("All y_max values must be <= height (%s).", height))
  expect_true((x_max > torch_tensor(1))$all()$item(),
              info = "x looks like a relative delta and shall be converted back to image width.")
  expect_true((y_max > torch_tensor(1))$all()$item(),
              info = "y looks like a relative delta and shall be converted back to image height.")
  ## bbox ordering checks
  expect_true((x_min <= x_max)$all()$item(),
              info = "Each x_min must be smaller than its x_max.")
  expect_true((y_min <= y_max)$all()$item(),
              info = "Each y_min must be smaller than its y_max.")

}

context("dataset-rf100-peixos")

t <- withr::local_tempdir()

test_that("rf100_peixos_segmentation_dataset handles missing files gracefully", {
  expect_error(
    rf100_peixos_segmentation_dataset(split = "train", root = t, download = FALSE),
    class = "runtime_error"
  )
})

test_that("rf100_peixos_segmentation_dataset 'test' split works", {
  # windows error on tar deep folder decompression on tempdir()
  skip_on_os("windows")
  expect_no_error(
    ds <- rf100_peixos_segmentation_dataset(split = "test", root = t, download = TRUE)
  )
  expect_length(ds, 118)
  item <- ds[1]
  expect_tensor(item$y$masks)
  expect_equal(item$y$masks$ndim, 3)
  expect_gt(item$y$masks$sum()$item(), 0)
})

test_that("rf100_peixos_segmentation_dataset 'val' split works", {
  # windows error on tar deep folder decompression on tempdir()
  skip_on_os("windows")
  expect_no_error(
    ds <- rf100_peixos_segmentation_dataset(split = "val", root = t, download = TRUE)
  )
  expect_length(ds, 251)
  item <- ds[1]
  expect_tensor(item$y$masks)
  expect_equal(item$y$masks$ndim, 3)
  expect_gt(item$y$masks$sum()$item(), 0)
})


test_that(paste0("rf100_peixos_segmentation_dataset loads 'train' split correctly"), {
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")
  # windows error on tar deep folder decompression on tempdir()
  skip_on_os("windows")
  ds <- rf100_peixos_segmentation_dataset(split = "train", root = t, download = TRUE)

  expect_s3_class(ds, "rf100_peixos_segmentation_dataset")
  expect_length(ds, 821)

  item <- ds[1]

  expect_type(item$y, "list")
  expect_named(item$y, c("masks", "labels"))
  expect_tensor(item$y$masks)
  expect_equal(item$y$masks$ndim, 3)
  expect_gt(item$y$masks$sum()$item(), 0)
  expect_equal(item$y$labels, 1L)
  expect_s3_class(item, "image_with_segmentation_mask")
})

test_that("places365_dataset loads and returns valid items", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- places365_dataset(
    split = "val",
    download = TRUE,
    transform = transform_to_tensor
  )

  expect_s3_class(ds, "places365_dataset")
  expect_length(ds, 36500)

  item <- ds[1]

  expect_type(item, "list")
  expect_named(item, c("x", "y"))

  # Check image
  expect_tensor(item$x)
  expect_equal(item$x$ndim, 3)
  expect_equal(item$x$shape[1], 3)
  expect_gt(item$x$shape[2], 0)
  expect_gt(item$x$shape[3], 0)
  expect_tensor_dtype(item$x, torch::torch_float())
  expect_gte(torch::torch_min(item$x)$item(), 0)
  expect_lte(torch::torch_max(item$x)$item(), 1)

  # Check label
  expect_type(item$y, "integer")
  expect_gte(item$y, 1)
  expect_lte(item$y, 365)
})

test_that("places365_dataset train split loads an item", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- places365_dataset(
    split = "train",
    download = TRUE,
    transform = transform_to_tensor
  )

  expect_gt(length(ds), 0)

  item <- ds[1]
  expect_type(item, "list")
  expect_named(item, c("x", "y"))
  expect_tensor(item$x)
  expect_tensor_dtype(item$x, torch::torch_float())
  expect_type(item$y, "integer")
  expect_gte(item$y, 1)
  expect_lte(item$y, 365)
})

test_that("places365 categories are correctly mapped", {
  ann_path <- file.path(tempdir(), "places365", "categories_places365.txt")
  skip_if_not(file.exists(ann_path), message = "Category file missing")

  categories <- readLines(ann_path, warn = FALSE)
  expect_length(categories, 365)

  ds <- places365_dataset(
    split = "val",
    download = FALSE,
    transform = transform_to_tensor
  )

  label <- ds[1]$y
  expect_type(label, "integer")
  expect_lte(label, length(categories))
})

test_that("places365_dataset test split returns image only", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- places365_dataset(
    split = "test",
    download = TRUE,
    transform = transform_to_tensor
  )

  expect_gt(length(ds), 0)

  item <- ds[1]
  expect_type(item, "list")
  expect_named(item, c("x", "y"))
  expect_tensor(item$x)
  expect_tensor_dtype(item$x, torch::torch_float())
  expect_true(is.na(item$y))
})


# Tests for the high-resolution Places365 variant

test_that("places365_dataset_large loads and returns valid items", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- places365_dataset_large(
    split = "val",
    download = TRUE,
    transform = transform_to_tensor
  )

  expect_s3_class(ds, "places365_dataset_large")
  expect_length(ds, 36500)

  item <- ds[1]

  expect_type(item, "list")
  expect_named(item, c("x", "y"))

  # Check image tensor has 3 channels and positive dims
  expect_tensor(item$x)
  expect_equal(item$x$ndim, 3)
  expect_equal(item$x$shape[1], 3)
  expect_gt(item$x$shape[2], 0)
  expect_gt(item$x$shape[3], 0)
  expect_tensor_dtype(item$x, torch::torch_float())
  expect_gte(torch::torch_min(item$x)$item(), 0)
  expect_lte(torch::torch_max(item$x)$item(), 1)

  # Check label range
  expect_type(item$y, "integer")
  expect_gte(item$y, 1)
  expect_lte(item$y, 365)
})

test_that("places365_dataset_large train split loads an item", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- places365_dataset_large(
    split = "train",
    download = TRUE,
    transform = transform_to_tensor
  )

  expect_gt(length(ds), 0)

  item <- ds[1]
  expect_type(item, "list")
  expect_named(item, c("x", "y"))
  expect_tensor(item$x)
  expect_tensor_dtype(item$x, torch::torch_float())
  expect_type(item$y, "integer")
  expect_gte(item$y, 1)
  expect_lte(item$y, 365)
})


test_that("places365_dataset_large test split returns image only", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- places365_dataset_large(
    split = "test",
    download = TRUE,
    transform = transform_to_tensor
  )

  expect_gt(length(ds), 0)

  item <- ds[1]
  expect_type(item, "list")
  expect_named(item, c("x", "y"))
  expect_tensor(item$x)
  expect_tensor_dtype(item$x, torch::torch_float())
  expect_true(is.na(item$y))
})


test_that("places365_dataset_large can be batched with resize_collate_fn", {
  skip_on_cran()
  skip_if(Sys.getenv("TEST_LARGE_DATASETS", unset = 0) != 1,
          "Skipping test: set TEST_LARGE_DATASETS=1 to enable tests requiring large downloads.")

  ds <- places365_dataset_large(
    split = "val",
    download = TRUE,
    transform = transform_to_tensor
  )

  dl <- torch::dataloader(
    dataset = ds,
    batch_size = 2,
    collate_fn = torchvision:::resize_collate_fn(c(256, 256))
  )

  iter <- dataloader_make_iter(dl)
  batch <- dataloader_next(iter)

  expect_named(batch, c("x", "y"))
  expect_tensor(batch$x)
  expect_tensor_shape(batch$x, c(2, 3, 256, 256))
  expect_tensor_dtype(batch$x, torch::torch_float())
  expect_tensor(batch$y)
  expect_tensor_shape(batch$y, 2)
  expect_tensor_dtype(batch$y, torch::torch_long())
})

set -ex



pip check
python test/smoke_test.py
pytest --verbose --durations=50 -k "not (test_internet or test_detection_model or test_quantized_classification_model or test_forward_backward or test_jit_forward_backward)" test/test_image.py
pytest --verbose --durations=50 -k "not (test_internet or test_detection_model or test_quantized_classification_model or test_forward_backward or test_jit_forward_backward)" test/test_transforms_v2.py
pytest --verbose --durations=50 -k "not (test_internet or test_detection_model or test_quantized_classification_model or test_forward_backward or test_jit_forward_backward)" test/test_datasets.py
pytest --verbose --durations=50 -k "not (test_internet or test_detection_model or test_quantized_classification_model or test_forward_backward or test_jit_forward_backward)" test/test_ops.py
exit 0

@echo on

set VC_VERSION_LOWER=17
set VC_VERSION_UPPER=18
if "%VC_YEAR%" == "2019" (
    set VC_VERSION_LOWER=16
    set VC_VERSION_UPPER=17
)
if "%VC_YEAR%" == "2017" (
    set VC_VERSION_LOWER=15
    set VC_VERSION_UPPER=16
)

for /f "usebackq tokens=*" %%i in (`"%ProgramFiles(x86)%\Microsoft Visual Studio\Installer\vswhere.exe" -legacy -products * -version [%VC_VERSION_LOWER%^,%VC_VERSION_UPPER%^) -property installationPath`) do (
    if exist "%%i" if exist "%%i\VC\Auxiliary\Build\vcvarsall.bat" (
        set "VS15INSTALLDIR=%%i"
        set "VS15VCVARSALL=%%i\VC\Auxiliary\Build\vcvarsall.bat"
        goto vswhere
    )
)

:vswhere
if "%VSDEVCMD_ARGS%" == "" (
    call "%VS15VCVARSALL%" x64 || exit /b 1
) else (
    call "%VS15VCVARSALL%" x64 %VSDEVCMD_ARGS% || exit /b 1
)

@echo on

set DISTUTILS_USE_SDK=1

set args=%1
shift
:start
if [%1] == [] goto done
set args=%args% %1
shift
goto start

:done
if "%args%" == "" (
    echo Usage: vc_env_helper.bat [command] [args]
    echo e.g. vc_env_helper.bat cl /c test.cpp
)

%args% || exit /b 1

@echo On

set TORCHVISION_INCLUDE=%LIBRARY_INC%
:: point the build system towards the .lib files
set LIB=%PREFIX%\\Lib\\site-packages\\torch\\lib;%LIB%
%PYTHON% -m pip install . -vv --no-deps --no-build-isolation

:: Set env vars that tell distutils to use the compiler that we put on path
SET DISTUTILS_USE_SDK=1
SET MSSdk=1

SET "VS_VERSION=16.0"
SET "VS_MAJOR=16"
SET "VS_YEAR=2019"

set "MSYS2_ARG_CONV_EXCL=/AI;/AL;/OUT;/out"
set "MSYS2_ENV_CONV_EXCL=CL"

:: For Python 3.5+, ensure that we link with the dynamic runtime.  See
:: http://stevedower.id.au/blog/building-for-python-3-5-part-two/ for more info
set "PY_VCRUNTIME_REDIST=%PREFIX%\\bin\\vcruntime140.dll"

for /f "usebackq tokens=*" %%i in (`"%ProgramFiles(x86)%\Microsoft Visual Studio\Installer\vswhere.exe" -legacy -products * -version [16^,17^) -property installationPath`) do (
    if exist "%%i" if exist "%%i\VC\Auxiliary\Build\vcvarsall.bat" (
        set "VSINSTALLDIR=%%i\"
        goto :vswhere
    )
)

:vswhere

:: Shorten PATH to avoid the `input line too long` error.
SET MyPath=%PATH%

setlocal EnableDelayedExpansion

SET TempPath="%MyPath:;=";"%"
SET var=
FOR %%a IN (%TempPath%) DO (
    IF EXIST %%~sa (
        SET "var=!var!;%%~sa"
    )
)

set "TempPath=!var:~1!"
endlocal & set "PATH=%TempPath%"

:: Shorten current directory too
FOR %%A IN (.) DO CD "%%~sA"

:: other things added by install_activate.bat at package build time

#!/bin/bash
set -ex -o pipefail

echo ""
echo "DIR: $(pwd)"

TEST_APP_PATH=$(dirname $(realpath $0))
cd ${TEST_APP_PATH}

PYTORCH_IOS_NIGHTLY_NAME=libtorch_ios_nightly_build.zip
VISION_IOS_NIGHTLY_NAME=libtorchvision_ops_ios_nightly_build.zip

echo "Downloading torch libs and vision libs..."
wget https://ossci-ios-build.s3.amazonaws.com/${PYTORCH_IOS_NIGHTLY_NAME}
wget https://ossci-ios-build.s3.amazonaws.com/${VISION_IOS_NIGHTLY_NAME}

mkdir -p ./library/torch
mkdir -p ./library/vision

echo "Unziping torch libs and vision libs..."
unzip -d ./library/torch ./${PYTORCH_IOS_NIGHTLY_NAME}
unzip -d ./library/vision ./${VISION_IOS_NIGHTLY_NAME}

cp ./library/vision/install/lib/*.a ./library/torch/install/lib
cp -r ./library/torch/install .

rm -rf ./library
rm -rf ./*.zip

echo "Generating the vision model..."
python ./make_assets.py

echo "Finished project setups."

#!/usr/bin/env bash

set -euo pipefail

./.github/scripts/setup-env.sh

# Activate conda environment
eval "$($(which conda) shell.bash hook)" && conda deactivate && conda activate ci

echo '::group::Install testing utilities'
# TODO: remove the <8 constraint on pytest when https://github.com/pytorch/vision/issues/8238 is closed
pip install --progress-bar=off "pytest<8" pytest-mock pytest-cov expecttest!=0.2.0 requests
echo '::endgroup::'

python test/smoke_test.py

# We explicitly ignore the video tests until we resolve https://github.com/pytorch/vision/issues/8162
pytest --ignore-glob="*test_video*" --junit-xml="${RUNNER_TEST_RESULTS_DIR}/test-results.xml" -v --durations=25

set -ex

if [[ "$cuda_compiler_version" == "None" ]]; then
  export FORCE_CUDA=0
else
  # +PTX should go to the oldest arch. There's a modest runtime performance
  # hit for (unlisted) newer arches on doing this, but the benefit is wider
  # compatibility.
  if [[ ${cuda_compiler_version} == 11.2* ]]; then
      export TORCH_CUDA_ARCH_LIST="3.5+PTX;5.0;6.0;6.1;7.0;7.5;8.0;8.6"
  elif [[ ${cuda_compiler_version} == 11.8 ]]; then
      export TORCH_CUDA_ARCH_LIST="3.5+PTX;5.0;6.0;6.1;7.0;7.5;8.0;8.6;9.0"
      export CUDA_TOOLKIT_ROOT_DIR=$CUDA_HOME
  elif [[ ${cuda_compiler_version} == 12.0 ]]; then
      export TORCH_CUDA_ARCH_LIST="5.0+PTX;6.0;6.1;7.0;7.5;8.0;8.6;9.0"
      # $CUDA_HOME not set in CUDA 12.0. Using $PREFIX
      export CUDA_TOOLKIT_ROOT_DIR="${PREFIX}"
  else
      echo "unsupported cuda version. edit build.sh"
      exit 1
  fi
  export FORCE_CUDA=1
fi

# remove pyproject.toml
rm -f pyproject.toml

export TORCHVISION_INCLUDE="${PREFIX}/include/"
${PYTHON} -m pip install . -vv --no-deps --no-build-isolation

#!/bin/bash

if [[ "$(uname)" == Darwin ]]; then
  # Uninstall Conflicting jpeg brew formulae
  jpeg_packages=$(brew list | grep jpeg)
  echo "Existing Jpeg-related Brew libraries"
  echo $jpeg_packages
  for pkg in $jpeg_packages; do
    brew uninstall --ignore-dependencies --force $pkg || true
  done

  conda install -yq wget
fi

if [[ "$(uname)" == Darwin || "$OSTYPE" == "msys" ]]; then
  conda install libpng libwebp -yq
  # Installing webp also installs a non-turbo jpeg, so we uninstall jpeg stuff
  # before re-installing them
  conda uninstall libjpeg-turbo libjpeg -y
  conda install -yq ffmpeg=4.2 libjpeg-turbo -c pytorch

  # Copy binaries to be included in the wheel distribution
  if [[ "$OSTYPE" == "msys" ]]; then
      python_exec="$(which python)"
      bin_path=$(dirname $python_exec)
      cp "$bin_path/Library/bin/libjpeg.dll" torchvision
  fi
else

  if [[ "$ARCH" == "aarch64" ]]; then
    conda install libpng -yq
    conda install -yq ffmpeg=4.2 libjpeg-turbo -c pytorch-nightly
  fi

  yum install -y libjpeg-turbo-devel libwebp-devel freetype gnutls
  pip install auditwheel
fi

pip install numpy pyyaml future ninja
pip install --upgrade setuptools==72.1.0

@echo on
set CL=/I"C:\Program Files (x86)\torchvision\include"
msbuild "-p:Configuration=Release" "-p:BuildInParallel=true" "-p:MultiProcessorCompilation=true" "-p:CL_MPCount=%1" run_model.vcxproj -maxcpucount:%1

#!/usr/bin/env bash

set -euxo pipefail

# Prepare conda
set +x && eval "$($(which conda) shell.bash hook)" && set -x

# Setup the OS_TYPE environment variable that should be used for conditions involving the OS below.
case $(uname) in
  Linux)
    OS_TYPE=linux
    ;;
  Darwin)
    OS_TYPE=macos
    ;;
  MSYS*)
    OS_TYPE=windows
    ;;
  *)
    echo "Unknown OS type:" $(uname)
    exit 1
    ;;
esac

echo '::group::Create build environment'
# See https://github.com/pytorch/vision/issues/7296 for ffmpeg
conda create \
  --name ci \
  --quiet --yes \
  python="${PYTHON_VERSION}" pip \
  ninja cmake \
  libpng \
  libwebp \
  'ffmpeg<4.3'
conda activate ci
conda install --quiet --yes libjpeg-turbo -c pytorch
pip install --progress-bar=off --upgrade setuptools==72.1.0

# See https://github.com/pytorch/vision/issues/6790
if [[ "${PYTHON_VERSION}" != "3.11" ]]; then
  pip install --progress-bar=off av!=10.0.0
fi

echo '::endgroup::'

if [[ "${OS_TYPE}" == windows && "${GPU_ARCH_TYPE}" == cuda ]]; then
  echo '::group::Install VisualStudio CUDA extensions on Windows'
  if [[ "${VC_YEAR:-}" == "2022" ]]; then
    TARGET_DIR="/c/Program Files (x86)/Microsoft Visual Studio/2022/BuildTools/MSBuild/Microsoft/VC/v170/BuildCustomizations"
  else
    TARGET_DIR="/c/Program Files (x86)/Microsoft Visual Studio/2019/BuildTools/MSBuild/Microsoft/VC/v160/BuildCustomizations"
  fi
  mkdir -p "${TARGET_DIR}"
  cp -r "${CUDA_HOME}/MSBuildExtensions/"* "${TARGET_DIR}"
  echo '::endgroup::'
fi

echo '::group::Install PyTorch'
# TODO: Can we maybe have this as environment variable in the job template? For example, `IS_RELEASE`.
if [[ (${GITHUB_EVENT_NAME} = 'pull_request' && (${GITHUB_BASE_REF} = 'release'*)) || (${GITHUB_REF} = 'refs/heads/release'*) ]]; then
  CHANNEL=test
else
  CHANNEL=nightly
fi

case $GPU_ARCH_TYPE in
  cpu)
    GPU_ARCH_ID="cpu"
    ;;
  cuda)
    VERSION_WITHOUT_DOT=$(echo "${GPU_ARCH_VERSION}" | sed 's/\.//')
    GPU_ARCH_ID="cu${VERSION_WITHOUT_DOT}"
    ;;
  *)
    echo "Unknown GPU_ARCH_TYPE=${GPU_ARCH_TYPE}"
    exit 1
    ;;
esac
PYTORCH_WHEEL_INDEX="https://download.pytorch.org/whl/${CHANNEL}/${GPU_ARCH_ID}"
pip install --progress-bar=off --pre torch --index-url="${PYTORCH_WHEEL_INDEX}"

if [[ $GPU_ARCH_TYPE == 'cuda' ]]; then
  python -c "import torch; exit(not torch.cuda.is_available())"
fi
echo '::endgroup::'

echo '::group::Install third party dependencies prior to TorchVision install'
# Installing with `easy_install`, e.g. `python setup.py install` or `python setup.py develop`, has some quirks when
# when pulling in third-party dependencies. For example:
# - On Windows, we often hit an SSL error although `pip` can install just fine.
# - It happily pulls in pre-releases, which can lead to more problems down the line.
#   `pip` does not unless explicitly told to do so.
# Thus, we use `easy_install` to extract the third-party dependencies here and install them upfront with `pip`.
python setup.py egg_info
# The requires.txt cannot be used with `pip install -r` directly. The requirements are listed at the top and the
# optional dependencies come in non-standard syntax after a blank line. Thus, we just extract the header.
sed -e '/^$/,$d' *.egg-info/requires.txt | tee requirements.txt
pip install --progress-bar=off -r requirements.txt
echo '::endgroup::'

echo '::group::Install TorchVision'
python setup.py develop
echo '::endgroup::'

echo '::group::Collect environment information'
conda list
python -m torch.utils.collect_env
echo '::endgroup::'

#!/usr/bin/env bash

set -euxo pipefail

./.github/scripts/setup-env.sh

# Activate conda environment
set +x && eval "$($(which conda) shell.bash hook)" && conda deactivate && conda activate ci && set -x

# Setup the OS_TYPE environment variable that should be used for conditions involving the OS below.
case $(uname) in
  Linux)
    OS_TYPE=linux
    ;;
  Darwin)
    OS_TYPE=macos
    ;;
  MSYS*)
    OS_TYPE=windows
    ;;
  *)
    echo "Unknown OS type:" $(uname)
    exit 1
    ;;
esac

if [[ $OS_TYPE == macos ]]; then
  JOBS=$(sysctl -n hw.logicalcpu)
else
  JOBS=$(nproc)
fi

TORCH_PATH=$(python -c "import pathlib, torch; print(pathlib.Path(torch.__path__[0]))")
if [[ $OS_TYPE == windows ]]; then
  PACKAGING_DIR="${PWD}/packaging"
  export PATH="${TORCH_PATH}/lib:${PATH}"
fi

Torch_DIR="${TORCH_PATH}/share/cmake/Torch"
if [[ "${GPU_ARCH_TYPE}" == "cuda" ]]; then
  WITH_CUDA=1
else
  WITH_CUDA=0
fi

echo '::group::Prepare CMake builds'
mkdir -p cpp_build

pushd examples/cpp
python script_model.py
mkdir -p build
mv resnet18.pt fasterrcnn_resnet50_fpn.pt build
popd

# This was only needed for the tracing above
pip uninstall -y torchvision
echo '::endgroup::'

echo '::group::Build and install libtorchvision'
pushd cpp_build


# On macOS, CMake is looking for the library (*.dylib) and the header (*.h) separately. By default, it prefers to load
# the header from other packages that install the library. This easily leads to a mismatch if the library installed
# from conda doesn't have the exact same version. Thus, we need to explicitly set CMAKE_FIND_FRAMEWORK=NEVER to force
# it to not load anything from other installed frameworks. Resources:
# https://stackoverflow.com/questions/36523911/osx-homebrew-cmake-libpng-version-mismatch-issue
# https://cmake.org/cmake/help/latest/variable/CMAKE_FIND_FRAMEWORK.html
cmake .. -DTorch_DIR="${Torch_DIR}" -DWITH_CUDA="${WITH_CUDA}" \
  -DCMAKE_PREFIX_PATH="${CONDA_PREFIX}" \
  -DCMAKE_FIND_FRAMEWORK=NEVER \
  -DCMAKE_INSTALL_PREFIX="${CONDA_PREFIX}"
if [[ $OS_TYPE == windows ]]; then
  "${PACKAGING_DIR}/windows/internal/vc_env_helper.bat" "${PACKAGING_DIR}/windows/internal/build_cmake.bat" $JOBS
else
  make -j$JOBS
  make install
fi

popd
echo '::endgroup::'

echo '::group::Build and run C++ example'
pushd examples/cpp/build

cmake .. -DTorch_DIR="${Torch_DIR}" \
  -DCMAKE_PREFIX_PATH="${CONDA_PREFIX}" \
  -DCMAKE_FIND_FRAMEWORK=NEVER \
  -DUSE_TORCHVISION=ON  # Needed for faster-rcnn since it's using torchvision ops like NMS.
if [[ $OS_TYPE == windows ]]; then
  "${PACKAGING_DIR}/windows/internal/vc_env_helper.bat" "${PACKAGING_DIR}/windows/internal/build_cpp_example.bat" $JOBS
  cd Release
  cp ../resnet18.pt .
  cp ../fasterrcnn_resnet50_fpn.pt .
else
  make -j$JOBS
fi

./run_model resnet18.pt
./run_model fasterrcnn_resnet50_fpn.pt

popd
echo '::endgroup::'

#!/bin/bash

if [ -z $1 ]
then
    echo "Commit hash is required to be passed when running this script."
    echo "./fbcode_to_main_sync.sh <commit_hash> <fork_name> <fork_main_branch>"
    exit 1
fi
commit_hash=$1

if [ -z $2 ]
then
    echo "Fork name is required to be passed when running this script."
    echo "./fbcode_to_main_sync.sh <commit_hash> <fork_name> <fork_main_branch>"
    exit 1
fi
fork_name=$2

if [ -z $3 ]
then
    fork_main_branch="main"
else
    fork_main_branch=$3
fi

from_branch="fbsync"
git stash
git checkout $from_branch
git pull
# Add random prefix in the new branch name to keep it unique per run
prefix=$RANDOM
IFS='
'
for line in $(git log --pretty=oneline "$commit_hash"..HEAD)
do
    if [[ $line != *\[fbsync\]* ]]
    then
        echo "Parsing $line"
        hash=$(echo $line | cut -f1 -d' ')
        git checkout $fork_main_branch
        git checkout -B cherrypick_${prefix}_${hash}
        git cherry-pick -x "$hash"
        git push $fork_name cherrypick_${prefix}_${hash}
        git checkout $from_branch
    fi
done
echo "Please review the PRs, add [FBCode->GH] prefix in the title and publish them."

#!/usr/bin/env bash
#
# Usage (run from root of project):
# TEST_INFRA_BRANCH=release/2.1 RELEASE_BRANCH=release/2.1 RELEASE_VERSION=2.1.0 packaging/cut_release.sh
#
# TEST_INFRA_BRANCH: The release branch of test-infra that houses all reusable
# workflows
#
# RELEASE_BRANCH: The name of the release branch for this repo
#
# RELEASE_VERSION: Version of this current release

set -eou pipefail

# Create and Check out to Release Branch
git checkout -b "${RELEASE_BRANCH}"

# Change all GitHub Actions to reference the test-infra release branch
# as opposed to main.
for i in .github/workflows/*.yml; do 
  if [[ "$OSTYPE" == "darwin"* ]]; then
    sed -i '' -e s#@main#@"${TEST_INFRA_BRANCH}"# $i;
    sed -i '' -e s#test-infra-ref:[[:space:]]main#"test-infra-ref: ${TEST_INFRA_BRANCH}"# $i;
  else
    sed -i -e s#@main#@"${TEST_INFRA_BRANCH}"# $i;
    sed -i -e s#test-infra-ref:[[:space:]]main#"test-infra-ref: ${TEST_INFRA_BRANCH}"# $i;
  fi
done

# Update the Release Version in version.txt
echo "${RELEASE_VERSION}" >version.txt

# Optional
# git add ./github/workflows/*.yml version.txt
# git commit -m "[RELEASE-ONLY CHANGES] Branch Cut for Release {RELEASE_VERSION}"
# git push origin "${RELEASE_BRANCH}"

#!/bin/bash

set -ex -o pipefail
echo ""
echo "DIR: $(pwd)"
VISION_IOS_ROOT=$(dirname $(realpath $0))

if ! [ -n "${LIBTORCH_HEADER_ROOT:-}" ]; then
  echo "Missing parameter: LIBTORCH_HEADER_ROOT"
  exit 1
fi

if [ -n "${IOS_ARCH:-}" ]; then
  if [ "${IOS_ARCH:-}" == "arm64" ]; then
    IOS_PLATFORM="OS"
  elif [ "${IOS_ARCH:-}" == "x86_64" ]; then
    IOS_PLATFORM="SIMULATOR"
  fi
fi

mkdir -p ${VISION_IOS_ROOT}/lib
mkdir -p ${VISION_IOS_ROOT}/build
cd ${VISION_IOS_ROOT}/build
cmake -DLIBTORCH_HEADER_ROOT=${LIBTORCH_HEADER_ROOT}  \
      -DCMAKE_TOOLCHAIN_FILE=${VISION_IOS_ROOT}/../cmake/iOS.cmake \
      -DIOS_ARCH=${IOS_ARCH} \
      -DIOS_PLATFORM=${IOS_PLATFORM} \
      ..
make
rm -rf ${VISION_IOS_ROOT}/build

set YEAR=2019
set VER=16

mkdir "%PREFIX%\etc\conda\activate.d"
COPY "%RECIPE_DIR%\activate.bat" "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"

IF "%cross_compiler_target_platform%" == "win-64" (
  set "target_platform=amd64"
  echo SET "CMAKE_GENERATOR=Visual Studio %VER% %YEAR% Win64" >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  echo pushd "%%VSINSTALLDIR%%" >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  IF "%VSDEVCMD_ARGS%" == "" (
    echo CALL "VC\Auxiliary\Build\vcvarsall.bat" x64 >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
    echo popd >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
    echo pushd "%%VSINSTALLDIR%%" >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
    echo CALL "VC\Auxiliary\Build\vcvarsall.bat" x86_amd64 >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  ) ELSE (
    echo CALL "VC\Auxiliary\Build\vcvarsall.bat" x64 %VSDEVCMD_ARGS% >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
    echo popd >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
    echo pushd "%%VSINSTALLDIR%%" >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
    echo CALL "VC\Auxiliary\Build\vcvarsall.bat" x86_amd64 %VSDEVCMD_ARGS% >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  )
  echo popd >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  ) else (
  set "target_platform=x86"
  echo SET "CMAKE_GENERATOR=Visual Studio %VER% %YEAR%" >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  echo pushd "%%VSINSTALLDIR%%" >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  echo CALL "VC\Auxiliary\Build\vcvars32.bat" >> "%PREFIX%\etc\conda\activate.d\vs%YEAR%_compiler_vars.bat"
  echo popd
  )

@echo on
msbuild "-p:Configuration=Release" "-p:BuildInParallel=true" "-p:MultiProcessorCompilation=true" "-p:CL_MPCount=%1" torchvision.vcxproj -maxcpucount:%1
msbuild "-p:Configuration=Release" "-p:BuildInParallel=true" "-p:MultiProcessorCompilation=true" "-p:CL_MPCount=%1" INSTALL.vcxproj -maxcpucount:%1

---
title: "RF100 Dataset Catalog"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{RF100 Dataset Catalog}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

The RoboFlow 100 (RF100) benchmark consists of 34 diverse object detection datasets organized into 6 collections. This vignette provides a comprehensive catalog to help you find the right dataset for your task.

The RF100 datasets cover a wide range of domains including:

- **Biology**: Microscopy, cells, bacteria, parasites (9 datasets)
- **Medical**: X-rays, MRI, pathology (8 datasets)  
- **Infrared**: Thermal imaging, FLIR cameras (4 datasets)
- **Damage**: Defect detection, infrastructure inspection (3 datasets)
- **Underwater**: Marine life, coral, infrastructure (4 datasets)
- **Document**: OCR, document parsing, diagrams (6 datasets)

## Quick Search

The easiest way to find datasets is using the search functions:

```{r eval=FALSE}
library(torchvision)

# Search for specific topics
search_rf100("cell")        # Find cell-related datasets
search_rf100("solar")       # Find solar panel datasets
search_rf100("x-ray")       # Find X-ray datasets

# List all datasets in a collection
search_rf100(collection = "biology")
search_rf100(collection = "medical")

# View complete catalog
catalog <- get_rf100_catalog()
View(catalog)
```

## Example: Finding a Photovoltaic Dataset

One of the motivations for this catalog was answering questions like: *"Is there a photovoltaic dataset in torchvision?"*

```{r eval=FALSE}
# Search for solar/photovoltaic datasets
search_rf100("solar")
search_rf100("photovoltaic")

# Result shows:
# - solar_panel in infrared collection
# - solar_panel in damage collection
```

## Complete Catalog

Here's the complete catalog of all RF100 datasets:

```{r eval=FALSE}
library(torchvision)
library(knitr)

catalog <- get_rf100_catalog()

# Display key columns
kable(catalog[, c("collection", "dataset", "description", "total_size_mb", "estimated_images")])
```

## Collections

### Biology Collection (9 datasets)

Microscopy and biological imaging datasets for research and diagnostics:

```{r eval=FALSE}
search_rf100(collection = "biology")
```

**Available datasets:**

- `stomata_cell`: Plant stomata cells for biology research
- `blood_cell`: Blood cell detection (RBC, WBC, platelets)
- `parasite`: Parasite detection in microscopy images
- `cell`: General cell detection in microscopy
- `bacteria`: Bacteria detection in microscopy images
- `cotton_disease`: Cotton plant disease detection
- `mitosis`: Mitosis phase detection in cell images
- `phage`: Bacteriophage detection in microscopy
- `liver_disease`: Liver disease pathology detection

### Medical Collection (8 datasets)

Medical imaging datasets for clinical and research applications:

```{r eval=FALSE}
search_rf100(collection = "medical")
```

**Available datasets:**

- `radio_signal`: Radio signal detection in medical imaging
- `rheumatology`: Rheumatology X-ray abnormality detection
- `knee`: ACL and knee X-ray analysis
- `abdomen_mri`: Abdomen MRI organ detection
- `brain_axial_mri`: Brain axial MRI structure detection
- `gynecology_mri`: Gynecology MRI structure detection
- `brain_tumor`: Brain tumor detection in MRI scans
- `fracture`: Bone fracture detection in X-rays

### Infrared Collection (4 datasets)

Thermal and infrared imaging datasets:

```{r eval=FALSE}
search_rf100(collection = "infrared")
```

**Available datasets:**

- `thermal_dog_and_people`: Thermal imaging of dogs and people
- `solar_panel`: Solar panel detection in infrared imagery
- `thermal_cheetah`: Thermal imaging of cheetahs
- `ir_object`: FLIR camera object detection

### Damage Collection (3 datasets)

Infrastructure damage and defect detection:

```{r eval=FALSE}
search_rf100(collection = "damage")
```

**Available datasets:**

- `liquid_crystals`: 4-fold defect detection in LCD displays
- `solar_panel`: Solar panel defect and damage detection
- `asbestos`: Asbestos detection for safety inspection

### Underwater Collection (4 datasets)

Marine and underwater imaging datasets:

```{r eval=FALSE}
search_rf100(collection = "underwater")
```

**Available datasets:**

- `pipes`: Underwater pipe detection for infrastructure
- `aquarium`: Aquarium fish and species detection
- `objects`: Underwater object detection
- `coral`: Coral reef detection and monitoring

### Document Collection (6 datasets)

Document analysis and OCR datasets:

```{r eval=FALSE}
search_rf100(collection = "document")
```

**Available datasets:**

- `tweeter_post`: Twitter post element detection
- `tweeter_profile`: Twitter profile element detection
- `document_part`: Document structure and part detection
- `activity_diagram`: Activity diagram element detection
- `signature`: Signature detection in documents
- `paper_part`: Academic paper structure detection

## Usage Example

Once you've found a dataset, loading it is straightforward:

```{r eval=FALSE}
library(torchvision)

# Search for blood cell dataset
search_rf100("blood")

# Load the dataset
ds <- rf100_biology_collection(
  dataset = "blood_cell",
  split = "train",
  download = TRUE
)

# Inspect a sample
item <- ds[1]
print(item$y$labels)  # Object classes
print(item$y$boxes)   # Bounding boxes

# Visualize with bounding boxes
boxed <- draw_bounding_boxes(item)
tensor_image_browse(boxed)
```

## Dataset Statistics

```{r eval=FALSE}
catalog <- get_rf100_catalog()

# Total size of all datasets
sum(catalog$total_size_mb) / 1024  # In GB

# Datasets by size
catalog[order(-catalog$total_size_mb), c("dataset", "collection", "total_size_mb")]

# Smallest and largest datasets
catalog[which.min(catalog$total_size_mb), ]
catalog[which.max(catalog$total_size_mb), ]

# Average size by collection
aggregate(total_size_mb ~ collection, data = catalog, FUN = mean)
```

## Filtering and Exploration

The catalog is a regular data frame, so you can use standard R operations:

```{r eval=FALSE}
# Find small datasets (< 20 MB total)
subset(catalog, total_size_mb < 20)

# Find large datasets (> 200 MB total)
subset(catalog, total_size_mb > 200)

# Find datasets with specific keywords
subset(catalog, grepl("tumor|cancer|disease", description, ignore.case = TRUE))

# Datasets with all three splits
subset(catalog, has_train & has_test & has_valid)
```

## Additional Resources

- **RoboFlow Universe**: Browse datasets at https://universe.roboflow.com/browse/
- **Collection Functions**: See `?rf100_biology_collection`, `?rf100_medical_collection`, etc.
- **Visualization**: See `?draw_bounding_boxes` for visualizing detections

## Citation

If you use RF100 datasets in your research, please cite:

```
@article{roboflow100,
  title={Roboflow 100: A Rich, Multi-Domain Object Detection Benchmark},
  author={Roboflow},
  journal={arXiv preprint},
  year={2022}
}
```

<!-- Before submitting a PR, please make sure to check our contributing guidelines regarding code formatting, tests, and documentation: https://github.com/pytorch/vision/blob/main/CONTRIBUTING.md -->

# Contributing to Torchvision

We want to make contributing to this project as easy and transparent as possible.

## TL;DR

We appreciate all contributions. If you are interested in contributing to Torchvision, there are many ways to help out.
Your contributions may fall into the following categories:

- It helps the project if you could
    - Report issues you're facing
    - Give a :+1: on issues that others reported and that are relevant to you

- Answering queries on the issue tracker, investigating bugs are very valuable contributions to the project.

- You would like to improve the documentation. This is no less important than improving the library itself!
If you find a typo in the documentation, do not hesitate to submit a GitHub pull request.

- If you would like to fix a bug
    - please pick one from the [list of open issues labelled as "help wanted"](https://github.com/pytorch/vision/issues?q=is%3Aopen+is%3Aissue+label%3A%22help+wanted%22)
    - comment on the issue that you want to work on this issue
    - send a PR with your fix, see below.

- If you plan to contribute new features, utility functions or extensions, please first open an issue and discuss the feature with us.

## Issues

We use GitHub issues to track public bugs. Please ensure your description is
clear and has sufficient instructions to be able to reproduce the issue.

## Development installation


### Dependencies

Start by installing the **nightly** build of PyTorch following the [official
instructions](https://pytorch.org/get-started/locally/). Note that the official
instructions may ask you to install torchvision itself. If you are doing development
on torchvision, you should not install prebuilt torchvision packages.

**Optionally**, install `libpng` and `libjpeg-turbo` if you want to enable
support for
native encoding / decoding of PNG and JPEG formats in
[torchvision.io](https://pytorch.org/vision/stable/io.html#image):

```bash
conda install libpng libjpeg-turbo -c pytorch
```

Note: you can use the `TORCHVISION_INCLUDE` and `TORCHVISION_LIBRARY`
environment variables to tell the build system where to find those libraries if
they are in specific locations. Take a look at
[setup.py](https://github.com/pytorch/vision/blob/main/setup.py) for more
details.

### Clone and install torchvision

```bash
git clone https://github.com/pytorch/vision.git
cd vision
python setup.py develop  # use install instead of develop if you don't care about development.
# or, for OSX
# MACOSX_DEPLOYMENT_TARGET=10.9 CC=clang CXX=clang++ python setup.py develop
# for C++ debugging, use DEBUG=1
# DEBUG=1 python setup.py develop
```

By default, GPU support is built if CUDA is found and `torch.cuda.is_available()` is true. It's possible to force
building GPU support by setting `FORCE_CUDA=1` environment variable, which is useful when building a docker image.

We don't officially support building from source using `pip`, but _if_ you do, you'll need to use the
`--no-build-isolation` flag.

#### Other development dependencies (some of these are needed to run tests):

```
pip install expecttest flake8 typing mypy pytest pytest-mock scipy requests
```

## Development Process

If you plan to modify the code or documentation, please follow the steps below:

1. Fork the repository and create your branch from `main`.
2. If you have modified the code (new feature or bug-fix), please add unit tests.
3. If you have changed APIs, update the documentation. Make sure the documentation builds.
4. Ensure the test suite passes.
5. Make sure your code passes the formatting checks (see below).

For more details about pull requests,
please read [GitHub's guides](https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/creating-a-pull-request).

If you would like to contribute a new model, please see [here](#New-architecture-or-improved-model-weights).

If you would like to contribute a new dataset, please see [here](#New-dataset).

### Code formatting and typing

#### Formatting

The torchvision code is formatted by [black](https://black.readthedocs.io/en/stable/),
and checked against pep8 compliance with [flake8](https://flake8.pycqa.org/en/latest/).
Instead of relying directly on `black` however, we rely on
[ufmt](https://github.com/omnilib/ufmt), for compatibility reasons with Facebook
internal infrastructure.

To format your code, install `ufmt` with `pip install ufmt==1.3.3 black==22.3.0 usort==1.0.2` and use e.g.:

```bash
ufmt format torchvision
```

For the vast majority of cases, this is all you should need to run. For the
formatting to be a bit faster, you can also choose to only apply `ufmt` to the
files that were edited in your PR with e.g.:

```bash
ufmt format `git diff main --name-only`
```

Similarly, you can check for `flake8` errors with `flake8 torchvision`, although
they should be fairly rare considering that most of the errors are automatically
taken care of by `ufmt` already.

##### Pre-commit hooks

For convenience and **purely optionally**, you can rely on [pre-commit
hooks](https://pre-commit.com/) which will run both `ufmt` and `flake8` prior to
every commit.

First install the `pre-commit` package with `pip install pre-commit`, and then
run `pre-commit install` at the root of the repo for the hooks to be set up -
that's it.

Feel free to read the [pre-commit docs](https://pre-commit.com/#usage) to learn
more and improve your workflow. You'll see for example that `pre-commit run
--all-files` will run both `ufmt` and `flake8` without the need for you to
commit anything, and that the `--no-verify` flag can be added to `git commit` to
temporarily deactivate the hooks.

#### Type annotations

The codebase has type annotations, please make sure to add type hints if required. We use `mypy` tool for type checking:
```bash
mypy --config-file mypy.ini
```

### Unit tests

Before running tests make sure to install [test dependencies](#other-development-dependencies-some-of-these-are-needed-to-run-tests).

If you have modified the code by adding a new feature or a bug-fix, please add unit tests for that. To run a specific
test:
```bash
pytest test/<test-module.py> -vvv -k <test_myfunc>
# e.g. pytest test/test_transforms.py -vvv -k test_center_crop
```

If you would like to run all tests:
```bash
pytest test -vvv
```

Tests that require internet access should be in
`test/test_internet.py`.

### Documentation

Torchvision uses [Google style](http://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html)
for formatting docstrings. Length of line inside docstrings block must be limited to 120 characters.

Please, follow the instructions to build and deploy the documentation locally.

#### Install requirements

```bash
cd docs
pip install -r requirements.txt
```

#### Build

```bash
cd docs
make html-noplot
```

Then open `docs/build/html/index.html` in your favorite browser.

The docs are also automatically built when you submit a PR. The job that
builds the docs is named `build_docs`. You can access the rendered docs by
clicking on that job and then going to the "Artifacts" tab.

You can clean the built docs and re-start the build from scratch by doing ``make
clean``.

#### Building the example gallery - or not

In most cases, running `make html-noplot` is enough to build the docs for your
specific use-case. The `noplot` part tells sphinx **not** to build the examples
in the [gallery](https://pytorch.org/vision/stable/auto_examples/index.html),
which saves a lot of building time.

If you need to build all the examples in the gallery, then you can use `make
html`.

You can also choose to only build a subset of the examples by using the
``EXAMPLES_PATTERN`` env variable, which accepts a regular expression. For
example ``EXAMPLES_PATTERN="transforms" make html`` will only build the examples
with "transforms" in their name.

### New architecture or improved model weights

Please refer to the guidelines in [Contributing to Torchvision - Models](https://github.com/pytorch/vision/blob/main/CONTRIBUTING_MODELS.md).

### New dataset

Please, do not send any PR with a new dataset without discussing
it in an issue as, most likely, it will not be accepted.

### Pull Request

If all previous checks (flake8, mypy, unit tests) are passing, please send a PR. Submitted PR will pass other tests on
different operating systems, python versions and hardware.

For more details about pull requests workflow,
please read [GitHub's guides](https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/creating-a-pull-request).

## License

By contributing to Torchvision, you agree that your contributions will be licensed
under the LICENSE file in the root directory of this source tree.

Contributors are also required to [sign our Contributor License Agreement](https://code.facebook.com/cla).

# Contributing to Torchvision - Models

- [New Model Architectures - Overview](#new-model-architectures---overview)

- [New Weights for Existing Model Architectures](#new-weights-for-existing-model-architectures)

## New Model Architectures - Overview

For someone who would be interested in adding a model architecture, it is also expected to train the model, so here are a few important considerations:

- Training big models requires lots of resources and the cost quickly adds up

- Reproducing models is fun but also risky as you might not always get the results reported on the paper. It might require a huge amount of effort to close the gap

- The contribution might not get merged if we significantly lack in terms of accuracy, speed etc

- Including new models in TorchVision might not be the best approach, so other options such as releasing the model through to [Pytorch Hub](https://pytorch.org/hub/) should be considered

So, before starting any work and submitting a PR there are a few critical things that need to be taken into account in order to make sure the planned contribution is within the context of TorchVision, and the requirements and expectations are discussed beforehand. If this step is skipped and a PR is submitted without prior discussion it will almost certainly be rejected.

### 1. Preparation work

- Start by looking into this [issue](https://github.com/pytorch/vision/issues/2707) in order to have an idea of the models that are being considered, express your willingness to add a new model and discuss with the community whether this model should be included in TorchVision. It is very important at this stage to make sure that there is an agreement on the value of having this model in TorchVision and there is no one else already working on it.

- If the decision is to include the new model, then please create a new ticket which will be used for all design and implementation discussions prior to the PR. One of the TorchVision maintainers will reach out at this stage and this will be your POC from this point onwards in order to provide support, guidance and regular feedback.

### 2.  Implement the model

Please take a look at existing models in TorchVision to get familiar with the idioms. Also, please look at recent contributions for new models. If in doubt about any design decisions you can ask for feedback on the issue created in step 1.  Example of things to take into account:

- The implementation should be as close as possible to the canonical implementation/paper
- The PR must include the code implementation, documentation and tests
- It should also extend the existing reference scripts used to train the model
- The weights need to reproduce closely the results of the paper in terms of accuracy, even though the final weights to be deployed will be those trained by the TorchVision maintainers
- The PR description should include commands/configuration used to train the model, so that the TorchVision maintainers can easily run them to verify the implementation and generate the final model to be released
- Make sure we re-use existing components as much as possible (inheritance)
- New primitives (transforms, losses, etc.) can be added if necessary, but the final location will be determined after discussion with the dedicated maintainer
- Please take a look at the detailed [implementation and documentation guidelines](https://github.com/pytorch/vision/issues/5319) for a fine grain list of things not to be missed

### 3. Train the model with reference scripts

To validate the new model against the common benchmark, as well as to generate pre-trained weights, you must use TorchVision’s reference scripts to train the model.

Make sure all logs and a final (or best) checkpoint are saved, because it is expected that a submission shows that a model has been successfully trained  and the results are in line with the original paper/repository. This will allow the reviewers to quickly check the validity of the submission, but please note that the final model to be released will be re-trained by the maintainers in order to verify reproducibility,  ensure that the changes occurred during the PR review did not introduce any bugs, and to avoid moving around a large amount of data (including all checkpoints and logs).

### 4. Submit a PR

Submit a PR and tag the assigned maintainer. This PR should:

- Link the original ticket
- Provide a link for the original paper and the original repository if available
- Highlight the important test metrics and how they compare to the original paper
- Highlight any design choices that deviate from the original paper/implementation and rationale for these choices

## New Weights for Existing Model Architectures

The process of improving existing models, for instance improving accuracy by retraining the model with a different set of hyperparameters or augmentations, is the following:

1. Open a ticket and discuss with the community and maintainers whether this improvement should be added to TorchVision. Note that to add new weights the improvement should be significant.

2. Train the model using TorchVision reference scripts. You can add new primitives (transforms, losses, etc) when necessary, but the final location will be determined after discussion with the dedicated maintainer.

3. Open a PR with the new weights, together with the training logs and the checkpoint chosen so the reviewers can verify the submission.  Details on how the model was trained, i.e., the training command using the reference scripts, should be included in the PR.

4. The PR reviewers should replicate the results on their side to verify the submission and if all goes well the new weights should be ready to be released!

# Torchvision maintainers guide

This document aims at documenting user-facing policies / principles used when
developing and maintaining torchvision. Other maintainer info (e.g. release
process) can be found in the meta-internal wiki.

### What is public and what is private?

For the Python API, torchvision largely follows the [PyTorch
policy](https://github.com/pytorch/pytorch/wiki/Public-API-definition-and-documentation)
which is consistent with other major packages
([numpy](https://numpy.org/neps/nep-0023-backwards-compatibility.html),
[scikit-learn](https://scikit-learn.org/dev/glossary.html#term-API) etc.).
We recognize that his policy is somewhat imperfect for some edge cases, and that
it's difficult to come up with an accurate technical definition. In broad terms,
which are usually well understood by users, the policy is that:

- modules that can be accessed without leading underscore are public
- objects in a public file that don't have a leading underscore are public
- class attributes are public iff they have no leading underscore
- the rest of the modules / objects / class attributes are considered private

The public API has backward-compatible (BC) guarantees defined in our
deprecation policy (see below). The private API has not BC guarantees.

For C++, code is private. For Meta employees: if a C++ change breaks fbcode, fix
fbcode or revert the change. We should be careful about models running in
production and relying on torchvision ops.

The `test` folder is not importable and is **private.** Even meta-internal
projects should *not* rely on it (it has happened in the past and is now
programmatically impossible).

The training references do not have BC guarantees. Breaking changes are
possible, but we should make sure that the tutorials are still running properly,
and that their intended narrative is preserved (by e.g. checking outputs,
etc.).

The rest of the folders (build, android, ios, etc.) are private and have no BC
guarantees.

### Deprecation policy.

Because they're disruptive, **deprecations should only be used sparingly**.

We largely follow the [PyTorch
policy](https://github.com/pytorch/pytorch/wiki/PyTorch's-Python-Frontend-Backward-and-Forward-Compatibility-Policy):
breaking changes require a deprecation period of at least 2 versions.

Deprecations should clearly indicate their deadline in the docs and warning
messages. Avoid not committing to a deadline, or keeping deprecated APIs for too
long: it gives no incentive for users to update their code, sends conflicting
messages ("why was this API removed while this other one is still around?"), and
accumulates debt in the project.

### Should this attribute be public? Should this function be private?

When designing an API it’s not always obvious what should be exposed as public,
and what should be kept as a private implementation detail. The following
guidelines can be useful:

* Functional consistency throughout the library is a top priority, for users and
  developers’ sake. In doubt and unless it’s clearly wrong, expose what other
  similar classes expose.
* Think really hard about the users and their use-cases, and try to expose what
  they would need to address those use-cases. Aggressively keep everything else
  private. Remember that the “private -> public” direction is way smoother than
  the “public -> private” one: in doubt, keep it private.
* When thinking about use-cases, the general API motto applies: make what’s
  simple and common easy, and make what’s complex possible (80% / 20% rule).
  There might be a ~1% left that’s not addressed: that’s OK. Also, **make what’s
  wrong very hard**, if not impossible.

As a good practice, always create new files and even classes with a leading
underscore in their name. This way, everything is private by default and the
only public surface is explicitly present in an `__init__.py` file.

---
title: Scheduled workflow failed
labels:
  - bug
  - "module: datasets"
---

Oh no, something went wrong in the scheduled workflow {{ env.WORKFLOW }}/{{ env.JOB }}. 
Please look into it:

https://github.com/{{ env.REPO }}/actions/runs/{{ env.ID }}

Feel free to close this if this was just a one-off error.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_center_crop}
\alias{transform_center_crop}
\title{Crops the given image at the center}
\usage{
transform_center_crop(img, size)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{size}{(sequence or int): Desired output size of the crop. If size is
an int instead of sequence like c(h, w), a square crop (size, size) is
made. If provided a tuple or list of length 1, it will be interpreted as
\code{c(size, size)}.}
}
\description{
The image can be a Magick Image or a torch Tensor, in which case it is
expected to have \verb{[..., H, W]} shape, where ... means an arbitrary number
of leading dimensions.
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-convnext.R
\name{model_convnext}
\alias{model_convnext}
\alias{model_convnext_tiny_1k}
\alias{model_convnext_tiny_22k}
\alias{model_convnext_small_22k}
\alias{model_convnext_small_22k1k}
\alias{model_convnext_base_1k}
\alias{model_convnext_base_22k}
\alias{model_convnext_large_1k}
\alias{model_convnext_large_22k}
\title{ConvNeXt Implementation}
\usage{
model_convnext_tiny_1k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 1000,
  ...
)

model_convnext_tiny_22k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 21841,
  ...
)

model_convnext_small_22k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 21841,
  ...
)

model_convnext_small_22k1k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 21841,
  ...
)

model_convnext_base_1k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 1000,
  ...
)

model_convnext_base_22k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 21841,
  ...
)

model_convnext_large_1k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 1000,
  ...
)

model_convnext_large_22k(
  pretrained = FALSE,
  progress = TRUE,
  channels = 3,
  num_classes = 21841,
  ...
)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{channels}{The number of channels in the input image.  Default: 3.}

\item{num_classes}{number of output classes (default: 1000).}

\item{...}{Other parameters passed to the model implementation.}
}
\description{
Implements the ConvNeXt architecture from \href{https://arxiv.org/pdf/2201.03545}{ConvNeXt: A ConvNet for the 2020s}
}
\section{Functions}{
\itemize{
\item \code{model_convnext_tiny_1k()}: ConvNeXt Tiny model trained on Imagenet 1k.

\item \code{model_convnext_tiny_22k()}: ConvNeXt Tiny model trained on Imagenet 22k.

\item \code{model_convnext_small_22k()}: ConvNeXt Small model trained on Imagenet 22k.

\item \code{model_convnext_small_22k1k()}: ConvNeXt Small model pretrained on Imagenet 1k
and fine-tuned on Imagenet 22k classes.

\item \code{model_convnext_base_1k()}: ConvNeXt Base model trained on Imagenet 1k.

\item \code{model_convnext_base_22k()}: ConvNeXt Base model trained on Imagenet 22k.

\item \code{model_convnext_large_1k()}: ConvNeXt Large model trained on Imagenet 1k.

\item \code{model_convnext_large_22k()}: ConvNeXt Large model trained on Imagenet 22k.

}}
\section{Variants}{

\subsection{Model Summary and Performance for pretrained weights}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{| Model                | Top-1 Acc| Params | GFLOPS | File Size | `num_classes`| image size |
|----------------------|----------|--------|--------|-----------|--------------|------------|
| convnext_tiny_1k     | 82.1\%    | 28M    | 4.5    | 109 MB    |         1000 | 224 x 224  |
| convnext_tiny_22k    | 82.9\%    | 29M    | 4.5    | 170 MB    |        21841 | 224 x 224  |
| convnext_small_22k   | 84.6\%    | 50M    | 8.7    | 252 MB    |        21841 | 224 x 224  |
| convnext_small_22k1k | 84.6\%    | 50M    | 8.7    | 252 MB    |        21841 | 224 x 224  |
| convnext_base_1k     | 85.1\%    | 89M    | 15.4   | 338 MB    |         1000 | 224 x 224  |
| convnext_base_22k    | 85.8\%    | 89M    | 15.4   | 420 MB    |        21841 | 224 x 224  |
| convnext_large_1k    | 84.3\%    | 198M   | 34.4   | 750 MB    |         1000 | 224 x 224  |
| convnext_large_22k   | 86.6\%    | 198M   | 34.4   | 880 MB    |        21841 | 224 x 224  |
}\if{html}{\out{</div>}}
}
}

\examples{
\dontrun{
# 1. Download sample image (dog)
norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants, see
# https://pytorch.org/vision/stable/models.html
norm_std  <- c(0.229, 0.224, 0.225)
img_url <- "https://en.wikipedia.org/wiki/Special:FilePath/Felis_catus-cat_on_snow.jpg"
img <- base_loader(img_url)

# 2. Convert to tensor (RGB only), resize and normalize
input <- img \%>\%
 transform_to_tensor() \%>\%
 transform_resize(c(224, 224)) \%>\%
 transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)

# 3. Load pretrained models
model_small <- convnext_tiny_1k(pretrained = TRUE, root = tempdir())
model_small$eval()

# 4. Forward pass
output_s <- model_small(batch)

# 5. Show Top-5 predictions
topk <- output_s$topk(k = 5, dim = 2)
indices <- as.integer(topk[[2]][1, ])
scores <- as.numeric(topk[[1]][1, ])
glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}\%)")
}

}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vision_utils.R
\name{draw_bounding_boxes}
\alias{draw_bounding_boxes}
\alias{draw_bounding_boxes.default}
\alias{draw_bounding_boxes.torch_tensor}
\alias{draw_bounding_boxes.image_with_bounding_box}
\title{Draws bounding boxes on image.}
\usage{
draw_bounding_boxes(x, ...)

\method{draw_bounding_boxes}{default}(x, ...)

\method{draw_bounding_boxes}{torch_tensor}(
  x,
  boxes,
  labels = NULL,
  colors = NULL,
  color = NULL,
  fill = FALSE,
  width = 1,
  font = c("serif", "plain"),
  font_size = 10,
  ...
)

\method{draw_bounding_boxes}{image_with_bounding_box}(x, ...)
}
\arguments{
\item{x}{Tensor of shape (C x H x W) and dtype \code{uint8} or dtype \code{float}.
In case of dtype float, values are assumed to be in range \eqn{[0, 1]}.
C value for channel can only be 1 (grayscale) or 3 (RGB).}

\item{...}{Additional arguments passed to methods.}

\item{boxes}{Tensor of size (N, 4) containing N bounding boxes in
c(\eqn{x_{min}}, \eqn{y_{min}}, \eqn{x_{max}}, \eqn{y_{max}}).
format. Note that the boxes coordinates are absolute with respect
to the image. In other words: \eqn{0  \leq x_{min} < x_{max} < Height } and
\eqn{0  \leq y_{min} < y_{max} < Width }.}

\item{labels}{character vector containing the labels of bounding boxes.}

\item{colors}{character vector containing the colors
of the boxes or single color for all boxes. The color can be represented as
strings e.g. "red" or "#FF00FF". By default, viridis colors are generated for boxes.}

\item{color}{Deprecated alias for \code{colors}.}

\item{fill}{If \code{TRUE} fills the bounding box with specified color.}

\item{width}{Width of text shift to the bounding box.}

\item{font}{NULL for the current font family, or a character vector of length 2 for Hershey vector fonts.}

\item{font_size}{The requested font size in points.}
}
\value{
torch_tensor of size (C, H, W) of dtype uint8: Image Tensor with bounding boxes plotted.
}
\description{
Draws bounding boxes on top of one image tensor
}
\examples{
if (torch::torch_is_installed()) {
\dontrun{
image_tensor <- torch::torch_randint(170, 250, size = c(3, 360, 360))$to(torch::torch_uint8())
x <- torch::torch_randint(low = 1, high = 160, size = c(12,1))
y <- torch::torch_randint(low = 1, high = 260, size = c(12,1))
boxes <- torch::torch_cat(c(x, y, x + 20, y +  10), dim = 2)
bboxed <- draw_bounding_boxes(image_tensor, boxes, colors = "black", fill = TRUE)
tensor_image_browse(bboxed)
}
}
}
\seealso{
Other image display: 
\code{\link{draw_keypoints}()},
\code{\link{draw_segmentation_masks}()},
\code{\link{tensor_image_browse}()},
\code{\link{tensor_image_display}()},
\code{\link{vision_make_grid}()}
}
\concept{image display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_normalize}
\alias{transform_normalize}
\title{Normalize a tensor image with mean and standard deviation}
\usage{
transform_normalize(img, mean, std, inplace = FALSE)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{mean}{(sequence): Sequence of means for each channel.}

\item{std}{(sequence): Sequence of standard deviations for each channel.}

\item{inplace}{(bool,optional): Bool to make this operation in-place.}
}
\description{
Given mean: \verb{(mean[1],...,mean[n])} and std: \verb{(std[1],..,std[n])} for \code{n}
channels, this transform will normalize each channel of the input
\code{torch_tensor} i.e.,
\code{output[channel] = (input[channel] - mean[channel]) / std[channel]}
}
\note{
This transform acts out of place, i.e., it does not mutate the input tensor.
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-efficientnetv2.R
\name{model_efficientnet_v2}
\alias{model_efficientnet_v2}
\alias{model_efficientnet_v2_s}
\alias{model_efficientnet_v2_m}
\alias{model_efficientnet_v2_l}
\title{EfficientNetV2 Models}
\usage{
model_efficientnet_v2_s(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_v2_m(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_v2_l(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{...}{Other parameters passed to the model implementation, such as
\code{num_classes} to change the output dimension.}
}
\description{
Constructs EfficientNetV2 model architectures as described in
\emph{EfficientNetV2: Smaller Models and Faster Training}.
}
\section{Functions}{
\itemize{
\item \code{model_efficientnet_v2_s()}: EfficientNetV2-S model

\item \code{model_efficientnet_v2_m()}: EfficientNetV2-M model

\item \code{model_efficientnet_v2_l()}: EfficientNetV2-L model

}}
\section{Task}{

Image classification with 1000 output classes by default (ImageNet).
}

\section{Input Format}{

The models expect input tensors of shape \code{(batch_size, 3, H, W)}.
Typical values for \code{H} and \code{W} are 384 for V2-S, 480 for V2-M,
and 512 for V2-L.
}

\section{Variants}{

\tabular{lllll}{
\strong{Model} \tab \strong{Resolution} \tab \strong{Params (M)} \tab
\strong{GFLOPs} \tab \strong{Top-1 Acc.} \cr
V2-S \tab 384 \tab 24 \tab 8.4 \tab 83.9 \cr
V2-M \tab 480 \tab 55 \tab 24 \tab 85.1 \cr
V2-L \tab 512 \tab 119 \tab 55 \tab 85.7
}
}

\examples{
\dontrun{
model <- model_efficientnet_v2_s()
input <- torch::torch_randn(1, 3, 224, 224)
output <- model(input)

# Show Top-5 predictions
topk <- output$topk(k = 5, dim = 2)
indices <- as.integer(topk[[2]][1, ])
scores <- as.numeric(topk[[1]][1, ])
glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}\%)")
}
}
\seealso{
\code{\link{model_efficientnet}}

Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vision_utils.R
\name{draw_keypoints}
\alias{draw_keypoints}
\title{Draws Keypoints}
\usage{
draw_keypoints(
  image,
  keypoints,
  connectivity = NULL,
  colors = NULL,
  radius = 2,
  width = 3
)
}
\arguments{
\item{image}{Tensor of shape (3 x H x W) and dtype \code{uint8} or dtype \code{float}.
In case of dtype float, values are assumed to be in range \eqn{[0, 1]}.}

\item{keypoints}{Tensor of shape (N, K, 2) the K keypoints location for each of the N detected poses instance,}

\item{connectivity}{List of integer pairs \code{c(i, j)} specifying which
keypoints to connect with a line, e.g. \code{list(c(1, 2), c(2, 3))}.
\code{NULL} (default) draws no connecting lines.}

\item{colors}{character vector containing the colors
of the keypoints or single color for all keypoints. The color can be represented as
strings e.g. "red" or "#FF00FF". By default, rainbow colors are generated for keypoints}

\item{radius}{radius of the plotted keypoint.}

\item{width}{width of line connecting keypoints.}
}
\value{
Image Tensor of dtype uint8 with keypoints drawn.
}
\description{
Draws Keypoints, an object describing a body part (like rightArm or leftShoulder), on given RGB tensor image.
}
\examples{
if (torch::torch_is_installed()) {
\dontrun{
image <- torch::torch_randint(190, 255, size = c(3, 360, 360))$to(torch::torch_uint8())
keypoints <- torch::torch_randint(low = 60, high = 300, size = c(4, 5, 2))
keypoint_image <- draw_keypoints(image, keypoints)
tensor_image_browse(keypoint_image)
}
}
}
\seealso{
Other image display: 
\code{\link{draw_bounding_boxes}()},
\code{\link{draw_segmentation_masks}()},
\code{\link{tensor_image_browse}()},
\code{\link{tensor_image_display}()},
\code{\link{vision_make_grid}()}
}
\concept{image display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-box_convert.R
\name{box_xyxy_to_cxcywh}
\alias{box_xyxy_to_cxcywh}
\title{box_xyxy_to_cxcywh}
\usage{
box_xyxy_to_cxcywh(boxes)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format which will be converted.}
}
\value{
boxes (Tensor(N, 4)): boxes in \eqn{(c_x, c_y, w, h)} format.
}
\description{
Converts bounding boxes from  \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format to \eqn{(c_x, c_y, w, h)} format.
(x1, y1) refer to top left of bounding box
(x2, y2) refer to bottom right of bounding box
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-box_convert.R
\name{box_xywh_to_xyxy}
\alias{box_xywh_to_xyxy}
\title{box_xywh_to_xyxy}
\usage{
box_xywh_to_xyxy(boxes)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes in (x, y, w, h) which will be converted.}
}
\value{
boxes (Tensor[N, 4]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
}
\description{
Converts bounding boxes from  (x, y, w, h) format to \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
(x, y) refers to top left of bouding box.
(w, h) refers to width and height of box.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_resize}
\alias{transform_resize}
\title{Resize the input image to the given size}
\usage{
transform_resize(img, size, interpolation = 2)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{size}{(sequence or int): Desired output size. If size is a sequence
like c(h, w), output size will be matched to this. If size is an int,
smaller edge of the image will be matched to this number.
i.e, if height > width, then image will be rescaled to
(size * height / width, size).}

\item{interpolation}{(int, optional) Desired interpolation. An integer
\code{0 = nearest}, \code{2 = bilinear}, and \code{3 = bicubic} or a name from
\code{\link[magick:filter_types]{magick::filter_types()}}.}
}
\description{
The image can be a Magic Image or a torch Tensor, in which case it is
expected to have \verb{[..., H, W]} shape, where ... means an arbitrary number
of leading dimensions
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_erasing}
\alias{transform_random_erasing}
\title{Randomly selects a rectangular region in an image and erases its pixel values}
\usage{
transform_random_erasing(
  img,
  p = 0.5,
  scale = c(0.02, 0.33),
  ratio = c(0.3, 3.3),
  value = 0,
  inplace = FALSE
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{p}{probability that the random erasing operation will be performed.}

\item{scale}{range of proportion of erased area against input image.}

\item{ratio}{range of aspect ratio of erased area.}

\item{value}{erasing value. Default is 0. If a single int, it is used to
erase all pixels. If a tuple of length 3, it is used to erase
R, G, B channels respectively.
If a str of 'random', erasing each pixel with random values.}

\item{inplace}{boolean to make this transform inplace. Default set to FALSE.}
}
\description{
'Random Erasing Data Augmentation' by Zhong \emph{et al.}
See \url{https://arxiv.org/pdf/1708.04896}
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_affine}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-fcn.R
\name{model_fcn_resnet}
\alias{model_fcn_resnet}
\alias{model_fcn_resnet50}
\alias{model_fcn_resnet101}
\title{Fully Convolutional Network for Semantic Segmentation}
\usage{
model_fcn_resnet50(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 21,
  aux_loss = NULL,
  pretrained_backbone = TRUE,
  ...
)

model_fcn_resnet101(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 21,
  aux_loss = NULL,
  pretrained_backbone = TRUE,
  ...
)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{num_classes}{Number of output classes. Default: 21.}

\item{aux_loss}{If TRUE, includes the auxiliary classifier. If NULL, defaults to TRUE when \code{pretrained = TRUE}.}

\item{pretrained_backbone}{If TRUE, uses a backbone pre-trained on ImageNet.}

\item{...}{Additional arguments passed to the backbone implementation.}
}
\value{
An \code{nn_module} representing the FCN model.
}
\description{
Constructs an FCN (Fully Convolutional Network) model for semantic image segmentation,
based on a ResNet backbone as described in
\href{https://arxiv.org/abs/1411.4038}{Fully Convolutional Networks for Semantic Segmentation}.
}
\details{
The 21 output classes follow the PASCAL VOC convention:
\code{background}, \code{aeroplane}, \code{bicycle}, \code{bird}, \code{boat},
\code{bottle}, \code{bus}, \code{car}, \code{cat}, \code{chair},
\code{cow}, \verb{dining table}, \code{dog}, \code{horse}, \code{motorbike},
\code{person}, \verb{potted plant}, \code{sheep}, \code{sofa}, \code{train},
\code{tv/monitor}.

Pretrained weights require \code{num_classes = 21}.
}
\examples{
\dontrun{
library(magrittr)
norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants, see
# https://pytorch.org/vision/stable/models.html
norm_std  <- c(0.229, 0.224, 0.225)
img_url <- "https://en.wikipedia.org/wiki/Special:FilePath/Felis_catus-cat_on_snow.jpg"
img <- base_loader(img_url)

input <- img \%>\%
 transform_to_tensor() \%>\%
 transform_resize(c(520, 520)) \%>\%
 transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)

model <- model_fcn_resnet50(pretrained = TRUE)
model$eval()
output <- model(batch)

# visualize the result
# `draw_segmentation_masks()` turns the torch_float output into a boolean mask internaly:
segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
tensor_image_display(segmented)


model <- model_fcn_resnet101(pretrained = TRUE)
model$eval()
output <- model(batch)

# visualize the result
segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
tensor_image_display(segmented)
}
}
\seealso{
Other semantic_segmentation_model: 
\code{\link{model_convnext_segmentation}},
\code{\link{model_deeplabv3}}
}
\concept{semantic_segmentation_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_affine}
\alias{transform_random_affine}
\title{Random affine transformation of the image keeping center invariant}
\usage{
transform_random_affine(
  img,
  degrees,
  translate = NULL,
  scale = NULL,
  shear = NULL,
  interpolation = 0,
  fill = 0,
  resample,
  fillcolor
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{degrees}{(sequence or float or int): Range of degrees to select from.
If degrees is a number instead of sequence like c(min, max), the range of
degrees will be (-degrees, +degrees).}

\item{translate}{(tuple, optional): tuple of maximum absolute fraction for
horizontal and vertical translations. For example \code{translate=c(a, b)}, then
horizontal shift is randomly sampled in the range
-img_width * a < dx < img_width * a and vertical shift is randomly sampled
in the range -img_height * b < dy < img_height * b. Will not translate by
default.}

\item{scale}{(tuple, optional): scaling factor interval, e.g c(a, b), then
scale is randomly sampled from the range a <= scale <= b. Will keep
original scale by default.}

\item{shear}{(sequence or float or int, optional): Range of degrees to select
from. If shear is a number, a shear parallel to the x axis in the range
(-shear, +shear) will be applied. Else if shear is a tuple or list of 2
values a shear parallel to the x axis in the range \verb{(shear[1], shear[2])}
will be applied. Else if shear is a tuple or list of 4 values, a x-axis
shear in \verb{(shear[1], shear[2])} and y-axis shear in \verb{(shear[3], shear[4])}
will be applied. Will not apply shear by default.}

\item{interpolation}{(int or character, optional): Interpolation mode.
Supported values are 0 / "nearest" and 2 / "bilinear". Default is 0.}

\item{fill}{(tuple or int): Fill color for the area outside the transform.
Default is 0. This option is not supported for Tensor input.}

\item{resample}{Deprecated. Use interpolation instead.}

\item{fillcolor}{Deprecated. Use fill instead.}
}
\description{
Random affine transformation of the image keeping center invariant
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_affine}
\alias{transform_affine}
\title{Apply affine transformation on an image keeping image center invariant}
\usage{
transform_affine(
  img,
  angle,
  translate,
  scale,
  shear,
  interpolation = 0,
  fill = NULL,
  resample,
  fillcolor,
  center = NULL
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{angle}{(float or int): rotation angle value in degrees,
counter-clockwise.}

\item{translate}{(sequence of int) – horizontal and vertical translations
(post-rotation translation)}

\item{scale}{(float) – overall scale}

\item{shear}{(float or sequence) – shear angle value in degrees between -180 to 180,
clockwise direction. If a sequence is specified, the first value corresponds
to a shear parallel to the x-axis, while the second value corresponds to a
shear parallel to the y-axis.}

\item{interpolation}{(int or character): Interpolation mode.
Supported values are 0 / "nearest" and 2 / "bilinear". Default is 0.}

\item{fill}{Fill color for area outside the transform. Default is NULL.}

\item{resample}{Deprecated. Use interpolation instead.}

\item{fillcolor}{Deprecated. Use fill instead.}

\item{center}{Optional center of rotation, c(x, y). Default is image center.}
}
\description{
Apply affine transformation on an image keeping image center invariant
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collection-catalog.R
\name{search_collection}
\alias{search_collection}
\title{Search Collection Catalog}
\usage{
search_collection(keyword = NULL, collection = NULL)
}
\arguments{
\item{keyword}{Character string to search for (case-insensitive). Searches
in both dataset names and descriptions. If NULL, returns all datasets
(optionally filtered by collection).}

\item{collection}{Filter by collection name. One of: "biology", "medical",
"infrared", "damage", "underwater", "document". If NULL, searches all collections.}
}
\value{
A data frame with matching datasets and their metadata. Returns NULL
invisibly if no matches are found.
}
\description{
Search through all Collection datasets by keywords in name or description,
or filter by collection. This makes it easy to discover datasets relevant
to your task without browsing each collection individually.
}
\examples{
\dontrun{
# Find all medical datasets
search_collection(collection = "medical")

# Find datasets about cells
search_collection("cell")

# Find photovoltaic/solar datasets
search_collection("solar")
search_collection("photovoltaic")

# Find all biology datasets with "cell" in name/description
search_collection("cell", collection = "biology")

# List all available datasets
search_collection()
}

}
\seealso{
\code{\link[=get_collection_catalog]{get_collection_catalog()}}, \link{collection_catalog}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-boxes.R
\name{nms}
\alias{nms}
\title{Non-maximum Suppression (NMS)}
\usage{
nms(boxes, scores, iou_threshold)
}
\arguments{
\item{boxes}{(Tensor[N, 4])): boxes to perform NMS on. They are
expected to be in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
\itemize{
\item \eqn{0 \leq x_{min} < x_{max}} and
\item \eqn{0 \leq y_{min} < y_{max}}.
}}

\item{scores}{(Tensor[N]): scores for each one of the boxes}

\item{iou_threshold}{(float): discards all overlapping boxes with IoU > iou_threshold}
}
\value{
keep (Tensor): int64 tensor with the indices of the elements that
have been kept by NMS, sorted in decreasing order of scores.
}
\description{
Performs non-maximum suppression  (NMS) on the boxes according
to their intersection-over-union  (IoU). NMS iteratively removes
lower scoring boxes which have an IoU greater than iou_threshold
with another (higher scoring) box.
}
\details{
If multiple boxes have the exact same score and satisfy the IoU
criterion with respect to a reference box, the selected box is
not guaranteed to be the same between CPU and GPU. This is similar
to the behavior of argsort in torch when repeated values are present.

Current algorithm has a time complexity of O(n^2) and runs in native R.
It may be improve in the future by a Rcpp implementation or through alternative algorithm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-rf100-peixos.R
\name{rf100_peixos_segmentation_dataset}
\alias{rf100_peixos_segmentation_dataset}
\title{RF100 Peixos Segmentation Dataset}
\usage{
rf100_peixos_segmentation_dataset(
  split = c("train", "test", "valid"),
  root = tempdir(),
  download = FALSE,
  transform = NULL,
  target_transform = NULL
)
}
\arguments{
\item{split}{the subset of the dataset to choose between \code{c("train", "test", "valid")}.}

\item{root}{directory path to download the dataset.}

\item{download}{Logical. If TRUE, downloads the dataset if not present at \code{root}.}

\item{transform}{Optional transform function applied to the image.}

\item{target_transform}{Optional transform function applied to the target.}
}
\value{
A torch dataset. Each element is a named list with:
\itemize{
\item \code{x}: H × W × 3 array (use \code{transform_to_tensor()} in \code{transform} to get
C × H × W tensor).
\item \code{y}: a list with:
\itemize{
\item \code{masks}: boolean tensor of shape (1, H, W).
\item \code{labels}: integer vector with the class index (always 1 for "fish").
}
}

The returned item is given class \code{image_with_segmentation_mask} so it can be
visualised with helpers like \code{\link[=draw_segmentation_masks]{draw_segmentation_masks()}}.
}
\description{
Loads the Roboflow 100 "peixos" dataset for semantic segmentation.
"peixos" contains 3 splits of respectively 821 / 118 / 251 color images of size 640 x 640.
Segmentation masks are generated on-the-fly from polygon
annotations of the unique "fish" category.
}
\examples{
\dontrun{
ds <- rf100_peixos_segmentation_dataset(
  split = "train",
  transform = transform_to_tensor,
  download = TRUE
)
item <- ds[1]
overlay <- draw_segmentation_masks(item)
tensor_image_browse(overlay)
}

}
\seealso{
Other segmentation_dataset: 
\code{\link{coco_segmentation_dataset}()},
\code{\link{oxfordiiitpet_segmentation_dataset}()},
\code{\link{pascal_voc_datasets}}
}
\concept{segmentation_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-imagenet.R
\name{imagenet_classes}
\alias{imagenet_classes}
\alias{imagenet_1k_classes}
\alias{imagenet_21k_df}
\alias{imagenet_21k_classes}
\title{ImageNet Class Labels}
\usage{
imagenet_classes(class_id = 1:1000)

imagenet_1k_classes(class_id = 1:1000)

imagenet_21k_df(class_id = 1:21843)

imagenet_21k_classes(class_id)
}
\arguments{
\item{class_id}{Integer vector of 1-based class identifiers.}
}
\value{
A character vector with 1000 entries representing the ImageNet-1k
class labels.

A data.frame containing columns \code{id} and \code{label} representing
the ImageNet-21k class identifiers and labels. By default, returns all 21.8K rows.

A character vector with the labels associated with \code{class_id}.
}
\description{
Utilities for resolving ImageNet-1k class identifiers to their corresponding
human readable labels. The labels are retrieved from the same source used by
PyTorch's reference implementation.
}
\seealso{
Other class_resolution: 
\code{\link{caltech_classes}()},
\code{\link{coco_classes}()},
\code{\link{pascal_voc_classes}()}

Other class_resolution: 
\code{\link{caltech_classes}()},
\code{\link{coco_classes}()},
\code{\link{pascal_voc_classes}()}

Other class_resolution: 
\code{\link{caltech_classes}()},
\code{\link{coco_classes}()},
\code{\link{pascal_voc_classes}()}

Other class_resolution: 
\code{\link{caltech_classes}()},
\code{\link{coco_classes}()},
\code{\link{pascal_voc_classes}()}
}
\concept{class_resolution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-caltech.R
\name{caltech_dataset}
\alias{caltech_dataset}
\alias{caltech101_dataset}
\alias{caltech256_dataset}
\title{Caltech Datasets}
\usage{
caltech101_dataset(
  root = tempdir(),
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

caltech256_dataset(
  root = tempdir(),
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Character. Root directory for dataset storage. The dataset will be stored under \code{root/caltech256}.}

\item{transform}{Optional function to transform input images after loading. Default is \code{NULL}.}

\item{target_transform}{Optional function to transform labels. Default is \code{NULL}.}

\item{download}{Logical. Whether to download the dataset if not found locally. Default is \code{FALSE}.}
}
\value{
An object of class \code{caltech101_dataset}, which behaves like a torch dataset.
Each element is a named list with:
\itemize{
\item \code{x}: A H x W x 3 integer array representing an RGB image.
\item \code{y}: An Integer representing the label.
}

An object of class \code{caltech256_dataset}, which behaves like a torch dataset.
Each element is a named list with:
\itemize{
\item \code{x}: A H x W x 3 integer array representing an RGB image.
\item \code{y}: An Integer representing the label.
}
}
\description{
Caltech Datasets

Loads the Caltech-256 Object Category Dataset for image classification. It consists of 30,607 images across 256 distinct object categories.
Each category has at least 80 images, with variability in image size.
}
\details{
The Caltech-101 and Caltech-256 collections are \strong{classification} datasets
made of color images with varying sizes. They cover 101 and 256 object
categories respectively and are commonly used for evaluating visual
recognition models.

The Caltech-101 dataset contains around 9,000 images
spread over 101 object categories plus a background class. Images
have varying sizes.

Caltech-256 extends this to about 30,000 images
across 256 categories.
}
\examples{
\dontrun{
caltech101 <- caltech101_dataset(download = TRUE)

first_item <- caltech101[1]
first_item$x  # Image array
first_item$y  # Integer label
}

}
\seealso{
Other classification_dataset: 
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-efficientnet.R
\name{model_efficientnet}
\alias{model_efficientnet}
\alias{model_efficientnet_b0}
\alias{model_efficientnet_b1}
\alias{model_efficientnet_b2}
\alias{model_efficientnet_b3}
\alias{model_efficientnet_b4}
\alias{model_efficientnet_b5}
\alias{model_efficientnet_b6}
\alias{model_efficientnet_b7}
\title{EfficientNet Models}
\usage{
model_efficientnet_b0(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_b1(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_b2(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_b3(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_b4(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_b5(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_b6(pretrained = FALSE, progress = TRUE, ...)

model_efficientnet_b7(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{...}{Other parameters passed to the model implementation, such as
\code{num_classes} to change the output dimension.}
}
\description{
Constructs EfficientNet model architectures as described in
\emph{EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks}.
These models are designed for image classification tasks and provide a balance
between accuracy and computational efficiency through compound scaling.
}
\section{Functions}{
\itemize{
\item \code{model_efficientnet_b0()}: EfficientNet B0 model

\item \code{model_efficientnet_b1()}: EfficientNet B1 model

\item \code{model_efficientnet_b2()}: EfficientNet B2 model

\item \code{model_efficientnet_b3()}: EfficientNet B3 model

\item \code{model_efficientnet_b4()}: EfficientNet B4 model

\item \code{model_efficientnet_b5()}: EfficientNet B5 model

\item \code{model_efficientnet_b6()}: EfficientNet B6 model

\item \code{model_efficientnet_b7()}: EfficientNet B7 model

}}
\section{Task}{

Image classification with 1000 output classes by default (ImageNet).
}

\section{Input Format}{

The models expect input tensors of shape \code{(batch_size, 3, H, W)}, where H and W
should typically be 224 for B0 and scaled versions for B1–B7 (e.g., B7 uses 600x600).
}

\section{Variants and Scaling}{

\tabular{lllllll}{
\strong{Model} \tab \strong{Width} \tab \strong{Depth} \tab
\strong{Resolution} \tab \strong{Params (M)} \tab \strong{GFLOPs} \tab
\strong{Top-1 Acc.} \cr
B0 \tab 1.0 \tab 1.0 \tab 224 \tab 5.3 \tab 0.39 \tab 77.1 \cr
B1 \tab 1.0 \tab 1.1 \tab 240 \tab 7.8 \tab 0.70 \tab 79.1 \cr
B2 \tab 1.1 \tab 1.2 \tab 260 \tab 9.2 \tab 1.00 \tab 80.1 \cr
B3 \tab 1.2 \tab 1.4 \tab 300 \tab 12.0 \tab 1.80 \tab 81.6 \cr
B4 \tab 1.4 \tab 1.8 \tab 380 \tab 19.0 \tab 4.20 \tab 82.9 \cr
B5 \tab 1.6 \tab 2.2 \tab 456 \tab 30.0 \tab 9.90 \tab 83.6 \cr
B6 \tab 1.8 \tab 2.6 \tab 528 \tab 43.0 \tab 19.0 \tab 84.0 \cr
B7 \tab 2.0 \tab 3.1 \tab 600 \tab 66.0 \tab 37.0 \tab 84.3
}
}

\examples{
\dontrun{
model <- model_efficientnet_b0()
image_batch <- torch::torch_randn(1, 3, 224, 224)
output <- model(image_batch)
imagenet_classes(which.max(as.numeric(output)))
}

\dontrun{
# Example of using EfficientNet-B5 with its native image size
model <- model_efficientnet_b5()
image_batch <- torch::torch_randn(1, 3, 456, 456)
output <- model(image_batch)
imagenet_classes(which.max(as.numeric(output)))
}

}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_linear_transformation}
\alias{transform_linear_transformation}
\title{Transform a tensor image with a square transformation matrix and a
mean_vector computed offline}
\usage{
transform_linear_transformation(img, transformation_matrix, mean_vector)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{transformation_matrix}{(Tensor): tensor \verb{[D x D]}, D = C x H x W.}

\item{mean_vector}{(Tensor): tensor \link{D}, D = C x H x W.}
}
\description{
Given \code{transformation_matrix} and \code{mean_vector}, will flatten the
\code{torch_tensor} and subtract \code{mean_vector} from it which is then followed by
computing the dot product with the transformation matrix and then reshaping
the tensor to its original shape.
}
\section{Applications}{

whitening transformation: Suppose X is a column vector zero-centered data.
Then compute the data covariance matrix \verb{[D x D]} with torch.mm(X.t(), X),
perform SVD on this matrix and pass it as \code{transformation_matrix}.
}

\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_choice}
\alias{transform_random_choice}
\title{Apply single transformation randomly picked from a list}
\usage{
transform_random_choice(img, transforms)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{transforms}{(list or tuple): list of transformations.}
}
\description{
Apply single transformation randomly picked from a list
}
\seealso{
Other combining_transforms: 
\code{\link{transform_five_crop}()},
\code{\link{transform_random_apply}()},
\code{\link{transform_random_order}()},
\code{\link{transform_resized_crop}()},
\code{\link{transform_ten_crop}()}
}
\concept{combining_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-coco.R
\name{coco_classes}
\alias{coco_classes}
\title{MS COCO Class Labels}
\usage{
coco_classes(class_id = 1:90)
}
\arguments{
\item{class_id}{Integer vector of 1-based class identifiers.}
}
\value{
A character vector with the COCO class names
}
\description{
Utilities for resolving COCO 90 class identifiers to their corresponding
human readable labels. The labels are retrieved from pytorch/vision source to be compliant
with torchvision pretrained models.
}
\seealso{
Other class_resolution: 
\code{\link{caltech_classes}()},
\code{\link{imagenet_classes}()},
\code{\link{pascal_voc_classes}()}
}
\concept{class_resolution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_adjust_contrast}
\alias{transform_adjust_contrast}
\title{Adjust the contrast of an image}
\usage{
transform_adjust_contrast(img, contrast_factor)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{contrast_factor}{(float): How much to adjust the contrast. Can be any
non negative number. 0 gives a solid gray image, 1 gives the
original image while 2 increases the contrast by a factor of 2.}
}
\description{
Adjust the contrast of an image
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_rgb_to_grayscale}
\alias{transform_rgb_to_grayscale}
\title{Convert RGB Image Tensor to Grayscale}
\usage{
transform_rgb_to_grayscale(img)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}
}
\description{
For RGB to Grayscale conversion, ITU-R 601-2 luma transform is performed
which is L = R * 0.2989 + G * 0.5870 + B * 0.1140
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-mobilenetv3.R,
%   R/models-mobilenetv3_large.R
\name{model_mobilenet_v3}
\alias{model_mobilenet_v3}
\alias{model_mobilenet_v3_large}
\alias{model_mobilenet_v3_small}
\alias{model_mobilenet_v3_large_quantized}
\title{MobileNetV3 Model}
\usage{
model_mobilenet_v3_large(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 1000,
  width_mult = 1
)

model_mobilenet_v3_small(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 1000,
  width_mult = 1
)

model_mobilenet_v3_large_quantized(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{num_classes}{number of output classes (default: 1000).}

\item{width_mult}{width multiplier for model scaling (default: 1.0).}

\item{...}{Other parameters passed to the model implementation.}
}
\description{
MobileNetV3 is a state-of-the-art lightweight convolutional neural network architecture
designed for mobile and embedded vision applications. This implementation follows the
design and optimizations presented in the original paper:\href{https://arxiv.org/abs/1905.02244}{MobileNetV3: Searching for MobileNetV3}

This function mirrors \verb{torchvision::quantization::mobilenet_v3_large} and
loads quantized weights when \code{pretrained} is \code{TRUE}.
}
\details{
The model includes two variants:
\itemize{
\item \code{model_mobilenet_v3_large()}
\item \code{model_mobilenet_v3_small()}
}

Both variants utilize efficient blocks such as inverted residuals, squeeze-and-excitation (SE) modules,
and hard-swish activations for improved accuracy and efficiency.
\subsection{Model Summary and Performance for pretrained weights}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{| Model                  | Top-1 Acc | Top-5 Acc | Params  | GFLOPS | File Size | Notes                               |
|------------------------|-----------|-----------|---------|--------|-----------|-------------------------------------|
| MobileNetV3 Large      | 74.04\%    | 91.34\%    | 5.48M   | 0.22   | 21.1 MB   | Trained from scratch, simple recipe |
| MobileNetV3 Small      | 67.67\%    | 87.40\%    | 2.54M   | 0.06   | 9.8 MB    | Improved recipe over original paper |
}\if{html}{\out{</div>}}
}
}
\section{Functions}{
\itemize{
\item \code{model_mobilenet_v3_large()}: MobileNetV3 Large model with about 5.5 million parameters.

\item \code{model_mobilenet_v3_small()}: MobileNetV3 Small model with about 2.5 million parameters.

}}
\examples{
\dontrun{
# 1. Download sample image (dog)
norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants, see
# https://pytorch.org/vision/stable/models.html
norm_std  <- c(0.229, 0.224, 0.225)
img_url <- "https://en.wikipedia.org/wiki/Special:FilePath/Felis_catus-cat_on_snow.jpg"
img <- base_loader(img_url)

# 2. Convert to tensor (RGB only), resize and normalize
input <- img \%>\%
 transform_to_tensor() \%>\%
 transform_resize(c(224, 224)) \%>\%
 transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)

# 3. Load pretrained models
model_small <- model_mobilenet_v3_small(pretrained = TRUE)
model_small$eval()

# 4. Forward pass
output_s <- model_small(batch)

# 5. Top-5 printing helper
topk <- output_s$topk(k = 5, dim = 2)
indices <- as.integer(topk[[2]][1, ])
scores <- as.numeric(topk[[1]][1, ])

# 6. Show Top-5 predictions
glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}\%)")

# 7. Same with large model
model_large <- model_mobilenet_v3_large(pretrained = TRUE)
model_large$eval()
output_l <- model_large(input)
topk <- output_l$topk(k = 5, dim = 2)
indices <- as.integer(topk[[2]][1, ])
scores <- as.numeric(topk[[1]][1, ])
glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}\%)")
}

}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}

Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-caltech.R
\name{caltech_classes}
\alias{caltech_classes}
\title{Caltech Class Labels}
\usage{
caltech_classes(class_id = 1:257)
}
\arguments{
\item{class_id}{Integer vector of 1-based class identifiers.}
}
\value{
A character vector with 257 entries representing the Caltech 257
class labels.
}
\description{
Utilities for resolving Caltech class identifiers to their corresponding
human readable labels.
}
\seealso{
Other class_resolution: 
\code{\link{coco_classes}()},
\code{\link{imagenet_classes}()},
\code{\link{pascal_voc_classes}()}
}
\concept{class_resolution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collection-catalog.R
\name{get_collection_catalog}
\alias{get_collection_catalog}
\title{Get Complete Collection Catalog}
\usage{
get_collection_catalog()
}
\value{
A data frame with all datasets and their metadata.
}
\description{
Returns the complete catalog of datasets in collections with their metadata.
This is a convenience function that loads and returns the collection_catalog data.
}
\examples{
\dontrun{
# Get complete catalog
catalog <- get_collection_catalog()

# View in RStudio
View(catalog)

# Summary statistics
summary(catalog$total_size_mb)
table(catalog$collection)

# Find smallest dataset
catalog[which.min(catalog$total_size_mb), ]

# Find largest dataset
catalog[which.max(catalog$total_size_mb), ]
}

}
\seealso{
\code{\link[=search_collection]{search_collection()}}, \link{collection_catalog}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_perspective}
\alias{transform_random_perspective}
\title{Random perspective transformation of an image with a given probability}
\usage{
transform_random_perspective(
  img,
  distortion_scale = 0.5,
  p = 0.5,
  interpolation = 2,
  fill = 0
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{distortion_scale}{(float): argument to control the degree of distortion
and ranges from 0 to 1. Default is 0.5.}

\item{p}{(float): probability of the image being transformed. Default is 0.5.}

\item{interpolation}{(int, optional) Desired interpolation. An integer
\code{0 = nearest}, \code{2 = bilinear}, and \code{3 = bicubic} or a name from
\code{\link[magick:filter_types]{magick::filter_types()}}.}

\item{fill}{(int or str or tuple): Pixel fill value for constant fill.
Default is 0. If a tuple of length 3, it is used to fill R, G, B channels
respectively. This value is only used when the padding_mode is constant.
Only int value is supported for Tensors.}
}
\description{
Performs a random perspective transformation of the given image with a given
probability
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_affine}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-boxes.R
\name{clip_boxes_to_image}
\alias{clip_boxes_to_image}
\title{Clip Boxes to Image}
\usage{
clip_boxes_to_image(boxes, size)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format
with
\itemize{
\item \eqn{0 \leq x_{min} < x_{max}} and
\item \eqn{0 \leq y_{min} < y_{max}}.
}}

\item{size}{(Tuple[height, width]): size of the image}
}
\value{
clipped_boxes (Tensor[N, 4])
}
\description{
Clip boxes so that they lie inside an image of size \code{size}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-convnext_segmentation.R
\name{model_convnext_segmentation}
\alias{model_convnext_segmentation}
\alias{model_convnext_tiny_fcn}
\alias{model_convnext_small_fcn}
\alias{model_convnext_base_fcn}
\alias{model_convnext_tiny_upernet}
\alias{model_convnext_small_upernet}
\alias{model_convnext_base_upernet}
\title{ConvNeXt Segmentation Models}
\usage{
model_convnext_tiny_fcn(
  num_classes = 21,
  aux_loss = FALSE,
  pretrained_backbone = FALSE,
  ...
)

model_convnext_small_fcn(
  num_classes = 21,
  aux_loss = FALSE,
  pretrained_backbone = FALSE,
  ...
)

model_convnext_base_fcn(
  num_classes = 21,
  aux_loss = FALSE,
  pretrained_backbone = FALSE,
  ...
)

model_convnext_tiny_upernet(
  num_classes = 21,
  aux_loss = FALSE,
  pretrained = FALSE,
  pretrained_backbone = FALSE,
  pool_scales = c(1, 2, 3, 6),
  ...
)

model_convnext_small_upernet(
  num_classes = 21,
  aux_loss = FALSE,
  pretrained = FALSE,
  pretrained_backbone = FALSE,
  pool_scales = c(1, 2, 3, 6),
  ...
)

model_convnext_base_upernet(
  num_classes = 21,
  aux_loss = FALSE,
  pretrained = FALSE,
  pretrained_backbone = FALSE,
  pool_scales = c(1, 2, 3, 6),
  ...
)
}
\arguments{
\item{num_classes}{Number of output segmentation classes. Default: 21 (PASCAL VOC).}

\item{aux_loss}{If TRUE, includes an auxiliary classifier branch. Default: FALSE.}

\item{pretrained_backbone}{If TRUE, loads ImageNet pretrained
weights for the ConvNeXt backbone. Default: FALSE.}

\item{...}{Additional arguments passed to the backbone.}

\item{pretrained}{If TRUE, loads convnext pretrained weights of backbone and segmentation heads.}

\item{pool_scales}{Numeric vector. Pooling scales used in the Pyramid Pooling
Module for UPerNet models. Default: c(1, 2, 3, 6).}
}
\value{
An \code{nn_module} representing the segmentation model.
}
\description{
Semantic segmentation models that use a ConvNeXt backbone with either
an FCN (Fully Convolutional Network) head or a UPerNet (Unified Perceptual
Parsing Network) head.

These models follow the architecture patterns from mmsegmentation and
can be used for semantic segmentation tasks.
}
\section{Functions}{
\itemize{
\item \code{model_convnext_tiny_fcn()}: ConvNeXt Tiny with FCN head

\item \code{model_convnext_small_fcn()}: ConvNeXt Small with FCN head

\item \code{model_convnext_base_fcn()}: ConvNeXt Base with FCN head

\item \code{model_convnext_tiny_upernet()}: ConvNeXt Tiny with UPerNet head

\item \code{model_convnext_small_upernet()}: ConvNeXt Small with UPerNet head

\item \code{model_convnext_base_upernet()}: ConvNeXt Base with UPerNet head

}}
\section{Available FCN Models}{

\itemize{
\item \code{model_convnext_tiny_fcn()}
\item \code{model_convnext_small_fcn()}
\item \code{model_convnext_base_fcn()}
}
}

\section{Available UPerNet Models}{

\itemize{
\item \code{model_convnext_tiny_upernet()}
\item \code{model_convnext_small_upernet()}
\item \code{model_convnext_base_upernet()}
}
}

\examples{
\dontrun{
library(magrittr)
norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants
norm_std  <- c(0.229, 0.224, 0.225)

# Use a publicly available image
wmc <- "https://upload.wikimedia.org/wikipedia/commons/thumb/"
url <- "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg"
img <- base_loader(paste0(wmc, url))

input <- img \%>\%
  transform_to_tensor() \%>\%
  transform_resize(c(520, 520)) \%>\%
  transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)

# ConvNeXt Tiny FCN segmentation
model <- model_convnext_tiny_fcn(num_classes = 21, pretrained_backbone = TRUE)
model$eval()
output <- model(batch)

# Visualize result
segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
tensor_image_display(segmented)

# ConvNeXt Tiny UPerNet segmentation
model <- model_convnext_tiny_upernet(num_classes = 21, pretrained_backbone = TRUE)
model$eval()
output <- model(batch)

# Visualize result
segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
tensor_image_display(segmented)
}

}
\seealso{
Other semantic_segmentation_model: 
\code{\link{model_deeplabv3}},
\code{\link{model_fcn_resnet}}
}
\concept{semantic_segmentation_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_ten_crop}
\alias{transform_ten_crop}
\title{Crop an image and the flipped image each into four corners and a central crop}
\usage{
transform_ten_crop(img, size, vertical_flip = FALSE)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{size}{(sequence or int): Desired output size. If size is a sequence
like c(h, w), output size will be matched to this. If size is an int,
smaller edge of the image will be matched to this number.
i.e, if height > width, then image will be rescaled to
(size * height / width, size).}

\item{vertical_flip}{(bool): Use vertical flipping instead of horizontal}
}
\description{
Crop the given image into four corners and the central crop, plus the flipped
version of these (horizontal flipping is used by default). This transform
returns a tuple of images and there may be a mismatch in the number of
inputs and targets your Dataset returns.
}
\seealso{
Other combining_transforms: 
\code{\link{transform_five_crop}()},
\code{\link{transform_random_apply}()},
\code{\link{transform_random_choice}()},
\code{\link{transform_random_order}()},
\code{\link{transform_resized_crop}()}
}
\concept{combining_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-flickr.R
\name{flickr_caption_dataset}
\alias{flickr_caption_dataset}
\alias{flickr8k_caption_dataset}
\alias{flickr30k_caption_dataset}
\title{Flickr Caption Datasets}
\usage{
flickr8k_caption_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

flickr30k_caption_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Character. Root directory where the dataset will be stored under \code{root/flickr30k}.}

\item{train}{: If \code{TRUE}, loads the training set. If \code{FALSE}, loads the test set. Default is \code{TRUE}.}

\item{transform}{Optional function to transform input images after loading. Default is \code{NULL}.}

\item{target_transform}{Optional function to transform labels. Default is \code{NULL}.}

\item{download}{Logical. Whether to download the dataset if not found locally. Default is \code{FALSE}.}
}
\value{
A torch dataset of class \code{flickr8k_caption_dataset}.
Each element is a named list:
\itemize{
\item \code{x}: a H x W x 3 integer array representing an RGB image.
\item \code{y}: a character vector containing all five captions associated with the image.
}

A torch dataset of class \code{flickr30k_caption_dataset}.
Each element is a named list:
\itemize{
\item \code{x}: a H x W x 3 integer array representing an RGB image.
\item \code{y}: a character vector containing all five captions associated with the image.
}
}
\description{
Flickr8k Dataset
}
\details{
The Flickr8k and Flickr30k collections are \strong{image captionning} datasets
composed of 8,000 and 30,000 color images respectively, each paired with five
human-annotated captions. The images are in RGB format with varying spatial
resolutions, and these datasets are widely used for training and evaluating
vision-language models.
}
\examples{
\dontrun{
# Load the Flickr8k caption dataset
flickr8k <- flickr8k_caption_dataset(download = TRUE)

# Access the first item
first_item <- flickr8k[1]
first_item$x  # image array with shape {3, H, W}
first_item$y  # character vector containing five captions.

# Load the Flickr30k caption dataset
flickr30k <- flickr30k_caption_dataset(download = TRUE)

# Access the first item
first_item <- flickr30k[1]
first_item$x  # image array with shape {3, H, W}
first_item$y  # character vector containing five captions.
}

}
\seealso{
Other caption_dataset: 
\code{\link{coco_caption_dataset}()}
}
\concept{caption_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-vgg.R
\name{model_vgg}
\alias{model_vgg}
\alias{model_vgg11}
\alias{model_vgg11_bn}
\alias{model_vgg13}
\alias{model_vgg13_bn}
\alias{model_vgg16}
\alias{model_vgg16_bn}
\alias{model_vgg19}
\alias{model_vgg19_bn}
\title{VGG implementation}
\usage{
model_vgg11(pretrained = FALSE, progress = TRUE, ...)

model_vgg11_bn(pretrained = FALSE, progress = TRUE, ...)

model_vgg13(pretrained = FALSE, progress = TRUE, ...)

model_vgg13_bn(pretrained = FALSE, progress = TRUE, ...)

model_vgg16(pretrained = FALSE, progress = TRUE, ...)

model_vgg16_bn(pretrained = FALSE, progress = TRUE, ...)

model_vgg19(pretrained = FALSE, progress = TRUE, ...)

model_vgg19_bn(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet}

\item{progress}{(bool): If TRUE, displays a progress bar of the download
to stderr}

\item{...}{other parameters passed to the VGG model implementation.}
}
\description{
VGG models implementations based on
\href{https://arxiv.org/pdf/1409.1556}{Very Deep Convolutional Networks For Large-Scale Image Recognition}
}
\section{Functions}{
\itemize{
\item \code{model_vgg11()}: VGG 11-layer model (configuration "A")

\item \code{model_vgg11_bn()}: VGG 11-layer model (configuration "A") with batch normalization

\item \code{model_vgg13()}: VGG 13-layer model (configuration "B")

\item \code{model_vgg13_bn()}: VGG 13-layer model (configuration "B") with batch normalization

\item \code{model_vgg16()}: VGG 13-layer model (configuration "D")

\item \code{model_vgg16_bn()}: VGG 13-layer model (configuration "D") with batch normalization

\item \code{model_vgg19()}: VGG 19-layer model (configuration "E")

\item \code{model_vgg19_bn()}: VGG 19-layer model (configuration "E") with batch normalization

}}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/folder-dataset.R
\name{image_folder_dataset}
\alias{image_folder_dataset}
\title{Create an image folder dataset}
\usage{
image_folder_dataset(
  root,
  transform = NULL,
  target_transform = NULL,
  loader = NULL,
  is_valid_file = NULL
)
}
\arguments{
\item{root}{Root directory path.}

\item{transform}{A function/transform that takes in an PIL image and returns
a transformed version. E.g, \code{\link[=transform_random_crop]{transform_random_crop()}}.}

\item{target_transform}{A function/transform that takes in the target and
transforms it.}

\item{loader}{A function to load an image given its path.}

\item{is_valid_file}{A function that takes path of an Image file and check if
the file is a valid file (used to check of corrupt files)}
}
\description{
A generic data loader for images stored in folders.
See \code{Details} for more information.
}
\details{
This function assumes that the images for each class are contained
in subdirectories of \code{root}. The names of these subdirectories are stored
in the \code{classes} attribute of the returned object.

An example folder structure might look as follows:

\if{html}{\out{<div class="sourceCode">}}\preformatted{root/dog/xxx.png
root/dog/xxy.png
root/dog/xxz.png

root/cat/123.png
root/cat/nsdf3.png
root/cat/asd932_.png
}\if{html}{\out{</div>}}
}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-vit.R
\name{model_vit}
\alias{model_vit}
\alias{model_vit_b_16}
\alias{model_vit_b_32}
\alias{model_vit_l_16}
\alias{model_vit_l_32}
\alias{model_vit_h_14}
\title{Vision Transformer Implementation}
\usage{
model_vit_b_16(pretrained = FALSE, progress = TRUE, ...)

model_vit_b_32(pretrained = FALSE, progress = TRUE, ...)

model_vit_l_16(pretrained = FALSE, progress = TRUE, ...)

model_vit_l_32(pretrained = FALSE, progress = TRUE, ...)

model_vit_h_14(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{...}{Other parameters passed to the model implementation.}
}
\description{
Vision Transformer (ViT) models implement the architecture proposed in the paper
\href{https://arxiv.org/abs/2010.11929}{An Image is Worth 16x16 Words}.
These models are designed for image classification tasks and operate by treating
image patches as tokens in a Transformer model.
}
\details{
\subsection{Model Variants and Performance (ImageNet-1k)}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{| Model     | Top-1 Acc | Top-5 Acc | Params  | GFLOPS | File Size | Weights Used              | Notes                  |
|-----------|-----------|-----------|---------|--------|-----------|---------------------------|------------------------|
| vit_b_16  | 81.1\%     | 95.3\%     | 86.6M   | 17.56  | 346 MB    | IMAGENET1K_V1             | Base, 16x16 patches    |
| vit_b_32  | 75.9\%     | 92.5\%     | 88.2M   | 4.41   | 353 MB    | IMAGENET1K_V1             | Base, 32x32 patches    |
| vit_l_16  | 79.7\%     | 94.6\%     | 304.3M  | 61.55  | 1.22 GB   | IMAGENET1K_V1             | Large, 16x16 patches   |
| vit_l_32  | 77.0\%     | 93.1\%     | 306.5M  | 15.38  | 1.23 GB   | IMAGENET1K_V1             | Large, 32x32 patches   |
| vit_h_14  | 88.6\%     | 98.7\%     | 633.5M  | 1016.7 | 2.53 GB   | IMAGENET1K_SWAG_E2E_V1    | Huge, 14x14 patches    |
}\if{html}{\out{</div>}}
\itemize{
\item \strong{TorchVision Recipe}: \url{https://github.com/pytorch/vision/tree/main/references/classification}
\item \strong{SWAG Recipe}: \url{https://github.com/facebookresearch/SWAG}
}

\strong{Weights Selection}:
\itemize{
\item All models use the default \code{IMAGENET1K_V1} weights for consistency, stability, and official support from TorchVision.
\item These are supervised weights trained on ImageNet-1k.
\item For \code{vit_h_14}, the default weight is \code{IMAGENET1K_SWAG_E2E_V1}, pretrained on SWAG and fine-tuned on ImageNet.
}
}
}
\section{Functions}{
\itemize{
\item \code{model_vit_b_16()}: ViT-B/16 model (Base, 16×16 patch size)

\item \code{model_vit_b_32()}: ViT-B/32 model (Base, 32×32 patch size)

\item \code{model_vit_l_16()}: ViT-L/16 model (Base, 16×16 patch size)

\item \code{model_vit_l_32()}: ViT-L/32 model (Base, 32×32 patch size)

\item \code{model_vit_h_14()}: ViT-H/14 model (Base, 14×14 patch size)

}}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vision_utils.R
\name{tensor_image_browse}
\alias{tensor_image_browse}
\title{Display image tensor}
\usage{
tensor_image_browse(image, browser = getOption("browser"))
}
\arguments{
\item{image}{\code{torch_tensor()} of shape (1, W, H) for grayscale image or (3, W, H) for
color image to display}

\item{browser}{argument passed to \link{browseURL}}
}
\description{
Display image tensor into browser
}
\seealso{
Other image display: 
\code{\link{draw_bounding_boxes}()},
\code{\link{draw_keypoints}()},
\code{\link{draw_segmentation_masks}()},
\code{\link{tensor_image_display}()},
\code{\link{vision_make_grid}()}
}
\concept{image display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_adjust_gamma}
\alias{transform_adjust_gamma}
\title{Adjust the gamma of an RGB image}
\usage{
transform_adjust_gamma(img, gamma, gain = 1)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{gamma}{(float): Non negative real number, same as \eqn{\gamma} in the
equation. gamma larger than 1 make the shadows darker, while gamma smaller
than 1 make dark regions lighter.}

\item{gain}{(float): The constant multiplier.}
}
\description{
Also known as Power Law Transform. Intensities in RGB mode are adjusted
based on the following equation:
\deqn{
    I_{\mbox{out}} = 255 \times \mbox{gain} \times \left
    (\frac{I_{\mbox{in}}}{255}\right)^{\gamma}
    }
}
\details{
Search for Gamma Correction for more details.
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-deeplabv3.R
\name{model_deeplabv3}
\alias{model_deeplabv3}
\alias{model_deeplabv3_resnet50}
\alias{model_deeplabv3_resnet101}
\title{DeepLabV3 Semantic Segmentation Models}
\usage{
model_deeplabv3_resnet50(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 21,
  aux_loss = NULL,
  pretrained_backbone = FALSE,
  ...
)

model_deeplabv3_resnet101(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 21,
  aux_loss = NULL,
  pretrained_backbone = FALSE,
  ...
)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{num_classes}{Integer. Number of output segmentation classes including
background. Default: \code{21} (Pascal VOC). Set to \code{NULL} to infer from
pretrained weights.}

\item{aux_loss}{Logical or \code{NULL}. If \code{TRUE}, adds an auxiliary FCN classifier
head at an intermediate backbone layer, used as a secondary loss during
training. If \code{NULL} (default), inferred from pretrained weights.}

\item{pretrained_backbone}{Logical. If \code{TRUE} and \code{pretrained = FALSE}, loads
\code{IMAGENET1K_V1} weights for the backbone only. Ignored when \code{pretrained = TRUE}.
Default: \code{TRUE}.}

\item{...}{Other parameters passed to the resnet model.}
}
\description{
Semantic segmentation models implementing the DeepLabV3 architecture from
\href{https://arxiv.org/abs/1706.05587}{Rethinking Atrous Convolution for Semantic Image Segmentation}.
These models use Atrous Spatial Pyramid Pooling (ASPP) to capture multi-scale
context, and are available with ResNet-50 and ResNet-101 backbones.
\subsection{Available Models}{
\itemize{
\item \code{model_deeplabv3_resnet50()}
\item \code{model_deeplabv3_resnet101()}
}
}

\subsection{Model Variants and Performance (COCO val2017, VOC labels)}{

All models are trained on a 20-class subset of COCO that corresponds to
Pascal VOC categories, plus background (21 classes total).

\if{html}{\out{<div class="sourceCode">}}\preformatted{| Model                     | mIoU  | Pixel Acc | Params | GFLOPS | File Size | Weights Used              |
|---------------------------|-------|-----------|--------|--------|-----------|---------------------------|
| model_deeplabv3_resnet50  | 66.4\% | 92.4\%     | 42.0M  | 178.72 | 161 MB    | COCO_WITH_VOC_LABELS_V1   |
| model_deeplabv3_resnet101 | 67.4\% | 92.4\%     | 61.0M  | 258.74 | 233 MB    | COCO_WITH_VOC_LABELS_V1   |
}\if{html}{\out{</div>}}
}

\subsection{Weights Selection}{
\itemize{
\item All models use \code{COCO_WITH_VOC_LABELS_V1} weights, trained on COCO with the
20 Pascal VOC categories (+ background = 21 classes).
\item Backbone weights default to \code{IMAGENET1K_V1} (supervised ImageNet-1k) when
\code{pretrained = FALSE} and \code{pretrained_backbone = TRUE}.
\item When \code{pretrained = TRUE}, backbone weights are overridden by the full
segmentation model weights and \code{pretrained_backbone} is ignored.
\item The auxiliary classifier branch (\code{aux_loss}) is automatically enabled when
loading pretrained weights; set explicitly when training from scratch.
}
}

\subsection{Input Format}{

Models expect input tensors of shape \verb{(batch_size, 3, H, W)}, normalized
with ImageNet mean \code{c(0.485, 0.456, 0.406)} and std \code{c(0.229, 0.224, 0.225)}.
Training resolution is 520x520.
}

\subsection{Output Format}{

Returns a named list with:
\itemize{
\item \verb{$out} — main segmentation logits, shape \verb{(batch, num_classes, H, W)}
\item \verb{$aux} — auxiliary logits from an intermediate backbone layer (only when \code{aux_loss = TRUE})
}
}
}
\section{Functions}{
\itemize{
\item \code{model_deeplabv3_resnet50()}: DeepLabV3 with ResNet-50 backbone

\item \code{model_deeplabv3_resnet101()}: DeepLabV3 with ResNet-101 backbone

}}
\examples{
\dontrun{
library(magrittr)
norm_mean <- c(0.485, 0.456, 0.406)
norm_std  <- c(0.229, 0.224, 0.225)

url <- paste0("https://upload.wikimedia.org/wikipedia/commons/thumb/",
           "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg")
img <- base_loader(url)

input <- img \%>\%
  transform_to_tensor() \%>\%
  transform_resize(c(520, 520)) \%>\%
  transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)    # Add batch dimension: (1, 3, H, W)

# --- ResNet-50 backbone ---
model <- model_deeplabv3_resnet50(pretrained = TRUE)
model$eval()
output <- model(batch)

segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
tensor_image_browse(segmented)

# Show most frequent class
mask_id <- output$out$argmax(dim = 2)  # (1, H, W)
class_contingency_with_background <- mask_id$view(-1)$bincount()
class_contingency_with_background[1] <- 0L # we clean the counter for background class id 1
top_class_index <- class_contingency_with_background$argmax()$item()
cli::cli_inform("Majority class {.pkg ResNet-50}: {.emph {pascal_voc_classes(top_class_index)}}")

# --- ResNet-101 backbone ---
model <- model_deeplabv3_resnet101(pretrained = TRUE)
model$eval()
output <- model(batch)

segmented <- draw_segmentation_masks(input, output$out$squeeze(1))
tensor_image_browse(segmented)

# Show most frequent class
mask_id <- output$out$argmax(dim = 2)  # (1, H, W)
class_contingency_with_background <- mask_id$view(-1)$bincount()
class_contingency_with_background[1] <- 0L # we clean the counter for background class id 1
top_class_index <- class_contingency_with_background$argmax()$item()
cli::cli_inform("Majority class {.pkg ResNet-50}: {.emph {pascal_voc_classes(top_class_index)}}")
}

}
\seealso{
Other semantic_segmentation_model: 
\code{\link{model_convnext_segmentation}},
\code{\link{model_fcn_resnet}}
}
\concept{semantic_segmentation_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-coco.R
\name{coco_segmentation_dataset}
\alias{coco_segmentation_dataset}
\title{COCO Segmentation Dataset}
\usage{
coco_segmentation_dataset(
  root = tempdir(),
  train = TRUE,
  year = c("2017", "2014"),
  download = FALSE,
  transform = NULL,
  target_transform = NULL
)
}
\arguments{
\item{root}{Root directory where the dataset is stored or will be downloaded to.}

\item{train}{Logical. If TRUE, loads the training split; otherwise, loads the validation split.}

\item{year}{Character. Dataset version year. One of \code{"2014"} or \code{"2017"}.}

\item{download}{Logical. If TRUE, downloads the dataset if it's not already present in the \code{root} directory.}

\item{transform}{Optional transform function applied to the image.}

\item{target_transform}{Optional transform function applied to the target.
Use \code{target_transform_coco_masks} to convert polygon annotations to binary masks.}
}
\value{
An object of class \code{coco_segmentation_dataset}. Each item is a list:
\itemize{
\item \code{x}: a \verb{(C, H, W)} array representing the image.
\item \code{y$labels}: an integer \code{torch_tensor} with the class label for each object.
\item \code{y$iscrowd}: a boolean \code{torch_tensor}, where \code{TRUE} marks the object as part of a crowd.
\item \code{y$segmentation}: a list of segmentation polygons for each object.
\item \code{y$masks}: a \verb{(N, H, W)} boolean \code{torch_tensor} containing binary segmentation masks (when using target_transform_coco_masks).
}

The returned object has S3 class \code{"image_with_segmentation_mask"}
to enable automatic dispatch by visualization functions such as \code{draw_segmentation_masks()}.

For object detection tasks without segmentation, use \code{\link{coco_detection_dataset}} instead.
}
\description{
Loads the MS COCO dataset for instance segmentation tasks.
}
\details{
The returned image \code{x} is in CHW format (channels, height, width), matching the torch convention.
The dataset \code{y} offers instance segmentation annotations including labels,
crowd indicators, and segmentation masks from the official COCO annotations.

Files are downloaded to a \code{coco} subdirectory in the torch cache directory for better organization.
}
\examples{
\dontrun{
# Load dataset for instance segmentation
ds <- coco_segmentation_dataset(
  train = FALSE,
  year = "2017",
  download = TRUE,
  target_transform = target_transform_coco_masks
)

item <- ds[1]

# Visualize segmentation masks
masked <- draw_segmentation_masks(item)
tensor_image_browse(masked)
}
}
\seealso{
\code{\link{coco_detection_dataset}} for object detection tasks

Other segmentation_dataset: 
\code{\link{oxfordiiitpet_segmentation_dataset}()},
\code{\link{pascal_voc_datasets}},
\code{\link{rf100_peixos_segmentation_dataset}()}
}
\concept{segmentation_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_perspective}
\alias{transform_perspective}
\title{Perspective transformation of an image}
\usage{
transform_perspective(
  img,
  startpoints,
  endpoints,
  interpolation = 2,
  fill = NULL
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{startpoints}{(list of list of ints): List containing four lists of two
integers corresponding to four corners
\verb{[top-left, top-right, bottom-right, bottom-left]} of the original image.}

\item{endpoints}{(list of list of ints): List containing four lists of two
integers corresponding to four corners
\verb{[top-left, top-right, bottom-right, bottom-left]} of the transformed
image.}

\item{interpolation}{(int, optional) Desired interpolation. An integer
\code{0 = nearest}, \code{2 = bilinear}, and \code{3 = bicubic} or a name from
\code{\link[magick:filter_types]{magick::filter_types()}}.}

\item{fill}{(int or str or tuple): Pixel fill value for constant fill.
Default is 0. If a tuple of length 3, it is used to fill R, G, B channels
respectively. This value is only used when the padding_mode is constant.
Only int value is supported for Tensors.}
}
\description{
Perspective transformation of an image
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-plankton.R
\name{whoi_small_coralnet_dataset}
\alias{whoi_small_coralnet_dataset}
\title{Coralnet Dataset}
\usage{
whoi_small_coralnet_dataset(
  split = "val",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{split}{One of \code{"train"}, \code{"val"}, or \code{"test"}. Default is \code{"val"}.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\description{
Small Coralnet dataset is an image \strong{classification dataset}
of very large submarine coral reef images annotated into 3 classes
and produced by  \href{https://coralnet.ucsd.edu}{CoralNet},
a resource for benthic images classification.
}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_crop}
\alias{transform_crop}
\title{Crop the given image at specified location and output size}
\usage{
transform_crop(img, top, left, height, width)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{top}{(int): Vertical component of the top left corner of the crop box.}

\item{left}{(int): Horizontal component of the top left corner of the crop
box.}

\item{height}{(int): Height of the crop box.}

\item{width}{(int): Width of the crop box.}
}
\description{
Crop the given image at specified location and output size
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_five_crop}
\alias{transform_five_crop}
\title{Crop image into four corners and a central crop}
\usage{
transform_five_crop(img, size)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{size}{(sequence or int): Desired output size. If size is a sequence
like c(h, w), output size will be matched to this. If size is an int,
smaller edge of the image will be matched to this number.
i.e, if height > width, then image will be rescaled to
(size * height / width, size).}
}
\description{
Crop the given image into four corners and the central crop. This transform
returns a tuple of images and there may be a mismatch in the number of
inputs and targets your Dataset returns.
}
\seealso{
Other combining_transforms: 
\code{\link{transform_random_apply}()},
\code{\link{transform_random_choice}()},
\code{\link{transform_random_order}()},
\code{\link{transform_resized_crop}()},
\code{\link{transform_ten_crop}()}
}
\concept{combining_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collection-rf100-doc.R
\name{rf100_document_collection}
\alias{rf100_document_collection}
\title{RF100 Document Collection Datasets}
\usage{
rf100_document_collection(
  dataset,
  split = c("train", "test", "valid"),
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{dataset}{Dataset to select within \code{c("tweeter_post", "tweeter_profile", "document_part",
"activity_diagram", "signature", "paper_part", "tabular_data", "paragraph", "currency", "wine_label")}.}

\item{split}{the subset of the dataset to choose between \code{c("train", "test", "valid")}.}

\item{transform}{Optional transform function applied to the image.}

\item{target_transform}{Optional transform function applied to the target.}

\item{download}{Logical. If TRUE, downloads the dataset if not present at \code{root}.}
}
\value{
A torch dataset. Each element is a named list with:
\itemize{
\item \code{x}: H x W x 3 array representing the image, auto-oriented and stretched to 640 x 640.
\item \code{y}: a list containing the target with:
\itemize{
\item \code{image_id}: numeric identifier of the x image.
\item \code{labels}: numeric identifier of the N bounding-box object class.
\item \code{boxes}: a torch_tensor of shape (N, 4) with bounding boxes, each in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
}
}

The returned item inherits the class \code{image_with_bounding_box} so it can be
visualised with helper functions such as \code{\link[=draw_bounding_boxes]{draw_bounding_boxes()}}.
}
\description{
RoboFlow 100 Document dataset Collection
}
\details{
Loads one of the RoboFlow 100 Document datasets with
bounding box annotations for object-detection task.
}
\examples{
\dontrun{
ds <- rf100_document_collection(
  dataset = "tweeter_post",
  split = "train",
  transform = transform_to_tensor,
  download = TRUE
)

# Retrieve a sample and inspect annotations
item <- ds[1]
item$y$labels
item$y$boxes

# Draw bounding boxes and display the image
boxed_img <- draw_bounding_boxes(item)
tensor_image_browse(boxed_img)
}

}
\seealso{
Other detection_dataset: 
\code{\link{coco_detection_dataset}()},
\code{\link{pascal_voc_datasets}},
\code{\link{rf100_biology_collection}()},
\code{\link{rf100_damage_collection}()},
\code{\link{rf100_infrared_collection}()},
\code{\link{rf100_medical_collection}()},
\code{\link{rf100_underwater_collection}()}
}
\concept{detection_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-oxfordiiitpet.R
\name{oxfordiiitpet_segmentation_dataset}
\alias{oxfordiiitpet_segmentation_dataset}
\title{Oxford-IIIT Pet Segmentation Dataset}
\usage{
oxfordiiitpet_segmentation_dataset(
  root = tempdir(),
  train = TRUE,
  target_type = "category",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Character. Root directory where the dataset is stored or will be downloaded to. Files are placed under \code{root/oxfordiiitpet}.}

\item{train}{Logical. If TRUE, use the training set; otherwise, use the test set. Not applicable to all datasets.}

\item{target_type}{Character. One of \code{"category"} or \code{"binary-category"} (default: \code{"category"}).}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\value{
A torch dataset object \code{oxfordiiitpet_dataset}. Each item is a named list:
\itemize{
\item \code{x}: a H x W x 3 integer array representing an RGB image.
\item \code{y$masks}: a boolean tensor of shape (3, H, W), representing the segmentation trimap as one-hot masks.
\item \code{y$label}: an integer representing the class label, depending on the \code{target_type}:
\itemize{
\item \code{"category"}: an integer in 1–37 indicating the pet breed.
\item \code{"binary-category"}: 1 for Cat, 2 for Dog.
}
}
}
\description{
The Oxford-IIIT Pet Dataset is a \strong{segmentation} dataset consisting of color images
of 37 pet breeds (cats and dogs). Each image is annotated with a pixel-level
trimap segmentation mask, identifying pet, background, and outline regions.
It is commonly used for evaluating models on object segmentation tasks.
}
\examples{
\dontrun{
# Load the Oxford-IIIT Pet dataset with basic tensor transform
oxfordiiitpet <- oxfordiiitpet_segmentation_dataset(
   transform = transform_to_tensor,
   download = TRUE
)

# Retrieve the image tensor and label (trimap in raw format)
first_item <- oxfordiiitpet[1]
first_item$x  # RGB image tensor of shape (3, H, W)
first_item$y$trimap  # (H, W) integer tensor: 1=pet, 2=background, 3=outline
first_item$y$label  # Integer label (1–37 or 1–2 depending on target_type)
oxfordiiitpet$classes[first_item$y$label] # Class name of the label

# Load dataset with explicit segmentation mask transformation
oxfordiiitpet_masked <- oxfordiiitpet_segmentation_dataset(
   transform = transform_to_tensor,
   target_transform = target_transform_trimap_masks,
   download = TRUE
)

masked_item <- oxfordiiitpet_masked[1]
masked_item$y$masks  # (3, H, W) bool tensor: pet, background, outline

# Visualize segmentation masks
overlay <- draw_segmentation_masks(masked_item)
tensor_image_browse(overlay)
}

}
\seealso{
Other segmentation_dataset: 
\code{\link{coco_segmentation_dataset}()},
\code{\link{pascal_voc_datasets}},
\code{\link{rf100_peixos_segmentation_dataset}()}
}
\concept{segmentation_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-oxfordiiitpet.R
\name{oxfordiiitpet_dataset}
\alias{oxfordiiitpet_dataset}
\alias{oxfordiiitpet_binary_dataset}
\title{Oxford-IIIT Pet Classification Datasets}
\usage{
oxfordiiitpet_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

oxfordiiitpet_binary_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Character. Root directory where the dataset is stored or will be downloaded to. Files are placed under \code{root/oxfordiiitpet}.}

\item{train}{Logical. If TRUE, use the training set; otherwise, use the test set. Not applicable to all datasets.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\value{
A torch dataset object \code{oxfordiiitpet_dataset} or \code{oxfordiiitpet_binary_dataset}.
Each element is a named list with:
\itemize{
\item \code{x}: A H x W x 3 integer array representing an RGB image.
\item \code{y}: An integer label:
\itemize{
\item For \code{oxfordiiitpet_dataset}: a value from 1–37 representing the breed.
\item For \code{oxfordiiitpet_binary_dataset}: 1 for Cat, 2 for Dog.
}
}
}
\description{
Oxford-IIIT Pet Datasets
}
\details{
The Oxford-IIIT Pet collection is a \strong{classification} dataset consisting of high-quality
images of 37 cat and dog breeds. It includes two variants:
\itemize{
\item \code{oxfordiiitpet_dataset}: Multi-class classification across 37 pet breeds.
\item \code{oxfordiiitpet_binary_dataset}: Binary classification distinguishing cats vs dogs.
}

The Oxford-IIIT Pet dataset contains over 7,000 images across 37 categories,
with roughly 200 images per class. Each image is labeled with its breed and species (cat/dog).
}
\examples{
\dontrun{
# Multi-class version
oxford <- oxfordiiitpet_dataset(download = TRUE)
first_item <- oxford[1]
first_item$x  # RGB image
first_item$y  # Label in 1–37
oxford$classes[first_item$y]  # Breed name

# Binary version
oxford_bin <- oxfordiiitpet_binary_dataset(download = TRUE)
first_item <- oxford_bin[1]
first_item$x  # RGB image
first_item$y  # 1 for Cat, 2 for Dog
oxford_bin$classes[first_item$y]  # "Cat" or "Dog"
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_rotation}
\alias{transform_random_rotation}
\title{Rotate the image by angle}
\usage{
transform_random_rotation(
  img,
  degrees,
  interpolation = 0,
  expand = FALSE,
  center = NULL,
  fill = NULL,
  resample
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{degrees}{(sequence or float or int): Range of degrees to select from.
If degrees is a number instead of sequence like c(min, max), the range of
degrees will be (-degrees, +degrees).}

\item{interpolation}{(int, optional): Interpolation mode. 0 for nearest, 2 for bilinear.
Default is 0 (nearest).}

\item{expand}{(bool, optional): Optional expansion flag. If true, expands the
output to make it large enough to hold the entire rotated image. If false
or omitted, make the output image the same size as the input image. Note
that the expand flag assumes rotation around the center and no translation.}

\item{center}{(list or tuple, optional): Optional center of rotation, c(x, y).
Origin is the upper left corner. Default is the center of the image.}

\item{fill}{(n-tuple or int or float): Pixel fill value for area outside the
rotated image. If int or float, the value is used for all bands
respectively. Defaults to 0 for all bands. This option is only available
for Pillow>=5.2.0. This option is not supported for Tensor input. Fill
value for the area outside the transform in the output image is always 0.}

\item{resample}{Deprecated. Use interpolation instead.}
}
\description{
Rotate the image by angle
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_affine}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-box_convert.R
\name{box_cxcywh_to_xyxy}
\alias{box_cxcywh_to_xyxy}
\title{box_cxcywh_to_xyxy}
\usage{
box_cxcywh_to_xyxy(boxes)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes in \eqn{(c_x, c_y, w, h)} format which will be converted.}
}
\value{
boxes (Tensor(N, 4)): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
}
\description{
Converts bounding boxes from  \eqn{(c_x, c_y, w, h)} format to \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format.
\eqn{(c_x, c_y)} refers to center of bounding box
(w, h) are width and height of bounding box
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_resized_crop}
\alias{transform_resized_crop}
\title{Crop an image and resize it to a desired size}
\usage{
transform_resized_crop(img, top, left, height, width, size, interpolation = 2)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{top}{(int): Vertical component of the top left corner of the crop box.}

\item{left}{(int): Horizontal component of the top left corner of the crop
box.}

\item{height}{(int): Height of the crop box.}

\item{width}{(int): Width of the crop box.}

\item{size}{(sequence or int): Desired output size. If size is a sequence
like c(h, w), output size will be matched to this. If size is an int,
smaller edge of the image will be matched to this number.
i.e, if height > width, then image will be rescaled to
(size * height / width, size).}

\item{interpolation}{(int, optional) Desired interpolation. An integer
\code{0 = nearest}, \code{2 = bilinear}, and \code{3 = bicubic} or a name from
\code{\link[magick:filter_types]{magick::filter_types()}}.}
}
\description{
Crop an image and resize it to a desired size
}
\seealso{
Other combining_transforms: 
\code{\link{transform_five_crop}()},
\code{\link{transform_random_apply}()},
\code{\link{transform_random_choice}()},
\code{\link{transform_random_order}()},
\code{\link{transform_ten_crop}()}
}
\concept{combining_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-vggface2.R
\name{vggface2_dataset}
\alias{vggface2_dataset}
\title{VGGFace2 Dataset}
\usage{
vggface2_dataset(
  root = tempdir(),
  split = "val",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Character. Root directory where the dataset will be stored under \code{root/vggface2}.}

\item{split}{One of \code{"train"}, \code{"val"}, or \code{"test"}. Default is \code{"val"}.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\value{
A torch dataset object \code{vggface2_dataset}:
\itemize{
\item \code{x}: RGB image array.
\item \code{y}: Integer label (1…N) for the identity.
}

\code{ds$classes} is a named list mapping integer labels to a list with:
\itemize{
\item \code{name}: Character name of the person.
\item \code{gender}: "Male" or "Female".
}
}
\description{
The VGGFace2 dataset is a large-scale face recognition dataset containing images
of celebrities from a wide range of ethnicities, professions, and ages.
Each identity has multiple images with variations in context, pose, age, and illumination.
}
\examples{
\dontrun{
#Load the training set
ds <- vggface2_dataset(download = TRUE)
item <- ds[1]
item$x      # image array RGB
item$y      # integer label
ds$classes[item$y]  # list(name=..., gender=...)

#Load the test set
ds <- vggface2_dataset(download = TRUE, train = FALSE)
item <- ds[1]
item$x      # image array RGB
item$y      # integer label
ds$classes[item$y]  # list(name=..., gender=...)
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collection-catalog.R
\docType{data}
\name{collection_catalog}
\alias{collection_catalog}
\title{Dataset Collection Catalog}
\format{
A data frame with datasets as rows and 17 columns:
\describe{
\item{collection}{Collection name (biology, medical, infrared, damage, underwater, document, mnist)}
\item{dataset}{Dataset identifier used in collection functions}
\item{description}{Brief description of the dataset and its purpose}
\item{task}{Machine learning task type (currently all "object_detection")}
\item{num_classes}{Number of different object classes}
\item{num_images}{Total images across all splits}
\item{image_width}{Typical image width in pixels}
\item{image_height}{Typical image height in pixels}
\item{train_size_mb}{Size of training split in megabytes}
\item{test_size_mb}{Size of test split in megabytes}
\item{valid_size_mb}{Size of validation split in megabytes}
\item{total_size_mb}{Total size across all splits in megabytes}
\item{has_train}{Is training split available}
\item{has_test}{Is test split available}
\item{has_valid}{Is validation split available}
\item{function_name}{R function name to load this dataset's collection}
\item{roboflow_url}{URL to the collection on RoboFlow Universe}
}
}
\usage{
collection_catalog
}
\description{
A comprehensive catalog of all collections RF100 (RoboFlow 100) and EMNIST datasets available in torchvision.
This data frame contains metadata about each dataset including descriptions, sizes,
available splits, and collection information.
}
\examples{
\dontrun{
# View the complete catalog
data(collection_catalog)
View(collection_catalog)

# See all biology datasets
subset(collection_catalog, collection == "biology")

# Find large datasets (> 100 MB)
subset(collection_catalog, total_size_mb > 100)
}

}
\seealso{
\code{\link[=search_collection]{search_collection()}}, \code{\link[=get_collection_catalog]{get_collection_catalog()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-segmentation.R
\name{target_transform_coco_masks}
\alias{target_transform_coco_masks}
\title{Target Transform: COCO Polygon Segmentation to Masks}
\usage{
target_transform_coco_masks(y)
}
\arguments{
\item{y}{list being COCO dataset target variable, with names \code{segmentation}, \code{image_height}, \code{image_width}.}
}
\value{
Modified \code{y} list with added \code{masks} field (N, H, W) boolean tensor, N being the number of
classes.
}
\description{
Converts COCO-style polygon segmentation annotations from target \verb{$segmentation} variable
into boolean mask tensors as target \verb{$masks} variable in order to ease later-on visualisation
via \code{draw_segmentation_mask()}.
Use as \code{target_transform} in \code{coco_detection_dataset()}.
}
\examples{
\dontrun{
ds <- coco_detection_dataset(
  root = "data",
  target_transform = target_transform_coco_masks
)
item <- ds[1]
draw_segmentation_masks(item)
}

}
\seealso{
Other target_transforms: 
\code{\link{target_transform_trimap_masks}()}
}
\concept{target_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collection-catalog.R
\name{list_collection_datasets}
\alias{list_collection_datasets}
\title{List Datasets in a Collection}
\usage{
list_collection_datasets(collection)
}
\arguments{
\item{collection}{Collection name. One of: "biology", "medical", "infrared",
"damage", "underwater", "document".}
}
\value{
Character vector of dataset names in the collection.
}
\description{
List all available datasets within a specific RF100 collection.
}
\examples{
\dontrun{
# List all biology datasets
list_collection_datasets("biology")

# List all medical datasets
list_collection_datasets("medical")
}

}
\seealso{
\code{\link[=search_collection]{search_collection()}}, \code{\link[=get_collection_catalog]{get_collection_catalog()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-flowers.R
\name{flowers102_dataset}
\alias{flowers102_dataset}
\title{Oxford Flowers 102 Dataset}
\usage{
flowers102_dataset(
  root = tempdir(),
  split = "train",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Root directory for dataset storage. The dataset will be stored under \code{root/flowers102}.}

\item{split}{One of \code{"train"}, \code{"val"}, or \code{"test"}. Default is \code{"train"}.}

\item{transform}{Optional function to transform input images after loading. Default is \code{NULL}.}

\item{target_transform}{Optional function to transform labels. Default is \code{NULL}.}

\item{download}{Logical. Whether to download the dataset if not found locally. Default is \code{FALSE}.}
}
\value{
An object of class \code{flowers102_dataset}, which behaves like a torch dataset.
Each element is a named list:
\itemize{
\item \code{x}: a W x H x 3 numeric array representing an RGB image.
\item \code{y}: an integer label indicating the class index.
}
}
\description{
Loads the Oxford 102 Category Flower Dataset. This dataset consists of 102 flower categories,
with between 40 and 258 images per class. Images in this dataset are of variable sizes.
}
\details{
This is a \strong{classification} dataset where the goal is to assign each image to one of the 102 flower categories.

The dataset is split into:
\itemize{
\item \code{"train"}: training subset with labels.
\item \code{"val"}: validation subset with labels.
\item \code{"test"}: test subset with labels (used for evaluation).
}
}
\examples{
\dontrun{
# Load the dataset with inline transforms
flowers <- flowers102_dataset(
  split = "train",
  download = TRUE,
  transform = . \%>\% transform_to_tensor() \%>\% transform_resize(c(224, 224))
)

# Create a dataloader
dl <- dataloader(
  dataset = flowers,
  batch_size = 4
)

# Access a batch
batch <- dataloader_next(dataloader_make_iter(dl))
batch$x  # Tensor of shape (4, 3, 224, 224)
batch$y  # Tensor of shape (4,) with numeric class labels
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_pad}
\alias{transform_pad}
\title{Pad the given image on all sides with the given "pad" value}
\usage{
transform_pad(img, padding, fill = 0, padding_mode = "constant")
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{padding}{(int or tuple or list): Padding on each border. If a single
int is provided this is used to pad all borders. If tuple of length 2 is
provided this is the padding on left/right and top/bottom respectively.
If a tuple of length 4 is provided this is the padding for the left, right,
top and bottom borders respectively.}

\item{fill}{(int or str or tuple): Pixel fill value for constant fill.
Default is 0. If a tuple of length 3, it is used to fill R, G, B channels
respectively. This value is only used when the padding_mode is constant.
Only int value is supported for Tensors.}

\item{padding_mode}{Type of padding. Should be: constant, edge, reflect or
symmetric. Default is constant. Mode symmetric is not yet supported for
Tensor inputs.
\itemize{
\item constant: pads with a constant value, this value is specified with fill
\item edge: pads with the last value on the edge of the image
\item reflect: pads with reflection of image (without repeating the last
value on the edge) padding \verb{[1, 2, 3, 4]} with 2 elements on both sides
in reflect mode will result in \verb{[3, 2, 1, 2, 3, 4, 3, 2]}
\item symmetric: pads with reflection of image (repeating the last value on
the edge) padding \verb{[1, 2, 3, 4]} with 2 elements on both sides in
symmetric mode will result in \verb{[2, 1, 1, 2, 3, 4, 4, 3]}
}}
}
\description{
The image can be a Magick Image or a torch Tensor, in which case it is
expected to have \verb{[..., H, W]} shape, where ... means an arbitrary number
of leading dimensions.
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_color_jitter}
\alias{transform_color_jitter}
\title{Randomly change the brightness, contrast and saturation of an image}
\usage{
transform_color_jitter(
  img,
  brightness = 0,
  contrast = 0,
  saturation = 0,
  hue = 0
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{brightness}{(float or tuple of float (min, max)): How much to jitter
brightness. \code{brightness_factor} is chosen uniformly from
\verb{[max(0, 1 - brightness), 1 + brightness]} or the given \verb{[min, max]}.
Should be non negative numbers.}

\item{contrast}{(float or tuple of float (min, max)): How much to jitter
contrast. \code{contrast_factor} is chosen uniformly from
\verb{[max(0, 1 - contrast), 1 + contrast]} or the given \verb{[min, max]}. Should
be non negative numbers.}

\item{saturation}{(float or tuple of float (min, max)): How much to jitter
saturation. \code{saturation_factor} is chosen uniformly from
\verb{[max(0, 1 - saturation), 1 + saturation]} or the given \verb{[min, max]}.
Should be non negative numbers.}

\item{hue}{(float or tuple of float (min, max)): How much to jitter hue.
\code{hue_factor} is chosen uniformly from \verb{[-hue, hue]} or the given
\verb{[min, max]}. Should have 0<= hue <= 0.5 or -0.5 <= min <= max <= 0.5.}
}
\description{
Randomly change the brightness, contrast and saturation of an image
}
\seealso{
Other random_transforms: 
\code{\link{transform_random_affine}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-alexnet.R
\name{model_alexnet}
\alias{model_alexnet}
\title{AlexNet Model Architecture}
\usage{
model_alexnet(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{...}{other parameters passed to the model intializer. currently only
\code{num_classes} is used.}
}
\description{
AlexNet model architecture from the
\href{https://arxiv.org/abs/1404.5997}{One weird trick...} paper.
}
\seealso{
Other classification_model: 
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-eurosat.R
\name{eurosat_dataset}
\alias{eurosat_dataset}
\alias{eurosat_all_bands_dataset}
\alias{eurosat100_dataset}
\title{EuroSAT datasets}
\usage{
eurosat_dataset(
  root = tempdir(),
  split = "val",
  download = FALSE,
  transform = NULL,
  target_transform = NULL
)

eurosat_all_bands_dataset(
  root = tempdir(),
  split = "val",
  download = FALSE,
  transform = NULL,
  target_transform = NULL
)

eurosat100_dataset(
  root = tempdir(),
  split = "val",
  download = FALSE,
  transform = NULL,
  target_transform = NULL
)
}
\arguments{
\item{root}{(Optional) Character. The root directory where the dataset will be stored.
if empty, will use the default \code{rappdirs::user_cache_dir("torch")}.}

\item{split}{One of \code{"train"}, \code{"val"}, or \code{"test"}. Default is \code{"val"}.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}
}
\value{
A \code{torch::dataset} object. Each item is a list with:
\itemize{
\item \code{x}: a 64x64 image tensor with 3 (RGB) or 13 (all bands) channels
\item \code{y}: the class label
}
}
\description{
A collection of Sentinel-2 satellite images for land-use \strong{classification}.
The standard version contains 27,000 RGB thumbnails (64x64) across 10
classes. Variants include the full 13 spectral bands and a small 100-image
subset useful for demos.

Downloads and prepares the EuroSAT dataset with 13 spectral bands.

A subset of 100 images with 13 spectral bands useful for workshops and demos.
}
\details{
\code{eurosat_dataset()} provides a total of 27,000 RGB labeled images.

\code{eurosat_all_bands_dataset()} provides a total of 27,000 labeled images with 13 spectral channel bands.

\code{eurosat100_dataset()} provides a subset of 100 labeled images with 13 spectral channel bands.
}
\examples{
\dontrun{
# Initialize the dataset
ds <- eurosat100_dataset(split = "train", download = TRUE)

# Access the first item
head <- ds[1]
print(head$x) # Image
print(head$y) # Label
}
}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-mobilenetv2.R
\name{model_mobilenet_v2}
\alias{model_mobilenet_v2}
\title{MobileNetV2 Model}
\usage{
model_mobilenet_v2(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{...}{Other parameters passed to the model implementation.}
}
\description{
Constructs a MobileNetV2 architecture from
\href{https://arxiv.org/abs/1801.04381}{MobileNetV2: Inverted Residuals and Linear Bottlenecks}.
}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-boxes.R
\name{box_convert}
\alias{box_convert}
\title{Box Convert}
\usage{
box_convert(boxes, in_fmt, out_fmt)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes which will be converted.}

\item{in_fmt}{(str): Input format of given boxes. Supported formats are ['xyxy', 'xywh', 'cxcywh'].}

\item{out_fmt}{(str): Output format of given boxes. Supported formats are ['xyxy', 'xywh', 'cxcywh']}
}
\value{
boxes (Tensor[N, 4]): Boxes into converted format.
}
\description{
Converts boxes from given in_fmt to out_fmt.
}
\details{
Supported in_fmt and out_fmt are:
\itemize{
\item 'xyxy': boxes are represented via corners,
\itemize{
\item \eqn{x_{min}, y_{min}} being top left and
\item \eqn{x_{max}, y_{max}} being bottom right.
}
\item 'xywh' : boxes are represented via corner, width and height,
\itemize{
\item \eqn{x_{min}, y_{min}} being top left,
\item w, h being width and height.
}
\item 'cxcywh' : boxes are represented via centre, width and height,
\itemize{
\item \eqn{c_x, c_y} being center of box,
\item w, h  being width and height.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_grayscale}
\alias{transform_random_grayscale}
\title{Randomly convert image to grayscale with a given probability}
\usage{
transform_random_grayscale(img, p = 0.1)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{p}{(float): probability that image should be converted to grayscale
(default 0.1).}
}
\description{
Convert image to grayscale with a probability of \code{p}.
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_affine}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-boxes.R
\name{box_area}
\alias{box_area}
\title{Box Area}
\usage{
box_area(boxes)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes for which the area will be computed. They
are expected to be in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
\itemize{
\item \eqn{0 \leq x_{min} < x_{max}} and
\item \eqn{0 \leq y_{min} < y_{max}}.
}}
}
\value{
area (Tensor[N]): area for each box
}
\description{
Computes the area of a set of bounding boxes, which are specified by its
\eqn{(x_{min}, y_{min}, x_{max}, y_{max})} coordinates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-pascal.R
\name{pascal_voc_classes}
\alias{pascal_voc_classes}
\title{Pascal VOC Segmentation Dataset}
\usage{
pascal_voc_classes(class_id = 1:21)
}
\arguments{
\item{class_id}{Integer vector of 1-based class identifiers. Must be within [1, 21].}
}
\description{
The Pascal Visual Object Classes (VOC) dataset is a widely used benchmark for object detection and semantic segmentation tasks in computer vision.
}
\details{
This dataset provides RGB images along with per-pixel class segmentation masks for 20 object categories, plus a background class.
Each pixel in the mask is labeled with a class index corresponding to one of the predefined semantic categories.

The VOC dataset was released in yearly editions (2007 to 2012), with slight variations in data splits and annotation formats.
Notably, only the 2007 edition includes a separate \code{test} split; all other years (2008–2012) provide only the \code{train}, \code{val}, and \code{trainval} splits.

The dataset defines 21 semantic classes: \code{"background"}, \code{"aeroplane"}, \code{"bicycle"}, \code{"bird"}, \code{"boat"}, \code{"bottle"}, \code{"bus"}, \code{"car"}, \code{"cat"}, \code{"chair"},
\code{"cow"}, \code{"dining table"}, \code{"dog"}, \code{"horse"}, \code{"motorbike"}, \code{"person"}, \code{"potted plant"}, \code{"sheep"}, \code{"sofa"}, \code{"train"}, and \code{"tv/monitor"}.
They are available through the \code{classes} variable of the dataset object.
}
\seealso{
Other class_resolution: 
\code{\link{caltech_classes}()},
\code{\link{coco_classes}()},
\code{\link{imagenet_classes}()}
}
\concept{class_resolution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-fgvc.R
\name{fgvc_aircraft_dataset}
\alias{fgvc_aircraft_dataset}
\title{FGVC Aircraft Dataset}
\usage{
fgvc_aircraft_dataset(
  root = tempdir(),
  split = "train",
  annotation_level = "variant",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Character. Root directory for dataset storage. The dataset will be stored under \verb{root/fgvc-aircraft-2013b}.}

\item{split}{Character. One of \code{"train"}, \code{"val"}, \code{"trainval"}, or \code{"test"}. Default is \code{"train"}.}

\item{annotation_level}{Character. Level of annotation to use for classification. Default is \code{"variant"}.
One of \code{"variant"}, \code{"family"}, \code{"manufacturer"}, or \code{"all"}. See \emph{Details}.}

\item{transform}{Optional function to transform input images after loading. Default is \code{NULL}.}

\item{target_transform}{Optional function to transform labels. Default is \code{NULL}.}

\item{download}{Logical. Whether to download the dataset if not found locally. Default is \code{FALSE}.}
}
\value{
An object of class \code{fgvc_aircraft_dataset}, which behaves like a torch-style dataset.
Each element is a named list with:
\itemize{
\item \code{x}: an array of shape (H, W, C) with pixel values in the range (0, 255). Please note that images have varying sizes.
\item \code{y}: for single-level annotation (\code{"variant"}, \code{"family"}, \code{"manufacturer"}): an integer class label.
for multi-level annotation (\code{"all"}): a vector of three integers \code{c(manufacturer_idx, family_idx, variant_idx)}.
}
}
\description{
The FGVC-Aircraft dataset supports the following official splits:
\itemize{
\item \code{"train"}: training subset with labels.
\item \code{"val"}: validation subset with labels.
\item \code{"trainval"}: combined training and validation set with labels.
\item \code{"test"}: test set with labels (used for evaluation).
}
}
\details{
The \code{annotation_level} determines the granularity of labels used for classification and supports four values:
\itemize{
\item \code{"variant"}: the most fine-grained level, e.g., \code{"Boeing 737-700"}. There are 100 visually distinguishable variants.
\item \code{"family"}: a mid-level grouping, e.g., \code{"Boeing 737"}, which includes multiple variants. There are 70 distinct families.
\item \code{"manufacturer"}: the coarsest level, e.g., \code{"Boeing"}, grouping multiple families under a single manufacturer. There are 30 manufacturers.
\item \code{"all"}: multi-label format that returns all three levels as a vector of class indices \code{c(manufacturer_idx, family_idx, variant_idx)}.
}

These levels form a strict hierarchy: each \code{"manufacturer"} consists of multiple \code{"families"}, and each \code{"family"} contains several \code{"variants"}.
Not all combinations of levels are valid — for example, a \code{"variant"} always belongs to exactly one \code{"family"}, and a \code{"family"} to exactly one \code{"manufacturer"}.

When \code{annotation_level = "all"} is used, the \verb{$classes} field is a named list with three components:
\itemize{
\item \code{classes$manufacturer}: a character vector of manufacturer names
\item \code{classes$family}: a character vector of family names
\item \code{classes$variant}: a character vector of variant names
}
}
\examples{
\dontrun{
# Single-label classification
fgvc <- fgvc_aircraft_dataset(transform = transform_to_tensor, download = TRUE)

# Create a custom collate function to resize images and prepare batches
resize_collate_fn <- function(batch) {
  xs <- lapply(batch, function(item) {
    torchvision::transform_resize(item$x, c(768, 1024))
  })
  xs <- torch::torch_stack(xs)
  ys <- torch::torch_tensor(sapply(batch, function(item) item$y), dtype = torch::torch_long())
  list(x = xs, y = ys)
}
dl <- torch::dataloader(dataset = fgvc, batch_size = 2, collate_fn = resize_collate_fn)
batch <- dataloader_next(dataloader_make_iter(dl))
batch$x  # batched image tensors with shape (2, 3, 768, 1024)
batch$y  # class labels as integer tensor of shape 2

# Multi-label classification
fgvc <- fgvc_aircraft_dataset(split = "test", annotation_level = "all")
item <- fgvc[1]
item$x  # a double vector representing the image
item$y  # an integer vector of length 3: manufacturer, family, and variant indices
fgvc$classes$manufacturer[item$y[1]]  # e.g., "Boeing"
fgvc$classes$family[item$y[2]]        # e.g., "Boeing 707"
fgvc$classes$variant[item$y[3]]       # e.g., "707-320"
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vision_utils.R
\name{vision_make_grid}
\alias{vision_make_grid}
\title{A simplified version of torchvision.utils.make_grid}
\usage{
vision_make_grid(
  tensor,
  scale = TRUE,
  num_rows = 8,
  padding = 2,
  pad_value = 0
)
}
\arguments{
\item{tensor}{tensor of shape (B x C x H x W) to arrange in grid.}

\item{scale}{whether to normalize (min-max-scale) the input tensor.}

\item{num_rows}{number of rows making up the grid (default 8).}

\item{padding}{amount of padding between batch images (default 2).}

\item{pad_value}{pixel value to use for padding.}
}
\value{
a 3d torch_tensor of shape \eqn{\approx(C , num\_rows \times H , num\_cols \times W)} of all images arranged in a grid.
}
\description{
Arranges a batch B of (image) tensors in a grid, with optional padding between
images. Expects a 4d mini-batch tensor of shape (B x C x H x W).
}
\seealso{
Other image display: 
\code{\link{draw_bounding_boxes}()},
\code{\link{draw_keypoints}()},
\code{\link{draw_segmentation_masks}()},
\code{\link{tensor_image_browse}()},
\code{\link{tensor_image_display}()}
}
\concept{image display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-lfw.R
\name{lfw_dataset}
\alias{lfw_dataset}
\alias{lfw_people_dataset}
\alias{lfw_pairs_dataset}
\title{LFW Datasets}
\usage{
lfw_people_dataset(
  root = tempdir(),
  transform = NULL,
  split = "original",
  target_transform = NULL,
  download = FALSE
)

lfw_pairs_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  split = "original",
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Root directory for dataset storage. The dataset will be stored under \code{root/lfw_people} or \code{root/lfw_pairs}.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{split}{Which version of the dataset to use. One of \code{"original"} or \code{"funneled"}. Defaults to \code{"original"}.}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}

\item{train}{For \code{lfw_pairs_dataset}, whether to load the training (\code{pairsDevTrain.txt}) or test (\code{pairsDevTest.txt}) split.}
}
\value{
A torch dataset object \code{lfw_people_dataset} or \code{lfw_pairs_dataset}.
Each element is a named list with:
\itemize{
\item \code{x}:
\itemize{
\item For \code{lfw_people_dataset}: a H x W x 3 numeric array representing a single RGB image.
\item For \code{lfw_pairs_dataset}: a list of two H x W x 3 numeric arrays representing a pair of RGB images.
}
\item \code{y}:
\itemize{
\item For \code{lfw_people_dataset}: an integer index from 1 to the number of identities in the dataset.
\item For \code{lfw_pairs_dataset}: 1 if the pair shows the same person, 2 if different people.
}
}
}
\description{
Labelled Faces in the Wild (LFW) Datasets
}
\details{
The LFW dataset collection provides facial images for evaluating face recognition systems.
It includes two variants:
\itemize{
\item \code{lfw_people_dataset}: A \strong{multi-class classification} dataset where each image is labelled by person identity.
\item \code{lfw_pairs_dataset}: A \strong{face verification} dataset containing image pairs with binary labels (same or different person).
}

This R implementation of the LFW dataset is based on the \code{fetch_lfw_people()} and \code{fetch_lfw_pairs()} functions from the \code{scikit-learn} library,
but deviates in a few key aspects due to dataset availability and R API conventions:
\itemize{
\item The \code{color} and \code{resize} arguments from Python are not directly exposed. Instead, all images are RGB with a fixed size of 250x250.
\item The \code{split} argument in Python (e.g., \code{train}, \code{test}, \verb{10fold}) is simplified to a \code{train} boolean flag in R.
The \verb{10fold} split is not supported, as the original protocol files are unavailable or incompatible with clean separation of image-label pairs.
\item The \code{split} parameter in R controls which version of the dataset to use: \code{"original"} (unaligned) or \code{"funneled"} (aligned using funneling).
The funneled version contains geometrically normalized face images, offering better alignment and typically improved performance for face recognition models.
\item The dataset is downloaded from Figshare,
which hosts the same files referenced in \code{scikit-learn}'s dataset utilities.
\item \code{lfw_people_dataset}: 13,233 images across multiple identities (using either \code{"original"} or \code{"funneled"} splits)
\item \code{lfw_pairs_dataset}:
\itemize{
\item Training split (\code{train = TRUE}): 2,200 image pairs
\item Test split (\code{train = FALSE}): 1,000 image pairs
}
}
}
\examples{
\dontrun{
# Load data for LFW People Dataset
lfw <- lfw_people_dataset(download = TRUE)
first_item <- lfw[1]
first_item$x  # RGB image
first_item$y  # Label index
lfw$classes[first_item$y]  # person's name (e.g., "Aaron_Eckhart")

# Load training data for LFW Pairs Dataset
lfw <- lfw_pairs_dataset(download = TRUE, train = TRUE)
first_item <- lfw[1]
first_item$x  # List of 2 RGB Images
first_item$x[[1]]  # RGB Image
first_item$x[[2]]  # RGB Image
first_item$y  # Label index
lfw$classes[first_item$y]  # Class Name (e.g., "Same" or "Different")

# Load test data for LFW Pairs Dataset
lfw <- lfw_pairs_dataset(download = TRUE, train = FALSE)
first_item <- lfw[1]
first_item$x  # List of 2 RGB Images
first_item$x[[1]]  # RGB Image
first_item$x[[2]]  # RGB Image
first_item$y  # Label index
lfw$classes[first_item$y]  # Class Name (e.g., "Same" or "Different")
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-convnext_detection.R
\name{model_convnext_detection}
\alias{model_convnext_detection}
\alias{model_convnext_tiny_detection}
\alias{model_convnext_small_detection}
\alias{model_convnext_base_detection}
\title{ConvNeXt Detection Models (Faster R-CNN style)}
\usage{
model_convnext_tiny_detection(
  num_classes = 91,
  pretrained_backbone = FALSE,
  ...
)

model_convnext_small_detection(
  num_classes = 91,
  pretrained_backbone = FALSE,
  ...
)

model_convnext_base_detection(
  num_classes = 91,
  pretrained_backbone = FALSE,
  ...
)
}
\arguments{
\item{num_classes}{Number of output classes excluding background (default: 90 for COCO).}

\item{pretrained_backbone}{Logical. If \code{TRUE}, loads ImageNet-pretrained
ConvNeXt backbone weights. Default: \code{FALSE}.}

\item{...}{Other arguments (unused).}
}
\description{
Object detection models combining a ConvNeXt backbone with a Feature Pyramid
Network (FPN) and the Faster R-CNN detection head. The architecture mirrors
\code{\link[=model_fasterrcnn_resnet50_fpn]{model_fasterrcnn_resnet50_fpn()}}, with the ResNet backbone replaced by
ConvNeXt variants. The design follows the paper
\href{https://arxiv.org/abs/2201.03545}{A ConvNet for the 2020s}.
\subsection{Available Models}{
\itemize{
\item \code{model_convnext_tiny_detection()}
\item \code{model_convnext_small_detection()}
\item \code{model_convnext_base_detection()}
}
}

\subsection{Backbone Performance (ImageNet-1k)}{

Accuracy metrics reflect backbone classification performance only.
Detection head weights are randomly initialized and must be fine-tuned
on task-specific labelled data before meaningful predictions are produced.

\if{html}{\out{<div class="sourceCode">}}\preformatted{| Model                             | Top-1 Acc | Top-5 Acc | Params  | GFLOPS | File Size | Backbone Weights              | Notes                    |
|-----------------------------------|-----------|-----------|---------|--------|-----------|-------------------------------|--------------------------|
| model_convnext_tiny_detection     | 82.5\%     | 96.1\%     | 28.6M   | 4.46   | 109 MB    | IMAGENET1K_V1                 | Tiny backbone, FPN head  |
| model_convnext_small_detection    | 83.6\%     | 96.7\%     | 50.2M   | 8.68   | 192 MB    | IMAGENET1K_V1 (22k pretrain)  | Small backbone, FPN head |
| model_convnext_base_detection     | 84.1\%     | 96.9\%     | 88.6M   | 15.36  | 338 MB    | IMAGENET1K_V1                 | Base backbone, FPN head  |
}\if{html}{\out{</div>}}
}

\subsection{FPN Channel Configuration}{

Each ConvNeXt variant produces four feature maps (C2–C5) fed into the FPN.
Channel widths differ between Tiny/Small and Base:

\if{html}{\out{<div class="sourceCode">}}\preformatted{| Variant | FPN in_channels          | FPN out_channels |
|---------|--------------------------|------------------|
| Tiny    | c(96, 192, 384, 768)     | 256              |
| Small   | c(96, 192, 384, 768)     | 256              |
| Base    | c(128, 256, 512, 1024)   | 256              |
}\if{html}{\out{</div>}}
}

\subsection{Weights Selection}{
\itemize{
\item All variants use \code{IMAGENET1K_V1} backbone weights by default (supervised ImageNet-1k).
\item The Small variant backbone (\code{model_convnext_small_22k}) was additionally
pretrained on ImageNet-22k prior to fine-tuning on ImageNet-1k.
\item Detection head weights are \strong{randomly initialized} — bounding-box
predictions are meaningless without fine-tuning on labelled detection data.
\item Set \code{pretrained_backbone = TRUE} to load ImageNet backbone weights.
}
}
}
\section{Functions}{
\itemize{
\item \code{model_convnext_tiny_detection()}: ConvNeXt Tiny with FPN detection head

\item \code{model_convnext_small_detection()}: ConvNeXt Small with FPN detection head

\item \code{model_convnext_base_detection()}: ConvNeXt Base with FPN detection head

}}
\note{
Detection head weights are randomly initialized. Predicted bounding boxes
will be arbitrary until the detection head is trained on labelled data.
Only the backbone benefits from \code{pretrained_backbone = TRUE}.
}
\examples{
\dontrun{
library(magrittr)
norm_mean <- c(0.485, 0.456, 0.406) # ImageNet normalization constants
norm_std  <- c(0.229, 0.224, 0.225)

url <- paste0("https://upload.wikimedia.org/wikipedia/commons/thumb/",
              "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg")
img <- base_loader(url) \%>\%
  transform_to_tensor() \%>\%
  transform_resize(c(520, 520))

input <- img \%>\% transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)    # Add batch dimension: (1, 3, H, W)

# ConvNeXt Tiny detection
model <- model_convnext_tiny_detection(pretrained_backbone = TRUE)
model$eval()
# Please wait 2 mins + on CPU
pred     <- model(batch)$detections[[1]]
num_boxes <- as.integer(pred$boxes$size()[1])
topk     <- pred$scores$topk(k = 5)[[2]]
boxes    <- pred$boxes[topk, ]
labels   <- imagenet_classes(as.integer(pred$labels[topk]))

# `draw_bounding_box()` may fail if bbox values are not consistent.
if (num_boxes > 0) {
  boxed <- draw_bounding_boxes(img, boxes, labels = labels)
  tensor_image_browse(boxed)
}
}

}
\seealso{
Other object_detection_model: 
\code{\link{model_facenet}},
\code{\link{model_fasterrcnn}},
\code{\link{model_maskrcnn}}
}
\concept{object_detection_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_apply}
\alias{transform_random_apply}
\title{Apply a list of transformations randomly with a given probability}
\usage{
transform_random_apply(img, transforms, p = 0.5)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{transforms}{(list or tuple): list of transformations.}

\item{p}{(float): probability.}
}
\description{
Apply a list of transformations randomly with a given probability
}
\seealso{
Other combining_transforms: 
\code{\link{transform_five_crop}()},
\code{\link{transform_random_choice}()},
\code{\link{transform_random_order}()},
\code{\link{transform_resized_crop}()},
\code{\link{transform_ten_crop}()}
}
\concept{combining_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/folder-dataset.R
\name{base_loader}
\alias{base_loader}
\title{Base loader}
\usage{
base_loader(path)
}
\arguments{
\item{path}{path or URL to load the image from.}
}
\value{
an channel-last array of image values with dim Height x Width x 3
}
\description{
Loads an image using \code{jpeg}, \code{png} or \code{tiff} packages depending on the
file extension.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-boxes.R
\name{batched_nms}
\alias{batched_nms}
\title{Batched Non-maximum Suppression (NMS)}
\usage{
batched_nms(boxes, scores, idxs, iou_threshold)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes where NMS will be performed. They are expected to be
in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
\itemize{
\item \eqn{0 \leq x_{min} < x_{max}} and
\item \eqn{0 \leq y_{min} < y_{max}}.
}}

\item{scores}{(Tensor[N]): scores for each one of the boxes}

\item{idxs}{(Tensor[N]): indices of the categories for each one of the boxes.}

\item{iou_threshold}{(float): discards all overlapping boxes with IoU > \code{iou_threshold}}
}
\value{
keep (Tensor): int64 tensor with the indices of
the elements that have been kept by NMS, sorted
in decreasing order of scores
}
\description{
Performs non-maximum suppression in a batched fashion.
Each index value correspond to a category, and NMS
will not be applied between elements of different categories.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-coco.R
\name{coco_caption_dataset}
\alias{coco_caption_dataset}
\title{COCO Caption Dataset}
\usage{
coco_caption_dataset(
  root = tempdir(),
  train = TRUE,
  year = c("2014"),
  download = FALSE,
  transform = NULL,
  target_transform = NULL
)
}
\arguments{
\item{root}{Root directory where the dataset is stored or will be downloaded to.}

\item{train}{Logical. If TRUE, loads the training split; otherwise, loads the validation split.}

\item{year}{Character. Dataset version year. One of \code{"2014"}.}

\item{download}{Logical. If TRUE, downloads the dataset if it's not already present in the \code{root} directory.}

\item{transform}{Optional transform function applied to the image.}

\item{target_transform}{Optional transform function applied to the target (labels, boxes, etc.).}
}
\value{
An object of class \code{coco_caption_dataset}. Each item is a list:
\itemize{
\item \code{x}: an \verb{(H, W, C)} numeric array containing the RGB image.
\item \code{y}: a character string with the image caption.
}
}
\description{
Loads the MS COCO dataset for image captioning.
}
\examples{
\dontrun{
ds <- coco_caption_dataset(
  train = FALSE,
  download = TRUE
)
example <- ds[1]

# Access image and caption
x <- example$x
y <- example$y

# Prepare image for plotting
image_array <- as.numeric(x)
dim(image_array) <- dim(x)

plot(as.raster(image_array))
title(main = y, col.main = "black")
}
}
\seealso{
Other caption_dataset: 
\code{\link{flickr_caption_dataset}}
}
\concept{caption_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-pascal.R
\name{pascal_voc_datasets}
\alias{pascal_voc_datasets}
\alias{pascal_segmentation_dataset}
\alias{pascal_detection_dataset}
\title{Pascal VOC Datasets}
\usage{
pascal_segmentation_dataset(
  root = tempdir(),
  year = "2012",
  split = "train",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

pascal_detection_dataset(
  root = tempdir(),
  year = "2012",
  split = "train",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{Character. Root directory where the dataset will be stored under \verb{root/pascal_voc_<year>}.}

\item{year}{Character. VOC dataset version to use. One of \code{"2007"}, \code{"2008"}, \code{"2009"}, \code{"2010"}, \code{"2011"}, or \code{"2012"}. Default is \code{"2012"}.}

\item{split}{Character. One of \code{"train"}, \code{"val"}, \code{"trainval"}, or \code{"test"}. Determines the dataset split. Default is \code{"train"}.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\value{
A torch dataset of class \code{pascal_segmentation_dataset}.

The returned list inherits class \code{image_with_segmentation_mask}, which allows generic visualization
utilities to be applied.

Each element is a named list with the following structure:
\itemize{
\item \code{x}: a H x W x 3 array representing the RGB image.
\item \code{y}: A named list containing:
\itemize{
\item \code{masks}: A \code{torch_tensor} of dtype \code{bool} and shape \verb{(21, H, W)}, representing a multi-channel segmentation mask.
Each of the 21 channels corresponds to a Pascal VOC classes
\item \code{labels}: An integer vector indicating the indices of the classes present in the mask.
}
}

A torch dataset of class \code{pascal_detection_dataset}.

The returned list inherits class \code{image_with_bounding_box}, which allows generic visualization
utilities to be applied.

Each element is a named list:
\itemize{
\item \code{x}: a H x W x 3 array representing the RGB image.
\item \code{y}: a list with:
\itemize{
\item \code{labels}: a character vector with object class names.
\item \code{boxes}: a tensor of shape (N, 4) with bounding box coordinates in \verb{(xmin, ymin, xmax, ymax)} format.
}
}
}
\description{
This dataset is frequently used for training and evaluating semantic segmentation models, and supports tasks requiring dense, per-pixel annotations.
}
\examples{
\dontrun{
# Load Pascal VOC segmentation dataset (2007 train split)
pascal_seg <- pascal_segmentation_dataset(
 transform = transform_to_tensor,
 download = TRUE,
 year = "2007"
)

# Access the first image and its mask
first_item <- pascal_seg[1]
first_item$x  # Image
first_item$y$masks  # Segmentation mask
first_item$y$labels  # Unique class labels in the mask
pascal_voc_classes(first_item$y$labels)  # Class names

# Visualise the first image and its mask
masked_img <- draw_segmentation_masks(first_item)
tensor_image_browse(masked_img)

# Load Pascal VOC detection dataset (2007 train split)
pascal_det <- pascal_detection_dataset(
 transform = transform_to_tensor,
 download = TRUE,
 year = "2007"
)

# Access the first image and its bounding boxes
first_item <- pascal_det[1]
first_item$x  # Image
first_item$y$labels  # Object labels
first_item$y$boxes  # Bounding box tensor

# Visualise the first image with bounding boxes
boxed_img <- draw_bounding_boxes(first_item)
tensor_image_browse(boxed_img)
}

}
\seealso{
Other segmentation_dataset: 
\code{\link{coco_segmentation_dataset}()},
\code{\link{oxfordiiitpet_segmentation_dataset}()},
\code{\link{rf100_peixos_segmentation_dataset}()}

Other detection_dataset: 
\code{\link{coco_detection_dataset}()},
\code{\link{rf100_biology_collection}()},
\code{\link{rf100_damage_collection}()},
\code{\link{rf100_document_collection}()},
\code{\link{rf100_infrared_collection}()},
\code{\link{rf100_medical_collection}()},
\code{\link{rf100_underwater_collection}()}
}
\concept{detection_dataset}
\concept{segmentation_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-fer.R
\name{fer_dataset}
\alias{fer_dataset}
\title{FER-2013 Facial Expression Dataset}
\usage{
fer_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{(string, optional): Root directory for dataset storage,
the dataset will be stored under \code{root/fer2013}.}

\item{train}{Logical. If TRUE, use the training set; otherwise, use the test set. Not applicable to all datasets.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\value{
A torch dataset of class \code{fer_dataset}.
Each element is a named list:
\itemize{
\item \code{x}: a 48x48 grayscale array
\item \code{y}: an integer from 1 to 7 indicating the class index
}
}
\description{
Loads the FER-2013 dataset for facial expression recognition. The dataset contains grayscale images
(48x48) of human faces, each labeled with one of seven emotion categories:
\code{"Angry"}, \code{"Disgust"}, \code{"Fear"}, \code{"Happy"}, \code{"Sad"}, \code{"Surprise"}, and \code{"Neutral"}.
}
\details{
The dataset is split into:
\itemize{
\item \code{"Train"}: training images labeled as \code{"Training"} in the original CSV.
\item \code{"Test"}: includes both \code{"PublicTest"} and \code{"PrivateTest"} entries.
}
}
\examples{
\dontrun{
fer <- fer_dataset(train = TRUE, download = TRUE)
first_item <- fer[1]
first_item$x  # 48x48 grayscale array
first_item$y  # 4
fer$classes[first_item$y]  # "Happy"
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/folder-dataset.R
\name{magick_loader}
\alias{magick_loader}
\title{Load an Image using ImageMagick}
\usage{
magick_loader(path)
}
\arguments{
\item{path}{path or URL to load the image from.}
}
\value{
an magick-image object as result of \code{image_read()}
}
\description{
Load an image located at \code{path} using the \code{{magick}} package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_adjust_hue}
\alias{transform_adjust_hue}
\title{Adjust the hue of an image}
\usage{
transform_adjust_hue(img, hue_factor)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{hue_factor}{(float):  How much to shift the hue channel. Should be in
\verb{[-0.5, 0.5]}. 0.5 and -0.5 give complete reversal of hue channel in
HSV space in positive and negative direction respectively.
0 means no shift. Therefore, both -0.5 and 0.5 will give an image
with complementary colors while 0 gives the original image.}
}
\description{
The image hue is adjusted by converting the image to HSV and
cyclically shifting the intensities in the hue channel (H).
The image is then converted back to original image mode.
}
\details{
\code{hue_factor} is the amount of shift in H channel and must be in the
interval \verb{[-0.5, 0.5]}.

Search for Hue for more details.
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_convert_image_dtype}
\alias{transform_convert_image_dtype}
\title{Convert a tensor image to the given \code{dtype} and scale the values accordingly}
\usage{
transform_convert_image_dtype(img, dtype = torch::torch_float())
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{dtype}{(torch.dtype): Desired data type of the output.}
}
\description{
Convert a tensor image to the given \code{dtype} and scale the values accordingly
}
\note{
When converting from a smaller to a larger integer \code{dtype} the maximum
values are \strong{not} mapped exactly. If converted back and forth, this
mismatch has no effect.
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_to_tensor}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_crop}
\alias{transform_random_crop}
\title{Crop the given image at a random location}
\usage{
transform_random_crop(
  img,
  size,
  padding = NULL,
  pad_if_needed = FALSE,
  fill = 0,
  padding_mode = "constant"
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{size}{(sequence or int): Desired output size. If size is a sequence
like c(h, w), output size will be matched to this. If size is an int,
smaller edge of the image will be matched to this number.
i.e, if height > width, then image will be rescaled to
(size * height / width, size).}

\item{padding}{(int or tuple or list): Padding on each border. If a single
int is provided this is used to pad all borders. If tuple of length 2 is
provided this is the padding on left/right and top/bottom respectively.
If a tuple of length 4 is provided this is the padding for the left, right,
top and bottom borders respectively.}

\item{pad_if_needed}{(boolean): It will pad the image if smaller than the
desired size to avoid raising an exception. Since cropping is done
after padding, the padding seems to be done at a random offset.}

\item{fill}{(int or str or tuple): Pixel fill value for constant fill.
Default is 0. If a tuple of length 3, it is used to fill R, G, B channels
respectively. This value is only used when the padding_mode is constant.
Only int value is supported for Tensors.}

\item{padding_mode}{Type of padding. Should be: constant, edge, reflect or
symmetric. Default is constant. Mode symmetric is not yet supported for
Tensor inputs.
\itemize{
\item constant: pads with a constant value, this value is specified with fill
\item edge: pads with the last value on the edge of the image
\item reflect: pads with reflection of image (without repeating the last
value on the edge) padding \verb{[1, 2, 3, 4]} with 2 elements on both sides
in reflect mode will result in \verb{[3, 2, 1, 2, 3, 4, 3, 2]}
\item symmetric: pads with reflection of image (repeating the last value on
the edge) padding \verb{[1, 2, 3, 4]} with 2 elements on both sides in
symmetric mode will result in \verb{[2, 1, 1, 2, 3, 4, 4, 3]}
}}
}
\description{
The image can be a Magick Image or a Tensor, in which case it is expected
to have \verb{[..., H, W]} shape, where ... means an arbitrary number of leading
dimensions.
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_affine}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-mask_rcnn.R
\name{model_maskrcnn}
\alias{model_maskrcnn}
\alias{model_maskrcnn_resnet50_fpn}
\alias{model_maskrcnn_resnet50_fpn_v2}
\title{Mask R-CNN Models}
\usage{
model_maskrcnn_resnet50_fpn(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 90,
  score_thresh = 0.05,
  nms_thresh = 0.5,
  detections_per_img = 100,
  ...
)

model_maskrcnn_resnet50_fpn_v2(
  pretrained = FALSE,
  progress = TRUE,
  num_classes = 90,
  score_thresh = 0.05,
  nms_thresh = 0.5,
  detections_per_img = 100,
  ...
)
}
\arguments{
\item{pretrained}{Logical. If TRUE, loads pretrained weights from local file.}

\item{progress}{Logical. Show progress bar during download (unused).}

\item{num_classes}{Number of output classes excluding background (default: 90 for COCO).}

\item{score_thresh}{Numeric. Minimum score threshold for detections (default: 0.05).}

\item{nms_thresh}{Numeric. Non-Maximum Suppression (NMS) IoU threshold for removing overlapping boxes (default: 0.5).}

\item{detections_per_img}{Integer. Maximum number of detections per image (default: 100).}

\item{...}{Other arguments (unused).}
}
\value{
A \code{maskrcnn_model} nn_module.
}
\description{
Construct Mask R-CNN model variants for instance segmentation task.
Mask R-CNN extends Faster R-CNN by adding a mask prediction branch that
outputs segmentation masks for each detected object.
}
\section{Functions}{
\itemize{
\item \code{model_maskrcnn_resnet50_fpn()}: Mask R-CNN with ResNet-50 FPN

\item \code{model_maskrcnn_resnet50_fpn_v2()}: Mask R-CNN with ResNet-50 FPN V2

}}
\section{Task}{

Instance segmentation over images with bounding boxes, class labels, and segmentation masks.
}

\section{Input Format}{

Input images should be \code{torch_tensor}s of shape
\verb{(batch_size, 3, H, W)} where \code{H} and \code{W} are typically around 800.
}

\section{Output Format}{

Returns a list with:
\itemize{
\item \code{features}: Feature maps from the backbone
\item \code{detections}: List containing:
\itemize{
\item \code{boxes}: Bounding boxes (N, 4)
\item \code{labels}: Class labels (N)
\item \code{scores}: Confidence scores (N)
\item \code{masks}: Segmentation masks (N, 28, 28)
}
}
}

\section{Available Models}{

\itemize{
\item \code{model_maskrcnn_resnet50_fpn()}
\item \code{model_maskrcnn_resnet50_fpn_v2()}
}
}

\examples{
\dontrun{
library(magrittr)
# ImageNet normalization constants, see https://pytorch.org/vision/stable/models.html
norm_mean <- c(0.485, 0.456, 0.406)
norm_std  <- c(0.229, 0.224, 0.225)

# Load an image
url <- paste0("https://upload.wikimedia.org/wikipedia/commons/thumb/",
              "e/ea/Morsan_Normande_vache.jpg/120px-Morsan_Normande_vache.jpg")
img <- base_loader(url)

input <- img \%>\%
  transform_to_tensor() \%>\%
  transform_resize(c(800, 800)) \%>\%
  transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)

# Mask R-CNN ResNet-50 FPN
model <- model_maskrcnn_resnet50_fpn(pretrained = TRUE, , detections_per_img = 5)
model$eval()

torch::with_no_grad({pred <- model(batch)$detections[[1]]})

# Visualize boxes
labels <- coco_classes(as.integer(pred$labels))
boxed <- draw_bounding_boxes(image, pred$boxes, labels = labels)
tensor_image_browse(boxed)
}

}
\seealso{
Other object_detection_model: 
\code{\link{model_convnext_detection}},
\code{\link{model_facenet}},
\code{\link{model_fasterrcnn}}
}
\concept{object_detection_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vision_utils.R
\name{coco_polygon_to_mask}
\alias{coco_polygon_to_mask}
\title{Convert COCO polygon to mask tensor (Robust Version)}
\usage{
coco_polygon_to_mask(segmentation, height, width)
}
\arguments{
\item{segmentation}{A list of polygons from COCO annotations (e.g., \code{anns$segmentation[[i]]}).}

\item{height}{Height of the image}

\item{width}{Width of the image}
}
\value{
A torch_bool() tensor of shape (height, width)
}
\description{
Converts a COCO-style polygon annotation (list of coordinates) into a binary mask tensor.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_horizontal_flip}
\alias{transform_random_horizontal_flip}
\title{Horizontally flip an image randomly with a given probability}
\usage{
transform_random_horizontal_flip(img, p = 0.5)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{p}{(float): probability of the image being flipped.
Default value is 0.5}
}
\description{
Horizontally flip an image randomly with a given probability. The image can
be a Magick Image or a torch Tensor, in which case it is expected to have
\verb{[..., H, W]} shape, where ... means an arbitrary number of leading
dimensions
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_affine}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_resized_crop}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-imagenet.R
\name{tiny_imagenet_dataset}
\alias{tiny_imagenet_dataset}
\title{Tiny ImageNet dataset}
\usage{
tiny_imagenet_dataset(root, split = "train", download = FALSE, ...)
}
\arguments{
\item{root}{directory path to download the dataset.}

\item{split}{dataset split, \code{train}, \code{validation} or \code{test}.}

\item{download}{whether to download or not the dataset.}

\item{...}{other arguments passed to \code{\link[=image_folder_dataset]{image_folder_dataset()}}.}
}
\description{
Prepares the Tiny ImageNet dataset and optionally downloads it.
}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-segmentation.R
\name{target_transform_trimap_masks}
\alias{target_transform_trimap_masks}
\title{Target Transform: Trimap to Boolean Masks}
\usage{
target_transform_trimap_masks(y)
}
\arguments{
\item{y}{List containing \code{trimap} field (H, W) tensor with values 1, 2, 3}
}
\value{
Modified y list with added \code{masks} field (3, H, W) boolean tensor
}
\description{
Converts Oxford-IIIT Pet dataset target \verb{$trimap} variable (values 1,2,3) into
3-channel boolean masks tensors as target \verb{$masks} variable in order to ease later-on visualisation
via \code{draw_segmentation_mask()}.
Use as \code{target_transform} in \code{oxfordiiitpet_segmentation_dataset()}.
}
\details{
Creates three mutually exclusive masks:
\itemize{
\item Channel 1: Pet pixels (trimap == 1)
\item Channel 2: Background pixels (trimap == 2)
\item Channel 3: Outline pixels (trimap == 3)
}
}
\examples{
\dontrun{
ds <- oxfordiiitpet_segmentation_dataset(
  root = "data",
  target_transform = target_transform_trimap_masks
)
item <- ds[1]
draw_segmentation_masks(item)
}

}
\seealso{
Other target_transforms: 
\code{\link{target_transform_coco_masks}()}
}
\concept{target_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-places365.R
\name{places365_dataset}
\alias{places365_dataset}
\alias{places365_dataset_large}
\title{Places365 Dataset}
\usage{
places365_dataset(
  root = tempdir(),
  split = c("train", "val", "test"),
  transform = NULL,
  target_transform = NULL,
  download = FALSE,
  loader = magick_loader
)

places365_dataset_large(
  root = tempdir(),
  split = c("train", "val", "test"),
  transform = NULL,
  target_transform = NULL,
  download = FALSE,
  loader = magick_loader
)
}
\arguments{
\item{root}{Root directory for dataset storage. The dataset will be stored under \verb{root/<dataset-name>}. Defaults to \code{tempdir()}.}

\item{split}{One of \code{"train"}, \code{"val"}, or \code{"test"}.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}

\item{loader}{A function to load an image given its path. Defaults to
\code{\link[=magick_loader]{magick_loader()}}, which uses the \code{{magick}} package.}
}
\value{
A torch dataset of class \code{places365_dataset}. Each element is a named
list with:
\itemize{
\item \code{x}: the image as loaded (or transformed if \code{transform} is set).
\item \code{y}: the integer class label. For the \code{test} split, no labels are available
and \code{y} will always be \code{NA}.
}
}
\description{
Loads the MIT Places365 dataset for scene classification.
}
\details{
The dataset provides three splits: \code{"train"}, \code{"val"}, and \code{"test"}.
Folder structure and image layout on disk are handled internally by the loader.

This function downloads and prepares the smaller 256x256 image version (~30 GB).
For the high-resolution variant (~160 GB), use \code{\link[=places365_dataset_large]{places365_dataset_large()}}.
Note that images in the large version come in varying sizes, so resizing may be
needed before batching.

The \code{test} split corresponds to the \emph{private} evaluation set used in the
Places365 challenge. Annotation files are not publicly released, so only the
images are provided.
}
\section{Functions}{
\itemize{
\item \code{places365_dataset_large()}: High resolution variant (~160 GB).

}}
\examples{
\dontrun{
ds <- places365_dataset(
  split = "val",
  download = TRUE,
  transform = transform_to_tensor
)
item <- ds[1]
tensor_image_browse(item$x)

# Show class index and label
label_idx <- item$y
label_name <- ds$classes[label_idx]
label_idx  # Class Index
label_name  # Name of the Label

dl <- dataloader(ds, batch_size = 2)
batch <- dataloader_next(dataloader_make_iter(dl))
batch$x

ds_large <- places365_dataset_large(
  split = "val",
  download = TRUE,
  transform = . \%>\% transform_to_tensor() \%>\% transform_resize(c(256, 256))
)
dl <- torch::dataloader(dataset = ds_large, batch_size = 2)
batch <- dataloader_next(dataloader_make_iter(dl))
batch$x
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_to_tensor}
\alias{transform_to_tensor}
\title{Convert an image to a tensor}
\usage{
transform_to_tensor(img)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}
}
\description{
Converts a Magick Image or array (H x W x C) in the range \verb{[0, 255]} to a
\code{torch_tensor} of shape (C x H x W) in the range \verb{[0.0, 1.0]}. In the
other cases, tensors are returned without scaling.
}
\note{
Because the input image is scaled to \verb{[0.0, 1.0]}, this transformation
should not be used when transforming target image masks.
}
\seealso{
Other unitary_transforms: 
\code{\link{transform_adjust_brightness}()},
\code{\link{transform_adjust_contrast}()},
\code{\link{transform_adjust_gamma}()},
\code{\link{transform_adjust_hue}()},
\code{\link{transform_adjust_saturation}()},
\code{\link{transform_affine}()},
\code{\link{transform_center_crop}()},
\code{\link{transform_convert_image_dtype}()},
\code{\link{transform_crop}()},
\code{\link{transform_grayscale}()},
\code{\link{transform_hflip}()},
\code{\link{transform_linear_transformation}()},
\code{\link{transform_normalize}()},
\code{\link{transform_pad}()},
\code{\link{transform_perspective}()},
\code{\link{transform_resize}()},
\code{\link{transform_rgb_to_grayscale}()},
\code{\link{transform_rotate}()},
\code{\link{transform_vflip}()}
}
\concept{unitary_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-inception.R
\name{model_inception_v3}
\alias{model_inception_v3}
\title{Inception v3 model}
\usage{
model_inception_v3(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If \code{TRUE}, returns a model pre-trained on ImageNet}

\item{progress}{(bool): If \code{TRUE}, displays a progress bar of the download to stderr}

\item{...}{Used to pass keyword arguments to the Inception module:
\itemize{
\item aux_logits (bool): If \code{TRUE}, add an auxiliary branch that can improve training.
Default: \emph{TRUE}
\item transform_input (bool): If \code{TRUE}, preprocess the input according to the method with which it
was trained on ImageNet. Default: \emph{FALSE}
}}
}
\description{
Architecture from \href{https://arxiv.org/abs/1512.00567}{Rethinking the Inception Architecture for Computer Vision}
The required minimum input size of the model is 75x75.
}
\note{
\strong{Important}: In contrast to the other models the inception_v3 expects tensors with a size of
N x 3 x 299 x 299, so ensure your images are sized accordingly.
}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-mnist.R
\name{mnist_dataset}
\alias{mnist_dataset}
\alias{kmnist_dataset}
\alias{qmnist_dataset}
\alias{fashion_mnist_dataset}
\alias{emnist_collection}
\alias{emnist_dataset}
\title{MNIST and Derived Datasets}
\usage{
mnist_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

kmnist_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

qmnist_dataset(
  root = tempdir(),
  split = "train",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

fashion_mnist_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

emnist_collection(
  root = tempdir(),
  split = "test",
  dataset = "balanced",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

emnist_dataset(kind, ...)
}
\arguments{
\item{root}{Root directory for dataset storage. The dataset will be stored under \verb{root/<dataset-name>}. Defaults to \code{tempdir()}.}

\item{train}{Logical. If TRUE, use the training set; otherwise, use the test set. Not applicable to all datasets.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}

\item{split}{Character. Used in \code{emnist_dataset()} and \code{qmnist_dataset()} to specify the subset. See individual descriptions for valid values.}

\item{dataset}{to select within \code{c("byclass", "bymerge", "balanced", "letters", "digits", "mnist")}
representing the subset of emnist collection made of a set of classes.
You can look at dataset attribute \verb{$classes} to see the actual classes.}

\item{kind}{the \code{dataset} in \code{emnist_collection}.}

\item{...}{the other \code{emnist_collection} parameters.}
}
\value{
A torch dataset object, where each items is a list of \code{x} (image) and \code{y} (label).
}
\description{
Prepares various MNIST-style image classification datasets and optionally downloads them.
Images are thumbnails images of 28 x 28 pixels of grayscale values encoded as integer.
}
\details{
\itemize{
\item \strong{MNIST}: Original handwritten digit dataset.
\item \strong{Fashion-MNIST}: Clothing item images for classification.
\item \strong{Kuzushiji-MNIST}: Japanese cursive character dataset.
\item \strong{QMNIST}: Extended MNIST with high-precision NIST data.
\item \strong{EMNIST}: A collection of letters and digits with multiple datasets and  splits.
}
}
\section{Functions}{
\itemize{
\item \code{kmnist_dataset()}: Kuzushiji-MNIST cursive Japanese character dataset.

\item \code{qmnist_dataset()}: Extended MNIST dataset with high-precision test data (QMNIST).

\item \code{fashion_mnist_dataset()}: Fashion-MNIST clothing image dataset.

\item \code{emnist_collection()}: EMNIST collection with digits and letters arranged in multiple datasets.

\item \code{emnist_dataset()}: Deprecated. Please use emnist_collection.

}}
\section{Supported \code{dataset}s for \code{emnist_collection()}}{

\itemize{
\item \code{"byclass"}: 62 classes (digits + uppercase + lowercase)
\item \code{"bymerge"}: 47 classes (merged uppercase and lowercase)
\item \code{"balanced"}: 47 classes, balanced digits and letters
\item \code{"letters"}: 26 uppercase letters
\item \code{"digits"}: 10 digit classes
\item \code{"mnist"}: Standard MNIST digit classes
}
}

\section{Supported \code{split}s for \code{qmnist_dataset()}}{

\itemize{
\item \code{"train"}: 60,000 training samples (MNIST-compatible)
\item \code{"test"}: Extended test set
\item \code{"nist"}: Full NIST digit set
}
}

\examples{
\dontrun{
ds <- mnist_dataset(download = TRUE)
item <- ds[1]
item$x  # image
item$y  # label

qmnist <- qmnist_dataset(split = "train", download = TRUE)
item <- qmnist[1]
item$x
item$y

emnist <- emnist_collection(dataset = "balanced", split = "test", download = TRUE)
item <- emnist[1]
item$x
item$y

kmnist <- kmnist_dataset(download = TRUE, train = FALSE)
fmnist <- fashion_mnist_dataset(download = TRUE, train = TRUE)
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-boxes.R
\name{box_iou}
\alias{box_iou}
\title{Box IoU}
\usage{
box_iou(boxes1, boxes2)
}
\arguments{
\item{boxes1}{(Tensor[N, 4])}

\item{boxes2}{(Tensor[M, 4])}
}
\value{
iou (Tensor[N, M]): the NxM matrix containing the pairwise IoU values for every element in boxes1 and boxes2
}
\description{
Return intersection-over-union  (Jaccard index) of boxes.
Both sets of boxes are expected to be in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format with
\eqn{0 \leq x_{min} < x_{max}} and \eqn{0 \leq y_{min} < y_{max}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-plankton.R
\name{whoi_plankton_dataset}
\alias{whoi_plankton_dataset}
\alias{whoi_small_plankton_dataset}
\title{WHOI Plankton Datasets}
\usage{
whoi_small_plankton_dataset(
  split = "val",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

whoi_plankton_dataset(
  split = "val",
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{split}{One of \code{"train"}, \code{"val"}, or \code{"test"}. Default is \code{"val"}.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\value{
A torch dataset with a
\itemize{
\item \code{classes} attribute providing the vector of class names.
}

Each element is a named list:
\itemize{
\item \code{x}: a H x W x 1 integer array representing an grayscale image.
\item \code{y}: the class id of the image.
}
}
\description{
WHOI-Plankton Dataset
}
\details{
The WHOI-Plankton and WHOI-Plankton small are \strong{image classification} datasets
from the Woods Hole Oceanographic Institution (WHOI) of microscopic marine plankton.
https://hdl.handle.net/10.1575/1912/7341
Images were collected in situ by automated submersible imaging-in-flow cytometry
with an instrument called Imaging FlowCytobot (IFCB). They are small grayscale images
of varying size.
Images are classified into 100 classes, with an overview  available in
\href{https://whoigit.github.io/whoi-plankton/}{project Wiki page}
Dataset size is 957k and 58k respectively, and each provides a train / val / test split.
}
\examples{
\dontrun{
# Load the small plankton dataset and turn images into tensor images
plankton <- whoi_small_plankton_dataset(download = TRUE, transform = transform_to_tensor)

# Access the first item
first_item <- plankton[1]
first_item$x  # a tensor grayscale image with shape {1, H, W}
first_item$y  # id of the plankton class.
plankton$classes[first_item$y] # name of the plankton class

# Load the full plankton dataset
plankton <- whoi_plankton_dataset(download = TRUE)

# Access the first item
first_item <- plankton[1]
first_item$x  # grayscale image array with shape {H, W}
first_item$y  # id of the plankton class.
}

}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{cifar10_dataset}()},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-resnet.R
\name{model_resnet}
\alias{model_resnet}
\alias{model_resnet18}
\alias{model_resnet34}
\alias{model_resnet50}
\alias{model_resnet101}
\alias{model_resnet152}
\alias{model_resnext50_32x4d}
\alias{model_resnext101_32x8d}
\alias{model_wide_resnet50_2}
\alias{model_wide_resnet101_2}
\title{ResNet implementation}
\usage{
model_resnet18(pretrained = FALSE, progress = TRUE, ...)

model_resnet34(pretrained = FALSE, progress = TRUE, ...)

model_resnet50(pretrained = FALSE, progress = TRUE, ...)

model_resnet101(pretrained = FALSE, progress = TRUE, ...)

model_resnet152(pretrained = FALSE, progress = TRUE, ...)

model_resnext50_32x4d(pretrained = FALSE, progress = TRUE, ...)

model_resnext101_32x8d(pretrained = FALSE, progress = TRUE, ...)

model_wide_resnet50_2(pretrained = FALSE, progress = TRUE, ...)

model_wide_resnet101_2(pretrained = FALSE, progress = TRUE, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{...}{Other parameters passed to the resnet model.}
}
\description{
ResNet models implementation from
\href{https://arxiv.org/pdf/1512.03385}{Deep Residual Learning for Image Recognition} and later
related papers (see Functions)
}
\section{Functions}{
\itemize{
\item \code{model_resnet18()}: ResNet 18-layer model

\item \code{model_resnet34()}: ResNet 34-layer model

\item \code{model_resnet50()}: ResNet 50-layer model

\item \code{model_resnet101()}: ResNet 101-layer model

\item \code{model_resnet152()}: ResNet 152-layer model

\item \code{model_resnext50_32x4d()}: ResNeXt-50 32x4d model from \href{https://arxiv.org/pdf/1611.05431}{"Aggregated Residual Transformation for Deep Neural Networks"}
with 32 groups having each a width of 4.

\item \code{model_resnext101_32x8d()}: ResNeXt-101 32x8d model from \href{https://arxiv.org/pdf/1611.05431}{"Aggregated Residual Transformation for Deep Neural Networks"}
with 32 groups having each a width of 8.

\item \code{model_wide_resnet50_2()}: Wide ResNet-50-2 model from \href{https://arxiv.org/pdf/1605.07146}{"Wide Residual Networks"}
with width per group of 128.

\item \code{model_wide_resnet101_2()}: Wide ResNet-101-2 model from \href{https://arxiv.org/pdf/1605.07146}{"Wide Residual Networks"}
with width per group of 128.

}}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-mobilenetv3.R
\name{.rename_mobilenet_v3_state_dict}
\alias{.rename_mobilenet_v3_state_dict}
\title{Rename state_dict keys to match current MobileNetV3 model structure}
\usage{
.rename_mobilenet_v3_state_dict(state_dict)
}
\arguments{
\item{state_dict}{Named list of model parameters}
}
\value{
Named list with updated key names
}
\description{
Rename state_dict keys to match current MobileNetV3 model structure
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vision_utils.R
\name{draw_segmentation_masks}
\alias{draw_segmentation_masks}
\alias{draw_segmentation_masks.default}
\alias{draw_segmentation_masks.torch_tensor}
\alias{draw_segmentation_masks.image_with_segmentation_mask}
\title{Draw segmentation masks}
\usage{
draw_segmentation_masks(x, ...)

\method{draw_segmentation_masks}{default}(x, ...)

\method{draw_segmentation_masks}{torch_tensor}(x, masks, alpha = 0.8, colors = NULL, ...)

\method{draw_segmentation_masks}{image_with_segmentation_mask}(x, alpha = 0.5, colors = NULL, ...)
}
\arguments{
\item{x}{Tensor of shape (C x H x W) and dtype \code{uint8} or dtype \code{float}.
In case of dtype float, values are assumed to be in range \eqn{[0, 1]}.
C value for channel can only be 1 (grayscale) or 3 (RGB).}

\item{...}{Additional arguments passed to methods.}

\item{masks}{torch_tensor of shape (num_masks, H, W) or (H, W) and dtype bool.}

\item{alpha}{number between 0 and 1 denoting the transparency of the masks.}

\item{colors}{character vector containing the colors
of the boxes or single color for all boxes. The color can be represented as
strings e.g. "red" or "#FF00FF". By default, viridis colors are generated for masks}
}
\value{
torch_tensor of shape (3, H, W) and dtype uint8 of the image with segmentation masks drawn on top.
}
\description{
Draw segmentation masks with their respective colors on top of a given RGB tensor image
}
\examples{
\dontshow{if (torch::torch_is_installed() && rlang::is_installed("magick")) withAutoprint(\{ # examplesIf}
image_tensor <- torch::torch_randint(170, 250, size = c(3, 360, 360))$to(torch::torch_uint8())
mask <- torch::torch_tril(torch::torch_ones(c(360, 360)))$to(torch::torch_bool())
masked_image <- draw_segmentation_masks(image_tensor, mask, alpha = 0.2)
tensor_image_browse(masked_image)
\dontshow{\}) # examplesIf}
}
\seealso{
Other image display: 
\code{\link{draw_bounding_boxes}()},
\code{\link{draw_keypoints}()},
\code{\link{tensor_image_browse}()},
\code{\link{tensor_image_display}()},
\code{\link{vision_make_grid}()}
}
\concept{image display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataset-cifar.R
\name{cifar10_dataset}
\alias{cifar10_dataset}
\alias{cifar100_dataset}
\title{CIFAR datasets}
\usage{
cifar10_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)

cifar100_dataset(
  root = tempdir(),
  train = TRUE,
  transform = NULL,
  target_transform = NULL,
  download = FALSE
)
}
\arguments{
\item{root}{(string): Root directory of dataset where directory
\code{cifar-10-batches-bin} exists or will be saved to if download is set to TRUE.}

\item{train}{Logical. If TRUE, use the training set; otherwise, use the test set. Not applicable to all datasets.}

\item{transform}{Optional. A function that takes an image and returns a transformed version (e.g., normalization, cropping).}

\item{target_transform}{Optional. A function that transforms the label.}

\item{download}{Logical. If TRUE, downloads the dataset to \verb{root/}. If the dataset is already present, download is skipped.}
}
\value{
A torch::dataset object. Each item is a list with:
\itemize{
\item \code{x}: a 32x32x3 integer array
\item \code{y}: the class label
}
}
\description{
The CIFAR datasets are benchmark \strong{classification} datasets composed of
60,000 RGB thumbnail images of size 32x32 pixels. The
\href{https://www.cs.toronto.edu/~kriz/cifar.html}{CIFAR10} variant contains
10 classes while CIFAR100 provides 100 classes. Images are split into
50,000 training samples and 10,000 test samples.

Downloads and prepares the
\href{https://www.cs.toronto.edu/~kriz/cifar.html}{CIFAR100} dataset.
}
\details{
Downloads and prepares the CIFAR archives.
}
\examples{
\dontrun{
ds <- cifar10_dataset(root = tempdir(), download = TRUE)
item <- ds[1]
item$x
item$y
}
}
\seealso{
Other classification_dataset: 
\code{\link{caltech_dataset}},
\code{\link{eurosat_dataset}()},
\code{\link{fer_dataset}()},
\code{\link{fgvc_aircraft_dataset}()},
\code{\link{flowers102_dataset}()},
\code{\link{image_folder_dataset}()},
\code{\link{lfw_dataset}},
\code{\link{mnist_dataset}()},
\code{\link{oxfordiiitpet_dataset}()},
\code{\link{places365_dataset}()},
\code{\link{tiny_imagenet_dataset}()},
\code{\link{vggface2_dataset}()},
\code{\link{whoi_plankton_dataset}()},
\code{\link{whoi_small_coralnet_dataset}()}
}
\concept{classification_dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-maxvit.R
\name{model_maxvit}
\alias{model_maxvit}
\title{MaxViT Model}
\usage{
model_maxvit(pretrained = FALSE, progress = TRUE, num_classes = 1000, ...)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{num_classes}{(integer) Number of output classes.}

\item{...}{Additional parameters passed to the model initializer.}
}
\description{
Implementation of the MaxViT architecture described in
\href{https://arxiv.org/abs/2204.01697}{MaxViT: Multi-Axis Vision Transformer}.
The model performs image classification and by default returns logits for
1000 ImageNet classes.
}
\examples{
\dontrun{
library(magrittr)
# 1. Load the basketball image
img_url <- "https://upload.wikimedia.org/wikipedia/commons/7/7a/Basketball.png"
img <- base_loader(img_url)

# 2. Define normalization (ImageNet)
norm_mean <- c(0.485, 0.456, 0.406)
norm_std <- c(0.229, 0.224, 0.225)

# 3. Preprocess: convert to tensor, resize, Normalize
input <- img \%>\%
  transform_to_tensor() \%>\%
  transform_resize(c(400, 400)) \%>\%
  transform_normalize(norm_mean, norm_std)
batch <- input$unsqueeze(1)    # Add batch dimension (1, 3, H, W)

# 4. Display the image before normalization
tensor_image_browse(input)

# 5. Load MaxViT model
model <- model_maxvit(pretrained = TRUE)
model$eval()

# 6. Run inference
output <- model(batch)
topk <- output$topk(k = 5, dim = 2)
indices <- as.integer(topk[[2]][1, ])
scores <- as.numeric(topk[[1]][1, ])

# 7. Show Top-5 predictions
glue::glue("{seq_along(indices)}. {imagenet_classes(indices)} ({round(scores, 2)}\%)")
}

}
\seealso{
Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_facenet}},
\code{\link{model_inception_v3}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms-generics.R
\name{transform_random_resized_crop}
\alias{transform_random_resized_crop}
\title{Crop image to random size and aspect ratio}
\usage{
transform_random_resized_crop(
  img,
  size,
  scale = c(0.08, 1),
  ratio = c(3/4, 4/3),
  interpolation = 2
)
}
\arguments{
\item{img}{A \code{magick-image}, \code{array} or \code{torch_tensor}.}

\item{size}{(sequence or int): Desired output size. If size is a sequence
like c(h, w), output size will be matched to this. If size is an int,
smaller edge of the image will be matched to this number.
i.e, if height > width, then image will be rescaled to
(size * height / width, size).}

\item{scale}{(tuple of float): range of size of the origin size cropped}

\item{ratio}{(tuple of float): range of aspect ratio of the origin aspect
ratio cropped.}

\item{interpolation}{(int, optional) Desired interpolation. An integer
\code{0 = nearest}, \code{2 = bilinear}, and \code{3 = bicubic} or a name from
\code{\link[magick:filter_types]{magick::filter_types()}}.}
}
\description{
Crop the given image to a random size and aspect ratio. The image can be a
Magick Image or a Tensor, in which case it is expected to have
\verb{[..., H, W]} shape, where ... means an arbitrary number of leading
dimensions
}
\details{
A crop of random size (default: of 0.08 to 1.0) of the original size
and a random aspect ratio (default: of 3/4 to 4/3) of the original aspect
ratio is made. This crop is finally resized to given size. This is
popularly used to train the Inception networks.
}
\seealso{
Other random_transforms: 
\code{\link{transform_color_jitter}()},
\code{\link{transform_random_affine}()},
\code{\link{transform_random_crop}()},
\code{\link{transform_random_erasing}()},
\code{\link{transform_random_grayscale}()},
\code{\link{transform_random_horizontal_flip}()},
\code{\link{transform_random_perspective}()},
\code{\link{transform_random_rotation}()},
\code{\link{transform_random_vertical_flip}()}
}
\concept{random_transforms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-boxes.R
\name{remove_small_boxes}
\alias{remove_small_boxes}
\title{Remove Small Boxes}
\usage{
remove_small_boxes(boxes, min_size)
}
\arguments{
\item{boxes}{(Tensor[N, 4]): boxes in \eqn{(x_{min}, y_{min}, x_{max}, y_{max})} format
with
\itemize{
\item \eqn{0 \leq x_{min} < x_{max}} and
\item \eqn{0 \leq y_{min} < y_{max}}.
}}

\item{min_size}{(float): minimum size}
}
\value{
keep (Tensor[K]): indices of the boxes that have both sides
larger than min_size
}
\description{
Remove boxes which contains at least one side smaller than min_size.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models-facenet.R
\name{model_facenet}
\alias{model_facenet}
\alias{model_facenet_pnet}
\alias{model_facenet_rnet}
\alias{model_facenet_onet}
\alias{model_mtcnn}
\alias{model_facenet_inception_resnet_v1}
\title{MTCNN Face Detection Networks}
\usage{
model_facenet_pnet(pretrained = TRUE, progress = FALSE, ...)

model_facenet_rnet(pretrained = TRUE, progress = FALSE, ...)

model_facenet_onet(pretrained = TRUE, progress = FALSE, ...)

model_mtcnn(pretrained = TRUE, progress = TRUE, ...)

model_facenet_inception_resnet_v1(
  pretrained = NULL,
  classify = FALSE,
  num_classes = 10,
  dropout_prob = 0.6,
  ...
)
}
\arguments{
\item{pretrained}{(bool): If TRUE, returns a model pre-trained on ImageNet.}

\item{progress}{(bool): If TRUE, displays a progress bar of the download to
stderr.}

\item{...}{Other parameters passed to the model implementation.}

\item{classify}{Logical, whether to include the classification head. Default is FALSE.}

\item{num_classes}{Integer, number of output classes for classification. Default is 10.}

\item{dropout_prob}{Numeric, dropout probability applied before classification. Default is 0.6.}
}
\value{
\code{model_mtcnn()} returns a named list with three elements:
\itemize{
\item \code{boxes}: A tensor of shape \code{(N, 4)} with bounding box coordinates \code{[x1, y1, x2, y2]}.
\item \code{landmarks}: A tensor of shape \code{(N, 10)} with (x, y) coordinates of 5 facial landmarks:
left eye, right eye, nose, left mouth corner, right mouth corner.
\item \code{cls}: A tensor of shape \code{(N, 2)} with face classification probabilities
(face / non-face). The \code{cls} head has two classes:
\itemize{
\item \code{1}: Non-face probability (background)
\item \code{2}: Face probability — use this value for thresholding detections
}
}
(Here, \code{N} is the number of detected faces in the input image.)

\code{model_facenet_inception_resnet_v1()} returns a tensor output depending on the \code{classify} argument:
\itemize{
\item When \code{classify = FALSE} (default):
A tensor of shape \code{(N, 512)}, where each row is a normalized embedding
vector (L2 norm = 1).
These 512-dimensional FaceNet embeddings can be compared using cosine
similarity or Euclidean distance for face verification and clustering.

\item When \code{classify = TRUE}:
A tensor of shape \code{(N, num_classes)} containing class logits.
}
}
\description{
These models implement the three-stage Multi-task Cascaded Convolutional Networks (MTCNN)
architecture from the paper
\href{https://arxiv.org/abs/1604.02878}{Joint Face Detection and Alignment using Multi-task Cascaded Convolutional Networks}.
}
\details{
MTCNN detects faces and facial landmarks in an image through a coarse-to-fine pipeline:
\itemize{
\item \strong{PNet} (Proposal Network): Generates candidate face bounding boxes at multiple scales.
\item \strong{RNet} (Refine Network): Refines candidate boxes, rejecting false positives.
\item \strong{ONet} (Output Network): Produces final bounding boxes and 5-point facial landmarks.
}
\subsection{Model Variants}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{| Model | Input Size     | Parameters | File Size | Outputs                       | Notes                             |
|-------|----------------|------------|-----------|-------------------------------|-----------------------------------|
| PNet  | ~12×12+        | ~3k        | 30 kB     | 2-class face prob + bbox reg  | Fully conv, sliding window stage  |
| RNet  | 24×24          | ~30k       | 400 kB    | 2-class face prob + bbox reg  | Dense layers, higher recall       |
| ONet  | 48×48          | ~100k      | 2 MB      | 2-class prob + bbox + 5-point | Landmark detection stage          |
}\if{html}{\out{</div>}}

Inception-ResNet-v1 is a convolutional neural network architecture combining Inception modules
with residual connections, designed for face recognition tasks. The model achieves high accuracy
on standard face verification benchmarks such as LFW (Labeled Faces in the Wild).
}

\subsection{Model Variants and Performance (LFW accuracy)}{

\if{html}{\out{<div class="sourceCode">}}\preformatted{|    Weights     | LFW Accuracy | File Size |
|----------------|--------------|-----------|
| CASIA-Webface  | 99.05\%       | 111 MB    |
| VGGFace2       | 99.65\%       | 107 MB    |
}\if{html}{\out{</div>}}
\itemize{
\item The CASIA-Webface pretrained weights provide strong baseline accuracy.
\item The VGGFace2 pretrained weights achieve higher accuracy, benefiting from a larger, more diverse dataset.
}
}
}
\section{Functions}{
\itemize{
\item \code{model_facenet_pnet()}: PNet (Proposal Network) — small fully-convolutional network for candidate face box generation.

\item \code{model_facenet_rnet()}: RNet (Refine Network) — medium CNN with dense layers for refining and rejecting false positives.

\item \code{model_facenet_onet()}: ONet (Output Network) — deeper CNN that outputs final bounding boxes and 5 facial landmark points.

\item \code{model_mtcnn()}: MTCNN (Multi-task Cascaded Convolutional Networks) — face detection and alignment using a cascade of three neural networks

\item \code{model_facenet_inception_resnet_v1()}: Inception-ResNet-v1 — high-accuracy face recognition model combining Inception modules with residual connections, pretrained on VGGFace2 and CASIA-Webface datasets

}}
\examples{
\dontrun{
# Example usage of PNet
model_pnet <- model_facenet_pnet(pretrained = TRUE)
model_pnet$eval()
input_pnet <- torch_randn(1, 3, 224, 224)
output_pnet <- model_pnet(input_pnet)
output_pnet

# Example usage of RNet
model_rnet <- model_facenet_rnet(pretrained = TRUE)
model_rnet$eval()
input_rnet <- torch_randn(1, 3, 24, 24)
output_rnet <- model_rnet(input_rnet)
output_rnet

# Example usage of ONet
model_onet <- model_facenet_onet(pretrained = TRUE)
model_onet$eval()
input_onet <- torch_randn(1, 3, 48, 48)
output_onet <- model_onet(input_onet)
output_onet

# Example usage of MTCNN
mtcnn <- model_mtcnn(pretrained = TRUE)
mtcnn$eval()
image_tensor <- torch_randn(c(1, 3, 224, 224))
out <- mtcnn(image_tensor)
out

# Load an image from the web
url <- paste0("https://upload.wikimedia.org/wikipedia/commons",
              "/b/b4/Catherine_Bell_200101233d_hr_\%28cropped\%29.jpg")
tmp_file <- tempfile(fileext = ".jpg")
download.file(url, tmp_file, mode = "wb")
img <- jpeg::readJPEG(tmp_file)

# Convert to torch tensor [C, H, W] normalized
input <- transform_to_tensor(img)  # [C, H, W]
batch <- input$unsqueeze(1)   # [1, C, H, W]

# Load pretrained model
model <- model_facenet_inception_resnet_v1(pretrained = "vggface2")
model$eval()
output <- model(batch)
output

# Example usage of Inception-ResNet-v1 with CASIA-Webface Weights
model <- model_facenet_inception_resnet_v1(pretrained = "casia-webface")
model$eval()
output <- model(batch)
output
}

}
\seealso{
Other object_detection_model: 
\code{\link{model_convnext_detection}},
\code{\link{model_fasterrcnn}},
\code{\link{model_maskrcnn}}

Other object_detection_model: 
\code{\link{model_convnext_detection}},
\code{\link{model_fasterrcnn}},
\code{\link{model_maskrcnn}}

Other classification_model: 
\code{\link{model_alexnet}()},
\code{\link{model_convnext}},
\code{\link{model_efficientnet}},
\code{\link{model_efficientnet_v2}},
\code{\link{model_inception_v3}()},
\code{\link{model_maxvit}()},
\code{\link{model_mobilenet_v2}()},
\code{\link{model_mobilenet_v3}},
\code{\link{model_resnet}},
\code{\link{model_vgg}},
\code{\link{model_vit}}
}
\concept{classification_model}
\concept{object_detection_model}
